/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#ifndef ACTIVITYGROUP
#define ACTIVITYGROUP

#include <Inventor/SbBasic.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoGroup.h>

class InternalVariable;
class InternalVariableUpdateAction;
class BehaviorUpdateAction;
class BaseActivity;

class ActivityGroup : public SoGroup {
    SO_NODE_HEADER(ActivityGroup);
    
    public:
    enum ShowState {
	SHOWALWAYS, 
	SHOWNEVER, 
	SHOWONSWITCH
    };
    SoSFFloat errTolerance;
    SoSFEnum showState;
    
    int activeActivityIndex;
    BaseActivity *activeActivity;
    
    static void initClass();
    void turnOffChildren();
     
    virtual void doAction(SoAction *act);
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    void printState();
    
    /* this has children print into each entry of buffs and returns # of lines */ 
    int printState(char **bufs);
    
    ActivityGroup();       // Note only should be call during a read
    virtual ~ActivityGroup();
    
    private:
    int switchedThisTick;
    float findAndSetActiveChild();
    float getDeltaAndResetValueForAll();
    float jiggle();
    void iterateToEquilibrium(int count, SoAction *action);

};


#endif  







