/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/
#ifndef BASEACTIVITY
#define BASEACTIVITY

#include <Inventor/SbBasic.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoGroup.h>

#include "Pronome.h"
#include "BaseBehaviorNode.h"

class InternalVariable;
class InternalVariableUpdateAction;
class BaseActivity;
class ActivityGroup;

class BaseActivity : public BaseBehaviorNode {
    SO_NODE_HEADER(BaseActivity);
    public:
    SoMFFloat inhibitoryGains;
    SoSFBool active;
    SoSFBool leaf;
    SoSFBool ignoreSuggestions;
    SoSFString leafCBName;
    SoSFString suggestionCBName;
    SoSFString leafNotifyCBName;
    float previousValue;
    
//    Pronome objectOfInterest; moved to BaseBehaviorNode
    
    activityCB leafCallback;
    
    activityCB suggestionCallback;
    
    activityCB leafNotifyCallback;
    
    static void initClass();
    
    float calcInhibitionOn(int i);
    float sumInhibition(SoNodeList &otherActivities);
    void  updateValueAfterInhibition(SoNodeList &otherActivities);
    float getDeltaInValueAndReset(SbBool shouldReset = TRUE);
    float getValueBeforeInhibition();
    void  setActive(SbBool isOn);
    void setLeafCallback(activityCB callback);
    void setSuggestionCallback(activityCB callback);
    ActivityGroup *getActivityGroup() {
	return (leaf.getValue())?NULL : (ActivityGroup *)getChild(activityGroupIndex);}
    
    virtual void doAction(SoAction *act);
    virtual void doLeafAction(SoAction *action);
    virtual void doRecommendation(SoAction *a);
    virtual void printState(char *buf= NULL);
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    
    void removePronome(SoAction *action);
    void installPronome(SoAction *action);
    void doLearning(SoAction *action);
    void addADependsOnVariable(SoNode *aDov);
    InternalVariable *getMaxUnderlyingMotivation();
    
    BaseActivity();       // Note only should be call during a read
    virtual ~BaseActivity();
    
    private:
    float inhibition;
    int indexIntoChildList; // where this guy lives in parent's childList
    int dovIndex;	    // where DependsOnVariables lives in childList
    int activityGroupIndex; // where ActivityGroup lives in childList
    int reinforcementGroupIndex; // where ReinforcementGroup lives in childList
    

};


#endif  
