/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#include <math.h>
#include <Inventor/SoDB.h>
#include <Inventor/misc/SoChildList.h>
#include "InternalVariable.h"
#include "BehaviorAction.h"
#include "ReleasingMechanism.h"
#include "BaseBehaviorNode.h"

#define MAXVALUE 50.0
#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a

/*****************************************************************
 * 
 *		    BaseBehaviorNode
 * 
 ****************************************************************/
SbPList BaseBehaviorNode::callBackList;

SO_NODE_SOURCE(BaseBehaviorNode);

BaseBehaviorNode::BaseBehaviorNode() : SoGroup()
{
    SO_NODE_CONSTRUCTOR(BaseBehaviorNode);
    SO_NODE_ADD_FIELD(pname, ("BaseBehaviorNode"));
    SO_NODE_ADD_FIELD(value, (0));
    SO_NODE_ADD_FIELD(printDebug, (FALSE));
    enableNotify(FALSE);
    data = NULL;
}

void
BaseBehaviorNode::initClass()
{
    SO_NODE_INIT_CLASS(BaseBehaviorNode, SoGroup, "Group");
}

BaseBehaviorNode *BaseBehaviorNode::readFromPrototypeFile(const char *protoFileName)
{
    SoInput mySceneInput;
    SoNode *fn;
    if (!mySceneInput.openFile(protoFileName)) {
       printf("Cannot open file\n");
       return NULL;
    }
    SoDB::read(&mySceneInput, fn);
    if (!(fn->isOfType(BaseBehaviorNode::getClassTypeId()))) {
	printf("%s does not contain a valid prototype \n",protoFileName );
	return NULL;
    }
    else {
	return (BaseBehaviorNode *)fn;
    }    
}

BaseBehaviorNode::~BaseBehaviorNode()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
BaseBehaviorNode::readInstance(SoInput *in, unsigned short flags)
#else
BaseBehaviorNode::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoGroup::readInstance(in, flags)) {
#else
    if (SoGroup::readInstance(in)) {
#endif
	ref();
	return (1);
    }
    else {
	return (0);
    }
}

void BaseBehaviorNode::doAction(SoAction *action)
{
    int numIndices;
    int lastChildIndex;
    const int *indices;
    int i;
    float v;
    
    if(action->getPathCode(numIndices, indices) == 
	SoAction::IN_PATH)
	lastChildIndex = indices[numIndices -1];
    else
	lastChildIndex = getNumChildren();
    if(action->isOfType(InternalVariableUpdateAction::getClassTypeId())){
	v = value.getValue();
	for(i=0;i<lastChildIndex;i++){
	    InternalVariableUpdateAction *ba = (InternalVariableUpdateAction *)action;
	    ba->value = v;
//	    printf("doAction of BaseBehavior \n");
	    children->traverse(action, i);
	    ba->value = v;
	}   
    }
    else {
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);    
    }
}


rmCBData *BaseBehaviorNode::findCallbackByName(SbString &fname)
{
    int i;
    for(i=0;i<BaseBehaviorNode::callBackList.getLength();i++)
	if(((rmCBData *)BaseBehaviorNode::callBackList[i])->fname == fname)
	    return ((rmCBData *)BaseBehaviorNode::callBackList[i]);
    printf("WARNING CALLBACK %s NOT FOUND \n", fname.getString());
    return NULL;
}
