/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#ifndef BEHAVIORNODE
#define BEHAVIORNODE

#include <Inventor/SbBasic.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoGroup.h>
#include "Pronome.h"

class InternalVariable;
class InternalVariableUpdateAction;
class BaseActivity;
class ReleasingMechanism;
class BehaviorUpdateAction;

/********************************************************************
 *
 *	Callback mechanism for ReleasingMechanisms & Activities
 * 
 ********************************************************************/
typedef SbBool (*rmFilterCB)(ReleasingMechanism *rm, BehaviorUpdateAction *action);
typedef float (*rmWeightCB)(ReleasingMechanism *rm, BehaviorUpdateAction *action);
typedef void (*activityCB)(BaseActivity *act, BehaviorUpdateAction *action);
typedef void (*genericCB)(void *genericArg);

class rmCBData {
    public:
    SbString fname;
    rmCBData(const char *name, rmFilterCB func){fname = SbString(name);
	filterCB = func;}
    rmCBData(const char *name, rmWeightCB func){fname = SbString(name); 
	weightCB = func;}
    rmCBData(const char *name, activityCB func){fname = SbString(name); 
	actCB = func;}
    rmCBData(const char *name, genericCB func){fname = SbString(name); 
	genCB = func;}
    rmFilterCB  &getFilterCB() {return filterCB;}
    rmWeightCB  &getWeightCB() {return weightCB;}	
    activityCB &getActCB() {return actCB;}
    genericCB &getGenCB() {return genCB;}
    private:
    union {
	rmFilterCB filterCB;
	rmWeightCB weightCB;
	activityCB actCB;
	genericCB genCB;
    };
};

/********************************************************************
 *
 *	Superclass for ReleasingMechanisms & Activities
 * 
 ********************************************************************/

class BaseBehaviorNode : public SoGroup {
    SO_NODE_HEADER(BaseBehaviorNode);
    public:
    static SbPList callBackList;
    
    SoSFFloat value;
    SoSFString pname;
    Pronome objectOfInterest;
    SoSFBool printDebug;
    float v_tmp;
    void *data;
    static void initClass();
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual void doAction(SoAction *act);
    static BaseBehaviorNode *readFromPrototypeFile(const char *protoFileName);

    BaseBehaviorNode();       // Note only should be call during a read
    virtual ~BaseBehaviorNode();
    
    // Callback stuff
    public:
    static void addCallback(const char *name, rmFilterCB func){
	BaseBehaviorNode::callBackList.append(new rmCBData(name, func));
    }

    static void addCallback(const char *name, rmWeightCB func){
	BaseBehaviorNode::callBackList.append(new rmCBData(name, func));
    }

    static void addCallback(const char *name, activityCB func){
	BaseBehaviorNode::callBackList.append(new rmCBData(name, func));
    }
    
    static void addCallback(const char *name, genericCB func){
	BaseBehaviorNode::callBackList.append(new rmCBData(name, func));
    }
    
    rmFilterCB getFilterCBByName(SbString fname){
	rmCBData *cb;
	return (cb = BaseBehaviorNode::findCallbackByName(fname))
	    ?cb->getFilterCB():NULL;
    }

    rmWeightCB getWeightCBByName(SbString fname){
	rmCBData *cb;
	return (cb = BaseBehaviorNode::findCallbackByName(fname))
	    ?cb->getWeightCB():NULL;
    }
    
    activityCB getActCBByName(SbString fname){
	rmCBData *cb;
	return (cb = BaseBehaviorNode::findCallbackByName(fname))
	    ?cb->getActCB():NULL;
    }
    genericCB getGenericCBByName(SbString fname){
	rmCBData *cb;
	return (cb = BaseBehaviorNode::findCallbackByName(fname))
	    ?cb->getGenCB():NULL;
    }
    protected:
    static rmCBData *findCallbackByName(SbString &fname);

};

#endif  
