/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/


#include <Inventor/elements/SoElements.h>
#include <Inventor/elements/SoSwitchElement.h>
#include <Inventor/elements/SoUnitsElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/nodes/SoGroup.h>
#include "BehaviorAction.h"
#include "InternalVariable.h"
#include "ReleasingMechanism.h"

SO_ACTION_SOURCE(InternalVariableUpdateAction);

//
// Initializes the InternalVariableUpdateAction class. This is a one-time
// thing that is done after database initialization and before
// any instance of this class is constructed.
//

void
InternalVariableUpdateAction::initClass()
{
    // Initialize the run-time type variables
    SO_ACTION_INIT_CLASS(InternalVariableUpdateAction, SoAction);

    SO_ENABLE(InternalVariableUpdateAction, SoSwitchElement);
    SO_ACTION_ADD_METHOD(SoNode, nullAction);
    SO_ACTION_ADD_METHOD(InternalVariable, callDoAction);
    SO_ACTION_ADD_METHOD(ModifyVariable, callDoAction);
    SO_ACTION_ADD_METHOD(SoGroup, callDoAction);
}

//
// Constructor
//

InternalVariableUpdateAction::InternalVariableUpdateAction()
{
   SO_ACTION_CONSTRUCTOR(InternalVariableUpdateAction);
   tick = 0;
}

//
// Destructor. Does nothing.
//

InternalVariableUpdateAction::~InternalVariableUpdateAction()
{
}

//
// Initiates action on a graph. This is called when the action
// is applied to a node, a path, or a path list. It gives us a
// chance to initialize things before beginning traversal.
//

void
InternalVariableUpdateAction::beginTraversal(SoNode *node)
{
   tick++;
   // Begin traversal at the given root node.
   traverse(node);
}


//
// This method implements the action for an SoSphere node.
//

void
InternalVariableUpdateAction::callDoAction(SoAction *action, SoNode *node)
{
   node->doAction(action);
}

SO_ACTION_SOURCE(BehaviorUpdateAction);

//
// Initializes the BehaviorUpdateAction class. This is a one-time
// thing that is done after database initialization and before
// any instance of this class is constructed.
//

void
BehaviorUpdateAction::initClass()
{
    // Initialize the run-time type variables
    SO_ACTION_INIT_CLASS(BehaviorUpdateAction, SoAction);

    SO_ENABLE(BehaviorUpdateAction, SoSwitchElement);
    SO_ACTION_ADD_METHOD(SoNode, nullAction);
    SO_ACTION_ADD_METHOD(DependsOnVariables, callDoAction);
    SO_ACTION_ADD_METHOD(ReinforcementGroup, callDoAction);
    SO_ACTION_ADD_METHOD(SoGroup, callDoAction);
}

//
// Constructor
//

BehaviorUpdateAction::BehaviorUpdateAction()
{
   SO_ACTION_CONSTRUCTOR(BehaviorUpdateAction);
   tick = 0;
}

//
// Destructor. Does nothing.
//

BehaviorUpdateAction::~BehaviorUpdateAction()
{
}

//
// Initiates action on a graph. This is called when the action
// is applied to a node, a path, or a path list. It gives us a
// chance to initialize things before beginning traversal.
//

void
BehaviorUpdateAction::beginTraversal(SoNode *node)
{
   // Begin traversal at the given root node.
   tick++;
   // empty pronomes
   pronomes.empty();
   traverse(node);
}


//
// This method implements the action for an SoSphere node.
//

void
BehaviorUpdateAction::callDoAction(SoAction *action, SoNode *node)
{
   node->doAction(action);
}

SO_ACTION_SOURCE(BehaviorResetAction);

//
// Initializes the class. This is a one-time
// thing that is done after database initialization and before
// any instance of this class is constructed.
//

void
BehaviorResetAction::initClass()
{
    // Initialize the run-time type variables
    SO_ACTION_INIT_CLASS(BehaviorResetAction, SoAction);
    SO_ENABLE(BehaviorResetAction, SoSwitchElement);
    SO_ACTION_ADD_METHOD(SoNode, callDoAction);
}

//
// Constructor
//

BehaviorResetAction::BehaviorResetAction()
{
   SO_ACTION_CONSTRUCTOR(BehaviorResetAction);
}

//
// Destructor. Does nothing.
//

BehaviorResetAction::~BehaviorResetAction()
{
}

//
// Initiates action on a graph. This is called when the action
// is applied to a node, a path, or a path list. It gives us a
// chance to initialize things before beginning traversal.
//

void
BehaviorResetAction::beginTraversal(SoNode *node)
{
   printf("reseting all internal variables \n");
   SoAction::beginTraversal(node);
}


//
// This method implements the action for an SoSphere node.
//

void
BehaviorResetAction::callDoAction(SoAction *action, SoNode *node)
{
   node->doAction(action);
}
