/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#ifndef _BEHAVIORACTION_
#define _BEHAVIORACTION_

#include <Inventor/actions/SoSubAction.h>
#include "Pronome.h"
class InternalVariable;
class InternalVariableUpdateAction : public SoAction {

   SO_ACTION_HEADER(InternalVariableUpdateAction);

    public:
    float value;
    int tick;
   
    // Initializes this action class for use with scene graphs
    static void    initClass();

    // Constructor and destructor
    InternalVariableUpdateAction();
    virtual ~InternalVariableUpdateAction();

 protected:
    // Initiates action on graph
    virtual void   beginTraversal(SoNode *node);

 private:
    static void    callDoAction(SoAction *, SoNode *);

};

class BehaviorUpdateAction : public SoAction {

   SO_ACTION_HEADER(BehaviorUpdateAction);

    public:
    enum Passes{
	INITIALPASS, 
	ITERATIONPASS, 
	RECOMMENDERPASS, 
	DISPATCHPASS
    };
    float value;
    float tick;
    int whichPass;
    PronomeList pronomes;
    SoNodeList *currentActivityGroup;
    InternalVariable *maxUnderlyingMotivation; // used for reinforcement & forgetting
   
    // Initializes this action class for use with scene graphs
    static void    initClass();

    // Constructor and destructor
    BehaviorUpdateAction();
    virtual ~BehaviorUpdateAction();

 protected:
    // Initiates action on graph
    virtual void   beginTraversal(SoNode *node);

 private:
    static void    callDoAction(SoAction *, SoNode *);

};

class BehaviorResetAction : public SoAction {

   SO_ACTION_HEADER(BehaviorResetAction);

    public:
    // Initializes this action class for use with scene graphs
    static void    initClass();

    // Constructor and destructor
    BehaviorResetAction();
    virtual ~BehaviorResetAction();

 protected:
    // Initiates action on graph
    virtual void   beginTraversal(SoNode *node);

 private:
    static void    callDoAction(SoAction *, SoNode *);

};

#endif
