/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#include <math.h>
#include <Inventor/SoDB.h>
#include <Inventor/misc/SoChildList.h>
#include "InternalVariable.h"
#include "BehaviorAction.h"
#include "ReleasingMechanism.h"
#include "DiscoveryGroup.h"

#define MAXVALUE 50.0
#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a

/*****************************************************************
 * 
 *		    DiscoveryGroup
 * 
 ****************************************************************/
SO_NODE_SOURCE(DiscoveryGroup);

DiscoveryGroup::DiscoveryGroup() : BaseBehaviorNode()
{
    SO_NODE_CONSTRUCTOR(DiscoveryGroup);
    watchSize = 0;
}

void
DiscoveryGroup::initClass()
{
    SO_NODE_INIT_CLASS(DiscoveryGroup, BaseBehaviorNode, "BaseBehaviorNode");
}

DiscoveryGroup::~DiscoveryGroup()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
DiscoveryGroup::readInstance(SoInput *in, unsigned short flags)
#else
DiscoveryGroup::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (BaseBehaviorNode::readInstance(in, flags)) {
#else
    if (BaseBehaviorNode::readInstance(in)) {
#endif
	ref();
	return (1);
    }
    else {
	return (0);
    }
}

void DiscoveryGroup::addChildToWatch(BaseBehaviorNode *bbn)
{
    if(watchSize < MAXWATCHSIZE){
	addChild(bbn);
	watchSize++;
    }
}

void DiscoveryGroup::updateIntervalData()
{
    static k = 0;
    int i, j;
    BaseBehaviorNode *childI;
    BaseBehaviorNode *childJ;
    for(i=0;i<watchSize;i++){
	childI = (BaseBehaviorNode *)getChild(i);
	if(childI->value.getValue() > 1.0e-4){
	    intervalData[i][i]++;
	    for(j=i+i;j<watchSize;j++){
		childJ = (BaseBehaviorNode *)getChild(i);
		if(childJ->value.getValue() > 1.0e-4)
		    intervalData[i][j]++;
	    }
	}
    }
    k++;
    if(k % 20)
	printIntervalData();
}

void DiscoveryGroup::printIntervalData()
{
    int i, j;
    for(i=0;i<watchSize;i++)
	for(j=0;j<watchSize;j++)
	    printf(" %d ", intervalData[i][j]);
	printf("\n");
}

void DiscoveryGroup::doAction(SoAction *action)
{
    int numIndices;
    int lastChildIndex;
    const int *indices;
    int i;
    float v;
    
    if(action->getPathCode(numIndices, indices) == 
	SoAction::IN_PATH)
	lastChildIndex = indices[numIndices -1];
    else
	lastChildIndex = getNumChildren();
    if(action->isOfType(BehaviorUpdateAction::getClassTypeId())){
	for(i=0;i<lastChildIndex;i++){
	    children->traverse(action, i);
	}
	updateIntervalData();   
    }
    else {
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);    
    }
}

