/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or  modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#include <math.h>
#include <Inventor/misc/SoChildList.h>
#include "InternalVariable.h"
#include "BehaviorAction.h"
#include "ReleasingMechanism.h"
#include "BaseBehaviorNode.h"

#define MAXVALUE 50.0
#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a
#define NOTAVALUE -1.0e6
/*****************************************************************
 * 
 *		    DependsOnVariables
 * 
 ****************************************************************/

SO_NODE_SOURCE(DependsOnVariables);

DependsOnVariables::DependsOnVariables() : SoGroup()
{
    SO_NODE_CONSTRUCTOR(DependsOnVariables);
    SO_NODE_ADD_FIELD(combineRule, (DependsOnVariables::COMB_ADD));
    SO_NODE_ADD_FIELD(combineRMandIVRule, (DependsOnVariables::COMB_ADD));
    SO_NODE_DEFINE_ENUM_VALUE(CombineRuleType, COMB_ADD);
    SO_NODE_DEFINE_ENUM_VALUE(CombineRuleType, COMB_MULT);
    SO_NODE_DEFINE_ENUM_VALUE(CombineRuleType, COMB_RMTRUE);
    SO_NODE_DEFINE_ENUM_VALUE(CombineRuleType, COMB_RMMAX);
    SO_NODE_SET_SF_ENUM_TYPE(combineRule, CombineRuleType);
    SO_NODE_SET_SF_ENUM_TYPE(combineRMandIVRule, CombineRuleType);
    enableNotify(FALSE);
}

void
DependsOnVariables::initClass()
{
    SO_NODE_INIT_CLASS(DependsOnVariables, SoGroup, "Group");
}


DependsOnVariables::~DependsOnVariables()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
DependsOnVariables::readInstance(SoInput *in, unsigned short flags)
#else
DependsOnVariables::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoGroup::readInstance(in, flags)) {
#else
    if (SoGroup::readInstance(in)) {
#endif
	return (1);
    }
    else {
	return (0);
    }
}

ReleasingMechanism *DependsOnVariables::getRMwithMaxValue()
{
    int i;
    float maxRM = -1;
    ReleasingMechanism *rmMax = NULL;
    SoNode *child;
    int lastChildIndex = getNumChildren();
    for(i=0;i<lastChildIndex;i++){
	child = getChild(i);
	if(child->isOfType(ReleasingMechanism::getClassTypeId())){
	    if(((ReleasingMechanism *)child)->value.getValue()> maxRM){
		maxRM = ((ReleasingMechanism *)child)->value.getValue();
		rmMax = (ReleasingMechanism *)child;
	    }
	}
    }
    return rmMax;
}

void combVar(float &result, float newValue, int rule)
{
    if(result ==NOTAVALUE)
	result = newValue;
    else {
	if(rule == DependsOnVariables::COMB_ADD)
	    result += newValue;
	else if(rule == DependsOnVariables::COMB_RMMAX)
	    result = (newValue> result)?newValue : result;
	else
	    result *= newValue;
    }
}

void DependsOnVariables::combRMandIV()
{
    if(releasingMechanismSum == NOTAVALUE)
	releasingMechanismSum = 0;
    if(internalVariablesSum == NOTAVALUE)
	internalVariablesSum = 0;
    /*
     * This rule says that if the combine rule is COMB_RMTRUE,  then
     * rm has to be non-zero. 
     */
    if((combineRMandIVRule.getValue() == DependsOnVariables::COMB_RMTRUE) &&
	(fabs(releasingMechanismSum)<1.0e-4)){
	    value = 0;
    }
    else {
	value = (combineRMandIVRule.getValue() == DependsOnVariables::COMB_MULT)
	    ? releasingMechanismSum * internalVariablesSum
	    : releasingMechanismSum + internalVariablesSum;
    }
}

void DependsOnVariables::doAction(SoAction *action)
{
    int numIndices;
    int lastChildIndex;
    const int *indices;
    int i;
    float v;
    SoNode *child;
        
    if(action->getPathCode(numIndices, indices) == 
	SoAction::IN_PATH)
	lastChildIndex = indices[numIndices -1];
    else
	lastChildIndex = getNumChildren();
    if(action->isOfType(BehaviorUpdateAction::getClassTypeId())){
	releasingMechanismSum = internalVariablesSum = NOTAVALUE;
	levelOfInterest = 1;
	for(i=0;i<lastChildIndex;i++){
	    child = getChild(i);
	    if(child->isOfType(LevelOfInterest::getClassTypeId()))
		levelOfInterest = ((LevelOfInterest *)child)->value.getValue();
	    else if(child->isOfType(ReleasingMechanism::getClassTypeId())){
		children->traverse(action, i);
		combVar(releasingMechanismSum, ((ReleasingMechanism *)child)->value.getValue(), 
		    combineRule.getValue());
		// releasingMechanismSum += ((ReleasingMechanism *)child)->value.getValue();
	    }
	    else if(child->isOfType(InternalVariable::getClassTypeId()))
		combVar(internalVariablesSum, ((InternalVariable *)child)->value.getValue(), 
		    combineRule.getValue());
		//internalVariablesSum += ((InternalVariable *)child)->value.getValue();
	}
	combRMandIV();
/*	
	if(releasingMechanismSum == NOTAVALUE)
	    releasingMechanismSum = 0;
	if(internalVariablesSum == NOTAVALUE)
	    internalVariablesSum = 0;
*/
//	printf("%f %f %f \n",levelOfInterest, internalVariablesSum, releasingMechanismSum);    
    }
    else {
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);    
    }
}
/*****************************************************************
 * 
 *		    ModifyVariable
 * 
 ****************************************************************/

SO_NODE_SOURCE(ModifyVariable);

ModifyVariable::ModifyVariable() : SoNode()
{
    SO_NODE_CONSTRUCTOR(ModifyVariable);
    SO_NODE_ADD_FIELD(gain, (0));
    SO_NODE_ADD_FIELD(valueDependentEffect, (TRUE));
    SO_NODE_ADD_FIELD(who, (0));
    enableNotify(FALSE);
}

void
ModifyVariable::initClass()
{
    SO_NODE_INIT_CLASS(ModifyVariable, SoNode, "Node");
}


ModifyVariable::~ModifyVariable()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
ModifyVariable::readInstance(SoInput *in, unsigned short flags)
#else
ModifyVariable::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoNode::readInstance(in, flags)) {
#else
    if (SoNode::readInstance(in)) {
#endif
	return (1);
    }
    else {
	return (0);
    }
}

void ModifyVariable::doAction(SoAction *action)
{
    if(action->isOfType(InternalVariableUpdateAction::getClassTypeId())){
//	printf("doAction of ModifyVariable \n");
	who.getValue()->doAction(action);
	InternalVariableUpdateAction *ba = (InternalVariableUpdateAction *)action;
	float adjustValue = (valueDependentEffect.getValue())
	    ? ba->value *gain.getValue()
	    : (ba->value > 0.0)?gain.getValue():0.0;
	((InternalVariable *)who.getValue())->adjustValueBy(adjustValue);
    }
    else SoNode::doAction(action);
}

/*****************************************************************
 * 
 *		    Internal Variable
 * 
 ****************************************************************/
SO_NODE_SOURCE(InternalVariable);

InternalVariable::InternalVariable() : SoNode()
{
    SO_NODE_CONSTRUCTOR(InternalVariable);
    SO_NODE_ADD_FIELD(value, (0));
    SO_NODE_ADD_FIELD(growth, (0));
    SO_NODE_ADD_FIELD(damping, (0));
    SO_NODE_ADD_FIELD(pname, ("a variable"));
    SO_NODE_ADD_FIELD(clampToPositive, (1));
    SO_NODE_ADD_FIELD(showState, (0));
    SO_NODE_ADD_FIELD(maxValue, (1000.0));
    enableNotify(FALSE); 
}

void
InternalVariable::initClass()
{
    SO_NODE_INIT_CLASS(InternalVariable, SoNode, "Node");
}

SbBool	
#ifdef SO_VERSION_REVISION
InternalVariable::readInstance(SoInput *in, unsigned short flags)
#else
InternalVariable::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoNode::readInstance(in, flags)) {
#else
    if (SoNode::readInstance(in)) {
#endif
	initialValue = value.getValue();
	return (1);
    }
    else {
	return (0);
    }
}

InternalVariable::~InternalVariable()
{
    lastUpdateTick = -1;
}

float InternalVariable::getInitialUpdateValue()
{
    return value.getValue();    
}

void InternalVariable::doAction(SoAction *action)
{
    if(action->isOfType(InternalVariableUpdateAction::getClassTypeId())){
	InternalVariableUpdateAction *ba = (InternalVariableUpdateAction *)action;
	if(ba->tick != lastUpdateTick){
//	    printf("doAction of InternalVariable \n");
	    lastUpdateTick = ba->tick;
	    float v = getInitialUpdateValue();
	    v += growth.getValue() - damping.getValue()*v;
	    if(clampToPositive.getValue())
		v = MAX(v, 0.0);
	    v = MIN(v, maxValue.getValue());
	    value.setValue(v);
	    if(showState.getValue())
		reportStateAt(lastUpdateTick);
	}
    }
    else if(action->isOfType(BehaviorResetAction::getClassTypeId())){
	if(value.getValue() != initialValue){
		value.setValue(initialValue);
		printf("reseting internal variable %s to %f \n", 
	    	pname.getValue().getString(),value.getValue());
	}
    }
    else SoNode::doAction(action);
    return;
}

float InternalVariable::adjustValueBy(float effect)
{
    float v = value.getValue() + effect;
    if(clampToPositive.getValue())
	v = MAX(v, 0.0);
    v = MIN(v, maxValue.getValue());
    value.setValue(v);
//  printf("inside adjustValue -> %s %f \n", pname.getValue().getString(),value.getValue());
    return v;
}

void InternalVariable::reportStateAt(int t)
{
  float f = value.getValue();
  printf("name =>  %s  value => %f \n", pname.getValue().getString(), f);
}

/*****************************************************************
 * 
 *		    LevelOfInterest
 * 
 ****************************************************************/

SO_NODE_SOURCE(LevelOfInterest);

LevelOfInterest::LevelOfInterest() : InternalVariable()
{
    SO_NODE_CONSTRUCTOR(LevelOfInterest);
}

void
LevelOfInterest::initClass()
{
    SO_NODE_INIT_CLASS(LevelOfInterest, InternalVariable, "InternalVariable");
}


LevelOfInterest::~LevelOfInterest()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
LevelOfInterest::readInstance(SoInput *in, unsigned short flags) 
#else
LevelOfInterest::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (InternalVariable::readInstance(in, flags)) {
#else
    if (InternalVariable::readInstance(in)) {
#endif
	value.setValue(1.0);
	if(damping.getValue() == 0.0)
	    damping.setValue(growth.getValue());
	return (1);
    }
    else {
	return (0);
    }
}

/*****************************************************************
 * 
 *		    Motivation
 * 
 ****************************************************************/

SO_NODE_SOURCE(Motivation);

Motivation::Motivation() : InternalVariable()
{
    SO_NODE_CONSTRUCTOR(Motivation);
}

void
Motivation::initClass()
{
    SO_NODE_INIT_CLASS(Motivation, InternalVariable, "InternalVariable");
}


Motivation::~Motivation()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
Motivation::readInstance(SoInput *in, unsigned short flags) 
#else
Motivation::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (InternalVariable::readInstance(in, flags)) {
#else
    if (InternalVariable::readInstance(in)) {
#endif
	printf("motivational variable -> %s \n", 
	   getName().getString());
	return (1);
    }
    else {
	return (0);
    }
}

/*****************************************************************
 * 
 *		    Reinforcement
 * 
 ****************************************************************/

SO_NODE_SOURCE(Reinforcement);

Reinforcement::Reinforcement() : InternalVariable()
{
    SO_NODE_CONSTRUCTOR(Reinforcement);
    SO_NODE_ADD_FIELD(underlyingMotivation, (0));
    
}

void
Reinforcement::initClass()
{
    SO_NODE_INIT_CLASS(Reinforcement, InternalVariable, "InternalVariable");
}


Reinforcement::~Reinforcement()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
Reinforcement::readInstance(SoInput *in, unsigned short flags) 
#else
Reinforcement::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (InternalVariable::readInstance(in, flags)) {
#else
    if (InternalVariable::readInstance(in)) {
#endif
	printf("reinforcement variable -> %s \n", 
	   getName().getString());
	return (1);
    }
    else {
	return (0);
    }
}

float Reinforcement::getInitialUpdateValue()
{
    return 0.0;    
}
  
/*****************************************************************
 * 
 *		    ReinforcementGroup
 * 
 ****************************************************************/

SO_NODE_SOURCE(ReinforcementGroup);

ReinforcementGroup::ReinforcementGroup() : SoGroup()
{
    SO_NODE_CONSTRUCTOR(ReinforcementGroup);
    enableNotify(FALSE);
}

void
ReinforcementGroup::initClass()
{
    SO_NODE_INIT_CLASS(ReinforcementGroup, SoGroup, "Group");
}


ReinforcementGroup::~ReinforcementGroup()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
ReinforcementGroup::readInstance(SoInput *in, unsigned short flags)
#else
ReinforcementGroup::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoGroup::readInstance(in, flags)) {
#else
    if (SoGroup::readInstance(in)) {
#endif
	return (1);
    }
    else {
	return (0);
    }
}

void ReinforcementGroup::doAction(SoAction *action)
{
    int numIndices;
    int lastChildIndex;
    const int *indices;
    int i;
    SoNode *child;
        
    if(action->getPathCode(numIndices, indices) == 
	SoAction::IN_PATH)
	lastChildIndex = indices[numIndices -1];
    else
	lastChildIndex = getNumChildren();
    if(action->isOfType(BehaviorUpdateAction::getClassTypeId())){
	rsum = 0.0;
	for(i=0;i<lastChildIndex;i++){
	    child = getChild(i);
	    if(child->isOfType(InternalVariable::getClassTypeId()))
		rsum +=  ((InternalVariable *)child)->value.getValue();
	}
//	printf("rsum -> %f \n", rsum);
    }
    else {
	for(i=0;i<lastChildIndex;i++)
	    children->traverse(action, i);    
    }
}

InternalVariable *ReinforcementGroup::getMaxUnderlyingMotivation()
{
    float tv, maxSoFar = -1.0e6;
    int maxIndex = -1;
    Reinforcement *rchild = NULL;
    int i, lastChildIndex = getNumChildren();
    for(i=0;i<lastChildIndex;i++){
	rchild = (Reinforcement *)getChild(i);
	if(rchild->underlyingMotivation.getValue() != NULL){
	    tv =  ((InternalVariable *)rchild->
		underlyingMotivation.getValue())->value.getValue();
	    if(tv > maxSoFar){
		maxSoFar = tv;
		maxIndex = i;
	    }  
	}
    }
    return (maxIndex> -1)?
	((InternalVariable *)
	    ((Reinforcement *)getChild(maxIndex))->
		underlyingMotivation.getValue()) : NULL;
}







