/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/
#ifndef ENDO
#define ENDO

#include <Inventor/SbBasic.h>
#include <Inventor/fields/SoMFFloat.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoGroup.h>
class InternalVariable;
class InternalVariableUpdateAction;
class ReleasingMechanism;
class Reinforcement;

class DependsOnVariables : public SoGroup {
    SO_NODE_HEADER(DependsOnVariables);
    public:
    float levelOfInterest;
    float releasingMechanismSum;
    float internalVariablesSum;
    float value;
    enum CombineRuleType{
	COMB_ADD, 
	COMB_MULT, 
	COMB_RMTRUE,    // if(rmSum and ivSum non-zero) add else 0
	COMB_RMMAX	// use maximum rm only for rm sum
    }; 
    SoSFEnum combineRule;   // default COMB_ADD
    SoSFEnum combineRMandIVRule; // default COMB_ADD
    ReleasingMechanism *getRMwithMaxValue();
    void combRMandIV();
    virtual void doAction(SoAction *action);
    static void initClass();
    
    DependsOnVariables();
    
    private:
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~DependsOnVariables();
};

class ReinforcementGroup : public SoGroup {
    SO_NODE_HEADER(ReinforcementGroup);
    public:
    float rsum;
    virtual void doAction(SoAction *action);
    static void initClass();
    ReinforcementGroup();
    InternalVariable *getMaxUnderlyingMotivation();
    private:
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~ReinforcementGroup();
};

class ModifyVariable : public SoNode {
    SO_NODE_HEADER(ModifyVariable);
    public:
    SoSFFloat gain;
    SoSFNode who;
    SoSFBool valueDependentEffect;
    static void initClass();
    private:
    ModifyVariable();
    virtual void doAction(SoAction *action);
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~ModifyVariable();
};

class InternalVariable: public SoNode { 
    SO_NODE_HEADER(InternalVariable);
    public:
    SoSFFloat value;
    SoSFString pname;
    SoSFFloat growth;
    SoSFFloat damping;
    SoSFBool clampToPositive;
    SoSFBool showState;
    SoSFFloat maxValue;
    
    static void initClass();
    float adjustValueBy(float effect);
    virtual void doAction(SoAction *action);
    virtual float getInitialUpdateValue();
    void reportStateAt(int t);
#ifdef SO_VERSION_REVISION
    SbBool readInstance(SoInput *in, unsigned short flags);
#else
    SbBool readInstance(SoInput *in);
#endif
    
    protected:
    int lastUpdateTick;
    float initialValue;
    InternalVariable();
    virtual ~InternalVariable();
};

class LevelOfInterest: public InternalVariable { 
    SO_NODE_HEADER(InternalVariable);
    public:
    static void initClass();
    
    
    private:
    LevelOfInterest();
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~LevelOfInterest();
};

class Motivation: public InternalVariable { 
    SO_NODE_HEADER(Motivation);
    public:
    static void initClass();
    
    
    private:
    Motivation();
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~Motivation();
};

class Reinforcement: public InternalVariable { 
    SO_NODE_HEADER(Reinforcement);
    public:
    SoSFNode underlyingMotivation; // points to motivational variable
    static void initClass();
    virtual float getInitialUpdateValue();
    
    private:
    Reinforcement();
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~Reinforcement(); 
};
#endif  







