/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/

#include "Pronome.h"

#include <stdlib.h>

Link::Link(Pronome *pobj) { next = 0; pronome = pobj; }

Link::~Link() {delete next; }


PronomeList::PronomeList() 
{
    length = 0;
    head = tail = 0;
}
PronomeList::~PronomeList() 
{
    delete head; // this will delete every node
}
  
void PronomeList::add(Pronome *pronome) 
{
    if (head) {
      tail->next = new Link(pronome);
      tail = tail->next;
      length++;
    }
    else {
      head = tail = new Link(pronome);
      length = 1;
    }
}
  
Pronome *PronomeList::currentPronome() 
{
    return ((tail)? tail->pronome : NULL);
}

Link *PronomeList::find(Pronome *pronome)
{
    Link *curr;
    
    for(curr = head;curr;curr=curr->next){
	if(curr->pronome == pronome)
	    return curr;
    }
    return NULL;
}

void *PronomeList::remove(Pronome *pronome)
{
    Link *curr;
    Link *prev;
    
    for(curr =head, prev=NULL;curr;prev = curr, curr=curr->next){
	if(curr->pronome == pronome){
	    if(curr == head)
		head = head->next;
	    else
		prev->next = curr->next;
	    
	    if(curr == tail)
		tail = prev;
	    
	    curr->next = NULL;
	    delete curr;
	    length --;
	    return pronome;    
	}
    }
    return NULL;
}

void PronomeList::empty()
{
    Link *curr;
    Link *next;
    
    delete head;
    head = tail = NULL;
    length = 0;
    return;
}
