/**********************************************************************
***********************************************************************
Copyright 1995 by the Massachusetts Institute of Technology.  All 
rights reserved.

Developed by Bruce M. Blumberg at the Media Laboratory, MIT, Cambridge, 
Massachusetts, with support from Sega North America.

For use by Dan Weld, University of Washington.
  
This distribution is approved by Nicholas Negroponte, Director of 
the Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without fee is hereby
granted, provided that this copyright notice and the original authors'
names appear on all copies and supporting documentation.  If individual 
files are separated from this distribution directory structure, this 
copyright notice must be included.  For any other uses of this software, 
in original or modified form, including but not limited to distribution 
in whole or in part, specific prior permission must be obtained from 
MIT.  These programs shall not be used, rewritten, or adapted as the 
basis of a commercial software or hardware product without first 
obtaining appropriate licenses from MIT.  MIT makes no representations 
about the suitability of this software for any purpose.  It is provided 
"as is" without express or implied warranty.
************************************************************************
************************************************************************/
#ifndef TRELEASER
#define TRELEASER
#include <Inventor/SbBasic.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoMFString.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/nodes/SoNode.h>
#include "InternalVariable.h"
#include "BaseBehaviorNode.h"

class BehaviorUpdateAction;
class ReleasingMechanism;

 

/********************************************************************
 *
 *	ReleasingMechanism
 * 
 ********************************************************************/

class ReleasingMechanism : public BaseBehaviorNode {
    SO_NODE_HEADER(ReleasingMechanism);
    public:
    
    enum ReleaserType{
	STDRELEASER, 
	AVGINGRELEASER, 
	SUMMINGRELEASER, 
	TRIGGERRELEASER 
    }; 
    
    SoSFInt32 typeTag;
    SoSFFloat minValue;
    SoSFFloat maxValue;
    SoMFFloat activeRange;
    SoSFFloat period;
    SoSFNode sensor;
    SoSFEnum type;
    SoSFString filterCBName;
    SoSFString findCBName;
    SoSFString weightCBName; 
    rmFilterCB rmFilterCallback;
    rmWeightCB rmWeightCallback;
    rmFilterCB rmFindCallback;
    SoSFFloat reinforcementWindow;
    SoSFFloat reinforcedMaxValue;
    SoSFFloat experimentGain;
    
    protected:
    float *hist;
    int lastAdjustedAt;
    int maxValueAt;
    int next;
    int reinforcementReceivedAt;
    float initialMaxValue;
    
    
    public:
    static void initClass();
    void adjustValueFromHistoryAt(int t);
    
    void setRMFilterCallback(rmFilterCB callback);
    void setRMWeightCallback(rmWeightCB callback);
    void setRMFindCallback(rmFilterCB callback);
    
    float linearWeightBasedOnRange(float r); // linear weight
    float calcWeightBasedOnRange(float r);  // spline-weight
    virtual void doAction(SoAction *action);
    
    /* You can either subclass any of the following 3 methods or
     * you can set the appropriate callback. 
     */
    virtual SbBool find(BehaviorUpdateAction *action); // returns FALSE by default
    virtual SbBool filter(BehaviorUpdateAction *action);// returns TRUE by default
    virtual float  weight(BehaviorUpdateAction *action);// does nothing by default
    
    /*
     *  This is a trivial method which does the following: 
     *	if(find(act)&& filter(act))
     *	    weight(act);
     */
    virtual void  _updateValue(BehaviorUpdateAction *action);
    
    /*
     * calls _updateValue() so you should subclass that if you
     * want to create a new kind of releasing Mechanism,  or you
     * can set callbacks to use if you don't want to subclass.
     */
    virtual float updateValue(BehaviorUpdateAction *action);
    
    void reportStateAt(int t);
    
    virtual void doLearning(float r, BehaviorUpdateAction *action);
    virtual void forget(BehaviorUpdateAction *action);
    
    ReleasingMechanism();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool readInstance(SoInput *in);
#endif
    virtual ~ReleasingMechanism();

};
 
#endif  
