//
//	tclCreatures.c++	tag@media-lab  Wed Aug 24 07:20:14 EDT 1994
//
#include <iostream.h>
#include <assert.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <stdio.h>
#include <tcl.h>
#include "CreatureLib.h"
#include "DogLib.h"
#include "MotorLib.h"
#include "TextLib.h"

#include "Command.h"
#include "Alive.h"
#include "AliveDispatch.h"
#define ERROR(S) \
	return((sprintf(interp->result, "%s%s", (S), ((CommandList *)(clientData))->help), \
	TCL_ERROR))

#define OK return(TCL_OK)

extern double atof();
extern int continuousMode;
extern int monitorTclFlag;

//
//  a hash table for linking event names to class instances.
//
static Tcl_HashTable creatureTable;

BaseCreature *Tcl_GetCreature(char *name)
{
   Tcl_HashEntry *entryPtr = Tcl_FindHashEntry(&creatureTable, name);
   if (entryPtr == NULL) {
       return (BaseCreature *)NULL;
   }
   return (BaseCreature *)Tcl_GetHashValue(entryPtr);
}

//
//
//
#include "globals.h"
extern Globals globals;

static int 
Creature(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{	
    if (argc != 3) {
	ERROR("");
    } else {
	int new_flag;
	Tcl_HashEntry *entryPtr;
    
	entryPtr = Tcl_CreateHashEntry(&creatureTable, argv[1], &new_flag);
	if (!new_flag) {
	    sprintf(interp->result, "Creature %s already exisits\n", argv[1]);
	    return(TCL_ERROR);
	}
	SoNode *fn =  globals.creatureCache->addBaseCreatureFromFile(argv[2]);
	if(!fn)
	    return(TCL_ERROR);
    
	Tcl_SetHashValue(entryPtr, fn);
    
	OK;
    }
}


static int 
CreatureDo(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{	
   if (argc < 3) {
       ERROR("");
   } else {

        BaseCreature *c = Tcl_GetCreature(argv[1]);
        if (!c) {
	    sprintf(interp->result, "No creature named %s\n", argv[1]);
            return(TCL_ERROR);
	}

       if (!(c->isOfType(MotorCreature::getClassTypeId()))) {
	  sprintf(interp->result, "%s is not a MotorCreature\n", argv[1]);
	  return ( TCL_ERROR );
       }

       MotorCreature *mc = (MotorCreature *)c;
       MotorCommandBlock mcb;


       mcb.cmd = mc->controller->getCommandId(argv[2]);
       for(int i = 3; i < argc; i++) 
	mcb.setNextArgGeneric((float)atof(argv[i]));
       mc->controller->executeMotorCommand(mcb);
      OK;
  }
}

static int 
CreatureMove(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{	
   if (argc < 4) {
       ERROR("");
   } else {

        BaseCreature *c = Tcl_GetCreature(argv[1]);
        if (!c) {
	    sprintf(interp->result, "No creature named %s\n", argv[1]);
            return(TCL_ERROR);
	}
	SbVec3f np((float)atof(argv[2]), 
		(float)atof(argv[3]), 
		(float)atof(argv[4]));
	c->setPosition(np);
      OK;
  }
}

static int 
run(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{	
   if (argc < 1) {
       ERROR("");
    } 
    else {
	continuousMode = 1;
	OK;
    }
}

static int 
tclOff(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{	
   if (argc < 1) {
       ERROR("");
    } 
    else {
	monitorTclFlag = 0;
	OK;
    }
}

static int 
step(ClientData clientData, Tcl_Interp *interp, int argc, char **)
{	
   if (argc < 1) {
       ERROR("");
    } 
    else {
	continuousMode = 0;
	globals.creatureCache->updateStateAt(.1);
	OK;
    }
}

//
//
//

static doEcho(int argc, char **argv)
{
    Text *c = (Text *)Tcl_GetCreature("echo");
    if(c != NULL){
	char *commandStr = Tcl_Merge(argc, argv);
	CommandData *cData = c->controller->getCommandData("setString");
	MotorCommandBlock mcb(cData->commandId);
	mcb.setNextArgGeneric(commandStr);
	c->controller->executeMotorCommand(mcb);
	printf("%s \n", commandStr);
	free(commandStr);
    }
}
static int 
creatureNameDo(ClientData, Tcl_Interp *interp, int argc, char **argv)
{	
  if (argc < 2) 
    {
      sprintf(interp->result, "No command given for %s", argv[0]);
      return(TCL_ERROR);
    } 
  else 
    {
      BaseCreature *c = Tcl_GetCreature(argv[0]);
      if (!c) 
	{
	  sprintf(interp->result, "No creature named %s\n", argv[0]);
	  return(TCL_ERROR);
	}
      
      if (!(c->isOfType(MotorCreature::getClassTypeId()))) 
	{
	  sprintf(interp->result, "%s is not a MotorCreature\n", argv[0]);
	  return ( TCL_ERROR );
	}
      
      doEcho(argc, argv);
      
      MotorCreature *mc = (MotorCreature *)c;
      
      CommandData *cData = mc->controller->getCommandData(argv[1]);
      if (cData)
	{
	  MotorCommandBlock mcb;
	  mcb.cmd = cData->commandId;
	  int argX = 2; // Current command line arg.
	  SbPList defaults = cData->getDefaults();
	  
	  // Use list of defaults to decide how to parse commmand line.
	  for (int i = 0; i < defaults.getLength(); i++)
	    {
	      ArgType *argType = (ArgType *)defaults[i];
	      switch (argType->getType()) 
		{
		case ArgType::FLOAT:
		  if (argX + 1 <= argc)
		    mcb.setNextArgGeneric((float)atof(argv[argX++]));
		  break;
		case ArgType::INT:
		  if (argX + 1 <= argc)
		    mcb.setNextArgGeneric((int)atoi(argv[argX++]));
		  break;
		case ArgType::SBVEC3F:
		  if (argX + 3 <= argc)
		    mcb.setNextArgGeneric(SbVec3f(atof(argv[argX++]),
						  atof(argv[argX++]),
						  atof(argv[argX++])));
		  break;
		case ArgType::CHARPTR:
		  if (argX + 1 <= argc)
		    mcb.setNextArgGeneric(argv[argX++]);
		  break;
		case ArgType::VOIDPTR:
		  fprintf(stderr,"ERROR: creatureNameDo: Cannot parse void *\n");
		  break;
		case ArgType::BASECREATUREPTR:
		  if (argX + 1 <= argc)
		    mcb.setNextArgGeneric((BaseCreature *)Tcl_GetCreature(argv[argX++]));
		  break;
		default:
		  assert(0);
		};
	   }
	  if (Alive::isDistributed == TRUE)
	    {
	      if (mc->motorCreatureType.getValue() == MotorCreature::MASTER)
		{
		  AliveDispatch::beginMCBSerialization(mc);
		}
	    }
	  mc->controller->executeMotorCommand(mcb);
	  if (Alive::isDistributed == TRUE)
	    {
	      if (mc->motorCreatureType.getValue() == MotorCreature::MASTER)
		{
		  AliveDispatch::endMCBSerialization();
		}
	    }
	  for (i = 0; i < mcb.getRetArgCount(); i++){
	    ArgType retArg= mcb.getReturnArg(i);
	    retArg.print();
	    }
	}
      else
	sprintf(interp->result,
		"The creature %s does not understand the command %s.\n",
		argv[0],argv[1]);
      OK;
    }
}
          


static int 
creature(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{	
   if (argc != 3) {
       ERROR("");
   } else {
	int new_flag;
	Tcl_HashEntry *entryPtr;

	SoInput mySceneInput;
	SoNode *fn;
	if (!mySceneInput.openFile(argv[2])) {
	   ERROR("Cannot open file\n");
	}
	SoDB::read(&mySceneInput, fn);
       
       entryPtr = Tcl_CreateHashEntry(&creatureTable, argv[1], &new_flag);
       if (!new_flag) {
	   sprintf(interp->result, "Creature %s already exisits\n", argv[1]);
	   return(TCL_ERROR);
       }

	if (fn->isOfType(BaseCreature::getClassTypeId())) {
            Tcl_SetHashValue(entryPtr, fn);

//            if (fn->isOfType(Camera::getClassTypeId()))
//             globals.creatureCache->addBaseCreatureToFront((BaseCreature *)fn);
//	    else
             globals.creatureCache->addBaseCreature((BaseCreature *)fn);


	    Tcl_CreateCommand (globals.interp, argv[1], creatureNameDo,
			    (ClientData) NULL, (Tcl_CmdDeleteProc *)NULL);

       } else {
            SoSeparator *sep = new SoSeparator;
            SoTransform *trans = new SoTransform;
            sep->addChild(trans);
            sep->addChild(fn);

            Tcl_SetHashValue(entryPtr, sep);
            globals.geomSep->addChild(sep);
       }

       OK;
   }
}


static CommandList creatures_cmds[] =
{
{  "load", creature,
   "load <name> <filename.iv> "
},
{  "CreatureDo", CreatureDo,
   "CreatureDo <name> <command-index(int)> "
},
{  "CreatureMove", CreatureMove,
   "CreatureMove <name> <command-index(int)> "
},
{  "run", run,
   "run"
},
{  "step", step,
   "step "
},
{  "tclOff", tclOff,
   "tclOff"
},
{  NULL }
};

void
AddCommands_creature_cmds(Tcl_Interp *interp, Commands *coms)
{

     //
     // Init the hash table
     //
     Tcl_InitHashTable(&creatureTable, TCL_STRING_KEYS);

     coms->addCommandList(interp, creatures_cmds);
}



