#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/sensors/SoAlarmSensor.h>
#include <bstring.h>
#include "AliveDog.h"
#include "DogMotorSkills.h"
#include "CreatureLib.h"
#include "DogController.h"
#include "SensorLib.h"
#include "../CreatureLib/TextGauge.h"
#include "MoreSkills.h"
#include "TimeStat.h"
// Change this to 1 if you want a faster but more glidy walk
#define USEAVERAGESPEED 0
#define GC_SETBVAR 2000
#define GC_SETBFIELD 2001
#define GC_SETRMTYPETAG 2002
#define GC_GETACTIVEBEV 2003
#define GC_GETACTIVERM 2004
#define GC_BUILDRMS 2005
#define GC_BUILDACTRM 2006
#define GC_GETMOT 2007
#define GC_GETMOTS 2008
int showState = 0;
char *activityfile = "sample.iv";
char *gaugebufs[20];

SO_NODE_SOURCE(AliveDog);

void doResetAction(BaseCreature *who)
{
    MotorCreature *theCreature = (MotorCreature *)who;
    if(theCreature->behaviorRoot != NULL){
	BehaviorResetAction *resetAction = new BehaviorResetAction();
	resetAction->apply(theCreature->behaviorRoot);
	delete resetAction;
    }
}
Timestat *ts = NULL;

void
AliveDog::initClass()
{
    SO_NODE_INIT_CLASS(AliveDog, Dog, "Dog");
    BaseCreature::addCallback("resetAction",doResetAction);
}

AliveDog::AliveDog() : Dog()
{
    SO_NODE_CONSTRUCTOR(AliveDog);
    SO_NODE_ADD_FIELD(gaugeOrigin, (0, 0, 0));
    SO_NODE_ADD_FIELD(traceLevel, (0));
    SO_NODE_ADD_FIELD(useBrainProbe, (FALSE));
    behaviorGauge = NULL;
    int i;
    for(i=0;i<20;i++){
	gaugebufs[i] = new char[256];
	bzero(gaugebufs[i], 256);
    }

    motivationN = -1;
    personFieldN = 0;
    timestamp = -1;
    commFlag = 1;
    ts = new Timestat("behavior update time", 20);
}

AliveDog::~AliveDog()
{

}

void AliveDog::buildBehaviorSystem()
{
    MotorCreature::buildBehaviorSystem();
    if(behaviorRoot){
	/* now allocate actions */
	ivAction = new InternalVariableUpdateAction();
	dbAction = new CreatureBehaviorUpdateAction(this);
	
	int i, agFound=0; 
	int len = ((SoGroup *)behaviorRoot)->getNumChildren();
	SoNode *aChild;
	for(i=0;i<len;i++){
	    aChild = ((SoGroup *)behaviorRoot)->getChild(i);
	    if(aChild->isOfType(ActivityGroup::getClassTypeId())){
		topActivityGroup = (ActivityGroup *)aChild;
		agFound++;
	    }
	}
	printf("activity groups found %d \n", agFound);
/*	cache pointer to top level activityGroup 
	SoSearchAction sa;
	SoPath *p;
	sa.setType(ActivityGroup::getClassTypeId());
	sa.apply(behaviorRoot);
	p = sa.getPath();
	if(!p){
	    printf("Error no ActivityGroup found \n");
	}
	else
	    topActivityGroup = (ActivityGroup *)p->getTail();
*/	    
	// Add callbacks
	// Added defaults.  hardts
	SetBehaviorVariable *sbv = new SetBehaviorVariable(this,"setmotvar");
	getController()->addCommand( GC_SETBVAR, "setMotivation", sbv, TRUE,
				     ArgType("none"),ArgType((float)0));
	MotorSkill *ms = getMotorSkillByName("getfloatfieldinnode");
	getController()->addCommand( GC_GETMOT, "getMotivation", ms, TRUE,
				     ArgType("none"),ArgType("value"), ArgType((int)1));
	
	getController()->addCommand( GC_SETRMTYPETAG, "setObjectType", sbv, TRUE,
				     ArgType("none"),ArgType((float)0));
	SetBoolField *sbf = new SetBoolField(this,"setboolfield");
	getController()->addCommand( GC_SETBFIELD, "set", sbf, TRUE,
				     ArgType("none"),ArgType((int)0));
	
	GetBehaviorStateVector *gbsv = new GetBehaviorStateVector(this,"getBehaviorStateVector");
	getController()->addCommand( GC_GETACTIVEBEV, "getActiveBehaviors", gbsv,
				     TRUE);
	GetRMStateVector *grmv = new GetRMStateVector(this,"getRMStateVector");
	getController()->addCommand( GC_GETACTIVERM, "getActiveRMs", grmv,
				     TRUE,ArgType((int)0), ArgType((char *)NULL) );
	GetMotStateVector *gmmv = new GetMotStateVector(this,"getMotStateVector");
	getController()->addCommand( GC_GETMOTS, "getAllMotivations", gmmv,
				     TRUE );
	BuildBooleanRMs *bbrm = new BuildBooleanRMs(this,"buildBooleanRMs");
	getController()->addCommand( GC_BUILDRMS, "buildTestRMs", bbrm,
				     TRUE,ArgType((BaseCreature *)NULL) );
	BuildWatchActRM *bwarm = new BuildWatchActRM(this,"buildWatchActRM");
	getController()->addCommand( GC_BUILDACTRM, "buildActRM", bwarm,
				     TRUE,ArgType((char *)NULL), ArgType((char *)NULL) );			     
    }
}

void AliveDog::traceBehaviorSystemAt(UpdateStateData &usd)
{
    if(!behaviorGauge){
	behaviorGauge = new TextGauge();
	behaviorGauge->moveTextTo(gaugeOrigin.getValue());
	((SoSeparator *)usd.currSceneGraph)->addChild(behaviorGauge->root);
    }
    int i;
    ActivityGroup *traceGroup = topActivityGroup;
    BaseActivity *act;
    
    for(i=0;i<traceLevel.getValue();i++){
	act = traceGroup->activeActivity;
	if(act->leaf.getValue())
	    break;
	else
	    traceGroup = act->getActivityGroup();
    }
    int numActs = traceGroup->printState(gaugebufs);
    for(i=0;i<numActs;i++){
	behaviorGauge->set1Value(i, gaugebufs[i]);
    }
    for(i=numActs;i<20;i++){
	behaviorGauge->set1Value(i,"");
    }
}

SoAction *AliveDog::getBehaviorUpdateAction()
{
    return dbAction;
}

void AliveDog::updateBehaviorSystemAt(UpdateStateData &usd)
{
//    ts->startTime();
    ivAction->apply(behaviorRoot);
    dbAction->apply(behaviorRoot);
//    ts->endTime();
}

SniffSensor *AliveDog::getSniffSensor()
{
    return (SniffSensor *)getSensorOfType(SniffSensor::getClassTypeId());
}

PotentialFieldSensor *AliveDog::getPFSensor()
{
    return (PotentialFieldSensor *)getSensorOfType(PotentialFieldSensor::getClassTypeId());
}

RayData *AliveDog::getDataForClosestUsingSniff(int whichTag)
{
    SniffSensor *ss = getSniffSensor();
    RayData *rd = NULL;
    if(ss)
	rd = getSniffSensor()->getDataForClosestWithTag(whichTag);
    return rd;
}

InternalVariable *AliveDog::getInternalVariable(char *nodeName)
{
	SoNode *n = behaviorRoot->getByName(nodeName);
	if (!(n->isOfType(InternalVariable::getClassTypeId()))) {
		fprintf(stderr, "No internal variable by name >%s<\n", nodeName);
		return (InternalVariable *)NULL;
	}
	return (InternalVariable *)n;
}

float AliveDog::getMotivation(const char *nodeName)
{
	MotorCommandBlock mcb(controller->getCommandId("getMotivation"),
	    ArgType((char *)nodeName));
	controller->executeMotorCommand(mcb);
	if(mcb.getRetArgCount()){
	    return mcb.getRetFloatAt(0);
	}
	else {
	    printf("motivation %s not found \n");
	    return -1;
	}   
}

void AliveDog::setMotivation(const char *nodeName, float val)
{
	MotorCommandBlock mcb(controller->getCommandId("setMotivation"),
	    ArgType((char *)nodeName), 
	    ArgType(val));
	controller->executeMotorCommand(mcb);
}

void AliveDog::getActiveBehaviors(char **behaviors,int &num)
{
	MotorCommandBlock mcb(controller->getCommandId("getActiveBehaviors"));
	controller->executeMotorCommand(mcb);
	int i, k =0;
	for(i = 0;i<mcb.getRetArgCount();i +=2){
	    sprintf(behaviors[k++],"%s",mcb.getRetCharPtrAt(i)); 
	}
	num = mcb.getRetArgCount()/2;
}

SbBool AliveDog::getPersonFieldValue(const char *fieldName)
{
	MotorCommandBlock mcb(controller->getCommandId("getBool"), 
	    ArgType((char *)fieldName));
	MotorCreature *mc = (MotorCreature *)myCache->findBaseCreatureWithTag(300);
	if(mc){
	    mc->controller->executeMotorCommand(mcb);
	    return (mcb.getRetIntAt(0)== 1)? TRUE : FALSE;
	}
	else return FALSE;
}

//////////////////////////////////////////////////////////
// By Akira
//   For Tcl/Tk Interface of dog's internal state
//////////////////////////////////////////////////////////


// Create a child Tk process and establish a pair of pipe (for read and write, 
// respectively). 

void AliveDog::openTkInterface () 
{
    FILE *fpr, *fpw;
    int pid;
    // RW... parent reads, child writes. vice versa.
    int pipefdsRW[2], pipefdsWR[2];  

    if(pipe(pipefdsRW) < 0) {
	perror("pipe");
	exit(1);
    }

    if(pipe(pipefdsWR) < 0) {
	perror("pipe");
	exit(1);
    }

    if((pid = fork()) < 0) {
	perror("fork");
	exit(1);
    }

    if(pid == 0) {
	// This portion is executed in the child process (Tcl/Tk)

	close(0);
	dup(pipefdsWR[0]);
	close(pipefdsWR[0]);
	close(pipefdsWR[1]);

	close(1);
	dup(pipefdsRW[1]);
	close(pipefdsRW[1]);
	close(pipefdsRW[0]);

	// Since the 'select' system call is needed, I need wishx (TclX versin of wish)
	chdir("../BrainProbe");
	execl("./wishx","wishx","-f","silas",0);
	// If failed to execl... perhaps wishx is not found
	fprintf(stderr, "AliveDog::openTkInterface: failed to execl ../BrainProbe/wishx\n");
	fprintf(stderr, "Please check that  ../BrainProbe/wishx exists\n");
	exit(1);
    }

    // main (parent) process (ALIVE)
    fpr = fdopen(pipefdsRW[0],"r");
    close(pipefdsRW[1]);
    fpw = fdopen(pipefdsWR[1],"w");
    close(pipefdsWR[0]);

    fdR = pipefdsRW[0];
    pipeR = fpr;
    pipeW = fpw;
}

// Examine whether there is a message in the input pipe.
// Return value: 1 ... a message is found.
//               0 ... no message is found.
int AliveDog::peekInput()
{
    int nfds;
    fd_set readfds;
    struct timeval tv;
    
    FD_ZERO(&readfds);
    FD_SET(fdR,&readfds);            // specify input pipe
    // Don't wait ... set 0 second.
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    nfds = select(fdR+1,&readfds,NULL,NULL,&tv);
    if(nfds == 0) {
	return 0;
    } else {
	if(FD_ISSET(fdR,&readfds)) {
	    return 1;
	} else {
	    fprintf(stderr,"AliveDog::peekInput: something wrong.\n");
	    return 0;
	}
    }
}

// Read the list of motivations the Tk interface needs.
// This is called only once per dog.
// The first input from the pipe should be this list.
// Format: 
//   "MotivationList"
//   motivation1
//   motivation2
//   ...........
//   "End"
void AliveDog::readMotivationListFromPipe() 
{
    char buf[50];
    int i;
    fgets(buf,49,pipeR);
    if(strcmp(buf,"MotivationList\n") != 0) {
	fprintf(stderr,"readMotivationListFromPipe: invalid line:%s\n",buf);
	return;
    }
    for(i=0,fgets(buf,49,pipeR); strcmp(buf,"End\n") != 0; i++,fgets(buf,49,pipeR)) {
	if(strlen(buf) > 0)
	    buf[strlen(buf)-1] = NULL;  // remove the CR.
	strcpy(motivationNames[i],buf);
    }
    motivationN = i;
}

// Read the list of person fields from the Tk interface needs.
// This is called only once per dog.
// The first input from the pipe should be this list.
// Format: 
//   "PersonFieldList"
//   personfield1
//   personfield2
//   ............
//   "End"
void AliveDog::readPersonFieldListFromPipe() 
{
    char buf[50];
    int i;
    fgets(buf,49,pipeR);
    if(strcmp(buf,"PersonFieldList\n") != 0) {
	fprintf(stderr,"readPersonFieldListFromPipe: invalid line:%s\n",buf);
	return;
    }
    for(i=0,fgets(buf,49,pipeR); strcmp(buf,"End\n") != 0; i++,fgets(buf,49,pipeR)) {
	if(strlen(buf) > 0)
	    buf[strlen(buf)-1] = NULL;  // remove the CR.
	strcpy(personFieldNames[i],buf);
    }
    personFieldN = i;
}

// Read one motivation from pipe.
// Format: 
//    motivationName:value
void AliveDog::readMotivationInternal(char *buf)
{
    float value;
    static char *bp,name[50];

    bp = strchr(buf, (int)':');
    if (bp == NULL) return;   // blank line
    name[0]=NULL;
    strncat(name,buf,bp-buf);
    value = atof(bp+1);
    setMotivation(name, value);
//    printf("Read name = %s value = %f\n",name,value);
}

void AliveDog::readMotivationFromPipe()
{
    static char buf[50];

    // Get timestamp
    fgets(buf,49,pipeR);

    // I don't know why, but when a slider on Tk interface is touched
    // more than once with very short time interval (that is, when more than one 
    // messages are sent from Tk interface at one time), peekInput only find the
    // first input and ignores later input (when new message is sent from Tk
    // interface, peekInput works again).
    // In order to avoid this problem, Tk interface sends a blank line when this 
    // inconsistency is found, so that peekInput can catch up. Since this 
    // blank line is not needed as a data, this should be ignored.
    if(strcmp(buf, "\n") == 0) {
//	printf("BLANK LINE FOUND\n");
	// Ignoring a blank line
	return;
    }

    timestamp = atoi(buf);

    fgets(buf, 49, pipeR);
    if(strcmp(buf,"Motivation\n") == 0) {
	while(fgets(buf, 49, pipeR) != NULL) {
	    if(strcmp(buf,"End\n") == 0) 
		break;
	    readMotivationInternal(buf);
	}
    } else {
	readMotivationInternal(buf);
    }
}

void AliveDog::writeMotivationToPipe()
{
    if(motivationN == 0){
	// In this case, it is unknown which motivations the Tk interface 
	// need to display. 
//	printf("writeMotivationToPipe: debug: returning");
	return;
    }

    int i = 0;
    float value;

    // Write timestamp
    if(fprintf(pipeW,"%d\n",timestamp) < 0) {
	fprintf(stderr,"writeActiveBehaviorPipe: failed to write\n");return;
    }

    if(fprintf(pipeW, "Motivation\n") < 0) {
	fprintf(stderr,"writeMotivationPipe: failed to write\n");return;
    }
    for(i=0; i<motivationN; i++) {
	value = getMotivation(motivationNames[i]);
	if(fprintf(pipeW,"%s:%f\n",motivationNames[i],value) < 0) {
	    fprintf(stderr,"writeMotivationPipe: failed to write\n");return;
	}
//	printf("name = %s value = %f\n",motivationNames[i],value);
    }
    if(fprintf(pipeW,"End\n") < 0) {
	fprintf(stderr,"writeMotivationPipe: failed to write\n");return;
    }
    fflush(pipeW);
}


// Write active behavior to pipe stream
// Format:
//   "ActiveBehavior"
//   activebehavior1
//   activebehavior2
//   ...............
//   "End"
// Deactivated behaviors are not recorded.
void AliveDog::writeActiveBehaviorToPipe()
{
    int i,num;
    //Assuming that the number of behaviors and the length of their names...

    static char *behaviors[20];
    static char buf[20][20];
    static int initflag=1;

    // Prepare the behaviors buffer.
    // This is kind of dirty.
    if(initflag) {
	for(i=0; i<20; i++) {
	    behaviors[i] = buf[i];
	}
	initflag = 0;
    }

    getActiveBehaviors(behaviors,num);

    if(fprintf(pipeW,"ActiveBehavior\n") < 0) {
	fprintf(stderr,"writeActiveBehaviorPipe: failed to write\n");return;
    }
    for(i=0; i<num; i++) {
	if(fprintf(pipeW,"%s\n",behaviors[i]) < 0) {
	    fprintf(stderr,"writeActiveBehaviorPipe: failed to write\n");return;
	}
//	printf("name = %s\n",behaviors[i]);
    }
    if(fprintf(pipeW,"End\n") < 0) {
	fprintf(stderr,"writeActiveBehaviorPipe: failed to write\n");return;
    }
    fflush(pipeW);
}

void AliveDog::writePersonFieldToPipe()
{
    if(personFieldN == 0){
	// In this case, it is unknown which person field the Tk interface 
	// need to display. 
	printf("writePersonFieldToPipe: debug: returning");
	return;
    }

    int i = 0;
    SbBool value;

    if(fprintf(pipeW, "PersonField\n") < 0) {
	fprintf(stderr,"writePersonFieldToPipe: failed to write\n");return;
    }
    for(i=0; i<personFieldN; i++) {
	value = getPersonFieldValue(personFieldNames[i]);
	if(fprintf(pipeW,"%s:%d\n",personFieldNames[i],value) < 0) {
	    fprintf(stderr,"writePersonFieldToPipe: failed to write\n");return;
	}
//	printf("name = %s value = %d\n",personFieldNames[i],value);
    }
    if(fprintf(pipeW,"End\n") < 0) {
	fprintf(stderr,"writePersonFieldToPipe: failed to write\n");return;
    }
    fflush(pipeW);
}

void synchronizeCB(void *data, SoSensor *)
{
    int *commFlag = (int *)data;
    if(*commFlag == 0)
	*commFlag = 1;
}

void AliveDog::updateStateAt(UpdateStateData &usd)
{
  if ( ! (myCache->isDistributed == TRUE &&
	  motorCreatureType.getValue() == MotorCreature::DROID))  /* Skip opening brain probe if
								     we're running distributed mode and
								     the dog is a droid on this machine */
    {
      if(useBrainProbe.getValue()) {
	if(motivationN == -1) {  // First time
	  openTkInterface();
	  motivationN = 0;
	}
	if(commFlag == 1) {
	  SoAlarmSensor *as = new SoAlarmSensor(synchronizeCB, &commFlag);
	  as->setTimeFromNow(3.0); // do it every 3 seconds
	  as->schedule();
	  
	  while(peekInput() != 0) {
	    if(motivationN == 0) {
	      readMotivationListFromPipe();
	      readPersonFieldListFromPipe();
	    }
	    else 
	      readMotivationFromPipe();
	  }
	}
      }
    }

    // Do the ordinary thing...
    MotorCreature::updateStateAt(usd);

    if(useBrainProbe.getValue() && commFlag == 1) {
//    if(commFlag == 1) {
	if(motivationN == 0 || personFieldN == 0) {
	    // not ready yet.
	    return;
	} else {
	    writeMotivationToPipe();
	    writeActiveBehaviorToPipe();
	    writePersonFieldToPipe();
	    commFlag = 0;
	}
    }
}
