#ifndef ALIVEDOG_1DEF
#define ALIVEDOG_1DEF

#include <Inventor/SbPList.h>
#include "Dog.h"
#include "ActivityLib_1.h"
#include "BehaviorCreatureLib.h"
#include <Inventor/fields/SoSFString.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFBool.h>
class SniffSensor;
class AliveRaySensor;
class PotentialFieldSensor;

class RayData;
class SbMatrix;
class SoCamera;
class DogController;
class TextGauge;


class AliveDog : public Dog {
    SO_NODE_HEADER(AliveDog);

public:
    static void initClass();
    SoSFVec3f gaugeOrigin;
    SoSFInt32 traceLevel;
    SoSFBool useBrainProbe;
    
    /* these guys are for the behavior system */
    InternalVariableUpdateAction *ivAction;
    CreatureBehaviorUpdateAction *dbAction;
    
    TextGauge *behaviorGauge;
    
    /* purely convenience functions */
    PotentialFieldSensor *getPFSensor();
    SniffSensor *getSniffSensor();
    RayData *getDataForClosestUsingSniff(int whichTag);
    
    virtual void traceBehaviorSystemAt(UpdateStateData &usd);
    
    /* over-rides of methods defined in MotorCreature */
    virtual SoAction *getBehaviorUpdateAction();
    void updateBehaviorSystemAt(UpdateStateData &usd);
    
    virtual void buildBehaviorSystem();

    AliveDog();       // Note only should be call during a read
    
    float getMotivation(const char *nodeName);
    void  setMotivation(const char *nodeName, float val);
    void  getActiveBehaviors(char **behaviors,int &num);
    SbBool getPersonFieldValue(const char *fieldName);
    InternalVariable *getInternalVariable(char *nodeName);
    
    // By Akira: for Tcl/Tk interface
    virtual void updateStateAt(UpdateStateData &);
    int motivationN;
    int personFieldN;
    //This table is read from motivation file only once.
    //I'm assuming the number of of possible motivations is less than 20.
    char motivationNames[20][20];  
    char personFieldNames[30][20];  
    // Timestamp is generated by Tk interface. ALIVE never modifies the 
    // value of timestamp. The idea is that ALIVE puts current timestamp to 
    // the message it writes to the stream, so that the Tk interface can ignore 
    // any old message from ALIVE. This will prevent the slider values on Tk 
    // interface to be backdated.
    // This timestamp management is only for motivations.
    int timestamp;
    int commFlag;
    int fdR;             // file descriptor ... corresponds to pipeR.
    FILE *pipeR, *pipeW;
    void openTkInterface();
    int peekInput();
    void readMotivationListFromPipe();
    void readPersonFieldListFromPipe();
    void readMotivationFromPipe();
    void readMotivationInternal(char *);
    void writeMotivationToPipe();
    void writeActiveBehaviorToPipe();
    void writePersonFieldToPipe();

protected:
    ActivityGroup *topActivityGroup;
    virtual ~AliveDog();
    
}; 

#endif
