#include <assert.h>
#include "SensorLib.h"
#include "DogLib.h"
#include "DurationTimer.h"
#include "DogMotorSkills.h"
#include "AliveDogLib_2.h"
#include "ActivityLib_1.h"
#include "BehaviorCreature.h"
#include "util.h"
#include "PotentialField.h"
#include "BasicDefines.h"
#include "UniformRandomVariable.h"

#include <Inventor/actions/SoGetMatrixAction.h>

void barkAction(MotorCreature* dog);
char* pickRandomBark(BaseActivity* act, MotorCreature* mc);
void suggestBarkCB(BaseActivity *act, BehaviorUpdateAction *action);
void suggestPantCB(BaseActivity *act, BehaviorUpdateAction *action);
void drinkSoundSuggestionCB(BaseActivity *act, BehaviorUpdateAction *action);
void howlCB(BaseActivity *act, BehaviorUpdateAction *action);
void howlSoundCB(BaseActivity *act, BehaviorUpdateAction *action);
void gazeAtObjectOfInterest(float bearing, BaseActivity *act, BehaviorUpdateAction *action);
void whimperCB(BaseActivity *act, BehaviorUpdateAction *action);


void issueSuggestion(BaseActivity *act, MotorCreature &mc, MotorCommandBlock &mcbToExecute, int whichKind)
{
//  mc.controller->suggestMotorCommand(mcbToExecute, whichKind, 
//    act, act->value.getValue(), FALSE);
    mc.controller->suggestMotorCommand(mcbToExecute, whichKind, 
        act, act->getValueBeforeInhibition(), FALSE);
}

void wanderCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    static count = 0;
    static val = 0;
    
    CBUA(dba, action);
    PFSENSOR(s, dba);
    LeafActivity *la = (LeafActivity *)act;
    if(!dba->creature->checkIfReadyToMove(la->mcb))
	return;
    else {

    if (count == 0) {
	val = rand() / 10923;
	if (val == 0) {
		count = rand() / 1638; // do for a max of 20 times
	} else {
		count = rand() / 3276; // do for a max of 10 times
	}
    } else {
	count--;
    }
    float tmp;
    switch (val) {
	case 0:
	    la->mcb.cmd = GC_GOSTRAIGHT;
	    break;	
	case 1:
	    la->mcb.cmd = GC_LFT;
	    tmp = 1.0;
	    la->mcb.setNextArgGeneric((float)tmp);
	    la->mcb.setNextArgGeneric((float).025);
	    break;	
	case 2:
	    la->mcb.cmd = GC_LFT;
	    tmp = 0.0;
	    la->mcb.setNextArgGeneric((float)tmp);
	    la->mcb.setNextArgGeneric((float).025);
	    break;	
    }

    }  

}
void followGaitSuggestion(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    SuggestedMCB *suggestion = dba->creature->controller->getSuggestionFor(GC_FWD);
    if(suggestion){
	LeafActivity *la = (LeafActivity *)act;
	la->issueMotorCommand(*dba->creature,*suggestion);
    }
}

void simpleWanderCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    static int timesCalled;
    static int currCmd = GC_STRAIGHT;

    
    CBUA(dba, action);
    PFSENSOR(s, dba);
    PRONOME(pronome, dba);
    pronome->location[0] = 30;
    pronome->location[1] = 11;
    pronome->location[2] = 0;
    pronome->location = s->getVecInWorldFrame(pronome->location);
    s->calcBearingAndRange(pronome->location, pronome->bearing, pronome->range);
    s->updateGradient();
    gazeAtObjectOfInterest(pronome->bearing, act, action);
    LeafActivity *la = (LeafActivity *)act;
    if(!dba->creature->checkIfReadyToMove(la->mcb))
	return;
    else {
	followGaitSuggestion(act, action);
	la->mcb.cmd = currCmd;
    }  

}


// aries 9/21/95
char *
pickRandomBark(MotorCreature* mc)
{
  int i, num = mc->sounds.getNum();
  char* barks[128]; // no more than 128 barks I hope 
  int barkptr = 0;
  
  if (num < 2)
    return;
  for(i = 0; i < num; i +=2) 
    {
//      printf("Sound: %s\n", mc->sounds[i].getString());
      if (!strncmp(mc->sounds[i].getString(),
		  "bark",
		  4))
	{
	  barks[barkptr++] = (char*) mc->sounds[i].getString();
	}
    }
  
  if (!barkptr) return NULL;
  
  UniformRandomVariable urv(0.0, float(barkptr));
  return barks[int(urv())];
  
}

    
void
barkAction(BaseActivity* act, MotorCreature* dog)
{
  char* bark = pickRandomBark(dog);
  if (!bark)
    return;
  //issue secondary actions
//  issueSuggestion(act, *dog,
//		  MotorCommandBlock(GC_SOUND_ON, ArgType(bark)),
//		  SuggestedMCB::DO);

  // make head jerk up for bark
//  dog->controller->executeMotorCommand(
//	MotorCommandBlock(GC_MOVEMOUTH, 
//	    ArgType((float).1), 
//	    ArgType((float)0.0)));

  dog->controller->executeMotorCommand(MotorCommandBlock(GC_BARK, 
    ArgType((float).1), ArgType((float)1.0), ArgType((char *)bark)));
   issueSuggestion(act, *dog,
		  MotorCommandBlock(GC_NOD, 
				    ArgType(float(.5)),
				    ArgType(float(0.0)),
				    ArgType(float(1.0))),
		  SuggestedMCB::DO);

}


// aries 9/21/95
void suggestBarkCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(pronome, dba);
  
  UniformRandomVariable urv = UniformRandomVariable(0.0, 1.0);
  urv.seed();
 
  // aggression is a 0 to 1 thing, 0 is completely submissive, 1 is
  // completely aggressive

  AliveDog* doggy = (AliveDog*) dba->creature;
  InternalVariable* angst = doggy->getInternalVariable("NEEDTOBARK");
  
  if (urv() < angst->value.getValue()) // jackpot 
    {
      barkAction(act, doggy);
      // printf("Woof!\n");
      angst->value.setValue(0.0); // null it
    }
}


// aries 9/21/95
void 
suggestPantCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  static int panting = 0;
  float thirst_thresh = 20.0;
  float thirst_max = 40.0;
  
  CBUA(dba, action);
  PRONOME(pronome, dba);
  
  AliveDog* doggy = (AliveDog*) dba->creature;
  InternalVariable* thirst_iv = doggy->getInternalVariable("NEEDTODRINK");
  float thirst = thirst_iv->value.getValue();

  // printf("Thirst: %f\n", thirst);

  if (panting)
    {

      float vol;

      vol = thirst/thirst_max;
      if (vol > 1.0) vol = 1.0;

      SoundSkill::setSoundVolume("pant", vol);
    }

  if (thirst > thirst_thresh && !panting)
    {
      float vol;

      vol = thirst/thirst_max;
      if (vol > 1.0) vol = 1.0;

      panting = 1;
      
      SoundSkill::setSoundVolume("pant", vol);
      
      issueSuggestion(act, *dba->creature, 
		      MotorCommandBlock(GC_SOUND_LOOP, ArgType("pant"), ArgType(1)),
		      SuggestedMCB::DO);
      issueSuggestion(act, *dba->creature, 
		      MotorCommandBlock(GC_SOUND_ON, ArgType("pant")),
		      SuggestedMCB::DO);

      // printf("Started panting\n");
    }

  if (thirst < thirst_thresh && panting)
    {
      panting = 0;
      issueSuggestion(act, *dba->creature, 
		      MotorCommandBlock(GC_SOUND_OFF, "pant"),
		      SuggestedMCB::DO);
      // printf("Stopped panting\n");
    }
}

void
drinkSoundSuggestionCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(pronome, dba);
  
  issueSuggestion(act, *dba->creature, 
		  MotorCommandBlock(GC_SOUND_ON, "drink"),
		  SuggestedMCB::DO);
  
}


void whimperCB(BaseActivity *act, BehaviorUpdateAction *action)
{
	
  CBUA(dba, action);
  PRONOME(pronome, dba);
  UniformRandomVariable urv;

  if (urv() < .05)
    {
      issueSuggestion(act, *dba->creature, 
		      MotorCommandBlock(GC_SOUND_ON, "whimper"),
		      SuggestedMCB::DO);
    }
  
}


void 
howlCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  float speed = .5; // raising head for howl

  CBUA(dba, action);
  PRONOME(pronome, dba);
  LeafActivity* la = (LeafActivity*) act;
  AliveDog* doggy = (AliveDog*) dba->creature;
  SbVec3f pos;
  MotorCommandBlock mcb(GC_LOOKAT);
  MotorCommandBlock sit(GC_SIT);
  static DurationTimer howl_time;

//  printf("Called howlCB\n");

  // make him look at a spot way over his head
  doggy->getPosition(pos);
  mcb.setNextArgGeneric(pos[0]+float(20.0));
  mcb.setNextArgGeneric(pos[1] + float(1000.0)); // look way over head
  mcb.setNextArgGeneric(pos[2]);
  mcb.setNextArgGeneric(float(speed));
  
  la->issueMotorCommand(*doggy, mcb);
  if(doggy->speed != 0)
	la->mcb.cmd = GC_HALT;
    else
	la->mcb.cmd = GC_SIT;
  //la->issueMotorCommand(*doggy, sit);

  
  Sitting* sitting = (Sitting*) doggy->getMotorSkillByName("sitting");
  if (!sitting) 
    {
      fprintf(stderr, "howlCB: Can't find sitting motorskill\n");
      return;
    }
  
  if(fabs(sitting->counter-1) < 1.0e-3)
    {
      //sucker is sitting
      MotorCommandBlock howl(GC_SOUND_ON, ArgType("howl"));
 //     la->issueMotorCommand(*doggy, howl); 
      la->mcb.cmd = GC_BARK;
      la->mcb.setNextArgGeneric((float).02);
      la->mcb.setNextArgGeneric((float)1.0); 
      la->mcb.setNextArgGeneric("howl");
    }
  else
    {
      // start the timer while sitting
      // howl_time.setTime();      
    }

}


void
howlSoundCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  
  CBUA(dba, action);
  PRONOME(pronome, dba);
  LeafActivity* la = (LeafActivity*) act;
  AliveDog* doggy = (AliveDog*) dba->creature;
  MotorCommandBlock howl(GC_SOUND_ON, "howl");

  la->issueMotorCommand(*doggy, howl);  
}


void aggSuggestionCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(p, dba);
    PFSENSOR(s, dba);
    if(act->data == NULL){
	act->data = ((AliveDog *)dba->creature)->getInternalVariable("AGGRESSION");
    }
    float agg; 
    if(act->data){
	agg = ((InternalVariable *)act->data)->value.getValue();
    }
    else{
	printf("AGGRESSION not found \n");
	return;
    }
    
    // aggression is a 0 to 1 thing, 0 is completely submissive, 1 is
    // completely aggressive
    
    //issue secondary actions
    // lower/raise the head
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_LOWERNECK, ArgType((float).1), 
				      ArgType((float)((1-agg)*2.0))),  
		    SuggestedMCB::DO);
    // wag tail
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVETAIL, ArgType((float)1-agg), 
				      ArgType((float)(.1 - (.1*fabs(agg-.5))/.5)), 
				      ArgType((float).1)), 
		    SuggestedMCB::DO);
    // ears 
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVEEARS, ArgType((float).1), 
				      ArgType((float)agg)), 
		    SuggestedMCB::DO);
    // eyebrows 
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVEBROWS, ArgType((float).1), 
				      ArgType((float)agg)), 
		    SuggestedMCB::DO);
    // bugeyes 
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_BUGEYES, ArgType((float).1), 
				      ArgType((float)1-agg)), 
		    SuggestedMCB::DO);
    // showTeeth 
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_SHOWTEETH, ArgType((float).1), 
				      ArgType((float)agg)), 
		    SuggestedMCB::DO);
    // move mouth
    if(agg> .7)
	issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVEMOUTH, ArgType((float).1), 
				      ArgType((float)agg)), 
		    SuggestedMCB::DO);

    if (agg > .9)
      {
	issueSuggestion(act, *dba->creature,
			MotorCommandBlock(GC_SOUND_ON, ArgType("growl")),
			SuggestedMCB::DO);
      }
      if((act->active.getValue()) && (dba->creature->isMoving.getValue() != TRUE)){
	    float bearing = p->bearing;
	    if(fabs(bearing)> 1){
		bearing = (bearing>1)?1 : bearing;
		bearing = (bearing< -1)?-1 : bearing;
		p->location = s->getVecInSensorFrame(p->location);
		p->location[0] = cos(bearing)*p->range;
		p->location[1] = 10.0;
		p->location[2] = sin(bearing)*p->range;
		p->location = s->getVecInWorldFrame(p->location); 
	    }
	    gazeAtObjectOfInterest(p->bearing, act, action);
      }
}
void guiltSuggestionCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    if(act->data == NULL){
	act->data = ((AliveDog *)dba->creature)->getInternalVariable("GUILT");
    }
    float guilt; 
    if(act->data){
	guilt = ((InternalVariable *)act->data)->value.getValue()/2.0;
    }
    else{
	printf("GUILT not found \n");
	return;
    }

    // wag tail
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVETAIL, ArgType((float)(.1)), 
				      ArgType((float)(guilt)), 
				      ArgType((float).1)), 
		    SuggestedMCB::DO);
    // showTeeth 
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_SHOWTEETH, ArgType((float).1), 
				      ArgType((float)guilt)), 
		    SuggestedMCB::DO);
    
}

void gaitSuggestionCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    AliveDog* doggy = (AliveDog*) dba->creature;
    LeafActivity *la = (LeafActivity *)act;

    MotorCommandBlock mcb(GC_FWD);
   
    // always go forward
    mcb.setNextArgGeneric((float)(1.0));

// aries 10/8/95
    float speed = doggy->speed;
    if (speed > 0.2)
      {
	/* makeing this a issueMotorCommand caused machine hangs.  WTF?! 
	 Fixed BMB */
	 dba->creature->controller->executeMotorCommand(
	    MotorCommandBlock(GC_SOUND_ON,ArgType("tags")));
	  
      }
    
    
    // choose gait
    float vbi = act->getValueBeforeInhibition();
//    if(vbi>.8) vbi = .8;
//        else if(vbi<.2) vbi = .2;
    mcb.setNextArgGeneric((float)((vbi>.8)?0.0:0.0));
    
    // use standard speed
    mcb.setNextArgGeneric((float)(dba->creature->defaultFwdSpeed.getValue()));
    
    // issue suggestion
    issueSuggestion(act, *dba->creature,mcb, SuggestedMCB::MODIFY);
    
    // issue secondary actions
    float wagspeed = (vbi>.5)?((vbi-.5)/.5)*.1 :0;
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_LOWERNECK, ArgType((float).1), 
				      ArgType((float)1-vbi)), 
		    SuggestedMCB::DO);
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVETAIL, wagspeed, 
				      ArgType((float)1-vbi), 
				      ArgType((float).1)), 
		    SuggestedMCB::DO);
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVEEARS, ArgType((float).1), 
				      ArgType((float)1-vbi)), 
		    SuggestedMCB::DO);
     if(vbi < .4)
    issueSuggestion(act, *dba->creature, 
		    MotorCommandBlock(GC_MOVEMOUTH, ArgType((float).1), 
				      ArgType((float)vbi)), 
		    SuggestedMCB::DO);
    
}

void lowerNeckSuggestionCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    
    issueSuggestion(act, *dba->creature, 
    MotorCommandBlock(GC_LOWERNECK, ArgType((float).1), ArgType((float)1.0))
	,SuggestedMCB::DO);
}

void movetoSchemaCallback(void *schemaPtr)
{
    PotentialField *schema = (PotentialField *)schemaPtr;
//   printf("testing schema callback \n");
    schema->decay = .50;
    schema->spreadPotential(SbVec3f(0, 0, 0), -16);  
}

void movetoSchemaCallback_1(void *schemaPtr)
{
    PotentialField *schema = (PotentialField *)schemaPtr;
//    printf("testing schema callback \n");
    schema->decay = .5;
    schema->spreadPotential(SbVec3f(0, 0, 0), -16);
    schema->decay = 4;
    schema->spreadPotential(SbVec3f(0, 0, 0), -64);  
}

void gazeAtObjectOfInterest(float bearing, BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(p, dba);
    PFSENSOR(s, dba);
//    CHECKPRONOME(p);
    LeafActivity *la = (LeafActivity *)act;
    // now do gaze
    MotorCommandBlock mcb(GC_LOOKAT);
/*
    if((fabs(p->bearing)>bearing)||
	p->type == Pronome::LOCATION){
	p->location = s->getVecInSensorFrame(p->location);
	p->location[0] = cos(bearing)*p->range;
	p->location[2] = sin(bearing)*p->range;
	p->location = s->getVecInWorldFrame(p->location);
    }
*/
    mcb.setNextArgGeneric(p->location[0]);
    mcb.setNextArgGeneric(p->location[1]);
    mcb.setNextArgGeneric(p->location[2]);
    mcb.setNextArgGeneric((float).25);
//    la->issueMotorCommand(*dba->creature,mcb);
    issueSuggestion(la,*dba->creature,mcb,SuggestedMCB::DO); 
}

extern float testbearing;

void followGradient(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(p, dba);
    PFSENSOR(s, dba);
    static float lastBearing = 0.0;
    
    LeafActivity *la = (LeafActivity *)act;

    float bearing = s->pf->weightedBearing;
    
    //float bearing = testbearing;
    float mg = s->pf->maxGrad;
    float sp = .05;
 //   printf("mg -> %f %f \n", mg, s->range);
    if((dba->creature->isMoving.getValue() == TRUE)||
	(s->range>9) || // was 9
	(fabs(bearing-lastBearing)<.01)){
	followGaitSuggestion(act, action);
    }
    lastBearing = bearing;

//    printf("minBearing -> %f maxGradient %f \n",bearing, mg);
    
    bearing = (bearing>1)?1 : bearing;
    bearing = (bearing< -1)?-1 : bearing;
	
    la->mcb.cmd = GC_LFT;
    float tmp = (float)(.5 -( bearing/2));
    la->mcb.setNextArgGeneric(tmp);
    sp = (fabs(mg)>1.0)?fabs(mg*mg)*.010 :.010;
    sp = (sp>.06)?.06 : sp;
    la->mcb.setNextArgGeneric(sp);
    
    gazeAtObjectOfInterest(bearing, act, action);
    
    
}

void turnTowardMinDensityCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
//    CHECKPRONOME(pronome);
    
    LeafActivity *la = (LeafActivity *)act;
    float bearing; 
    
    if(!dba->creature->checkIfReadyToMove(la->mcb)){
	gazeAtObjectOfInterest(pronome->bearing, act, action);
	return;
    }
    //followGaitSuggestion(act, action);
    // experimental gaze stuff
    s->updateGradient();
    bearing = s->pf->weightedBearing;
    
    pronome->location[0] = cos(.6*bearing)*30.0;
    pronome->location[1] = -5.0;
    pronome->location[2] = sin(.6*bearing)*30.0;
    
    pronome->location = s->getVecInWorldFrame(pronome->location);
    followGradient(act, action);
}

void approachEdgeMarkerCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb)){
	gazeAtObjectOfInterest(pronome->bearing, act, action);
	return;
    }
    
    if(pronome->status == Pronome::PASSEDFILTER){
	followGaitSuggestion(act, action);
	float bearing = s->markerEdgeBearing;
	if(s->closestEdge == s->markerMinX){
	    float delta = s->markerMinX - 54;
	    // if delta is < 0 then turn left
	    bearing = delta/10.0;
	}
	else {
	    float delta = 42 - s->markerMaxX;
	    // if delta is < 0 then turn right
	    bearing = delta/10.0;
	}
	bearing = (bearing>1)?1 : bearing;
        bearing = (bearing< -1)?-1 : bearing; 
	   
	la->mcb.cmd = GC_LFT;
	float tmp = (float)(.5 -( bearing/2));
	la->mcb.setNextArgGeneric(tmp);
	la->mcb.setNextArgGeneric((float).05);
    
	gazeAtObjectOfInterest(bearing, act, action);
	/* wag tail as a side - effect */
	MotorCommandBlock mb;
	mb.cmd = GC_WAGTAIL;
	la->issueMotorCommand(*dba->creature, mb);
    }
}

void approachMarkerCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb)){
	gazeAtObjectOfInterest(pronome->bearing, act, action);
	return;
    }
    
    if(pronome->status == Pronome::PASSEDFILTER){
	followGaitSuggestion(act, action);
	float bearing = pronome->bearing;
	bearing = (bearing>1)?1 : bearing;
	bearing = (bearing< -1)?-1 : bearing;
	    
	la->mcb.cmd = GC_LFT;
	float tmp = (float)(.5 -( bearing/2));
	la->mcb.setNextArgGeneric(tmp);
	la->mcb.setNextArgGeneric((float).025);
    
	gazeAtObjectOfInterest(bearing, act, action);
	/* wag tail as a side - effect */
	MotorCommandBlock mb;
	mb.cmd = GC_WAGTAIL;
	la->issueMotorCommand(*dba->creature, mb);
    }
}
void searchCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
 //   CHECKPRONOME(pronome);
    
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb)){
	gazeAtObjectOfInterest(pronome->bearing, act, action);
	return;
    }
    
    if(pronome->status == Pronome::PASSEDFILTER){
	// position schema
	//SbVec3f lpt = s->getVecInSensorFrame(pronome->location);
	float b, r;
	
	SbVec3f lpt = s->calcBearingAndRange(pronome->location, b, r);
//	printf("bearing before pf -> %f \n", b);
	int d = la->schema->depth;
	int w = la->schema->width;
	float schemaDist = (d/2 - s->pf->depth/2);
	lpt[1] = 0;
	if(pronome->range > schemaDist){
	    lpt.normalize();
	    lpt *= schemaDist;
	}
	
	s->pf->addPotentialFieldAtOffset((*la->schema->pfield), 
	    (int)lpt[0], (int)lpt[2]); 
	
	// update the gradient and follow it
	s->updateGradient();
	
	followGradient(act, action);

	/* wag tail as a side - effect */
	MotorCommandBlock mb;
	mb.cmd = GC_WAGTAIL;
	la->issueMotorCommand(*dba->creature, mb);
    }
}
void approachCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
 //   CHECKPRONOME(pronome);
    
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb)){
	gazeAtObjectOfInterest(pronome->bearing, act, action);
	return;
    }
    
    if(pronome->status == Pronome::PASSEDFILTER){
	// position schema
	SbVec3f lpt = s->getVecInSensorFrame(pronome->location);
	int d = la->schema->depth;
	int w = la->schema->width;
	float schemaDist = (d/2 - s->pf->depth/2);
	lpt[1] = 0;
	if(pronome->range > schemaDist){
	    lpt.normalize();
	    lpt *= schemaDist;
	}
	//
	float r, b;
	s->calcBearingAndRange(pronome->location, b, r);
	//printf("bearing and dist %f %f \n", b,r);
	if(r< 20.0 && (fabs(b)>1.57)){
	   lpt[2] *= -1;
	   pronome->location = s->getVecInWorldFrame(lpt);
	}

/*
	float r, b;
	static int persist = 0;
	s->calcBearingAndRange(pronome->location, b, r);
	printf("bearing and dist %f %f \n", b,r);
	if(r< 20.0 && ((fabs(b)> 1.5)||(persist>0))){
	    la->issueMotorCommand(*dba->creature, MotorCommandBlock(GC_BACK)); 
	    la->mcb.cmd = GC_STRAIGHT;
	    persist = (fabs(b)> 1.2)? 5 : persist-- ;
	}
	else 
*/
	{
	    s->pf->clear();
	    s->pf->addPotentialFieldAtOffset((*la->schema->pfield), 
		(int)lpt[0], (int)lpt[2]); 
	    
	    // update the gradient and follow it
	    s->updateGradient();
	    followGradient(act, action);
	}

	/* wag tail as a side - effect */
	MotorCommandBlock mb;
	mb.cmd = GC_WAGTAIL;
	la->issueMotorCommand(*dba->creature, mb);
    }
}
void fetchSuggCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    MotorCommandBlock mcb(GC_FWD, 
	ArgType((float)1.0), 
	ArgType((float)1.0),
	ArgType((float)dba->creature->defaultFwdSpeed.getValue()) );
    
    // issue suggestion
    dba->creature->controller->suggestMotorCommand(mcb, SuggestedMCB::MODIFY, 
        act, 100, FALSE);
	
   // issueSuggestion(act, *dba->creature,mcb, SuggestedMCB::MODIFY);
}

void fetchCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    fetchSuggCallback(act, action);
    searchCallback(act, action);
}

void leftGoAwayCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    pronome->location[0] -= 100;
    pronome->location[1] = 0;
    s->calcBearingAndRange(pronome->location, pronome->bearing, pronome->range);
    searchCallback(act, action);
}
void rightGoAwayCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    pronome->location[0] += 100;
    pronome->location[1] = 0;
    s->calcBearingAndRange(pronome->location, pronome->bearing, pronome->range);
    searchCallback(act, action);
}

void stopNearCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    
    LeafActivity *la = (LeafActivity *)act;
    la->mcb.cmd = GC_STOP;
    
    if(((Dog *)dba->creature)->speed == 0){
	MotorCommandBlock mb(GC_LIFTLEG);
	mb.setNextArgGeneric((float)((pronome->bearing <0.0)?0:1));
	la->issueMotorCommand(*dba->creature, mb);
    }
    gazeAtObjectOfInterest(pronome->bearing, act, action);
}

void crouchNearCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    la->mcb.cmd = GC_HALT;
    
    if(((Dog *)dba->creature)->speed == 0){
	MotorCommandBlock mb(GC_CROUCH);
	la->issueMotorCommand(*dba->creature, mb);
    }
    gazeAtObjectOfInterest(pronome->bearing, act, action);
}

void stopNearAndDoCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    
    if(((Dog *)dba->creature)->speed != 0)
	la->mcb.cmd = GC_HALT;
    else
	if(!dba->creature->checkIfNeedToChangePosture(la->mcb,la->defaultCmd))
	    la->mcb.cmd = la->defaultCmd;
	
    gazeAtObjectOfInterest(pronome->bearing, act, action);
}

void doCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    
    MotorCommandBlock mcb(GC_LFT);
    MotorSkill *ms = dba->creature->getMotorSkillByName("turning");
    if(ms){
	mcb.setNextArgGeneric(ms->counter);
	issueSuggestion(la,*dba->creature,mcb,SuggestedMCB::DO);
    }
    else {
	printf("motor skill not found \n");
    }
    if(!dba->creature->checkIfNeedToChangePosture(la->mcb,la->defaultCmd))
	la->mcb.cmd = la->defaultCmd;
    gazeAtObjectOfInterest(pronome->bearing, act, action);
}

void dropAndDoCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    dba->creature->controller->executeMotorCommand(MotorCommandBlock(GC_UNGRAB));
    doCallback(act, action);
}    
    
    
void grabCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
/*
    
    if(((Dog *)dba->creature)->speed != 0)
	la->mcb.cmd = GC_HALT;
    else
*/
    {
	la->mcb.cmd = GC_GRAB;
	la->mcb.setNextArgGeneric( (BaseCreature *)pronome->object);
	la->mcb.setNextArgGeneric( "MOUTH_JOINT");
	la->mcb.setNextArgGeneric((int) 1);
    }
    gazeAtObjectOfInterest(pronome->bearing, act, action);
}

void dropCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    
    if(((Dog *)dba->creature)->speed != 0)
	la->mcb.cmd = GC_HALT;
    else{
	la->mcb.cmd = GC_UNGRAB;
    }
    gazeAtObjectOfInterest(pronome->bearing, act, action);
    
}

void sleepCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(p, dba);
    PFSENSOR(s, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    
    if(((Dog *)dba->creature)->speed != 0)
	la->mcb.cmd = GC_HALT;
    else{
	la->mcb.cmd = GC_LIEDOWN;
	p->bearing = 0;
	p->type = Pronome::LOCATION;
	p->location = s->getVecInSensorFrame(p->location);
	p->location[0] += 60;
	p->location[1] = 10;
	p->location = s->getVecInWorldFrame(p->location);
    }
    gazeAtObjectOfInterest(p->bearing, act, action);
}

void shakePawCallback(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
//    CHECKPRONOME(pronome);
    LeafActivity *la = (LeafActivity *)act;
    la->defaultCmd = GC_SIT;
    stopNearAndDoCallback(act, action);
    // now figure out which leg to shake
    MotorCommandBlock mb(GC_SHAKEPAW,
			 ArgType((float)((pronome->bearing <0.0)?0:1)), 
			 ArgType((float).1));
    la->issueMotorCommand(*dba->creature, mb);
}
/*
 *	STEVE'S NEW CALLBACKS
 */
static void grabCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
  
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb))
	return;
    
    MotorCommandBlock mcb(GC_GRAB);
    mcb.setNextArgGeneric((BaseCreature *)pronome->object);
    mcb.setNextArgGeneric("MOUTH_JOINT");
    mcb.setNextArgGeneric(1);
    
    // new style of setting transform on grab
    SoTransform *xf = new SoTransform;
    xf->ref();
    static SoGetMatrixAction *gma = new SoGetMatrixAction(SbViewportRegion());
    xf->translation.setValue(SbVec3f(2,-2,10));
    xf->rotation.setValue(SbRotation(SbVec3f(1,0,0),-M_PI / 2));
    gma->apply(xf);
    SbMatrix sbm = gma->getMatrix();
    mcb.arglist.append(ArgType(sbm));
    
    // old style 
    //SoTransform *trans = new SoTransform;
    //trans->ref();
    //trans->rotation.setValue(SbVec3f(0,1,0),-M_PI / 2);
    //Values specific for Silas and the hamster.
    //trans->translation.setValue(1,-2,0);
    //mcb.setNextArgGeneric((void *)trans);
    
    la->issueMotorCommand(*dba->creature,mcb);
}

static void grabSimpleCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
  
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb))
	return;
    
    MotorCommandBlock mcb(GC_GRAB);
    mcb.setNextArgGeneric((BaseCreature *)pronome->object);
    mcb.setNextArgGeneric("MOUTH_JOINT");
    mcb.setNextArgGeneric(1);
    
    // new style of setting transform on grab
    SoTransform *xf = new SoTransform;
    xf->ref();
    static SoGetMatrixAction *gma = new SoGetMatrixAction(SbViewportRegion());
    xf->translation.setValue(SbVec3f(1,-2,0));
    xf->rotation.setValue(SbRotation(SbVec3f(0,1,0),-M_PI/2));
    gma->apply(xf);
    SbMatrix sbm = gma->getMatrix();
    mcb.arglist.append(ArgType(sbm));
    
    // old style 
    //SoTransform *trans = new SoTransform;
    //trans->ref();
    //trans->rotation.setValue(SbVec3f(0,1,0),-M_PI / 2);
    //Values specific for Silas and the hamster.
    //trans->translation.setValue(1,-2,0);
    //mcb.setNextArgGeneric((void *)trans);
    
    la->issueMotorCommand(*dba->creature,mcb);
}

static void testVelCB(BaseActivity *, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(p, dba);
  
  if (p && p->status == Pronome::PASSEDFILTER)
    {
      const float *vals = p->vel.getValue();
      
      printf("Test Vel: valid %d (%f,%f,%f)\n",p->validVel,
	     vals[0],vals[1],vals[2]);
    }
  else
    printf("Failed filter\n");
}

static void predictCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(p, dba);
  
  if((p->status == Pronome::PASSEDFILTER) &&
    (p->validVel)){
	p->location += p->vel;
  }
  searchCallback(act, action);
}

static void secondariesCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  
  LeafActivity *la = (LeafActivity *)act;
  
  float vbi = act->getValueBeforeInhibition();

  // always go forward
//  MotorCommandBlock mcb(GC_FWD);
//  mcb.setNextArgGeneric((float)(1.0));
  // choose gait
//  mcb.setNextArgGeneric((float)((vbi>.8)?0.0:0.0));
  // use standard speed
//  mcb.setNextArgGeneric((float)(dba->creature->defaultFwdSpeed.getValue()));
//  issueSuggestion(act, *dba->creature,mcb, SuggestedMCB::MODIFY);
    
  // issue secondary actions
  float wagspeed = (vbi>.5)?((vbi-.5)/.5)*.1 :0;

  MotorCommandBlock moveTail(GC_MOVETAIL, wagspeed, 
			     ArgType((float)1-vbi), 
			     ArgType((float).1));
  issueSuggestion(act, *dba->creature, moveTail, SuggestedMCB::DO);
  MotorCommandBlock moveEars(GC_MOVEEARS, 
			     ArgType((float).1), 
			     ArgType((float)1-vbi));  
  issueSuggestion(act, *dba->creature, moveEars, SuggestedMCB::DO);
  
  SniffSensor *sens = ((AliveDog *)(dba->creature))->getSniffSensor();
  assert(sens);
  // 875 == HAMSTER
  RayData *rd = sens->getDataForClosestWithTag(875);

  if (!rd)
    {
      fprintf(stderr,"Could not find 875(Hamster).\n");
      return;
    }
  if (!rd->who->isGrabbed.getValue())
    {
      if (rd->range <= 40)
	{
	  MotorCommandBlock lowerNeck(GC_LOWERNECK,
				      ArgType((float).5), 
				      ArgType((float)1.9));
	  issueSuggestion(act, *dba->creature,lowerNeck,SuggestedMCB::DO); 
	}
      else
	{
	  MotorCommandBlock lowerNeck(GC_LOWERNECK,
				      ArgType((float).5), 
				      ArgType((float)0));
	  issueSuggestion(act, *dba->creature,lowerNeck,SuggestedMCB::DO); 
	}
    }
  
}



static void shakeHamsterCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    LeafActivity *la = (LeafActivity *)act;


    if (((Dog *)dba->creature)->speed != 0)
      la->mcb.cmd = GC_HALT;
    else
      {
	MotorCommandBlock mcb(GC_LIEDOWN);
	la->issueMotorCommand(*dba->creature,mcb);
	
	MotorCommandBlock lowerNeck(GC_LOWERNECK,
				      ArgType((float).5), 
				      ArgType((float)0));
	la->issueMotorCommand(*dba->creature,lowerNeck);
	la->mcb.cmd = GC_SHAKEHEAD;
	la->mcb.setNextArgGeneric((float).8);
	la->mcb.setNextArgGeneric((float)0.0);
	
	dba->creature->controller->removeSuggestions(GC_LOOKAT);
	dba->creature->controller->removeSuggestions(GC_SHAKEHEAD);
      }
}

static void keepGrabbingCB(BaseActivity *, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  dba->creature->controller->executeMotorCommand(MotorCommandBlock(GC_GRAB));
}

static void keepGrabbingSuggCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    if(act->getValueBeforeInhibition()> 0.001)
	issueSuggestion(act,*dba->creature,MotorCommandBlock(GC_GRAB), SuggestedMCB::DO);
//    else 
//	issueSuggestion(act,*dba->creature,MotorCommandBlock(GC_UNGRAB), SuggestedMCB::DO);
}

static void suggestWalkSlowCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);

  MotorCommandBlock mcb(GC_FWD,ArgType((float)1.0),ArgType((float)0.0),
			ArgType((float)18));
  issueSuggestion(act, *dba->creature,mcb, SuggestedMCB::MODIFY);
}

static void suggestWalkCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);

  MotorCommandBlock mcb(GC_FWD,ArgType((float)1.0),ArgType((float)0.0),
			ArgType((float)(dba->creature->defaultFwdSpeed.
					getValue())));
  issueSuggestion(act, *dba->creature,mcb, SuggestedMCB::MODIFY);
}

static void suggestBoundCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);

  MotorCommandBlock mcb(GC_FWD,ArgType((float)1.0),ArgType((float)1.0),
			ArgType((float)5));
  issueSuggestion(act, *dba->creature,mcb, SuggestedMCB::MODIFY);
}

static void magicRotateCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(p, dba);
  LeafActivity *la = (LeafActivity *)act;
  
  SbVec3f loc = p->location;
  loc[1] = 11;
  la->mcb.cmd = GC_HALT;
  MotorCommandBlock mcb(GC_HEADTOWARD);
  mcb.setNextArgGeneric(loc[0]);
  mcb.setNextArgGeneric(loc[1]);
  mcb.setNextArgGeneric(loc[2]);
  issueSuggestion(act, *dba->creature, mcb, SuggestedMCB::DO);
}

void initDogActionCallbackList()
{
    BaseBehaviorNode::addCallback("search", searchCallback);
    BaseBehaviorNode::addCallback("approachMarker",approachMarkerCallback);
    BaseBehaviorNode::addCallback("approachEdgeMarker",approachEdgeMarkerCallback);
    BaseBehaviorNode::addCallback("simplewander", simpleWanderCallback);
    BaseBehaviorNode::addCallback("wander", wanderCallback);
    BaseBehaviorNode::addCallback("stopnear",stopNearCallback);
    BaseBehaviorNode::addCallback("stopNearAndDo",stopNearAndDoCallback);
    BaseBehaviorNode::addCallback("do",doCallback);
    BaseBehaviorNode::addCallback("dropAndDo",dropAndDoCallback);
    BaseBehaviorNode::addCallback("crouchnear",crouchNearCallback);
    BaseBehaviorNode::addCallback("turnaway",turnTowardMinDensityCallback);
    BaseBehaviorNode::addCallback("movetoschema",movetoSchemaCallback);
    BaseBehaviorNode::addCallback("closeschema",movetoSchemaCallback_1);
    BaseBehaviorNode::addCallback("gaitsuggestion",gaitSuggestionCallback);
    BaseBehaviorNode::addCallback("showAgression",aggSuggestionCallback);
    BaseBehaviorNode::addCallback("showGuilt",guiltSuggestionCallback);
    BaseBehaviorNode::addCallback("goawayleft",leftGoAwayCallback);
    BaseBehaviorNode::addCallback("goawayright",rightGoAwayCallback);
    BaseBehaviorNode::addCallback("stopnear",stopNearCallback);
    BaseBehaviorNode::addCallback("shakepaw",shakePawCallback);
    BaseBehaviorNode::addCallback("fetch",fetchCallback);
    BaseBehaviorNode::addCallback("fetchsugg",fetchSuggCallback);
//    BaseBehaviorNode::addCallback("grab",grabCallback);
    BaseBehaviorNode::addCallback("drop",dropCallback);
    BaseBehaviorNode::addCallback("approach",approachCallback);
    BaseBehaviorNode::addCallback("sleep",sleepCallback);
    BaseBehaviorNode::addCallback("lowerNeckCB",lowerNeckSuggestionCallback);
    BaseBehaviorNode::addCallback("turnaround",magicRotateCB);
    
      // hardts
    BaseBehaviorNode::addCallback("secondaries",secondariesCB);
    BaseBehaviorNode::addCallback("grab",grabCB);
    BaseBehaviorNode::addCallback("grabSimple",grabSimpleCB);
    BaseBehaviorNode::addCallback("testVel",testVelCB);
    BaseBehaviorNode::addCallback("predict",predictCB);
    BaseBehaviorNode::addCallback("shakeHamster",shakeHamsterCB);
    BaseBehaviorNode::addCallback("keepGrabbing",keepGrabbingCB);
    BaseBehaviorNode::addCallback("suggKeepGrabbing",keepGrabbingSuggCB);
    BaseBehaviorNode::addCallback("suggestWalkSlow",suggestWalkSlowCB);
    BaseBehaviorNode::addCallback("suggestWalk",suggestWalkCB);
    BaseBehaviorNode::addCallback("suggestBound",suggestBoundCB);
    BaseBehaviorNode::addCallback("suggestBark",suggestBarkCB);
    BaseBehaviorNode::addCallback("suggestPant",suggestPantCB);
    BaseBehaviorNode::addCallback("drinkSoundSuggestion",drinkSoundSuggestionCB);
    BaseBehaviorNode::addCallback("howl", howlCB);
    BaseBehaviorNode::addCallback("howlSound", howlSoundCB);
    BaseBehaviorNode::addCallback("whimperCB", whimperCB);
}




