#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoSelection.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/manips/SoTransformBoxManip.h>
#include <Inventor/SoPath.h>
#include <Inventor/SoOutput.h>
#include <Inventor/sensors/SoSensor.h>
#include <Inventor/sensors/SoIdleSensor.h>
#include <Inventor/sensors/SoTimerSensor.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>


#include "Alive.h"
#include "AliveCamera.h"
#include "VisionServices.h"
#include "Shadow.h"
#include "CreatureLib.h"
#include "DurationTimer.h"
#include "AliveClient.h"
#include "AliveServer.h"
#include "AliveDispatch.h"
#include "MotorCreature.h"

/***********************************************************************
 *
 *		Definition and Initialization of statics 
 *
 ***********************************************************************/
Alive *Alive::globalAlive = NULL;

float Alive::worldExpireTime = 5.0;
int Alive::chromakey = 0;
int Alive::gestureText = 0;
int Alive::useShadow = 0;
int Alive::connectToVision = 0;
int Alive::doPartialRestart = 0;
char *Alive::visionServerName = "blackfriars";
int Alive::shouldRestart = -1;
VisionServices *Alive::visionServices = NULL;
AliveCamera *Alive::aliveCamera = NULL;
DurationTimer *timerForFullReset = NULL;
float timeForFullReset = 3600;
SoTransformBoxManip *tbm;
int Alive::isDistributed = 0;
int Alive::iAmServer = 1;
float Alive::maxTimeBetweenRenders = 0.5;

/***********************************************************************
 *
 *		Call Backs 
 *
 ***********************************************************************/
 
// This function will be called by the Timer sensor step through a time step
static void myCallback( void *data, SoSensor *is )
{
    /* restart the world if necessary */
    if (Alive::shouldRestart == 1){
	printf("time since last restart %f \n",timerForFullReset->duration()); 
	if(timerForFullReset->duration()>timeForFullReset)
	    Alive::doPartialRestart = 0;
	Alive::globalAlive->cleanupAndRestart();
    }
    else{
	//Alive::globalAlive->creatureCache->updateStateAt(-1.0);
	//Alive::globalAlive->currentViewer->render();
    }	
}

void selectionCallback(
   void *, // user data is not used
   SoPath *selectionPath)
{
   // Attach the manipulator.
   // Use the convenience routine to get a path to
   // the transform that effects the selected object.
   SoSearchAction sa;
   sa.setType(BaseCreature::getClassTypeId(), TRUE);
   sa.apply(selectionPath->getHead());
   SoPath *xformPath = sa.getPath();
   if(xformPath != NULL){
	printf("something found \n");
	sa.setType(SoTransform::getClassTypeId(), TRUE);
	sa.setNode(((BaseCreature *)xformPath->getTail())->getBodyTransform());
	sa.apply(selectionPath->getHead());
	xformPath = sa.getPath();
	xformPath->ref();
	tbm->replaceNode(xformPath);
    }
	
}

void
distributedRenderCB(void *userData, SoSensor *sensor)
{
  Alive *alive = (Alive *) userData;
  int counter;
  static struct timeval lastRenderTime = {0, 0};
  struct timeval currentTime;
  float timediff;

  if (lastRenderTime.tv_sec == 0)
    gettimeofday(&lastRenderTime);

  gettimeofday(&currentTime);

  timediff = ((float) (currentTime.tv_sec - lastRenderTime.tv_sec)) + 
    (((float) (currentTime.tv_usec - lastRenderTime.tv_usec)) / 1000000.0);

  if (timediff > Alive::maxTimeBetweenRenders)
    {
      // force a render, since the last one occurred longer than maxTimeBetweenRenders ago
//      printf("forcing render due to timeout\n");
      alive->currentViewer->render();
      gettimeofday(&lastRenderTime);
      return;
    }
  else
    {
//      printf("timediff was %f\n", timediff);
    }

  alive->creatureCache->distributedCatchUpFlag = FALSE;

  for (counter = 0; counter < alive->creatureCache->creatures.getLength(); counter++)
    {
      if (((BaseCreature *) alive->creatureCache->creatures[counter])->isOfType(MotorCreature::getClassTypeId()))
	{
	  if ((((MotorCreature *) alive->creatureCache->creatures[counter])->motorCreatureType.getValue() == MotorCreature::DROID) &&
	      (((MotorCreature *) alive->creatureCache->creatures[counter])->remoteMCBList.getLength() > 0))
	    {
	      // this means that this creature still has more MCBs to apply from a remote master.
	      // skip the render for this clock tick.
	      alive->creatureCache->distributedCatchUpFlag = TRUE;
	      return;
	    }
	}
    }
  // do a render this tick
  alive->currentViewer->render();
  gettimeofday(&lastRenderTime);
}
  

/***********************************************************************
 *
 *		Alive Class Definition 
 *
 ***********************************************************************/

extern char **global_argv;  /* Defined in main.c++ */

void cleanup (...)
{	
    Alive::visionServices->abortConnection();
    exit(0);
}

void Alive::cleanupAndRestart()
{
    if(Alive::doPartialRestart){
	VisionServices::gesture->userGone();
	printf("doing partial restart \n");
	Alive::shouldRestart = -1;
	//Alive::globalAlive->creatureCache->restart();
#ifdef PFC
	Alive::visionServices->resetConnection();
#endif
    }
    else {
#ifndef PFC
	Alive::visionServices->resetConnection();
#endif
	printf("global_argv[0] = %s, [1] = %s.\n", global_argv[0], global_argv[1]);	/* DEBUG */
	execv("AliveApp", global_argv);
    }
}

Alive::Alive(Widget appWindow, char *camParamFile,  char *gestFile, char *aliveServerName)
{
    SoTimerSensor *mySensor = new SoTimerSensor( myCallback, this );
    mySensor->setInterval(1.0);
    mySensor->schedule();
    
    currentViewer = new SoXtExaminerViewer(appWindow);
    currentViewer->setSize(SbVec2s(640, 512));
    currentViewer->setHeadlight(0);
    currentViewer->setViewing(0);
    currentViewer->setDecoration(0);
    currentViewer->saveHomePosition();
    //currentViewer->setAutoRedraw(0);
    if(Alive::chromakey){
	currentViewer->setBackgroundColor(SbColor(1, .3, 1));
	//currentViewer->setBackgroundColor(SbColor(.6, .3, .8));
    }
    else
	currentViewer->setTransparencyType(SoGLRenderAction::BLEND);
    currentViewer->setAutoClipping(0);
   
    aliveInitialization(appWindow, camParamFile, gestFile);
    
    // now add creatureCache
    creatureCache = new CreatureCache( currentViewer);
    creatureCache->attachSceneGraph(cacheRoot);
    currentViewer->setSceneGraph(root);
    Alive::globalAlive = this;
    timerForFullReset = new DurationTimer();
    timerForFullReset->setTime();

    // Initialization of Distributed ALIVE if requested

    if (Alive::isDistributed)
      {
	if (Alive::iAmServer)
	  {
	    creatureCache->isDistributed = TRUE;
	    creatureCache->iAmServer = TRUE;
	    AliveServer *as = new AliveServer;
	    as->threadPacketForwarder();
	    AliveDispatch::setAliveServer(as);
	  }
	else
	  {
	    creatureCache->isDistributed = TRUE;
	    creatureCache->iAmServer = FALSE;
	    AliveClient *ac = new AliveClient;
	    ac->connectToServer(aliveServerName);
	    AliveDispatch::setAliveClient(ac);
	  }
	AliveDispatch::setCreatureCache(creatureCache);

	if (!Alive::iAmServer)
	  {
	    // synchronize my creature database with the server's.

	    if (AliveDispatch::synchronizeCreatureDatabaseWithServer() < 0)
	      {
		cerr << "Creature database synchronization was unsuccessful!" << endl;
	      }
	  }

	// Code which allows individual machines in Distributed ALIVE
	// to render at different rates

	SoNodeSensor *renderSensor = new SoNodeSensor(distributedRenderCB, this);
	currentViewer->setAutoRedraw(FALSE);
	renderSensor->setPriority(currentViewer->getDefaultRedrawPriority() - 1);
	renderSensor->attach(root);  // currentViewer's scene graph
      }
}

void Alive::aliveInitialization(Widget myWindow, char *camParamFile, 
    char *gestFile)
{
    // Build enabled window    
    
    static int attributeList[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
	GLX_GREEN_SIZE, 1,
	GLX_BLUE_SIZE, 1,
	GLX_DEPTH_SIZE, 1,
	GLX_DOUBLEBUFFER,
//	GLX_STENCIL_SIZE, 1, 
//	GLX_ACCUM_RED_SIZE,  4,
//	GLX_ACCUM_GREEN_SIZE,  4,
//	GLX_ACCUM_BLUE_SIZE,  4,
	None,
    };

    static int attributeList_chroma[] = {
	GLX_RGBA,
	GLX_RED_SIZE, 1,
	GLX_GREEN_SIZE, 1,
	GLX_BLUE_SIZE, 1,
	GLX_DEPTH_SIZE, 1,
	GLX_DOUBLEBUFFER,
	None,
    };
    
    XVisualInfo *vis = NULL;
    
    vis = (Alive::chromakey)
	?AliveCamera::buildEnabledVisual(attributeList_chroma)
	:AliveCamera::buildEnabledVisual(attributeList);

    
    if(vis)
	currentViewer->setNormalVisual(vis);
    currentViewer->setAutoClipping(0);
   
    if(!Alive::chromakey)
   	currentViewer->setTransparencyType(SoGLRenderAction::BLEND);
    
    // now build default scenegraph 
    root = new SoSelection;
    root->ref();
//    ((SoSelection *)root)->addSelectionCallback(selectionCallback, NULL);
//    selectionRoot->
//      addDeselectionCallback(deselectionCallback, NULL);
//    root = new SoSeparator();

    tbm = new SoTransformBoxManip;
    tbm->ref();
    
    // Camera 
    SoPerspectiveCamera *graphicsCamera = new SoPerspectiveCamera;
    Alive::aliveCamera = new AliveCamera(graphicsCamera, currentViewer);
    Alive::aliveCamera->readCameraParametersFromFile(camParamFile);
    root->addChild(graphicsCamera);
    

    // lights
    root->addChild(new SoDirectionalLight);
    SoDirectionalLight *sdl = new SoDirectionalLight;
    SbVec3f dir;
    dir.setValue(.6, -.6, 0);
    dir.normalize();
    sdl->direction.setValue(dir);
    sdl->intensity = .8;
    root->addChild(sdl);
    
    // Start VisionServices and add person contour to scenegraph 
    if(Alive::connectToVision){
	Alive::visionServices = new VisionServices(
	    Alive::visionServerName, 
	    gestFile,
	    root, 
	    Alive::aliveCamera, 
	    0); 
	signal(SIGINT,cleanup);
//	Alive::visionServices->attachContourToSceneGraph(root);
    }
    
    // add separator for CreatureCache and Shadow if needed 
    cacheRoot = new SoSeparator();
    if(Alive::useShadow){
	SbVec4f shadColor = (Alive::chromakey)? SbVec4f(0, 1, 0,0) :
		SbVec4f(0, 1, 0,.1);
	ShadowNode *shad = new ShadowNode(SbVec3f(.4, -.6, .2),shadColor, cacheRoot);
	root->addChild(shad->sep);
    }
    else   
	root->addChild(cacheRoot);
}
