#ifndef ALIVECAM
#define ALIVECAM

#include <Inventor/SbLinear.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SbPList.h>
#include <Inventor/SbLinear.h>

class BackProject {
    public:
    SbMatrix spm_static;
    SbMatrix spm1;
    SoTransform *xf;
    SbVec3f cp;
    SbMatrix matrix;
    SbMatrix m2;
    SbPlane plane;
    SbMatrix spm;
    SbMatrix proj;
    float planeZ;
    BackProject(SoPerspectiveCamera *cam);
    void setProjectionPlane(float z);
    void backProjectPt(const SbVec3f &sp1, SbVec3f &projPt, float z);
};

class AliveCamera {
    public:
    SoCamera *camera;
    SoXtRenderArea *renderArea;
    SoSeparator *markers;
    float vXwScale;
    BackProject *bp;
    int showMouseMarkers;
   
    
    static XVisualInfo *buildEnabledVisual(int *attributeList);
    static void AliveCameraCB(void *userData, SoEventCallback *eventCB);
    
    AliveCamera(SoCamera *cam, SoXtRenderArea *renderArea);
    AliveCamera(SoCamera *cam, SoXtRenderArea *renderArea, 
	SbVec3f camPos,SbMatrix camO,float vscale);
    void commonInit(SoCamera *cam, SoXtRenderArea *renderArea);
    void readCameraParametersFromFile(char *fn);
    void printCameraParameters();
    void convertToLUV(SbColor color);
    void adjustCamera(SbVec3f disp,SbMatrix rm, float vscale, SbVec2s size);
    void setShowMouseMarkers(int yn, SoSeparator *where);
    void addMarker(SbVec2f xy);
    void projectYtoZFloor(SbVec2f pt2d,SbVec3f &pt3d);
    void projectYtoZFloor(float *pt2df, float *pt3df);
    void projectScreenYtoWorldY(SbVec2f pt2d,SbVec3f &pt3d,float z);
    void projectScreenYtoWorldY(float *pts,int npts, int stride, SbVec3f *pt3d,float z);
    void projectScreenYtoWorldY(float *pts,int npts,int stride,  
	float *pt3ds,float z);
    void projectToScreen(SbVec3f &wpt2d,SbVec3f &spt3d);
    void trackMouseEvents();
};


#endif
