/*
 * RCS $Header: /tmp_mnt/mas/disks/vm20/cwren/projects/Creature2.1/AliveLib/RCS/Avatar.c++,v 1.5 1995/10/31 17:54:36 bruce Exp $
 *
 * Avatar.c++
 *
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <Inventor/SoDB.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoOutput.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/nodes/SoNodes.h>
#include "image.h"
#include "Avatar.h"
#include "CreatureLib.h"
#if MALLOC
#include "dmalloc.h"
#endif
#include "math.h"
const float RADTODEG(float rad) {
      return (fmodf((rad)/(2*M_PI),1.0)*360);
}

const float DEGTORAD(float deg) {
      return (fmodf((deg)/360,1.0)*2*M_PI);
}

static const float INITIAL_AVATAR_SCALE = 45.0;

static const float AVATAR_ONE_OVER_ASPECT_RATIO = 0.75;

static const int MAX_FACES = 5;
static const int NUM_INDICES = 10;
static const int LEFT_ARM_INDEX = 0;
static const int RIGHT_ARM_INDEX = 1;

static const int AB_GRP_XF_INDEX = 0;
static const int AB_GRP_ARM_INDEX = 1;

static const int LEFT_ARM_SIGN = -1;
static const int RIGHT_ARM_SIGN = 1;

static const float LIMB_RADIUS = 0.01;
  
#define LEFT_START0 0 
#define LEFT_MID0 1
#define LEFT_TIP 2
#define LEFT_MID1 3
#define LEFT_START1 4
#define RIGHT_START0 5
#define RIGHT_MID0 6
#define RIGHT_TIP 7
#define RIGHT_MID1 8
#define RIGHT_START1 9


void
AvatarInit()
{
  AvatarFrameInfo::initClass();
  Avatar::initClass();
}


SO_NODE_SOURCE(AvatarFrameInfo);

AvatarFrameInfo::AvatarFrameInfo() : SoNode()
{
  SO_NODE_CONSTRUCTOR(AvatarFrameInfo);

  SO_NODE_ADD_FIELD(annotationIndices, (0));
  SO_NODE_ADD_FIELD(useLeftArm, (FALSE));
  SO_NODE_ADD_FIELD(useRightArm, (FALSE));
  SO_NODE_ADD_FIELD(leftShoulderPos, (0, 0, 0));
  SO_NODE_ADD_FIELD(rightShoulderPos, (0, 0, 0));
  SO_NODE_ADD_FIELD(leftElbowPos, (0, 0, 0));
  SO_NODE_ADD_FIELD(rightElbowPos, (0, 0, 0));
  SO_NODE_ADD_FIELD(leftEndPos, (0, 0, 0));
  SO_NODE_ADD_FIELD(rightEndPos, (0, 0, 0));

  SO_NODE_ADD_FIELD(leftAngles, (0, 0));
  SO_NODE_ADD_FIELD(rightAngles, (0, 0));
  SO_NODE_ADD_FIELD(leftShoulderRange, (3.14, 1.33)); // center, dist
  SO_NODE_ADD_FIELD(leftElbowRange, (-0.78, 1.57));    // center, dist
  SO_NODE_ADD_FIELD(rightShoulderRange, (0, 1.33));   // center, dist
  SO_NODE_ADD_FIELD(rightElbowRange, (0.78, 1.57));   // center, dist


  SO_NODE_ADD_FIELD(centroid, (0, 0, 0));
  SO_NODE_ADD_FIELD(head, (0, 0, 0));
  SO_NODE_ADD_FIELD(leftHand, (0, 0, 0));
  SO_NODE_ADD_FIELD(rightHand, (0, 0, 0));

  SO_NODE_ADD_FIELD(texscale, (1, 1));
  SO_NODE_ADD_FIELD(texoffset, (0, 0));

}

AvatarFrameInfo::~AvatarFrameInfo()
{
}

void
AvatarFrameInfo::initClass()
{
  SO_NODE_INIT_CLASS(AvatarFrameInfo, SoNode, "SoNode");
}

SbBool
#ifdef SO_VERSION_REVISION
AvatarFrameInfo::readInstance(SoInput *in, unsigned short flags)
#else
AvatarFrameInfo::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoNode::readInstance(in, flags)) {
#else
    if (SoNode::readInstance(in)) {
#endif
    return TRUE;
    }
  return FALSE;
}


AvatarCache::AvatarCache()
{
  afi = NULL;
  originalCoords = NULL;
  newCoords = NULL;
  leftABXform = NULL;
  rightABXform = NULL;
  faceSet = NULL;
  leftArm = NULL;
  rightArm = NULL;
}

AvatarCache::~AvatarCache()
{
}


SO_NODE_SOURCE(Avatar);

Avatar::Avatar() : SoSeparator()
{
  SO_NODE_CONSTRUCTOR(Avatar);

  SO_NODE_ADD_FIELD(sceneGraphFilename, ("none"));
  SO_NODE_ADD_FIELD(avatarStandingFrameCount, (0));
  SO_NODE_ADD_FIELD(avatarNormalHeight, (0));

  avatarScale = NULL;
  originalCoordinateSwitch = NULL;
  frameSwitch = NULL;
  currentFrame = NULL;

  currentFrame = new AvatarCache;

  doSelect = doWarp = 1;
  annoEditMode = 0;
}

Avatar::~Avatar()
{
}

void
Avatar::initClass()
{
  SO_NODE_INIT_CLASS(Avatar, SoSeparator, "SoSeparator");
}

SoPath *
Avatar::getPathToName(SoNode *from, const SbName &name)
{
    SoNode *namedNode = getByName(name);
    SoPath *pathToName = NULL;
    
    if(namedNode){
	SoSearchAction sa= SoSearchAction();
	sa.setNode(namedNode);
	sa.apply(from);
	pathToName = sa.getPath();
	if (pathToName) pathToName->ref();
    }
    return pathToName;
}

SbBool Avatar::writeToIVFile(char *scenefilename) {
  SoOutput out;
  if (out.openFile(scenefilename) == FALSE)
    {
      cerr << "Error writing scene graph!" << endl;
      return FALSE;
    }
  else
    {
      SoWriteAction wa(&out);

      annoEditMode = 0;
      buildFaceSetsFromAnnotations();

      restoreAllCoords();
      originalCoordinateSwitch->ref();
      removeChild(originalCoordinateSwitch);
      wa.apply( this );
      insertChild(originalCoordinateSwitch,1);
      originalCoordinateSwitch->unref();
    }
  out.closeFile();
  return TRUE;
}

int
Avatar::getNearestContourPoint(SbVec3f &v) 
{
  // returns index of closest contour point to v.
  // sets v equal to coordinates of this point.

  double minv; 
  int mini;
  int32_t s = frameSwitch->whichChild.getValue();
  SoCoordinate3 *contour = (SoCoordinate3 *) originalCoordinateSwitch->getChild(s);
  int n = contour->point.getNum();
  for (int i = 0; i < n; i++) {
    SbVec3f contour_vert = contour->point[i];
    SbVec3f tmp = v - contour_vert;
    double d = tmp.length();
    
    if (i == 0 || d < minv) minv = d, mini = i;
  }
  v = contour->point[mini];
  return mini;
}


void Avatar::restoreAllCoords() {
  int c1;

  for (c1 = 0; c1 < frameSwitch->getNumChildren(); c1++)
    {
	  currentFrame->newCoords->copyFieldValues(currentFrame->originalCoords);
	  resetArmAngles();
    }
}

int
Avatar::lcmp(const void *a, const void *b) {
  int32_t *ia = (int32_t *)a, *ib = (int32_t *)b;
  if (*ia==*ib) return 0;
  if (*ia<*ib) return -1;
  return 1;
}

void
Avatar::sortAnnotationIndices()
{
  int c1;

  for (c1 = 0; c1 < frameSwitch->getNumChildren(); c1++)
    {
      switchToFrame(c1);
      int32_t *arm_indices = currentFrame->afi->annotationIndices.startEditing();
      qsort((void *)arm_indices,NUM_INDICES,sizeof(int32_t),lcmp);
      currentFrame->afi->annotationIndices.finishEditing();
    }
}

void Avatar::buildFaceSetsFromAnnotations() { 
  int32_t child = frameSwitch->whichChild.getValue();
  int c1;
  
  for (c1 = 0; c1 < frameSwitch->getNumChildren(); c1++) {
      switchToFrame(c1);

      SoIndexedFaceSet *fs = currentFrame->faceSet;
      int i,ni;
      int numIndices = currentFrame->originalCoords->point.getNum();
      int32_t *idx = new int32_t[numIndices * MAX_FACES];
      ni=0;
      if (annoEditMode) {
	    // single face: for degugging when indicies are shot...
	    for (i=0;
		 i<numIndices;
		 idx[ni++] = i++) ;
	    idx[ni++] = SO_END_FACE_INDEX;
      } else {
	    // BODY FACE:
	    // Four cases: both, left, right, none.
	    if (!currentFrame->afi->useLeftArm.getValue() &&
		!currentFrame->afi->useRightArm.getValue()) {
		// NONE: single face
		for (i=0;
		     i<numIndices;
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    } else if (currentFrame->afi->useLeftArm.getValue() &&
		       currentFrame->afi->useRightArm.getValue()) {
		// BOTH: 
		// body: 0...(left start0),(left start1),
		// (right start0),(right start1)...n
		for (i=0;
		     i<=(int)currentFrame->afi->annotationIndices[LEFT_START0];
		     idx[ni++] = i++) ;
		for (i=(int)currentFrame->afi->annotationIndices[LEFT_START1];
		     i<=(int)currentFrame->afi->annotationIndices[RIGHT_START0];
		     idx[ni++] = i++) ;
		for (i=(int)currentFrame->afi->annotationIndices[RIGHT_START1];
		     i<numIndices;
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    } else if (currentFrame->afi->useLeftArm.getValue() &&
		       !currentFrame->afi->useRightArm.getValue()) {
		// body face when only left arm used
		for (i=0;
		     i<=(int)currentFrame->afi->annotationIndices[LEFT_START0];
		     idx[ni++] = i++) ;
		for (i=(int)currentFrame->afi->annotationIndices[LEFT_START1];
		     i<numIndices;
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    } else if (!currentFrame->afi->useLeftArm.getValue() &&
		       currentFrame->afi->useRightArm.getValue()) {
		// body face when only right arm used
		for (i=0;
		     i<=(int)currentFrame->afi->annotationIndices[RIGHT_START0];
		     idx[ni++] = i++) ;
		for (i=(int)currentFrame->afi->annotationIndices[RIGHT_START1];
		     i<numIndices;
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    }

	    // LEFT ARM FACES:
	    if (currentFrame->afi->useLeftArm.getValue()) {
		// left upper arm: (left start0),...,(left mid0)
		// (left mid1)...(left start1)
		for (i=(int)currentFrame->afi->annotationIndices[LEFT_START0];
		     i<=(int)currentFrame->afi->annotationIndices[LEFT_MID0];
		     idx[ni++] = i++) ;
		for (i=(int)currentFrame->afi->annotationIndices[LEFT_MID1];
		     i<=(int)currentFrame->afi->annotationIndices[LEFT_START1];
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    
		// left lower arm: (left mid0),...,(left mid1)
		for (i=(int)currentFrame->afi->annotationIndices[LEFT_MID0];
		     i<=(int)currentFrame->afi->annotationIndices[LEFT_MID1];
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    }

	    if (currentFrame->afi->useRightArm.getValue()) {
		// right upper arm: (right start0),...,(right mid0),
		// (right mid1)...(right start1)
		for (i=(int)currentFrame->afi->annotationIndices[RIGHT_START0];
		     i<=(int)currentFrame->afi->annotationIndices[RIGHT_MID0];
		     idx[ni++] = i++) ;
		for (i=(int)currentFrame->afi->annotationIndices[RIGHT_MID1];
		     i<=(int)currentFrame->afi->annotationIndices[RIGHT_START1];
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;

		// right lower arm: (right mid0),...,(right mid1)
		for (i=(int)currentFrame->afi->annotationIndices[RIGHT_MID0];
		     i<=(int)currentFrame->afi->annotationIndices[RIGHT_MID1];
		     idx[ni++] = i++) ;
		idx[ni++] = SO_END_FACE_INDEX;
	    }
      }
      if (ni>=MAX_FACES * numIndices) cerr<<"MAXIDX overflow!"<<endl;
      fs->coordIndex.setNum(ni);
      fs->coordIndex.setValues(0,ni,idx);
      delete[] idx;
    }
  switchToFrame(child);
}

void Avatar::buildCaches() {
      int c1;
      
      for (c1 = 0; c1 < frameSwitch->getNumChildren(); c1++) {
	    
	    AvatarCache *cache = new AvatarCache;
	    cache->originalCoords = (SoCoordinate3 *) originalCoordinateSwitch->getChild(c1);

	    SoGroup *grp = (SoGroup *) frameSwitch->getChild(c1);

	    cache->newCoords = (SoCoordinate3 *) getNthChildOfType(grp, 0, SoCoordinate3::getClassTypeId());
	    cache->texCoords = (SoTextureCoordinate2 *) getNthChildOfType(grp, 0, SoTextureCoordinate2::getClassTypeId());
	    cache->faceSet = (SoIndexedFaceSet *) getNthChildOfType(grp, 0, SoIndexedFaceSet::getClassTypeId());
	    cache->leftABXform = (SoTransform *) ((SoSeparator *) getNthChildOfType(grp, LEFT_ARM_INDEX,SoSeparator::getClassTypeId()))->getChild(AB_GRP_XF_INDEX);
	    cache->rightABXform = (SoTransform *) ((SoSeparator *) getNthChildOfType(grp, RIGHT_ARM_INDEX,SoSeparator::getClassTypeId()))->getChild(AB_GRP_XF_INDEX);
	    cache->leftArm = (ArticulatedBody *) ((SoSeparator *) getNthChildOfType(grp, LEFT_ARM_INDEX, SoSeparator::getClassTypeId()))->getChild(AB_GRP_ARM_INDEX);
	    cache->rightArm = (ArticulatedBody *) ((SoSeparator *) getNthChildOfType(grp, RIGHT_ARM_INDEX, SoSeparator::getClassTypeId()))->getChild(AB_GRP_ARM_INDEX);
	    cache->afi = (AvatarFrameInfo *) getNthChildOfType(grp, 0, AvatarFrameInfo::getClassTypeId());




	    frameCaches.append(cache);
      }
}

AvatarCache *Avatar::frame(int i) { return (AvatarCache *)frameCaches[i]; }

void
Avatar::buildArms()
{
  int c1;

  for (c1 = 0; c1 < frameSwitch->getNumChildren(); c1++)
    {
      switchToFrame(c1);
      buildLeftArm();
      buildRightArm();
    }
}

SbBool
#ifdef SO_VERSION_REVISION
Avatar::readInstance(SoInput *in, unsigned short flags)
#else
Avatar::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoSeparator::readInstance(in, flags)) {
#else
    if (SoSeparator::readInstance(in)) {
#endif
      if (sceneGraphFilename.getValue() != SbString("none"))
	{
	  if (getNumChildren() > 0)
	    {
	      cerr << "Avatar::readInstance: can't set the scene graph file name when there's a scene in the .iv file!" << endl;
	      return FALSE;
	    }
	  else
	    {
	      // read in from another file.
	      // This is like the case where you specify the avatar file name in the BehaviorPerson's .iv file
	      // and actually store the avatar's information in another file (to prevent cutting and pasting).
	      
	      SoNode *node;
	      SoInput otherInput;
	      if (otherInput.openFile(sceneGraphFilename.getValue().getString()) == FALSE)
		{
		  cerr << "Avatar::readInstance: Error reading file " << sceneGraphFilename.getValue().getString() << endl;
		  return FALSE;
		}
	      else
		{
		  // read in into node "node"
		  node = SoDB::readAll(&otherInput);
		  otherInput.closeFile();

		  if (node->isOfType(Avatar::getClassTypeId()))
		    {
		      Avatar *av = (Avatar *) node;
		      // copy all children of "node" into "this"
		      while (av->getNumChildren() > 0)
			{
			  SoNode *tempNode = av->getChild(0);
			  addChild(tempNode);
			  av->removeChild(tempNode);
			}
		      // copy all fields over as well
		      copyFieldValues(av);
		    }
		  else
		    {
		      cerr << "Avatar::readInstance: File " << sceneGraphFilename.getValue().getString() <<
			" does not contain an avatar" << endl;
		      return FALSE;
		    }
		}
	    }
	}

      // set up pointers
      SoPath *tmp_path = NULL;

      if (tmp_path = getPathToName(this, "FRAME_SWITCH")) {
	    frameSwitch = (SoSwitch *)tmp_path->getTail();
      } else {
	    cerr << "Failed to get path to FRAME_SWITCH!" << endl;
      }

      if (tmp_path = getPathToName(this, "AVATAR_SCALE")) {
	    avatarScale = (SoScale *)tmp_path->getTail();
      } else {
	    cerr << "Failed to get path to AVATAR_SCALE!" << endl;
      }

      if (tmp_path = getPathToName(this, "TAIL_SEP")) {
	    tailSep = (SoSeparator *)tmp_path->getTail();
	    cerr << "Found a tail on this fish" << endl;
      } else {
	    tailSep = NULL;
      }

      originalCoordinateSwitch = new SoSwitch;
      originalCoordinateSwitch->setName(SbName("ORIGINAL_COORDINATE_SWITCH"));
      insertChild(originalCoordinateSwitch,1);

      for (int c1 = 0; c1 < frameSwitch->getNumChildren(); c1++) {
	    SoGroup *grp = (SoGroup *) frameSwitch->getChild(c1);
	    SoCoordinate3 *coords = new SoCoordinate3;
	    coords->copyFieldValues((SoCoordinate3 *) getNthChildOfType(grp, 0, SoCoordinate3::getClassTypeId()));
	    originalCoordinateSwitch->addChild(coords);
      }

      buildCaches();

      sortAnnotationIndices();
      buildFaceSetsFromAnnotations();

      buildArms();
      return TRUE;
    }
  else
    return FALSE;
}


void
Avatar::combineRawImageWithSegmentation(unsigned char *dumpbuf, unsigned char *segbuf, unsigned short *outbuf) {
  int c1, c2, c3;


  int usesegbuf = (seg_filename && *seg_filename != ':');
  int threshalpha = (seg_filename && !strcmp(seg_filename,":thresh"));

  printf(" usesegbuf = %d threshalpha = %d \n", usesegbuf, threshalpha);

  // RGB
  for (c1 = 0; c1 < 3; c1++) {
	for (c2 = 0; c2 < image_height; c2++) {
	      for (c3 = 0; c3 < image_width; c3++) {
		    outbuf[(image_width*image_height*c1) + 
			   (image_width*c2) + (c3)] =
#define FLIPV
#ifdef FLIPV			  
			   dumpbuf[(3*image_width*(image_height-c2-1))+(3*c3)+c1];
#else
			   dumpbuf[(3*image_width*c2)+(3*c3)+c1];
#endif
	      } 
	}
  }

  // ALPHA CHANNEL
   for (c2 = 0; c2 < image_height; c2++) {
	for (c3 = 0; c3 < image_width; c3++) {
	      int val;
	      if (usesegbuf) {
		    val =  segbuf[(image_width * c2) + (c3)];
	      } else {
		    int r = dumpbuf[(3*image_width*c2)+(3*c3)];
		    int g = dumpbuf[(3*image_width*c2)+(3*c3)+1];
		    int b = dumpbuf[(3*image_width*c2)+(3*c3)+2];
		    
		    if (/* r == 0 && g == 0 && b == 255*/ 
			(float)b/(float)(r+g) > 0.75 && b>100) {
			  val = 0;
		    } else {
			  if (threshalpha) {
				val = 255;
			  } else {
				val = (r + g + b) / 3;
			  }
		    }
	      }
	      outbuf[(image_width*image_height*3) + 
#ifdef FLIPV
		     (image_width * (image_height-c2-1)) + (c3)] = val;
#else
		     (image_width * c2) + (c3)] = val;
#endif
	}
  }
}

void
Avatar::convertPfinderRawImagesToTextures()
{
  FILE *dfp;
  FILE *sfp;
  IMAGE *outimage;
  int width = image_width;
  int height = image_height;
  int framecount = 0;
  char rgboutfilename[256];

  unsigned char *dumpbuf = new unsigned char[width * height * 3];
  unsigned char *segbuf = new unsigned char[width * height];
  unsigned short *outbuf = new unsigned short[width*height*4];
  
  if ((dfp = fopen(raw_filename, "r")) == NULL)
    {
      cerr << "Avatar::convertPfinderRawImagesToTextures: Error opening raw image file" << endl;
      return;
    }

  if (seg_filename && *seg_filename != ':') {
	if ((sfp = fopen(seg_filename, "r")) == NULL) {
	      cerr << "Avatar::convertPfinderRawImagesToTextures: Error opening segmentation file" << endl;
	      exit(1);
	}
  } else {
	cerr << "Using chroma key to compute seg" << endl;
  }

  int num_bytes_read;

  while (1)
    {
      if (num_frames && (framecount == num_frames)) {
	    // DONE
	  delete[] dumpbuf;
	  delete[] segbuf;
	  delete[] outbuf;
	  return;
      } // if num_frames is 0, then read as many full frames as possible

      if ((num_bytes_read = fread(dumpbuf, sizeof(char), width * height * 3, dfp)) < (width * height * 3))
	{
	  if (num_frames == 0 && num_bytes_read != 0)
	    cerr << "Avatar::convertPfinderRawImagesToTextures: Got unexpected EOF of raw image file at frame " << framecount << endl;
	  num_frames = framecount;
	  delete[] dumpbuf;
	  delete[] segbuf;
	  delete[] outbuf;
	  return;
	}

      if (seg_filename && *seg_filename != ':') {
	    if ((num_bytes_read = fread(segbuf, sizeof(char), width*height, sfp)) < (width*height)) {
		  if (num_bytes_read != 0)
		    cerr << "Avatar::convertPfinderRawImagesToTextures: Got unexpected EOF of segmentation file at frame " << framecount << endl;
		  delete[] dumpbuf;
		  delete[] segbuf;
		  delete[] outbuf;
		  return;
	    }
      }

      // combine dumpbuf and segbuf into outbuf
      combineRawImageWithSegmentation(dumpbuf, segbuf, outbuf);

      /// adds random header as number of frames increases.

      sprintf(rgboutfilename, "%s-%d.rgb", output_base_filename, framecount);

      if ((outimage = iopen(rgboutfilename, "w", RLE(1), 4, width, height, 4)) == NULL)
	{
	  cerr << "Error opening output image file" << endl;
	  exit(1);
	}

      for (int c1 = 0; c1 < 4; c1++)
	{
	  for (int c2 = 0; c2 < height; c2++)
	    {
	      putrow(outimage, &(outbuf[(width*height*c1)+(width*c2)]), height - c2 - 1, c1);
	    }
	}
      iclose(outimage);
      cerr << "Converted frame " << rgboutfilename << endl;

      framecount++;
    }
}


SbVec3f
Avatar::setGoal(ArticulatedBody *b, SbVec3f g)
{
  b->moveToConstrainedXYGoal(g);
  SbVec3f endPt;
  b->getEndPoint(endPt);
  return endPt;
}

void
Avatar::removeLimbs(ArticulatedBody *b)
{ 
      b->rootJoint->getLimb()->removeAllChildren();
      b->rootJoint->getNextJoint()->getLimb()->removeAllChildren();
}
void
Avatar::setLengths(ArticulatedBody *b, float l1, float l2)
{ 
  b->rootJoint->getNextJoint()->getInner()->translation.setValue(l1,0,0);
  b->rootJoint->getLimb()->setLengthRadius(l1,LIMB_RADIUS);
  
  b->rootJoint->getNextJoint()->getNextJoint()->
    getInner()->translation.setValue(l2,0,0);
  b->rootJoint->getNextJoint()->getLimb()->setLengthRadius(l2,LIMB_RADIUS);
}

void
Avatar::getLengths(ArticulatedBody *b, float &l1, float &l2) { 
      l1 = b->rootJoint->getNextJoint()->getInner()->translation.getValue()[0];
      l2 = b->rootJoint->getNextJoint()->
	getNextJoint()->getInner()->translation.getValue()[0];
}

void Avatar::getAngles(ArticulatedBody *b, float &a1, float &a2) { 
      a1 = b->rootJoint->angle.getValue();
      a2 = b->rootJoint->getNextJoint()->angle.getValue();
}
void
Avatar::setAngles(ArticulatedBody *b, float a1, float a2) { 
      static const SHOULDER_CENTER_GAIN = 0.2;

      b->rootJoint->setAngleImmediate(a1);
      b->rootJoint->center.setValue(a1);
      b->rootJoint->centerGain = SHOULDER_CENTER_GAIN;
      b->rootJoint->getNextJoint()->setAngleImmediate(a2);
}

void
Avatar::setElbowCenter(ArticulatedBody *b, float a)
{ 
  static const ELBOW_CENTER_GAIN = 0.8;
  b->rootJoint->getNextJoint()->center = a;
  b->rootJoint->getNextJoint()->centerGain = ELBOW_CENTER_GAIN;
}

SbVec3f
Avatar::getMid(ArticulatedBody *b) { 
  float ang = b->rootJoint->angle.getValue();
  SbVec3f t = b->rootJoint->getNextJoint()->getInner()->translation.getValue();

  float r = t[0];
  float x = r * fcos(ang);
  float y = r * fsin(ang);
  return SbVec3f(x,y,0);
}

//
// simple inverse kinematics for 2 link arm case.

SbBool
Avatar::inverse2DOF(ArticulatedBody *b, SbVec3f g, int sign, float &a1, float &a2) {  

      // note g already in shoulder relative coordinates

      // geometric solution to 2 DOF manipulator. 
      // From Craig, p. 110

      static const float ARM_SLOP = 0.0001;

      g[2]=0; // force solution in the plane

      float l1,l2;

      getLengths(b, l1, l2);

      float arm_max = l1 + l2 - ARM_SLOP;
      if (g.length() > arm_max) {
	    g.normalize();
	    g *= arm_max;
      }
      float arm_min = (l1 > l2 ? l1-l2+ARM_SLOP : l2-l1+ARM_SLOP);
      if (g.length() < arm_min) {
	    g.normalize();
	    g *= arm_min;
      }

      float x = g[0], y = g[1];

      float c2 = (x*x + y*y - l1*l1 - l2*l2) / (2 * l1 * l2);
      if (c2 < -1.0 || c2 > 1.0) {
	    //cerr << "inverse2DOF c2 out of range" << endl;
	    return FALSE;
      }
      a2 = facos(c2);
      a2 *= (sign < 0 ? -1 : 1);
      float beta = fatan2(y,x);
      float cpsi =  (x*x + y*y + l1*l1 - l2*l2) / (2 * l1 * fsqrt(x*x + y*y));
      if (cpsi < -1.0 || cpsi > 1.0)  {
	    // cerr << "inverse2DOF cpsi out of range" << endl;
	    return FALSE;
      }
      float psi = facos(cpsi);

      if (a2 > 0.0) { 
	    a1 = beta - psi; 
      } else {
	    a1 = beta + psi;
      }
      return TRUE;
}
SbVec3f
Avatar::constrainedSetGoal(AvatarCache *frame, SbVec3f g, int sign) {  
      ArticulatedBody *b = (sign == LEFT_ARM_SIGN ? 
			    frame->leftArm : frame->rightArm);

      float a1,a2;

      if (sign == LEFT_ARM_SIGN) { 
	    if (g[0] > 0) g[0] = 0;
      } else {
	    if (g[0] < 0) g[0] = 0;
      }
      SbBool inworkspace = inverse2DOF(b, g, sign, a1, a2);

      float a1base, a2base, a1r, a2r;

      while (a1 < -M_PI) a1 += 2*M_PI;
      while (a2 < -M_PI) a2 += 2*M_PI;
      while (a1 > M_PI) a1 -= 2*M_PI;
      while (a2 > M_PI) a2 -= 2*M_PI;

      if (sign == LEFT_ARM_SIGN) { // left arm
	    a1base = frame->afi->leftShoulderRange.getValue()[0];
	    a2base = frame->afi->leftElbowRange.getValue()[0];
	    a1r = frame->afi->leftShoulderRange.getValue()[1];
	    a2r = frame->afi->leftElbowRange.getValue()[1];
	    if (!inworkspace) a1 = a1base, a2 = a2base;
	    //if (g[0] > 0) inworkspace=FALSE;
	    if (a1 > 0 && a1 < a1base - a1r) 
	      a1 = a1base-a1r, a2 = 0, inworkspace=FALSE;
	    if (a1 < 0 && a1 > (a1base + a1r)-2*M_PI) 
	      a1 = (a1base + a1r)-2*M_PI, inworkspace=FALSE;
	    if (a2 > 0 ||  a2 > a2base + a2r)
	      a2 = a2base+a2r, inworkspace=FALSE;
	    if (a2 < 0 && a2 < a2base - a2r)	
	      a2 = a2base-a2r, inworkspace=FALSE;	  
      } else {
	    a1base = frame->afi->rightShoulderRange.getValue()[0];
	    a2base = frame->afi->rightElbowRange.getValue()[0];
	    a1r = frame->afi->rightShoulderRange.getValue()[1];
	    a2r = frame->afi->rightElbowRange.getValue()[1];
	    if (!inworkspace) a1 = a1base, a2 = a2base;
	    //if (g[0] < 0) inworkspace=FALSE;
	    if (a1 < 0 && a1 < a1base - a1r) 
	      a1 = a1base-a1r, inworkspace=FALSE;
	    if (a1 > 0 && a1 > a1base + a1r) 
	      a1 = a1base + a1r, a2 = 0, inworkspace=FALSE;
	    if (a2 > 0 && a2 > a2base + a2r)
	      a2 = a2base+a2r, inworkspace=FALSE;
	    if (a2 < 0 ||  a2 < a2base - a2r)	
	      a2 = a2base-a2r, inworkspace=FALSE;	  
      }

//      if (inworkspace) setAngles(b, a1, a2);
//      printf(" %d %g %g \n", sign, a1, a2);
      setAngles(b, a1, a2);
      SbVec3f endPt;
      b->getEndPoint(endPt);
      return endPt;
}


void Avatar::resetArmAngles() {
      SbVec2f tmp = currentFrame->afi->leftAngles.getValue();
      setAngles(currentFrame->leftArm,tmp[0],tmp[1]);

      tmp = currentFrame->afi->rightAngles.getValue();
      setAngles(currentFrame->rightArm,tmp[0],tmp[1]);
}


void
Avatar::buildLeftArm()
{
  currentFrame->leftArm->rootJoint->shouldNotify.setValue(TRUE);
  currentFrame->leftArm->rootJoint->getNextJoint()->shouldNotify.setValue(TRUE);

  if (! currentFrame->afi->useLeftArm.getValue()) {
	currentFrame->leftABXform->translation.setValue(SbVec3f(0,0,0));
	//removeLimbs(currentFrame->leftArm);
	return;
  }

  // left arm first
  SbVec3f feat = currentFrame->afi->leftHand.getValue();

  SoCoordinate3 *contour = currentFrame->originalCoords;
  SbVec3f start0 = contour->point[currentFrame->afi->annotationIndices[LEFT_START0]];
  SbVec3f start1 = contour->point[currentFrame->afi->annotationIndices[LEFT_START1]];
  SbVec3f shoulder = (start0 + start1) / 2;
  currentFrame->leftABXform->translation.setValue(shoulder);
  currentFrame->afi->leftShoulderPos.setValue(shoulder);

  SbVec3f mid0=contour->point[currentFrame->afi->annotationIndices[LEFT_MID0]]-shoulder;
  SbVec3f mid1=contour->point[currentFrame->afi->annotationIndices[LEFT_MID1]]-shoulder;
  SbVec3f mid = (mid0 + mid1) / 2;
  currentFrame->afi->leftElbowPos.setValue(mid);

  feat -= shoulder;
  currentFrame->afi->leftEndPos.setValue(feat);

  SbVec3f tmp = mid;
  float l1 = sqrt(tmp[0]*tmp[0]+tmp[1]*tmp[1]);
  float a1 = fatan2(tmp[1],tmp[0]);
  tmp = feat-mid;
  float l2 = sqrt(tmp[0]*tmp[0]+tmp[1]*tmp[1]);
  float a2 = fatan2(tmp[1],tmp[0]) - a1;
  currentFrame->afi->leftAngles.setValue(SbVec2f(a1,a2));

  setLengths(currentFrame->leftArm,l1,l2);
  setAngles(currentFrame->leftArm,a1,a2);
  setElbowCenter(currentFrame->leftArm, -.57);  // 90 degrees in radians
}

void
Avatar::buildRightArm()
{

  currentFrame->rightArm->rootJoint->shouldNotify.setValue(TRUE);
  currentFrame->rightArm->rootJoint->getNextJoint()->shouldNotify.setValue(TRUE);

  if (! currentFrame->afi->useRightArm.getValue()) {
	currentFrame->rightABXform->translation.setValue(SbVec3f(0,0,0));
	//removeLimbs(currentFrame->rightArm);
	return;
  } 

  // now do right hand
  SbVec3f feat = currentFrame->afi->rightHand.getValue();

  SoCoordinate3 *contour = currentFrame->originalCoords;
  SbVec3f start0 = contour->point[currentFrame->afi->annotationIndices[RIGHT_START0]];
  SbVec3f start1 = contour->point[currentFrame->afi->annotationIndices[RIGHT_START1]];
  SbVec3f shoulder = (start0 + start1) / 2;
  currentFrame->rightABXform->translation.setValue(shoulder);
  currentFrame->afi->rightShoulderPos.setValue(shoulder);

  SbVec3f mid0=contour->point[currentFrame->afi->annotationIndices[RIGHT_MID0]]-shoulder;
  SbVec3f mid1=contour->point[currentFrame->afi->annotationIndices[RIGHT_MID1]]-shoulder;
  SbVec3f mid = (mid0 + mid1) / 2;
  currentFrame->afi->rightElbowPos.setValue(mid);

  feat -= shoulder;
  currentFrame->afi->rightEndPos.setValue(feat);

  SbVec3f tmp = mid;
  float l1 = sqrt(tmp[0]*tmp[0]+tmp[1]*tmp[1]);
  float a1 = fatan2(tmp[1],tmp[0]);
  tmp = feat-mid;
  float l2 = sqrt(tmp[0]*tmp[0]+tmp[1]*tmp[1]);
  float a2 = fatan2(tmp[1],tmp[0]) - a1;
  currentFrame->afi->rightAngles.setValue(SbVec2f(a1,a2));

  setLengths(currentFrame->rightArm,l1,l2);
  setAngles(currentFrame->rightArm,a1,a2);
  setElbowCenter(currentFrame->rightArm, 1.57);  // 90 degrees in radians
}
  

ArticulatedBody *
Avatar::loadArm()
{
  char *armpath = "Arm.iv";
  char *absarmpath = "/mas/vision/users/trevor/projects/avatar/Arm.iv";

  SoInput mySceneInput;
  SoNode *fn;
  if (!mySceneInput.openFile(armpath))  {
	if (!mySceneInput.openFile(absarmpath))  {
	      fprintf(stderr,"Cannot open Arm.iv file\n");
	      return NULL;
	} 
  } 

  SoDB::read(&mySceneInput, fn);
  return (ArticulatedBody *)fn;
}

SoSeparator *
Avatar::buildArmArticulatedBody()
{
  SoSeparator *base = new SoSeparator;
  base->ref();

  SoTransform *armxf = new SoTransform;
  base->addChild(armxf);     // ARM XF is child 0 (AB_GRP_XF_INDEX)
  base->addChild(loadArm()); // ARM is child 1 (AB_GRP_ARM_INDEX)

  base->unrefNoDelete();
  return base;
}
  

void
Avatar::computeAvatarNormalHeight()
{
  int c1;
  avatarNormalHeight = 0.0;

  // compute normal height of avatar.
  
  if (avatarStandingFrameCount.getValue() > frameSwitch->getNumChildren())
    {
      cerr << "BehaviorPerson:: standing frame count is larger "
	<< "than number of frames read in!" << endl;
      return;
    }

  for (c1 = 0; c1 < avatarStandingFrameCount.getValue(); c1++)
    {
      switchToFrame(c1);
      avatarNormalHeight = avatarNormalHeight.getValue() + currentFrame->afi->head.getValue()[1];
    }

  avatarNormalHeight = avatarNormalHeight.getValue() / (float) avatarStandingFrameCount.getValue();
}  

void
Avatar::normalizeAvatarContours(float old_scale,
				float new_scale,
				float aspect_ratio)
{
  SoCoordinate3 *current_contour;
  SbVec3f *vectors;
  int c1, c2;

  for (c1 = 0; c1 < originalCoordinateSwitch->getNumChildren(); c1++)
    {
      current_contour = (SoCoordinate3 *) originalCoordinateSwitch->getChild(c1);

      vectors = current_contour->point.startEditing();
      
      for (c2 = 0; c2 < current_contour->point.getNum(); c2++)
	{
	  vectors[c2] /= old_scale;
	  vectors[c2] *= new_scale;
	  
	  vectors[c2][0] /= aspect_ratio;
	}
      current_contour->point.finishEditing();
    }
}

void Avatar::normalizeAvatarSF(SoSFVec3f *field,
			 float old_scale,
			 float new_scale,
			 float aspect_ratio) {
      SbVec3f v = field->getValue();
      v /= old_scale;
      v *= new_scale;
      v[0] /= aspect_ratio;
      field->setValue(v);
}

  
SbVec2f Avatar::curTexLoc(SbVec3f p) {
      SbVec2f texscale = currentFrame->afi->texscale.getValue();
      SbVec2f texoffset = currentFrame->afi->texoffset.getValue();
      SbVec2f r = SbVec2f((p[0]/texscale[0]) + texoffset[0], 
		     (p[1]/texscale[1]) + texoffset[1]);
      return r;
}

void Avatar::normalizeAvatarAFIs(float old_scale,
				float new_scale,
				 float aspect_ratio)
{
  int c1;

  for (c1 = 0; c1 < frameSwitch->getNumChildren(); c1++)
    {
      AvatarFrameInfo *afi = frame(c1)->afi;

      normalizeAvatarSF(&(afi->leftHand),
			old_scale,new_scale,aspect_ratio);
      normalizeAvatarSF(&(afi->rightHand),
			old_scale,new_scale,aspect_ratio);
      normalizeAvatarSF(&(afi->head),
			old_scale,new_scale,aspect_ratio);
      afi->texoffset.setValue(SbVec2f(afi->centroid.getValue()[0],0));
      SbVec2f s = afi->texscale.getValue();
      s /= old_scale;
      s *= new_scale;
      s[0] /= aspect_ratio;
      afi->texscale.setValue(s);
    }




}

void Avatar::centerHeads() {
      int num;
      if (!(num = frameSwitch->getNumChildren())) return;
      SbVec3f head0 = frame(0)->afi->head.getValue();

      for (int c1 = 0; c1 < num; c1++) {    

	    AvatarFrameInfo *afi = frame(c1)->afi;

	    SbVec3f head = frame(c1)->afi->head.getValue();

	    float xoff = head0[0] - head[0];

	    afi->head.setValue(afi->head.getValue() + SbVec3f(xoff, 0, 0));
	    afi->leftHand.setValue(afi->leftHand.getValue() + SbVec3f(xoff, 0, 0));
	    afi->rightHand.setValue(afi->rightHand.getValue() + SbVec3f(xoff, 0, 0));

	    int numpts = frame(c1)->originalCoords->point.getNum();
	    SbVec3f *vectors = frame(c1)->originalCoords->point.startEditing();

	    for (int c2 = 0; c2 < numpts; c2++) {
		  vectors[c2] +=  SbVec3f(xoff, 0, 0);
	    }
	    frame(c1)->originalCoords->point.finishEditing();

	    float texoffset_x = xoff / (afi->texscale.getValue()[0]);
	    SbVec2f texoffset = afi->texoffset.getValue();
	    texoffset[0] += texoffset_x;
	    afi->texoffset.setValue(texoffset);
      }
}


void
Avatar::setAvatarScale(float newScale)
{
  avatarScale->scaleFactor.setValue(newScale, newScale, newScale);
}

void
Avatar::addGroupsToSceneGraph()
{
  int num_pts;
  float current_point[3];
  float current_tex_coords[2];
  float centroid[2];
  float head[2];
  float left_hand[2];
  float right_hand[2];
  float garbage;
  char texture_filename[256];

  FILE *cfp;
  cfp = fopen(cont_filename, "r");
  if (!cfp)
    {
      cerr << "Avatar::addGroupsToSceneGraph: Error opening contour file" << endl;
      return;
    }

  FILE *tfp;
  if ((tfp = fopen(stat_filename, "r")) == NULL)
    {
      cerr << "Avatar::addGroupsToSceneGraph: Error opening statistics file" << endl;
      return;
    }

  frameSwitch->whichChild.setValue(0);


  for (int frame_counter = 0; frame_counter < num_frames; frame_counter++)
    {
      SoGroup *grp = new SoGroup;
      frameSwitch->addChild(grp);

      char namebuf[100];
      sprintf(namebuf, "FRAME_%s_%d", output_base_filename, frame_counter);
      grp->setName(SbName(namebuf));
      printf("Added %s group\n", namebuf);

      sprintf(texture_filename, "%s-%d.rgb", output_base_filename, frame_counter);
      SoTexture2 *tex = new SoTexture2;
      tex->filename.setValue(texture_filename);
      SoCoordinate3 *cont_coords = new SoCoordinate3;
      SoTextureCoordinate2 *tex_coords = new SoTextureCoordinate2;

      SoNormal *norm = new SoNormal;
      SoNormalBinding *normb = new SoNormalBinding;
      SoTextureCoordinateBinding *texb = new SoTextureCoordinateBinding;
      norm->vector.setValue(0.0,0.0,1.0);
      normb->value = SoNormalBinding::OVERALL;
      texb->value = SoTextureCoordinateBinding::PER_VERTEX_INDEXED;

      // read in this frame's statistics
      fscanf(tfp, "%e %e", &centroid[0], &centroid[1]);
      fscanf(tfp, "%e %e %e %e", &garbage, &garbage, &garbage, &garbage);
      fscanf(tfp, "%e %e", &right_hand[0], &right_hand[1]);
      fscanf(tfp, "%e %e %e %e", &garbage, &garbage, &garbage, &garbage);
      fscanf(tfp, "%e %e", &left_hand[0], &left_hand[1]);
      fscanf(tfp, "%e %e %e %e", &garbage, &garbage, &garbage, &garbage);
      fscanf(tfp, "%e %e", &head[0], &head[1]);
      fscanf(tfp, "%e %e %e %e", &garbage, &garbage, &garbage, &garbage);
      
      AvatarFrameInfo *afi = new AvatarFrameInfo;
      afi->centroid.setValue(SbVec3f(centroid[0],centroid[1],0));
      afi->head.setValue(SbVec3f(head[0]-centroid[0],1.-head[1],0));
      afi->leftHand.setValue(SbVec3f(left_hand[0]-centroid[0],1.-left_hand[1],0));
      afi->rightHand.setValue(SbVec3f(right_hand[0]-centroid[0],1.-right_hand[1],0));


      // set the annotation indices all to -1
      for (int ac = 0; ac < NUM_INDICES; ac++)
	{
	  afi->annotationIndices.set1Value(ac, -1);
	}

      // now read in a 2D contour and convert it to 3D
      fread(&num_pts, sizeof(int), 1, cfp);
      for (int point_counter = 0; point_counter < num_pts; point_counter++)
	{
	  if (fread(current_tex_coords, sizeof(float), 2, cfp) < 2)
	    {
	      cerr << "Avatar::loadContoursFromPfinderContourFile: incomplete contour data set" << endl;
	      fclose(cfp);
	      return;
	    }
	  
	  current_point[0] = current_tex_coords[0] - centroid[0];
	  current_point[1] = 1. - current_tex_coords[1];
	  current_point[2] = 0;
	  current_tex_coords[1] = 1. - current_tex_coords[1];
	  
	  if (point_counter != num_pts - 1)   // last vertex is sometimes repeated in contour, screwing up Inventor
	    {
	      cont_coords->point.set1Value(point_counter, current_point);
	      tex_coords->point.set1Value(point_counter, current_tex_coords);
	    }
	}

      // add the original contour to the scene graph
      originalCoordinateSwitch->addChild(cont_coords);

      // add in the arms
      grp->addChild(buildArmArticulatedBody());  // left arm
      grp->addChild(buildArmArticulatedBody());  // right arm

      // add a temporary coordinate node to the group;
      // this is calculated each frame during the warping.
      grp->addChild(new SoCoordinate3);

      // add all the relevant nodes to the group 
      grp->addChild(afi);
      grp->addChild(tex);
      grp->addChild(tex_coords);
      grp->addChild(texb);
      grp->addChild(norm);
      grp->addChild(normb);
      grp->addChild(new SoIndexedFaceSet);
    }

  fclose(cfp);
  fclose(tfp);

  buildCaches();
  //normalize(AVATAR_ONE_OVER_ASPECT_RATIO);
  normalize(1.0);
}

void Avatar::normalize(float aspect_ratio) {
  // normalize the avatar's stored feature vectors and contours
  computeAvatarNormalHeight();
  printf("computed normal height = %g\n", avatarNormalHeight.getValue());
  normalizeAvatarAFIs(avatarNormalHeight.getValue(), 1.0, aspect_ratio);
  normalizeAvatarContours(avatarNormalHeight.getValue(), 1.0, aspect_ratio);
  setAvatarScale(INITIAL_AVATAR_SCALE);
  centerHeads();
  buildArms();
}      

SoNode *
Avatar::getNthChildOfType(SoGroup *parent,
			  int whichChild,
			  SoType classId)
{
  int temp_idx = 0;

  if (whichChild < 0)
    return NULL;

  for (int c1 = 0; c1 < parent->getNumChildren(); c1++)
    {
      SoNode *node = parent->getChild(c1);
      if (node->isOfType(classId))
	{
	  if (whichChild == temp_idx)
	    return node;
	  else
	    temp_idx++;
	}
    }
  return NULL;
}

  


void
Avatar::switchToFrame(int whichFrame)
{
  // check to make sure this is a valid frame

  if ((whichFrame < 0) || (whichFrame > frameSwitch->getNumChildren()))
    {
      cerr << "Avatar::switchToFrame: invalid frame index (should be between 0 and " << frameSwitch->getNumChildren() - 1 << ")" << endl;
      return;
    }

  currentFrame = frame(whichFrame);

  // copies the stored coordinates for the current frame
  // into the local "temporary" copy under the group node
  // which is stored under the frameSwitch node.
  currentFrame->newCoords->copyFieldValues(currentFrame->originalCoords);

  frameSwitch->whichChild.setValue(whichFrame);

}


void
Avatar::buildSceneGraph()
{
  // build the initial scene graph from the contour files.

  avatarScale = new SoScale;
  avatarScale->setName(SbName("AVATAR_SCALE"));
  addChild(avatarScale);

  originalCoordinateSwitch = new SoSwitch;
  originalCoordinateSwitch->setName(SbName("ORIGINAL_COORDINATE_SWITCH"));
  addChild(originalCoordinateSwitch);
  SoShapeHints *hints = new SoShapeHints;
  hints->faceType.setValue(SoShapeHints::UNKNOWN_FACE_TYPE);
  addChild(hints);


  frameSwitch = new SoSwitch;
  frameSwitch->setName(SbName("FRAME_SWITCH"));
  addChild(frameSwitch);


  // add everything else to the scene graph
  addGroupsToSceneGraph();
}

void
Avatar::buildFromPfinderStatisticsFiles(char *user_raw_filename,
					char *user_seg_filename,
					char *user_stat_filename,
					char *user_cont_filename,
					int user_image_width,
					int user_image_height,
					int user_num_frames, // 0 == use all 
					int standing_num_frames, // 0 == use all 
					char *user_output_base_filename)
{
  raw_filename = user_raw_filename;
  seg_filename = user_seg_filename;
  stat_filename = user_stat_filename;
  cont_filename = user_cont_filename;
  output_base_filename = user_output_base_filename;
  image_width = user_image_width;
  image_height = user_image_height;
  num_frames = user_num_frames;



  convertPfinderRawImagesToTextures(); // sets num_frames if 0
  if (!standing_num_frames) {
	avatarStandingFrameCount.setValue(num_frames); 
  } else {
	avatarStandingFrameCount.setValue(standing_num_frames);
  }
  buildSceneGraph();
}  




void
Avatar::updateToNewPositions(float normal_height,
			     SbVec3f left_hand_pos,
			     SbVec3f right_hand_pos,
			     SbVec3f head_pos,
			     SbVec3f centroid_pos) {
      static SoTranslation *tmpTranslation = NULL;

      SbVec3f frame_centroid_pos, frame_left_hand_pos, frame_right_hand_pos, frame_head_pos, oldstart, oldend, newgoal, newmid, newend;
      SbVec2f oldang, newang, dang;

      // weights for frame distance selection:

      // on head offset
      static const float HEAD_WEIGHT = 4.0; 
      // on offset between closest interpolatable hand position and user hand
      static const float HAND_WEIGHT = 1.0; 
      // on angle change to achieve closest hand position
      static const float ARM_ANG_WEIGHT = 0.005;
      // on end position change to achieve closest hand position
      static const float ARM_POS_WEIGHT = 0.15;
	    
      if (normal_height == 0.0) {
	    cerr << "Avatar::updateToNewPositions: normal height was zero!"
		 << endl;
	    return;
      }
      left_hand_pos[0] -= centroid_pos[0];
      right_hand_pos[0] -= centroid_pos[0];
      head_pos[0] -= centroid_pos[0];
      left_hand_pos /= normal_height;
      right_hand_pos /=normal_height;
      head_pos /= normal_height;

      float dist, min;
      int min_idx = 0;

      if (tmpTranslation) {
	    // for tail bending seq -- adjust hands
	    left_hand_pos -= tmpTranslation->translation.getValue();
	    right_hand_pos -= tmpTranslation->translation.getValue();
      }
      if (doSelect) {
	    for (int c1 = 0; c1 < frameSwitch->getNumChildren(); c1++) {
		  AvatarFrameInfo *afi = frame(c1)->afi;
		  frame_left_hand_pos = afi->leftHand.getValue();
		  frame_right_hand_pos = afi->rightHand.getValue();
		  frame_head_pos = afi->head.getValue();
		  
		  SbVec3f tmp = frame_head_pos - head_pos;
		  dist = HEAD_WEIGHT * tmp.length();

		  if (doWarp && afi->useLeftArm.getValue()) {
			oldstart = afi->leftShoulderPos.getValue();
			newgoal = left_hand_pos-oldstart;
			oldend = afi->leftEndPos.getValue();
			newend = constrainedSetGoal(frame(c1),newgoal,LEFT_ARM_SIGN);
			frame_left_hand_pos = newend + oldstart;
			tmp = frame_left_hand_pos - left_hand_pos;
			dist += HAND_WEIGHT * tmp.length();

			oldang = afi->leftAngles.getValue();
			getAngles(frame(c1)->leftArm,newang[0],newang[1]);
			dang = newang - oldang; 
			dist += ARM_ANG_WEIGHT * dang.length();
			tmp = oldend - newend;
			dist += ARM_POS_WEIGHT * tmp.length();
		  } else {
			tmp = frame_left_hand_pos - left_hand_pos;
			dist += tmp.length();
		  }

		  if (doWarp && afi->useRightArm.getValue()) {
			oldstart = afi->rightShoulderPos.getValue();
			newgoal = right_hand_pos-oldstart;
			oldend = afi->rightEndPos.getValue();
			newend = constrainedSetGoal(frame(c1),newgoal,RIGHT_ARM_SIGN);
			frame_right_hand_pos = newend + oldstart;
			tmp = frame_right_hand_pos - right_hand_pos;
			dist += HAND_WEIGHT * tmp.length();

			oldang = afi->rightAngles.getValue();
			getAngles(frame(c1)->rightArm,newang[0],newang[1]);
			dang = newang - oldang; 
			dist += ARM_ANG_WEIGHT * dang.length();

			tmp = oldend - newend;
			dist += ARM_POS_WEIGHT * tmp.length();
		  } else {
			tmp = frame_right_hand_pos - right_hand_pos;
			dist += tmp.length();
		  }


		  if (!c1 || dist < min) min = dist, min_idx = c1;
	    }
	    if (min_idx != frameSwitch->whichChild.getValue()) {
		  switchToFrame(min_idx);
		  resetArmAngles();
	    }
      }

      if (doWarp) {
	    oldstart = currentFrame->afi->leftShoulderPos.getValue();
	    newgoal = left_hand_pos-oldstart;
	    newend = constrainedSetGoal(currentFrame, newgoal, LEFT_ARM_SIGN);
	    newmid = getMid(currentFrame->leftArm);
	    
	    leftArmWarp(newend, newmid);

	    oldstart = currentFrame->afi->rightShoulderPos.getValue();
	    newgoal = right_hand_pos-oldstart;
	    newend = constrainedSetGoal(currentFrame, newgoal, RIGHT_ARM_SIGN);
	    newmid = getMid(currentFrame->rightArm);
	    
	    rightArmWarp(newend, newmid);
      }

      if (tailSep) {
	    if (!tmpTranslation) {
		  cerr << "inserting tmp translation for tail" << endl;
		  // insert after the AVATAR_SCALE
		  this->insertChild(tmpTranslation = new SoTranslation, 1);
	    }

	    //static const TAIL_XF_IDX = 0; 
	    static const TAIL_SEQ_SWITCH_IDX = 1;
	    static const STANDING_SEQ_IDX = 0;
	    static const BENDING_SEQ_IDX = 1;

	    if (tailSep->getNumChildren() < 2) 
	      cerr << "Wrong number of children in tail!" << cerr;

	    SoSwitch *seqswitch = (SoSwitch *) tailSep->getChild(TAIL_SEQ_SWITCH_IDX);
	    int numseqs = seqswitch->getNumChildren();
	    int whichseq = seqswitch->whichChild.getValue();

	    frame_head_pos = frame(min_idx)->afi->head.getValue();
	    float hy = head_pos[1];

	    static const float BENDING_BOTTOM = 0.4;

	    printf(" tail hy = %g ", hy);
	    if (hy >= 1.0) { 
		  tmpTranslation->translation.setValue(SbVec3f(0,0,0));

		  seqswitch->whichChild.setValue(STANDING_SEQ_IDX);

		  SoSwitch *tailframeswitch = (SoSwitch *) 
		    seqswitch->getChild(STANDING_SEQ_IDX);

		  int numtailframes = tailframeswitch->getNumChildren();
		  int whichtailframe = tailframeswitch->whichChild.getValue();

		  whichtailframe++;
		  whichtailframe %= numtailframes;
		  
		  tailframeswitch->whichChild.setValue(whichtailframe);

		  printf(" standing frame %d\n", whichtailframe);

	    } else {
		  seqswitch->whichChild.setValue(BENDING_SEQ_IDX);

		  SoSwitch *tailframeswitch = (SoSwitch *) 
		    seqswitch->getChild(BENDING_SEQ_IDX);

		  int numtailframes = tailframeswitch->getNumChildren();
		  int whichtailframe = tailframeswitch->whichChild.getValue();

		  float bendfactor = (1.0 - ((hy - BENDING_BOTTOM) / 
					     (1.0 - BENDING_BOTTOM)));
		    
		  if (bendfactor > 1.0) bendfactor = 1.0;
		  if (bendfactor < 0.0) bendfactor = 0.0;
		  
		  whichtailframe = (int) ((float)(numtailframes-1) * bendfactor);
		  tailframeswitch->whichChild.setValue(whichtailframe);
		  
		  float ady =  - (bendfactor * BENDING_BOTTOM);
		  tmpTranslation->translation.setValue(SbVec3f(0,ady,0));
		  printf(" bending factor %g frame %d ady %g \n", bendfactor, whichtailframe, ady);

	    }






      }

}
      

SbVec2f inline rotate2(SbVec2f p, float da) {
      SbVec2f newv;
      float ang = fatan2(p[1],p[0]);
      float r = p.length();

      ang += da;

      newv[0] = r * fcos(ang);
      newv[1] = r * fsin(ang);
      return newv;
}

void
Avatar::leftArmWarp(SbVec3f newend, SbVec3f newmid)
{ 

  SbVec3f oldstart = currentFrame->afi->leftShoulderPos.getValue();

  if (!currentFrame->afi->useLeftArm.getValue()) return;
  SbVec3f tmp;
  SbVec2f warp[NUM_INDICES];
  
  SoCoordinate3 *contour = currentFrame->originalCoords;
  SoCoordinate3 *newcontour = currentFrame->newCoords;

  SbVec3f oldgoal = currentFrame->afi->leftEndPos.getValue();
  SbVec3f oldmid = currentFrame->afi->leftElbowPos.getValue();

  // define warp vectors at start, middle and tip locations on contour
  // start
  warp[LEFT_START0] = warp[LEFT_START1] = SbVec2f(0,0);

  // middle
  tmp = newend-newmid;
  float a2 = fatan2(tmp[1],tmp[0]);
  tmp = -newmid;
  float a1 = fatan2(tmp[1],tmp[0]);
  float av = (a1+a2)/2;
  SbVec3f unit = SbVec3f(cos(av),sin(av),0);
  if (unit[1]<0) unit=-unit;
  SbVec3f oldmid0 = contour->point[currentFrame->afi->annotationIndices[LEFT_MID0]]-oldstart;
  SbVec3f oldmid1 = contour->point[currentFrame->afi->annotationIndices[LEFT_MID1]]-oldstart;
  SbVec3f newmid0, newmid1;
  tmp = oldmid0 - oldmid;
  float r = tmp.length();
  if (oldmid1[1] > oldmid0[1]) { 
    newmid1 = newmid + r * unit;
    newmid0 = newmid - r * unit;
  } else { 
    newmid1 = newmid - r * unit;
    newmid0 = newmid + r * unit;
  }
  tmp = newmid0 - oldmid0;
  warp[LEFT_MID0] = SbVec2f(tmp[0],tmp[1]);
  tmp = newmid1 - oldmid1;
  warp[LEFT_MID1] = SbVec2f(tmp[0],tmp[1]);
  tmp = newmid - oldmid;

  // tip
  // tip should be approximately colinear with last segment...
  SbVec3f oldtip = contour->point[currentFrame->afi->annotationIndices[LEFT_TIP]]-oldstart;
  tmp = oldtip - oldgoal;
  r = tmp.length();
  SbVec3f newtip = newend;
  newtip += SbVec3f(r * cos(a2),r * sin(a2),0);
  tmp = newtip-oldtip;
  warp[LEFT_TIP] = SbVec2f(tmp[0],tmp[1]);

  for (int k=LEFT_START0;k<LEFT_START1;k++) {
    int basei = currentFrame->afi->annotationIndices[k];
    SbVec3f base3 = contour->point[basei];
    SbVec2f base = SbVec2f(base3[0],base3[1]);
    SbVec2f basew = warp[k];
    SbVec2f newbase = base + basew;

    int endi = currentFrame->afi->annotationIndices[k+1];
    SbVec3f end3 = contour->point[endi];
    SbVec2f end = SbVec2f(end3[0],end3[1]);
    SbVec2f endw = warp[k+1];
    SbVec2f newend = end + endw;

    SbVec2f oldv = end - base;
    SbVec2f newv = newend - newbase;
    float rotation = fatan2(newv[1],newv[0]) - fatan2(oldv[1],oldv[0]);
    float scale = newv.length() / oldv.length();

    for (int i=basei; i<endi; i++) {
	  SbVec3f oldpt = contour->point[i];
	  SbVec2f p = SbVec2f(oldpt[0],oldpt[1]); 
	  p -= base;
	  p = rotate2(p, rotation);
	  p *= scale;
	  p += newbase;
	  newcontour->point.set1Value(i,SbVec3f(p[0],p[1],0));
    }
  }
}



void
Avatar::rightArmWarp(SbVec3f newend, SbVec3f newmid)
{ 

  SbVec3f oldstart = currentFrame->afi->rightShoulderPos.getValue();

  if (!currentFrame->afi->useRightArm.getValue()) return;
  SbVec3f tmp;
  SbVec2f warp[NUM_INDICES];
  
  SoCoordinate3 *contour = currentFrame->originalCoords;
  SoCoordinate3 *newcontour = currentFrame->newCoords;

  SbVec3f oldgoal = currentFrame->afi->rightEndPos.getValue();
  SbVec3f oldmid = currentFrame->afi->rightElbowPos.getValue();

  // define warp vectors at start, middle and tip locations on contour
  // start
  warp[RIGHT_START0] = warp[RIGHT_START1] = SbVec2f(0,0);

  // middle
  tmp = newend-newmid;
  float a2 = fatan2(tmp[1],tmp[0]);
  tmp = -newmid;
  float a1 = fatan2(tmp[1],tmp[0]);
  float av = (a1+a2)/2;
  SbVec3f unit = SbVec3f(cos(av),sin(av),0);
  if (unit[1]<0) unit=-unit;
  SbVec3f oldmid0 = contour->point[currentFrame->afi->annotationIndices[RIGHT_MID0]]-oldstart;
  SbVec3f oldmid1 = contour->point[currentFrame->afi->annotationIndices[RIGHT_MID1]]-oldstart;
  SbVec3f newmid0, newmid1;
  tmp = oldmid0 - oldmid;
  float r = tmp.length();
  if (oldmid1[1] > oldmid0[1]) { 
    newmid1 = newmid + r * unit;
    newmid0 = newmid - r * unit;
  } else { 
    newmid1 = newmid - r * unit;
    newmid0 = newmid + r * unit;
  }
  tmp = newmid0 - oldmid0;
  warp[RIGHT_MID0] = SbVec2f(tmp[0],tmp[1]);
  tmp = newmid1 - oldmid1;
  warp[RIGHT_MID1] = SbVec2f(tmp[0],tmp[1]);
  tmp = newmid - oldmid;

  // tip
  // tip should be approximately colinear with last segment...
  SbVec3f oldtip = contour->point[currentFrame->afi->annotationIndices[RIGHT_TIP]]-oldstart;
  tmp = oldtip - oldgoal;
  r = tmp.length();
  SbVec3f newtip = newend;
  newtip += SbVec3f(r * cos(a2),r * sin(a2),0);
  tmp = newtip-oldtip;
  warp[RIGHT_TIP] = SbVec2f(tmp[0],tmp[1]);

  for (int k=RIGHT_START0;k<RIGHT_START1;k++) {
    int basei = currentFrame->afi->annotationIndices[k];
    SbVec3f base3 = contour->point[basei];
    SbVec2f base = SbVec2f(base3[0],base3[1]);
    SbVec2f basew = warp[k];
    SbVec2f newbase = base + basew;

    int endi = currentFrame->afi->annotationIndices[k+1];
    SbVec3f end3 = contour->point[endi];
    SbVec2f end = SbVec2f(end3[0],end3[1]);
    SbVec2f endw = warp[k+1];
    SbVec2f newend = end + endw;

    SbVec2f oldv = end - base;
    SbVec2f newv = newend - newbase;
    float rotation = fatan2(newv[1],newv[0]) - fatan2(oldv[1],oldv[0]);
    float scale = newv.length() / oldv.length();

    for (int i=basei; i<endi; i++) {
	  SbVec3f oldpt = contour->point[i];
	  SbVec2f p = SbVec2f(oldpt[0],oldpt[1]); 
	  p -= base;
	  p = rotate2(p, rotation);
	  p *= scale;
	  p += newbase;
	  newcontour->point.set1Value(i,SbVec3f(p[0],p[1],0));
    }
  }
}

