#include <iostream.h>
#include <string.h>
#include "BehaviorPerson.h"
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoLists.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoText3.h>

#include "VisionServices.h"
#include "util.h"
#include "Alive.h"
#include "BasicDefines.h"
#include "PersonMotorSkills.h"
#include "Kalman.h"
#include "Vector.h"
#include "RaySensor.h"

#include <rpc/rpc.h>
#include "pfinder_clnt.h"

/* PERSON LABELS */


#define RIGHTHAND   1
#define LEFTHAND    0
#define HEAD    2
#define NOTHING     -1
#define DELTALOW    4.0
#define FEET        2


#define CCLASS(x) (remote_person->cc3d.pfcc3d_u.cc.cc_val[x])
#define NCCLASSES ((float)(remote_person->cc3d.pfcc3d_u.cc.cc_len))
#define FEATURE(x) SbVec3f((float *)(&(remote_person->f3.pffeat3d_u.feats->x)))
#define ISDOING(x) (remote_person->g.pfgest_u.bits->x)

extern pfperson *remote_person;
extern CLIENT *remote_person_server;
static int inReadInstance = 0;
static SoSeparator *colorClassNodes = NULL;
/*****************************************************************************
 * 
 *			Definition of BehaviorPerson 
 * 
 ****************************************************************************/
SO_NODE_SOURCE(BehaviorPerson);

static void personGrabCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
    BehaviorPerson *bp = (BehaviorPerson *)dba->creature;
    LeafActivity *la = (LeafActivity *)act;
    MotorCommandBlock mcb1(GC_SETBOOLFIELD_GEN);
    mcb1.setNextArgGeneric("ballInHand");
    mcb1.setNextArgGeneric(1);
    bp->getController()->executeMotorCommand(mcb1);
    //  bp->ballInHand.setValue(TRUE);
    MotorCommandBlock mcb(GC_GRAB);
    mcb.setNextArgGeneric((BaseCreature *)pronome->object);
    mcb.setNextArgGeneric("LHAND");
    mcb.setNextArgGeneric(1);
    SbMatrix sbm;
    sbm.makeIdentity();
    mcb.arglist.append(ArgType(sbm));
    
    //  static SoTransform *trans = new SoTransform;
    //  trans->ref();
    //  mcb.setNextArgGeneric((void *)trans);
    la->issueMotorCommand(*dba->creature,mcb);
}

static void personGrabBirdCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
    BehaviorPerson *bp = (BehaviorPerson *)dba->creature;
    LeafActivity *la = (LeafActivity *)act;
    MotorCommandBlock mcb(GC_GRAB);
    mcb.setNextArgGeneric((BaseCreature *)pronome->object);
    mcb.setNextArgGeneric((bp->handOverHeadLeft.getValue())? "LHAND" : "RHAND" );
    mcb.setNextArgGeneric(1);
    SbMatrix sbm;
    sbm.makeIdentity();
    mcb.arglist.append(ArgType(sbm));
    
    //  static SoTransform *trans = new SoTransform;
    //  trans->ref();
    //  mcb.setNextArgGeneric((void *)trans);
    la->issueMotorCommand(*dba->creature,mcb);
}

static void personDropCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
  
    LeafActivity *la = (LeafActivity *)act;

  
    BehaviorPerson *bp = (BehaviorPerson *)dba->creature;
    MotorCommandBlock mcb1(GC_SETBOOLFIELD_GEN);
    mcb1.setNextArgGeneric("ballInHand");
    mcb1.setNextArgGeneric(0);
    bp->getController()->executeMotorCommand(mcb1);
//    bp->ballInHand.setValue(FALSE);
    la->mcb.cmd = 30001;
    SbVec3f posToUse = bp->leftHand.getValue();
    posToUse[1] = (posToUse[1]<30.0)? 30.0 :posToUse[1];
    la->mcb.setNextArgGeneric(posToUse); 
    la->mcb.setNextArgGeneric((float)(-bp->throwingSpeed.getValue()));
    la->mcb.setNextArgGeneric((float)(0.0));
    la->mcb.setNextArgGeneric((float)(0.0));
    
    MotorCommandBlock mcb(GC_UNGRAB);
    la->issueMotorCommand(*dba->creature,mcb);

}

static void personThrowCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
  
    LeafActivity *la = (LeafActivity *)act;

    BehaviorPerson *bp = (BehaviorPerson *)dba->creature;
    SbVec3f pos = bp->position();
    if(pos[0]>38.0)
	personDropCB(act, action);
    else {
//	MotorCommandBlock mcb(GC_UNGRAB);
//	la->issueMotorCommand(*dba->creature,mcb);
      
	MotorCommandBlock mcb1(GC_SETBOOLFIELD_GEN);
	mcb1.setNextArgGeneric("ballInHand");
	mcb1.setNextArgGeneric(0);
	bp->getController()->executeMotorCommand(mcb1);
//	bp->ballInHand.setValue(FALSE);
	la->mcb.cmd = 30001;
	la->mcb.setNextArgGeneric((bp->throwingLeft.getValue())?
	    bp->leftHand.getValue() : bp->rightHand.getValue()); 
	la->mcb.setNextArgGeneric((float)((bp->throwingLeft.getValue())
	    ?-bp->throwingSpeed.getValue() : -bp->throwingSpeed.getValue()));
	la->mcb.setNextArgGeneric((float)(0.0));
	la->mcb.setNextArgGeneric((float)(0.0));
	
	MotorCommandBlock mcb(GC_UNGRAB);
	la->issueMotorCommand(*dba->creature,mcb);
    }
}

SbBool findObjectInScreenCoord(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
//    SNIFFSENSOR(ss, dua);
    SniffSensor *ss = (SniffSensor *)creatureRM->getSensor(dua->creature);
    int inScreenRange = 0;
    int in3DRange = 0;
    float len;
    RayData *rd = NULL;
    if(ss)
	rd = ss->getDataForClosestWithTag(creatureRM->typeTag.getValue());
    if(rd){
	SbVec3f p1, p1s, p2, p2s;
	p1 = ss->getVecInWorldFrame(SbVec3f(0, 0, 0));
	p2 = rd->pt;
	rd->range = fabs(p1[2] - p2[2]);
	Alive::aliveCamera->projectToScreen(p1, p1s);
	Alive::aliveCamera->projectToScreen(p2, p2s);
	len = (p1s-p2s).length();
	BehaviorPerson *bp = (BehaviorPerson *)dua->creature;
	inScreenRange = (len < bp->screenTolerance.getValue())? 1:0;
    }
    in3DRange = TESTRANGE(rd, creatureRM);
    //printf(" 3d -> %d %f 2d-> %d %f\n", in3DRange, rd->range, inScreenRange, len);
    if(inScreenRange && in3DRange)
    {
	p->object = rd->who;
	p->location = rd->pt;
	p->status = Pronome::PASSEDFIND;
	p->type = Pronome::OBJECT;
	p->range = rd->range;
	p->bearing = rd->bearing;
	
	p->vel = rd->vel;
	p->validVel = rd->validVel;
	
	creatureRM->value.setValue(creatureRM->maxValue.getValue());
    }
    else p->status = Pronome::FAILEDFIND;
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
    
}

void
BehaviorPerson::initClass()
{
    SO_NODE_INIT_CLASS(BehaviorPerson, BehaviorCreature, "BehaviorCreature");
    BaseBehaviorNode::addCallback("pgrab", personGrabCB);
    BaseBehaviorNode::addCallback("pgrabbird", personGrabBirdCB);
    BaseBehaviorNode::addCallback("pthrow", personThrowCB);
    BaseBehaviorNode::addCallback("pdrop", personThrowCB);
    BaseBehaviorNode::addCallback("screenFind", findObjectInScreenCoord);
}


BehaviorPerson::BehaviorPerson() : BehaviorCreature()
{
    SO_NODE_CONSTRUCTOR(BehaviorPerson);
    
    SO_NODE_ADD_FIELD(updateStateFromVision, (FALSE));
    SO_NODE_ADD_FIELD(extendedLeft, (FALSE));
    SO_NODE_ADD_FIELD(extendedRight, (FALSE));
    SO_NODE_ADD_FIELD(pointingLeft, (FALSE));
    SO_NODE_ADD_FIELD(pointingRight, (FALSE));
    SO_NODE_ADD_FIELD(handOverHeadLeft, (FALSE));
    SO_NODE_ADD_FIELD(handOverHeadRight, (FALSE));
    SO_NODE_ADD_FIELD(handLowLeft, (FALSE));
    SO_NODE_ADD_FIELD(handLowRight, (FALSE));
    SO_NODE_ADD_FIELD(kickingLeft, (FALSE));
    SO_NODE_ADD_FIELD(kickingRight, (FALSE));
    SO_NODE_ADD_FIELD(throwingLeft, (FALSE));
    SO_NODE_ADD_FIELD(throwingRight, (FALSE));
    SO_NODE_ADD_FIELD(throwingEnabled, (FALSE));
    SO_NODE_ADD_FIELD(standing, (FALSE));
    SO_NODE_ADD_FIELD(sitting, (FALSE));
    SO_NODE_ADD_FIELD(bendingOver, (FALSE));
    SO_NODE_ADD_FIELD(handsOverHead, (FALSE));
    SO_NODE_ADD_FIELD(leftHand, (0, 0, 0));
    SO_NODE_ADD_FIELD(rightHand, (0, 0, 0));
    SO_NODE_ADD_FIELD(head, (0, 0, 0));
    SO_NODE_ADD_FIELD(bodyCentroid, (0, 0, 0));
    SO_NODE_ADD_FIELD(gazeTranslation, (0, 0, 0));
    SO_NODE_ADD_FIELD(throwingSpeed, (60.0));
    SO_NODE_ADD_FIELD(sysNoise, (0, 0, 0));
    SO_NODE_ADD_FIELD(obsNoise, (0, 0, 0));
    SO_NODE_ADD_FIELD(useSpeech, (FALSE));
    SO_NODE_ADD_FIELD(personExists, (FALSE));
    SO_NODE_ADD_FIELD(ballInHand, (FALSE));
    SO_NODE_ADD_FIELD(somethingInHand, (FALSE));
    SO_NODE_ADD_FIELD(screenTolerance, (.1));
    SO_NODE_ADD_FIELD(isUtteringSit, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringShake, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringUp, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringDown, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringGoAway, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringFetch, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringBruce, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringNewt, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringGoodbye, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringGoodDog, (FALSE));
    SO_NODE_ADD_FIELD(isUtteringBadDog, (FALSE));
    SO_NODE_ADD_FIELD(normalWidth, (0));
    SO_NODE_ADD_FIELD(normalHeight, (0));
    SO_NODE_ADD_FIELD(normalValid, (FALSE));
    SO_NODE_ADD_FIELD(userGesturing, (FALSE));
    
    
    // THIS SHOULD BE SET IN FILE
    SO_NODE_ADD_FIELD(numReps, (0));
    
    partsInstalled = FALSE;
    audioMan = NULL;
    lhxf = NULL;
    rhxf = NULL;
    hhxf = NULL;
    gbxf = NULL;
    leftText = NULL;
    rightText = NULL;
    bodyText = NULL;
	
    num_colors = 0;
    colors = NULL;
}

SbBool
#ifdef SO_VERSION_REVISION
BehaviorPerson::readInstance(SoInput *in, unsigned short flags)
#else
BehaviorPerson::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (BehaviorCreature::readInstance(in, flags)) {
#else
    if (BehaviorCreature::readInstance(in)) {
#endif
	SoSearchAction sa;
	SoPath *tmp_path = NULL;
	tmp_path = getPathToName(this, "RHAND");
	if(tmp_path){
	    rhxf = (SoTransform *)tmp_path->getTail();
	}
	tmp_path = getPathToName(this, "LHAND");
	if(tmp_path){
	    lhxf = (SoTransform *)tmp_path->getTail();
	}
//	tmp_path = getPathToName(this, "KHAND");
//	if(tmp_path){
//	    klhxf = (SoTransform *)tmp_path->getTail();
//	}
	tmp_path = getPathToName(this, "HEAD");
	if(tmp_path){
	    hhxf = (SoTransform *)tmp_path->getTail();
	}
	tmp_path = getPathToName(this, "GAZEBLOCK");
	if(tmp_path){
	    gbxf = (SoTransform *)tmp_path->getTail();
	}
	inReadInstance = 1;
	changeRep(RepSwitch::REP_3);
	tmp_path = getPathToName(this, "LEFTTEXT");
	if(tmp_path){
	    leftText = (SoText3 *)tmp_path->getTail();
	}
	tmp_path = getPathToName(this, "RIGHTTEXT");
	if(tmp_path){
	    rightText = (SoText3 *)tmp_path->getTail();
	}
	tmp_path = getPathToName(this, "BODYTEXT");
	if(tmp_path){
	    bodyText = (SoText3 *)tmp_path->getTail();
	}
	tmp_path = getPathToName(this, "COLORCLASSNODES");
	if(tmp_path){
	    colorClassNodes = (SoSeparator *)tmp_path->getTail();
	}
	changeRep(RepSwitch::REP_0);
	inReadInstance = 0;
	
	
//	((SoSwitch *)switchNode)->whichChild.setValue(defaultSwitchValue.getValue());
	
	if(useSpeech.getValue() == TRUE){
	    audioMan = new AudioMan(14159);
	}
	return (1);
   } else {
	return(0);
   }
}

void BehaviorPerson::addMotorSkills()
{
    BehaviorCreature::addMotorSkills();
    BallThrow *bt = new BallThrow((Person *)this, "throwball");
    getController()->addCommand(30001, "throwball", bt, TRUE, 
				ArgType((float)10), ArgType((float)0), 
				ArgType((float).95));
    PersonUpdateVisionFromMCB *puvm = new PersonUpdateVisionFromMCB(this, "updateVisionFromMCB");
    SbPList arglist;
    static int tempint = 0;
    static ArgType tempintarg(tempint);
    static SbVec3f tempvec(0,0,0);
    static ArgType tempvecarg(tempvec);
    static char personNotExistString[2];
    personNotExistString[0] = 1;
    personNotExistString[1] = 0;
    static ArgType tempchararg((char *)personNotExistString);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempintarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempvecarg);
    arglist.append(&tempchararg);
    getController()->addCommand(30002, "updateVisionFromMCB", puvm, TRUE, arglist);
    PersonUpdateAudioFromMCB *puam = new PersonUpdateAudioFromMCB(this, "updateAudioFromMCB");
    getController()->addCommand(30003, "updateAudioFromMCB", puam, TRUE, ArgType((int)-1));
}

BehaviorPerson::~BehaviorPerson()
{
}

void BehaviorPerson::changeRep(int newRep)
{
    // This is a hack so droid of person always has diagnostics
    //
    //
    if((motorCreatureType.getValue() == MotorCreature::DROID) &&
	(newRep == RepSwitch::REP_0))
	newRep = RepSwitch::REP_3;
	BehaviorCreature::changeRep(newRep);
    
    // add special code here
    // just for debugging purposes
    if((newRep != RepSwitch::REP_REVERT) &&
	(newRep != RepSwitch::REP_SENSING) &&
	(newRep != RepSwitch::REP_SENSING_OO)){
	switch (newRep){
	    // normal composited user
	    case RepSwitch::REP_0 : { 
		tag.setValue(300); 
		break; 
	    }
	    // guide
	    case RepSwitch::REP_1 :{ 
		tag.setValue(876); 
		break ;
	    }
	    // hands
	    case RepSwitch::REP_2 :{ 
		tag.setValue(875); 
		break ;
	    }
	    // diagnostic
	    case RepSwitch::REP_3 :{
		tag.setValue(300); 
		break ;
	    }
	    default: { tag.setValue(300); break; } 
	}
    }
    if(VisionServices::contourServer != NULL){
	if((newRep != RepSwitch::REP_REVERT) &&
	    (newRep != RepSwitch::REP_SENSING) &&
	    (newRep != RepSwitch::REP_SENSING_OO)){
	    
	    currentRep = newRep;
	    SbBool useContour = FALSE;
	    if((newRep ==  RepSwitch::REP_0)||
		(newRep == RepSwitch::REP_3)||
		((newRep == RepSwitch::REP_1) && motorCreatureType.getValue() == MotorCreature::DROID))
		useContour = TRUE;
		
	    VisionServices::contourServer->useContour(useContour);
	    if((newRep == RepSwitch::REP_3) &&
		(motorCreatureType.getValue() == MotorCreature::DROID))
		 VisionServices::contourServer->showContour(FALSE);
	    else VisionServices::contourServer->showContour(
		(newRep !=  RepSwitch::REP_3)? FALSE  : TRUE);
	}
    }
    return;
}



void BehaviorPerson::updateBehaviorSystemAt(UpdateStateData &usd)
{
    if(updateStateFromVision.getValue() && 
	Alive::connectToVision)
	updateFromVisionServices();
    /*
     * Now do any processing that is independent of source
     */
    
    if(useSpeech.getValue() == TRUE)
	updateFromSpeech();
	
    BehaviorCreature::updateBehaviorSystemAt(usd);
}

void BehaviorPerson::speechDiagnostic(char *s)
{
    /*
     *  Assumes presence of a text creature with a tag of 302
     */
    MotorCreature *mc = (MotorCreature *)myCache->findBaseCreatureWithTag(302);
    if(mc != NULL){
	mc->controller->executeMotorCommand(
	    MotorCommandBlock(mc->controller->getCommandId("setString"), 
	    ArgType((char *)s))); 
    }
    return;
}

void BehaviorPerson::updateFromSpeech()
{
// old AudioMan code

/*
    AudioManRet *amr;
    if ((amr = audioMan->CheckSpeech()) != NULL) {
      printf("Received uterance\n");
      printf("Received %s \n", amr->utterance);
    }
*/
    // new code for new AudioMan
    MotorCommandBlock mcb(30003); // PersonUpdateAudioFromMCB's magic number
    // check for utterances
    // set the first argument to a number between 0 and 10
    // corresponding to one of the 11 boolean fields

    audioMan->CheckSpeech();
    AudioEvent *ae = audioMan->GetLastAudioEvent();
    
    if (ae)
      {
	if (strstr(ae->utterance, "SIT"))
	  {
	    mcb.setNextArgGeneric((int)0);
	  }
	else if (strstr(ae->utterance, "SHAKE"))
	  {
	    mcb.setNextArgGeneric((int)1);
	  }
	else if (strstr(ae->utterance, "UP"))
	  {
	    mcb.setNextArgGeneric((int)2);
	  }
	else if (strstr(ae->utterance, "DOWN"))
	  {
	    mcb.setNextArgGeneric((int)3);
	  }
	else if (strstr(ae->utterance, "GO AWAY"))
	  {
	    mcb.setNextArgGeneric((int)4);
	  }
	else if (strstr(ae->utterance, "FETCH"))
	  {
	    mcb.setNextArgGeneric((int)5);
	  }
	else if (strstr(ae->utterance, "BRUCE"))
	  {
	    mcb.setNextArgGeneric((int)6);
	  }
	else if (strstr(ae->utterance, "NEWT"))
	  {
	    mcb.setNextArgGeneric((int)7);
	  }
	else if (strstr(ae->utterance, "GOODBYE"))
	  {
	    mcb.setNextArgGeneric((int)8);
	  }
	else if (strstr(ae->utterance, "GOOD DOG"))
	  {
	    mcb.setNextArgGeneric((int)9);
	  }
	else if (strstr(ae->utterance, "BAD DOG"))
	  {
	    mcb.setNextArgGeneric((int)10);
	  }
	  printf("got here %s \n",ae->utterance);
	  speechDiagnostic(ae->utterance); 
      }
    else {
	// speechDiagnostic("no utterance");
    }
    getController()->executeMotorCommand(mcb);
}

void BehaviorPerson::updateTextDiag()
{
    if(currentRep == RepSwitch::REP_3){
	int index = 0;
	if(leftText){
	    index = 0;
	    if(extendedLeft.getValue()){
		leftText->string.set1Value(index++, "extendedLeft");
	    }
	    if(pointingLeft.getValue()){
		leftText->string.set1Value(index++, "pointingLeft");
	    }
	    if(throwingLeft.getValue()){
		leftText->string.set1Value(index++, "throwingLeft");
	    }
	    if(handOverHeadLeft.getValue()){
		leftText->string.set1Value(index++, "handOverHeadLeft");
	    }
	    if(index == 0){
		leftText->string.set1Value(index++, "");
		leftText->string.set1Value(index, "");
	    }
	    
	}
	if(rightText){
	    index = 0;
	    if(extendedRight.getValue()){
		rightText->string.set1Value(index++, "extendedRight");
	    }
	    if(pointingRight.getValue()){
		rightText->string.set1Value(index++, "pointingRight");
	    }
	    if(throwingRight.getValue()){
		rightText->string.set1Value(index++, "throwingRight");
	    }
	    if(handOverHeadRight.getValue()){
		rightText->string.set1Value(index++, "handOverHeadRight");
	    }
	    if(index == 0){
		rightText->string.set1Value(index++, "");
		rightText->string.set1Value(index++, "");
	    }
	}
	if(bodyText){
	    index = 0;
	    if(standing.getValue()){
		bodyText->string.set1Value(index++, "standing");
	    }
	    if(sitting.getValue()){
		bodyText->string.set1Value(index++, "sitting");
	    }
	    if(bendingOver.getValue()){
		bodyText->string.set1Value(index++, "bendingOver");
	    }
	    if(index == 0){
		bodyText->string.set1Value(index, "");
	    }
	}
	if(colorClassNodes){
	    int numClasses = colorClassNodes->getNumChildren();
	    int nc = (numClasses< num_colors)?
		numClasses : num_colors;
	    int ni;
	    SbVec3f p = position();
//	    printf("number of classes -> %d \n", nc);
	    for(ni = 0;ni<nc;ni++)
		((ColorClassNode *)colorClassNodes->getChild(ni))->
		    updateStateFrom(colors[ni], p);
	    if(ni<numClasses){
		for(ni = num_colors ;ni<numClasses;ni++)
		    ((ColorClassNode *)colorClassNodes->getChild(ni))->
		    updateStateNoData(p);
	    }
	}
    }
}
void BehaviorPerson::updateFromVisionServices()
{
    int i;
    static int ifUserExists = 0;

//    cerr << "updating from vision services" << endl;
    /* get new visiondata */
    Alive::visionServices->updateStateAt();
    
    // we're going to update the person using a MotorCommandBlock.
    MotorCommandBlock mcb(30002); // PersonUpdateFromMCB's magic identity number
    static char boolFields[18];
    personExists.setValue(ISDOING(user_exists)?TRUE : FALSE);
    if(personExists.getValue()){
	boolFields[0] = 2;   // person exists

	/* calculate positions for body parts  and project on floor */
	
	SbVec3f parts[3];
	SbVec3f proj, np;

	parts[LEFTHAND] = FEATURE(left_hand) + SbVec3f(0, 0, .50);
	parts[RIGHTHAND] = FEATURE(right_hand)+ SbVec3f(0, 0, .50);
	parts[HEAD] = FEATURE(head);
	np = FEATURE(centroid);

	mcb.setNextArgGeneric(parts[LEFTHAND]);

	// update right hand, head, and centroid

	mcb.setNextArgGeneric(parts[RIGHTHAND]);
	mcb.setNextArgGeneric(parts[HEAD]);
	mcb.setNextArgGeneric(np);

	// update normal width and height of user
	mcb.setNextArgGeneric(FEATURE(normal_size));
	mcb.setNextArgGeneric((int) remote_person->f3.pffeat3d_u.feats->normal_valid);

	// update color blobs
	if (num_colors < NCCLASSES)
	  {
	    delete [] colors;
	    num_colors = NCCLASSES;
	    colors = new ColorClass[num_colors];
	  }
	
	{
	  int i;
	  for(i=0; i<num_colors; i++)
	    {
	      colors[i].color_mean = SbVec3f(CCLASS(i).cmean[0],
					     CCLASS(i).cmean[1],
					     CCLASS(i).cmean[2]);
	      colors[i].geom_mean = SbVec3f(CCLASS(i).gmean[0],
					    CCLASS(i).gmean[1],
					    CCLASS(i).gmean[2]);
	      colors[i].eigenvalues = SbVec3f(CCLASS(i).gL[0],
					      CCLASS(i).gL[1],
					      CCLASS(i).gL[2]);
	      colors[i].eigenvecs[0] = SbVec3f(CCLASS(i).gS[0].x,
					       CCLASS(i).gS[0].y,
					       CCLASS(i).gS[0].z);
	      colors[i].eigenvecs[1] = SbVec3f(CCLASS(i).gS[1].x,
					       CCLASS(i).gS[1].y,
					       CCLASS(i).gS[1].z);
	      colors[i].eigenvecs[2] = SbVec3f(CCLASS(i).gS[2].x,
					       CCLASS(i).gS[2].y,
					       CCLASS(i).gS[2].z);
	    }
	}

	// calculate local transforms for 
	// kalman hand, right hand, and head

	SbVec3f lhxft = parts[LEFTHAND] - SbVec3f(np[0], 0, np[2]);
	SbVec3f rhxft = parts[RIGHTHAND] - SbVec3f(np[0], 0, np[2]);
	SbVec3f hhxft = parts[HEAD] - SbVec3f(np[0], 0, np[2]);
	mcb.setNextArgGeneric(lhxft);
	mcb.setNextArgGeneric(rhxft);
	mcb.setNextArgGeneric(hhxft);

	mcb.setNextArgGeneric(FEATURE(gaze_pos));

	// calculate local transform for the creature

	proj = SbVec3f(0, position()[1], 0) + SbVec3f(np[0], 0, np[2]);
	mcb.setNextArgGeneric(proj);

	// calculate boolean fields.
	// note that we have gotten rid of ifUserExists flag 
	// which buffered period (maybe?) when user first walked
	// into the scene

//	cerr << "standing was " << ISDOING(standing) << endl;

	boolFields[1] = ((ISDOING(extend_left)) ? 2 : 1);
	boolFields[2] = ((ISDOING(extend_right)) ? 2 : 1);
	boolFields[3] = (ISDOING(point_left) ? 2 : 1);
	boolFields[4] = (ISDOING(point_right) ? 2 : 1);
	boolFields[5] = (ISDOING(hand_moving_left) ? 2 : 1);
	boolFields[6] = (ISDOING(hand_moving_right) ? 2 : 1);
	// false advertising, these are now handOverHeadLeft,handOverHeadRight
	// and handsOverHead
	boolFields[7] = (ISDOING(kick_left) ? 2 : 1);
	boolFields[8] = (ISDOING(kick_right) ? 2 : 1);
	
	boolFields[11] = (ISDOING(standing) ? 2 : 1);
	boolFields[12] = (ISDOING(sitting) ? 2 : 1);
	boolFields[13] = (ISDOING(bending_over) ? 2 : 1);
	boolFields[15] = ((ISDOING(extend_left))? 2 : 1);
	boolFields[16] = ((ISDOING(extend_right))? 2 : 1);
	
	boolFields[9] = boolFields[10]= boolFields[14] = 1;

//	NOT USED, DERIVED FIELDS NOW	
//	boolFields[9] = ((ISDOING(point_left) && (ballInHand.getValue()))? 2:1);
//	boolFields[10] = (ISDOING(throw_right) ? 1 : 1);
//	boolFields[14] = (((boolFields[7] + boolFields[8])==4) ? 2 : 1);
	
	
	// now calculate and set values
	boolFields[17] = 0;
	
	mcb.setNextArgGeneric(boolFields);
	
	// now we execute this motor command
	getController()->executeMotorCommand(mcb);
      }
    else
      {
//	cerr << "executing null mcb" << endl;
	getController()->executeMotorCommand(mcb);  /* executing this MCB with no arguments
						       should cause the MotorSkill to realize
						       that personExists should be false. */
      }
     
	    
}
