#ifndef BEHPERSON
#define BEHPERSON
#include <Inventor/SbBasic.h>
#include "MotorLib.h"
#include "BehaviorCreatureLib.h"
#include "AudioMan.h"
#include "ColorClassNode.h"
class SoText3;



class BehaviorPerson: public BehaviorCreature {
    SO_NODE_HEADER(BehaviorPerson);

    public:
    SoSFBool updateStateFromVision;
    
    // gestural fields that get set by vision

    SoSFBool extendedLeft;
    SoSFBool extendedRight;
    SoSFBool pointingLeft;
    SoSFBool pointingRight;
    SoSFBool handOverHeadLeft;
    SoSFBool handOverHeadRight;
    
    SoSFBool handsOverHead;
    SoSFBool handLowLeft;
    SoSFBool handLowRight;
    
    SoSFBool kickingLeft;
    SoSFBool kickingRight;
    
    SoSFBool throwingEnabled;
    SoSFBool throwingLeft;
    SoSFBool throwingRight;
    
    SoSFBool standing;
    SoSFBool sitting;
    SoSFBool bendingOver;
    
    SoSFFloat throwingSpeed;
    SoSFBool personExists;
    SoSFBool ballInHand;
    SoSFBool somethingInHand;
    // fields that get set by audio input

    SoSFBool isUtteringSit;
    SoSFBool isUtteringShake;
    SoSFBool isUtteringUp;
    SoSFBool isUtteringDown;
    SoSFBool isUtteringGoAway;
    SoSFBool isUtteringFetch;
    SoSFBool isUtteringBruce;
    SoSFBool isUtteringNewt;
    SoSFBool isUtteringGoodbye;
    SoSFBool isUtteringGoodDog;
    SoSFBool isUtteringBadDog;
    
    SoSFBool userGesturing;

    // positions of head, hands and centroid
    SoSFVec3f head;
    SoSFVec3f rightHand;
    SoSFVec3f leftHand;
    SoSFVec3f bodyCentroid;
    
    // notion of how big the person is
    SoSFFloat normalWidth;
    SoSFFloat normalHeight;
    SoSFBool lastNormalValid;  // keeps last frame's value
    SoSFBool normalValid;

    // translation for the gaze block's transform
    SoSFVec3f gazeTranslation;

    SoSFVec3f sysNoise;
    SoSFVec3f obsNoise;
    
    SoSFInt32 numReps;
    
    // use audio-input
    SoSFBool useSpeech;
    AudioMan *audioMan;
    
    // screen tolerance for picking stuff up
    SoSFFloat screenTolerance;
    
    // transformations for graphical representations
    // of head and hands

    SoTransform *lhxf;
    SoTransform *rhxf;
    SoTransform *hhxf;
    
    // transform for the gaze block
    
    SoTransform *gbxf;

    SoText3 *leftText;
    SoText3 *rightText;
    SoText3 *bodyText;
        
    int num_colors;
    ColorClass *colors;
    
    void speechDiagnostic(char *s);
    void updateTextDiag();

    static void initClass();
    virtual void addMotorSkills();
    virtual void updateBehaviorSystemAt(UpdateStateData &usd);
    virtual void changeRep(int newRep);
    void updateFromSpeech();
    void updateFromVisionServices();
    protected:
    SbBool partsInstalled;
    
    BehaviorPerson();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    SbBool readInstance(SoInput *in, unsigned short flags);
#else
    SbBool readInstance(SoInput *in);
#endif
    virtual ~BehaviorPerson();
    
};

#endif      

