#include "ColorClassNode.h"
#include "BaseCreature.h"
#include "BehaviorPerson.h"
#include <Inventor/nodes/SoScale.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/SbLinear.h>
#include <Inventor/fields/SoSFString.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/elements/SoElements.h>
#include <Inventor/elements/SoSwitchElement.h>

SO_NODE_SOURCE(ColorClassNode);

void
ColorClassNode::initClass()
{
    SO_NODE_INIT_CLASS(ColorClassNode, SoSeparator, "Separator");
}

ColorClassNode::ColorClassNode() : SoSeparator()
{
    SO_NODE_CONSTRUCTOR(ColorClassNode);
    addChild(new SoTransform);
    addChild(new SoBaseColor);
    addChild(new SoSphere);
	
}

ColorClassNode::~ColorClassNode()
{
}

SbBool
#ifdef SO_VERSION_REVISION
ColorClassNode::readInstance(SoInput *in, unsigned short flags)
#else
ColorClassNode::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoSeparator::readInstance(in, flags)) {
#else
    if (SoSeparator::readInstance(in)) {
#endif
    return (1);
    }
    return (0);
}

void ColorClassNode::print()
{
	printf("geom_mean %f %f \n", 
	data.geom_mean[0], data.geom_mean[1]);
	
	printf("color_mean %f %f %f \n", 
	data.color_mean[0], data.color_mean[1],  data.color_mean[2]);
	
	printf("eigenvalue & vec 1 %f %f %f \n", 
	data.eigenvalues[0], 
	data.eigenvecs[0][0],  data.eigenvecs[0][1]);
	
	printf("eigenvalue & vec 2 %f %f %f \n", 
	data.eigenvalues[1], 
	data.eigenvecs[1][0],  data.eigenvecs[1][1]); 
}

void ColorClassNode::updateStateFrom(ColorClass &cc, SbVec3f base)
{
    SoTransform *xf = (SoTransform *)getChild(0);
    SoBaseColor *bc = (SoBaseColor *)getChild(1);
    
    data = cc;
    xf->translation.setValue(SbVec3f(data.geom_mean[0] - base[0],
    data.geom_mean[1] - base[1], 
    0 ));
    xf->rotation.setValue(SbRotation(SbVec3f(1, 0, 0), 
	SbVec3f(data.eigenvecs[0][0],data.eigenvecs[0][1], 0)));
    xf->scaleFactor.setValue(data.eigenvalues[0],data.eigenvalues[1], 1); 
    
    bc->rgb.setValue((1.0/255.0)*data.color_mean);
    
//    print();
    
}

void ColorClassNode::updateStateNoData(SbVec3f base)
{
    SoTransform *xf = (SoTransform *)getChild(0);
    SoBaseColor *bc = (SoBaseColor *)getChild(1);
    
    xf->translation.setValue(0, 0, 0);
    xf->rotation.setValue(SbRotation(SbVec3f(1, 0, 0), 
	SbVec3f(data.eigenvecs[0][0],data.eigenvecs[0][1], 0)));
    xf->scaleFactor.setValue(.1, .1, .1); 
    
    bc->rgb.setValue(SbVec3f(1, 1, 1));
    
//   print();
    
}
