#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoFaceSet.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/Xt/SoXtRenderArea.h>
#include <Inventor/nodes/SoCallback.h>
#include "TimeStat.h"
#include "AliveCamera.h"
#include "ContourServer.h"
#include "Shadow.h"
#include <Inventor/elements/SoCacheElement.h>

// #define DOTIMING
extern int chromakey;
static int showContour = 0;
int contourserver_use_kalman = 1;
#ifdef PFC
#include "Alive.h"
#include <rpc/rpc.h>
extern "C" {
#include "pfinder_clnt.h"
}
#endif


/* constructor */

ContourServer::ContourServer(char *serverName, SoSeparator *scene, 
    AliveCamera *cameraMapping, int layers)
{
    contourServerName = serverName;
    mapping = cameraMapping;
    
    buildContourFaceSet();
    if(scene)
	scene->addChild(sep);

    openServer();
    swf = makeSWFilter(FILTER_WINDOW);	
    scale = 1.0;
}

#ifdef PFC
CLIENT *remote_person_server = NULL;
pfperson *remote_person = NULL;
#endif


void ContourServer::openServer()
{

#ifdef PFC
	if (Alive::visionServerName) {
		if (!(remote_person_server =
		      clnt_create(Alive::visionServerName,
				  PFINDERPROG, PFINDERVERS, "tcp"))) {
			clnt_pcreateerror( Alive::visionServerName );
			exit(1);
		}
	}
#endif
}

void initremotepersonserver() {
#ifdef PFC
      int ntries;
      if (!remote_person_server) {
	    printf("Connecting to pfinder host %s...",Alive::visionServerName);
	    for (ntries = 0; !remote_person_server && ntries<20; ntries++) {
		  remote_person_server = clnt_create( Alive::visionServerName, PFINDERPROG, 
				   PFINDERVERS, "tcp");
		  if (!remote_person_server) {
			clnt_pcreateerror( Alive::visionServerName );
			sleep(1);
		  }
	    }
	    if (!remote_person_server) puts("error, couldn't connect, aborting..."), exit(-1);
      }
#endif
}

void ContourServer::closeServer()
{
}

void ContourServer::resetServer()
{
#ifdef PFC
  pfinit_1( NULL, remote_person_server );
  clnt_destroy(remote_person_server);
  remote_person_server = NULL;
#endif
}

void ContourServer::abortServer()
{
}

int ContourServer::grabAndGetContour()
{
#ifdef PFC
    int i;
    pfopts pfo;
    pfv3f *pts;
    bzero(&pfo, sizeof(pfo));
    pfo.sync = 0; 
    pfo.kalman = contourserver_use_kalman;
    pfo.cc3d = pfo.c3 = pfo.f3 = pfo.g = pfo.alive_coords = 1;
    if (!remote_person_server)
      initremotepersonserver();  // for pfinder reconnection capability
  retry:
    remote_person = pf_1(&pfo, remote_person_server);
    if (!remote_person) {
      clnt_perror(remote_person_server, Alive::visionServerName);
      clnt_destroy(remote_person_server);
      remote_person_server = NULL;
      initremotepersonserver();
      goto retry;
    }
    numPts = remote_person->c3.pfcont3d_u.points.points_len;
    pts =  remote_person->c3.pfcont3d_u.points.points_val;
    //printf("numPts -> %d \n", numPts);
    for (i=0; i<numPts; i++) {
	    coords3d[i][0]=pts[i].x;
	    coords3d[i][1]=pts[i].y;
	    coords3d[i][2]=pts[i].z;
    }
    xf->translation.setValue(remote_person->f3.pffeat3d_u.feats->centroid.x,
			     remote_person->f3.pffeat3d_u.feats->centroid.y,
			     remote_person->f3.pffeat3d_u.feats->centroid.z);
//    pfgest_s *pg = remote_person->g.pfgest_u.bits;
//    printf("pg->extended_left = %d\n", pg->extend_left);

    if(((numPts-1)<0)&& vp->vertex.getNum()){
	vp->vertex.deleteValues(0, -1);
	fs->numVertices.set1Value(0,-1);
    }
    else if(numPts-1>0){
	vp->vertex.setValues(0,numPts-1,coords3d);
	fs->numVertices.set1Value(0,numPts-1);
    }
    return numPts;
#endif
}

void
ContourServer::setScale(float sc)
{
  scale = sc;
}
void ContourServer::useContour(SbBool yesOrNo)
{
    contourSwitch->whichChild.setValue((yesOrNo)?-3 : -1);
}

void ContourServer::showContour(SbBool yesOrNo)
{
    ::showContour = (yesOrNo)?1 :0;
    printf(" ::showContour %d \n", ::showContour);
    if(::showContour){
	vp->orderedRGBA.set1Value(0, SbColor(0, 1, 1).getPackedValue(0));
    }
    else{
	if(Alive::chromakey)
	    vp->orderedRGBA.set1Value(0, SbColor(1, .3, 1).getPackedValue(0));
	else
	    vp->orderedRGBA.set1Value(0, SbColor(0, 0, 0).getPackedValue(1));
    }
}	

void ContourServer::buildContourFaceSet()
{
    int numvertices[1];
    SoDrawStyle *ds = new SoDrawStyle;
    sep = new SoSeparator;
    contourSwitch = new SoSwitch;
    contourSwitch->whichChild.setValue(-3);
    sep->addChild(contourSwitch);
    sep->renderCaching.setValue(SoSeparator::OFF);
    
    ds->style.setValue(SoDrawStyle::FILLED);
    contourSwitch->addChild(ds);
    
    xf = new SoTransform;
    xf->translation.setValue(0, 0, 0);
    xf->scaleFactor.setValue(.95, .95, 1);
    contourSwitch->addChild(xf);
    contourSwitch->ref();
    
    // use base color as lighting model 
    SoLightModel *lm = new SoLightModel;
    lm->model = SoLightModel::BASE_COLOR;
    contourSwitch->addChild(lm);
  
    SoShapeHints *shapeHints = new SoShapeHints;
    shapeHints->faceType = SoShapeHints::UNKNOWN_FACE_TYPE;
    shapeHints->vertexOrdering = SoShapeHints::CLOCKWISE;
    contourSwitch->addChild(shapeHints);
   
    vp = new SoVertexProperty;
    // set bindings
    vp->normalBinding.setValue(SoVertexProperty::OVERALL);
    vp->materialBinding.setValue(SoVertexProperty::OVERALL);
    vp->normal.set1Value(0,SbVec3f(0,0, 1));
    
    // do color
    vp->orderedRGBA.set1Value(0, SbColor(0, 0, 0).getPackedValue(1));
    contourSwitch->addChild(vp);
    
    fs = new SoFaceSet;
    fs->numVertices.set1Value(0,-1);
    contourSwitch->addChild(fs);
}
