#ifndef CONTOURCLASS
#define CONTOURCLASS

#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoVertexProperty.h>
#include "common.h"
#include "packet.h"
#include "ring_buffer.h"
#include "filter.h"
#include "contour.h"
#include "Gesture.h"
#ifndef PFC
#include "SocketClient.h"
#endif
// can these stay outside and compile under C++?
extern "C" {
#include "contour.proto"
#include "ring_buffer.proto"
#include "filter.proto"
#include "common.proto"
#include "packet.proto"
#include "find_features.proto"
#include "track_features.proto"
}


#define NUMFLOATSPERVERTEX COORDS_PER_POINT // incldued in packet.h
#define MAXALPHALAYERS 10
#define FEATHER .05
#define FILTER_WINDOW 3 /* Number of points to use in average bottom */

class SoSeparator;
class AliveCamera;
class SoFaceSet;
class SoNode;
class SoTransform;
class SoCallback;

struct ScreenFeatures
{
  float centroid[2];
  float boxMin[2];
  float boxMax[2];
  float robotCentroid[2];
};

class ContourServer {
    public:
    char *contourServerName;
    Contour cont;
    ContourPacket cp;

    SWFilter *swf; /* stores the filter bounding box bottom */
    AliveCamera *mapping;
    Features features; // world
    ScreenFeatures screen_features; // screen
    int numExtremities;
    int numPts; /* in contour */
    float contourCoords[5000*COORDS_PER_POINT]; // screen
    float extremityCoords[NUMBER_OF_EXTREMA*COORDS_PER_POINT]; // screen
    SbVec3f robot3d;  /* World coordinates of robot */
    SbVec3f coords3d[5000]; /* World coords of contour */
    float scale; // camera to world 
    
    // drawing stuff
    SoSeparator *sep;
    SoSwitch *contourSwitch;
    SoTransform *xf;
    SoFaceSet *fs;
    SoVertexProperty *vp;
    SbVec3f centroid;

    // protos

    // builds initial contour and opens connection
    ContourServer(char *serverName, SoSeparator *scene, 
	AliveCamera *cameraMapping, int layers = 0);
    
    // open and closes connection to server
    void openServer();
    void closeServer();
   

    // these preempt closeServer above to provide
    // new functionality for resetting the worlds
    void resetServer();
    void abortServer();

    // turns contour on or off
    void useContour(SbBool yesOrNo);
    void showContour(SbBool yesOrNo);
	
    // does both a grab and a get
    int grabAndGetContour();
	
    // units/inch
    void ContourServer::setScale(float sc);
 
 private:
    void buildContourFaceSet();   
};
#endif

















