#include "Gesture.h"


Gesture::Gesture(char *filename)
{
  gp = makeGestureParams();
  param_file = stringCopy(filename);
  loadGestureParams(gp, filename);
  model = makeUserModel(gp, makeCameraToWorld());
}


Gesture::~Gesture()
{
  freeGestureParams(gp);
  freeUserModel(model);
}


void Gesture::reloadParameters(void)
{
  loadGestureParams(gp, param_file);
}


void
Gesture::updateModel(Features *features, float scale)
{
  static float lookup_table[] = {0.0, 0.75, 1.0};
  float left_confidence, right_confidence;

  left_confidence = lookup_table[features->left_hand_valid];
  right_confidence = lookup_table[features->right_hand_valid];

  updateUserModel(model, 
		  features->centroid[0]/scale, 
		  features->centroid[1]/scale, 
		  features->topLeft[0]/scale,
		  features->bottomRight[0]/scale,
		  features->topLeft[1]/scale,
		  features->bottomRight[1]/scale,
		  features->left_hand[0]/scale,
		  features->left_hand[1]/scale,
		  features->right_hand[0]/scale,
		  features->right_hand[1]/scale,
		  features->left_foot[0]/scale,
		  features->left_foot[1]/scale,
		  features->right_foot[0]/scale,
		  features->right_foot[1]/scale,
		  left_confidence,
		  right_confidence);
}


void
Gesture::userGone(void)
{
	resetUserModel(model);
	
}

void
Gesture::findGestures(GestureFlags *desired, GestureFlags *found)
{
  findGesturesUserModel(model, desired, found);
}





