#ifndef __GESTURE__
#define __GESTURE__

#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoCoordinate3.h>

#include "common.h"
#include "packet.h"
#include "ring_buffer.h"
#include "filter.h"
#include "user_model.h"

extern "C" {
#include "ring_buffer.proto"
#include "filter.proto"
#include "common.proto"
#include "user_model.proto"
}


struct Features
{
  SbVec3f left_hand;
  SbVec3f right_hand;
  SbVec3f left_foot;
  SbVec3f right_foot;
  SbVec3f head;
  SbVec3f centroid;
  SbVec3f topLeft;
  SbVec3f bottomRight;

  boolean left_hand_valid;
  boolean right_hand_valid;
  boolean left_foot_valid;
  boolean right_foot_valid;
  boolean head_valid;
    
};


class Gesture {
public:
	Gesture(char *param_file);
	~Gesture();
	
	// called every step to update model
	void updateModel(Features *features, float scale);
	
	// called to get gestures
	void findGestures(GestureFlags *desired, GestureFlags *found);
	
	// call when the user leaves the FOV of the camera.
	void  Gesture::userGone(void);
	
	// reload gesture parameters from file
	void reloadParameters(void);

	// observers for average height and width, in inches
  float userHeight() { return model->normal_height; };
  float userWidth() { return model->normal_width; };
  
private:
	UserModel *model;
	GestureParams *gp;
	char *param_file;
};

#endif
