#include "Person.h"
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoLists.h>
#include <Inventor/nodes/SoSwitch.h>
#define USEVISION
#ifdef USEVISION
#include "VisionServices.h"
#include "Alive.h"
#include "PersonMotorSkills.h"
#ifdef PFC
#include <rpc/rpc.h>
#include "pfinder_clnt.h"
#endif
#endif

SoTransform *rhxf = NULL;
SoTransform *lhxf = NULL;

/* PERSON LABELS */


#define RIGHTHAND   1
#define LEFTHAND    0
#define HEAD    2
#define NOTHING     -1
#define DELTALOW    4.0
#define FEET        2


#ifdef USEVISION
#ifndef PFC
#define FEATURE(which) Alive::visionServices->contourServer->features.which

#define ISDOING(which) ((((unsigned char)Alive::visionServices->found.which) == 1)?1:0)
#define EXTENDEDLEFT (ISDOING(extend_left)&& (!ISDOING(hand_moving_left)) && !(ISDOING(point_left)))
#define EXTENDEDRIGHT (ISDOING(extend_right) &&( !ISDOING(hand_moving_right))&& (!ISDOING(point_right)))
#else
#define FEATURE(x) cvt_pfv3f_sbvec3f(x)
#define ISDOING(x) x
#endif
#endif

#ifdef PFC
extern pfperson *remote_person;
extern CLIENT *remote_person_server;
SbVec3f cvt_pfv3f_sbvec3f(pfv3f v) {
	SbVec3f r;
	r[0] = v.x, r[1] = v.y, r[2] = v.z;
	return r;
}
#endif


#define CHECKMAINBODY(x) ((mainBody != NULL) && (mainBody->x))?1:0
/*****************************************************************************
 * 
 *			Definition of PersonPart 
 * 
 ****************************************************************************/
SO_NODE_SOURCE(PersonPart);

void
PersonPart::initClass()
{
    SO_NODE_INIT_CLASS(PersonPart, BaseCreature, "BaseCreature");
}

PersonPart::PersonPart() : BaseCreature()
{
    SO_NODE_CONSTRUCTOR(PersonPart);
    SO_NODE_ADD_FIELD(whichPart, (0));   
}

PersonPart::~PersonPart()
{
}

void PersonPart::updateStateAt(UpdateStateData &usd)
{
    BaseCreature::updateStateAt(usd);
}

int PersonPart::isPatting()
{
   return CHECKMAINBODY(sitting.getValue());
}

int PersonPart::isThisExtended()
{
    return (whichPart.getValue()== LEFTHAND)?
	CHECKMAINBODY(extendedLeft.getValue())
	:CHECKMAINBODY(extendedRight.getValue());
}

int PersonPart::isAnyExtendedLeft()
{
    return CHECKMAINBODY(extendedLeft.getValue());
}

int PersonPart::isAnyExtendedRight()
{
    return CHECKMAINBODY(extendedRight.getValue());
}

int PersonPart::isAnyExtendedLeftAndLow()
{
    int r1 = CHECKMAINBODY(handLowLeft.getValue());
    int r2 = isAnyExtendedLeft();
    return (r1 && r2)?TRUE : FALSE;
}

int PersonPart::isAnyExtendedRightAndLow()
{
    int r1 = CHECKMAINBODY(handLowRight.getValue());
    int r2 = isAnyExtendedRight();
    return (r1 && r2)?TRUE : FALSE;
}

int PersonPart::isThisExtendedAndLow()
{
    return (whichPart.getValue()== LEFTHAND)
	? isAnyExtendedLeftAndLow()
	: isAnyExtendedRightAndLow();
}

int PersonPart::isThisHandMoving()
{
    return (whichPart.getValue()== LEFTHAND)?
	CHECKMAINBODY(handMovingLeft.getValue())
	:CHECKMAINBODY(handMovingRight.getValue());
}

int PersonPart::isThisPointing()
{
    return (whichPart.getValue()== LEFTHAND)?
	CHECKMAINBODY(pointingLeft.getValue())
	:CHECKMAINBODY(pointingRight.getValue());
}

int PersonPart::isAnyPointingLeft()
{
    return CHECKMAINBODY(pointingLeft.getValue());
}

int PersonPart::isAnyPointingRight()
{
    return CHECKMAINBODY( pointingRight.getValue());
}

int PersonPart::handsOverHead()
{
    return CHECKMAINBODY(handsOverHead.getValue());
}

/*****************************************************************************
 * 
 *			Definition of Person 
 * 
 ****************************************************************************/
SO_NODE_SOURCE(Person);

void
Person::initClass()
{
    SO_NODE_INIT_CLASS(Person, MotorCreature, "MotorCreature");
}

Person::Person() : MotorCreature()
{
    SO_NODE_CONSTRUCTOR(Person);
    
    SO_NODE_ADD_FIELD(updateStateFromVision, (FALSE));
    SO_NODE_ADD_FIELD(extendedLeft, (FALSE));
    SO_NODE_ADD_FIELD(extendedRight, (FALSE));
    SO_NODE_ADD_FIELD(pointingLeft, (FALSE));
    SO_NODE_ADD_FIELD(pointingRight, (FALSE));
    SO_NODE_ADD_FIELD(handMovingLeft, (FALSE));
    SO_NODE_ADD_FIELD(handMovingRight, (FALSE));
    SO_NODE_ADD_FIELD(handLowLeft, (FALSE));
    SO_NODE_ADD_FIELD(handLowRight, (FALSE));
    SO_NODE_ADD_FIELD(kickingLeft, (FALSE));
    SO_NODE_ADD_FIELD(kickingRight, (FALSE));
    SO_NODE_ADD_FIELD(throwingLeft, (FALSE));
    SO_NODE_ADD_FIELD(throwingRight, (FALSE));
     SO_NODE_ADD_FIELD(throwingEnabled, (FALSE));
    SO_NODE_ADD_FIELD(standing, (FALSE));
    SO_NODE_ADD_FIELD(sitting, (FALSE));
    SO_NODE_ADD_FIELD(bendingOver, (FALSE));
    SO_NODE_ADD_FIELD(handsOverHead, (FALSE));
    SO_NODE_ADD_FIELD(leftHand, (0, 0, 0));
    SO_NODE_ADD_FIELD(rightHand, (0, 0, 0));
    SO_NODE_ADD_FIELD(head, (0, 0, 0));
    SO_NODE_ADD_FIELD(bodyCentroid, (0, 0, 0));
    SO_NODE_ADD_FIELD(throwingSpeed, (60.0));
    SO_NODE_ADD_FIELD(personExists, (FALSE));
    partsInstalled = FALSE;   
}

#ifdef SO_VERSION_REVISION
SbBool Person::readInstance(SoInput *in, unsigned short flags) 
#else
SbBool Person::readInstance(SoInput *in) 
#endif
{
    PersonPart *pp;
    
#ifdef SO_VERSION_REVISION
    if (MotorCreature::readInstance(in, flags)) {
#else
    if (MotorCreature::readInstance(in)) {
#endif
	SoSearchAction sa;
	sa.setType(PersonPart::getClassTypeId());
	sa.setInterest(SoSearchAction::ALL);
	sa.apply(this);
	SoPathList ucp;
	ucp = sa.getPaths();
	printf(" %d person parts found \n", ucp.getLength());
	int i;
	for(i=0;i<ucp.getLength();i++){
	    pp = (PersonPart *)ucp[i]->getTail();
	    pp->ref();
	    pp->ref();
	    pp->whichPart.setValue(i);
	    removeChild(pp);
	    personParts.append(pp);
	}
	((SoSwitch *)switchNode)->whichChild.setValue(-3);
	SoPath *tmp_path = NULL;
	tmp_path = getPathToName(this, "RHAND");
	if(tmp_path){
	    rhxf = (SoTransform *)tmp_path->getTail();
	}
	tmp_path = getPathToName(this, "LHAND");
	if(tmp_path){
	    lhxf = (SoTransform *)tmp_path->getTail();
	}
	((SoSwitch *)switchNode)->whichChild.setValue(defaultSwitchValue.getValue());
	return (1);
   } else {
	return(0);
   }
}

void Person::addMotorSkills()
{
    MotorCreature::addMotorSkills();
    BallThrow *bt = new BallThrow(this, "throwball");
    getController()->addCommand(30001, "throwball", bt, TRUE, 
				ArgType((float)10), ArgType((float)0), 
				ArgType((float).95));
}

Person::~Person()
{
}

void Person::updateStateAt(UpdateStateData &usd)
{
    int i;
    if(partsInstalled == FALSE){
	partsInstalled = TRUE;
	for(i=0;i<personParts.getLength();i++){
	    PersonPart *pp = (PersonPart *)personParts[i];
	    pp->mainBody = this;
	    usd.currCreatureCache->addBaseCreature(pp);
	}
    }
    BaseCreature::updateStateAt(usd);
    if(updateStateFromVision.getValue()||
	Alive::connectToVision)
	updateFromVisionServices();
    /*
     * Now do any processing that is independent of source
     */
    if(throwingEnabled.getValue() &&
    (throwingLeft.getValue() || throwingRight.getValue())){
	// issue throwing command
	MotorCommandBlock mcb;
	mcb.cmd = 30001; 
	mcb.setNextArgGeneric((throwingLeft.getValue())?
	    leftHand.getValue() : rightHand.getValue()); 
	mcb.setNextArgGeneric((float)((throwingLeft.getValue())
	    ?-throwingSpeed.getValue() : throwingSpeed.getValue()));
	mcb.setNextArgGeneric((float)(0.0));
	mcb.setNextArgGeneric((float)(0.0));
	controller->executeMotorCommand(mcb);
    }
    else {
	throwingLeft.setValue(FALSE);
	throwingRight.setValue(FALSE);
    }
    updateMotorSkillsAt(usd);
}


void Person::updateFromVisionServices()
{
    int i;
    static int ifUserExists = 0;
#ifdef USEVISION    
#ifdef PFC
    Alive::visionServices->updateStateAt();

    if (!remote_person) {
	    fprintf(stderr,"NO RP!\n");
	    return;
    }
    pffeat3d_s *pf3 = remote_person->f3.pffeat3d_u.feats;
    pfgest_s *pg = remote_person->g.pfgest_u.bits;
    
    /* calculate positions for body parts  and project on floor */
    SbVec3f parts[3];
    SbVec3f proj, np;
    
    parts[LEFTHAND] = FEATURE(pf3->left_hand) + SbVec3f(0, 0, .50);
    parts[RIGHTHAND] = FEATURE(pf3->right_hand)+ SbVec3f(0, 0, .50);
    parts[HEAD] = FEATURE(pf3->head);
    np = FEATURE(pf3->centroid);
     
    leftHand.setValue(parts[LEFTHAND]);
    rightHand.setValue(parts[RIGHTHAND]);
    if(lhxf){
	lhxf->translation.setValue(leftHand.getValue()-SbVec3f(np[0], 0, np[2]));
	
    }
    if(rhxf){
	rhxf->translation.setValue(rightHand.getValue()-SbVec3f(np[0], 0, np[2]));
    }
    head.setValue(parts[HEAD]);
    bodyCentroid.setValue(np);
    
#define NEWPOSCODE
#ifdef NEWPOSCODE

    /* first body */
    proj = SbVec3f(0, position()[1], 0) + SbVec3f(np[0], 0, np[2]);
    setPosition(proj + SbVec3f(0, 0, 0));
    
    /* now rest of parts
    for(i=0;i<2;i++){
	pp = (PersonPart *)personParts[i];
	proj = SbVec3f(0, pp->position()[1], 0) + SbVec3f(parts[i][0], 0, parts[i][2]);
	pp->setPosition(proj);
    }
    if(personParts.getLength()>2){
	pp = (PersonPart *)personParts[2];
	pp->setPosition(Alive::visionServices->contourServer->robot3d);
    }
    */
#else
    /* first body */
    SbVec3f pos = basexf->translation.getValue();
    proj = SbVec3f(0, pos[1], 0) + SbVec3f(np[0], 0, np[2]);
    basexf->translation.setValue(proj + SbVec3f(0, 0, 0));
#endif    
    
    /* now set flags */
    ifUserExists += ISDOING(pg->user_exists)?1 : 0;
    if(ifUserExists>20.0){
	extendedLeft.setValue(ISDOING(pg->extend_left)?TRUE : FALSE);
	extendedRight.setValue(ISDOING(pg->extend_right)?TRUE : FALSE);
	pointingLeft.setValue(ISDOING(pg->point_left)?TRUE : FALSE);
	pointingRight.setValue(ISDOING(pg->point_right)?TRUE : FALSE);
	handMovingLeft.setValue(ISDOING(pg->hand_moving_left)?TRUE : FALSE);
	handMovingRight.setValue(ISDOING(pg->hand_moving_right)?TRUE : FALSE);
	
	kickingLeft.setValue(ISDOING(pg->kick_left)?TRUE : FALSE);
	kickingRight.setValue(ISDOING(pg->kick_right)?TRUE : FALSE);
	
	throwingLeft.setValue(ISDOING(pg->throw_left)?TRUE : FALSE);
	throwingRight.setValue(ISDOING(pg->throw_right)?TRUE : FALSE);
	
	standing.setValue(ISDOING(pg->standing)?TRUE : FALSE);
	sitting.setValue(ISDOING(pg->sitting)?TRUE : FALSE);
	bendingOver.setValue(ISDOING(pg->bending_over)?TRUE : FALSE);
    }
    personExists.setValue(ISDOING(pg->user_exists)?TRUE : FALSE);
    
#ifdef USEVALID
    rightHand_valid = /*FEATURE*/(pf3->right_hand_valid);
    leftHand_valid =  /*FEATURE*/(pf3->left_hand_valid);
    head_valid =      /*FEATURE*/(pf3->head_valid);
#endif
    /* calculated values */
    SbVec3f hh = FEATURE(pf3->head);
    SbVec3f d1 = FEATURE(pf3->left_hand)-FEATURE(pf3->head);
    SbVec3f d2 = FEATURE(pf3->right_hand)-FEATURE(pf3->head);
    int result = (d1[1]>-(DELTALOW) && d2[1]>-(DELTALOW) &&
	(ISDOING(pg->user_exists)?1:0));
    handsOverHead.setValue((result)?1:0);
    
    SbVec3f diff = FEATURE(pf3->left_hand)-FEATURE(pf3->centroid);
    
    handLowLeft.setValue((diff[1]<DELTALOW)?TRUE:FALSE);
    diff = FEATURE(pf3->right_hand)-FEATURE(pf3->centroid);
    handLowRight.setValue((diff[1]<DELTALOW)?TRUE:FALSE);
#else /* PFC */

    /* get new visiondata */
    Alive::visionServices->updateStateAt();
    
    /* calculate positions for body parts  and project on floor */
    SbVec3f parts[3];
    SbVec3f proj, np;
    PersonPart *pp;
    
    parts[LEFTHAND] = FEATURE(left_hand) + SbVec3f(0, 0, .50);
    parts[RIGHTHAND] = FEATURE(right_hand)+ SbVec3f(0, 0, .50);
    parts[HEAD] = FEATURE(head);
    np = FEATURE(centroid);
     
    leftHand.setValue(parts[LEFTHAND]);
    rightHand.setValue(parts[RIGHTHAND]);
    if(lhxf){
	lhxf->translation.setValue(leftHand.getValue()-SbVec3f(np[0], 0, np[2]));
	
    }
    if(rhxf){
	rhxf->translation.setValue(rightHand.getValue()-SbVec3f(np[0], 0, np[2]));
    }
    head.setValue(parts[HEAD]);
    bodyCentroid.setValue(np);
    
   

    /* first body */
    proj = SbVec3f(0, position()[1], 0) + SbVec3f(np[0], 0, np[2]);
    setPosition(proj + SbVec3f(0, 0, 0));
    
    /* now rest of parts
    for(i=0;i<2;i++){
	pp = (PersonPart *)personParts[i];
	proj = SbVec3f(0, pp->position()[1], 0) + SbVec3f(parts[i][0], 0, parts[i][2]);
	pp->setPosition(proj);
    }
    if(personParts.getLength()>2){
	pp = (PersonPart *)personParts[2];
	pp->setPosition(Alive::visionServices->contourServer->robot3d);
    }
    */
    /* now set flags */
    ifUserExists += ISDOING(user_exists)?1 : 0;
    if(ifUserExists>20.0){
	extendedLeft.setValue((EXTENDEDLEFT)?TRUE : FALSE);
	extendedRight.setValue((EXTENDEDRIGHT)?TRUE : FALSE);
	pointingLeft.setValue(ISDOING(point_left)?TRUE : FALSE);
	pointingRight.setValue(ISDOING(point_right)?TRUE : FALSE);
	handMovingLeft.setValue(ISDOING(hand_moving_left)?TRUE : FALSE);
	handMovingRight.setValue(ISDOING(hand_moving_right)?TRUE : FALSE);
	
	kickingLeft.setValue(ISDOING(kick_left)?TRUE : FALSE);
	kickingRight.setValue(ISDOING(kick_right)?TRUE : FALSE);
	
	throwingLeft.setValue(ISDOING(throw_left)?TRUE : FALSE);
	throwingRight.setValue(ISDOING(throw_right)?TRUE : FALSE);
	
	
	
	standing.setValue(ISDOING(standing)?TRUE : FALSE);
	sitting.setValue(ISDOING(sitting)?TRUE : FALSE);
	bendingOver.setValue(ISDOING(bending_over)?TRUE : FALSE);
    }
    personExists.setValue(ISDOING(user_exists)?TRUE : FALSE);
    
    /* calculated values */
    SbVec3f hh = FEATURE(head);
    SbVec3f d1 = FEATURE(left_hand)-FEATURE(head);
    SbVec3f d2 = FEATURE(right_hand)-FEATURE(head);
    int result = (d1[1]>-(DELTALOW) && d2[1]>-(DELTALOW) &&
	(ISDOING(user_exists)?1:0));
    handsOverHead.setValue((result)?1:0);
    
    SbVec3f diff = FEATURE(left_hand)-FEATURE(centroid);
    
    handLowLeft.setValue((diff[1]<DELTALOW)?TRUE:FALSE);
    diff = FEATURE(right_hand)-FEATURE(centroid);
    handLowRight.setValue((diff[1]<DELTALOW)?TRUE:FALSE);
#endif /* PFC */
#endif /* USEVISION */
    
}
