#ifndef PERSON
#define PERSON
#include <Inventor/SbBasic.h>
#include "MotorLib.h"
#include "CreatureLib.h"

class Person;

class PersonPart: public BaseCreature {
    SO_NODE_HEADER(BaseCreature);

    public:
    SoSFInt32 whichPart;
    
    Person *mainBody;
    
    int isPatting();
    int isThisExtended();
    int isThisPointing();  
    int isAnyPointingLeft();
    int isAnyPointingRight();
    int isAnyExtendedLeft();
    int isAnyExtendedRight();
    int isAnyExtendedLeftAndLow();
    int isAnyExtendedRightAndLow();
    int isThisExtendedAndLow();
    int isThisHandMoving();
    int handsOverHead();
    static void initClass();
    virtual void updateStateAt(UpdateStateData &usd);
    
    protected:
    
    PersonPart();       // Note only should be call during a read
    virtual ~PersonPart();

};

class Person: public MotorCreature {
    SO_NODE_HEADER(MotorCreature);

    public:
    SoSFBool updateStateFromVision;
    
    SoSFBool extendedLeft;
    SoSFBool extendedRight;
    SoSFBool pointingLeft;
    SoSFBool pointingRight;
    SoSFBool handMovingLeft;
    SoSFBool handMovingRight;
    
    SoSFBool handsOverHead;
    SoSFBool handLowLeft;
    SoSFBool handLowRight;
    
    SoSFBool kickingLeft;
    SoSFBool kickingRight;
    
    SoSFBool throwingEnabled;
    SoSFBool throwingLeft;
    SoSFBool throwingRight;
    
    SoSFBool standing;
    SoSFBool sitting;
    SoSFBool bendingOver;
    
    SoSFFloat throwingSpeed;
    SoSFBool personExists;
    
    // positions of head, hands and centroid
    SoSFVec3f head;
    SoSFVec3f rightHand;
    SoSFVec3f leftHand;
    SoSFVec3f bodyCentroid;
    
    SbPList personParts;
    static void initClass();
    virtual void addMotorSkills();
    virtual void updateStateAt(UpdateStateData &usd);
    void updateFromVisionServices();
    protected:
    SbBool partsInstalled;
    
    Person();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    SbBool Person::readInstance(SoInput *in, unsigned short flags);
#else
    SbBool Person::readInstance(SoInput *in);
#endif
    virtual ~Person();

};

#endif      

