#include <iostream.h>
#include <string.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/SoPath.h>

#include "MotorLib.h"
#include "BasicDefines.h"
#include "PersonMotorSkills.h"
#include "CreatureLib.h"
#include "BehaviorBall.h"
#include "BehaviorPerson.h"

/*********************************************************************
 * 
 * BallThrow Skill
 * arg0 : speed horizontal (+ or - x)
 * arg1 : vertical speed
 * arg2 : restitution
 * 
 *********************************************************************/

BallThrow::BallThrow(Person *p, char *pname) : MotorSkill(p, pname)
{
   timeSince = 100.0;
   hspeed = 0.0;
   vspeed = 0.0;
   zspeed = 0.0;
}

static MotorCreature *grabbee = NULL;

int BallThrow::prepareToRun(void *arg)
{
    int okToTurnOn = (timeSince>5.0)?1:0;

    if(okToTurnOn){
	timeSince = 0.0;
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	handPos = argVals->getSbVec3fAt(0);
	hspeed = argVals->getFloatAt(1);
	vspeed = argVals->getFloatAt(2);
	zspeed = argVals->getFloatAt(3);
	GrabberSkill *grab = (GrabberSkill *)parent->getMotorSkillByName("grabskill");
	if(grab->grabbedList.getLength())
	    grabbee = (MotorCreature *)((grabbedConstraint *)grab->grabbedList[0])
		->getGrabbedCreature();
	else {
	    grabbee = NULL;
	    printf("grabbee is NULL \n");
	}
    }
    return okToTurnOn;
}


int BallThrow::modifyRun(void *arg)
{
	return 0;  
}

int BallThrow::prepareToTurnOff(void *arg)
{
    int okToTurnOff = 1;
    return okToTurnOff;
}

void BallThrow::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {
	if(timeSince == 0.0){
	    GrabberSkill *grab = (GrabberSkill *)parent->getMotorSkillByName("grabskill");
	    if(grabbee){
		grabbee->setPosition(handPos);
		grabbee->isGrabbed.setValue(FALSE);
		if(grabbee->isOfType(BehaviorBall::getClassTypeId()))
		    ((BehaviorBall *)grabbee)->hasBeenThrown.setValue(TRUE);
//		printf("%f %f %f \n", hspeed, vspeed, zspeed);
		MotorCommandBlock mcb(GC_MOVE_GEN, hspeed, vspeed, zspeed);
		grabbee->controller->executeMotorCommand(mcb);
	    }
	}
	timeSince += usd.time;
	if(timeSince > 5.0)
	    turnOff();
    }

    MotorSkill::updateStateAt(usd);
    return;
}


/*********************************************************************
 * 
 * PersonUpdateVisionFromMCB skill
 *  arg0: SbVec3f = pos. of left hand
 *  arg1: SbVec3f = pos. of right hand
 *  arg2: SbVec3f = pos. of head
 *  arg3: SbVec3f = pos. of body centroid
 *  arg4: SbVec3f = normal width (x) and height (y) of person (z is ignored)
 *  arg5:     int = normal valid flag (0 if false, 1 if true)
 *  arg6: SbVec3f = translation of left hand local transform
 *  arg7: SbVec3f = translation of right hand local transform
 *  arg8: SbVec3f = translation of head local transform
 *  arg9: SbVec3f = translation of gaze block local transform
 * arg10: SbVec3f = local transform for creature
 * arg11: "char. string" = concatenated SoSFBools from BehaviorPerson.h:
          arg8[0] = person exists
	  arg8[1] = extended left
	  arg8[2] = extended right
	  arg8[3] = pointing left
	  arg8[4] = pointing right
	  arg8[5] = handOverHeadLeft
	  arg8[6] = handOverHeadRight
	  arg8[7] = kick left
	  arg8[8] = kick right
	  arg8[9] = throw left
	  arg8[10] = throw right
	  arg8[11] = standing
	  arg8[12] = sitting
	  arg8[13] = bending over
	  arg8[14] = hands over head
	  arg8[15] = hand low left
	  arg8[16] = hand low right
	  arg8[17] = 0 ; end of char string
 *********************************************************************/

PersonUpdateVisionFromMCB::PersonUpdateVisionFromMCB(BehaviorPerson *p, char *pname) : MotorSkill(p, pname)
{
  isImmediate = TRUE;
}

void
PersonUpdateVisionFromMCB::updateImmediate(void *arg)
{
  BehaviorPerson *bp = (BehaviorPerson *) parent;
  MotorCommandBlock *mcb = (MotorCommandBlock *) arg;
  const char *boolFields = mcb->getCharPtrAt(11);

  // now run through the boolean fields, setting each one
    bp->enableNotify(FALSE);
  if (strlen(boolFields) < 1)
    {
      cerr << "PersonUpdateVisionFromMCB: boolean field length is less than 1!" << endl;
      return;
    }
  
  if (boolFields[0] == 1)  // person does not exist
    {
      bp->personExists.setValue(FALSE);
      bp->setPosition(SbVec3f(30.0, 0, -10));
      bp->enableNotify(TRUE);
      bp->touch();
      return;
    }

  bp->personExists.setValue(TRUE);

  if (strlen(boolFields) != 17)
    {
      cerr << "PersonUpdateVisionFromMCB: boolean field length is other than expected: " << strlen(boolFields) << endl;
      return;
    }
  else
    {
      bp->leftHand = mcb->getSbVec3fAt(0);
      bp->rightHand = mcb->getSbVec3fAt(1);
      bp->head = mcb->getSbVec3fAt(2);
      bp->bodyCentroid = mcb->getSbVec3fAt(3);

      SbVec3f normalSize = mcb->getSbVec3fAt(4);
      bp->normalWidth = normalSize[0];
      bp->normalHeight = normalSize[1];
      bp->lastNormalValid = bp->normalValid;
      bp->normalValid = (mcb->getIntAt(5) ? TRUE : FALSE);

      if (bp->lhxf)
	bp->lhxf->translation.setValue(mcb->getSbVec3fAt(6));
      if (bp->rhxf)
	bp->rhxf->translation.setValue(mcb->getSbVec3fAt(7));
      if (bp->hhxf)
	bp->hhxf->translation.setValue(mcb->getSbVec3fAt(8));
      bp->setPosition(mcb->getSbVec3fAt(10));
      if (bp->gbxf)
	{
	  SbVec3f gbxft = mcb->getSbVec3fAt(9);
	  bp->vecToBody(gbxft);
	  bp->gbxf->translation.setValue(gbxft);
	}

      bp->extendedLeft.setValue((boolFields[1] == 2) ? TRUE : FALSE);
      bp->extendedRight.setValue((boolFields[2] == 2) ? TRUE : FALSE);
      bp->pointingLeft.setValue((boolFields[3] == 2) ? TRUE : FALSE);
      bp->pointingRight.setValue((boolFields[4] == 2) ? TRUE : FALSE);
      bp->handOverHeadLeft.setValue((boolFields[7] == 2) ? TRUE : FALSE);
      bp->handOverHeadRight.setValue((boolFields[8] == 2) ? TRUE : FALSE);
      bp->standing.setValue((boolFields[11] == 2) ? TRUE : FALSE);
      bp->sitting.setValue((boolFields[12] == 2) ? TRUE : FALSE);
      bp->bendingOver.setValue((boolFields[13] == 2) ? TRUE : FALSE);
      bp->handLowLeft.setValue((boolFields[15] == 2) ? TRUE : FALSE);
      bp->handLowRight.setValue((boolFields[16] == 2) ? TRUE : FALSE);
      
      // the following are derived fields
      bp->handsOverHead.setValue(
	   (bp->handOverHeadLeft.getValue() &&
	   bp->handOverHeadRight.getValue())?1:0);
      bp->throwingLeft.setValue(
	   bp->pointingLeft.getValue() && bp->ballInHand.getValue());
      bp->throwingRight.setValue(FALSE);
      
	if( (boolFields[1] == 2) ||
	    (boolFields[2] == 2) ||
	    (boolFields[3] == 2) ||
	    (boolFields[4] == 2) ||
	    (boolFields[7] == 2) ||
	    (boolFields[8] == 2) ||
	    (boolFields[15] == 2) ||
	    (boolFields[16] == 2) )
	    bp->userGesturing.setValue(TRUE);
	else
	    bp->userGesturing.setValue(FALSE);
      
      // for historical reference only
      //bp->throwingLeft.setValue((boolFields[9] == 2) ? TRUE : FALSE);
      //bp->throwingRight.setValue((boolFields[10] == 2) ? TRUE : FALSE);
      //bp->handsOverHead.setValue((boolFields[14] == 2) ? TRUE : FALSE);
      
      bp->enableNotify(TRUE);
      bp->touch();
      
      if(bp->currentRep == RepSwitch::REP_3)
	 bp->updateTextDiag();

   }  
}

/*********************************************************************
 * 
 * PersonUpdateAudioFromMCB skill
 * arg0: int = which boolean field (0...10) to update
         (value of -1 indicates no utterance received)
 *********************************************************************/

PersonUpdateAudioFromMCB::PersonUpdateAudioFromMCB(BehaviorPerson *p, char *pname) : MotorSkill(p, pname)
{
  isImmediate = TRUE;
}

void
PersonUpdateAudioFromMCB::updateImmediate(void *arg)
{
  BehaviorPerson *bp = (BehaviorPerson *) parent;
  MotorCommandBlock *mcb = (MotorCommandBlock *) arg;
  int whichField = mcb->getIntAt(0);

  // all of these fields ordinarily get set to 0

  bp->isUtteringSit.setValue(FALSE);
  bp->isUtteringShake.setValue(FALSE);
  bp->isUtteringUp.setValue(FALSE);
  bp->isUtteringDown.setValue(FALSE);
  bp->isUtteringGoAway.setValue(FALSE);
  bp->isUtteringFetch.setValue(FALSE);
  bp->isUtteringBruce.setValue(FALSE);
  bp->isUtteringNewt.setValue(FALSE);
  bp->isUtteringGoodbye.setValue(FALSE);
  bp->isUtteringGoodDog.setValue(FALSE);
  bp->isUtteringBadDog.setValue(FALSE);

  switch (whichField)
    {
    case -1:
      break;
      
    case 0:
      bp->isUtteringSit.setValue(TRUE);
      break;

    case 1:
      bp->isUtteringShake.setValue(TRUE);
      break;

    case 2:
      bp->isUtteringUp.setValue(TRUE);
      break;

    case 3:
      bp->isUtteringDown.setValue(TRUE);
      break;

    case 4:
      bp->isUtteringGoAway.setValue(TRUE);
      break;

    case 5:
      bp->isUtteringFetch.setValue(TRUE);
      break;

    case 6:
      bp->isUtteringBruce.setValue(TRUE);
      break;

    case 7:
      bp->isUtteringNewt.setValue(TRUE);
      break;

    case 8:
      bp->isUtteringGoodbye.setValue(TRUE);
      break;

    case 9:
      bp->isUtteringGoodDog.setValue(TRUE);
      break;

    case 10:
      bp->isUtteringBadDog.setValue(TRUE);
      break;

    default:
      cerr << "PersonUpdateAudioFromMCB: unexpected field number: " << whichField << endl;
      break;
    }
}
