#ifndef PERSONSKILLS
#define PERSONSKILLS

#include "MotorSkill.h"
#include "Person.h"
#include "BehaviorPerson.h"

/**********************************************************************
 *
 * BallThrow
 *
 * This MotorSkill implements moving in what ever direction the wheels are
 * pointing.
 *
 *  args (optional):
 *    0 - hand position (SbVec3f) 
 *    1 - horizontal speed
 *    2 - vertical speed
 *    3 - restitution (0 to 1)
 **********************************************************************/

class BallThrow: public MotorSkill {
  public:
    BallThrow(Person *p, char *pname);
    float vspeed;
    float hspeed;
    float zspeed;
    float timeSince;
    SbVec3f handPos;
    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg);
    void updateStateAt(UpdateStateData &usd);
    int modifyRun(void *args);
};

/*********************************************************************
 *
 * PersonUpdateVisionFromMCB
 *
 * Implements a person's ability to update its vision state from an MCB.
 * Will hopefully be used in Distributed ALIVE so avatars can
 * get the same information as the person running locally from
 * a pfinder.
 *
 ********************************************************************/

class PersonUpdateVisionFromMCB: public MotorSkill {
 public:
  PersonUpdateVisionFromMCB(BehaviorPerson *p, char *pname);
  void updateImmediate(void *arg);
};

/*********************************************************************
 *
 * PersonUpdateAudioFromMCB
 *
 * Implements a person's ability to update its audio state from an MCB.
 * Will hopefully be used in Distributed ALIVE so avatars can
 * get the same information as the person running locally from
 * a pfinder.
 *
 ********************************************************************/

class PersonUpdateAudioFromMCB: public MotorSkill {
 public:
  PersonUpdateAudioFromMCB(BehaviorPerson *p, char *pname);
  void updateImmediate(void *arg);
};

#endif











