//===========================================================================
// TextAnnonation.c++
//===========================================================================

#include "TextAnnotation.h"

#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoFont.h>

#define TEXT_SCALE 0.85
#define TEXT_SPACING -5*TEXT_SCALE
#define TEXT_SPACING_BASE 40
#define TEXT_POS_LEFT -40*TEXT_SCALE
#define TEXT_POS_CENTER 0*TEXT_SCALE
#define TEXT_POS_RIGHT 40*TEXT_SCALE
#define TEXT_Z_POS 0

//----------------------------------------------------------------------
// Left			General			Right
// ----------		----------		----------
// Pointing		Standing		Pointing
// Kicking		Sitting			Kicking
// Waving		Hand Over Head		Waving
// Throwing		User Exists		Throwing
// Extending					Extending
// Hand Moving					Hand Moving
//----------------------------------------------------------------------

static float greenish[] = {0.2, 1.0, 0.2};  // Color when selected
static float blueish[] = {0.2, 0.2, 1.0};  // Color when selected
static float white[] = {0.2, 0.2, 0.2};    // Color when not selected


//======================================================================
// TextAnnotation constructor
//======================================================================
TextAnnotation::TextAnnotation()
{
  material_center = new SoMaterial;  
  material_standing = new SoMaterial;
  material_sitting = new SoMaterial;
  material_hand_over_head = new SoMaterial;
  material_bending = new SoMaterial;
  material_user_exists = new SoMaterial;
  material_left = new SoMaterial;
  material_left_pointing = new SoMaterial;
  material_left_kicking = new SoMaterial;
  material_left_waving = new SoMaterial;
  material_left_throwing = new SoMaterial;
  material_left_extending = new SoMaterial;
  material_left_hand_moving = new SoMaterial;
  material_right = new SoMaterial;
  material_right_pointing = new SoMaterial;
  material_right_kicking = new SoMaterial;
  material_right_waving = new SoMaterial;
  material_right_throwing = new SoMaterial;
  material_right_extending = new SoMaterial;
  material_right_hand_moving = new SoMaterial;
}

//======================================================================
// TextAnnotation::makeText
//======================================================================
SoSeparator *TextAnnotation::makeText
(const SbVec3f &textPosition, 
 SoMaterial *materialPointer, 
 const SbColor &textColor,
 const SoText2::Justification &textJustification,
 char *actualText)
{
  //--------------------------------------------------
  // Create text piece #1
  //--------------------------------------------------
  SoSeparator *text_piece = new SoSeparator;

  //--------------------------------------------------
  // Components of text piece #1
  //--------------------------------------------------
  SoTransform *textTransform = new SoTransform;
  textTransform->translation.setValue(textPosition);
  text_piece->addChild(textTransform);
  //--------------------------------------------------
  // materialPointer = new SoMaterial;
  materialPointer->diffuseColor.setValue(textColor);
  text_piece->addChild(materialPointer);
  //--------------------------------------------------
  SoText2 *myText = new SoText2;
  myText->string = actualText;
  myText->justification = textJustification;
  // myText->parts = SoText2::FRONT | SoText2::BACK;
  text_piece->addChild(myText);
  //--------------------------------------------------
  return(text_piece);
}


//======================================================================
// TextAnnotation::createTextSetup
//======================================================================
SoSeparator *TextAnnotation::createTextSetup()
{
  //--------------------------------------------------
  // Create Root Text
  //--------------------------------------------------
  SoSeparator *rootText = new SoSeparator;


#if 0
  //--------------------------------------------------
  // Draw Style
  //--------------------------------------------------
  SoDrawStyle *drawStyle = new SoDrawStyle;
  drawStyle->style = SoDrawStyle::LINES;
  rootText->addChild(drawStyle);
#endif

  //--------------------------------------------------
  // Font Style
  //--------------------------------------------------
  SoFont *font = new SoFont;
  font->name.setValue("Times-Roman");
  font->size.setValue(12);
  rootText->addChild(font);

  //--------------------------------------------------
  // LEFT
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      0*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left,
	      SbColor(blueish),
	      SoText2::LEFT,
	      "LEFT"));
  //--------------------------------------------------
  // Pointing
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      1*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left_pointing,
	      SbColor(white),
	      SoText2::LEFT,
	      "Pointing"));
  //--------------------------------------------------
  // Kicking
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      2*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left_kicking,
	      SbColor(white),
	      SoText2::LEFT,
	      "Kicking"));
  //--------------------------------------------------
  // Waving
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      3*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left_waving,
	      SbColor(white),
	      SoText2::LEFT,
	      "Waving"));
  //--------------------------------------------------
  // Throwing
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      4*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left_throwing,
	      SbColor(white),
	      SoText2::LEFT,
	      "Throwing"));
  //--------------------------------------------------
  // Extending
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      5*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left_extending,
	      SbColor(white),
	      SoText2::LEFT,
	      "Extending"));
  //--------------------------------------------------
  // Hand Moving
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_LEFT, 
		      6*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_left_hand_moving,
	      SbColor(white),
	      SoText2::LEFT,
	      "Hand Moving"));

  //--------------------------------------------------
  // RIGHT
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      0*TEXT_SPACING + TEXT_SPACING_BASE,
		      TEXT_Z_POS), 
	      material_right,
	      SbColor(blueish),
	      SoText2::RIGHT,
	      "RIGHT"));
  //--------------------------------------------------
  // Pointing
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      1*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_right_pointing,
	      SbColor(white),
	      SoText2::RIGHT,
	      "Pointing"));
  //--------------------------------------------------
  // Kicking
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      2*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_right_kicking,
	      SbColor(white),
	      SoText2::RIGHT,
	      "Kicking"));
  //--------------------------------------------------
  // Waving
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      3*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_right_waving,
	      SbColor(white),
	      SoText2::RIGHT,
	      "Waving"));
  //--------------------------------------------------
  // Throwing
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      4*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_right_throwing,
	      SbColor(white),
	      SoText2::RIGHT,
	      "Throwing"));
  //--------------------------------------------------
  // Extending
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      5*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_right_extending,
	      SbColor(white),
	      SoText2::RIGHT,
	      "Extending"));
  //--------------------------------------------------
  // Hand Moving
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_RIGHT, 
		      6*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_right_hand_moving,
	      SbColor(white),
	      SoText2::RIGHT,
	      "Hand Moving"));

  //--------------------------------------------------
  // GENERAL
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_CENTER, 
		      0*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_center,
	      SbColor(blueish),
	      SoText2::CENTER,
	      "GENERAL"));
  //--------------------------------------------------
  // Standing
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_CENTER, 
		      1*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_standing,
	      SbColor(white),
	      SoText2::CENTER,
	      "Standing"));
  //--------------------------------------------------
  // Sitting
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_CENTER, 
		      2*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_sitting,
	      SbColor(white),
	      SoText2::CENTER,
	      "Sitting"));
  //--------------------------------------------------
  // Hand Over Head
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_CENTER, 
		      3*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_hand_over_head,
	      SbColor(white),
	      SoText2::CENTER,
	      "Hand Over Head"));
  //--------------------------------------------------
  // Bending
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_CENTER, 
		      4*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_bending,
	      SbColor(white),
	      SoText2::CENTER,
	      "Bending"));

  //--------------------------------------------------
  // User Exists
  //--------------------------------------------------
  rootText->addChild
    (makeText(SbVec3f(TEXT_POS_CENTER, 
		      5*TEXT_SPACING + TEXT_SPACING_BASE, 
		      TEXT_Z_POS), 
	      material_user_exists,
	      SbColor(white),
	      SoText2::CENTER,
	      "User Exists"));

  //--------------------------------------------------
  // return rootText
  //--------------------------------------------------
  return(rootText);
}


//======================================================================
// TextAnnotation::updateText()
//----------------------------------------------------------------------
// typedef struct
// {
//   unsigned char standing;
//   unsigned char sitting;
//   unsigned char hand_over_head;
//
//   unsigned char point_left;
//   unsigned char point_right;
//
//   unsigned char kick_left;
//   unsigned char kick_right;
//
//   unsigned char wave_left;
//   unsigned char wave_right;
//
//   unsigned char throw_left;
//   unsigned char throw_right;
//
//   unsigned char extend_left;
//   unsigned char extend_right;
//  
//   unsigned char hand_moving_left;
//   unsigned char hand_moving_right;
//  
//   unsigned char user_exists;
// } GestureFlags;
//======================================================================
void TextAnnotation::updateText(GestureFlags *gesture)
{
  //============================================================
  // General Annotation
  //============================================================
  //--------------------------------------------------
  // sitting
  //--------------------------------------------------
  if (gesture->sitting) {
    material_sitting->diffuseColor.setValue(greenish);
  } else {
    material_sitting->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // standing
  //--------------------------------------------------
  if (gesture->standing) {
    material_standing->diffuseColor.setValue(greenish);
  } else {
    material_standing->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // head_over_head
  //--------------------------------------------------
  if (gesture->hand_over_head) {
    material_hand_over_head->diffuseColor.setValue(greenish);
  } else {
    material_hand_over_head->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // bending_over
  //--------------------------------------------------
  if (gesture->bending_over) {
    material_bending->diffuseColor.setValue(greenish);
  } else {
    material_bending->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // user_exists
  //--------------------------------------------------
  if (gesture->user_exists) {
    material_user_exists->diffuseColor.setValue(greenish);
  } else {
    material_user_exists->diffuseColor.setValue(white);
  }
  
  //============================================================
  // Left
  //============================================================
  //--------------------------------------------------
  // pointing
  //--------------------------------------------------
  if (gesture->point_left) {
    material_left_pointing->diffuseColor.setValue(greenish);
  } else {
    material_left_pointing->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // kicking
  //--------------------------------------------------
  if (gesture->kick_left) {
    material_left_kicking->diffuseColor.setValue(greenish);
  } else {
    material_left_kicking->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // waving
  //--------------------------------------------------
  if (gesture->wave_left) {
    material_left_waving->diffuseColor.setValue(greenish);
  } else {
    material_left_waving->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // throwing
  //--------------------------------------------------
  if (gesture->throw_left) {
    material_left_throwing->diffuseColor.setValue(greenish);
  } else {
    material_left_throwing->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // extending
  //--------------------------------------------------
  if (gesture->extend_left) {
    material_left_extending->diffuseColor.setValue(greenish);
  } else {
    material_left_extending->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // hand_moving
  //--------------------------------------------------
  if (gesture->hand_moving_left) {
    material_left_hand_moving->diffuseColor.setValue(greenish);
  } else {
    material_left_hand_moving->diffuseColor.setValue(white);
  }

  //============================================================
  // Right
  //============================================================
  //--------------------------------------------------
  // pointing
  //--------------------------------------------------
  if (gesture->point_right) {
    material_right_pointing->diffuseColor.setValue(greenish);
  } else {
    material_right_pointing->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // kicking
  //--------------------------------------------------
  if (gesture->kick_right) {
    material_right_kicking->diffuseColor.setValue(greenish);
  } else {
    material_right_kicking->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // waving
  //--------------------------------------------------
  if (gesture->wave_right) {
    material_right_waving->diffuseColor.setValue(greenish);
  } else {
    material_right_waving->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // throwing
  //--------------------------------------------------
  if (gesture->throw_right) {
    material_right_throwing->diffuseColor.setValue(greenish);
  } else {
    material_right_throwing->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // extending
  //--------------------------------------------------
  if (gesture->extend_right) {
    material_right_extending->diffuseColor.setValue(greenish);
  } else {
    material_right_extending->diffuseColor.setValue(white);
  }
  //--------------------------------------------------
  // hand_moving
  //--------------------------------------------------
  if (gesture->hand_moving_right) {
    material_right_hand_moving->diffuseColor.setValue(greenish);
  } else {
    material_right_hand_moving->diffuseColor.setValue(white);
  }
}
