#include "VisionServices.h"
#include "AliveCamera.h"
#include "DurationTimer.h"
#include "Alive.h"
#include <bstring.h>
#include <iostream.h>

#include <Inventor/nodes/SoSeparator.h>

int gestureText = 1;

ContourServer *VisionServices::contourServer=NULL;
Gesture *VisionServices::gesture=NULL;
float worldExpireTime = 5.;
float timePersonToBeInFrame = 5.;

 /* this should be called exactly once */

VisionServices::VisionServices(char *machine, char *paramFile, SoSeparator *sep,AliveCamera *aMapping, 
int layers)
{
    if(VisionServices::contourServer != NULL)
	printf("contour server already running \n");
    else {
      if (gestureText) {
	//------------------------------------------------------------
	// attach my Text stuff here -- johnny
	// cout << "*** before texta = new TextAnnotation;" << endl;
	texta = new TextAnnotation;
	// cout << "*** after texta = new TextAnnotation;" << endl;
	
      }
	VisionServices::contourServer = new ContourServer(machine, NULL, aMapping,layers);
	VisionServices::gesture = new Gesture(paramFile);
	flagAllGestures(&desired);
	//desired.kick_left = FALSE;
   	//desired.kick_right = FALSE;
	setCurrentMapping(aMapping);
	if(sep)
	    attachContourToSceneGraph(sep);
    }
}

void VisionServices::setCurrentMapping(AliveCamera *aMapping)
{
    currentMapping = aMapping;
    VisionServices::contourServer->mapping =  currentMapping;
    VisionServices::contourServer->setScale(
	1.0/VisionServices::contourServer->mapping->vXwScale);
}

void VisionServices::attachContourToSceneGraph(SoSeparator *sep)
{
    sep->addChild(VisionServices::contourServer->sep);
    if (gestureText) {
      //------------------------------------------------------------
      // attach my Text stuff here -- johnny
      // cout << "*** before createTextSetup" << endl;
      sep->addChild(texta->createTextSetup());
      // cout << "*** after createTextSetup" << endl;
    }
}

void VisionServices::updateStateAt()
{
  static unsigned char personWasInFrame = 0;
  static unsigned char personHereLastFrame = 0;
  static DurationTimer *timePersonInFrame = new DurationTimer;
  static DurationTimer *timePersonOutOfFrame = new DurationTimer;

  if(VisionServices::contourServer){
    bzero((unsigned char *) &found, sizeof(GestureFlags)); /* This should be a function!  --- Foner 14 Jul 94 */
    if ( VisionServices::contourServer->grabAndGetContour() ) {
      //printf("Got contour\n");
#ifndef PFC
      VisionServices::gesture->updateModel(&(VisionServices::contourServer->features), 
					   VisionServices::contourServer->scale);
      VisionServices::gesture->findGestures(&desired, &found);

#endif
  // stuff to reset the world if the person has at any point been in the frame
  // more than 5 seconds and has also left the frame for more than 5 seconds

      if (!personHereLastFrame)  // person just walked in
        timePersonInFrame->setTime();
      else
	{
	  if (timePersonInFrame->duration() > timePersonToBeInFrame)
	    personWasInFrame = 1;
	}
      personHereLastFrame = 1;

      //	printf("Found gestures\n");
      //printGestureFlags(stdout, &found);
    } else {
/* Following line removed at Mike's suggestion */
//      VisionServices::gesture->userGone();

      if (personWasInFrame)
        {
          if (personHereLastFrame)   // person just stepped out
            {
              timePersonOutOfFrame->setTime();
            }
          else
            {
              if (timePersonOutOfFrame->duration() > worldExpireTime)
                {
                  Alive::shouldRestart = 1;
		  personWasInFrame = 0;
                }
            }
        }
      personHereLastFrame = 0;

    }
  }

  if (gestureText) {
    //------------------------------------------------------------
    // update my Text stuff here -- johnny
    // cout << "*** before updateText" << endl;
    texta->updateText(&found);
    // cout << "*** after updateText" << endl;
  }
}

void VisionServices::openConnection()
{
    contourServer->openServer();
}

void VisionServices::closeConnection()
{
    contourServer->closeServer();
}

void VisionServices::resetConnection()
{
    contourServer->resetServer();
    contourServer->closeServer();
}

void VisionServices::abortConnection()
{
    contourServer->abortServer();
    contourServer->closeServer();
}
