#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/SoPath.h>

#include "MotorLib.h"
#include "BallMotorSkills.h"
#include "CreatureLib.h"

/*********************************************************************
 * 
 * BallMOVE Skill
 * 
 *********************************************************************/

BallMove::BallMove(Ball *p, char *pname) : MotorSkill(p, pname)
{
   restitution = 1;
}

int BallMove::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	modifyRun(arg);
    }
    return okToTurnOn;
}


int BallMove::modifyRun(void *arg)
{
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	parent->vel.v0[0] = argVals->getFloatAt(0);
	parent->vel.v0[1] = argVals->getFloatAt(1);
	restitution = ((Ball *)parent)->restitution.getValue();
        return 1;  
}



int BallMove::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;

    if (okToTurnOff) {
	parent->vel.v0.setValue(0,0, 0);
    }
	
    return okToTurnOff;
}

void BallMove::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {
	parent->vel.v0[1] -= .5*(32.0*12.0)*usd.time;
	SbVec3f pos = parent->position();
	if(pos[1]<2.0){
	    parent->vel.v0[1] *= -restitution;
	    parent->vel.v0[0] *= restitution;
	    parent->setPosition(SbVec3f(pos[0], 0.0, pos[2]));
	}
    }

    MotorSkill::updateStateAt(usd);
    return;
}

BallChangeColor::BallChangeColor(Ball *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
    dofs[numDofs++] = ((Dof *)parent->dofs[0]);
}

void BallChangeColor::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    float alpha  = argVals->getFloatAt(0);
    if(dofs[0]->interpolator){
	dofs[0]->setAlpha(alpha);
    }
    else printf("no color dof \n");
}

BallChangeAlpha::BallChangeAlpha(Ball *p, char *pname):
MotorSkill((MotorCreature *)p, pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[1]);
    speed = .1;
}

int BallChangeAlpha::prepareToRun(void *arg)
{
    counter = 0;
    target = 1;
    speed = .05;
    return 1;
}

int BallChangeAlpha::prepareToTurnOff(void *arg)
{
    target = 0;
    speed = -.05;
    return (counter == target)? 1:0;
}

void BallChangeAlpha::updateStateAt(UpdateStateData &usd)
{
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	counter += speed;
	if(dofs[0]->interpolator){
	    dofs[0]->setAlpha(counter);
	}
	if((!requestedThisTick) && (parent->runBehavior.getValue()))
	    turnOff();	
    }
    MotorSkill::updateStateAt(usd);
}

