#include <math.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/nodes/SoMaterial.h>
#include "BehaviorBall.h"
#include "BallMotorSkills.h"
#include "CreatureLib.h"
#include "MotorLib.h"
#include "AliveDogLib_2.h"
#include "BehaviorCreatureLib.h"
#include "../BehaviorCreatureLib/util.h"
#include "../MotorLib/BasicDefines.h"
#define BALL_CHANGECOLOR 40000
#define BALL_STOPMOVE 40001
#define BALL_CHANGEALPHA 40002
SO_NODE_SOURCE(BehaviorBall);

void ungrabbedCallback(BaseCreature *bc)
{
    BehaviorBall *bb = (BehaviorBall *)bc;
    // On a grab
    if(bb->isGrabbed.getValue() == TRUE){
	if(bb->grabbedBy->isOfType( AliveDog::getClassTypeId())){
	    printf("I WAS GRABBED BY DOG\n");
	    bb->pickedUpByDog = 1;
	    bb->shouldReturnToHomePosition.setValue(TRUE);
	    MotorCommandBlock mcb(GC_MOVE_GEN);
	    bb->controller->executeMotorCommand(mcb);
	}
	else bb->pickedUpByDog = 0;
    }
    // On an ungrab
    else if(bb->pickedUpByDog){
	bb->shouldReturnToHomePosition.setValue(TRUE);
	bb->pickedUpByDog = 0;
	printf("I WAS UN_GRABBED BY DOG\n");
    } 
}

void returnToHome(BaseActivity *act, BehaviorUpdateAction *action)
{
    static int ct;
    CBUA(dba, action);
    LeafActivity *la = (LeafActivity *)act;
    
    // let it drop to near 0
    SbVec3f pp = dba->creature->position();
    if(pp[1]<2){
	ct--;
	if(ct>0){
	    la->mcb.cmd = BALL_CHANGEALPHA;
	}
	else {
	    // first tell the sucker to stop moving
	    //la->mcb.cmd = BALL_STOPMOVE;
	    ((BehaviorBall *)dba->creature)->shouldReturnToHomePosition.setValue(FALSE);
	    la->mcb.cmd = GC_POSITION;
	    // now have it move to home position
	    SbVec3f hp = ((BehaviorBall *)dba->creature)->homePosition.getValue();
	    la->mcb.setNextArgGeneric((float)hp[0]);
	    la->mcb.setNextArgGeneric((float)hp[1]);
	    la->mcb.setNextArgGeneric((float)hp[2]);
	}
    }
    else ct = 20;
}

void changeColor(BaseActivity *act, BehaviorUpdateAction *action)
{
    LeafActivity *la = (LeafActivity *)act;

    la->mcb.cmd = BALL_CHANGECOLOR;
    float tmp = act->value.getValue()/10.0;
    la->mcb.setNextArgGeneric(tmp);
}

void
BehaviorBall::initClass()
{
    SO_NODE_INIT_CLASS(BehaviorBall, BehaviorCreature, "BehaviorCreature");
    BaseBehaviorNode::addCallback("changeColor", changeColor);
    BaseBehaviorNode::addCallback("returnHome", returnToHome);
    BaseCreature::addCallback("ballUngrabbed", ungrabbedCallback);
}

BehaviorBall::BehaviorBall() : BehaviorCreature()
{
    SO_NODE_CONSTRUCTOR(BehaviorBall);
    SO_NODE_ADD_FIELD(restitution, (1.0));
    SO_NODE_ADD_FIELD(hasBeenThrown, (FALSE));
    SO_NODE_ADD_FIELD(homePosition, (0,  0,  0));
    SO_NODE_ADD_FIELD(shouldReturnToHomePosition, (FALSE));
    pickedUpByDog = 0;
}

SbBool	
#ifdef SO_VERSION_REVISION
BehaviorBall::readInstance(SoInput *in, unsigned short flags) 
#else
BehaviorBall::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
   if (BehaviorCreature::readInstance(in, flags)) {
#else
   if (BehaviorCreature::readInstance(in)) {
#endif

	return(1);
   } else {
	return(0);
   }
}


BehaviorBall::~BehaviorBall()
{
}

void BehaviorBall::addMotorSkills()
{
    BehaviorCreature::addMotorSkills();

    BallMove *move = new BallMove((Ball *)this, "move");
    getController()->addCommand( GC_MOVE_GEN, "move", move, 1, ArgType((float)0), ArgType((float)0), ArgType((float)1));
    getController()->addCommand( BALL_STOPMOVE, "stop", move, 0);
    
    BallChangeColor *bcc = new BallChangeColor((Ball *)this, "changeColor");
    getController()->addCommand( BALL_CHANGECOLOR, "color", bcc, 1, ArgType((float)0));

    BallChangeAlpha *bca = new BallChangeAlpha((Ball *)this, "changeAlpha");
    getController()->addCommand( BALL_CHANGEALPHA, "alpha", bca, 1, ArgType((float)0));
}

void BehaviorBall::getDOFInterpolators()
{
    SoInterpolateVec3f *svf=NULL;
    SoMaterial *bc =  (SoMaterial *)getByName("COLOR");
    if(bc){
	svf = (SoInterpolateVec3f *)getConnectedEngine(&bc->diffuseColor);
	dofs.append(new Dof(svf, bc));
	SoInterpolateFloat *sf=NULL;
	sf = (SoInterpolateFloat *)getConnectedEngine(&bc->transparency);
	dofs.append(new Dof(sf, bc));
    }
}
    
int BehaviorBall::stepTo(UpdateStateData &usd)
{
    if(isGrabbed.getValue() == TRUE){
	vel.v0 = SbVec3f(0, 0, 0);
	return 1;
    }
    else{
	SbVec3f p;
	getPosition(p);
	if((p - homePosition.getValue()).length() < .2)
	    vel.v0[1] = 0;
	else {
	    p += vel.v0*usd.time;
	    setPosition(p);
	}
	return 1;
    }
}











