#include <math.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/nodes/SoMaterial.h>

#include "Bird.h"
#include "BallMotorSkills.h"
#include "CreatureLib.h"
#include "MotorLib.h"
#include "AliveDogLib_2.h"
#include "../AliveLib/BehaviorPerson.h"
#include "BehaviorCreatureLib.h"
#include "BaseBehaviorNode.h"
#include "../BehaviorCreatureLib/util.h"
#include "../MotorLib/BasicDefines.h"

SO_NODE_SOURCE(Bird);

static void forceReleaseCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    CHECKPRONOME(pronome);
    Bird *bird = (Bird *)dba->creature;
    LeafActivity *la = (LeafActivity *)act;
    
    // Set Mad flag
    MotorCommandBlock mcb1(GC_SETBOOLFIELD_GEN);
    mcb1.setNextArgGeneric("isMad");
    mcb1.setNextArgGeneric((int)1);
    bird->getController()->executeMotorCommand(mcb1);
    
    // Set color 
    la->mcb.cmd = 1783;
    la->mcb.setNextArgGeneric((float)1);
    
    // force release
//    if(bird->grabber){
//	MotorCommandBlock mcb(GC_UNGRAB);
//	bird->grabber->getController()->executeMotorCommand(mcb);
//    }
//    else printf("NO GRABBER \n");
}

void ungrabbedBirdCallback(BaseCreature *bc)
{
    Bird *bb = (Bird *)bc;
    
    // On a grab
    if(bb->isGrabbed.getValue() == TRUE){
	if(bb->grabbedBy->isOfType( BehaviorPerson::getClassTypeId())){
	    printf("I WAS GRABBED BY PERSON\n");
	    bb->grabber = (MotorCreature *)bb->grabbedBy;
	    MotorCommandBlock mcb1(GC_SETBOOLFIELD_GEN);
	    mcb1.setNextArgGeneric("somethingInHand");
	    mcb1.setNextArgGeneric(1);
	    bb->grabber->getController()->executeMotorCommand(mcb1);
	}
	else {
	    printf("I WAS GRABBED BY DOG\n");
	    bb->grabber = (MotorCreature *)bb->grabbedBy;
	}
    }
    else {
	// turn off somethingInHand
	MotorCommandBlock mcb1(GC_SETBOOLFIELD_GEN);
	mcb1.setNextArgGeneric("somethingInHand");
	mcb1.setNextArgGeneric(0);
	bb->grabber->getController()->executeMotorCommand(mcb1);
	printf("I WAS UN_GRABBED BY PERSON\n");
	
	// if mad and released when hand is below head
	if(!((BehaviorPerson *)bb->grabber)->handsOverHead.getValue()){
	    if(bb->isMad.getValue()){
		// ok here we will change the reps and the worlds
		bb->nextRep++;
		if(bb->nextRep == 3)
		    bb->nextRep++;
		int numReps = ((BehaviorPerson *)bb->grabber)->numReps.getValue();
		if(numReps)
		    bb->nextRep = (numReps)? bb->nextRep % numReps : 0;
		
		MotorCommandBlock mcb(GC_CHANGEREP, ArgType(((int)(RepSwitch::REP_0)+bb->nextRep)));
		bb->grabber->controller->executeMotorCommand(mcb);
		printf("I'M MAD AT PERSON\n");
	    }
	}
	
	// reduce Mad to 0 and change color 
	MotorCommandBlock mcb2(GC_SETBOOLFIELD_GEN);
	mcb2.setNextArgGeneric("isMad");
	mcb2.setNextArgGeneric(0);
	bb->getController()->executeMotorCommand(mcb2);
	bb->grabber = NULL;
	MotorCommandBlock mcb3(1783, ArgType((float)0.0));
	bb->getController()->executeMotorCommand(mcb3);
	
	MotorCommandBlock mcb22(2000, ArgType("IMCOOL"), ArgType((float)19.0));
	bb->getController()->executeMotorCommand(mcb22);
	
	// warp to new position to avoid an immediate pickup
	SbMatrix sbm;
	SbVec3f pos = bb->position();
	sbm.setTranslate(pos);
	MotorCommandBlock mcb4(GC_SETTRANSFORM, ArgType((SbMatrix)sbm));
	bb->getController()->executeMotorCommand(mcb4);
    } 
}

void
Bird::initClass()
{
    SO_NODE_INIT_CLASS(Bird, BehaviorCreature, "BehaviorCreature");
    BaseCreature::addCallback("birdgrab", ungrabbedBirdCallback);
    BaseBehaviorNode::addCallback("forceRelease", forceReleaseCB);
}

Bird::Bird() : BehaviorCreature()
{
    SO_NODE_CONSTRUCTOR(Bird);
    SO_NODE_ADD_FIELD(target, (0,  0,  0));
    SO_NODE_ADD_FIELD(isMad, (TRUE));
    vspeed = 0.0;
    nextRep = 0;
}

SbBool	
#ifdef SO_VERSION_REVISION
Bird::readInstance(SoInput *in, unsigned short flags) 
#else
Bird::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
   if (BehaviorCreature::readInstance(in, flags)) {
#else
   if (BehaviorCreature::readInstance(in)) {
#endif

	return(1);
   } else {
	return(0);
   }
}


Bird::~Bird()
{
}

void Bird::getDOFInterpolators()
{
    SoInterpolateVec3f *svf=NULL;
    SoMaterial *bc =  (SoMaterial *)getByName("COLOR");
    if(bc){
	svf = (SoInterpolateVec3f *)getConnectedEngine(&bc->diffuseColor);
	dofs.append(new Dof(svf, bc));
    }
}

void Bird::addMotorSkills()
{
    BehaviorCreature::addMotorSkills();
    GenericClimb *climb = new  GenericClimb(this, "climb");
    getController()->addCommand(1782, "climb", climb, 1, ArgType((float)0), ArgType((float)0));
    BirdChangeColor *bcc = new  BirdChangeColor(this, "birdChangeColor");
    getController()->addCommand(1783, "changeColor", bcc, 1, ArgType((float)0));
}

int Bird::stepTo(UpdateStateData &usd)
{
    MotorCreature::stepTo(usd);
    float t = usd.time;
    float vd = (useRealTime.getValue())? vspeed * t :vspeed;
    SbVec3f pos = position();
    pos[1] += vd;
    if(pos[1]<0.0)
	pos[1] = 0.0;
    else if(pos[1]>70)
	pos[1] = 70;
    setPosition(pos);
    return 1;
}    
/*************************************************************
 * 
 * 
 *************************************************************/
GenericClimb::GenericClimb(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    
}

int GenericClimb::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	start_speed = ((Bird *)parent)->vspeed;
	time_since  = 0.0; 

	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_speed = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);
    }

//fprintf(stderr, "Prepare To run : ds %f, secs %f\n", desired_speed, secs_to_reach);    

    return okToTurnOn;
}


int GenericClimb::modifyRun(void *arg)
{
	start_speed = ((Bird *)parent)->vspeed;
	time_since  = 0.0; 


	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_speed = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);

        return 1;  
}



int GenericClimb::prepareToTurnOff(void *)
{
    int okToTurnOff;

    if ((fabs(((Bird *)parent)->vspeed) <= .05) && (desired_speed == 0.0)) 
	okToTurnOff = 1;
    else 
	okToTurnOff = 0;

    
    if (okToTurnOff) {
	((Bird *)parent)->vspeed = 0.0;
        desired_speed = 0.0;
        start_speed = 0.0;
        secs_to_reach = 0.0;
    }
	
    return okToTurnOff;
}

void GenericClimb::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {

    time_since += usd.time;

    if (parent->speed != desired_speed) {
	if (time_since < secs_to_reach) {
	    float tmp = time_since / secs_to_reach;
	    ((Bird *)parent)->vspeed = start_speed + (tmp * (desired_speed - start_speed));
	} else {
	    ((Bird *)parent)->vspeed = desired_speed;
	}
    } else 
	turnOff();

    }

    MotorSkill::updateStateAt(usd);
    return;
}


bool_t GenericClimb::toOrFromStream(XDR *xdrs)
{
  if (!xdr_float(xdrs, &desired_speed))
    return FALSE;
  if (!xdr_float(xdrs, &start_speed))
    return FALSE;
  if (!xdr_float(xdrs, &time_since))
    return FALSE;
  if (!xdr_float(xdrs, &secs_to_reach))
    return FALSE;
  return MotorSkill::toOrFromStream(xdrs);
}

BirdChangeColor::BirdChangeColor(Bird *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
    dofs[numDofs++] = ((Dof *)parent->dofs[0]);
}

void BirdChangeColor::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    float alpha  = argVals->getFloatAt(0);
    if(dofs[0]->interpolator){
	dofs[0]->setAlpha(alpha);
    }
    else printf("no color dof \n");
}
