#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <bstring.h>
#include "BehaviorCreature.h"
#include "CreatureLib.h"
#include "MoreSkills.h"
#include "SensorLib.h"
#include "TextGauge.h"
#include "BehaviorCreature.h"
#include "ActivityLib_1.h"

#define GC_SETBVAR 2000
#define GC_SETBFIELD 2001
#define GC_SETRMTYPETAG 2002
#define GC_GETACTIVEBEV 2003
SO_NODE_SOURCE(BehaviorCreature);

void doGenericResetAction(BaseCreature *who)
{
    MotorCreature *theCreature = (MotorCreature *)who;
    if(theCreature->behaviorRoot != NULL){
	BehaviorResetAction *resetAction = new BehaviorResetAction();
	resetAction->apply(theCreature->behaviorRoot);
	delete resetAction;
    }
}

void
BehaviorCreature::initClass()
{
    SO_NODE_INIT_CLASS(BehaviorCreature, MotorCreature, "MotorCreature");
    BaseCreature::addCallback("genericResetAction",doGenericResetAction);
}

BehaviorCreature::BehaviorCreature() : MotorCreature()
{
    SO_NODE_CONSTRUCTOR(BehaviorCreature);
    SO_NODE_ADD_FIELD(gaugeOrigin, (0, 0, 0));
    SO_NODE_ADD_FIELD(traceLevel, (0));
    behaviorGauge = NULL;
    int i;
    for(i=0;i<20;i++){
	gaugebufs[i] = new char[256];
	bzero(gaugebufs[i], 256);
    }
}

BehaviorCreature::~BehaviorCreature()
{

}

void BehaviorCreature::buildBehaviorSystem()
{
    MotorCreature::buildBehaviorSystem();
    if(behaviorRoot) {
	/* now allocate actions */
	ivAction = new InternalVariableUpdateAction();
	dbAction = new CreatureBehaviorUpdateAction(this);
	
	/* cache pointer to top level activityGroup */
	SoSearchAction sa;
	SoPath *p;
	sa.setType(ActivityGroup::getClassTypeId());
	sa.apply(behaviorRoot);
	p = sa.getPath();
	if(!p){
	    printf("Error no ActivityGroup found \n");
	}
	else
	    topActivityGroup = (ActivityGroup *)p->getTail();
	    
	// Add callbacks
	SetBehaviorVariable *sbv = new SetBehaviorVariable(this,"setmotvar");
	getController()->addCommand( GC_SETBVAR, "setMotivation", sbv, 1,  ArgType("none"), ArgType((float)0.0));
	getController()->addCommand( GC_SETRMTYPETAG, "setObjectOfInterest", sbv, 1, ArgType("none"), ArgType((float)0.0));
	SetBoolField *sbf = new SetBoolField(this,"setboolfield");
	getController()->addCommand( GC_SETBFIELD, "set", sbf, 1);
	
	GetBehaviorStateVector *gbsv = new GetBehaviorStateVector(this,"getBehaviorStateVector");
	getController()->addCommand( GC_SETBFIELD, "getActiveBehaviors", gbsv, 1);
    }
}

void BehaviorCreature::traceBehaviorSystemAt(UpdateStateData &usd)
{
    if(!behaviorGauge){
	behaviorGauge = new TextGauge();
	behaviorGauge->moveTextTo(gaugeOrigin.getValue());
	((SoSeparator *)usd.currSceneGraph)->addChild(behaviorGauge->root);
    }
    int i;
    ActivityGroup *traceGroup = topActivityGroup;
    BaseActivity *act;
    
    for(i=0;i<traceLevel.getValue();i++){
	act = traceGroup->activeActivity;
	if(act->leaf.getValue())
	    break;
	else
	    traceGroup = act->getActivityGroup();
    }
    int numActs = traceGroup->printState(gaugebufs);
    for(i=0;i<numActs;i++){
	behaviorGauge->set1Value(i, gaugebufs[i]);
    }
    for(i=numActs;i<20;i++){
	behaviorGauge->set1Value(i,"");
    }
}

void BehaviorCreature::changeRep(int newRep)
{
    BaseCreature::changeRep(newRep);
    if(	(newRep != RepSwitch::REP_REVERT) &&
	(newRep != RepSwitch::REP_SENSING) &&
	(newRep != RepSwitch::REP_SENSING_OO)){
	currentRep = newRep;
    }
    return;
}
void BehaviorCreature::updateBehaviorSystemAt(UpdateStateData &usd)
{
    ivAction->apply(behaviorRoot);
    dbAction->apply(behaviorRoot);
}

SniffSensor *BehaviorCreature::getSniffSensor()
{
    return (SniffSensor *)getSensorOfType(SniffSensor::getClassTypeId());
}

PotentialFieldSensor *BehaviorCreature::getPFSensor()
{
    return (PotentialFieldSensor *)getSensorOfType(PotentialFieldSensor::getClassTypeId());
}

RayData *BehaviorCreature::getDataForClosestUsingSniff(int whichTag)
{
    SniffSensor *ss = getSniffSensor();
    RayData *rd = NULL;
    if(ss)
	rd = getSniffSensor()->getDataForClosestWithTag(whichTag);
    return rd;
}







