/*
 * Copyright (c) 1991-94 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*--------------------------------------------------------------
 *  This is an example from the Inventor Toolmaker,
 *  chapter 4, example 2.
 *
 *  Source file for "BehaviorAction" action.
 *------------------------------------------------------------*/

#include <Inventor/elements/SoElements.h>
#include <Inventor/elements/SoSwitchElement.h>
#include <Inventor/elements/SoUnitsElement.h>
#include <Inventor/elements/SoViewVolumeElement.h>
#include <Inventor/nodes/SoGroup.h>
#include "CreatureBehavior.h"
#include "MotorLib.h"
#include "InternalVariable.h"
#include "ReleasingMechanism.h"

SO_ACTION_SOURCE(CreatureBehaviorUpdateAction);

//
// Initializes the class. This is a one-time
// thing that is done after database initialization and before
// any instance of this class is constructed.
//

void
CreatureBehaviorUpdateAction::initClass()
{
    // Initialize the run-time type variables
    SO_ACTION_INIT_CLASS(CreatureBehaviorUpdateAction, BehaviorUpdateAction);
}

//
// Constructor
//

CreatureBehaviorUpdateAction::CreatureBehaviorUpdateAction(MotorCreature *theCreature)
{
   SO_ACTION_CONSTRUCTOR(CreatureBehaviorUpdateAction);
   creature = theCreature;
}

//
// Destructor. Does nothing.
//

CreatureBehaviorUpdateAction::~CreatureBehaviorUpdateAction()
{
}

//
// Initiates action on a graph. This is called when the action
// is applied to a node, a path, or a path list. It gives us a
// chance to initialize things before beginning traversal.
//

void
CreatureBehaviorUpdateAction::beginTraversal(SoNode *node)
{
   creature->controller->clearSuggestions();
   Pronome *p = pronomes.currentPronome();
   if(p && (p->type == Pronome::OBJECT) && (p->object != NULL)){
	objectOfInterestTag = ((BaseCreature *)p->object)->tag.getValue();
   }
//   else
//	objectOfInterestTag = -1;
   BehaviorUpdateAction::beginTraversal(node);
   creature->controller->executeSuggestions();
}


//
// This method implements the action for an SoSphere node.
//

void
CreatureBehaviorUpdateAction::callDoAction(SoAction *action, SoNode *node)
{
   node->doAction(action);
}
