#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/SoDB.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFNode.h>
#include "BehaviorCreature.h"
#include "CreatureFieldRM.h"
#include "CreatureBehavior.h"
#include "SensorLib.h"
#include "BehaviorAction.h"
#include "util.h"

#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a
float useFieldAsValue(ReleasingMechanism *rm, BehaviorUpdateAction *act);

SO_NODE_SOURCE(CreatureFieldRM);

CreatureFieldRM::CreatureFieldRM() : CreatureRM()
{
    SO_NODE_CONSTRUCTOR(CreatureFieldRM);
    SO_NODE_ADD_FIELD(findFieldName, ("none"));
    SO_NODE_ADD_FIELD(filterFieldNames, ("none"));
    SO_NODE_ADD_FIELD(weightFieldName, ("none"));
    
    SO_NODE_ADD_FIELD(logicalOp, (CreatureFieldRM::IF_ANY));
    SO_NODE_DEFINE_ENUM_VALUE(LogicalOpType, IF_ANY);
    SO_NODE_DEFINE_ENUM_VALUE(LogicalOpType, IF_ALL);
    SO_NODE_DEFINE_ENUM_VALUE(LogicalOpType, IF_NOT_ANY);
    SO_NODE_SET_SF_ENUM_TYPE(logicalOp, LogicalOpType);
}

void
CreatureFieldRM::initClass()
{
    SO_NODE_INIT_CLASS(CreatureFieldRM, CreatureRM, "CreatureRM");
}


CreatureFieldRM::~CreatureFieldRM()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
CreatureFieldRM::readInstance(SoInput *in, unsigned short flags) 
#else
CreatureFieldRM::readInstance(SoInput *in) 
#endif
{
    SbString ffn;
#ifdef SO_VERSION_REVISION
    if (CreatureRM::readInstance(in, flags)) {
#else
    if (CreatureRM::readInstance(in)) {
#endif
	ffn = filterFieldNames[0];
	if((ffn != SbString("none"))&&
	   (filterCBName.getValue() == SbString("none"))){
		setRMFilterCallback(filterBoolFields);
	} 
	if((findFieldName.getValue() != SbString("none"))&&
	   (findCBName.getValue() == SbString("none"))){
	    setRMFindCallback(findAndUseFieldAsPosition);
	}
	if((weightFieldName.getValue() != SbString("none"))&&
	   (weightCBName.getValue() == SbString("none"))){
	    setRMWeightCallback(useFieldAsValue);
	}
//	type.setValue( ReleasingMechanism::AVGINGRELEASER);
//	period.setValue( 10);
			    
	return (1);
    }
    else {
	return (0);
    }
}

SbBool filterBoolFields(ReleasingMechanism *rm, BehaviorUpdateAction *act)
{
    CBUA(dua, act);
    PRONOME_RM(p, rm);
    CreatureFieldRM *frm = (CreatureFieldRM *)rm;
    int i;
    SbBool logicalValue;
    SbBool flValue;
    
    const SbString *fn = frm->filterFieldNames.getValues(0);
    SoSFBool *boolField = NULL;
    
    int fnl = frm->filterFieldNames.getNum();
    // this will initially 
    logicalValue = (frm->logicalOp.getValue() == CreatureFieldRM::IF_ALL)?TRUE:FALSE;
    
    if((p->type == Pronome::OBJECT) && (p->object != NULL) &&
	(fn[0] != "")){
	// iterate through the field names
	for(i=0;i<fnl;i++){
	    boolField = ((SoSFBool *)p->object->getField(SbName(fn[i])));
	    flValue = (!boolField )? FALSE : boolField->getValue();
	    logicalValue = (frm->logicalOp.getValue() == CreatureFieldRM::IF_ALL)
		    ? (flValue && logicalValue)
		    : (flValue );
	    if(!logicalValue && (frm->logicalOp.getValue() == CreatureFieldRM::IF_ALL))
		return FALSE;
	    else if(flValue && (frm->logicalOp.getValue() == CreatureFieldRM::IF_ANY))
		return TRUE;
	    else if(flValue && (frm->logicalOp.getValue() == CreatureFieldRM::IF_NOT_ANY))
		return FALSE;
	}
	return (frm->logicalOp.getValue() == CreatureFieldRM::IF_ALL ||
	    frm->logicalOp.getValue() == CreatureFieldRM::IF_NOT_ANY )?TRUE:FALSE;
    }
    return FALSE;
}
SbBool useSelfAsTarget(ReleasingMechanism *rm, BehaviorUpdateAction *act)
{
    CBUA(dua, act);
    CreatureFieldRM *frm = (CreatureFieldRM *)rm;
    const SbString fn = frm->findFieldName.getValue();
    PRONOME_RM(p, rm);
    SNIFFSENSOR(ss,frm,dua);
    SoSFVec3f *vecField = NULL;
    
    vecField = ((SoSFVec3f *)dua->creature->getField(SbName(fn)));
    SbVec3f pt = (vecField)?vecField->getValue() : dua->creature->position();
	
    p->object = dua->creature;
    p->location = pt;
    p->status = Pronome::PASSEDFIND;
    p->type = Pronome::OBJECT;
    p->range = 0.0;
    p->bearing = 0.0;
    
    p->vel = dua->creature->vel.v0;  // hardts 2/1/95
    p->validVel = TRUE;
    frm->v_tmp = (frm->maxValue.getValue());
    //frm->value.setValue(frm->maxValue.getValue());
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}

SbBool findAndUseFieldAsPosition(ReleasingMechanism *rm, BehaviorUpdateAction *act)
{
    CBUA(dua, act);
    CreatureFieldRM *frm = (CreatureFieldRM *)rm;
    const SbString fn = frm->findFieldName.getValue();
    PRONOME_RM(p, rm);
    SNIFFSENSOR(ss,frm, dua);
    SoSFVec3f *vecField = NULL;
    
    RayData *rd = NULL;
    if(ss)
	rd = ss->getDataForClosestWithTag((int)frm->typeTag.getValue());
    if(rd && rd->who){
	vecField = ((SoSFVec3f *)rd->who->getField(SbName(fn)));
	rd->pt = (vecField)?vecField->getValue() : rd->pt;
	ss->calcBearingAndRange(rd->pt, rd->bearing, rd->range);
    }
    if(TESTRANGE(rd, frm))
    {
	p->object = rd->who;
	p->location = rd->pt;
	p->status = Pronome::PASSEDFIND;
	p->type = Pronome::OBJECT;
	p->range = rd->range;
	p->bearing = rd->bearing;
	frm->v_tmp = (frm->maxValue.getValue());
	//frm->value.setValue(frm->maxValue.getValue());
    }
    else p->status = Pronome::FAILEDFIND;
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}
float useFieldAsValue(ReleasingMechanism *rm, BehaviorUpdateAction *act)
{
    CBUA(dua, act);
    CreatureFieldRM *frm = (CreatureFieldRM *)rm;
    const SbString fn = frm->weightFieldName.getValue();
    PRONOME_RM(p, rm);
    SNIFFSENSOR(ss,frm, dua);
    SoSFFloat *fField = NULL;
    
    if(p->object && (p->status == Pronome::PASSEDFILTER)){
	fField = ((SoSFFloat *)((SoNode *)p->object)->getField(SbName(fn)));
	return (fField!= NULL)? (float)fField->getValue() : 0.0;
    }
    else return 0.0;
}
/********************************************************************
 * 
 *	    ProprioceptiveFieldRM
 * 
 ********************************************************************/
SO_NODE_SOURCE(ProprioceptiveFieldRM);

ProprioceptiveFieldRM::ProprioceptiveFieldRM() : CreatureFieldRM()
{
    SO_NODE_CONSTRUCTOR(ProprioceptiveFieldRM);
    senseNode = NULL;
}

void
ProprioceptiveFieldRM::initClass()
{
    SO_NODE_INIT_CLASS(ProprioceptiveFieldRM, CreatureFieldRM, "CreatureFieldRM");
}


ProprioceptiveFieldRM::~ProprioceptiveFieldRM()
{
}

void
ProprioceptiveFieldRM::setSenseNode(BaseBehaviorNode *node)
{
    senseNode = node;
    filterFieldNames.set1Value(0, "active");
}

SbBool useSenseNodeAsTarget(ReleasingMechanism *rm, BehaviorUpdateAction *act)
{
    CBUA(dua, act);
    PRONOME_RM(p, rm);
    ProprioceptiveFieldRM *frm = (ProprioceptiveFieldRM *)rm;
    if(frm->senseNode == NULL){
	printf("installing pointer to sense node \n");
	SoNode *sn = ((MotorCreature *)dua->creature)
	    ->behaviorRoot->getByName(frm->pname.getValue());
	if(sn)
	   frm->senseNode = (BaseBehaviorNode *)sn;
	else {
	    printf("sense node named %s\n", frm->pname.getValue().getString());
	    p->object = NULL;
	    p->status = Pronome::FAILEDFIND;
	    return FALSE;
	} 
    }
    SbVec3f pt = dua->creature->position();
    p->object = frm->senseNode;
    p->location = pt;
    p->status = Pronome::PASSEDFIND;
    p->type = Pronome::OBJECT;
    p->range = 0.0;
    p->bearing = 0.0;
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}

SbBool	
#ifdef SO_VERSION_REVISION
ProprioceptiveFieldRM::readInstance(SoInput *in, unsigned short flags) 
#else
ProprioceptiveFieldRM::readInstance(SoInput *in) 
#endif
{
    SbString ffn;
#ifdef SO_VERSION_REVISION
    if (CreatureFieldRM::readInstance(in, flags)) {
#else
    if (CreatureFieldRM::readInstance(in)) {
#endif
	setRMFindCallback(useSenseNodeAsTarget);
	return TRUE;
    }
    else return FALSE;
}



void initCreatureFieldRMCallbackList()
{
    BaseBehaviorNode::addCallback("filterBoolFields", filterBoolFields);
    BaseBehaviorNode::addCallback("findAndUseFieldAsPosition",  findAndUseFieldAsPosition);
    BaseBehaviorNode::addCallback("useSelf",  useSelfAsTarget);
    BaseBehaviorNode::addCallback("useSenseNode",  useSenseNodeAsTarget);
}
    
