#include <assert.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoNode.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoDB.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFNode.h>
#include "CreatureRM.h"
#include "BehaviorAction.h"
#include "creatureRMCallbacks.h"
#include "SensorLib.h"
#include "MotorLib.h"

#define MIN(a,b)((a)<(b))?a:b
#define MAX(a,b)((a)<(b))?b:a

SO_NODE_SOURCE(CreatureRM);

CreatureRM::CreatureRM() : ReleasingMechanism()
{
    SO_NODE_CONSTRUCTOR(CreatureRM);
    SO_NODE_ADD_FIELD(weightByDistance, (TRUE));
    SO_NODE_ADD_FIELD(useSensors, (CreatureRM::USESNIFFONLY));
    SO_NODE_ADD_FIELD(sensorToUse, ("BODYSENSOR"));
    
    SO_NODE_DEFINE_ENUM_VALUE(SensingType, USEVISIBLEONLY);
    SO_NODE_DEFINE_ENUM_VALUE(SensingType, USESNIFFONLY);
    SO_NODE_DEFINE_ENUM_VALUE(SensingType, USEBEST);
    SO_NODE_SET_SF_ENUM_TYPE(useSensors, SensingType);
    sSensor = NULL;
    enableNotify(FALSE);
}

void
CreatureRM::initClass()
{
    SO_NODE_INIT_CLASS(CreatureRM, ReleasingMechanism, "ReleasingMechanism");
}


CreatureRM::~CreatureRM()
{
}

SbBool	
#ifdef SO_VERSION_REVISION
CreatureRM::readInstance(SoInput *in, unsigned short flags) 
#else
CreatureRM::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (ReleasingMechanism::readInstance(in,  flags)) {
#else
    if (ReleasingMechanism::readInstance(in)) {
#endif
	if(findCBName.getValue() == SbString("none"))
	    setRMFindCallback(findObject);
	if(weightCBName.getValue() == SbString("none"))
	    setRMWeightCallback(weightByDist);
	return (1);
    }
    else {
	return (0);
    }
}

void CreatureRM::_updateValue(BehaviorUpdateAction *act)
{
    static char* buff[2] = {"FAILED", "PASSED"};
    SbBool pOrf;
    if(printDebug.getValue() == FALSE){
	if(find(act)&& filter(act))
	    weight(act);
    }
    else {
	pOrf = find(act);
	printf(" %s %s find \n", pname.getValue().getString(), 
	    (pOrf == TRUE)?buff[1]:buff[0]);
	if(pOrf){
	    pOrf = filter(act);
	    printf(" %s %s filter \n", pname.getValue().getString(), 
	    (pOrf == TRUE)?buff[1]:buff[0]);
	    if(pOrf){
		float ww = weight(act);
		printf(" %s weight %f \n", pname.getValue().getString(),ww); 
	    }   
	}
    }
}

BaseSensor *CreatureRM::getSensor(MotorCreature *creature)
{
    if (!sSensor){
	sSensor = (BaseSensor *)creature->getSensorWithName(SbName(sensorToUse.getValue()));
	if(!sSensor)
	    sSensor = (BaseSensor *)creature->getSensorOfType(SniffSensor::getClassTypeId());
    }
    return sSensor;
}

