#include <Inventor/actions/SoSearchAction.h>
#include "MoreSkills.h"
#include "BehaviorCreatureLib.h"
#include "MotorLib.h"

SetBehaviorVariable::SetBehaviorVariable(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetBehaviorVariable::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(((MotorCreature *)parent)->behaviorRoot){
	SbName nodeName(argVals->getCharPtrAt(0));
	float value = argVals->getFloatAt(1);
	SoNode *namedNode = ((MotorCreature *)parent)->behaviorRoot->
	getByName(nodeName);
	if(namedNode){
	    if(namedNode->isOfType(InternalVariable::getClassTypeId()))
		((InternalVariable *)namedNode)->value.setValue(value);
	    else if(namedNode->isOfType(ReleasingMechanism::getClassTypeId()))
		((ReleasingMechanism *)namedNode)->typeTag.setValue((int)value);
	    else if(namedNode->isOfType(BaseBehaviorNode::getClassTypeId()))
		((BaseBehaviorNode *)namedNode)->value.setValue(value);
	}	
    }
    else 
	printf("variable not found \n");
}

SetBoolField::SetBoolField(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetBoolField::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    SbName fieldName(argVals->getCharPtrAt(0));
    if(fieldName == SbName("runBehavior"))
	((MotorCreature *)parent)->runBehavior.setValue(argVals->getIntAt(1));
    else 
	printf("variable not found \n");
}

/*********************************************************************
 * 
 *	This returns the active leaf behavior
 * 
 *********************************************************************/
 
BaseBehaviorNode *getActiveLeafBehavior(SoNode *node)
{
    SoSearchAction sa;
    SoPath *path;
    ActivityGroup *ag = NULL;
    
    sa.setType(ActivityGroup::getClassTypeId());
    sa.apply(node);
    path = sa.getPath();
    if(!path){
	printf("Error no ActivityGroup found \n");
	ag = NULL;
	return NULL;
    }
    else {
	ag = (ActivityGroup *)path->getTail();
	BaseActivity *act = NULL;
	while((ag != NULL)&&(ag->activeActivity != NULL)){
	    act = ag->activeActivity;
	    ag = (act->leaf.getValue())?NULL : act->getActivityGroup();
	}
	return act;
    }
}

/*********************************************************************
 * 
 *	This returns a list of all active behaviors 
 * 
 *********************************************************************/
GetBehaviorStateVector::GetBehaviorStateVector(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    int i;
    int len = ((SoGroup *)parent->behaviorRoot)->getNumChildren();
    SoNode *aChild;
    topActivityGroup = NULL;
    for(i=0;i<len;i++){
	aChild = ((SoGroup *)parent->behaviorRoot)->getChild(i);
	if(aChild->isOfType(ActivityGroup::getClassTypeId())){
	    topActivityGroup = (ActivityGroup *)aChild;
	}
    }
    isImmediate = TRUE;
}

void GetBehaviorStateVector::updateImmediate(void *arg)
{
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;

    MotorSkill *ms;
    int i;
    ActivityGroup *traceGroup = topActivityGroup;
    BaseActivity *act;
    while((traceGroup != NULL)&&(traceGroup->activeActivity != NULL)){
	act = traceGroup->activeActivity;
	retVals->setNextReturnArg((char *)act->pname.getValue().getString());
	retVals->setNextReturnArg((void *)act);
	traceGroup = (act->leaf.getValue())?NULL : act->getActivityGroup();
    }
    return;
}

/***************************************************************************
 *
 *  Purpose of this motor skill is to return print name of all active (non-zero) 
 *  Releasing Mechanisms
 *
 ***************************************************************************/
 
GetRMStateVector::GetRMStateVector(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
   sa = new SoSearchAction;
   isImmediate = TRUE;
}

void GetRMStateVector::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int shouldReaquire = argVals->getIntAt(0);
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;
    int i, j;
    ReleasingMechanism *rm;
    SoNode *root;
    
    root = (argVals->getArgCount() > 1)
	? parent->behaviorRoot->getByName(argVals->getCharPtrAt(1))
	: parent->behaviorRoot;
    
    if((rmlist.getLength() == 0) || (shouldReaquire != 0)){
	SoPathList paths;
	sa->setType(ReleasingMechanism::getClassTypeId());
	sa->setInterest(SoSearchAction::ALL);
	sa->apply(root);
	paths= sa->getPaths();
	
	for(i=0;i<paths.getLength();i++){
	  rm = (ReleasingMechanism *)paths[i]->getTail();
	  if(!rm->isOfType(ReleasingMechanism::getClassTypeId()))
	    printf("trouble in paradise \n");
	  for(j=0;j<rmlist.getLength();j++){
	      if((void *)rm == rmlist[j])
		break;
	  }
	  if(j==rmlist.getLength())
		rmlist.append(rm);
	}
    }
    j = 0;
    for(i=0;i<rmlist.getLength();i++){
	rm = (ReleasingMechanism *)rmlist[i];
	if(fabs(rm->value.getValue() - rm->minValue.getValue())>1.0e-4){
	    j++;
	    retVals->setNextReturnArg((char *)rm->pname.getValue().getString());
	}
    }
    printf(" %d active releasing mechanisms found \n", j);
    return;
}

/***************************************************************************
 *
 *  Purpose of this motor skill is to return print name of all active (non-zero) 
 *  Releasing Mechanisms
 *
 ***************************************************************************/
 
GetMotStateVector::GetMotStateVector(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
   sa = new SoSearchAction;
   isImmediate = TRUE;
}

void GetMotStateVector::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int shouldReaquire = argVals->getIntAt(0);
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;
    int i, j;
    Motivation *mot;
    SoNode *root;
    
    root = parent->behaviorRoot;
    
    if((motlist.getLength() == 0)){
	SoPathList paths;
	sa->setType(Motivation::getClassTypeId());
	sa->setInterest(SoSearchAction::ALL);
	sa->apply(root);
	paths= sa->getPaths();
	
	for(i=0;i<paths.getLength();i++){
	  mot = (Motivation *)paths[i]->getTail();
	  if(!mot->isOfType(Motivation::getClassTypeId()))
	    printf("trouble in paradise \n");
	  for(j=0;j<motlist.getLength();j++){
	      if((void *)mot == motlist[j])
		break;
	  }
	  if(j==motlist.getLength())
		motlist.append(mot);
	}
    }
    j = 0;
    for(i=0;i<motlist.getLength();i++){
	mot = (Motivation *)motlist[i];
	retVals->setNextReturnArg((char *)mot->getName().getString());
	retVals->setNextReturnArg((float)mot->value.getValue());
    }
    return;
}

/*
 *  Purpose of this motor skill is to build a set of releasing mechanisms
 *  for sensing all of the boolean fields of another creature. The releasing
 *  mechanisms are inserted into a named separator.
 */
BuildBooleanRMs::BuildBooleanRMs(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
   isImmediate = TRUE;
}

void BuildBooleanRMs::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    MotorCreature *mc = (MotorCreature *)argVals->getBaseCreaturePtrAt(0);
    
    MotorCommandBlock mcb((int)(parent->controller->getCommandId("getboolfields")));
    mc->controller->executeMotorCommand(mcb);
    
    int i;
    CreatureFieldRM *cfrm;
    DiscoveryGroup *sep = new DiscoveryGroup;
    sep->setName(SbName("DISCOVERYNODE"));
    const char *fname;
    
    for(i=0;i<mcb.getRetArgCount();i++){
	cfrm = (CreatureFieldRM *)BaseBehaviorNode::readFromPrototypeFile(
	    "CreatureFieldRMProto.iv");
	cfrm->typeTag.setValue(mc->tag.getValue());
	fname = (const char *)mcb.getRetCharPtrAt(i);
	cfrm->pname.setValue(fname);
	cfrm->filterFieldNames.set1Value(0, fname);
	sep->addChild(cfrm);
    }
    
    ((SoSeparator *)parent->behaviorRoot)->addChild(sep);
}

/*
 *  Purpose of this motor skill is to build a releasing mechanisms
 *  which will watch a named activity node,  and use the active
 *  field to monitor whether it is active.
 */
BuildWatchActRM::BuildWatchActRM(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
   isImmediate = TRUE;
}

void BuildWatchActRM::updateImmediate(void *arg)
{
    BaseActivity *ba;
    
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    const char *behaviorName = argVals->getCharPtrAt(0);
    if(behaviorName != NULL)
	ba = (BaseActivity *)parent->behaviorRoot->getByName(SbName(behaviorName));
    else{
	ba = (BaseActivity *)getActiveLeafBehavior(parent->behaviorRoot);
	behaviorName = ba->pname.getValue().getString();  
    }
    if(ba){
	SoNode *root;
	if((argVals->getArgCount() > 1) &&
	    (argVals->getCharPtrAt(1) != NULL))
	    root = parent->behaviorRoot->getByName(argVals->getCharPtrAt(1));
	else {
	    DiscoveryGroup *dg = new DiscoveryGroup();
	    ((SoGroup *)parent->behaviorRoot)->addChild(dg);
	    root = dg;
	}
	
	ProprioceptiveFieldRM *cfrm;
	cfrm = (ProprioceptiveFieldRM *)BaseBehaviorNode::readFromPrototypeFile(
		"ProprioceptiveFieldRMProto.iv");
	cfrm->pname.setValue(behaviorName);
	cfrm->senseNode = ba;
	cfrm->filterFieldNames.set1Value(0, "active");
	cfrm->type.setValue(ReleasingMechanism::STDRELEASER);
	((DiscoveryGroup *)root)->addChildToWatch(cfrm);
    }
    else printf("behavior node named %s not found \n", behaviorName);
}    
    
