#include <math.h>
#include <Inventor/SbLinear.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/nodes/SoTransform.h>
#include "BehaviorCreature.h"
#include "BaseBehaviorNode.h"
#include "CreatureRM.h"
#include "CreatureBehavior.h"
#include "LeafActivity.h"
#include "SensorLib.h"
#include "BasicDefines.h"
#include "creatureRMCallbacks.h"
#include "util.h" // macros defined in file live here

/*
 * Function types:
 *	typedef SbBool (*rmFilterCB)(ReleasingMechanism *rm, 
 *	    BehaviorUpdateAction *action);
 *	typedef float (*rmWeightCB)(ReleasingMechanism *rm, 
 *	    BehaviorUpdateAction *action);
 *	typedef void (*activityCB)(BaseActivity *act, 
 *	    BehaviorUpdateAction *action);
 *
 * Use these to add callbacks:
 *	addCallback(const char *name, rmFilterCB func);
 *	
 */

/****************************************************************
 *
 *  Generic Find and Weight callbacks
 * 
 ****************************************************************/
 
/****************************************************************
 *
 *  Generic Find and Weight callbacks
 * 
 ****************************************************************/
SbBool copyPronome(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    rm->objectOfInterest = *action->pronomes.currentPronome();
    return TRUE;
}
SbBool copyPronomeCheckRange(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    rm->objectOfInterest = *action->pronomes.currentPronome();
    if(TESTRANGEV(rm->objectOfInterest.range, rm))
	return TRUE;
    else {
	rm->objectOfInterest.status = Pronome::FAILEDFIND;
	return FALSE;
    }
}

SbBool testAInsideB(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    
    if((p->type == Pronome::OBJECT) &&
	(p->object != NULL)){
	SbBox3f bbA, bbB;
	SbVec2f before, after;
	float dummy;
	// get the two bounding boxes
	dua->creature->calcBoundingBox(bbA);
	((MotorCreature *)p->object)->calcBoundingBox(bbB);
	SbVec3f m1, m2;
#ifdef 0
	m1 = bbA.getMin();
	m2 = bbA.getMax();
	printf(" A: %f %f %f %f %f %f \n", m1[0], m1[1], m1[2], m2[0], m2[1], m2[2]);
	
	m1 = bbB.getMin();
	m2 = bbB.getMax();
	printf(" B: %f %f %f %f %f %f \n", m1[0], m1[1], m1[2], m2[0], m2[1], m2[2]);
#endif	
	// get the size of the B's bounding box
	bbB.getSize(before[0],dummy,before[1] );
	
	// now make B contain A and see if it grew
	bbB.extendBy(bbA);
	bbB.getSize(after[0],dummy,after[1] );
	if((after-before).length()< 5){
//	    printf("by golly miss molly,  i'm inside \n");
	    p->status = Pronome::PASSEDFILTER;
	    return 1;
	}
//	else {
//	   printf("i'm outside %f \n", (after-before).length()); 
//	}
    }
    p->status = Pronome::FAILEDFILTER;
    return 0;
}

SbBool testPtInsideB(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    
    if((p->type == Pronome::OBJECT) &&
	(p->object != NULL)){
	SbBox3f bbA, bbB;
	SbVec2f before, after;
	float dummy;
	// get the two bounding boxes
	((MotorCreature *)p->object)->calcBoundingBox(bbB);
	
	// get the size of the B's bounding box
	bbB.getSize(before[0],dummy,before[1] );
	
	// now make B contain A and see if it grew
	bbB.extendBy(dua->creature->position());
	bbB.getSize(after[0],dummy,after[1] );
	if((after-before).length()< 1){
//	    printf("by golly miss molly,  body inside \n");
	    p->status = Pronome::PASSEDFILTER;
	    return 1;
	}
//	else {
//	   printf("i'm outside %f \n", (after-before).length()); 
//	}
    }
    p->status = Pronome::FAILEDFILTER;
    return 0;
}

SbBool isFacing(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    if(p != NULL){
	//printf("bearing -> %f \n", p->bearing);
	if(fabs(p->bearing)<1.4){
	    p->status = Pronome::PASSEDFILTER;
	    return 1;
	}
    }
    p->status = Pronome::FAILEDFILTER;
    return 0;
}
        
SbBool findObject(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
//    SNIFFSENSOR(ss, dua);
    SniffSensor *ss = (SniffSensor *)creatureRM->getSensor(dua->creature);
    
    RayData *rd = NULL;
    if(ss)
	rd = ss->getDataForClosestWithTag(creatureRM->typeTag.getValue());
    if(TESTRANGE(rd, creatureRM))
    {
	p->object = rd->who;
	p->location = rd->pt;
	p->status = Pronome::PASSEDFIND;
	p->type = Pronome::OBJECT;
	p->range = rd->range;
	p->bearing = rd->bearing;
	
	p->vel = rd->vel;
	p->validVel = rd->validVel;
	creatureRM->v_tmp = (creatureRM->maxValue.getValue());
	//creatureRM->value.setValue(creatureRM->maxValue.getValue());
    }
    else p->status = Pronome::FAILEDFIND;
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}


SbBool findMarker(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    PFSENSOR(pf, dua);
    static SbVec3f mp;
    RayData rd;
    if(pf->markerFnd){
	mp = rd.pt = pf->markerPos;
	pf->calcBearingAndRange(rd.pt, rd.bearing, rd.range);
    }
    if(TESTMARKERRANGE(pf->markerFnd, rd.range, creatureRM))
    {
	p->object = NULL;
	p->location = rd.pt;
	p->status = Pronome::PASSEDFIND;
	p->type = Pronome::LOCATION;
	p->range = rd.range;
	p->bearing = rd.bearing;
	
//	p->vel = 0;
	p->validVel = rd.validVel;
	creatureRM->v_tmp = (creatureRM->maxValue.getValue());
	//creatureRM->value.setValue(creatureRM->maxValue.getValue());
    }
    else {
	p->status = Pronome::FAILEDFIND;
    }
	
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}

SbBool proximityOnly(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    PFSENSOR(s, dua);
    RayData rd;
    if(s->range<1.0e6){
	rd.pt = s->closestPt;
	rd.range = s->range;
	if(TESTRANGE((&rd), creatureRM)){
	    p->object = NULL;
	    p->location = rd.pt;
	    p->status = Pronome::PASSEDFIND;
	    p->type = Pronome::LOCATION;
	    p->range = rd.range;
	    p->bearing = rd.bearing; // not really valid.  hardts.
	    
	    p->vel = rd.vel;
	    p->validVel = rd.validVel;
	    creatureRM->v_tmp = (creatureRM->maxValue.getValue());
	   // creatureRM->value.setValue(creatureRM->maxValue.getValue());
	}
    }
    else p->status = Pronome::FAILEDFIND;
    return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}

float weightByDist(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    
    float w = rm->calcWeightBasedOnRange(p->range);
    return (w * ((creatureRM->maxValue.getValue() - creatureRM->minValue.getValue())))+
	creatureRM->minValue.getValue();
}

float linearWeightByDist(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    
    float w = rm->linearWeightBasedOnRange(p->range);
    return (w * ((creatureRM->maxValue.getValue() - creatureRM->minValue.getValue())))+
	creatureRM->minValue.getValue();
}

float weightOne(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CRM(creatureRM, rm);
    
    float w = 1.0;
    return w*(creatureRM->maxValue.getValue());
}

float inverseWeightByDist(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    
    float w = rm->calcWeightBasedOnRange(p->range);
    w = 1-w;
    return (w * ((creatureRM->maxValue.getValue() - creatureRM->minValue.getValue())))+
	creatureRM->minValue.getValue();
}

/**************************************************************************
 * 
 * 
 *			GENERIC ACTION CALLBACKS
 * 
 **************************************************************************/
 
static void moveStraightCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    static int currCmd = GC_MOVE_GEN;

    CBUA(dba, action);
    PFSENSOR(s, dba);
    s->updateGradient();
    LeafActivity *la = (LeafActivity *)act;
    if(!dba->creature->checkIfReadyToMove(la->mcb))
      return;
    else {
      MotorCommandBlock noTurn(GC_TURN_GEN,ArgType((float)0),
			       ArgType((float).3));  // .3
      la->issueMotorCommand(*dba->creature,noTurn);
      la->mcb.cmd = currCmd;
      la->mcb.setNextArgGeneric((float)dba->creature->
				    defaultFwdSpeed.getValue());
      la->mcb.setNextArgGeneric((float)1.5);  // 3.0
    }  
}



static void turnWithGradient(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(p, dba);
  PFSENSOR(s, dba);
  static float lastBearing = 0.0;
  
  LeafActivity *la = (LeafActivity *)act;
  float bearing = s->pf->weightedBearing;
  float mg = s->pf->maxGrad;
  float sp = .05;
  
  bearing = (bearing>1)?1 : bearing;
  bearing = (bearing< -1)?-1 : bearing;
  
  la->mcb.cmd = GC_TURN_GEN;
  la->mcb.setNextArgGeneric(-bearing);
/*  sp = (fabs(mg)>1.0)?fabs(mg*mg)*.010 :.010;
    sp = (sp>.06)?.06 : sp; */
  sp = .02;
  la->mcb.setNextArgGeneric(sp);
}



static void turnAwayCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PFSENSOR(s, dba);
  LeafActivity *la = (LeafActivity *)act;
  
  if(!dba->creature->checkIfReadyToMove(la->mcb))
    return;
  s->updateGradient();
  turnWithGradient(act, action);
}



static void searchCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    LeafActivity *la = (LeafActivity *)act;
    
    if(!dba->creature->checkIfReadyToMove(la->mcb))
	return;
    
    if(pronome->status == Pronome::PASSEDFILTER){
	// position schema
	SbVec3f lpt = s->getVecInSensorFrame(pronome->location);
	int d = la->schema->depth;
	float schemaDist = (d/2 - s->pf->depth/2);
	if(pronome->range > schemaDist){
	    lpt.normalize();
	    lpt *= schemaDist;
	}
	s->pf->addPotentialFieldAtOffset((*la->schema->pfield), 
	    (int)lpt[0], (int)lpt[2]); 
	
	// update the gradient and follow it
	s->updateGradient();
//	if((la->ignoreSuggestions.getValue()))
//	    la->issueMotorCommand(*dba->creature, MotorCommandBlock(GC_FWD));
//	else
//	    followGaitSuggestion(act, action);   

	la->issueMotorCommand(*dba->creature,
	    MotorCommandBlock(GC_MOVE_GEN,
			      ArgType((float)dba->creature->defaultFwdSpeed.getValue()),
			      ArgType((float)1.0)));
	turnWithGradient(act, action);
    }
}



static void leaveLeftCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    pronome->location[0] -= 60;
    searchCB(act, action);
}



static void leaveRightCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    PFSENSOR(s, dba);
    pronome->location[0] += 60;
    searchCB(act, action);
}
 


static void avoidCreatureCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PFSENSOR(s, dba);
  PRONOME(p, dba);
  LeafActivity *la = (LeafActivity *)act;
  
  printf("avoidCreatureCB called\n");

  if(!dba->creature->checkIfReadyToMove(la->mcb))
    return;

  if (p->status == Pronome::PASSEDFILTER)
    {
      // BAD 
      if (p->validVel)  // Use velocity of pronome if available.
	{
	  float angle = atan2(p->vel[0],p->vel[2]);
	  la->mcb.cmd = GC_TURN_GEN;
	  la->mcb.setNextArgGeneric((float)angle);
	  la->mcb.setNextArgGeneric((float).1);
	}
      else
	printf("!p->validVel\n");
    }
}



static void dropToFloorCB(BaseCreature *creature)
{
    // Creature has just been released.
    if (!creature->isGrabbed.getValue()){
	SoTransform *bt = creature->getBodyTransform();
	const SbVec3f &trans = bt->translation.getValue();
	SoTransform *xf = new SoTransform;
	xf->ref();
	static SoGetMatrixAction *gma = new SoGetMatrixAction(SbViewportRegion());
	
	xf->translation.setValue(SbVec3f(trans[0],0,trans[2]));
	xf->rotation.setValue(SbRotation(SbVec3f(0,1,0),0));
	gma->apply(xf);
	SbMatrix sbm = gma->getMatrix();
	
	MotorCommandBlock mcb(GC_SETTRANSFORM, ArgType((SbMatrix)sbm));
	((MotorCreature *)creature)->controller->executeMotorCommand(mcb);
    
//	SoTransform *bt = creature->getBodyTransform();
//	const SbVec3f &trans = bt->translation.getValue();
//      bt->translation.setValue(trans[0],0,trans[2]);
//	Should really do something better with the rotation.
//      bt->rotation.setValue(SbVec3f(0,1,0),0);
    }
}



static void avoidPronomeCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(p, dba);
  LeafActivity *la = (LeafActivity *)act;
  
  if(!dba->creature->checkIfReadyToMove(la->mcb))
    return;

  if (p && p->status == Pronome::PASSEDFILTER)
    {
      SbVec3f pos; 
      dba->creature->getPosition(pos);
      SbVec3f heading;
      dba->creature->getHeading(heading);
      SbVec3f delta = pos - p->location;

      // Project into x-z plane
      heading.setValue(heading[0],0.0,heading[2]);
      delta.setValue(delta[0],0.0,delta[2]);
      heading.normalize();
      delta.normalize();

      float dot = heading.dot(delta);
      if (fabs(1.0 - dot) > .2) // turn away from pronome
	{
	  MotorCommandBlock mcb(GC_TURN_GEN);
	  float cross = heading.cross(delta)[1]; // Take y component.
	  mcb.setNextArgGeneric(ArgType((float)cross));
	  mcb.setNextArgGeneric(ArgType((float)0.0));
	  la->issueMotorCommand(*dba->creature,mcb);
	}
      else // go straight
	{
	  MotorCommandBlock mcb(GC_TURN_GEN,ArgType((float)0.0),
				ArgType((float)0.0));
	  la->issueMotorCommand(*dba->creature,mcb);
	}
    }

  la->mcb.cmd = GC_MOVE_GEN;
  la->mcb.setNextArgGeneric((float)dba->creature->defaultFwdSpeed.getValue());
  la->mcb.setNextArgGeneric((float).8);
}

void avoidWithSchemaCB(BaseActivity *act, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME(pronome, dba);
  PFSENSOR(s, dba);
  LeafActivity *la = (LeafActivity *)act;
  
  if(!dba->creature->checkIfReadyToMove(la->mcb))
    return;
  
  if(pronome->status == Pronome::PASSEDFILTER)
    {
      // position schema
      SbVec3f lpt = s->getVecInSensorFrame(pronome->location);
      int d = la->schema->depth;
      int w = la->schema->width;
      float schemaDist = (d/2 - s->pf->depth/2);
      lpt[1] = 0;
      if(pronome->range > schemaDist)
	{
	  lpt.normalize();
	  lpt *= schemaDist;
	}
      s->pf->addPotentialFieldAtOffset((*la->schema->pfield), 
				       (int)lpt[0], (int)lpt[2]); 
      
      // update the gradient and follow it
      s->updateGradient();

      float bearing = - s->pf->weightedBearing / 2;
      bearing = (bearing>1)?1 : bearing;
      bearing = (bearing< -1)?-1 : bearing;
      if (fabs(bearing) < .2)
	la->issueMotorCommand(*dba->creature, 
			      MotorCommandBlock(GC_TURN_GEN,
						ArgType((float)0),
						ArgType((float)0)));
      else
	la->issueMotorCommand(*dba->creature, 
			      MotorCommandBlock(GC_TURN_GEN,
						ArgType((float)bearing),
						ArgType((float).2)));
      
      // Go forward.
      la->mcb.cmd = GC_MOVE_GEN;
      la->mcb.setNextArgGeneric((float)dba->creature->defaultFwdSpeed.
				getValue());
      la->mcb.setNextArgGeneric((float).8);
    }
}

void moveAwaySchemaCB(void *schemaPtr)
{
    PotentialField *schema = (PotentialField *)schemaPtr;
    schema->decay = .25; // .5
    schema->spreadPotential(SbVec3f(0, 0, 0), 4);  
}

// Use "MOUTH_JOINT" in the tagged creature as a location pronome.
static SbBool findMouthCB(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  CRM(creatureRM, rm);
  PRONOME_RM(p, rm);
  SniffSensor *ss = (SniffSensor *)creatureRM->getSensor(dua->creature);
  
  RayData *rd = NULL;
  if(ss)
    rd = ss->getDataForClosestWithTag(creatureRM->typeTag.getValue());
  
  if (rd->who)
    {
      SoPath *path = rd->who->getPathToName(rd->who,"MOUTH_JOINT");
      SbViewportRegion vpr;
      SoGetMatrixAction mtxAction(vpr);
      mtxAction.apply(path);
      SbMatrix m = mtxAction.getMatrix();
      SbVec3f pt;
      m.multVecMatrix(SbVec3f(0,0,0),pt);
//      rd->who->vecToWorld(pt);
//    printf("(%f,%f,%f)\n",pt[0],pt[1],pt[2]);
      // At this point, pt should be the location of HEADJOINT in world coords.
      
      RayData dummyRd;
      SbVec3f myPosition;
      dua->creature->getPosition(myPosition);
      dummyRd.range = (pt - myPosition).length();
      if (TESTRANGE((&dummyRd), creatureRM))
	{
	  p->object = NULL;
	  p->location = pt;
	  p->status = Pronome::PASSEDFIND;
	  p->type = Pronome::LOCATION;
	  p->range = dummyRd.range;
	  // p->bearing is not set.

	  p->validVel = FALSE;
	  creatureRM->v_tmp = (creatureRM->maxValue.getValue());
	  //creatureRM->value.setValue(creatureRM->maxValue.getValue());
	}
      else 
	p->status = Pronome::FAILEDFIND;
    }
  else
    p->status = Pronome::FAILEDFIND;
  
  return (p->status == Pronome::PASSEDFIND) ? TRUE : FALSE;
}

void simpleTrackCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    LeafActivity *la = (LeafActivity *)act;
    MotorCommandBlock mcb(GC_TURN_GEN,ArgType((float)-(pronome->bearing/2.0)),
	ArgType((float)1));
    la->issueMotorCommand(*dba->creature,mcb);
    SbVec3f p = dba->creature->position();
    float vspeed = (pronome->location[1] -p[1])/10.0;
    MotorCommandBlock mcb1(1782,ArgType(vspeed),ArgType((float).5));
    la->issueMotorCommand(*dba->creature,mcb1);
    
    la->mcb.cmd = GC_MOVE_GEN;
    la->mcb.setNextArgGeneric((float)dba->creature->defaultFwdSpeed.getValue());
    la->mcb.setNextArgGeneric((float).8);
}

void simpleTrackFloorCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    LeafActivity *la = (LeafActivity *)act;
    MotorCommandBlock mcb(GC_TURN_GEN,ArgType((float)-(pronome->bearing/2.0)),
	ArgType((float)1));
    la->issueMotorCommand(*dba->creature,mcb);
    SbVec3f p = dba->creature->position();
//    float vspeed = (pronome->location - p).length()/
//	(2*(float)dba->creature->defaultFwdSpeed.getValue());
    float vspeed = (pronome->location - p).length();
    vspeed = (vspeed >(float)dba->creature->defaultFwdSpeed.getValue())?
	dba->creature->defaultFwdSpeed.getValue():
	vspeed;
    la->mcb.cmd = GC_MOVE_GEN;
    la->mcb.setNextArgGeneric(vspeed);
    la->mcb.setNextArgGeneric((float).8);
}

void simpleRunawayCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    LeafActivity *la = (LeafActivity *)act;
    float ang = .2;
    if((pronome->bearing > 0) && (pronome->bearing < 1.57))
	ang = .2;
    else if ((pronome->bearing < 0) && (pronome->bearing > -1.57))
	ang = -.2;
    
    float gain = 4*((60 - pronome->range)/60.0);
    gain = (gain < 1.0)? 1.0 : gain;
    gain = (gain > 4.0)? 4.0 : gain;
    MotorCommandBlock mcb(GC_TURN_GEN,ArgType((float)(ang)*gain),
	ArgType((float)1));
    la->issueMotorCommand(*dba->creature,mcb);
    
    la->mcb.cmd = GC_MOVE_GEN;
    la->mcb.setNextArgGeneric((float)(dba->creature->defaultFwdSpeed.getValue()*gain/2.0));
    la->mcb.setNextArgGeneric((float).8);
}  

void simpleCircleCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    LeafActivity *la = (LeafActivity *)act;
    float offset = (pronome->bearing >= 0.0)?1.0 : -1.0;
    MotorCommandBlock mcb(GC_TURN_GEN,ArgType((float)(-pronome->bearing + offset)),
	ArgType((float)1));
    la->issueMotorCommand(*dba->creature,mcb);
    
    SbVec3f p = dba->creature->position();
    float vspeed = (pronome->location[1] -p[1])/2.0;
    MotorCommandBlock mcb1(1782,ArgType(vspeed),ArgType((float).5));
    la->issueMotorCommand(*dba->creature,mcb1);
    
    la->mcb.cmd = GC_MOVE_GEN;
    la->mcb.setNextArgGeneric((float)(dba->creature->defaultFwdSpeed.getValue()/2.0));
    la->mcb.setNextArgGeneric((float).8);
}
  
void simpleStopCB(BaseActivity *act, BehaviorUpdateAction *action)
{
    CBUA(dba, action);
    PRONOME(pronome, dba);
    LeafActivity *la = (LeafActivity *)act;
    MotorCommandBlock mcb(GC_TURN_GEN,ArgType((float)0.0),
	ArgType((float).1));
    la->issueMotorCommand(*dba->creature,mcb);
    la->mcb.cmd = GC_MOVE_GEN;
    la->mcb.setNextArgGeneric((float)0.0);
    la->mcb.setNextArgGeneric((float).2);
}  
  
void initCreatureRMCallbackList()
{
    BaseBehaviorNode::addCallback("findObject", findObject);
    BaseBehaviorNode::addCallback("findMarker", findMarker);
    BaseBehaviorNode::addCallback("testIfInside", testAInsideB);
    BaseBehaviorNode::addCallback("testPtIfInside", testPtInsideB);
    BaseBehaviorNode::addCallback("testIfFacing", isFacing);
    BaseBehaviorNode::addCallback("weightByDist", weightByDist);
    BaseBehaviorNode::addCallback("weightOne", weightOne);
    BaseBehaviorNode::addCallback("linearWeight", linearWeightByDist);
    BaseBehaviorNode::addCallback("inverseWeightByDist", inverseWeightByDist);
    BaseBehaviorNode::addCallback("proximity", proximityOnly);
    BaseBehaviorNode::addCallback("copyPronome", copyPronome);
    BaseBehaviorNode::addCallback("copyPronomeCheckRange", copyPronomeCheckRange);
    
    // Generic action callbacks
    // hardts
    BaseBehaviorNode::addCallback("moveStraight", moveStraightCB);
    BaseBehaviorNode::addCallback("turnAway", turnAwayCB);
    BaseBehaviorNode::addCallback("avoidCreature", avoidCreatureCB);
    BaseBehaviorNode::addCallback("leaveRight", leaveRightCB);
    BaseBehaviorNode::addCallback("leaveLeft", leaveLeftCB);
    BaseCreature::addCallback("dropToFloor", dropToFloorCB);
    BaseBehaviorNode::addCallback("avoidPronome", avoidPronomeCB);
    BaseBehaviorNode::addCallback("avoidWithSchema", avoidWithSchemaCB);
    BaseBehaviorNode::addCallback("moveAwaySchema", moveAwaySchemaCB);
    BaseBehaviorNode::addCallback("findMouth", findMouthCB);
    
    // bird callbacks
    BaseBehaviorNode::addCallback("schematrack", searchCB);
    BaseBehaviorNode::addCallback("schemaavoid", turnAwayCB);
    BaseBehaviorNode::addCallback("simpleTrack", simpleTrackCB);
    BaseBehaviorNode::addCallback("simpleTrackFloor", simpleTrackFloorCB);
    BaseBehaviorNode::addCallback("simpleCircle", simpleCircleCB);
    BaseBehaviorNode::addCallback("simpleStop", simpleStopCB);
    BaseBehaviorNode::addCallback("simpleRunaway", simpleRunawayCB);
}
