#include <math.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include "MediaCreature.h"
#include "BehaviorBook.h"
#include "GenericMotorControl.h"
#include "BasicDefines.h"

SO_NODE_SOURCE(BehaviorBook);

// Routines to increment/decrement/examine the page number and return the
// new result, both in terms of page numbers and in SbString form
short
BehaviorBook::pageForward()
{
  loc++;
  if (loc == NUMFILES)
    loc = 0;
  return loc;
}

short
BehaviorBook::pageBackward()
{
  loc--;
  if (loc < 0)
    loc = NUMFILES-1;
  return loc;
}

short
BehaviorBook::currentPageNum()
{
  if (loc < 0)
    loc = 0;
  else if (loc >= NUMFILES)
    loc = NUMFILES;
  return loc;
}

SbString
BehaviorBook::pageForwardName()
{
  return imageFileNames[pageForward()];
}

SbString
BehaviorBook::pageBackwardName()
{
  return imageFileNames[pageBackward()];
}

SbString
BehaviorBook::currentPageName()
{
  return imageFileNames[currentPageNum()];
}

void
BehaviorBook::initClass()
{
  SO_NODE_INIT_CLASS(BehaviorBook, MediaCreature, "MediaCreature");
}

BehaviorBook::BehaviorBook() : MediaCreature()
{
  SO_NODE_CONSTRUCTOR(BehaviorBook);
  SO_NODE_ADD_FIELD(imageFileNames, ("/mas/disks/agents/users/wex/scr0"));
  SO_NODE_ADD_FIELD(startPosition,(0.,0.,0.));
  SO_NODE_ADD_FIELD(restPosition,(0.,0.,-30.));
  SO_NODE_ADD_FIELD(epsilon,(1.0));
  SO_NODE_ADD_FIELD(activated,(FALSE));
  SO_NODE_ADD_FIELD(opening,(FALSE));
  SO_NODE_ADD_FIELD(closing,(FALSE));
  SO_NODE_ADD_FIELD(moveSpeed,(8.0));
  SO_NODE_ADD_FIELD(moveTime,(0.3));
//  changeRep(RepSwitch::REP_0);
  loc = 0;
}

SbBool	
#ifdef SO_VERSION_REVISION
BehaviorBook::readInstance(SoInput *in, unsigned short flags)
#else
BehaviorBook::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
  if (BehaviorCreature::readInstance(in, flags))
#else
  if (BehaviorCreature::readInstance(in))
#endif
    return TRUE;
  else
    return FALSE;
}

BehaviorBook::~BehaviorBook()
{
}
