#include <iostream.h>
#include <stdio.h>
#include <Inventor/SoPath.h>
#include <Inventor/fields/SoSFPath.h>
#include "BaseBehaviorNode.h"
#include "BehaviorCreatureLib.h"
#include "BehaviorPerson.h"
#include "behaviorBookRMCB.h"
#include "BehaviorBook.h"
#include "util.h" // macros defined in file live here
#include "behaviorBookActionCB.h"
#include "CreatureBehavior.h"
#include "SensorLib.h"

// we'd like to do this with ==, but since movement isn't that precise we
// need to figure if we're near enough to the wall...
#define APPROX_EQ(a,b,EPSILON) (((a) == (b)) || ((a)+(EPSILON) >= (b) && (a)-(EPSILON) <= (b)))

SbBool
bookActivated (BehaviorUpdateAction *action)
{
  //cerr << "in bookActivated " << endl;
  CBUA(dba,action);
  BehaviorBook *self = (BehaviorBook *)dba->creature;
  return(self->activated.getValue());
}

SbBool
checkPositionCB (ReleasingMechanism *, BehaviorUpdateAction *action)
{
  CBUA(dba,action);
  BehaviorBook *self = (BehaviorBook *)dba->creature;
  float targetX, targetY, targetZ;
  float currX, currY, currZ;
  float ep = self->epsilon.getValue();

  //cerr << "in checkPositionCB " << endl;
  // if we're moving to the wall, we'll be in REP_1, otherwise don't
  // bother with doing all this
  if (self->currentRep != RepSwitch::REP_1)
    return(FALSE);
  self->restPosition.getValue().getValue(targetX, targetY, targetZ);
  self->position().getValue(currX, currY, currZ);
  return(APPROX_EQ(targetX, currX, ep) && APPROX_EQ(targetY, currY, ep)
	 && APPROX_EQ(targetZ, currZ, ep));
}

SbBool
personExtendedLeftBookCB (ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  PRONOME_RM(p, rm);

  BehaviorPerson *person = (BehaviorPerson *)(p->object);
/*
  cerr << "value of personExtendedLeft is " <<
    (person->extendedLeft.getValue()
     && !(person->extendedRight.getValue())
     && bookActivated(action)) << endl;
     */
  return (person->extendedLeft.getValue()
	  && !(person->extendedRight.getValue())
	  && bookActivated(action));

}

SbBool
personExtendedRightCB (ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dba, action);
  PRONOME_RM(p, rm);

  BehaviorPerson *person = (BehaviorPerson *)(p->object);
/*
  cerr << "value of personExtendedRight is " <<
    (person->extendedRight.getValue()
     && !(person->extendedRight.getValue())
     && bookActivated(action)) << endl;
     */
  return (person->extendedRight.getValue()
	  && !(person->extendedLeft.getValue())
	  && bookActivated(action));
}

SbBool
checkShelfCB (ReleasingMechanism *, BehaviorUpdateAction *action)
{
  CBUA(dba,action);
  BehaviorBook *self = (BehaviorBook *)dba->creature;
  float targetX, targetY, targetZ;
  float currX, currY, currZ;
  float ep = self->epsilon.getValue();

  // cerr << "in checkShelfCB " << endl;
  // if we're already on the wall, we'll be in REP_2 and don't need to
  // bother with doing all this
  if (self->currentRep != RepSwitch::REP_3)
    return(FALSE);
  self->startPosition.getValue().getValue(targetX, targetY, targetZ);
  self->position().getValue(currX, currY, currZ);
  return(APPROX_EQ(targetX, currX, ep) && APPROX_EQ(targetY, currY, ep)
	 && APPROX_EQ(targetZ, currZ, ep));
}

void
initBehaviorBookRMCallbackList()
{
  BaseBehaviorNode::addCallback("personExtendedLeftBook", personExtendedLeftBookCB);
  BaseBehaviorNode::addCallback("personExtendedRight", personExtendedRightCB);
  BaseBehaviorNode::addCallback("checkPosition", checkPositionCB);
  BaseBehaviorNode::addCallback("checkShelf", checkShelfCB);
}
