#include <math.h>
#include <Inventor/SbLinear.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include "Car.h"
#include "CarMotorSkills.h"
#include "CreatureLib.h"
#include "CarDefines.h"
#include "GenericMotorControl.h"


SO_NODE_SOURCE(Car);

void
Car::initClass()
{
    SO_NODE_INIT_CLASS(Car, MotorCreature, "MotorCreature");
}

Car::Car() : MotorCreature()
{
    SO_NODE_CONSTRUCTOR(Car);
    SO_NODE_ADD_FIELD(wheelRadius, (0.8673635));
    SO_NODE_ADD_FIELD(wheelBase,   (7.50640));
}

SbBool	
#ifdef SO_VERSION_REVISION
Car::readInstance(SoInput *in, unsigned short flags) 
#else
Car::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
   if (MotorCreature::readInstance(in, flags)) {
#else
   if (MotorCreature::readInstance(in)) {
#endif

	wheelCircum = 2.0 * 3.14159 * wheelRadius.getValue();
	turnBase.setValue(wheelBase.getValue());
	return(1);
   } else {
	return(0);
   }
}


Car::~Car()
{
}

void Car::addMotorSkills()
{
    MotorCreature::addMotorSkills();

    CarMove *move = new CarMove(this, "move");
    getController()->addCommand( GC_MOVE, "move", move, TRUE, 
				 ArgType((float)DEFAULT_SPEED), 
				 ArgType((float)DEFAULT_SECS_TO_SPEED));

    CarTurn *turn = new CarTurn(this, "turn");
    getController()->addCommand( GC_TURN, "turn", turn, TRUE, 
				 ArgType((float)DEFAULT_TURN), 
				 ArgType((float)DEFAULT_SECS_TO_TURN));
    
    CarTurnToward *turnToward = new CarTurnToward(this, "turnToward");
    getController()->addCommand( GC_TURN_TOWARD, "turnToward", turnToward,TRUE,
				 ArgType((float)DEFAULT_TURN_TOWARD_X), 
				 ArgType((float)DEFAULT_TURN_TOWARD_Y), 
				 ArgType((float)DEFAULT_TURN_TOWARD_Z));
    getController()->addCommand( GC_TURN_TOWARD_OFF, "turnTowardOff", 
				 turnToward, FALSE);

    CarOn *on = new CarOn(this, "on");
    getController()->addCommand( GC_ON, "on", on, TRUE);

    CarOff *off = new CarOff(this, "off");
    getController()->addCommand( GC_OFF, "off", off, TRUE);
}

void Car::addBehaviorSystem()
{
    
}

void Car::getDOFInterpolators()
{
    // NOTE !!!
    // Order must agree with values in DogDefines.h
    // NOTE !!!

    installXYZInterpolator("TILT_JOINT");
    installXYZInterpolator("PITCH_JOINT");
    installXYZInterpolator("BOUNCE_JOINT");

    installXYZInterpolator("ROLL_JOINT");
    installXYZInterpolator("TURN_JOINT");
}

#ifdef 0
int Car::stepTo(UpdateStateData &usd)
{
    float t = usd.time;
    float distance = speed * t;

    if (distance){
 	SbVec3f h, p;
	getHeading(h);
	getPosition(p);

	if (angle == 0.0) {

		p += h * distance;
		setPosition(p);
	} else {

	    SbVec3f center;  
	    SbRotation rot = getBodyTransform()->rotation.getValue();
	    rot.multVec(SbVec3f(0,0,-1), center); // vec pointing out right side
	    center.normalize();

   	    float the_angle = distance / (wheelBase.getValue() / fcos(1.570795 - angle));
	    center *= (wheelBase.getValue() * ftan(1.570795 - angle));
	    center += p;

	    SbMatrix mat;
	    mat.setTransform(SbVec3f(0,0,0), 
			     SbRotation(SbVec3f(0, 1, 0), the_angle), 
			     SbVec3f(1,1,1),
			     SbRotation(),
			     center);
	    mat.multDirMatrix(h, h);
	    mat.multVecMatrix(p, p);
	    setPosition(p);
	    setHeading(SbRotation(SbVec3f(1, 0, 0), h));
	    
	}
    }
    return 1;
}

#endif
void Car::getEyeView(SoCamera *xf)
{
    SbRotation r = getBodyTransform()->rotation.getValue();
    SbRotation rr = r*SbRotation(SbVec3f(0, 1, 0), M_PI);
    SbVec3f x = getBodyTransform()->translation.getValue();
    x = x*2;
    x[1]+=4.0;
    xf->orientation.setValue(rr);
    xf->position.setValue(x); 
}










