#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/SoPath.h>

#include "MotorLib.h"
#include "CarMotorSkills.h"
#include "CreatureLib.h"
#include "CarDefines.h"

/*********************************************************************
 * 
 * CarMOVE Skill
 * 
 *********************************************************************/

CarMove::CarMove(Car *p, char *pname) : MotorSkill(p, pname)
{
    dofs[numDofs++] = (Dof *)(parent->dofs[ROLL_INDEX]);
}

int CarMove::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	start_speed = ((Car *)parent)->speed;
	time_since  = 0.0; 

	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_speed = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);
    }

fprintf(stderr, "Prepare To run : ds %f, secs %f\n", desired_speed, secs_to_reach);    

    return okToTurnOn;
}


int CarMove::modifyRun(void *arg)
{
	start_speed = ((Car *)parent)->speed;
	time_since  = 0.0; 


	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_speed = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);

        return 1;  
}



int CarMove::prepareToTurnOff(void *)
{
    int okToTurnOff;

    if ((fabs(((Car *)parent)->speed) <= .05) && (desired_speed == 0.0)) 
	okToTurnOff = 1;
    else 
	okToTurnOff = 0;

    
    if (okToTurnOff) {
	((Car *)parent)->speed = 0.0;
        desired_speed = 0.0;
        start_speed = 0.0;
        secs_to_reach = 0.0;
    }
	
    return okToTurnOff;
}

void CarMove::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {

    time_since += usd.time;

    if (((Car *)parent)->speed != desired_speed) {
	if (time_since < secs_to_reach) {
	    float tmp = time_since / secs_to_reach;
	    ((Car *)parent)->speed = start_speed + (tmp * (desired_speed - start_speed));
	} else {
	    ((Car *)parent)->speed = desired_speed;
	}
/*
	float v;
  	char cmd[256];
	v = ((Car *)parent)->speed / 60.0;
	if (v > .5) v = .5;
	v = 1.0 - v;
	sprintf(cmd, "m CarMotor %f", v);
	((Car *)parent)->sendSoundCommand(usd, cmd);
*/
    } else 
	turnOff();

    }

    MotorSkill::updateStateAt(usd);
    return;
}


/*********************************************************************
 * 
 * CarTurn Skill
 * 
 *********************************************************************/

CarTurn::CarTurn(Car *p, char *pname) : MotorSkill(p, pname)
{
    dofs[numDofs++] = (Dof *)(parent->dofs[TURN_INDEX]);
}

int CarTurn::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	start_angle = ((Car *)parent)->angle;
	time_since  = 0.0; 

	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_angle = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);
    }

fprintf(stderr, "Prepare To run : da %f, secs %f\n", desired_angle, secs_to_reach);    

    return okToTurnOn;
}


int CarTurn::modifyRun(void *arg)
{
	start_angle = ((Car *)parent)->angle;
	time_since  = 0.0; 


	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_angle = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);

        return 1;  
}



int CarTurn::prepareToTurnOff(void *)
{
    int okToTurnOff;

    if (((Car *)parent)->angle == desired_angle)
	okToTurnOff = 1;
    else 
	okToTurnOff = 0;

    
    if (okToTurnOff) {
	((Car *)parent)->angle = desired_angle;
        secs_to_reach = 0.0;
    }
	
    return okToTurnOff;
}

void CarTurn::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {

    Car *c = (Car *)parent;
    SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[0]->interpolator;

    time_since += usd.time;

    if (c->angle != desired_angle) {
	if (time_since < secs_to_reach) {
	    float tmp = time_since / secs_to_reach;
	    c->angle = start_angle + (tmp * (desired_angle - start_angle));
	} else {
	    c->angle = desired_angle;
	}

        sif->alpha.setValue((c->angle + 0.7853975) / 1.570795);
    } else 
	turnOff();

    }

    MotorSkill::updateStateAt(usd);
    return;
}


/*********************************************************************
 * 
 * CarTurnToward Skill
 * 
 *********************************************************************/

CarTurnToward::CarTurnToward(Car *p, char *pname) : MotorSkill(p, pname)
{
    dofs[numDofs++] = (Dof *)(parent->dofs[TURN_INDEX]);
}

int CarTurnToward::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	x = argVals->getFloatAt(0);
	y = argVals->getFloatAt(1);
	z = argVals->getFloatAt(2);
	angle = 1.0;
    }

    return okToTurnOn;
}

int CarTurnToward::modifyRun(void *arg)
{
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	x = argVals->getFloatAt(0);
	y = argVals->getFloatAt(1);
	z = argVals->getFloatAt(2);

        return 1;  
}

int CarTurnToward::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;

    if (okToTurnOff) {
fprintf(stderr, "turnToward off\n");
    }
	
    return okToTurnOff;
}

void CarTurnToward::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {

    Car *c = (Car *)parent;
    SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[0]->interpolator;

    float angle, dis;
    c->calcBearingAndRange(SbVec3f(x,y,z), angle, dis);

    if (angle < -0.7) angle = -0.7;
    if (angle >  0.7) angle =  0.7;

    sif->alpha.setValue((-angle + 0.7853975) / 1.570795);
    c->angle = -angle;

    MotorSkill::updateStateAt(usd);
    }	
}

/*********************************************************************
 * 
 * CarOn Skill
 * 
 *********************************************************************/
CarOn::CarOn(Car *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void CarOn::updateImmediate(void *)
{
    SoundSkill::triggerSound("CarMotor");
    return;
}

/*********************************************************************
 * 
 * CarOff Skill
 * 
 *********************************************************************/
CarOff::CarOff(Car *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void CarOff::updateImmediate(void *)
{
    SoundSkill::turnoffSound("CarMotor");
    return;
}


