#ifndef CARSKILLS
#define CARSKILLS

#include "MotorSkill.h"
#include "Car.h"

/**********************************************************************
 *
 * CarMove
 *
 * This MotorSkill implements moving in what ever direction the wheels are
 * pointing.
 *
 *  args (optional):
 *    0 - new desired speed
 *    1 - second to reach desired speed
 *
 **********************************************************************/

class CarMove: public MotorSkill {
  public:
    float desired_speed;
    float start_speed;
    float time_since;
    float secs_to_reach;

    CarMove(Car *p, char *pname);

    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg);
    void updateStateAt(UpdateStateData &usd);
    int modifyRun(void *args);
};

/*********************************************************************
 * 
 *  CarTURN
 *
 *  args (optional):
 *	0 - new desired angle
 *	1 - seconds to reach that angle
 *
 *********************************************************************/

class CarTurn: public MotorSkill {
    public:

    float desired_angle;
    float start_angle;
    float time_since;
    float secs_to_reach;

    CarTurn(Car *p, char *pname);

    int modifyRun(void *args);
    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg); 
    void updateStateAt(UpdateStateData &usd);
};


/*********************************************************************
 * 
 *  CarTURNTOWARD
 *
 *  args (optional):
 *	0 - x
 *	1 - y
 *	2 - z
 *
 *********************************************************************/

class CarTurnToward: public MotorSkill {
    public:

    float x, y, z;
    float angle;

    CarTurnToward(Car *p, char *pname);

    int modifyRun(void *args);
    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg); 
    void updateStateAt(UpdateStateData &usd);
};

/*********************************************************************
 * 
 *  CarON
 *
 *  args (none):
 *
 *********************************************************************/

class CarOn: public MotorSkill {
    public:

    CarOn(Car *p, char *pname);
    void updateImmediate(void *arg);
};


/*********************************************************************
 * 
 *  CarOFF
 *
 *  args (none):
 *
 *********************************************************************/

class CarOff: public MotorSkill {
    public:

    CarOff(Car *p, char *pname);
    void updateImmediate(void *arg);
};

#endif











