/*
 */
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSurroundScale.h>
#include <Inventor/nodes/SoAntiSquish.h>
#include <Inventor/sensors/SoFieldSensor.h>

// Include files for child dragger classes.
#include <Inventor/draggers/SoRotateCylindricalDragger.h>
#include <Inventor/draggers/SoTranslate2Dragger.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>

// Include file for our new class.
#include "AnchorDragger.h"
#include "AnchorDraggerGeom.h"

SO_KIT_SOURCE(AnchorDragger);

void
AnchorDragger::initClass()
{
   SO_KIT_INIT_CLASS(AnchorDragger, SoDragger, "Dragger");	
}

AnchorDragger::AnchorDragger()
{
   SO_KIT_CONSTRUCTOR(AnchorDragger);

   //
   // Don't create "surroundScale" by default. It's only put 
   // to use if this dragger is used within a manipulator.
   //
   SO_KIT_ADD_CATALOG_ENTRY(surroundScale, SoSurroundScale,
      TRUE, topSeparator, geomSeparator,TRUE);
   //
   // Create an anti-squish node by default.
   //
   SO_KIT_ADD_CATALOG_ENTRY(antiSquish, SoAntiSquish,
      FALSE, topSeparator, geomSeparator,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(XYTranslator, SoTranslate2Dragger,
      TRUE, topSeparator, geomSeparator,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(XZTranslatorSep, SoSeparator,
      FALSE, topSeparator, geomSeparator,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(XZTranslatorRot, SoRotation,
      TRUE, XZTranslatorSep, ,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(XZTranslator, 
      SoTranslate2Dragger, TRUE, XZTranslatorSep, ,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(YZTranslatorSep, SoSeparator,
      FALSE, topSeparator, geomSeparator,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(YZTranslatorRot, SoRotation,
      TRUE, YZTranslatorSep, ,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(YZTranslator, 
      SoTranslate2Dragger, TRUE, YZTranslatorSep, ,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(XRotatorSep, SoSeparator,
      FALSE, topSeparator, geomSeparator,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(XRotatorRot, SoRotation,
      TRUE, XRotatorSep, ,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(XRotator, 
      SoRotateCylindricalDragger, TRUE, XRotatorSep, ,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(YRotator, 
      SoRotateCylindricalDragger, TRUE, topSeparator, 
      geomSeparator, TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(ZRotatorSep, SoSeparator,
      FALSE, topSeparator, geomSeparator,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(ZRotatorRot, SoRotation,
      TRUE, ZRotatorSep, ,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(ZRotator, 
      SoRotateCylindricalDragger, TRUE, ZRotatorSep, ,TRUE);

   // Read geometry resources. Only do this the first time we
   // construct one. 'geomBuffer' contains our compiled in
   // defaults. The user can override these by specifying new
   // scene graphs in the file:
   // $(SO_DRAGGER_DIR)/AnchorDragger.iv"
   if (SO_KIT_IS_FIRST_INSTANCE())
      readDefaultParts("AnchorDragger.iv", geomBuffer, 
      sizeof(geomBuffer));

   // Fields that always show current state of the dragger.
   SO_KIT_ADD_FIELD(rotation, (0.0, 0.0, 0.0, 1.0));
   SO_KIT_ADD_FIELD(translation, (0.0, 0.0, 0.0));

   // Creates parts list and default parts for this nodekit.
   SO_KIT_INIT_INSTANCE();

   // Make the anti-squish node surround the biggest dimension
   SoAntiSquish *myAntiSquish 
      = SO_GET_ANY_PART(this, "antiSquish", SoAntiSquish);
   myAntiSquish->sizing = SoAntiSquish::BIGGEST_DIMENSION;

   // Create the simple draggers that comprise this dragger.
   // This dragger has four simple pieces:  
   //    1 Translate2Dragger
   //    3 RotateCylindricalDraggers
   // In the constructor, we just call SO_GET_ANY_PART to
   // build each dragger.
   // Within the setUpConnections() method, we will
   // take care of giving these draggers new geometry and 
   // establishing their callbacks.

   // Create the translator dragger.    
   SoDragger *tXYDragger = SO_GET_ANY_PART(this, "XYTranslator", 
      SoTranslate2Dragger);
   SoDragger *tXZDragger = SO_GET_ANY_PART(this, "XZTranslator", 
      SoTranslate2Dragger);
   SoDragger *tYZDragger = SO_GET_ANY_PART(this, "YZTranslator", 
      SoTranslate2Dragger);

   // Create the XRotator dragger.    
   SoDragger *XDragger = SO_GET_ANY_PART(this, "XRotator", 
      SoRotateCylindricalDragger);

   // Create the YRotator dragger.    
   SoDragger *YDragger = SO_GET_ANY_PART(this, "YRotator", 
      SoRotateCylindricalDragger);

   // Create the ZRotator dragger.    
   SoDragger *ZDragger = SO_GET_ANY_PART(this, "ZRotator", 
      SoRotateCylindricalDragger);


   SoRotation *tYZRot = new SoRotation;
   tYZRot->rotation.setValue(
      SbRotation(SbVec3f(0,0,1), SbVec3f(1,0,0)));
   setAnyPartAsDefault("YZTranslatorRot", tYZRot);

   SoRotation *tXZRot = new SoRotation;
   tXZRot->rotation.setValue(
      SbRotation(SbVec3f(0,0,1), SbVec3f(0,1,0)));
   setAnyPartAsDefault("XZTranslatorRot", tXZRot);


   // Set rotations in "XRotatorRot" and "ZRotatorRot" parts.
   // These parts will orient the draggers from their default 
   // (rotating about Y) to the desired configurations.
   // By calling 'setAnyPartAsDefault' instead of 'setAnyPart'
   // we insure that they will not be written out, unless
   // they are changed later on.
   SoRotation *XRot = new SoRotation;
   XRot->rotation.setValue(
      SbRotation(SbVec3f(0,1,0), SbVec3f(1,0,0)));
   setAnyPartAsDefault("XRotatorRot", XRot);

   SoRotation *ZRot = new SoRotation;
   ZRot->rotation.setValue(
      SbRotation(SbVec3f(0,1,0), SbVec3f(0,0,1)));
   setAnyPartAsDefault("ZRotatorRot", ZRot);

   // Updates the fields when motionMatrix changes 
   addValueChangedCallback(&AnchorDragger::valueChangedCB);

   // Updates motionMatrix when either field changes.
   rotFieldSensor = new SoFieldSensor(
      &AnchorDragger::fieldSensorCB, this);
   rotFieldSensor->setPriority(0);
   translFieldSensor = new SoFieldSensor(
      &AnchorDragger::fieldSensorCB,this);
   translFieldSensor->setPriority(0);

   setUpConnections( TRUE, TRUE );
}

AnchorDragger::~AnchorDragger()
{
   if (rotFieldSensor)
      delete rotFieldSensor;
   if (translFieldSensor)
      delete translFieldSensor;
}


SbBool
AnchorDragger::setUpConnections( SbBool onOff, 
			    SbBool doItAlways)
{
   if ( !doItAlways && connectionsSetUp == onOff)
      return onOff;

   if (onOff) {
      // We connect AFTER base class.
      SoDragger::setUpConnections(onOff, doItAlways);


       SoDragger *tD 
	  = (SoDragger *) getAnyPart("XYTranslator", FALSE);
       // [a] Set up the parts in the child dragger...
       tD->setPartAsDefault("translator",
	  "AnchorTranslatorTranslator");
       tD->setPartAsDefault("translatorActive",
	  "AnchorTranslatorTranslatorActive");
       tD->setPartAsDefault("feedback",
	  "AnchorTranslatorFeedback");
       tD->setPartAsDefault("feedbackActive",
	  "AnchorTranslatorFeedbackActive");
       tD->addStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       tD->addFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       registerChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("YZTranslator", FALSE);
       // [a] Set up the parts in the child dragger...
       tD->setPartAsDefault("translator",
	  "AnchorTranslatorTranslator");
       tD->setPartAsDefault("translatorActive",
	  "AnchorTranslatorTranslatorActive");
       tD->setPartAsDefault("feedback",
	  "AnchorTranslatorFeedback");
       tD->setPartAsDefault("feedbackActive",
	  "AnchorTranslatorFeedbackActive");
       tD->addStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       tD->addFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       registerChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("XZTranslator", FALSE);
       // [a] Set up the parts in the child dragger...
       tD->setPartAsDefault("translator",
	  "AnchorTranslatorTranslator");
       tD->setPartAsDefault("translatorActive",
	  "AnchorTranslatorTranslatorActive");
       tD->setPartAsDefault("feedback",
	  "AnchorTranslatorFeedback");
       tD->setPartAsDefault("feedbackActive",
	  "AnchorTranslatorFeedbackActive");
       tD->addStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       tD->addFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       registerChildDragger(tD);


       SoDragger *XD 
	  = (SoDragger *) getAnyPart("XRotator", FALSE);
       // [a] Set up the parts in the child dragger...
       XD->setPartAsDefault("rotator",
	  "AnchorRotatorRotator");
       XD->setPartAsDefault("rotatorActive",
	  "AnchorRotatorRotatorActive");
       XD->setPartAsDefault("feedback",
	  "AnchorRotatorFeedback");
       XD->setPartAsDefault("feedbackActive",
	  "AnchorRotatorFeedbackActive");
       // [b] and [c] Add the callbacks and register the child
       XD->addStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       XD->addFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       registerChildDragger(XD);

       SoDragger *YD 
	  = (SoDragger *) getAnyPart("YRotator", FALSE);
       // [a] Set up the parts in the child dragger...
       YD->setPartAsDefault("rotator",
	  "AnchorRotatorRotator");
       YD->setPartAsDefault("rotatorActive",
	  "AnchorRotatorRotatorActive");
       YD->setPartAsDefault("feedback",
	  "AnchorRotatorFeedback");
       YD->setPartAsDefault("feedbackActive",
	  "AnchorRotatorFeedbackActive");
       // [b] and [c] Add the callbacks and register the child
       YD->addStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       YD->addFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       registerChildDragger(YD);

       SoDragger *ZD 
	  = (SoDragger *) getAnyPart("ZRotator", FALSE);
       // [b] Set up the parts in the child dragger...
       ZD->setPartAsDefault("rotator",
	  "AnchorRotatorRotator");
       ZD->setPartAsDefault("rotatorActive",
	  "AnchorRotatorRotatorActive");
       ZD->setPartAsDefault("feedback",
	  "AnchorRotatorFeedback");
       ZD->setPartAsDefault("feedbackActive",
	  "AnchorRotatorFeedbackActive");
       // [c] and [d] Add the callbacks and register the child
       ZD->addStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       ZD->addFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       registerChildDragger(ZD);


      // Call the sensor CB to make things up-to-date.
      fieldSensorCB(this, NULL);

      // Connect the field sensors
      if (translFieldSensor->getAttachedField() != &translation)
	 translFieldSensor->attach( &translation );
      if (rotFieldSensor->getAttachedField() != &rotation)
	 rotFieldSensor->attach( &rotation );
   }
   else {
      // We disconnect BEFORE base class.

      // Remove the callbacks from the child draggers,
      // and unregister them as children.
       SoDragger *tD 
	  = (SoDragger *) getAnyPart("XYTranslator", FALSE);
       tD->removeStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       tD->removeFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("XZTranslator", FALSE);
       tD->removeStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       tD->removeFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("YZTranslator", FALSE);
       tD->removeStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       tD->removeFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(tD);

       SoDragger *XD 
	  = (SoDragger *) getAnyPart("XRotator", FALSE);
       XD->removeStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       XD->removeFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(XD);

       SoDragger *YD 
	  = (SoDragger *) getAnyPart("YRotator", FALSE);
       YD->removeStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       YD->removeFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(YD);

       SoDragger *ZD 
	  = (SoDragger *) getAnyPart("ZRotator", FALSE);
       ZD->removeStartCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       ZD->removeFinishCallback(
	  &AnchorDragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(ZD);

      // Disconnect the field sensors.
      if (translFieldSensor->getAttachedField())
	translFieldSensor->detach();
      if (rotFieldSensor->getAttachedField())
	rotFieldSensor->detach();

      SoDragger::setUpConnections(onOff, doItAlways);
   }

   return !(connectionsSetUp = onOff);
}


// Called when the motionMatrix changes. Sets the "translation"
// and "rotation" fields based on the new motionMatrix
void
AnchorDragger::valueChangedCB(void *, SoDragger *inDragger)
{
   AnchorDragger *myself = (AnchorDragger *) inDragger;

   // Factor the motionMatrix into its parts
   SbMatrix motMat = myself->getMotionMatrix();
   SbVec3f   trans, scale;
   SbRotation rot, scaleOrient;
   motMat.getTransform(trans, rot, scale, scaleOrient);

   // Set the fields. Disconnect the sensors while doing so.
   myself->rotFieldSensor->detach();
   myself->translFieldSensor->detach();
   if (myself->rotation.getValue() != rot)
      myself->rotation = rot;
   if (myself->translation.getValue() != trans)
      myself->translation = trans;
   myself->rotFieldSensor->attach(&(myself->rotation));
   myself->translFieldSensor->attach(&(myself->translation));
}

// If the "translation" or "rotation" field changes, changes
// the motionMatrix accordingly.
void
AnchorDragger::fieldSensorCB(void *inDragger, SoSensor *)
{
   AnchorDragger *myself = (AnchorDragger *) inDragger;

   SbMatrix motMat = myself->getMotionMatrix();
   myself->workFieldsIntoTransform(motMat);

   myself->setMotionMatrix(motMat);
}

// When any child dragger starts or ends a drag, tell the
// "surroundScale" part (if it exists) to invalidate its 
// current bounding box value.
void 
AnchorDragger::invalidateSurroundScaleCB(void *parent, 
   SoDragger *)
{
   AnchorDragger *myParentDragger 
      = (AnchorDragger *) parent;

   // Invalidate the surroundScale, if it exists.
   SoSurroundScale *mySS = SO_CHECK_PART(
      myParentDragger, "surroundScale", SoSurroundScale);
   if (mySS != NULL)
      mySS->invalidate();
}

void
AnchorDragger::setDefaultOnNonWritingFields()
{
    // These nodes pointed to by these part-fields may 
    // change after construction, but we
    // don't want to write them out.
    surroundScale.setDefault(TRUE);
    antiSquish.setDefault(TRUE);

    SoDragger::setDefaultOnNonWritingFields();
}
