#ifndef _ARTICULATEDBODY_
#define _ARTICULATEDBODY_

#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/fields/SoMFVec3f.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/nodes/SoTransform.h>

#include "JointModel.h"
#include "../CreatureLib/Vector.h"

#define HINTSENSORPRIORITY JOINTSENSORPRIORITY + 1
#define STRENGTHSENSORPRIORITY HINTSENSORPRIORITY + 1
#define GOALSENSORPRIORITY STRENGTHSENSORPRIORITY + 1

class SoTransform;
class SoPath;
class SoSeparator;
class ArticulatedBody;
class SoPath;
class AnchorManip;

class ABCache {
    public:
    SbVec3f goal;
    Vector jointData;
    ABCache(SbVec3f ng, ArticulatedBody *ab);
};

class ABConstraint {
    public:
    SoPath *relativeToPath;
    SoNode *adjustNode;
    SbVec3f weight;
    SbVec3f constrainedPt;
    SbVec3f newPt;
    SbVec3f delta;
    SbBool applyFlag;
    ArticulatedBody *client;
    ABConstraint(ArticulatedBody *c, SoNode *root, SoNode *aNode,
	 SbBool applyOnEnd = TRUE);
    void beginConstraint();
    void endConstraint();
    void applyConstraint();
};
#define TOLERANCE 1.0e-3
#define ACCEPT 1.0e6

class ArticulatedBody : public SoSeparator {
    SO_NODE_HEADER(ArticulatedBody);

public:
    static void initClass();

    SoSFVec3f goal;
    SoSFFloat tol;
    SoMFVec3f angleHints; // use to set center angles, treated as packed data
    SoMFVec3f angleHintStrengths; // use to set center gains, treated as packed data
    SoSFNode interpolatorCache; // should contain pointers to EngineHelpers
    SoSFInt32 pickInterpolator; // index into interpolators
    SoSFBool useCache;
    
    JointModel *rootJoint;
    JointModel *firstActiveJoint;
    JointModel *lastJoint;
    SoPath *path;
    SbPList *ivkCache;
    Jacobian *jacobian;
    SbBool endpointFixedFlag;
    
    
    ArticulatedBody(SoTransform *anAnchor);

    virtual void setAnchorPosition(SbVec3f np);
    void attachAnchorAndJacobian(JointModel *newroot, JointModel *fa);
    void initCenterAngleAndGain(float g);
    ArticulatedBody *findMatchingShape(SoShape *testShape);
    
    Vector saveJointAngles();
    void restoreJointAngles(const Vector &sja);
    int checkCache(SbVec3f g, float cache_tol);
    void addToCache(SbVec3f g);
    void getEndPoint(SbVec3f &ep);
    void vecToBase(SbVec3f &ep);
    void vecToEnd(SbVec3f &ep);
    void vecToBaseUsing(SbVec3f &ep, SoPath *p);
    void vecToEndUsing(SbVec3f &ep, SoPath *p);
    
    float moveEndPointTo(SbVec3f *nep, float withinTol);
    float _moveEndPointTo(SbVec3f *nep, int iter, float testTol);

    void setFirstActiveJoint(JointModel *newFirst);
    SoPath *buildPath(SoNode *startNode);
    
    SoTransform *getAnchor() { return (SoTransform *)getChild(0); };
    int getAnchorIndex() { return 0; };

    JointModel *getLastJoint() { return lastJoint; };

    void addVisual();
    void removeVisual();

    void addAllVisuals();
    void removeAllVisuals();

    int getNumJoints();
    void setPickInterpolator(int which);
    void addJointToEnd(JointModel *j);

    void recenterJointLimb(JointModel *j);

    static ArticulatedBody *isSelection(SoPath *p);
    void addManipulator(SoPath *p);
    void removeManipulator();
    void moveToConstrainedXYGoal(SbVec3f &pt);
    void moveToGoalImmediate();

protected:
    
    ArticulatedBody();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool	readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool	readInstance(SoInput *in);
#endif

    virtual ~ArticulatedBody();
    SoFieldSensor *goalSensor;
    static void goalSensorCB(void *, SoSensor *); 

    SoFieldSensor *angleHintsSensor;
    SoFieldSensor *gainHintsSensor;
    static void hintsSensorCB(void *, SoSensor *);
    SoFieldSensor *pickInterpolatorSensor;
    static void pickInterpolatorSensorCB(void *, SoSensor *); 

    static void dragMotionCallback(void *data, SoDragger *);
    static void dragStartCallback(void *data, SoDragger *);

private:

    AnchorManip *myManip;
    SoPath     *manipPath;

    SbMatrix saveMat;
};

#endif

