#include <stdio.h>

#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoPath.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include "TimeStat.h"
#include "ArticulatedBody.h"
#include "JointModel.h"
#include "BaseCreature.h"
#include "TimeDependentComponent.h"
#include "RepSwitch.h"
/*********************************************************************
 * 
 *	    Helper function
 * 
 *********************************************************************/

float rayBoxIntersection(SbVec3f lu,SbVec3f ld,SbVec3f minB,SbVec3f maxB,float *far);

float rayBoxIntersection(SbVec3f lu,SbVec3f ld,SbVec3f minB,SbVec3f maxB,float *far)
{
    float tnear = -1e6;
    float tfar = 1e6;
    float t1,t2,tmp;
    int i;
    
    if((lu[0]>minB[0])&&(lu[0]<maxB[0])
	&&(lu[2]>minB[2])&&(lu[2]<maxB[2])){
	   // printf("inside box \n");
	    return -1000;
	}
     for(i=0;i<3;i++){
       if(ld[i]==0.0){
	    if((lu[i]<minB[i])||(lu[i]>maxB[i])){
		return -1;
	    }
	}
	else {
	    t1 = (minB[i]-lu[i])/ld[i];
	    t2 = (maxB[i]-lu[i])/ld[i];
	    if(t1>t2){
		tmp=t2;
		t2=t1;
		t1 =tmp;
	    }
	    if(t1>tnear)
		tnear = t1;
	    if(t2<tfar)
		tfar = t2;
	    if((tnear>tfar)||(tfar<0.0))
		return -1;
	}
    }
    //printf("ray enters %f and exits %f\n",tnear,tfar);
    *far = tfar;
    return tnear;
}
 
/*********************************************************************
 * 
 *	    BaseCreature  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(BaseCreature);

void BaseCreature::initClass()
{
    SO_NODE_INIT_CLASS(BaseCreature, SoSeparator, "Separator");
}

BaseCreature *BaseCreature::loadCreature(char *filename)
{
    SoInput mySceneInput;
    SoNode *fn;
    if (!mySceneInput.openFile(filename)) {
       printf("Cannot open file\n");
       return NULL;
    }
    SoDB::read(&mySceneInput, fn);
    if (!(fn->isOfType(BaseCreature::getClassTypeId()))) {
	printf("Does not contain a creature\n");
	return NULL;
    }
    else {
// below line now happens in readInstance -- kbrussel 6/9/95
//	((BaseCreature *)fn)->geometryFileName.setValue(filename);
	return (BaseCreature *)fn;
    }
}
 
BaseCreature::BaseCreature() : SoSeparator()
{
  SO_NODE_CONSTRUCTOR(BaseCreature);
  SO_NODE_ADD_FIELD(tag, (-1));
  SO_NODE_ADD_FIELD(realTag, (-1));
  SO_NODE_ADD_FIELD(registryNumber, (-1));
  SO_NODE_ADD_FIELD(defaultSwitchValue, (-3));
  SO_NODE_ADD_FIELD(isGrabbed, (FALSE));
  SO_NODE_ADD_FIELD(isMoving, (FALSE));
  SO_NODE_ADD_FIELD(updateCBName,("none"));
  SO_NODE_ADD_FIELD(isGrabbedCBName,("none"));
  SO_NODE_ADD_FIELD(resetCBName,("none"));
  SO_NODE_ADD_FIELD(geometryFileName,("none"));
  SbViewportRegion vpr;
  bba = new SoGetBoundingBoxAction(vpr);
  resetCB = NULL;
  updateCB = NULL;
  isGrabbedCB = NULL;
  grabbedBy = NULL;
}

BaseCreature::~BaseCreature()
{
}

#ifdef SO_VERSION_REVISION
SbBool BaseCreature::readInstance(SoInput *in, unsigned short flags) 
#else
SbBool BaseCreature::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoSeparator::readInstance(in, flags)){
#else
    if (SoSeparator::readInstance(in)){
#endif
	this->ref();
	realTag.setValue(tag.getValue());
	SoSearchAction sa;
	sa.setType(TimeDependentComponent::getClassTypeId());
	sa.setInterest(SoSearchAction::ALL);
	sa.apply(this);
	SoPathList ucp;
	ucp = sa.getPaths();
	//printf(" %d time dependent components found \n", ucp.getLength());
	int i;
	for(i=0;i<ucp.getLength();i++)
	{
	    ((TimeDependentComponent *)
	     ucp[i]->getTail())->initializeComponent(this);
	    updateComponents.append(ucp[i]->getTail());
	}
#ifdef 0
	//find and add repSwitches
	sa.setType(RepSwitch::getClassTypeId());
	sa.setInterest(SoSearchAction::ALL);
	sa.apply(this);
	ucp = sa.getPaths();
	for(i=0;i<ucp.getLength();i++){
	    repList.append(ucp[i]->getTail());
	}
#endif
	changeRep(RepSwitch::REP_REVERT);
	switchNode = (SoSwitch *)findFirstInstanceOf(SoSwitch::getClassTypeId());
//	if(switchNode)
//	switchNode->whichChild.setValue(defaultSwitchValue.getValue());
	
	// hardts 3/9/95
	if(resetCBName.getValue() != SbString("none"))
	{
	  resetCB = getCreatureCBByName(resetCBName.getValue());
	  if (resetCB)
	    printf("installing BaseCreature reset callback %s \n", 
		   resetCBName.getValue().getString());
	  else
	    printf("Warning: BaseCreature could not find callback %s \n", 
		   resetCBName.getValue().getString());
	}
	if(updateCBName.getValue() != SbString("none"))
	{
	  updateCB = getCreatureCBByName(updateCBName.getValue());
	  if (updateCB)
	    printf("installing BaseCreature update callback %s \n", 
		   updateCBName.getValue().getString());
	  else
	    printf("Warning: BaseCreature could not find callback %s \n", 
		   updateCBName.getValue().getString());
	}
	if(isGrabbedCBName.getValue() != SbString("none")){
	    isGrabbedCB = getCreatureCBByName(isGrabbedCBName.getValue());
	    if (isGrabbedCB)
		printf("installing BaseCreature isGrabbed callback %s \n", 
		   isGrabbedCBName.getValue().getString());
	    else
		printf("Warning: BaseCreature could not find callback %s \n", 
		   isGrabbedCBName.getValue().getString());
	}
	if(registryNumber.getValue() != -1){
	    printf("WARNING REGISTRY NUMBER IS BEING RESET\n");
	    registryNumber.setValue(-1);
	}
	
	// kbrussel 6/9/95
	if(geometryFileName.getValue() != SbString("none")){
	    printf("WARNING: geometry file name is being reset\n");
	}
	geometryFileName.setValue(in->getCurFileName());
	
	// Sensor for isGrabbed
	SoFieldSensor *sens = new SoFieldSensor(isGrabbedSensorCB,this);
	sens->attach(&isGrabbed);
	
	return 1;
    } 
    else {
	return 0 ;
    }
}

SoNode *BaseCreature::findFirstInstanceOf(SoType nodeType)
{
    SoSearchAction sa;
    sa.setType(nodeType);
    sa.setInterest(SoSearchAction::FIRST);
    sa.apply(this);
    SoPath *p = sa.getPath();
    return (p)?p->getTail():NULL;
}

ArticulatedBody *BaseCreature::getArticulatedBody(JointModel *j)
{
    SoSearchAction sa;

    sa.setNode(j);
    sa.apply(this);
    SoPath *p = sa.getPath();
    while (!(p->getTail()->isOfType(ArticulatedBody::getClassTypeId()))) 
	p->pop();
    return (ArticulatedBody *)p->getTail();
}

SbBool BaseCreature::bodyInPath(SoPath *path)
{
    return path->containsNode(this);
}

SoPath *BaseCreature::getBodyPath(SoNode *root)
{
    if(bodyPath)
	return bodyPath;
    else{
      SoSearchAction *sa= new SoSearchAction();
    
      sa->setNode(this);
      sa->apply(root);
      bodyPath = sa->getPath();
      bodyPath->ref();
      return bodyPath;
    }
}

SoPath *BaseCreature::getPathToName(SoNode *from, const SbName &name)
{
    SoNode *namedNode = getByName(name);
    SoPath *pathToName = NULL;
    
    if(namedNode){
	SoSearchAction sa= SoSearchAction();
	sa.setNode(namedNode);
	sa.apply(from);
	pathToName = sa.getPath();
	if(pathToName)
	    pathToName->ref();
	else printf(" named node not found -> %s\n", name.getString());
    }
    return pathToName;
}
/*
void BaseCreature::changeRep(int newRep)
{
    static ChangeRepAction *cra = NULL;
    if(!cra)
	cra = new ChangeRepAction();
    // This is more expensive than just iterating through list so
    // do not use it when getting the sensing rep (or reverting)
    if((newRep != RepSwitch::REP_SENSING)&&
	(newRep != RepSwitch::REP_SENSING_OO)&&
	(newRep != RepSwitch::REP_REVERT)){
	cra->setNewRep(newRep);
	cra->apply(this);
    }
    else {
	int i;
	for(i=0;i<repList.getLength();i++){
	    ((RepSwitch *)repList[i])->setCurrentRepImmediate(newRep);
	}
    }
}
*/

void BaseCreature::changeRep(int newRep)
{
    static ChangeRepAction *cra = NULL;
    if(!cra)
	cra = new ChangeRepAction();
    cra->setNewRep(newRep);
    cra->apply(this);
}

void BaseCreature::updateRepsAt(UpdateStateData &usd)
{
    int i;
    static SoSearchAction *repSearch = NULL;
    if(!repSearch){
	repSearch = new SoSearchAction;
	repSearch->setType(RepContainer::getClassTypeId());
	repSearch->setInterest(SoSearchAction::FIRST);
    }
    repSearch->apply(this);
    SoPath *p = repSearch->getPath();
    if(p)
	((RepContainer *)p->getTail())->updateStateAt(this, usd);
/*
    else {
	printf("no repcontainer found \n");
    }

    for(i=0;i<repList.getLength();i++){
	((RepSwitch *)repList[i])->updateStateAt(this, usd);
    }
*/
}

void BaseCreature::setIsGrabbed(SbBool soIsIt)
{
    isGrabbed.setValue(soIsIt);
}

void BaseCreature::setIsGrabbed(SbBool soIsIt, BaseCreature *byWhom)
{
    grabbedBy = (soIsIt == TRUE)? byWhom : NULL;
    setIsGrabbed( soIsIt);
}

SoSeparator *BaseCreature::getBody()
{
    return this;
}

SoTransform *BaseCreature::getBodyTransform()
{
    return (SoTransform *)this->getChild(0);
}

void BaseCreature::calcBoundingBox(SbBox3f &wbox)
{
    SbViewportRegion vr;
    if(!bodyPath)
	bodyPath = getBodyPath(myCache->sceneGraph);
    bba->apply(bodyPath);
    wbox = bba->getBoundingBox();
}
#ifdef 0
void BaseCreature::calcBoundingBox(SbBox3f &wbox)
{
    SbViewportRegion vr;
    static SoGetMatrixAction ma(vr);
    SbMatrix m;
    bba->apply(this);
    SbXfBox3f xwbox = bba->getXfBoundingBox();
    ma.apply(getBodyTransform());
    m = ma.getMatrix();
    xwbox.setTransform(m);
    wbox = xwbox.project();
}
#endif
void BaseCreature::calcCenter(SbVec3f &center)
{
    SbBox3f wbox;
    calcBoundingBox(wbox);
    center = wbox.getCenter();
}

int BaseCreature::testRay(SbLine &theRay, SbVec3f &thePt)
{
    SbVec3f lu = theRay.getPosition();
    SbVec3f ld = theRay.getDirection();
    
    /* get bounding box */
    SbBox3f wbox;
    calcBoundingBox(wbox);
    SbVec3f min = wbox.getMin();
    SbVec3f max = wbox.getMax();
    
    /* test for intersection */
    float far;
    float t = rayBoxIntersection(lu,ld,min,max,&far);
    if(t>=0){
	thePt = lu + ld*t;
    }
    return (t>0.0)?1:-1;
}

void BaseCreature::vecToBody(SbVec3f &pt)
{
  SbViewportRegion vpr;
  SoGetMatrixAction mtxAction(vpr);
  mtxAction.apply(getBodyTransform());
  SbMatrix m = mtxAction.getMatrix();
  m.inverse().multVecMatrix(pt, pt);
  
}

void BaseCreature::vecToWorld(SbVec3f &pt)
{
  SbViewportRegion vpr;
  SoGetMatrixAction mtxAction(vpr);
  mtxAction.apply(getBodyTransform());
  SbMatrix m = mtxAction.getMatrix();
  m.multVecMatrix(pt, pt);
}

void BaseCreature::dirToBody(SbVec3f &pt)
{
  SbViewportRegion vpr;
  SoGetMatrixAction mtxAction(vpr);
  mtxAction.apply(getBodyTransform());
  SbMatrix m = mtxAction.getMatrix();
  m.inverse().multDirMatrix(pt, pt);
  
}

void BaseCreature::dirToWorld(SbVec3f &pt)
{
  SbViewportRegion vpr;
  SoGetMatrixAction mtxAction(vpr);
  mtxAction.apply(getBodyTransform());
  SbMatrix m = mtxAction.getMatrix();
  m.multDirMatrix(pt, pt);
}

void BaseCreature::calcBearingAndRange(SbVec3f pt, float &bearing, float &range)
{
    SbVec3f localPt = pt;
    vecToBody(localPt);
    /* calc range */
    range = localPt.length();
    
    /* calc bearing */
    bearing = atan2(localPt[0], -localPt[2]);
    if(bearing>= -(M_PI/2.0))
	bearing -= M_PI/2.0;
    else 
	bearing += M_PI*1.5;
}

void BaseCreature::setPosition(SbVec3f &np)
{
    SbVec3f cc;
    getBodyTransform()->translation.setValue(np);
}

void BaseCreature::getPosition(SbVec3f &np)
{
    np = getBodyTransform()->translation.getValue();
}

const SbVec3f &BaseCreature::position()
{
    return getBodyTransform()->translation.getValue();
}

void BaseCreature::setHeading(float t)
{
    heading = t;
    if(heading> M_PI*2)
	heading -= M_PI*2;
    else if(heading<0)
	heading += M_PI*2;
    setHeading(SbRotation(SbVec3f(0,1,0),t));
}

void BaseCreature::setHeading(SbRotation rot)
{
    getBodyTransform()->rotation.setValue(rot);
}

float BaseCreature::getHeading()
{
    float theta;
    SbVec3f axis;
    return getHeading(axis, theta);
}

float BaseCreature::getHeading(SbVec3f &axis, float &ang)
{
    SbRotation rot = getBodyTransform()->rotation.getValue();
    rot.getValue(axis, ang);
    if (axis[1] < 0.0) {
      ang = 2*M_PI - ang;
      //ang = -ang;
      axis = SbVec3f(0, 1, 0);
    }	

    return ang;
}

void BaseCreature::getHeading(SbVec3f &vec)
{
    SbRotation rot = getBodyTransform()->rotation.getValue();
    rot.multVec(SbVec3f(1,0,0), vec);
    vec.normalize();
}

TimeDependentComponent *BaseCreature::getFirstTDCOfType(SoType atype)
{
   int udc = updateComponents.getLength();
   int i;
   for(i=0;i<udc;i++){
	if(((TimeDependentComponent *)updateComponents[i])->isOfType(atype))
	return (TimeDependentComponent *)updateComponents[i] ;
   }
   return NULL; 
}

TimeDependentComponent *BaseCreature::getFirstTDCWithName(SbName &aName)
{
   int udc = updateComponents.getLength();
   int i;
   for(i=0;i<udc;i++){
	if(((SoNode *)updateComponents[i])->getName()== aName)
	    return (TimeDependentComponent *)updateComponents[i] ;
   }
   return NULL; 
}

void BaseCreature::updateStateAt(UpdateStateData &usd)
{
  for (int i = 0; i < updateComponents.getLength(); i++)
    ((TimeDependentComponent *)updateComponents[i])->updateStateAt(usd);
  
  // Call registered update callback if it exists.
  if (updateCB)
    updateCB(this);
}

int BaseCreature::stepTo(UpdateStateData &)
{
/*
    SbBox3f wbox;
    SbVec3f mx, mn;
	
    calcBoundingBox(wbox);
    mx = wbox.getMax();
    mn = wbox.getMin();
    printf("%f %f %f %f %f %f \n",mn[0], mn[1], mn[2], mx[0], mx[1], mx[2]);    
*/
    return 1;
}

void BaseCreature::restart()
{
    if(resetCB != NULL)
	resetCB(this);
}

// kbrussel 8/2/95
bool_t
BaseCreature::syncToOrFromStream(XDR *xdrs)
{
  SbMatrix xform;

  if (xdrs->x_op == XDR_ENCODE)
    {
      SoGetMatrixAction gma(SbViewportRegion());
      gma.apply(getBodyTransform());
      xform = gma.getMatrix();
    }

  if (!xdr_SbMatrix(xdrs, &xform))
    {
      return FALSE;
    }

  if (xdrs->x_op == XDR_DECODE)
    {
      getBodyTransform()->setMatrix(xform);
    }
  
  return TRUE;
}

// hardts 3/9/95
SbPList BaseCreature::cbList;

creatureCB BaseCreature::getCreatureCBByName(const SbString &fname)
{
  for(int i = 0; i < cbList.getLength(); i++)
    if (((CreatureCBData *)cbList[i])->getFname() == fname)
      return ((CreatureCBData *)cbList[i])->getCB();
  return NULL;
}


void BaseCreature::isGrabbedSensorCB(void *data,SoSensor *)
{
  BaseCreature *creature = (BaseCreature *)data;
  if (creature->isGrabbedCB) 
    creature->isGrabbedCB(creature);
}
