#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCylinder.h>
#include <Inventor/SoPath.h>
#include <Inventor/draggers/SoDragger.h>

#include "Translate222Manip.h"
#include "ControlPoint.h"

int ControlPoint::translate_index  = 0;

/*********************************************************************
 * 
 *	    ControlPoint Class
 * 
 *********************************************************************/

SO_NODE_SOURCE(ControlPoint);

void
ControlPoint::initClass()
{
    SO_NODE_INIT_CLASS(ControlPoint, SoSeparator, "Separator");    
}

ControlPoint::ControlPoint() : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(ControlPoint);
    changeField = (SoMFVec3f *)NULL;

   myManip = (Translate222Manip *) NULL;
}

ControlPoint::ControlPoint(SbVec3f pos, int i, SoMFVec3f *cf) : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(ControlPoint);

    addChild(new SoTransform);

    SoMaterial *sphereMat = new SoMaterial;
    sphereMat->diffuseColor.setValue(.5, .8, .5);
    addChild(sphereMat);

    SoSphere *sphere = new SoSphere;
    sphere->radius.setValue(.5);
    addChild(sphere);

    getTransform()->translation = pos;
    index = i;
    changeField = cf;

   myManip = (Translate222Manip *) NULL;
}

ControlPoint::~ControlPoint()
{
}

ControlPoint *
ControlPoint::isSelection(SoPath *p)
{
   if (p->getNodeFromTail(1)->isOfType(ControlPoint::getClassTypeId())) {
	p->pop();
	return (ControlPoint *)p->getTail();
   } else {
	return (ControlPoint *) NULL;
   }
}

void
ControlPoint::addManipulator(SoPath *p)
{
   if (myManip != NULL)
	return;
	
   myManip = new Translate222Manip;
   myManip->ref();

   p->push(translate_index);
   manipPath = p;

   myManip->replaceNode(p);

   SoDragger *myDragger;
   myDragger = myManip->getDragger();
   myDragger->addMotionCallback(&ControlPoint::dragMotionCallback, this);
}

void
ControlPoint::removeManipulator()
{
   if (myManip == NULL)
	return;

   myManip->replaceManip(manipPath,NULL);
   myManip = (Translate222Manip *) NULL;
}

void
ControlPoint::dragMotionCallback(void *data, SoDragger *)
{
   ControlPoint *t = (ControlPoint *)data;

   float x, y, z;
   SbVec3f vec;

   vec = t->getTransform()->translation.getValue();
   vec.getValue(x, y, z);

   if (t->changeField) 
        t->changeField->set1Value(t->index, x, y, z);
}
