#include <iostream.h>   
#include <Inventor/Xt/viewers/SoXtWalkViewer.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoComplexity.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoMatrixTransform.h>
#include <Inventor/nodes/SoOrthographicCamera.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoOutput.h>
#include <Inventor/sensors/SoSensor.h>
#include <Inventor/SbPList.h>
#include <Inventor/SoPath.h>
#include <Inventor/nodes/SoSelection.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoLightModel.h>
#include "CreatureCache.h"
#include "MotorCreature.h"
#include "AliveDispatch.h"
#include "TimeStat.h"

void
CreatureConstraint::updateStateAt(UpdateStateData &)
{
}

//
//
//

void tCall(void *ud, SoAction *action)
{
    if(action->isOfType(SoGLRenderAction::getClassTypeId())){
	printf("render you sucker\n");
    }
}
    
CreatureCache::CreatureCache(SoXtViewer *v)
{
	viewer = v;
	sceneGraph = new SoSeparator;
	sceneGraph->ref();
//	SoCallback * precallback = new SoCallback;
//	precallback->setCallback(tCall,this);
//	sceneGraph->addChild(precallback);
	
	ticks = .1;
	wallClock = 0;
	creatureCount = 0;
	cameraCreature = NULL;
	isDistributed = FALSE;
	iAmServer = FALSE; // what do we want to do here?
	distributedCatchUpFlag = FALSE;
}

SoSeparator *CreatureCache::getSceneGraph()
{
    return sceneGraph;
}

void CreatureCache::attachSceneGraph(SoSeparator *root)
{
    root->addChild(sceneGraph);
}

void CreatureCache::detachSceneGraph(SoSeparator *root)
{
    SoSearchAction sa;
    SoPath *path = NULL;
    sa.setNode(sceneGraph);
    sa.apply(root);
    if((path = sa.getPath()) != NULL){
	path->pop();
	SoSeparator *parentNode = (SoSeparator *)path->getTail();
	parentNode->removeChild(sceneGraph);
    }
}

void CreatureCache::addBaseCreature(BaseCreature *c)
{
  if(findBaseCreature(c)<0){
    creatures.append(c);

    if (isDistributed && (iAmServer == 0))
      {
	// This if is here because we don't need to
	// register this creature if we are a client and
	// downloaded it from the server; in that case,
	// the creature already has a registry number
	// (set by AliveDispatch).

	if (c->registryNumber.getValue() == -1)
	  AliveDispatch::registerCreatureWithServer(c); 
	creatureCount++;  /* note that in the distributed client case
			     creatureCount != registryNumber.
			     we get registryNumber from the server's
			     creatureCount. */
	/* actually the above should not matter.
	   all references to creatureCount in this file
	   should really be to creatures.getLength(). */
      }
    else
      {
	creatureCount++;
	c->registryNumber.setValue(creatureCount);
	if (isDistributed)  // then we are a server
	  AliveDispatch::broadcastCreatureAnnouncement(c);
      }
    sceneGraph->addChild(c);
  }
  c->myCache = this;	
  return;
}

void CreatureCache::addBaseCreatureToFront(BaseCreature *c)
{
    if(findBaseCreature(c)<0){
	creatures.append(c);
	creatureCount++;
	SoNode *camInGraph = c->findFirstInstanceOf(SoCamera::getClassTypeId());
	if(camInGraph)
	    cameraCreature = c;
	sceneGraph->insertChild(c, 0);
    }
    c->myCache = this;	
    return;
}

BaseCreature *CreatureCache::addBaseCreatureFromFile(char *fn)
{
    BaseCreature *newCreature = BaseCreature::loadCreature(fn);
    if(newCreature != NULL)
	addBaseCreature(newCreature);
    return newCreature;
}

BaseCreature *
CreatureCache::addRegisteredBaseCreatureFromFile(char *fn,
						 int32_t registry_number)
{
    BaseCreature *newCreature = BaseCreature::loadCreature(fn);
    if(newCreature != NULL)
      {
	newCreature->registryNumber.setValue(registry_number);
	addBaseCreature(newCreature);
      }
    return newCreature;
}

BaseCreature *CreatureCache::addBaseCreatureToFrontFromFile(char *fn)
{
    BaseCreature *newCreature = BaseCreature::loadCreature(fn);
    if(newCreature != NULL)
	addBaseCreatureToFront(newCreature);
    return newCreature;
}

void CreatureCache::restart()
{
    int i;
    for(i=0;i<creatureCount;i++){
	((BaseCreature *)creatures[i])->restart();
    }
}

int CreatureCache::findBaseCreature(BaseCreature *ng)
{
    int i;
    for(i=0;i<creatureCount;i++){
	if(creatures[i]==ng)
	    return i;
    }
    return -1;
}

BaseCreature *CreatureCache::findBaseCreatureWithTag(int tag)
{
    int i;
    for(i=0;i<creatureCount;i++){
	if(((BaseCreature *)creatures[i])->tag.getValue() == tag)
	    break;
    }
    return (i<creatureCount)?((BaseCreature *)creatures[i]) : NULL;
}

BaseCreature *&
CreatureCache::getCreatureFromRegistryNumber(int32_t registry_number)
{
  static int counter;
  static BaseCreature *nullptr = NULL;

  for (counter = 0; counter < creatures.getLength(); counter++)
    {
      BaseCreature *bc = (BaseCreature *) creatures[counter];
      if (bc->registryNumber.getValue() == registry_number)
	return bc;
    }
  return nullptr;
}

void CreatureCache::removeBaseCreature(BaseCreature *ng)
{
    int j = findBaseCreature(ng);
    if(j != -1){
	sceneGraph->removeChild(ng );
	creatures.remove(j);
	creatureCount--;
    }
    return;
}

void CreatureCache::updateStateAt(float delta_time = -1.0) 
{
    int i;
    
//static Timestat *cc = new Timestat("CreatureCache update time", 20);
//static Timestat *u = new Timestat("updateStatAt time", 20);
//static Timestat *st = new Timestat("StepTo update time", 20);
//static Timestat *c = new Timestat("Constraints time", 20);   
//cc->startTime();
    if (!(isDistributed == TRUE &&
	  distributedCatchUpFlag == TRUE))   /* only increment wall clock if not catching up */
      {
	if (delta_time < 0.0) 
	  {
	    ticks +=1;
	    wallClock = ticks;
	  }
	else 
	  {
	    ticks = delta_time;
	    wallClock += ticks;
	  }
      }
//    printf("creature cache update \n");
    UpdateStateData usd;
    usd.time = ticks;
    usd.currSceneGraph = sceneGraph;
    usd.currViewer = viewer;
    usd.currCreatureCache = this;
//u->startTime();   
    // do the main AliveDispatch routine if distributed
    if (isDistributed == TRUE)
      AliveDispatch::mainDispatch(usd);
    for(i=0;i<creatureCount;i++)
      {
	if (!((isDistributed == TRUE) &&
	      ((BaseCreature *) creatures[i])->isOfType(MotorCreature::getClassTypeId()) &&
	      ((MotorCreature *) creatures[i])->motorCreatureType.getValue() == MotorCreature::MASTER &&
	      distributedCatchUpFlag == TRUE))
	  ((BaseCreature *)creatures[i])->updateStateAt(usd);
      }

//      if (!((isDistributed == TRUE) &&
//	    ((BaseCreature *) creatures[i])->isOfType(MotorCreature::getClassTypeId()) &&
//	    ((MotorCreature *) creatures[i])->motorCreatureType.getValue() == MotorCreature::DROID))
//	((BaseCreature *)creatures[i])->updateStateAt(usd);

//u->endTime();
//st->startTime();	
    for(i=0;i<creatureCount;i++)
      {
	if (!((isDistributed == TRUE) &&
	      ((BaseCreature *) creatures[i])->isOfType(MotorCreature::getClassTypeId()) &&
	      ((MotorCreature *) creatures[i])->motorCreatureType.getValue() == MotorCreature::MASTER &&
	      ((BaseCreature *) creatures[i])->myCache->distributedCatchUpFlag == TRUE))
	  ((BaseCreature *)creatures[i])->stepTo(usd);
      }

//      if (!((isDistributed == TRUE) &&
//	    ((BaseCreature *) creatures[i])->isOfType(MotorCreature::getClassTypeId()) &&
//	    ((MotorCreature *) creatures[i])->motorCreatureType.getValue() == MotorCreature::DROID))
//	((BaseCreature *)creatures[i])->stepTo(usd);



//st->endTime();
//c->startTime();
    int num_constraints = constraints.getLength();
    for(i=0; i<num_constraints; i++)
	((CreatureConstraint *)constraints[i])->updateStateAt(usd);
//c->endTime();
//cc->endTime();
}
