#ifndef CREATURECACHE
#define CREATURECACHE

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/manips/SoHandleBoxManip.h>
#include <Inventor/nodes/SoSelection.h>
#include "BaseCreature.h"

#include <Inventor/SbPList.h>
 
class SoCamera;
class BaseCreature;
class SoTransform;
class SoNode;
class SoXtViewer;

class CreatureConstraint {
  public:
    virtual void updateStateAt(UpdateStateData &usd);
};

class CreatureCache {
    public:
    SoSeparator *sceneGraph;
    SoXtViewer  *viewer;
    

    SbPList creatures;
    int creatureCount;
    float ticks;
    SbBool isDistributed;
    SbBool iAmServer;
    SbBool distributedCatchUpFlag;

    SbPList constraints;
    void addConstraint(CreatureConstraint *c) {constraints.append(c);};
    void removeConstraint(CreatureConstraint *c) {constraints.remove(constraints.find(c));};

    CreatureCache(SoXtViewer *v);
    void attachSceneGraph(SoSeparator *root);
    void detachSceneGraph(SoSeparator *root);
    SoSeparator *getSceneGraph();
    int findBaseCreature(BaseCreature *ng);
    BaseCreature *findBaseCreatureWithTag(int tag);
    void addBaseCreature(BaseCreature *ng);
    void addBaseCreatureToFront(BaseCreature *ng);
    BaseCreature *addBaseCreatureFromFile(char *fn);
    BaseCreature *addBaseCreatureToFrontFromFile(char *fn);
    void removeBaseCreature(BaseCreature *ng);
    void updateStateAt(float delta_time);
    BaseCreature *getCameraCreature() {return cameraCreature;}
    void restart();

    // new methods added by kbrussel for Distributed Alive
    BaseCreature *addRegisteredBaseCreatureFromFile(char *fn, int32_t registry_number);
    BaseCreature *&getCreatureFromRegistryNumber(int32_t registry_number);
    float wallClock;

    protected:
    BaseCreature *cameraCreature;

};
#endif


    
