#include <math.h>
#include <stdio.h>
#include "../nexttypes.h"
#include "../Creature.h"
#include "Integrator.h"

float maxReasonable = 1e4;
#define INVERR -2.0

Integrator::Integrator()
{
  nVar = 0;
}

float Integrator::dispatchDeriv(float time, vector y, vector ydot)
{
    derivStruct ytemp;
    Creature *creature = (Creature *)client;
    
    tTime = time;
    ytemp.time = time;
    ytemp.y = y;
    ytemp.ydot = ydot;
    
    creature->derivMethod(&ytemp);
    return dTime;
}

int goingSouth(int n,vector k1,float maxReasonable)
{
    int i;
    for(i= 0;i<n;i++)
	if(fabs(k1[i])>maxReasonable)
	    return YES;
    return NO;
}
	
int Integrator::rkf(vector yin,int n,vector yout,
    float tMin,float tMax,float eps,float hMin,
    float hMax)
{
    vector w = new_vector(0,n);
    vector wn = new_vector(0,n);
    vector k1 = new_vector(0,n);
    vector k2 = new_vector(0,n);
    vector k3 = new_vector(0,n);
    vector k4 = new_vector(0,n);
    vector k5 = new_vector(0,n);
    vector k6 = new_vector(0,n);
    vector r = new_vector(0,n);
    
    int i,result = 0,flag = YES;
    int cycleCount = 0;
    
    float t,h,delta,rmax;
    
    /* do some initialization */
    t = tMin;
    h = hMax;
    if(t>=tMax){
	flag = NO;
	vectorAssign(n,yin,yout);
	result = 0;
    }
    else
	vectorAssign(n,yin,w);
    
    while(flag){
	cycleCount++;
	/* k1 */
	dispatchDeriv(t,w,k1);
	vectorScalarMult(n,k1,h,k1);
	
	/*k2 */
	for(i=0;i<n;i++)
	    wn[i] = w[i] + (1.0/4.) *k1[i];
	dispatchDeriv(t+.25*h,wn,k2);
	if(goingSouth(n,k2,maxReasonable)){
	    h = .0625*h;
	    printf("going south 1\n");
	    goto wrapup;
	}
	vectorScalarMult(n,k2,h,k2);
	
	/*k3 */
	for(i=0;i<n;i++)
	    wn[i] = w[i] + (3.0/32.) *k1[i] + (9./32.)*k2[i];
	dispatchDeriv(t+(3./8.)*h,wn,k3);
	if(goingSouth(n,k3,maxReasonable)){
	    h = .1*h;
	    printf("going south 2\n");
	    goto wrapup;
	}
	vectorScalarMult(n,k3,h,k3);
	
	/*k4 */
	for(i=0;i<n;i++)
	    wn[i] = w[i] + (1932./2197.) *k1[i] - (7200./2197.)*k2[i] + 
		    (7296./2197.)*k3[i];
	dispatchDeriv(t+(12./13.)*h,wn,k4);
	if(goingSouth(n,k4,maxReasonable)){
	    h = (3/8.0)*h;
	    printf("going south 3\n");
	    goto wrapup;
	}
	vectorScalarMult(n,k4,h,k4);
	
	/*k5 */
	for(i=0;i<n;i++)
	    wn[i] = w[i] + (439./216.) *k1[i] - (8.)*k2[i] + 
		    (3680./513.)*k3[i] -(845./4104.)*k4[i];
	dispatchDeriv(t+h,wn,k5);
	if(goingSouth(n,k5,maxReasonable)){
	    h = .5*h;
	    printf("going south 4\n");
	    goto wrapup;
	}
	vectorScalarMult(n,k5,h,k5);
	
	/*k6 */
	for(i=0;i<n;i++)
	    wn[i] = w[i] - (8./27.) *k1[i] + (2.)*k2[i] - 
		    (3544./2565.)*k3[i] +(1859./4104.)*k4[i] -(11./40.)*k5[i];
	dispatchDeriv(t+.5*h,wn,k6);
	if(goingSouth(n,k2,maxReasonable)){
	    h = .5*h;
	    printf("going south 5\n");
	    goto wrapup;
	}
	vectorScalarMult(n,k6,h,k6);
	
	/* r */
	for(i=0;i<n;i++)
	    r[i] = fabs((1./360.) *k1[i] - (128./4275.)*k3[i] - 
	    (2197./75240.)*k4[i] +(1./50.)*k5[i] +(2./55.)*k6[i])/h;
	
	/* find max r */
	rmax = r[0];
	for(i=1;i<n;i++)
	    rmax = MAX(rmax,r[i]);
	delta = 0.84*pow(eps/rmax,.25);
		
	/* if rmax is less than allowable error accept results */
	if((rmax<=eps)){
	    t = t+h;
	    for(i=0;i<n;i++)
		w[i] = w[i] + (25./216.)*k1[i] + (1408./2565.)*k3[i] +
		(2197./4104.)*k4[i]-(1./5.)*k5[i];
	    goto wrapup;
	}
	
        /* calculate new h */
	    if(delta<=0.1)
		h = .1*h;
	    else 
		h = (delta>=4)?4*h:delta*h;
	    h = MIN(h,hMax);

	/* now wrap up */
wrapup:	
	//printf("t -> %f h -> %f tMax -> %f \n",t,h,tMax);
	if(t>=tMax){
	    vectorAssign(n,w,yout);
	    result = 0;
	    flag = NO;
	}
	else{
	    if(t+h>tMax)
		h = tMax-t;
	    else if(h<hMin){
		flag = NO;
		result = 1;
		printf("hmin exceeded\n");
	    }
	}
    }
    //printf("cycle count => %d \n",cycleCount);
    free_vector(w,0,n);
    free_vector(wn,0,n);
    free_vector(k1,0,n);
    free_vector(k2,0,n);
    free_vector(k3,0,n);
    free_vector(k4,0,n);
    free_vector(k5,0,n);
    free_vector(k6,0,n);
    free_vector(r,0,n);
    return result;
}


void Integrator::initForNumVariables(int n ,float minS ,float maxS,float me , void *aClient)
    
{
    minStepsize = minS;
    maxStepsize =maxS;
    maxError = me;
    client = aClient;
    restart = YES;
    okToOverShoot=YES;
    globalTimeStamp = 0;
    globalTime = 0.0;
    time = 0.0;
    if(n>nVar){
	if( nVar !=0){
	    free_vector(hist.wvalues,0,nVar);
	    free_vector(hist.fvalues,0,nVar);
	}
      	hist.wvalues = new_vector(0,n);
	hist.fvalues = new_vector(0,n);
    }
    nVar = n;
    
    return;
}
	
void Integrator::setClient(void *anObject)
{
    client = anObject;
    restart = YES;
}

void Integrator::setMinStepsize(float s)
{
    minStepsize = s;
    if(stepsize<minStepsize)
	restart = YES;
}

void Integrator::setMaxStepsize(float s)
{
    maxStepsize = s;
    if(stepsize>maxStepsize)
	restart =YES;
}

void Integrator::setMaxError(float e)
{
    maxError = e;
    restart =YES;
}


void Integrator::setTime(float t)
{
    time = t;
}

void Integrator::setOkToOverShoot(int yn)
{
    okToOverShoot = yn;
}

void Integrator::restartIntegrator()
{
    restart =YES;
    initializeIntegrator();
}

void Integrator::initializeIntegrator()
{
    Creature *c = (Creature *)client;
    restart  =YES;
    c->setInitialValues(hist.wvalues,(void *)this);
    return ;
}

float Integrator::stepToAndOk(float t ,int *yn)
{
   int result;
//printf("in stepToAndOk\n");
   Creature *c=(Creature *)client;
   globalTimeStamp++;
   globalTime = t;
   if(restart)
	initializeIntegrator();
    result = rkf(hist.wvalues,nVar,hist.wvalues, 
    time,t,maxError,minStepsize,maxStepsize);
    *yn=YES;
    restart = NO;
    c->stepAtWithVector(t,hist.wvalues);
    time = t;
    return time;
}
    
