#include "InventorXDRRoutines.h"
#include <Inventor/SbLinear.h>

bool_t
xdr_SbVec3f(XDR *xdrs, SbVec3f *vec)
{
  static float x, y, z;

  if (xdrs->x_op == XDR_ENCODE)
    {
      vec->getValue(x, y, z);
    }

  if (xdrs->x_op != XDR_FREE)
    {
      if (!xdr_float(xdrs, &x))
	return FALSE;
      if (!xdr_float(xdrs, &y))
	return FALSE;
      if (!xdr_float(xdrs, &z))
	return FALSE;
    }

  if (xdrs->x_op == XDR_DECODE)
    {
      vec->setValue(x, y, z);
    }

  return TRUE;
}

bool_t
xdr_SbVec4f(XDR *xdrs, SbVec4f *vec)
{
  static float x, y, z, w;

  if (xdrs->x_op == XDR_ENCODE)
    {
      vec->getValue(x, y, z, w);
    }

  if (xdrs->x_op != XDR_FREE)
    {
      if (!xdr_float(xdrs, &x))
	return FALSE;
      if (!xdr_float(xdrs, &y))
	return FALSE;
      if (!xdr_float(xdrs, &z))
	return FALSE;
      if (!xdr_float(xdrs, &w))
	return FALSE;
    }

  if (xdrs->x_op == XDR_DECODE)
    {
      vec->setValue(x, y, z, w);
    }

  return TRUE;
}

bool_t
xdr_SbMatrix(XDR *xdrs, SbMatrix *matrix)
{
  static SbMat mat;

  if (xdrs->x_op == XDR_ENCODE)
    {
      matrix->getValue(mat);
    }
  
  if (xdrs->x_op != XDR_FREE)
    {
      int c1, c2;
      
      for (c1 = 0; c1 < 4; c1++)
	{
	  for (c2 = 0; c2 < 4; c2++)
	    {
	      if (!xdr_float(xdrs, &(mat[c1][c2])))
		return FALSE;
	    }
	}
    }

  if (xdrs->x_op == XDR_DECODE)
    {
      matrix->setValue(mat);
    }
  
  return TRUE;
}
