#include "Kalman.h"
#include <stdio.h>

Kalman::Kalman(float ddt, Vector e, Vector n)
{
    /* build up the matrices */
    x = Vector(9);
    xstar = Vector(9);
    y = Vector(3);
    F = Matrix(9, 9);
    H = Matrix(3, 9);
    K = Matrix(9, 3);
    P = Matrix(9, 9);
    Q = Matrix(9, 9);
    P = Matrix(9, 9);
    Pstar = Matrix(9, 9);
    R = Matrix(3, 3);
    Phi = Matrix(9, 9);
    dt = ddt;
    initializeMatrices(e, n);
    numEstimates = 0;
}

void Kalman::initializeMatrices(Vector e, Vector n)
{
    int i;
    /* F */
    F.mm[0][1] = 1 ;F.mm[0][2] = dt/2.0;
    F.mm[1][2] = 1;
    F.mm[3][4] = 1 ;F.mm[3][5] = dt/2.0;
    F.mm[4][5] = 1;
    F.mm[6][7] = 1 ;F.mm[6][8] = dt/2.0;
    F.mm[7][8] = 1;
    
    /* H */
    H.mm[0][0] = 1; H.mm[0][1] = dt; H.mm[0][2] = dt*dt/2.0;
    H.mm[1][3] = 1; H.mm[1][4] = dt; H.mm[1][5] = dt*dt/2.0;
    H.mm[2][6] = 1; H.mm[2][7] = dt; H.mm[2][8] = dt*dt/2.0;
    
    /* Phi */
    Phi.identity();
    Phi = Phi +(F)*dt;
    
    /* Q */
    for(i=0;i<9;i++)
	Q.mm[i][i] = e.mm[i]*e.mm[i];
	
    /* R */
    for(i=0;i<3;i++)
	R.mm[i][i] = n.mm[i]*n.mm[i];
    
    /* P */
    P.identity();
}

void Kalman::initialState(Vector Xo)
{
    numEstimates = 0;
    x = Xo;
    prediction();
}

void Kalman::prediction()
{
    Matrix I(9, 9);
    P = (I.identity() - K*H)*Pstar;
    Pstar = (Phi *P *Phi.transpose() + Q);
   // Pstar.print("estimate covariance ");
    xstar = x + (F*dt)*x;
  //  xstar.print("prediction");   
}

void Kalman::update(float newx, float newy, float newz)
{  
    if(numEstimates<10){
	/* update K */
	Matrix pht(1, 1);
	Matrix hphti(1, 1);
	pht = (Pstar)*(H.transpose());
	hphti = (H*pht+ R);
	hphti = hphti.inverse();
	K = pht * hphti;
//	K.print("K");
    }
    
   /* state estimate */
    y.mm[0] = newx;
    y.mm[1] = newy;
    y.mm[2] = newz;
    
//    printf("delta %f %f %f\n",  y.mm[0]-xstar.mm[0], y.mm[1]-xstar.mm[3], y.mm[2]-xstar.mm[6]);
//    printf("predicted state %f %f %f\n", xstar.mm[0], xstar.mm[3], xstar.mm[6]);
    
    Vector ydiff = y - H*xstar;
//    ydiff.print("observed - predicted observation");
    x = xstar + K * (y - H*xstar);
  
   /* update prediction */
    prediction();
    
    //numEstimates++;
}

void Kalman::updateStateEstimate(SbVec3f &se)
{
    update(se[0], se[1], se[2]);
    getStateEstimate(se);
}

void Kalman::getStateEstimate(SbVec3f &se)
{
    se.setValue(xstar.mm[0], xstar.mm[3], xstar.mm[6]);
}

void Kalman::print()
{
    printf("F \n");
    F.print();
    
    printf("H \n");
    H.print();
    
    printf("Phi \n");
    Phi.print();
    
}
