#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCylinder.h>

#include "LimbModel.h"

/*********************************************************************
 * 
 *	    LimbModel Class
 * 
 *********************************************************************/

SO_NODE_SOURCE(LimbModel);

void
LimbModel::initClass()
{
    SO_NODE_INIT_CLASS(LimbModel, SoSeparator, "Separator");    
}

LimbModel::LimbModel() : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(LimbModel);
}

LimbModel::LimbModel(float l, float r) : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(LimbModel);
    buildLimb(l, r);
}

LimbModel::LimbModel(SoTransform *t, SoSeparator *s) : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(LimbModel);

    addChild(t);
    addChild(s);
}

LimbModel::~LimbModel()
{
}

SoSeparator *LimbModel::getShapeSep()
{
    if (getNumChildren() == 0) { 
       buildLimb();
    }
    return (SoSeparator *)getChild(1);
}

void LimbModel::setShapeSep(SoSeparator *s)
{
    if (getNumChildren() == 0) { 
       buildLimb();
    }
    replaceChild(1, (SoNode *)s);
}

SoTransform *LimbModel::getTransform()
{
    if (getNumChildren() == 0) { 
       buildLimb();
    }
    return (SoTransform *)getChild(0);
}

void LimbModel::setTransform(SoTransform *tx)
{
    if (getNumChildren() == 0) { 
       buildLimb();
    }
    replaceChild(0, (SoNode *)tx);
}

void LimbModel::setLengthRadius(float l, float r)
{
    buildLimb(l, r);
}

void LimbModel::buildLimb()
{
    buildLimb(1.0, 0.1);
}

void LimbModel::buildLimb(float length, float radius)
{
    removeAllChildren();
    
    //SoTransform *xt = new SoTransform;

    SoSeparator *ns = new SoSeparator;
    addChild(ns);
/*
    SoMaterial *mat = new SoMaterial;
    SoCylinder *cyl = new SoCylinder;

    xt->translation.setValue(SbVec3f(length/2, 0, 0));
    xt->rotation.setValue(SbRotation(SbVec3f(0,0, 1), M_PI/2));
    addChild(xt);

    addChild(ns);
    mat->ambientColor.setValue(1*.06, .69*.06, .0);
    mat->diffuseColor.setValue(1*.22, .69*.22, .0*.22);
    mat->specularColor.setValue(1*.71, .99*.71, .78*.71);
    mat->shininess = .16;
    ns->addChild(mat);
    
    cyl->height.setValue(length);
    cyl->radius.setValue(radius);
    ns->addChild(cyl);
*/
}
