#include "PendulumEiEo.h"

SO_ENGINE_SOURCE(PendulumEiEo);


void
PendulumEiEo::initClass()
{
   SO_ENGINE_INIT_CLASS(PendulumEiEo, SoEngine, "Engine");
}


PendulumEiEo::PendulumEiEo()
{
   SO_ENGINE_CONSTRUCTOR(PendulumEiEo);

   SO_ENGINE_ADD_INPUT(secSince, (0.0));
   SO_ENGINE_ADD_INPUT(secStart, (0.0));
   SO_ENGINE_ADD_INPUT(axis,  (0.0, 0.0, 1.0));
   SO_ENGINE_ADD_INPUT(angle0, (0.0));
   SO_ENGINE_ADD_INPUT(angle1, (0.0));
   SO_ENGINE_ADD_INPUT(period,  (1.0));
   SO_ENGINE_ADD_INPUT(ease,  (FALSE));

   SO_ENGINE_ADD_OUTPUT(output, SoSFRotation);
}

PendulumEiEo::~PendulumEiEo()
{
}

void
PendulumEiEo::evaluate()
{

   float p = period.getValue();
   float halfp = p * 0.5;
   float secs = secSince.getValue();
   float tmp = (secs - ((int)(secs / p) * p));

   if (secs < secStart.getValue())
	return;

   float ang;

   if (tmp > halfp) { 
      tmp = p - tmp;
   }
   float t = tmp / halfp;

   if (ease.getValue()) {
      float sqterm = 3.0 * t * t;
      float cuterm = 2.0 * t * t * t;
      ang = angle0.getValue() * (cuterm - sqterm + 1.0)  + 
	    angle1.getValue() * (sqterm - cuterm);
   } else {
      ang = (1.0 - t) * angle0.getValue() +  t * angle1.getValue();
   }

   SO_ENGINE_OUTPUT(output, SoSFRotation, setValue(axis.getValue(),ang));
}


