#include "RepSwitch.h"
#include "BaseCreature.h"
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/elements/SoElements.h>
#include <Inventor/elements/SoSwitchElement.h>
#include "JointModel.h"

SO_NODE_SOURCE(RepContainer);

void
RepContainer::initClass()
{
    SO_NODE_INIT_CLASS(RepContainer, SoSeparator, "Separator");
}

RepContainer::RepContainer() : SoSeparator(2)
{
    SO_NODE_CONSTRUCTOR(RepContainer);
}

RepContainer::~RepContainer()
{
}

SbBool
#ifdef SO_VERSION_REVISION
RepContainer::readInstance(SoInput *in, unsigned short flags)
#else
RepContainer::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoSeparator::readInstance(in, flags)){
#else
    if (SoSeparator::readInstance(in)){
#endif
	return (1);
    }
    return (0);
}

void RepContainer::updateStateAt(BaseCreature *bc, UpdateStateData &usd)
{
    return;
}

SoPath *RepContainer::buildPath(SoNode *startNode, SoNode *endNode)
{
    SoPath *pp;
    
    /* build path to sensorFrame */
    SoSearchAction sa;
    sa.setNode(endNode);
    sa.apply(startNode);
    pp = sa.getPath();
    pp->ref();
    if(pp==NULL)
        printf("path wasn't created \n");
//    else
//        printf("path created \n");
    return pp;
}


SO_NODE_SOURCE(RepSwitch);

RepSwitch::RepSwitch() : SoSwitch()
{
    SO_NODE_CONSTRUCTOR(RepSwitch);
    SO_NODE_ADD_FIELD(repBindings, (0));
    SO_NODE_ADD_FIELD(reps, (RepSwitch::REP_NONE));
    SO_NODE_ADD_FIELD(currentRep, (RepSwitch::REP_NONE));
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_NONE);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_ALL);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_REVERT);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_SENSING);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_SENSING_OO);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_0);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_1);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_2);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_3);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_4);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_5);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_6);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_7);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_8);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_9);
    SO_NODE_DEFINE_ENUM_VALUE(RepType, REP_10);
    SO_NODE_SET_SF_ENUM_TYPE(currentRep, RepType);
    SO_NODE_SET_SF_ENUM_TYPE(reps,  RepType);
    repSensor = new SoFieldSensor(&RepSwitch::repSensorCB, this);
    repSensor->attach( &currentRep );
    repSensor->setPriority(10);
    prevRep = RepSwitch::REP_ALL;
}

void
RepSwitch::initClass()
{
  SO_NODE_INIT_CLASS(RepSwitch, SoSwitch, "SoSwitch");
}

RepSwitch::~RepSwitch()
{
}

SbBool
#ifdef SO_VERSION_REVISION
RepSwitch::readInstance(SoInput *in, unsigned short flags)
#else
RepSwitch::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
    if (SoSwitch::readInstance(in, flags)){
#else
    if (SoSwitch::readInstance(in)){
#endif
	if(repBindings.getNum() != reps.getNum())
	    printf("WARNING DIFFERENT NUMBER OF REPS AND REP BINDINGS \n");
	// This is to force the switch. I know it is goofy
	//repSensorCB(this, NULL);
	//repSensor->unschedule();
	prevRep = currentRep.getValue();
	currentRep.setValue(RepSwitch::REP_ALL);
	repSensor->unschedule();
	whichChild.setValue(-3);
	return (1);
    }
    return (0);
}

void
RepSwitch::updateStateAt(BaseCreature *bc, UpdateStateData &usd)
{
    int whichRep = currentRep.getValue();
    if((whichRep >= RepSwitch::REP_0)){
	SoNode *child = getChild(whichChild.getValue());
	if((child != NULL) && 
	    (child->isOfType(RepContainer::getClassTypeId()))){
		((RepContainer *)child)->updateStateAt(bc, usd);
	}
    }
    return;
}
#define NOBINDINGFND -100

void RepSwitch::setCurrentRepImmediate(int i)
{
    if(((i == (int)RepSwitch::REP_REVERT)&&
	((currentRep.getValue() == RepSwitch::REP_SENSING) ||
	(currentRep.getValue() == RepSwitch::REP_SENSING_OO)))
	|| (currentRep.getValue() == RepSwitch::REP_ALL))
	i = prevRep;
    if(i != currentRep.getValue()){
	int binding = repBindingFor(i);
	if(binding != NOBINDINGFND){
	    if(currentRep.getValue() != RepSwitch::REP_ALL)
		prevRep = currentRep.getValue();
	    currentRep.setValue(i);
	    whichChild.setValue(binding);
	    repSensor->unschedule();
	}
    }
}
int RepSwitch::repBindingFor(int i)
{
    const int *theReps =reps.getValues(0);
    const int32_t *theBindings = repBindings.getValues(0);
    int j, repIndex = NOBINDINGFND;
    int numReps = reps.getNum();
    for(j=0;j<numReps;j++){
	if(theReps[j] == i){
	    repIndex = theBindings[j];
	    break;
	}
    }
    if((repIndex == NOBINDINGFND) &&
	(i== RepSwitch::REP_SENSING_OO)){
	for(j=0;j<numReps;j++){
	    if(theReps[j] == RepSwitch::REP_SENSING){
		repIndex = theBindings[j];
		break;
	    }
	}    
    }
    return repIndex; 
}

void RepSwitch::repSensorCB(void *in, SoSensor *)
{
    RepSwitch *j = (RepSwitch *)in;
    int binding = j->repBindingFor(j->currentRep.getValue());
    if((binding != NOBINDINGFND)
	&&(j->whichChild.getValue()!= binding)){
	j->whichChild.setValue(binding);
    }
}

#ifdef CRA
//
// Constructor
//
SO_ACTION_SOURCE(ChangeRepAction);

void
ChangeRepAction::initClass()
{
    // Initialize the run-time type variables
    SO_ACTION_INIT_CLASS(ChangeRepAction, SoAction);

    SO_ENABLE(ChangeRepAction, SoSwitchElement);
    
    SO_ACTION_ADD_METHOD(SoNode, nullAction);
    SO_ACTION_ADD_METHOD(SoGroup, callDoAction);
    SO_ACTION_ADD_METHOD(RepSwitch, callRepAction);
    SO_ACTION_ADD_METHOD(JointModel, callJointAction);
}

ChangeRepAction::ChangeRepAction()
{
    SO_ACTION_CONSTRUCTOR(ChangeRepAction);
}

//
// Destructor. Does nothing.
//

ChangeRepAction::~ChangeRepAction()
{
}

//
// Initiates action on a graph. This is called when the action
// is applied to a node, a path, or a path list. It gives us a
// chance to initialize things before beginning traversal.
//

void
ChangeRepAction::beginTraversal(SoNode *node)
{
    traverse(node);
}

void ChangeRepAction::setNewRep(int theRep)
{
    theNewRep = theRep;
}

//
// This method implements the action for an SoSphere node.
//

void
ChangeRepAction::callDoAction(SoAction *action, SoNode *node)
{
   node->doAction(action);
}

void
ChangeRepAction::callRepAction(SoAction *action, SoNode *node)
{
   RepSwitch *rep = (RepSwitch *)node;
   ChangeRepAction *cra = (ChangeRepAction *)action;
   int x = rep->whichChild.getValue();
   if((cra->theNewRep == RepSwitch::REP_SENSING) ||
	(cra->theNewRep == RepSwitch::REP_SENSING_OO) ||
	(cra->theNewRep == RepSwitch::REP_REVERT)){
	rep->doAction(action);
   	rep->setCurrentRepImmediate(cra->theNewRep);
   }
   else {
	rep->whichChild.setValue(-3);
	rep->doAction(action);
	rep->whichChild.setValue(x);
	rep->setCurrentRepImmediate(cra->theNewRep);
    }
}

void
ChangeRepAction::callJointAction(SoAction *action, SoNode *node)
{
   JointModel *ja = (JointModel *)node;
   ChangeRepAction *cra = (ChangeRepAction *)action;
   if((cra->theNewRep != RepSwitch::REP_SENSING)&&
	(cra->theNewRep != RepSwitch::REP_SENSING_OO)&&
	(cra->theNewRep != RepSwitch::REP_REVERT)){
	SbBool jenable = ja->isNotifyEnabled();
	if(jenable == FALSE){
	    ja->enableNotify(TRUE);
	    ja->doAction(action);
	    ja->enableNotify(FALSE);
	}
	else
	    ja->doAction(action);
    }
    else
	ja->doAction(action);
}

#endif 
