#ifndef _REP_SWITCH_
#define _REP_SWITCH_
/*********************************************************************
 * 
 * 
 * 
 ********************************************************************/
#include <Inventor/SbBasic.h>
#include <Inventor/SoInput.h>
#include <Inventor/actions/SoAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoMFVec2f.h>
#include <Inventor/fields/SoMFInt32.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/actions/SoSubAction.h>
#include <Inventor/fields/SoSFEnum.h>
#include <Inventor/fields/SoMFEnum.h>
#include "../CreatureLib/CreatureCache.h"
#include <Inventor/SoPath.h>
#define CRA 1
/*
 * A RepSwitch is used as a way of switching between representations,  where 
 * each representation is a direct child of the instance of RepSwitch. See
 * the kbperson.iv file in AliveApp/alive for an example of how to specify
 * a RepSwitch. Here is an example though:
 *	RepSwitch {
	    reps [REP_SENSING,  REP_0,  REP_1]
	    repBindings [0, 1, 3]
	    currentRep REP_0
	    Separator {} #graphics used when being sensed
	    Separator {} #graphics used when REP_0
	    Separator {} # random for the sake of being random
	    Separator {} #graphics used when REP_1
	}
 *  reps are an enum which are how you refer to the reps. That is in code
 *  they would be RepSwitch::REP_0. In the .iv file you can just use the
 *  enum label (e.g. REP_0).The first 4 reps are reserved ( REP_NONE, 
 *  REP_ALL, REP_REVERT, REP_SENSING). repBindings corresponds to the index 
 *  of the child which corresponds to the rep in reps list. For example, 
 *  REP_SENSING is associated with child 0,  REP_0 -> 1,  REP_1 -> 3. Thus
 *  obviously, reps and repBindings have to be the same length.
 *  To change reps, you either send the creature the appropriate motor command
 *  creature->controller->executeMotorCommand(
	MotorCommandBlock(creature->controller->getCommandId("changeRep"), 
	    ArgType((int)newRep)));
 *  or cheat and do: creature->changeRep(newRep). This latter method will
 *  not work when doing distributed alive!!!,  so will need to do the former.
    
    At the command line you can also type "foo changeRep newRep",  in this
    case though you have to remember that REP_0 = 4 and so on.
    
    The changeRep method of BaseCreature is virtual so it can be over-ridden
    as I have done in BehaviorPerson, to allow you to take special actions
    when the rep changes.
    
    The updateStateAt method of RepSwitch is called if someone invokes the
    updateReps method of BaseCreature. This is done automatically for a
    MotorCreature for whom updateReps is set to TRUE.
    
    See AvatarRep in AliveLib for more comments.    
 */

class RepSwitch : public SoSwitch {

    SO_NODE_HEADER(RepSwitch);

public:
    enum RepType{
	REP_NONE, 
	REP_ALL, 
	REP_REVERT, 
	REP_SENSING,
	REP_SENSING_OO, 
	REP_0, 
	REP_1,
	REP_2,
	REP_3,
	REP_4,
	REP_5,
	REP_6,
	REP_7,
	REP_8,
	REP_9,
	REP_10 
    };
    SoMFEnum reps;    // these are reps this guy knows about
    SoMFInt32 repBindings;    // these are the bindings (the child)
    SoSFEnum currentRep;
    RepSwitch();
    static void initClass();
    void setCurrentRepImmediate(int i);
    int repBindingFor(int i);
    virtual void updateStateAt(BaseCreature *parent,UpdateStateData &usd);
    
    protected:
    int prevRep;
    SoFieldSensor *repSensor;
    virtual ~RepSwitch();
#ifdef SO_VERSION_REVISION
    SbBool readInstance(SoInput *in, unsigned short flags);
#else
    SbBool readInstance(SoInput *in);
#endif
    static void repSensorCB(void *, SoSensor *);
};

class RepContainer : public SoSeparator {

    SO_NODE_HEADER(RepContainer);

public:
    RepContainer();
    static void initClass();
    virtual void updateStateAt(BaseCreature *parent,UpdateStateData &usd);
    SoPath *buildPath(SoNode *startNode, SoNode *endNode);
    protected:
    virtual ~RepContainer();
#ifdef SO_VERSION_REVISION
    SbBool readInstance(SoInput *in, unsigned short flags);
#else
    SbBool readInstance(SoInput *in);
#endif
};


#ifdef CRA
class ChangeRepAction : public SoAction {

    SO_ACTION_HEADER(ChangeRepAction);

    public:
    // Initializes this action class for use with scene graphs
    static void    initClass();

    // Constructor and destructor
    ChangeRepAction();
    virtual ~ChangeRepAction();
    void setNewRep(int theRep);
    int theNewRep;
    
 protected:
    // Initiates action on graph
    virtual void   beginTraversal(SoNode *node);

 private:
    static void    callDoAction(SoAction *, SoNode *);
    static void    callRepAction(SoAction *, SoNode *);
    static void	   callJointAction(SoAction *action, SoNode *node);
};
#endif
#endif
