#include "SpatialAlgebra.h"
#include "vector.h"

SbVec4f SbRotationDeriv(SbRotation &sb,SbVec3f  w)
{
  float q1[4];
  SbVec4f rq;

  sb.getValue(q1[0],q1[1],q1[2],q1[3]);
  rq[0] = -q1[1]*w[0] -q1[2] *w[1] -q1[3]*w[2];
  rq[1] = q1[0]*w[0] -q1[3] *w[1] +q1[2]*w[2];
  rq[2] = q1[3]*w[0] +q1[0] *w[1] -q1[1]*w[2];
  rq[3] = -q1[2]*w[0] +q1[1] *w[1] +q1[0]*w[2];
  rq *= .5;
  return SbVec4f(rq[0],rq[1],rq[2],rq[3]);
}

void symmetricRbiInverseMultSpv(matrix irbi,Spv *sv)
{
  int i;
  for(i=0;i<3;i++){
    sv->v[i]  = sv->v[i]*irbi[i][i];
    sv->v0[i] = sv->v0[i]*irbi[i+3][i+3];
  }
  return;
}

void vectorCrossCross(SbVec3f &a,SbVec3f &b, SbMatrix &r)
{
    double azbz = -a[2]*b[2];
    double ayby = -a[1]*b[1];
    double axbx = -a[0]*b[0];
    
    r[0][0] = azbz +ayby;
    r[0][1] = a[1] *b[0];
    r[0][2] = a[2] *b[0];
    r[1][0] = a[0]*b[1];
    r[1][1] = azbz+axbx;
    r[1][2] = a[2] *b[1];
    r[2][0] = a[0]*b[2];
    r[2][1] = a[1] *b[2];
    r[2][2] = ayby + axbx;
}

matrix rbiToMatrix(Rbi &r,matrix m)
{
    int i,j;
    float h[3];
    
    if(!m)
	m = new_matrix(0,5,0,5);
    else
	/* fill with zeros */
	matrixFill(6,m,0.0);
    
    m[0][4] = -r.h[2];
    m[0][5] =  r.h[1];
    
    m[1][3] =  r.h[2];
    m[1][5] =  -r.h[0];
    
    m[2][3] =  -r.h[1];
    m[2][4] =   r.h[0];
    
    m[0][1] =   r.h[2];
    m[0][2] =  -r.h[1];
    
    m[1][0] =   -r.h[2];
    m[1][2] =   r.h[0];
    
    m[2][0] =   r.h[1];
    m[2][4] =   -r.h[0];

    
    /*mass matrix */
    m[3][3] = m[4][4] = m[5][5] = r.m;
    
    /*inertia */
    for(i=0;i<3;i++)
	for(j=0;j<3;j++)
	    m[i][j] = r.I[i][j];
    
    return m;
}


/************************************************************************
 *
 *              SPX METHODS                                     
 *
 ***********************************************************************/
 
Spx::Spx(SbMatrix e1,SbVec3f r1)
{
    e = e1;
    r = r1;
}

Spx::Spx()
{
  e.makeIdentity();
  r.setValue(0,0,0);
  
}


Spx Spx::inverse()
{
    SbMatrix e1;
    SbVec3f r1;
    e.multMatrixVec(r,r1);
    r1.negate();
    e1 = e.transpose();
    e1.multMatrixVec(r,r1);
    r1.negate();
    return Spx(e1,r1);
}

/*
Spx Spx::inverse()
{
    SbMatrix e1;
    SbVec3f r1;
    e1 = e.transpose();
    r1 = r;
    r1.negate();
    return Spx(e1,r1);
}
*/
void Spx::multVec(SbVec3f *pt)
{
  SbVec3f a,b;
  
  e.multMatrixVec(r,a);
  e.multMatrixVec(*pt,b);
  *pt = b - a;
}


void Spx::multInvVec(SbVec3f *pt)
{
  SbVec3f a,b;

  e.multVecMatrix(*pt,a);
  *pt = a + r;
}


void Spx::multSpv(Spv *arg)
{
    SbVec3f r1;
    SbVec3f r2;
    
    e.multMatrixVec(arg->v0 - r.cross(arg->v),arg->v0);
    e.multMatrixVec(arg->v,arg->v);
}

void print_matrix(SbMatrix arg)
{
  int i;
  for(i=0;i<4;i++)
    printf("%f %f %f %f\n",arg[i][0],arg[i][1],arg[i][2],arg[i][3]);
  return;
}

void Spx::multInvSpv(Spv *arg)
{
    SbVec3f r1;
    SbVec3f r2;
   
    e.multVecMatrix(arg->v,r1);
    e.multVecMatrix(arg->v0,r2);
    
/* check
    SbVec3f rr1, rr2, rr3;
    float err;
    SbMatrix ei;

    ei = e.inverse();
    ei.multMatrixVec(arg->v,rr1);
    ei.multMatrixVec(arg->v0,rr2);
    rr3 = (rr1-r1) + (rr2 - r2);
    if((err = rr3[0]+rr3[1]+rr3[2])>1e-6)
	printf("MULTINVSPV screwed up \n");
    
*/    
    
    arg->v0 = r.cross(r1)+r2;
    arg->v = r1;
    return;
}

Spx &Spx::multSpx(Spx *arg)
{
  SbMatrix e2;
  SbVec3f rt;
  
  e2 = arg->e.inverse();
  e2.multMatrixVec(r,r);
  r += arg->r;
  e.multRight(arg->e);
  return *this;
}

Spx &Spx::multInvSpx(Spx *arg)
{
  Spx arg_inverse = arg->inverse();
  return multSpx(&arg_inverse);
}

void Spx::multRbi(Rbi *rbi)
{
    SbMatrix I,hmrr, e2;
    SbVec3f hmr;
    int i,j;
    
    /* rXhX */
    vectorCrossCross(r,rbi->h,I);
    
    /* h - mr */
    hmr = rbi->h  - rbi->m*r;
    vectorCrossCross(hmr,r,hmrr);
    
    /* e hmr */
    e.multMatrixVec(hmr,rbi->h);
    
    for(i=0;i<3;i++)
        for(j=0;j<3;j++){
            rbi->I[i][j] += I[i][j] + hmrr[i][j];
	    e2[i][j] = e[j][i];
	}    
    rbi->I.multRight(e2);
    rbi->I.multLeft(e);
    
    return;
}
    

void Spx::print()
{
  int i;
  for(i=0;i<3;i++)
    printf("%f %f %f\n",e[i][0],e[i][1],e[i][2]);
  for(i=0;i<3;i++)
    printf("%f\n",r[i]);
  return;
}

/************************************************************************
 *
 *              SPV Methods                                     
 *
 ***********************************************************************/
void Spv::print()
{
  int i;
  for(i=0;i<3;i++)
    printf("%f\n",v[i]);
  for(i=0;i<3;i++)
    printf("%f\n",v0[i]);
  return;
}

void Spv::setValue(float *vec)
{
    v.setValue(vec[0],vec[1],vec[2]);
    v0.setValue(vec[3],vec[4],vec[5]);
}

void Spv::getValue(float *vec)
{
    v.getValue(vec[0],vec[1],vec[2]);
    v0.getValue(vec[3],vec[4],vec[5]);
}

    
Spv Spv::cross(Spv *arg)
{
    SbVec3f a,b;
    b = v0.cross(arg->v) + v.cross(arg->v0) ;
    a  = v.cross(arg->v0);
    return Spv(a,b);
}

Spv Spv::operator+(Spv &arg)
{
  SbVec3f a,b;
  
  a = v + arg.v;
  b = v0 + arg.v0;
  return Spv(a,b);
}

Spv Spv::operator-(Spv &arg)
{
  SbVec3f a,b;
  
  a = v - arg.v;
  b = v0 - arg.v0;
  return Spv(a,b);
}

Spv::Spv(SbVec3f a,SbVec3f b)
{
  v = a;
  v0 = b;
}

Spv Spv::spvFromForce(SbVec3f force,SbVec3f offset)
{
   return Spv(force.cross(offset),force);
}

Spv::Spv()
{
  v.setValue(0,0,0);
  v0.setValue(0,0,0);
}

/************************************************************************
 *
 *              RBI METHODS                                     
 *
 ***********************************************************************/

Rbi::Rbi()
{
}
        
Rbi::Rbi(SbMatrix inertia, float mass, SbVec3f op)
{
        SbVec3f a;
        int k,j;
        
        /* calculate Inertia term */
        a = op * -mass;
        vectorCrossCross(a,op,I);
        for(k=0;k<3;k++)
            for(j=0;j<3;j++)
                I[k][j] += inertia[k][j];
	I[3][3] = 1;
	I[0][3] = I[1][3] = I[2][3]=0;
	 
        /* now h */
        h = op * mass;
        
        /* now m */
        m = mass;       
}

void Rbi::multSpv(Spv *arg)
{
    SbVec3f t,t1;

    I.multMatrixVec(arg->v,t); 
    t1 = h.cross(arg->v0);   
    arg->v0 = m*(arg->v0)+(-1 * h).cross(arg->v);
    arg->v  = t1+t;
}

float Rbi::kineticEnergy(Spv *arg)
{
    float ke;
    Spv rbiv = *arg;
    
    multSpv(&rbiv);
    ke = .5*(rbiv.v.dot(arg->v) + rbiv.v0.dot(arg->v0));
    return ke;    
}

void Rbi::print()
{
    int i;
    printf("mass -> %f \n",m);
    printf("inertia ->\n");
    for(i=0;i<4;i++)
        printf("%f %f %f %f\n",I[i][0],I[i][1],I[i][2],I[i][3]);
    printf("offset ->\n");
    for(i=0;i<3;i++)
        printf("%f \n",h[i]);
    return;
}

/************************************************************************
 *
 *              Screw methods                                               
 *
 ***********************************************************************/
 int Scr::XAXIS = 0;
 int Scr::ZAXIS = 1;
 int Scr::YAXIS = 2;
 
 Scr::Scr(float ang,float disp, int whichAxis)
 {
    a = ang;
    d = disp;
    axis = whichAxis;
    c = cos(a);
    s = sin(a);
    zeroTwist = (fabs(a)<1.e-6)? 1 :0;
    zeroDisplacement = (fabs(d)<1.e-6)? 1 :0;	   
 }
 
 void Scr::setTwist(float t)
 {
     if (a != t) {
         a = t;
         c = cos(a);
         s = sin(a);
         zeroTwist = (fabs(a)<1.e-6)? 1 :0;
     }
 }
 
 void Scr::setDisplacement(float disp)
 {
     d = disp;
     zeroDisplacement = (fabs(d)<1.e-6)? 1 :0;
 }
 
 void Scr::multSpv(Spv *arg)
 {
     Spv v;
     float vxd;
     float vyd;
     float vzd;
	 
     if(axis == Scr::ZAXIS){
	 v.v[0] = c*arg->v[0] + s*arg->v[1];
	 v.v[1] = c*arg->v[1] - s*arg->v[0];
	 v.v[2] = arg->v[2];
	 if(zeroDisplacement){
	     v.v0[0] = c*(arg->v0[0]) + s*(arg->v0[1]);
	     v.v0[1] = c*(arg->v0[1]) - s*(arg->v0[0]); 
	 }
	 else{
	     vxd = arg->v[0]*d;
	     vyd = arg->v[1]*d;
	     v.v0[0] = c*(arg->v0[0] + vyd) + s*(arg->v0[1] - vxd);
	     v.v0[1] = c*(arg->v0[1] - vxd) - s*(arg->v0[0] + vyd);
	 }
	 v.v0[2] = arg->v0[2];
     }
     else if(axis == Scr::XAXIS){
	 v.v[0] = arg->v[0];
	 v.v[1] = c*arg->v[1] + s*arg->v[2];
	 v.v[2] = c*arg->v[2] - s*arg->v[1];
	 
	 if(zeroDisplacement){
	     v.v0[0] = arg->v0[0];
	     v.v0[1] = c*(arg->v0[1]) + s*(arg->v0[2]);
	     v.v0[2] = c*(arg->v0[2]) - s*(arg->v0[1]);
	 }
	 else {
	     vzd = arg->v[2]*d;
	     vyd = arg->v[1]*d;
	     v.v0[0] = arg->v0[0];
	     v.v0[1] = c*(arg->v0[1] + vzd) + s*(arg->v0[2] - vyd);
	     v.v0[2] = c*(arg->v0[2] - vyd) - s*(arg->v0[1] + vzd);
	 }
     }
     else if(axis == Scr::YAXIS){
	 v.v[0] = c*arg->v[0] - s*arg->v[2];
	 v.v[1] = arg->v[1];
	 v.v[2] = c*arg->v[2] + s*arg->v[0];
	 if(zeroDisplacement){
	    v.v0[0] = c*arg->v0[0] - s*arg->v0[2];
	    v.v0[1] = arg->v0[1];
	    v.v0[2] = s*arg->v0[0] + c*arg->v0[2];
	 }
	 else {
	     v.v0[0] = -s*d*arg->v[0]- c*d*arg->v[2] + c*arg->v0[0] - s*arg->v0[2];
	     v.v0[1] = arg->v0[1];
	     v.v0[2] = c*d*arg->v[0]- s*d*arg->v[2] + s*arg->v0[0] + c*arg->v0[2]; 
	 }
     }
     *arg = v;
 }
 
 void Scr::multRevoluteZ(Spv *arg)
 {
     Spv v;
     float vzd;
	 
     if(axis == Scr::ZAXIS){
	 v = Spv(SbVec3f(0, 0, arg->v[2]), SbVec3f(0, 0, 0));
     }
     else if(axis == Scr::XAXIS){
	 v.v[1] = s*arg->v[2];
	 v.v[2] = c*arg->v[2];
	 v.v[0] = 0;
	 if(!zeroDisplacement){
	     vzd = arg->v[2]*d;
	     v.v0[1] = c*(vzd);
	     v.v0[2] = - s*(vzd);
	     v.v0[0] = 0;
	 }
	 else
	    v.v0[0] = v.v0[1] = v.v0[2];
     }
     *arg = v;
 }
 
 void Scr::multVec(SbVec3f *v)
 {
     SbVec3f vec = *v;
     SbVec3f v1;
      
     if(axis == Scr::ZAXIS){
	  vec = -SbVec3f(0, 0, d)+ *v;
	  v1[0] = c*(vec[0]) + s*(vec[1]);
	  v1[1] = c*(vec[1]) - s*(vec[0]);
	  v1[2] = vec[2];
     }
     else if(axis == Scr::XAXIS){
	 vec = -SbVec3f(d, 0, 0) + *v;
	 v1[0] = vec[0];
	 v1[1] = c*(vec[1]) + s*(vec[2]);
	 v1[2] = c*(vec[2]) - s*(vec[1]); 
     }
     else if(axis == Scr::YAXIS){
	 vec = -SbVec3f(0, d, 0) + *v;
	 v1[0] = c*(vec[0]) - s*(vec[2]);
	 v1[1] = vec[1];
	 v1[2] = c*(vec[2]) + s*(vec[0]); 
     }
     *v = v1;
 }

void Scr::multInvVec(SbVec3f *v)
{
     SbVec3f vec = *v;
     SbVec3f v1;
      
     if(axis == Scr::ZAXIS){
	  vec = *v + SbVec3f(0, 0, d);
	  v1[0] = c*(vec[0]) - s*(vec[1]);
	  v1[1] = c*(vec[1]) + s*(vec[0]);
	  v1[2] = vec[2];
     }
     else if(axis == Scr::XAXIS){
	 vec = SbVec3f(d, 0, 0) + *v;
	 v1[0] = vec[0];
	 v1[1] = c*(vec[1]) - s*(vec[2]);
	 v1[2] = c*(vec[2]) + s*(vec[1]); 
     }
     else if(axis == Scr::YAXIS){
	 vec = SbVec3f(0, d, 0) + *v;
	 v1[1] = vec[1];
	 v1[0] = c*(vec[0]) + s*(vec[2]);
	 v1[2] = c*(vec[2]) - s*(vec[0]); 
     }
     *v = v1;
}

void Scr::rotVec(SbVec3f *v)
 {
     SbVec3f vec = *v;
     SbVec3f v1;
      
     if(axis == Scr::ZAXIS){
	  v1[0] = c*(vec[0]) + s*(vec[1]);
	  v1[1] = c*(vec[1]) - s*(vec[0]);
	  v1[2] = vec[2];
     }
     else if(axis == Scr::XAXIS){
	 v1[0] = vec[0];
	 v1[1] = c*(vec[1]) + s*(vec[2]);
	 v1[2] = c*(vec[2]) - s*(vec[1]); 
     }
     *v = v1;
 }

void Scr::rotInvVec(SbVec3f *v)
{
     SbVec3f vec = *v;
     SbVec3f v1;
      
     if(axis == Scr::ZAXIS){
	  v1[0] = c*(vec[0]) - s*(vec[1]);
	  v1[1] = c*(vec[1]) + s*(vec[0]);
	  v1[2] = vec[2];
     }
     else if(axis == Scr::XAXIS){
	 v1[0] = vec[0];
	 v1[1] = c*(vec[1]) - s*(vec[2]);
	 v1[2] = c*(vec[2]) + s*(vec[1]); 
     }
     *v = v1;
} 
/************************************************************************
 *
 *              TEST CASE                                               
 *
 ***********************************************************************/       
void test()
{
  Spv a,b,c;
  Spx x1,x2;
  SbRotation axis;

  a.v.setValue(10,20,30);
  a.v0.setValue(30,40,50);
  x1.e.setRotate(SbRotation(SbVec3f(0,0,1),1.34));
  x1.r.setValue(10,20,30);
  x1.multSpv(&a);
  printf("check multSpv \n");
   a.print();
  
  printf("check assignment \n");
  c= a;
  c.print();
  
  // test spxSpx
  printf("check spxSpx \n");
  a.v.setValue(10,20,30);
  a.v0.setValue(30,40,50);
  x2.e.setRotate(SbRotation(SbVec3f(0,0,1),1.34));
  x2.r.setValue(10,20,30);
  x1.multSpx(&x2);
  x1.print();
  
  printf("check spxInverse \n");
  Spx x3 = x1.inverse();
  x3.print();
  
  printf("check spxRbi \n");
   
  SbMatrix i;
  SbVec3f h;
  h.setValue(80,60,20);

  i[0][0] = 2000;
  i[1][1] = 15670;
  i[2][2] = 2000;

  Rbi rbi = Rbi(i,32.0,h);
  
  rbi.print();
  x1.multRbi(&rbi);
  rbi.print();
  
  printf("check rbiMultSpv \n");
  c.print();
  rbi.multSpv(&c);
  c.print();
  
  return;
  
}

