/*
 */
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSurroundScale.h>
#include <Inventor/nodes/SoAntiSquish.h>
#include <Inventor/sensors/SoFieldSensor.h>

// Include files for child dragger classes.
#include <Inventor/draggers/SoRotateCylindricalDragger.h>
#include <Inventor/draggers/SoTranslate2Dragger.h>
#include <Inventor/draggers/SoTranslate1Dragger.h>

// Include file for our new class.
#include "Translate222Dragger.h"
#include "Translate222DraggerGeom.h"

SO_KIT_SOURCE(Translate222Dragger);

void
Translate222Dragger::initClass()
{
   SO_KIT_INIT_CLASS(Translate222Dragger, SoDragger, "Dragger");	
}

Translate222Dragger::Translate222Dragger()
{
   SO_KIT_CONSTRUCTOR(Translate222Dragger);

   //
   // Don't create "surroundScale" by default. It's only put 
   // to use if this dragger is used within a manipulator.
   //
   SO_KIT_ADD_CATALOG_ENTRY(surroundScale, SoSurroundScale,
      TRUE, topSeparator, geomSeparator,TRUE);
   //
   // Create an anti-squish node by default.
   //
   SO_KIT_ADD_CATALOG_ENTRY(antiSquish, SoAntiSquish,
      FALSE, topSeparator, geomSeparator,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(XYTranslator, SoTranslate2Dragger,
      TRUE, topSeparator, geomSeparator,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(XZTranslatorSep, SoSeparator,
      FALSE, topSeparator, geomSeparator,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(XZTranslatorRot, SoRotation,
      TRUE, XZTranslatorSep, ,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(XZTranslator, 
      SoTranslate2Dragger, TRUE, XZTranslatorSep, ,TRUE);

   SO_KIT_ADD_CATALOG_ENTRY(YZTranslatorSep, SoSeparator,
      FALSE, topSeparator, geomSeparator,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(YZTranslatorRot, SoRotation,
      TRUE, YZTranslatorSep, ,FALSE);
   SO_KIT_ADD_CATALOG_ENTRY(YZTranslator, 
      SoTranslate2Dragger, TRUE, YZTranslatorSep, ,TRUE);

   // Read geometry resources. Only do this the first time we
   // construct one. 'geomBuffer' contains our compiled in
   // defaults. The user can override these by specifying new
   // scene graphs in the file:
   // $(SO_DRAGGER_DIR)/Translate222Dragger.iv"
   if (SO_KIT_IS_FIRST_INSTANCE())
      readDefaultParts("Translate222Dragger.iv", geomBuffer, 
      sizeof(geomBuffer));

   // Fields that always show current state of the dragger.
   SO_KIT_ADD_FIELD(translation, (0.0, 0.0, 0.0));

   // Creates parts list and default parts for this nodekit.
   SO_KIT_INIT_INSTANCE();

   // Make the anti-squish node surround the biggest dimension
   SoAntiSquish *myAntiSquish 
      = SO_GET_ANY_PART(this, "antiSquish", SoAntiSquish);
   myAntiSquish->sizing = SoAntiSquish::BIGGEST_DIMENSION;


   SoDragger *tXYDragger = SO_GET_ANY_PART(this, "XYTranslator", 
      SoTranslate2Dragger);
   SoDragger *tXZDragger = SO_GET_ANY_PART(this, "XZTranslator", 
      SoTranslate2Dragger);
   SoDragger *tYZDragger = SO_GET_ANY_PART(this, "YZTranslator", 
      SoTranslate2Dragger);


   SoRotation *tYZRot = new SoRotation;
   tYZRot->rotation.setValue(
      SbRotation(SbVec3f(0,0,1), SbVec3f(1,0,0)));
   setAnyPartAsDefault("YZTranslatorRot", tYZRot);

   SoRotation *tXZRot = new SoRotation;
   tXZRot->rotation.setValue(
      SbRotation(SbVec3f(0,0,1), SbVec3f(0,1,0)));
   setAnyPartAsDefault("XZTranslatorRot", tXZRot);


   // Updates the fields when motionMatrix changes 
   addValueChangedCallback(&Translate222Dragger::valueChangedCB);

   // Updates motionMatrix when either field changes.
   translFieldSensor = new SoFieldSensor(
      &Translate222Dragger::fieldSensorCB,this);
   translFieldSensor->setPriority(0);

   setUpConnections( TRUE, TRUE );
}

Translate222Dragger::~Translate222Dragger()
{
   if (translFieldSensor)
      delete translFieldSensor;
}


SbBool
Translate222Dragger::setUpConnections( SbBool onOff, 
			    SbBool doItAlways)
{
   if ( !doItAlways && connectionsSetUp == onOff)
      return onOff;

   if (onOff) {
      // We connect AFTER base class.
      SoDragger::setUpConnections(onOff, doItAlways);


       SoDragger *tD 
	  = (SoDragger *) getAnyPart("XYTranslator", FALSE);
       // [a] Set up the parts in the child dragger...
       tD->setPartAsDefault("translator",
	  "Translate222TranslatorTranslator");
       tD->setPartAsDefault("translatorActive",
	  "Translate222TranslatorTranslatorActive");
       tD->setPartAsDefault("feedback",
	  "Translate222TranslatorFeedback");
       tD->setPartAsDefault("feedbackActive",
	  "Translate222TranslatorFeedbackActive");
       tD->addStartCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       tD->addFinishCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       registerChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("YZTranslator", FALSE);
       // [a] Set up the parts in the child dragger...
       tD->setPartAsDefault("translator",
	  "Translate222TranslatorTranslator");
       tD->setPartAsDefault("translatorActive",
	  "Translate222TranslatorTranslatorActive");
       tD->setPartAsDefault("feedback",
	  "Translate222TranslatorFeedback");
       tD->setPartAsDefault("feedbackActive",
	  "Translate222TranslatorFeedbackActive");
       tD->addStartCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       tD->addFinishCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       registerChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("XZTranslator", FALSE);
       // [a] Set up the parts in the child dragger...
       tD->setPartAsDefault("translator",
	  "Translate222TranslatorTranslator");
       tD->setPartAsDefault("translatorActive",
	  "Translate222TranslatorTranslatorActive");
       tD->setPartAsDefault("feedback",
	  "Translate222TranslatorFeedback");
       tD->setPartAsDefault("feedbackActive",
	  "Translate222TranslatorFeedbackActive");
       tD->addStartCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       tD->addFinishCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       registerChildDragger(tD);


      // Call the sensor CB to make things up-to-date.
      fieldSensorCB(this, NULL);

      // Connect the field sensors
      if (translFieldSensor->getAttachedField() != &translation)
	 translFieldSensor->attach( &translation );
   }
   else {
      // We disconnect BEFORE base class.

      // Remove the callbacks from the child draggers,
      // and unregister them as children.
       SoDragger *tD 
	  = (SoDragger *) getAnyPart("XYTranslator", FALSE);
       tD->removeStartCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       tD->removeFinishCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("XZTranslator", FALSE);
       tD->removeStartCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       tD->removeFinishCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(tD);

       tD 
	  = (SoDragger *) getAnyPart("YZTranslator", FALSE);
       tD->removeStartCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       tD->removeFinishCallback(
	  &Translate222Dragger::invalidateSurroundScaleCB, this);
       unregisterChildDragger(tD);

      // Disconnect the field sensors.
      if (translFieldSensor->getAttachedField())
	translFieldSensor->detach();

      SoDragger::setUpConnections(onOff, doItAlways);
   }

   return !(connectionsSetUp = onOff);
}


// Called when the motionMatrix changes. Sets the "translation"
// and "rotation" fields based on the new motionMatrix
void
Translate222Dragger::valueChangedCB(void *, SoDragger *inDragger)
{
   Translate222Dragger *myself = (Translate222Dragger *) inDragger;

   // Factor the motionMatrix into its parts
   SbMatrix motMat = myself->getMotionMatrix();
   SbVec3f   trans, scale;
   SbRotation rot, scaleOrient;
   motMat.getTransform(trans, rot, scale, scaleOrient);

   // Set the fields. Disconnect the sensors while doing so.
   myself->translFieldSensor->detach();
   if (myself->translation.getValue() != trans)
      myself->translation = trans;
   myself->translFieldSensor->attach(&(myself->translation));
}

// If the "translation" or "rotation" field changes, changes
// the motionMatrix accordingly.
void
Translate222Dragger::fieldSensorCB(void *inDragger, SoSensor *)
{
   Translate222Dragger *myself = (Translate222Dragger *) inDragger;

   SbMatrix motMat = myself->getMotionMatrix();
   myself->workFieldsIntoTransform(motMat);

   myself->setMotionMatrix(motMat);
}

// When any child dragger starts or ends a drag, tell the
// "surroundScale" part (if it exists) to invalidate its 
// current bounding box value.
void 
Translate222Dragger::invalidateSurroundScaleCB(void *parent, 
   SoDragger *)
{
   Translate222Dragger *myParentDragger 
      = (Translate222Dragger *) parent;

   // Invalidate the surroundScale, if it exists.
   SoSurroundScale *mySS = SO_CHECK_PART(
      myParentDragger, "surroundScale", SoSurroundScale);
   if (mySS != NULL)
      mySS->invalidate();
}

void
Translate222Dragger::setDefaultOnNonWritingFields()
{
    // These nodes pointed to by these part-fields may 
    // change after construction, but we
    // don't want to write them out.
    surroundScale.setDefault(TRUE);
    antiSquish.setDefault(TRUE);

    SoDragger::setDefaultOnNonWritingFields();
}
