/*------------------------------------------------------------
 *-----------------------------------------------------------*/
 
#ifndef  _TRANSLATE222_DRAGGER_
#define  _TRANSLATE222_DRAGGER_

#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/draggers/SoDragger.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFRotation.h>

class TranslateRadialDragger;
class SoRotateCylindricalDragger;

class Translate222Dragger : public SoDragger
{
   SO_KIT_HEADER(Translate222Dragger);

   // Makes the dragger surround other objects
   SO_KIT_CATALOG_ENTRY_HEADER(surroundScale);
   // Keeps the dragger evenly sized in all 3 dimensions
   SO_KIT_CATALOG_ENTRY_HEADER(antiSquish);

   // The translating dragger...
   SO_KIT_CATALOG_ENTRY_HEADER(XYTranslator);

   SO_KIT_CATALOG_ENTRY_HEADER(XZTranslatorSep);
   SO_KIT_CATALOG_ENTRY_HEADER(XZTranslatorRot);
   SO_KIT_CATALOG_ENTRY_HEADER(XZTranslator);

   SO_KIT_CATALOG_ENTRY_HEADER(YZTranslatorSep);
   SO_KIT_CATALOG_ENTRY_HEADER(YZTranslatorRot);
   SO_KIT_CATALOG_ENTRY_HEADER(YZTranslator);

  public:
  
   // Constructor
   Translate222Dragger();

   // These fields reflect state of the dragger at all times.
   SoSFVec3f   translation;

   // This should be called once after SoInteraction::init().
   static void initClass();

  protected:

   // These sensors insures that the motionMatrix is updated 
   // when the fields are changed from outside.
   SoFieldSensor *translFieldSensor;
   static void fieldSensorCB(void *, SoSensor *);

   // This function is invoked by the child draggers when they 
   // change their value.
   static void valueChangedCB(void *, SoDragger *);

   // Called at the beginning and end of each dragging motion.
   // Tells the "surroundScale" part to recalculate.
   static void invalidateSurroundScaleCB(void *, SoDragger *);

   // This will detach/attach the fieldSensor.
   // It is called at the end of the constructor (to attach).
   // and at the start/end of SoBaseKit::readInstance()
   // and on the new copy at the start/end of SoBaseKit::copy()
   // Returns the state of the node when this was called.
   virtual SbBool setUpConnections( SbBool onOff, 
				SbBool doItAlways = FALSE);

   // This allows us to specify that certain parts do not
   // write out. We'll use this on the antiSquish and
   // surroundScale parts.
   virtual void setDefaultOnNonWritingFields();

  private:

   static const char geomBuffer[];

   // Destructor.
   ~Translate222Dragger();
};   

#endif  /* _TRANSLATE222_DRAGGER_ */
