#include "../CreatureLib/Vector.h"
#include <stdio.h>
#include "lu.h"
#include <math.h>
void dummy()
{
    printf("dummy got called \n");
}

Vector::Vector(int row)
{
    m = row;
    int i;
    for(i=0;i<32;i++)
	mm[i] = 0.0;
}

Vector::Vector()
{
    m = 0;
}

Vector::~Vector()
{
}

Vector operator+(const Vector &m1, const Vector &m2)
{
    int r;
    Vector nm = Vector(m1.m);
    for(r = 0;r<m1.m;r++)
	nm.mm[r] = m1.mm[r] + m2.mm[r];
    return nm;
}

Vector operator-(const Vector &m1, const Vector &m2)
{
    int r;
    Vector nm = Vector(m1.m);
    for(r = 0;r<m1.m;r++)
	nm.mm[r] = m1.mm[r] - m2.mm[r];
    return nm;
}

float Vector::sum()
{
    int i;
    float r = 0;
    for(i=0;i<m;i++)
	r += mm[i];
    return r;
}

float Vector::abssum()
{
    int i;
    float r = 0;
    for(i=0;i<m;i++)
	r += fabs(mm[i]);
    return r;
}

float Vector::dot(const Vector &m1)
{
    return sqrt((*this * m1).sum());
}

Vector Vector::shift(int i)
{
    int j;
    Vector nm = Vector(m);
    if(i>0){
	for(j=0;j<m;j++)
	    nm.mm[j+i] = mm[j];
    }
    else {
	for(j=abs(i);j<m;j++)
	    nm.mm[j+i] = mm[j];
    }
    return nm;
}

Vector operator*(const Vector &m1, const Vector &m2)
{
    int i;
    Vector nm = Vector(m1.m);
    for(i=0;i<nm.m;i++)
	nm.mm[i] = m1.mm[i]*m2.mm[i];
    return nm;
}

Vector &Vector::operator=(const Vector &m2)
{
    int r;
    m = m2.m;
    for(r = 0;r<m2.m;r++)
	mm[r] = m2.mm[r];
    return *this;
}

void Vector::print()
{
    int i;
    printf("%d \n", m);
    for(i=0;i<m;i++){
	printf("%f ", mm[i]);
	printf("\n");
    }
}

void Vector::print(const char * t)
{
    printf("%s \n", t);
    print();
}

Matrix::Matrix(const Matrix &m1)
{
//    printf("Copy got called \n");
    mm = new_matrix(0, m1.m,0, m1.n);
    m = m1.m;
    n = m1.n; 
    for (int r = 0; r < m; r++)
      for (int c = 0; c < n; c++)
	mm[r][c] = m1.mm[r][c];
}
Matrix::Matrix()
{
    m = n = 0;
    mm = NULL;
}
Matrix::Matrix(int row, int col)
{
    m = row;
    n = col;
    mm = new_matrix(0, m, 0, n);
}

Matrix::~Matrix()
{
   // printf("freeing matrix %d %d \n", m, n);
    free_matrix(mm, 0, m, 0, n);
}

Matrix operator*(const Matrix &m1, const Matrix &m2)
{
    Matrix nm = Matrix(m1.m, m2.n);
    int r, c, k;
    for(r=0;r<nm.m;r++){
	for(c=0;c< nm.n; c++){
	    nm.mm[r][ c] = 0.0;
	    for(k=0;k<m1.n;k++){
		nm.mm[r][c] += m1.mm[r][k]*m2.mm[k][c];
	    }
	}
    }
    return nm;
}

Vector operator*(const Matrix &m1, const Vector &m2)
{
    int r, c;
    Vector nm = Vector(m1.m);
    for(r=0;r<m1.m;r++){
	for(c=0;c< m1.n; c++){
	    nm.mm[r] += m1.mm[r][c]*m2.mm[c];
	}
    }
    return nm;
}

Matrix &Matrix::operator=(const Matrix &m2)
{
    int r, c;
    if(mm)
	free_matrix(mm, 0, m, 0, n);
    mm = new_matrix(0, m2.m, 0, m2.n);
    m = m2.m;
    n = m2.n;
    for(r=0;r<m;r++){
	for(c=0;c< n; c++){
	    mm[r][c] = m2.mm[r][c];
	}
    }
    return *this;
}


Matrix Matrix::inverse()
{
    Matrix nm = Matrix(m, n);
    luinverse(mm,m, nm.mm);
    return nm;
}

Matrix Matrix::inverseSVD()
{
    Matrix nm = Matrix(m, n);
    svdInverse(mm, m, nm.mm);
    return nm;
}

Matrix operator+(const Matrix &m1, const Matrix &m2)
{
    int r, c;
    Matrix nm = Matrix(m1.m, m2.n);
    for(r = 0;r<nm.m;r++)
	for(c = 0;c<nm.n;c++)
	    nm.mm[r][c] = m1.mm[r][c] + m2.mm[r][c];
    return nm;

}

Matrix &Matrix::identity()
{
    int i;
    for(i=0;i<m;i++)
	mm[i][i] = 1;
    return *this;
}

Matrix operator*(const Matrix &m1, const float &s)
{
    int r, c;
    Matrix nm = Matrix(m1.m, m1.n);
    for(r = 0;r<nm.m;r++)
	for(c = 0;c<nm.n;c++)
	    nm.mm[r][c] = m1.mm[r][c]*s;
    return nm;
}

Matrix operator-(const Matrix &m1,const Matrix &m2)
{
    int r, c;
    Matrix nm = Matrix(m1.m, m2.n);
    for(r = 0;r<nm.m;r++)
	for(c = 0;c<nm.n;c++)
	    nm.mm[r][c] = m1.mm[r][c] - m2.mm[r][c];
    return nm;
}

Matrix Matrix::transpose()
{
    Matrix nm(n, m);
    int i, j;
    for(i=0;i<m;i++)
	for(j=0;j<n;j++)
	    nm.mm[j][i] = mm[i][j];
    return nm;
}

Vector Matrix::solve(const Vector &b)
{
    int i;
    Vector nv = Vector(n);
    Vector nb = Vector(b.m);
    for(i=0;i<b.m;i++)
	nb.mm[i] = b.mm[i];
    luSolve(mm, m, nb.mm);
     for(i=0;i<nv.m;i++)
	nv.mm[i] = nb.mm[i];
    return nb;
}

void Matrix::print()
{
    int i, j;
    printf("%d %d \n", m, n);
    for(i=0;i<m;i++){
	for(j=0;j<n;j++)
	    printf("%f ", mm[i][j]);
	printf("\n");
    }
}

void Matrix::print(const char * t)
{
    printf("%s \n", t);
    print();
}
