#include "../CreatureLib/vector.h"

/* LU decomposition Linear Equation Solver from Numerical Recipes
 * luSolve solves system ax=b, where a is a matrix created via matrix(0,n,0,n)
 * n is number of rows/columns (square matrix), b is a vector created
 * via vector(0,n). a should contain coefficients and b is righthand side, on
 * output b will contain values for x.  a is modified.
 * luInverse calculates inverse of a square matrix
 * ludcmp and lubksb are used by luSolve. However, if you have a constant
 * coefficient matrix but differing right-hand sides, call ludcmp once and
 * lubksb once for each right-hand vector
 *
 * luSolve and ludcmp return 1 on success and 0 on failure due to singular 
 * matrix
 */
 
int  luSolve(float **a, int n, float *b);
int ludcmp(float **a, int n, int *indx, float *d);
void lubksb(float **a, int n, int *indx, float *b);
int luinverse(float **a, int n, float **y);

void sparse(vector b,int n,vector x,float *rsq);

/* WARNING: svdcmp expects a square array and a matrix which goes from 1..n */
int svdSolve(matrix a,int n,vector b);
void svbksb(matrix u,vector w,matrix v,int m,int n,vector b,vector x);
int svdcmp(matrix a,int m,int n,vector w,matrix v);
matrix svdInverse(matrix a,int dim,matrix b);
