#include "../CreatureLib/vector.h"
#include <math.h>
#include <stdio.h>
#include <malloc.h>
#include <bstring.h>
#define X 0
#define Y 1
#define Z 2
float * new_vector(int nl, int nh)
{
    float *v;
    
    v = (float *) calloc((unsigned)(nh-nl+1),sizeof(float));
    return v-nl;
}

void free_vector(float *v, int nl,int nh)
{
    if(nl>1)
	nl =0;
    free((char *)(v+nl));
}

float ** new_matrix(int nrl, int nrh, int ncl, int nch)
{
    int i;
    float **m;
    
    m = (float **)malloc(
	(unsigned)(nrh-nrl+1)*sizeof(float *));
    m -=nrl;
   for(i=nrl;i<=nrh;i++){
	m[i] = (float *)calloc(
	    (unsigned)(nch-ncl+1),sizeof(float));
	m[i] -=ncl;
    }
    return m;
}

void free_matrix(float **m,int nrl,int nrh, int ncl,int nch)
{
    int i;
    for(i=nrh;i>=nrl;i--)
	free((char *)(m[i]+ncl));
    free((char *)(m+nrl));
    return;
}

matrix copyMatrix(int dim, matrix src)
{
    int i;
    matrix dest = new_matrix(0,dim-1,0,dim-1);
    
    for(i=0;i<dim;i++)
	bcopy(&src[i][0],&dest[i][0],dim*sizeof(float));
    return dest;
}

vector copyVector(int dim, vector src)
{
    int j;
    vector dest = new_vector(0,dim-1);
    
    bcopy(src,dest,dim*sizeof(float));
    return dest;
}

matrix identityMatrix(int dim,matrix dest)
{
    int i,j;
    if(!dest)
	dest = new_matrix(0,dim-1,0,dim-1);
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++){
	    if(i==j)
		dest[i][j] = 1.0;
	    else
		dest[i][j] = 0.0;
	}
    return dest;
}

matrix matrixAssign(int dim,matrix src,matrix dest)
{
    int i;
    for(i=0;i<dim;i++)
	bcopy(&src[i][0],&dest[i][0],dim*sizeof(float));
    return dest;
}

vector vectorAssign(int dim,vector src,vector dest)
{
    bcopy(src,dest,dim*sizeof(float));
    return dest;
}

matrix matrixTranspose(int dim,matrix src,matrix dest)
{
    int i,j;
    matrix tt=NULL;
    matrix tsrc=src;
    if(src==dest){
	tt = copyMatrix(dim,src);
	tsrc = tt;
    } 
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++)
	    dest[i][j] = tsrc[j][i];
    if(tt !=NULL)
	free_matrix(tt,0,dim-1,0,dim-1);
    return dest;
}



matrix tensorTransform(int dim,matrix xform,matrix tensor,matrix result)
{
    matrix copy = copyMatrix(dim,tensor);
    
    matrixMult(dim,0,INVERSE,copy,xform,result);
    matrixMult(dim,0,0,xform,result,result);
    free_matrix(copy,0,dim-1,0,dim-1);
    return result;
}
	
matrix  matrixMult( int dim, int invflag1,int invflag2, matrix m1,matrix m2, matrix m3)
{
    int r,c,k;
    matrix tt=NULL,t1=m1,t2=m2;
    
    if(m3==m1){
	tt = copyMatrix(dim,m1);
	t1 = tt;
    }
    else if(m2==m3){
	tt = copyMatrix(dim,m2);
	t2= tt;
    }
    if((!invflag1)&&(!invflag2)){
	for(r=0;r<dim;r++){
	    for(c=0;c<dim;c++){
		m3[r][c] = 0.0;
		for(k=0;k<dim;k++)
		    m3[r][c] += t1[r][k]*t2[k][c];
	    }
	}
    }
    else if ((invflag1)&&(!invflag2)){
	for(r=0;r<dim;r++){
	    for(c=0;c<dim;c++){
		m3[r][c] = 0.0;
		for(k=0;k<dim;k++)
		    m3[r][c] += t1[k][r]*t2[k][c];
	    }
	}
    }
    else if((!invflag1)&&(invflag2)){
	for(r=0;r<dim;r++){
	    for(c=0;c<dim;c++){
		m3[r][c] = 0.0;
		for(k=0;k<dim;k++)
		    m3[r][c] += t1[r][k]*t2[c][k];
	    }
	}
    }
    if(tt !=NULL)
	free_matrix(tt,0,dim-1,0,dim-1);
    return m3;
}

vector matrixVectorMult(int dim,int invflag, matrix m1,vector v1,vector v2)
{
    int i,j;
    vector tv=NULL,tv1 = v1;
    if(v1==v2){
	tv = copyVector(dim,v1);
	tv1 = tv;
    }
    if(!invflag){
	if(dim==3){
	    v2[0] = m1[0][0]*tv1[0] + m1[0][1]*tv1[1] +m1[0][2]*tv1[2];
	    v2[1] = m1[1][0]*tv1[0] + m1[1][1]*tv1[1] +m1[1][2]*tv1[2];
	    v2[2] = m1[2][0]*tv1[0] + m1[2][1]*tv1[1] +m1[2][2]*tv1[2];
	}
	else if(dim==6){
	    v2[0] = m1[0][0]*tv1[0] + m1[0][1]*tv1[1] +m1[0][2]*tv1[2]+
		m1[0][3]*tv1[3] + m1[0][4]*tv1[4] +m1[0][5]*tv1[5];
	    v2[1] = m1[1][0]*tv1[0] + m1[1][1]*tv1[1] +m1[1][2]*tv1[2]+
		m1[1][3]*tv1[3] + m1[1][4]*tv1[4] +m1[1][5]*tv1[5];
	    v2[2] = m1[2][0]*tv1[0] + m1[2][1]*tv1[1] +m1[2][2]*tv1[2]+
		m1[2][3]*tv1[3] + m1[2][4]*tv1[4] +m1[2][5]*tv1[5];
	    v2[3] = m1[3][0]*tv1[0] + m1[3][1]*tv1[1] +m1[3][2]*tv1[2]+
		m1[3][3]*tv1[3] + m1[3][4]*tv1[4] +m1[3][5]*tv1[5];
	    v2[4] = m1[4][0]*tv1[0] + m1[4][1]*tv1[1] +m1[4][2]*tv1[2]+
		m1[4][3]*tv1[3] + m1[4][4]*tv1[4] +m1[4][5]*tv1[5];
	    v2[5] = m1[5][0]*tv1[0] + m1[5][1]*tv1[1] +m1[5][2]*tv1[2]+
		m1[5][3]*tv1[3] + m1[5][4]*tv1[4] +m1[5][5]*tv1[5];
	}
	else {    
	    for(i=0;i<dim;i++){
		v2[i] = 0;
		for(j=0;j<dim;j++)
		    v2[i] += m1[i][j]*tv1[j];
	    }
	}
    }
    else {
	if(dim==3){
	    v2[0] = m1[0][0]*tv1[0] + m1[1][0]*tv1[1] +m1[2][0]*tv1[2];
	    v2[1] = m1[0][1]*tv1[0] + m1[1][1]*tv1[1] +m1[2][1]*tv1[2];
	    v2[2] = m1[0][2]*tv1[0] + m1[1][2]*tv1[1] +m1[2][2]*tv1[2];
	}
	else if(dim==6){
	    v2[0] = m1[0][0]*tv1[0] + m1[1][0]*tv1[1] +m1[2][0]*tv1[2]+
		m1[3][0]*tv1[3] + m1[4][0]*tv1[4] +m1[5][0]*tv1[5];
	    v2[1] = m1[0][1]*tv1[0] + m1[1][1]*tv1[1] +m1[2][1]*tv1[2]+
		m1[3][1]*tv1[3] + m1[4][1]*tv1[4] +m1[5][1]*tv1[5];
	    v2[2] = m1[0][2]*tv1[0] + m1[1][2]*tv1[1] +m1[2][2]*tv1[2]+
		m1[3][2]*tv1[3] + m1[4][2]*tv1[4] +m1[5][2]*tv1[5];
	    v2[3] = m1[0][3]*tv1[0] + m1[1][3]*tv1[1] +m1[2][3]*tv1[2]+
		m1[3][3]*tv1[3] + m1[4][3]*tv1[4] +m1[5][3]*tv1[5];
	    v2[4] = m1[0][4]*tv1[0] + m1[1][4]*tv1[1] +m1[2][4]*tv1[2]+
		m1[3][4]*tv1[3] + m1[4][4]*tv1[4] +m1[5][4]*tv1[5];
	    v2[5] = m1[0][5]*tv1[0] + m1[1][5]*tv1[1] +m1[2][5]*tv1[2]+
		m1[3][5]*tv1[3] + m1[4][5]*tv1[4] +m1[5][5]*tv1[5];
	}
	else{
	    for(i=0;i<dim;i++){
		v2[i] = 0;
		for(j=0;j<dim;j++)
		    v2[i] += m1[j][i]*tv1[j];
	    }
	}
    }
    if(tv !=NULL)
	free_vector(tv,0,dim-1);
    return v2; 
}
vector matrixVectorMult1(int dim,int invflag, matrix m1,vector v1,vector v2)
{
    int i,j;
    vector tv=NULL,tv1 = v1;
    if(v1==v2){
	tv = copyVector(dim,v1);
	tv1 = tv;
    }
    if(!invflag){
	for(i=0;i<dim;i++){
	    v2[i] = 0;
	    for(j=0;j<dim;j++)
		v2[i] += m1[i][j]*tv1[j];
	}
    }
    else {
	for(i=0;i<dim;i++){
	    v2[i] = 0;
	    for(j=0;j<dim;j++)
		v2[i] += m1[j][i]*tv1[j];
	}
    }
    if(tv !=NULL)
	free_vector(tv,0,dim-1);
    return v2; 
}

vector vectorNormalize(int dim,vector src,vector dest)
{
    float d;
    d = sqrt(vectorDotProduct(dim,src,src));
    d = (d==0.0)?1.0: 1/d;
    vectorScalarMult(dim,src,d,dest);
    return dest;
}

vector vectorNormal(vector v1,vector v2,vector v3,vector norm)
{
    float v13[3];
    float v23[3];
    
    vectorSubtract(3,v1,v3,v13);
    vectorSubtract(3,v2,v3,v23);
    vectorNormalize(3,v13,v13);
    vectorNormalize(3,v23,v23);
    vectorCrossProduct(v13,v23,norm);
    return norm;
}
    
vector vectorFill(int dim,vector v,float scalar)
{
    int i;
    if(scalar ==0.0)
	bzero(v,dim*sizeof(float));
    else
	for(i=0;i<dim;i++)
	    v[i] = scalar;
    return v;
}

int vectorFindMax(int dim,vector v)
{
    int i,max=0;
    float maxValue = 0.0;
    for(i=0;i<dim;i++)
	if(fabs(v[i])>fabs(maxValue)){
	    maxValue = v[i];
	    max = i;
	}
    return max;
}
vector vectorScalarMult(int dim,vector v,float scalar,vector r)
{
    int i;
    if(dim==3){
	r[0] = v[0]*scalar;
	r[1] = v[1]*scalar;
	r[2] = v[2]*scalar;
    }
    else if(dim==6){
	r[0] = v[0]*scalar;
	r[1] = v[1]*scalar;
	r[2] = v[2]*scalar;
	r[3] = v[3]*scalar;
	r[4] = v[4]*scalar;
	r[5] = v[5]*scalar;
    }
    else
	for(i=0;i<dim;i++)
	    r[i] = v[i]*scalar;
    return r;
}

matrix matrixFill(int dim,matrix m, float fill)
{
    int i,j;
    if(fill == 0.0){
	for(i=0;i<dim;i++)
	    bzero(&m[i][0],dim*sizeof(float));
    }
    else
	for(i=0;i<dim;i++)
	    for(j=0;j<dim;j++)
		m[i][j] = fill;
    return m;
}
	    
matrix matrixScalarMult(int dim,matrix m,float scalar,matrix r)
{
    int i,j;
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++)
	r[i][j] = m[i][j]*scalar;
    return r;
}

matrix matrixFromVectorCross(vector v1,matrix m)
{
    int i;
    if(m ==NULL)
	m = new_matrix(0,2,0,2);
    for(i=0;i<3;i++)
	m[i][i] = 0.0;
    m[0][1] = -v1[2];
    m[0][2] = v1[1];
    m[1][2] = -v1[0];
    m[1][0] = v1[2];
    m[2][0] = -v1[1];
    m[2][1] = v1[0];
    return m;
}

vector vectorCrossProduct(vector v1,vector v2,vector v3)
{
   vector tt=NULL,t1=v1,t2=v2;
    
    if(v3==v1){
	tt = copyVector(3,v1);
	t1 = tt;
    }
    else if(v2==v3){
	tt = copyVector(3,v2);
	t2= tt;
    }
    v3[X] = -t1[Z]*t2[Y] + t1[Y]*t2[Z];
    v3[Y] =  t1[Z]*t2[X] - t1[X]*t2[Z];
    v3[Z] = -t1[Y]*t2[X] + t1[X]*t2[Y];
    
    if(tt !=NULL)
	free_vector(tt,0,2);
    return v3;
}

matrix vectorMatrixCrossProduct(vector v1,matrix m1,matrix m2)
{
    int i,j;
    dtuple tv,tv1;
    matrix tm = m1,tt=NULL;
    
    if(m2==m1){
	tt = copyMatrix(3,m1);
	tm =tt;
    }
    for(i=0;i<3;i++){
	for(j=0;j<3;j++)
	    tv[j] = tm[j][i];
	vectorCrossProduct(v1,tv,tv1);
	for(j=0;j<3;j++)
	m2[j][i] = tv1[j];
    }
    if(tt !=NULL)
	free_matrix(tt,0,2,0,2);
    return m2;
}

matrix matrixVectorCrossProduct(matrix m1,vector v1,matrix m2)
{
    int i,j;
    dtuple tv,tv1;
    matrix tm = m1,tt=NULL;
    
    if(m2==m1){
	tt = copyMatrix(3,m1);
	tm =tt;
    }
    for(i=0;i<3;i++){
	for(j=0;j<3;j++)
	    tv[j] = tm[i][j];
	vectorCrossProduct(v1,tv,tv1);
	for(j=0;j<3;j++)
	    m2[i][j] = -tv1[j];
    }
    if(tt !=NULL)
	free_matrix(tt,0,2,0,2);
    return m2;
}

float vectorDistance(int dim,vector v1,vector v2)
{
    dtuple v3;
    vectorSubtract(3,v1,v2,v3);
    return sqrt(vectorDotProduct(3,v3,v3));
}
    
float vectorDotProduct(int dim,vector v1,vector v2)
{
    int i;
    float r=0.0;
    
    if(dim == 3){
	r = v1[0]*v2[0]+
	    v1[1]*v2[1]+
	    v1[2]*v2[2];
    }
    else if(dim ==6){
	r = v1[0]*v2[0]+
	    v1[1]*v2[1]+
	    v1[2]*v2[2]+
	    v1[3]*v2[3]+
	    v1[4]*v2[4]+
	    v1[5]*v2[5];
    }
    else 
	for(i=0;i<dim;i++)
	    r += v1[i] *v2[i];
    return r;
}

matrix vectorProduct(int dim,vector v1,vector v2,matrix m)
{
    int i,j;
    if(!m)
	m = new_matrix(0,dim-1,0,dim-1);
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++)
	    m[i][j] = v1[i]*v2[j];
    return m;
}

vector vectorAdd(int dim, vector v1,vector v2,vector v3)
{
    int i;
    if(dim == 3){
	v3[0] = v1[0]+v2[0];
	v3[1] = v1[1]+v2[1];
	v3[2] = v1[2]+v2[2];
    }
    else if(dim ==6){
	v3[0] = v1[0]+v2[0];
	v3[1] = v1[1]+v2[1];
	v3[2] = v1[2]+v2[2];
	v3[3] = v1[3]+v2[3];
	v3[4] = v1[4]+v2[4];
	v3[5] = v1[5]+v2[5];
    }
    else 
	for(i=0;i<dim;i++)
	    v3[i] = v1[i] + v2[i];
    return v3;
}

vector vectorSubtract(int dim, vector v1,vector v2,vector v3)
{
    int i;
    if(dim == 3){
	v3[0] = v1[0]-v2[0];
	v3[1] = v1[1]-v2[1];
	v3[2] = v1[2]-v2[2];
    }
    else if(dim ==6){
	v3[0] = v1[0]-v2[0];
	v3[1] = v1[1]-v2[1];
	v3[2] = v1[2]-v2[2];
	v3[3] = v1[3]-v2[3];
	v3[4] = v1[4]-v2[4];
	v3[5] = v1[5]-v2[5];
    }
    else
    for(i=0;i<dim;i++)
	v3[i] = v1[i] - v2[i];
    return v3;
}

int vectorEqual(int dim,vector v1,vector v2)
{
    int i;
    
    for(i=0;i<dim;i++)
	if(fabs(v1[i]-v2[i])>1.0e-12)
	    return 0;
    return 1;
}

matrix matrixSubtract(int dim, matrix m1,matrix m2,matrix m3)
{
    int i,j;
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++)
	    m3[i][j] = m1[i][j] - m2[i][j];
    return m3;
}

matrix matrixAdd(int dim, matrix m1,matrix m2,matrix m3)
{
    int i,j;
    for(i=0;i<dim;i++)
	for(j=0;j<dim;j++)
	    m3[i][j] = m1[i][j] + m2[i][j];
    return m3;
}

matrix vectorCrossVectorCross(vector a,vector b, matrix r)
{
    float azbz = -a[Z]*b[Z];
    float ayby = -a[Y]*b[Y];
    float axbx = -a[X]*b[X];
    
    r[0][0] = azbz +ayby;
    r[0][1] = a[Y] *b[X];
    r[0][2] = a[Z] *b[X];
    r[1][0] = a[X]*b[Y];
    r[1][1] = azbz+axbx;
    r[1][2] = a[Z] *b[Y];
    r[2][0] = a[X]*b[Z];
    r[2][1] = a[Y] *b[Z];
    r[2][2] = ayby + axbx;
    return r;
}

matrix vectorCrossAddMatrix(vector a,matrix m, matrix r)
{
    int i,j;
    if(r != m){
	for(i=0;i<3;i++)
	    for(j=0;j<3;j++) {
		r[i][j] = m[i][j];
	    }
    }
    r[0][1] -= a[Z];
    r[0][2]  += a[Y];
    r[1][2] -= a[X];
    r[1][0] += a[Z];
    r[2][0]  -= a[Y];
    r[2][1]  += a[X];
    return r;
}

void print_vector(int dim,vector v)
{
    int i;
    for(i=0;i<dim;i++)
	printf("%f\n",v[i]);
    return;
}

void print_matrix(int dim,matrix m)
{
    int i,j;
    for(i=0;i<dim;i++){
	for(j=0;j<dim;j++)
	    printf("  %f",m[i][j]);
	printf("\n");
    }
    return;
}

void print_matrix2(int dim1,int dim2,matrix m)
{
    int i,j;
    for(i=0;i<dim1;i++){
	for(j=0;j<dim2;j++)
	    printf("  %f",m[i][j]);
	printf("\n");
    }
    return;
}

void print_vector1(int dim,vector v)
{
    int i;
    for(i=1;i<=dim;i++)
	printf("%f\n",v[i]);
    return;
}

void print_matrix1(int dim1,int dim2,matrix m)
{
    int i,j;
    for(i=1;i<=dim1;i++){
	for(j=1;j<=dim2;j++)
	    printf("  %f",m[i][j]);
	printf("\n");
    }
    return;
}

vector axisAndAngleFromVector(vector v1,vector v2,vector origin,vector 
    axis,float *ang)
{
    float v11[3],v22[3],axis2;
    
    vectorSubtract(3,v1,origin,v11);
    vectorSubtract(3,v2,origin,v22);
    vectorNormalize(3,v11,v11);
    vectorNormalize(3,v22,v22);
    vectorCrossProduct(v11,v22,axis);
    axis2 = vectorDotProduct(3,axis,axis);
    *ang = asin(axis2);
    if(vectorDotProduct(3,v11,v22)<0.0)
	*ang =  3.14159 - *ang;
    if(axis2<1e-8)
	axis[2]=1.0;
    return axis;
}



   



