#ifndef VECTDEF
#define VECTDEF
#define INVERSE 1
#define STD 0

typedef float *vector;

typedef float *qvector;

typedef float **matrix;

typedef float dtuple[3];

/*  creating, freeing and copying vectors and matrices */

float * new_vector(int nl, int nh);

void free_vector(float *v, int nl,int nh);

vector copyVector(int dim, vector src);

float ** new_matrix(int nrl, int nrh, int ncl, int nch);

void free_matrix(float **m,int nrl,int nrh, int ncl,int nch);

matrix copyMatrix(int dim, matrix src);


/* scalar operations on vectors and matrices */
 
vector vectorScalarMult(int dim,vector v,float scalar,vector r);

matrix matrixScalarMult(int dim,matrix m,float scalar,matrix r);

vector vectorFill(int dim,vector v,float scalar);

matrix matrixFill(int dim,matrix m, float fill);
 

/*  arithmetic operations on vectors and matrices */

vector vectorAdd(int dim, vector v1,vector v2,vector v3);

vector vectorSubtract(int dim, vector v1,vector v2,vector v3);

float vectorDotProduct(int dim,vector v1,vector v2);

matrix vectorProduct(int dim,vector v1,vector v2,matrix m);

matrix matrixSubtract(int dim, matrix m1,matrix m2,matrix m3);

matrix matrixAdd(int dim, matrix m1,matrix m2,matrix m3);

matrix  matrixMult( int dim, int invflag1,int invflag2, matrix m1,matrix m2, matrix m3);

matrix identityMatrix(int dim,matrix dest);

matrix matrixAssign(int dim,matrix src,matrix dest);

vector vectorAssign(int dim,vector src,vector dest);

vector vectorNormalize(int dim,vector src,vector dest);

vector vectorNormal(vector v1,vector v2,vector v3,vector norm);

int vectorEqual(int dim,vector v1,vector v2);

int vectorFindMax(int dim,vector v);

float vectorDistance(int dim,vector v1,vector v2);

/* vector/vector and vector/matrix cross products */
matrix matrixFromVectorCross(vector v1,matrix m);

vector vectorCrossProduct(vector v1,vector v2,vector v3);

matrix vectorMatrixCrossProduct(vector v1,matrix m1,matrix m2);

matrix matrixVectorCrossProduct(matrix m1,vector v1,matrix m2);

matrix vectorCrossVectorCross(vector a,vector b, matrix r);

matrix vectorCrossAddMatrix(vector a,matrix m, matrix r);


/*  transforms and transposes */

matrix matrixTranspose(int dim,matrix src,matrix dest);

matrix tensorTransform(int dim,matrix xform,matrix tensor,matrix result);

vector matrixVectorMult(int dim,int invflag, matrix m1,vector v1,vector v2);

void print_vector(int dim,vector v);

void print_matrix(int dim,matrix m);

void print_matrix2(int dim1,int dim2,matrix m);

vector axisAndAngleFromVector(vector v1,vector v2,vector origin,vector axis,float *ang);
#endif
