#include "AliveServer.h"
#include "AliveClient.h"
#include "AliveDispatch.h"
#include "SocketBuf.h"
#include "GenericMotorControl.h"
#include "BaseCreature.h"
#include "MotorCreature.h"
#include "CreatureCache.h"
#include "ArgTypeArray_xdr.h"
#include <iostream.h>
#include <math.h>
#include <stdio.h>
#include <unistd.h>
#include <bstring.h>
//#define NETWORK_BANDWIDTH_TEST    // uncomment for network statistics
#ifdef NETWORK_BANDWIDTH_TEST
#endif

AliveClient *AliveDispatch::ac = NULL;
AliveServer *AliveDispatch::as = NULL;
CreatureCache *AliveDispatch::creature_cache = NULL;
SocketBuf *AliveDispatch::sb = NULL;
XDR AliveDispatch::xdrs;
int AliveDispatch::mcb_count = 0;
unsigned int AliveDispatch::mcb_xdr_repos = 0;
float AliveDispatch::time_between_syncs = 10.0;   // in seconds
struct timeval AliveDispatch::time_of_last_sync;

#ifdef NETWORK_BANDWIDTH_TEST
static int packets_per_second = 0;
static int bytes_per_second = 0;
static struct timeval last_time_stats_printed;
#endif

extern int continuousMode;  // defined in AliveApp/main.c++

float
AliveDispatch::timevalDifference(const struct timeval& tv2,
			     const struct timeval& tv1)
{
  return (float) (tv2.tv_sec - tv1.tv_sec) + (float) (tv2.tv_usec - tv1.tv_usec) * (float) (1.0e-6);
}

char *
AliveDispatch::longToDotFormat(int32_t addr)
{
  char *temp_s = new char[20];

  sprintf(temp_s, "%d.%d.%d.%d", ((addr & 0xff000000) >> 24),
	  ((addr & 0x00ff0000) >> 16), ((addr & 0x0000ff00) >> 8),
	  (addr & 0x000000ff));

  return temp_s;
}

int
AliveDispatch::getAliveDataType(SocketBuf *sb)
{
  static XDR xdrs;
  static char the_type;

  xdrmem_create(&xdrs,
		(caddr_t) sb->buf,
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  if (!xdr_char(&xdrs, &the_type))
    {
      cerr << "AliveDispatch::getAliveDataType: error reading data type" << endl;
      xdr_destroy(&xdrs);
      return -1;
    }
  xdr_destroy(&xdrs);
  return (int) the_type;
}

SocketBuf *
AliveDispatch::wrapConnectionAcknowledge(ConnectionAcknowledge *ca)
{
  static char ca_type = AliveDispatch::CLIENT_CONNECTION_ACKNOWLEDGE;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::wrapConnectionAcknowledge: error repositioning XDR stream" << endl;
      return NULL;
    }
  if (!xdr_char(&xdrs, &ca_type))
    {
      cerr << "AliveDispatch::wrapConnectionAcknowledge: error writing type" << endl;
      return NULL;
    }
  if (!xdr_ConnectionAcknowledge(&xdrs, ca))
    {
      cerr << "AliveDispatch::wrapConnectionAcknowledge: error writing ConnectionAcknowledge" << endl;
      return NULL;
    }
  sb->num_bytes = xdr_getpos(&xdrs);
  return sb;
}  

ConnectionAcknowledge *
AliveDispatch::getConnectionAcknowledge(SocketBuf *sb)
{
  static struct ConnectionAcknowledge *ca = NULL;
  static XDR xdrs;
  static char ca_type;

  if (!ca)
    {
      ca = new ConnectionAcknowledge;
    }
  
  xdrmem_create(&xdrs,
		(caddr_t) sb->buf,
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  if (!xdr_char(&xdrs, &ca_type))
    {
      cerr << "AliveDispatch::getConnectionAcknowledge: error reading type" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  if (!xdr_ConnectionAcknowledge(&xdrs, ca))
    {
      cerr << "AliveDispatch::getConnectionAcknowledge: error reading data" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  xdr_destroy(&xdrs);
  return ca;
}

void
AliveDispatch::initialize()
{
  if (!sb)
    {
      sb = new SocketBuf;
      sb->buf = new sm_byte[SB_MAX_UDP_PACKET_SIZE];
      xdrmem_create(&xdrs,
		    (caddr_t) sb->buf,
		    SB_MAX_UDP_PACKET_SIZE,
		    XDR_ENCODE);
#ifdef NETWORK_BANDWIDTH_TEST
      gettimeofday(&last_time_stats_printed);
#endif
    }
  gettimeofday(&time_of_last_sync);
}

int
AliveDispatch::setAliveServer(AliveServer *as)
{
  if ((AliveDispatch::as == NULL) && (AliveDispatch::ac == NULL))
    {
      initialize();
      AliveDispatch::as = as;
      return 0;
    }
  else
    {
      cerr << "Already have set the Alive Dispatch's server or client" << endl;
      return -1;
    }
}
  
int
AliveDispatch::setAliveClient(AliveClient *ac)
{
  if ((AliveDispatch::as == NULL) && (AliveDispatch::ac == NULL))
    {
      initialize();
      AliveDispatch::ac = ac;
      return 0;
    }
  else
    {
      cerr << "Already have set the Alive Dispatch's server or client" << endl;
      return -1;
    }
}
      
int
AliveDispatch::setCreatureCache(CreatureCache *cc)
{
  if (creature_cache == NULL)
    {
      creature_cache = cc;
      return 0;
    }
  else
    {
      cerr << "Already have set the Alive Dispatch's CreatureCache" << endl;
      return -1;
    }
}

SocketBuf *
AliveDispatch::pollForUpdate()
{
  SocketBuf *sb;
  
  if (as)
    sb = as->msc->readWithPollAndTimeout();
  else
    if (ac)
      sb = ac->udp_sc->readWithPollAndTimeout();
  return sb;
}
    
void
AliveDispatch::mainDispatch(UpdateStateData &usd)
{
  SocketBuf *sb;

  if (as)  // I am an ALIVE server
    {
      // accept up to 1 connection per tick
      as->acceptConnectionWithPoll();

      // Serve up to 1 request from each client per tick

      as->sm->rewindSockets();
      while (as->sm->nextSocket())
	{
	  sb = as->sm->readWithPollAndTimeout();
	  if (sb)
	    {
	      switch (getAliveDataType(sb))
		{
		case CREATURE_REGISTRY_REQUEST:
		  {
		    serviceCreatureRegistryRequest(sb);
		    delete sb;
		    break;
		  }
		  
		case CREATURE_DOWNLOAD_REQUEST:
		  {
		    serviceCreatureDownloadRequest(sb);
		    delete sb;
		    break;
		  }

		case CREATURE_DATABASE_SYNCHRONIZATION_REQUEST:
		  {
		    serviceCreatureDatabaseSynchronizationRequest(sb);
		    delete sb;
		    break;
		  }

		default:
		  {
		    cerr << "AliveDispatch::mainDispatch: " << 
		      "server received an invalid request " << getAliveDataType(sb) << endl;
		    delete sb;
		    break;
		  }
		}
	    }
	}
    }

  // Send synchronizations for all local master creatures if necessary
//  conditionalSendSyncForAllMasterCreatures();

  // Now service all multicast requests

  do
    {
      sb = AliveDispatch::pollForUpdate();
      if (sb)
	{
	  switch (getAliveDataType(sb))
	    {
	    case CLIENT_CONNECTION_ACKNOWLEDGE: /* ignored by everyone but client they're
						   destined for, and in that case they
						   are handled in a different place */
	      {
		delete sb;
		break;
	      }
	      
	    case CREATURE_REGISTRY_ACCEPTANCE:
	      {
		// ignore these if I am a server.
		// However, if I am a client,
		// I must go download this creature
		// if I don't have it locally (i.e.
		// I didn't just register it with the server).
		
		if (ac)
		  {
		    CreatureRegistryAcceptance *cra = getCreatureRegistryAcceptance(sb);
		    if (creature_cache->getCreatureFromRegistryNumber(cra->creature_id) == NULL)
		      {
			// must download it.
			
			CreatureDownloadRequest cdr;
			cdr.creature_id = cra->creature_id;
			downloadCreatureFromServer(&cdr);
		      }
		  }
		
		delete sb;
		break;
	      }
	      
	    case MCB_SERIALIZATION:
	      {
		dispatchMCBSeries(sb, usd);
		// following line is now done in MotorCreature::updateStateAt
//		delete sb;
		break;
	      }
	      
	    case MCB_SERIALIZATION_FOR_PARENT:
	      {
		dispatchMCBSeries(sb, usd, TRUE);
		// following line is now done in MotorCreature::updateStateAt
//		delete sb;
		break;
	      }
	      
	    case CREATURE_SYNCHRONIZATION_SERVICE:
	      {
		dispatchSyncToCreature(sb);
		break;
	      }

	    default:
	      {
		cerr << "AliveDispatch::mainDispatch: " <<
		  "got an invalid update: " << getAliveDataType(sb) << endl;
		delete sb;
		break;
	      }
	    }
	}
    }
  while (sb);
}

SocketBuf *
AliveDispatch::wrapCreatureRegistryRequest(CreatureRegistryRequest *crr)
{
  static char crr_type = AliveDispatch::CREATURE_REGISTRY_REQUEST;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::wrapCreatureRegistryRequest: error repositioning XDR stream" << endl;
      return NULL;
    }
  if (!xdr_char(&xdrs, &crr_type))
    {
      cerr << "AliveDispatch::wrapCreatureRegistryRequest: error writing type" << endl;
      return NULL;
    }
  if (!xdr_CreatureRegistryRequest(&xdrs, crr))
    {
      cerr << "AliveDispatch::wrapCreatureRegistryRequest: error writing data" << endl;
      return NULL;
    }
  sb->num_bytes = xdr_getpos(&xdrs);
  return sb;
}  
  
CreatureRegistryRequest *
AliveDispatch::getCreatureRegistryRequest(SocketBuf *sb)
{
  static struct CreatureRegistryRequest *crr = NULL;
  static XDR xdrs;
  static char crr_type;

  if (!crr)
    {
      crr = new CreatureRegistryRequest;
    }

  xdrmem_create(&xdrs,
		(caddr_t) sb->buf,
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  if (!xdr_char(&xdrs, &crr_type))
    {
      cerr << "AliveDispatch::getCreatureRegistryRequest: error reading type" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  if (!xdr_CreatureRegistryRequest(&xdrs, crr))
    {
      cerr << "AliveDispatch::getCreatureRegistryRequest: error reading data" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  xdr_destroy(&xdrs);
  return crr;
}

SocketBuf *
AliveDispatch::wrapCreatureRegistryAcceptance(CreatureRegistryAcceptance *cra)
{
  static char cra_type = AliveDispatch::CREATURE_REGISTRY_ACCEPTANCE;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::wrapCreatureRegistryAcceptance: error repositioning XDR stream" << endl;
      return NULL;
    }
  if (!xdr_char(&xdrs, &cra_type))
    {
      cerr << "AliveDispatch::wrapCreatureRegistryAcceptance: error writing type" << endl;
      return NULL;
    }
  if (!xdr_CreatureRegistryAcceptance(&xdrs, cra))
    {
      cerr << "AliveDispatch::wrapCreatureRegistryAcceptance: error writing data" << endl;
      return NULL;
    }
  return sb;
}  
  
CreatureRegistryAcceptance *
AliveDispatch::getCreatureRegistryAcceptance(SocketBuf *sb)
{
  static struct CreatureRegistryAcceptance *cra = NULL;
  static XDR xdrs;
  static char cra_type;

  if (!cra)
    {
      cra = new CreatureRegistryAcceptance;
    }

  xdrmem_create(&xdrs,
		(caddr_t) sb->buf,
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  if (!xdr_char(&xdrs, &cra_type))
    {
      cerr << "AliveDispatch::getCreatureRegistryAcceptance: error reading type" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  if (!xdr_CreatureRegistryAcceptance(&xdrs, cra))
    {
      cerr << "AliveDispatch::getCreatureRegistryAcceptance: error reading data" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  xdr_destroy(&xdrs);
  return cra;
}

SocketBuf *
AliveDispatch::wrapCreatureDownloadRequest(CreatureDownloadRequest *cdr)
{
  static char cdr_type = AliveDispatch::CREATURE_DOWNLOAD_REQUEST;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::wrapCreatureDownloadRequest: error repositioning XDR stream" << endl;
      return NULL;
    }
  if (!xdr_char(&xdrs, &cdr_type))
    {
      cerr << "AliveDispatch::wrapCreatureDownloadRequest: error writing type" << endl;
      return NULL;
    }
  if (!xdr_CreatureDownloadRequest(&xdrs, cdr))
    {
      cerr << "AliveDispatch::wrapCreatureDownloadRequest: error writing data" << endl;
      return NULL;
    }
  sb->num_bytes = xdr_getpos(&xdrs);
  return sb;
}  
  
CreatureDownloadRequest *
AliveDispatch::getCreatureDownloadRequest(SocketBuf *sb)
{
  static struct CreatureDownloadRequest *cdr = NULL;
  static XDR xdrs;
  static char cdr_type;

  if (!cdr)
    {
      cdr = new CreatureDownloadRequest;
    }

  xdrmem_create(&xdrs,
		(caddr_t) sb->buf,
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  if (!xdr_char(&xdrs, &cdr_type))
    {
      cerr << "AliveDispatch::getCreatureDownloadRequest: error reading type" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  if (!xdr_CreatureDownloadRequest(&xdrs, cdr))
    {
      cerr << "AliveDispatch::getCreatureDownloadRequest: error reading data" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  xdr_destroy(&xdrs);
  return cdr;
}

SocketBuf *
AliveDispatch::wrapCreatureDownloadService(CreatureDownloadService *cds)
{
  static char cds_type = AliveDispatch::CREATURE_DOWNLOAD_SERVICE;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::wrapCreatureDownloadService: error repositioning XDR stream" << endl;
      return NULL;
    }
  if (!xdr_char(&xdrs, &cds_type))
    {
      cerr << "AliveDispatch::wrapCreatureDownloadService: error writing type" << endl;
      return NULL;
    }
  if (!xdr_CreatureDownloadService(&xdrs, cds))
    {
      cerr << "AliveDispatch::wrapCreatureDownloadService: error writing data" << endl;
      return NULL;
    }
  sb->num_bytes = xdr_getpos(&xdrs);
  return sb;
}  
  
CreatureDownloadService *
AliveDispatch::getCreatureDownloadService(SocketBuf *sb)
{
  static struct CreatureDownloadService *cds = NULL;
  static XDR xdrs;
  static char cds_type;

  if (!cds)
    {
      cds = new CreatureDownloadService;
    }
  
  xdrmem_create(&xdrs,
		(caddr_t) sb->buf,
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  if (!xdr_char(&xdrs, &cds_type))
    {
      cerr << "AliveDispatch::getCreatureDownloadService: error reading type" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  if (!xdr_CreatureDownloadService(&xdrs, cds))
    {
      cerr << "AliveDispatch::getCreatureDownloadService: error reading data" << endl;
      xdr_destroy(&xdrs);
      return NULL;
    }
  xdr_destroy(&xdrs);
  return cds;
}

// Used by client to register a locally
// created creature with the server
int
AliveDispatch::registerCreatureWithServer(BaseCreature *bc)
{
  static CreatureRegistryRequest crr;

  
  if (!ac)
    return -1;

  const char *gfn = bc->geometryFileName.getValue().getString();
  bcopy(gfn, crr.creature_filename, strlen(gfn)+1);
  cerr << "Registering creature " << crr.creature_filename << " with server" << endl;
  SocketBuf *sb = wrapCreatureRegistryRequest(&crr);
  ac->tcp_sc->writeWithTimeout(sb);

  // now wait till we get the acceptance
  // this could (should?) be implemented to not block
  SocketBuf *sb1;
  CreatureRegistryAcceptance *cra;
  
  do
    {
      sb1 = ac->tcp_sc->readWithPollAndTimeout();
      if (sb1)
	{
	  if (getAliveDataType(sb1) == CREATURE_REGISTRY_ACCEPTANCE)
	    {
	      cra = getCreatureRegistryAcceptance(sb1);
	      cerr << "Got id " << cra->creature_id << " for this creature" << endl;
	      delete sb1;
	    }
	  else
	    {
	      cerr << "AliveDispatch::registerCreatureWithServer: got " <<
		"an unexpected packet" << endl;
	      delete sb1;
	      sb1 = NULL;
	    }
	}
    }
  while (sb1 == NULL);

  bc->registryNumber.setValue(cra->creature_id);
  return cra->creature_id;
}

// Server side companion for above
// See note inside about scheduling calls to this function
int
AliveDispatch::serviceCreatureRegistryRequest(SocketBuf *sb)
{
  CreatureRegistryRequest *crr;
  static CreatureRegistryAcceptance cra;

  if (!as)
    return -1;

  crr = getCreatureRegistryRequest(sb);
  cerr << "Got request for registering creature " << crr->creature_filename << endl;
  // This takes care of multicasting the
  // announcement of the creature's entering the world.
  // Below we tell the client that is registering with us
  // that he's gotten this number.
  BaseCreature *bc = creature_cache->addBaseCreatureFromFile(crr->creature_filename);
  // this field must be created in BaseCreature, I think
  int identity = bc->registryNumber.getValue();
  cra.creature_id = identity;
  SocketBuf *sb1 = wrapCreatureRegistryAcceptance(&cra);

  // note that this procedure must be called
  // immediately following reading in the CreatureRegistryRequest
  // in order that the AliveServer's SocketMan still be pointing
  // at the same client.
  as->sm->writeWithTimeout(sb1);

  // Since the client is running the behavior of the creature,
  // make this creature a droid
  if (bc->isOfType(MotorCreature::getClassTypeId()))
    ((MotorCreature *) bc)->motorCreatureType.setValue(MotorCreature::DROID);

  // return the new creature's identity number
  return identity;
}

// Server's function for broadcasting an announcement
// of a creature's entering the world
void
AliveDispatch::broadcastCreatureAnnouncement(BaseCreature *bc)
{
  if (!as)
    return;

  static CreatureRegistryAcceptance cra;
  cra.creature_id = bc->registryNumber.getValue();
  SocketBuf *sb = wrapCreatureRegistryAcceptance(&cra);
  as->msc->writeWithTimeout(sb);
}

// Client calls this function to download a creature from the server.
// Should be called when the server multicasts an announcement of
// entrance of a new creature (see serviceCreatureRegistryRequest)
// or when the client gets a MCB packet destined for a creature of
// which it has no knowledge.
// Returns -1 on failure, 0 if successful.
int
AliveDispatch::downloadCreatureFromServer(CreatureDownloadRequest *cdr)
{
  if (!ac)
    return -1;

  SocketBuf *sb = wrapCreatureDownloadRequest(cdr);
  ac->tcp_sc->writeWithTimeout(sb);
  
  // Now wait until the server sends us the creature back.
  // This probably should be made non-blocking sometime in the future

  SocketBuf *sb1;
  CreatureDownloadService *cds;

  do
    {
      sb1 = ac->tcp_sc->readWithPollAndTimeout();
      if (sb1)
	{
	  if (getAliveDataType(sb1) == CREATURE_DOWNLOAD_SERVICE)
	    {
	      cds = getCreatureDownloadService(sb1);
	      delete sb1;
	    }
	  else
	    {
	      cerr << "AliveDispatch::registerCreatureWithServer: got " <<
		"an unexpected packet" << endl;
	      delete sb1;
	      sb1 = NULL;
	    }
	}
    }
  while (sb1 == NULL);
  
  // add this creature to our local CreatureCache
  // Note we're using addRegisteredBaseCreatureFromFile here.
  BaseCreature *bc = creature_cache->addRegisteredBaseCreatureFromFile(cds->creature_filename,
								       cdr->creature_id);
  if (bc == NULL)
    {
      cerr << "AliveDispatch::downloadCreatureFromServer: error adding creature " << cdr->creature_id << " from file " <<
	cds->creature_filename << endl;
      return -1;
    }
  if (bc->isOfType(MotorCreature::getClassTypeId()))
    ((MotorCreature *) bc)->motorCreatureType.setValue(MotorCreature::DROID);
  return 0;
}

// Server calls this function to service the download requests that come in.
// NOTE that this function must be called immediately after reading in the
// CreatureDownloadRequest, so that the socket manager is still pointing
// at the correct client.
int
AliveDispatch::serviceCreatureDownloadRequest(SocketBuf *sb)
{
  if (!as)
    return -1;

  CreatureDownloadRequest *cdr = getCreatureDownloadRequest(sb);
  BaseCreature *bc = creature_cache->getCreatureFromRegistryNumber(cdr->creature_id);
  static CreatureDownloadService cds;

  const char *gfn = bc->geometryFileName.getValue().getString();
  bcopy(gfn, cds.creature_filename, strlen(gfn)+1);
  SocketBuf *sb1 = wrapCreatureDownloadService(&cds);
  as->sm->writeWithTimeout(sb1);
  return 0;
}

// Client calls this function to request that the server send back the
// contents of its creature database (creature identification numbers
// and geometry filenames).
// Currently called upon connection to the server.

int
AliveDispatch::synchronizeCreatureDatabaseWithServer()
{
  static char cdsr_type = AliveDispatch::CREATURE_DATABASE_SYNCHRONIZATION_REQUEST;

  if (!ac)
    return -1;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error repositioning XDR stream" << endl;
      return -1;
    }

  if (!xdr_char(&xdrs, &cdsr_type))
    {
      cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error writing type" << endl;
      return -1;
    }
  sb->num_bytes = xdr_getpos(&xdrs);
  ac->tcp_sc->writeWithTimeout(sb);

  // now wait for the server to come back with the creatures.
  // should probably be made non-blocking at some point in the future

  SocketBuf *sb1;

  do
    {
      sb1 = ac->tcp_sc->readWithPollAndTimeout();
      if (sb1)
	{
	  if (getAliveDataType(sb1) == CREATURE_DATABASE_SYNCHRONIZATION_SERVICE)
	    {
	      XDR xdrs;
	      xdrmem_create(&xdrs,
			    (caddr_t) sb1->buf,
			    SB_MAX_UDP_PACKET_SIZE,
			    XDR_DECODE);
	      
	      char cdss_type;
	      int num_creatures;

	      if (!xdr_char(&xdrs, &cdss_type))
		{
		  cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error reading type" << endl;
		  return -1;
		}

	      if (!xdr_int(&xdrs, &num_creatures))
		{
		  cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error reading number of creatures" << endl;
		  return -1;
		}

	      for (int counter = 0; counter < num_creatures; counter++)
		{
		  int creature_id;
		  char gfn[256];

		  if (!xdr_int(&xdrs, &creature_id))
		    {
		      cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error reading creature id" << endl;
		      return -1;
		    }
		      
		  if (!xdr_vector(&xdrs, (char *) gfn, 256, sizeof(char), (xdrproc_t) xdr_char))
		    {
		      cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error reading geometry file name" << endl;
		      return -1;
		    }
		  
		  // now add this creature to our cache if necessary

		  if (creature_cache->getCreatureFromRegistryNumber(creature_id) == NULL)
		    {
		      BaseCreature *bc = creature_cache->addRegisteredBaseCreatureFromFile(gfn,
											   creature_id);
		      if (bc == NULL)
			{
			  cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: error adding creature from file" << endl;
			  return -1;
			}

		      if (bc->isOfType(MotorCreature::getClassTypeId()))
			((MotorCreature *) bc)->motorCreatureType.setValue(MotorCreature::DROID);
		    }
		}
	      delete sb1;
	      xdr_destroy(&xdrs);
	    }
	  else
	    {
	      cerr << "AliveDispatch::synchronizeCreatureDatabaseWithServer: got " <<
		"an unexpected packet" << endl;
	      delete sb1;
	      sb1 = NULL;
	    }
	}
    }
  while (sb1 == NULL);
  return 0;
}


// Server calls this function to service database synchronization requests
// that come in. Right now these are only generated by clients when they
// connect to the server.

int
AliveDispatch::serviceCreatureDatabaseSynchronizationRequest(SocketBuf *)
{
  static char cdss_type = AliveDispatch::CREATURE_DATABASE_SYNCHRONIZATION_SERVICE;

  if (!as)
    return -1;

  cerr << "AliveDispatch::serviceCreatureDatabaseSynchronizationRequest: servicing sync request" << endl;

  // iterate over all the creatures we know about
  // and send their filenames to the client.

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::serviceCreatureDatabaseSynchronizationRequest: error repositioning XDR stream" << endl;
      return -1;
    }
  if (!xdr_char(&xdrs, &cdss_type))
    {
      cerr << "AliveDispatch::serviceCreatureDatabaseSynchronizationRequest: error writing type" << endl;
      return -1;
    }

  int creature_length = creature_cache->creatures.getLength();
  if (!xdr_int(&xdrs, &creature_length))
    {
      cerr << "AliveDispatch::serviceCreatureDatabaseSynchronizationRequest: error writing number of creatures" << endl;
      return -1;
    }

  for (int counter = 0; counter < creature_length; counter++)
    {
      BaseCreature *bc = (BaseCreature *) creature_cache->creatures[counter];
      const char *gfn = bc->geometryFileName.getValue().getString();
      long creature_id = bc->registryNumber.getValue();

      if (!xdr_long(&xdrs, &creature_id))
	{
	  cerr << "AliveDispatch::serviceCreatureDatabaseSynchronizationRequest: error writing id" << 
	    " for creature " << counter << endl;
	  return -1;
	}
      
      if (!xdr_vector(&xdrs, (char *) gfn, 256, sizeof(char), (xdrproc_t) xdr_char))
	{
	  cerr << "AliveDispatch::serviceCreatureDatabaseSynchronizationRequest: error writing filename" << 
	    " for creature " << counter << endl;
	  return -1;
	}
    }
  sb->num_bytes = xdr_getpos(&xdrs);
  as->sm->writeWithTimeout(sb);
  return 0;
}

// Called from mainDispatch each clock tick. Determines whether the time is right
// for sending out synchronization packets for all master creatures in the local
// CreatureCache and does so if necessary.

void
AliveDispatch::conditionalSendSyncForAllMasterCreatures()
{
  static struct timeval current_time;
  static char css_type = CREATURE_SYNCHRONIZATION_SERVICE;

  gettimeofday(&current_time);
  if (timevalDifference(current_time, time_of_last_sync) > time_between_syncs)
    {
      // send synchronization packets for all master creatures in my cache

      for (int i = 0; i < creature_cache->creatures.getLength(); i++)
	{
	  BaseCreature *bc = (BaseCreature *) creature_cache->creatures[i];
	  if (bc->isOfType(MotorCreature::getClassTypeId()))
	    {
	      MotorCreature *mc = (MotorCreature *) bc;

	      if (mc->motorCreatureType.getValue() == MotorCreature::MASTER)
		{
		  // send synchronization packet

		  long creature_id = mc->registryNumber.getValue();

		  cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: sending sync for creature " << creature_id << endl;

		  if (!xdr_setpos(&xdrs, 0))
		    {
		      cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: error repositioning XDR stream" << endl;
		      return;
		    }
		  
		  if (!xdr_char(&xdrs, &css_type))
		    {
		      cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: error writing type" << endl;
		      return;
		    }
		      
		  if (!xdr_long(&xdrs, &creature_id))
		    {
		      cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: error writing creature id" << endl;
		      return;
		    }
		  if (!xdr_float(&xdrs, &(mc->myCache->wallClock)))
		    {
		      cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: error writing tick number" << endl;
		      return;
		    }

		  if (!mc->syncToOrFromStream(&xdrs))
		    {
		      cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: error writing synchronization packet" << endl;
		      return;
		    }

		  sb->num_bytes = xdr_getpos(&xdrs);

		  if (as)
		    {
		      as->msc->writeWithTimeout(sb);
		      //      cerr << "Server multicasting MCB series" << endl;
		    }
		  else
		    {
		      if (ac)
			{
			  ac->udp_sc->writeWithTimeout(sb);
			  //	  cerr << "Client multicasting MCB series" << endl;
			}
		      else
			cerr << "AliveDispatch::conditionalSendSyncForAllMasterCreatures: unable to broadcast update" << endl;
		    }
		}
	    }
	}
      gettimeofday(&time_of_last_sync);
    }
}

void
AliveDispatch::dispatchSyncToCreature(SocketBuf *sb)
{
  static char css_type;
  static int32_t creature_id;
  static MotorCreature *mc;
  static float sync_tick;
  XDR xdrs;
  
  xdrmem_create(&xdrs,
		(caddr_t) sb->buf, 
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);

  if (!xdr_char(&xdrs, &css_type))
    {
      cerr << "AliveDispatch::dispatchSyncToCreature: unable to read type" << endl;
      xdr_destroy(&xdrs);
      return;
    }
  long creature_id_long = creature_id;
  if (!xdr_long(&xdrs, &creature_id_long))
    {
      cerr << "AliveDispatch::dispatchSyncToCreature: unable to read creature id" << endl;
      xdr_destroy(&xdrs);
      return;
    }
  else creature_id = creature_id_long;
  // The following cast is valid since the sync is only sent out for motor creatures
  mc = (MotorCreature *) creature_cache->getCreatureFromRegistryNumber(creature_id);

  if (mc == NULL)
    {
      cerr << "AliveDispatch::dispatchSyncToCreature: creature not in cache" << endl;
      xdr_destroy(&xdrs);
      return;
    }

  if (mc->motorCreatureType.getValue() == MotorCreature::MASTER)
    {
      // We were the source of this update, so throw it away
      xdr_destroy(&xdrs);
      return;
    }

  cerr << "AliveDispatch::dispatchSyncToCreature: dispatching sync to creature " << creature_id << endl;

  if (!xdr_float(&xdrs, &sync_tick))
    {
      cerr << "AliveDispatch::dispatchSyncToCreature: error reading sync tick number" << endl;
      return;
    }

  if (!mc->syncToOrFromStream(&xdrs))
    {
      cerr << "AliveDispatch::dispatchSyncToCreature: error reading sync into creature" << endl;
      xdr_destroy(&xdrs);
      return;
    }

  xdr_destroy(&xdrs);

  // now flush out the queued MCB updates from the creature if necessary.
  // They should not be applied now that the sync has occurred at this tick.
  // I'm doing some error checking here to make sure packets came in in the
  // correct order.

  if (mc->remoteMCBList.getLength() > 0)
    {
      cerr << "AliveDispatch::dispatchSyncToCreature: flushing MCB queue for creature " << creature_id << endl;
    }

  while (mc->remoteMCBList.getLength() > 0)
    {
      char mcb_type;
      float tick_number;
      
      // Pop the first one off the list

      SocketBuf *sb = (SocketBuf *) mc->remoteMCBList[0];
      mc->remoteMCBList.remove(0);
      
      // set up an XDR stream on this chunk of memory

      xdrmem_create(&xdrs,
		    (caddr_t) sb->buf, 
		    SB_MAX_UDP_PACKET_SIZE,
		    XDR_DECODE);

      if (!xdr_char(&xdrs, &mcb_type))  // ignored here
	{
	  cerr << "AliveDispatch::dispatchSyncToCreature: error reading MCB type" << endl;
	  xdr_destroy(&xdrs);
	  return;
	}
      long creature_id_long = creature_id;
      if (!xdr_long(&xdrs, &creature_id_long))  // ignored here
	{
	  cerr << "AliveDispatch::dispatchSyncToCreature: error getting MCB creature id" << endl;
	  xdr_destroy(&xdrs);
	  return;
	}
       else creature_id = creature_id_long;
      // check to see whether packets came in in the wrong order
		
      if (!xdr_float(&xdrs, &tick_number))
	{
	  cerr << "AliveDispatch::dispatchSyncToCreature: error reading MCB tick number" << endl;
	  return;
	}
      else
	{
	  if (tick_number >= sync_tick)
	    {
	      // This means that one of the queued MCBs has the same or greater clock tick
	      // than the synchronization packet. This is bad news and should not happen
	      // unless packets are coming in in the wrong order

	      cerr << "AliveDispatch::dispatchSyncToCreature: (nonfatal) error while flushing stored packets:" << endl <<
		"sync was from tick " << sync_tick << ", MCB tick was " << tick_number << endl;
	    }
	}
      xdr_destroy(&xdrs);
    }
}

void
AliveDispatch::beginMCBSerialization(BaseCreature *bc)
{
  static int32_t mcb_creature_id;
  static char mcb_type = AliveDispatch::MCB_SERIALIZATION;

  if (!xdr_setpos(&xdrs, 0))
    {
      cerr << "AliveDispatch::beginMCBSerialization: error repositioning XDR stream" << endl;
      return;
    }

  if (!xdr_char(&xdrs, &mcb_type))
    {
      cerr << "AliveDispatch::beginMCBSerialization: error writing type" << endl;
      return;
    }

  mcb_count = 0;
  
  mcb_creature_id = bc->registryNumber.getValue();
  long mcb_creature_id_long = mcb_creature_id;
  if (!xdr_long(&xdrs, &mcb_creature_id_long))
    {
      cerr << "AliveDispatch::beginMCBSerialization: error serializing creature ID" << endl;
      return;
    }
  else mcb_creature_id = mcb_creature_id_long;
  if (!xdr_float(&xdrs, &(bc->myCache->wallClock)))
    {
      cerr << "AliveDispatch::beginMCBSerialization: error writing tick number" << endl;
      return;
    }

  mcb_xdr_repos = xdr_getpos(&xdrs);
  if (!xdr_int(&xdrs, &mcb_count))
    {
      cerr << "AliveDispatch::beginMCBSerialization: error initializing number of MCBs" << endl;
      return;
    }
}

int
AliveDispatch::serializeMCB(MotorCommandBlock *mcb)
{
//  cerr << "Number of args to " << mcb->cmd << ": " << mcb->getArgCount() << endl;
//  for (int foo = 0; foo < mcb->getArgCount(); foo++)
//    {
//      cerr << "arg " << foo << ": ";
//      mcb->arglist.args[foo].print();
//    }
      
  if (!xdr_MotorCommandBlock(&xdrs, mcb))
    return (FALSE);
  mcb_count++;
  return (TRUE);
}

void
AliveDispatch::endMCBSerialization()
{
  sb->num_bytes = xdr_getpos(&xdrs);
#ifdef NETWORK_BANDWIDTH_TEST
  bytes_per_second += sb->num_bytes;
  packets_per_second++;
  struct timeval current_time;
  gettimeofday(&current_time);
  if (timevalDifference(current_time, last_time_stats_printed) > 1.0)
    {
      cerr << "MCB network stats last second: " <<
	bytes_per_second << " bytes, " <<
	  packets_per_second << " packets (maximum " << 8192*packets_per_second << ")" << endl;
      bytes_per_second = 0;
      packets_per_second = 0;
      gettimeofday(&last_time_stats_printed);
    }
#endif
  if (!xdr_setpos(&xdrs, mcb_xdr_repos))
    {
      cerr << "AliveDispatch::endMCBSerialization: error repositioning XDR stream" << endl;
    }
  else
    {
      if (!xdr_int(&xdrs, &mcb_count))
	{
	  cerr << "AliveDispatch::endMCBSerialization: error writing number of MCBs" << endl;
	}
    }

  if (as)
    {
      as->msc->writeWithTimeout(sb);
//      cerr << "Server multicasting MCB series" << endl;
    }
  else
    {
      if (ac)
	{
	  ac->udp_sc->writeWithTimeout(sb);
//	  cerr << "Client multicasting MCB series" << endl;
	}

      else
	cerr << "AliveDispatch::endMCBSerialization: unable to broadcast update" << endl;
    }
}

void
AliveDispatch::sendMCBToParentCreature(BaseCreature *bc,
				       MotorCommandBlock *mcb)
{
  static SocketBuf *sb = NULL;  /* need this here because this
				   function may get called in the
				   middle of another creature's behavior
				   system update  */
  static XDR xdrs;
  static char mcb_type = AliveDispatch::MCB_SERIALIZATION_FOR_PARENT;
  static int32_t mcb_creature_id;
  static int mcb_count = 1;
  static MotorCreature *mc;

  if (!sb)
    {
      sb = new SocketBuf;
      sb->buf = new sm_byte[SB_MAX_UDP_PACKET_SIZE];
      xdrmem_create(&xdrs,
		    (caddr_t) sb->buf,
		    SB_MAX_UDP_PACKET_SIZE,
		    XDR_ENCODE);
    }
  else
    {
      if (!xdr_setpos(&xdrs, 0))
	{
	  cerr << "AliveDispatch::sendMCBToParentCreature: error repositioning XDR stream" << endl;
	}
    }

  if (!xdr_char(&xdrs, &mcb_type))
    {
      cerr << "AliveDispatch::sendMCBToParentCreature: error writing type" << endl;
      return;
    }

  mcb_creature_id = bc->registryNumber.getValue();
  long mcb_creature_id_long = mcb_creature_id;
  if (!xdr_long(&xdrs, &mcb_creature_id_long))
    {
      cerr << "AliveDispatch::sendMCBToParentCreature: error serializing creature ID" << endl;
      return;
    }
  else mcb_creature_id = mcb_creature_id_long;
  if (!xdr_float(&xdrs, &(bc->myCache->wallClock)))
    {
      cerr << "AliveDispatch::sendMCBToParentCreature: error writing tick number" << endl;
      return;
    }

  if (!xdr_int(&xdrs, &mcb_count))
    {
      cerr << "AliveDispatch::sendMCBToParentCreature: error settng number of MCBs" << endl;
      return;
    }
  
  if (!xdr_MotorCommandBlock(&xdrs, mcb))
    {
      cerr << "AliveDispatch::sendMCBToParentCreature: error serializing MCB" << endl;
      return;
    }

  sb->num_bytes = xdr_getpos(&xdrs);

  // for efficiency, check to see if the passed creature is the desired
  // master. Avoids making network traffic unnecessarily.

  if (bc->isOfType(MotorCreature::getClassTypeId()) == FALSE)
    {
      cerr << "AliveDispatch::sendMCBToParentCreature: creature " << mcb_creature_id << " is not a motor creature" << endl;
      return;
    }
      
  mc = (MotorCreature *) bc;

  // if we're the server for this creature,
  // put this buffer on the creature's MCB queue.
  // Note that this involves a bcopy, because
  // the creature will destroy this buffer
  // after it's done executing the MCBs inside
  
  if (mc->motorCreatureType.getValue() == MotorCreature::MASTER)
    {
      SocketBuf *sb2 = new SocketBuf;
      sb2->buf = new sm_byte[SB_MAX_UDP_PACKET_SIZE];
      sb2->num_bytes = sb->num_bytes;
      bcopy(sb->buf, sb2->buf, SB_MAX_UDP_PACKET_SIZE);
      mc->remoteMCBList.append(sb2);
      return;
    }

  if (as)
    {
      as->msc->writeWithTimeout(sb);
//      cerr << "Server multicasting MCB series" << endl;
    }
  else
    {
      if (ac)
	{
	  ac->udp_sc->writeWithTimeout(sb);
//	  cerr << "Client multicasting MCB series" << endl;
	}

      else
	cerr << "AliveDispatch::sendMCBToParentCreature: unable to broadcast update" << endl;
    }
}

void
AliveDispatch::flushMCB(MotorCommandBlock *mcb)
{
  // delete all char * pointers in ArgTypes
  // to avoid memory leaks/bashing

  for (int counter = 0; counter < MAXARGS; counter++)
    {
      if (mcb->arglist.args[counter].getType() == ArgType::CHARPTR)
	{
	  xdr_free((xdrproc_t) xdr_wrapstring, &(mcb->arglist.args[counter].internal.ArgTypeInternal_u.cptr));
	}
    }
}
	  
// Dispatch the MCBs to the appropriate creature.
// If we don't know about this creature,
// download it from the server first.
void
AliveDispatch::dispatchMCBSeries(SocketBuf *sb,
				 UpdateStateData &usd,
				 int parent_only)
{
  static int32_t creature_id;
  static MotorCommandBlock mcb;
  static char mcb_type;

  XDR xdrs;
  BaseCreature *bc;
  MotorCreature *mc;
  xdrmem_create(&xdrs,
		(caddr_t) sb->buf, 
		SB_MAX_UDP_PACKET_SIZE,
		XDR_DECODE);
  
  if (!xdr_char(&xdrs, &mcb_type))
    {
      cerr << "AliveDispatch::dispatchMCBSeries: error reading type" << endl;
      xdr_destroy(&xdrs);
      return;
    }
  long creature_id_long = creature_id;
  if (!xdr_long(&xdrs, &creature_id_long))
    {
      cerr << "AliveDispatch::dispatchMCBSeries: error getting creature id" << endl;
      xdr_destroy(&xdrs);
      return;
    }
  else creature_id = creature_id_long;
  // get a pointer to this creature
  
  bc = creature_cache->getCreatureFromRegistryNumber(creature_id);
  if (!bc)
    {
      // download this creature
      CreatureDownloadRequest cdr;
      cdr.creature_id = creature_id;
      if (downloadCreatureFromServer(&cdr) < 0)
	{
	  cerr << "AliveDispatch::dispatchMCBSeries: unable to get creature " <<
	    creature_id << " for dispatch" << endl;
	  xdr_destroy(&xdrs);
	  return;
	}
    }

  // make sure this creature is a motor creature
  
  if (bc->isOfType(MotorCreature::getClassTypeId()) == FALSE)
    {
      cerr << "AliveDispatch::dispatchMCBSeries: creature " << creature_id << " is not a motor creature" << endl;
      xdr_destroy(&xdrs);
      return;
    }
      
  mc = (MotorCreature *) bc;

  // if we're the server for this creature,
  // ignore the MCB updates coming in for it,
  // unless this is a motor command destined
  // specifically for thie creature
  
  if (mc->motorCreatureType.getValue() == MotorCreature::MASTER)
    {
      if (!parent_only)
	{
	  xdr_destroy(&xdrs);
	  return;
	}
    }

  // put this buffer on the queue for the creature
  mc->remoteMCBList.append(sb);

  // destroy the XDR stream
  xdr_destroy(&xdrs);
}

