#include <stdio.h>
#include <iostream.h>
#include <unistd.h>
#include <bstring.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#define _BSD_SIGNALS
#include <signal.h>

#include "AliveServer.h"
#include "Alive.h"

static SocketMan *usm = NULL;
static SocketClient *msc2 = NULL;

AliveServer::AliveServer()
{
  sm = new SocketMan(ALIVE_TCP_PORT);
  msc = new SocketClient(ALIVE_MULTICAST_PORT);
  msc->setUsingMulticast(1);
  msc->connectToServer(ALIVE_GROUP);
  msc->setSuppressingMulticastLoopback(0);
}

AliveServer::~AliveServer()
{
  delete sm;
  delete msc;
}

void
cleanupPacketForwarder(...)
{
  if (usm) delete usm;
  if (msc2) delete msc2;
  cerr << "Packet forwarder cleaned up" << endl;
  exit(1);
}

void
runPacketForwarder(void *data)
{
  // for rebroadcasting multicast updates using UDP
  fd_set readfds;
  int ret;
  int width;

  // find the maximum number of open file descriptors
#ifndef __hpux
  width = getdtablesize();
#else
  width = FD_SETSIZE;
#endif

  usm = new SocketMan(ALIVE_UDP_PORT);
  usm->setUsingUDP(1);
  msc2 = new SocketClient(ALIVE_MULTICAST_PORT);
  msc2->setUsingMulticast(1);
  msc2->connectToServer(ALIVE_GROUP);
  msc2->setSuppressingMulticastLoopback(0);
  signal(SIGTERM, cleanupPacketForwarder);
  signal(SIGINT, cleanupPacketForwarder);

  while(1)
    {
      FD_ZERO(&readfds);
      usm->setFDSet(&readfds);
      msc2->setFDSet(&readfds);
      
      // go into blocking select on multicast socket and UDP SocketMan
      if ((ret = select(width, &readfds, NULL, NULL, NULL)) == -1)
	{
	  perror("runPacketForwarder");
	}
      else
	{
	  ret = usm->getPostSelectReadAction(&readfds);
	  if (ret == SocketMan::MSG_SS_ACCEPT_CONNECTION)
	    {
//	      cerr << "runPacketForwarder: accepting UDP connection" << endl;
	      usm->acceptConnectionWithPoll();
	    }
	  else
	    {
	      if (ret == SocketMan::MSG_SS_READ_FROM_CURRENT_SOCKET)
		{
		  SocketBuf *sb = usm->readWithPollAndTimeout();
		  if (sb)
		    {
//		      cerr << "runPacketForwarder: multicasting packet from client" << endl;
		      // multicast this packet
		      msc2->writeWithTimeout(sb);
		      delete sb;
		    }
		}
	      else
		{
		  ret = msc2->readyToRead(&readfds);
		  if (ret == SocketClient::MSG_SOCKET_READY_TO_READ)
		    {
		      SocketBuf *sb = msc2->readWithPollAndTimeout();
		      if (sb)
			{
//			  cerr << "Rebroadcasting multicast packet to clients" << endl;
			  // rebroadcast this packet to all connected clients
			  usm->rewindSockets();
			  while (usm->nextSocket())
			    {
			      usm->writeWithTimeout(sb);
			    }
			  
			  delete sb;
			}
		    }
		}
	    }
	}
    }
}

int
AliveServer::threadPacketForwarder()
{
  if (sproc(runPacketForwarder, NULL) != -1)
    {
      prctl(PR_SETEXITSIG, SIGTERM);
      signal(SIGTERM, cleanup);
      return 0;
    }
  return -1;
}

int
AliveServer::acceptConnectionWithPoll()
{
  int status;

  if ((status = sm->acceptConnectionWithPoll()) == SocketMan::MSG_NO_ERROR)
    {
      // accepted a TCP connection.
      // need to get the client's address so we
      // can multicast a client connection acknowledgement.

      ConnectionAcknowledge ca;
      ca.connection_address = sm->last_accepted_socket->address;
      SocketBuf *sb = AliveDispatch::wrapConnectionAcknowledge(&ca);
      if (msc->writeWithTimeout(sb) != SocketClient::MSG_NO_ERROR)
	{
	  cerr << "Error writing client connection acknowledgement" << endl;
	}
    }
  return status;
}


