#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include "Dog.h"
#include "DogMotorSkills.h"
#include "CreatureLib.h"
#include "DogController.h"
#include "../MotorLib/BasicDefines.h"

// Change this to 1 if you want a faster but more glidy walk
#define USEAVERAGESPEED 0

SO_NODE_SOURCE(Dog);

void
Dog::initClass()
{
    SO_NODE_INIT_CLASS(Dog, MotorCreature, "MotorCreature");
}

Dog::Dog() : MotorCreature()
{
    SO_NODE_CONSTRUCTOR(Dog);
    SO_NODE_ADD_FIELD(newStyleNeck, (TRUE));
}

Dog::~Dog()
{
}

GenericController *Dog::addController()
{
    return new GenericController(this);
}

void Dog::addMotorSkills()
{
    useRealTime.setValue(FALSE);
    turnBase.setValue(11.0);
    
    MotorCreature::addMotorSkills();

    Walking *w = new Walking(this, "walking", 16.0, defaultFwdSpeed.getValue(),USEAVERAGESPEED);
    Turning *t = new Turning(this, "turning", .1);
    WaggingAndMovingTail *wmt = new WaggingAndMovingTail(this,"moving tail",  .1);
    Sitting *sit = new Sitting(this,"sitting",  .1);
    LoweringAndRaisingHead *lrh = new LoweringAndRaisingHead(this,"moving head",  .1);
    ShakingAndTurningHead *sth = new ShakingAndTurningHead(this,"shaking head",  .1);
    RotatingEars *re = new RotatingEars(this,"moving ears",  .1);
    Sniffing *sniff = new Sniffing(this,"sniffing", 1.0);
    MovingEyeBrows *meb = new MovingEyeBrows(this,"moving eyebrows",  1.0);
    ShiftingEyes *se = new ShiftingEyes(this, "shifting eyes", 1.0);
    ShakePaw *sp = new ShakePaw(this,"shaking",  .1);
    Begging *beg = new Begging(this,"begging",  .1);
    TiltingHead *tilt = new TiltingHead(this,"tilting head",  .1);
    OpeningAndClosingMouth *ocm = new OpeningAndClosingMouth(this,"mouth",  .1);
    Bark *bark = new Bark(this,"bark",  .2);
    NoddingAndLoweringHead *nlh = new NoddingAndLoweringHead(this,"nodding",  0.1);
    Crouch *crouch = new Crouch(this,"crouching",  0.1);
    LyingDown *lyingDown = new LyingDown(this,"lying down",  0.1);
    LiftLeg *liftleg = new LiftLeg(this, "liftleg", .1);
    BugEyes *bugeyes = new BugEyes(this, "bugeyes", .1);
    ShowTeeth *showTeeth = new ShowTeeth(this, "showteeth", .1);
    LookAtAB *lookAtAB;
    ShakeHeadAB *shakeHeadAB;
    NodHeadAB *nodHeadAB;
    LookAt *lookAt;
    if(newStyleNeck.getValue()){
	lookAtAB = new LookAtAB(this,"looking at-ab",  0.1);
	shakeHeadAB = new ShakeHeadAB(this,"shakinghead-ab",  0.1);
	nodHeadAB = new NodHeadAB(this,"nodhead-ab",  0.1);
    }
    else
	lookAt = new LookAt(this,"looking at",  0.1);
    /*
     *  first over-rides of generic commands
     */
    getController()->addCommand( GC_MOVE_GEN,"move",w,TRUE, 
				 ArgType((float)1), 
				 ArgType((float)DOG_GAIT_WALK));
    getController()->addCommand( GC_TURN_GEN ,	      "turn", 		   t, TRUE, ArgType((float)DEFAULT_TURN), ArgType((float)DEFAULT_SPEED));
    
    /*
     * now unique commands
     */
     
    getController()->addCommand( GC_FWD,              "walk", 		   w,  1, ArgType((float)1), ArgType((float)DOG_GAIT_WALK));
    getController()->addCommand( GC_BACK,             "back", 	           w, 1, ArgType((float)-1), ArgType((float)DOG_GAIT_WALK));
    getController()->addCommand( GC_BOUND,            "bound", 	           w, 1, ArgType((float)1), ArgType((float)DOG_GAIT_BOUND));
    getController()->addCommand( GC_RGT ,	      "right", 		   t, 1, ArgType((float)DEFAULT_TURN), ArgType((float)DEFAULT_SPEED));
    getController()->addCommand( GC_LFT  ,            "left", 		   t, 1,ArgType((float)(1-DEFAULT_TURN)), ArgType((float)DEFAULT_SPEED));
    getController()->addCommand( GC_STRAIGHT,       "straight",           t, 1, ArgType((float)DEFAULT_STRAIGHT), ArgType((float)(DEFAULT_SPEED/2.0)));
    getController()->addCommand( GC_STOP ,            "stop", 		   w, 0, ArgType((float)1), ArgType((float)DOG_GAIT_WALK));
    getController()->addCommand( GC_WAGTAIL ,   	     "wagtail",       	   wmt,1, ArgType((float)DEFAULT_SPEED), ArgType((float)0), ArgType((float)0));
    getController()->addCommand( GC_STOPTAIL ,        "stoptail",	   wmt,0);
    getController()->addCommand( GC_MOVETAIL ,        "movetail",	   wmt, 1,ArgType((float)0), ArgType((float)1), ArgType((float)DEFAULT_SPEED) );
    getController()->addCommand( GC_SIT  ,            "sit",		   sit, 1, ArgType((float)DEFAULT_SIT), ArgType((float)(DEFAULT_SPEED*2)));
    getController()->addCommand( GC_STAND ,     	     "endsitting",              sit, 1, ArgType((float)DEFAULT_STAND), ArgType((float)(DEFAULT_SPEED*2))) ;
    if(newStyleNeck.getValue())
	getController()->addCommand( GC_SHAKEHEAD ,       "shakehead",          shakeHeadAB, 1, ArgType((float)1.0));
    else
	getController()->addCommand( GC_SHAKEHEAD ,       "shakehead",          sth, 1, ArgType((float)DEFAULT_SPEED),ArgType((float)0) );
    getController()->addCommand( GC_STOPSHAKE ,       "stopshake",          sth, 0);
    getController()->addCommand( GC_TURNHEAD ,  	     "turnhead",  	   sth, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)1), ArgType((float)0));
    getController()->addCommand( GC_STOPTURNHEAD ,     "stopturnhead",  	   sth, 0, ArgType((float)DEFAULT_SPEED), ArgType((float)0), ArgType((float)0));	
    getController()->addCommand( GC_LOWERNECK ,        "lowerhead", 	   lrh, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)1));
    getController()->addCommand( GC_RAISENECK,         "raisehead",	   lrh, 0);
    getController()->addCommand( GC_MOVEEARS ,        "moveears", 	   re, 1, 
				    ArgType((float)DEFAULT_SPEED),  // speed
				    ArgType((float)0.0),	    // ear position
				    ArgType((float)1.0));	    // tilted forward
    getController()->addCommand( GC_SNIFF ,           "sniff", 		   sniff, 1, ArgType((float)1.0));
    getController()->addCommand( GC_STOPSNIFF,            "stopsniff",          sniff, 0);	
    getController()->addCommand( GC_MOVEBROWS ,       "movebrows", 	   meb, 1, ArgType((float)DEFAULT_SPEED), ArgType((float).5), ArgType((float)1.0));
    getController()->addCommand( GC_SHIFTEYES ,       "shifteyes", 	   se, 1, ArgType((float)1.0));
    getController()->addCommand( GC_TILTHEAD ,        "tilthead", 	   tilt, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)0) );
    getController()->addCommand( GC_STOPTILTHEAD ,        "stoptilthead", 	   tilt, 0);	
    getController()->addCommand( GC_CROUCH ,          "crouch", 		   crouch, 1);
    getController()->addCommand( GC_STANDFROMCROUCH , "stopcrouch", 	   crouch, 0);
    getController()->addCommand( GC_SHAKEPAW ,        "shakepaw", 	   sp, 1);
    getController()->addCommand( GC_STOPSHAKEPAW ,    "stopshakepaw", 	   sp, 0);
    getController()->addCommand( GC_BEG ,	     "beg",          	   beg, 1, ArgType((float)DEFAULT_SIT), ArgType((float)DEFAULT_SPEED));
    getController()->addCommand( GC_STANDFROMBEG	 ,   "stopbegging",  	   beg, 1, ArgType((float)DEFAULT_STAND), ArgType((float)DEFAULT_SPEED));
    getController()->addCommand( GC_MOVEMOUTH ,       "movemouth",    	   ocm, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)1), ArgType((float)0));
    getController()->addCommand( GC_CHEW , 	     "chew", 	     	   ocm,  1, ArgType((float)DEFAULT_SPEED), ArgType((float)1), ArgType((float)1) );
    if(newStyleNeck.getValue())
	getController()->addCommand( GC_NOD ,       "nodhead",          nodHeadAB, 1, ArgType((float).1));
    else
	getController()->addCommand( GC_NOD , 	     "nod", 	     	   nlh, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)1) );
    getController()->addCommand( GC_MOVEHEADZ , 	     "moveheadz",    	   nlh, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)0) );
    getController()->addCommand( GC_STOPMOVEHEADZ ,"stopmovez",nlh, 0);
    getController()->addCommand( GC_LIEDOWN ,         "liedown",      	   lyingDown, 1, ArgType((float)DEFAULT_SPEED), ArgType((float)1) );
    getController()->addCommand( GC_STANDFROMLIEDOWN, "standfromlyingdown", lyingDown,0 );
    if(newStyleNeck.getValue())
	getController()->addCommand( GC_LOOKAT, "lookat", lookAtAB,1, 
				     ArgType((float)1),  
				     ArgType((float)0), 
				     ArgType((float)0), ArgType((float).1));
    else
	getController()->addCommand( GC_LOOKAT, "lookat", lookAt,1, 
				     ArgType((float)1),  
				     ArgType((float)0), 
				     ArgType((float)0), ArgType((float).1));
    getController()->addCommand( GC_LIFTLEG, "liftleg", liftleg,1, 
				 ArgType((float)0), 
				 ArgType((float).1));
    getController()->addCommand( GC_LOWERLEG, "lowerleg", liftleg,0, \
				 ArgType((float)0), 
				 ArgType((float).1));
    getController()->addCommand( GC_BUGEYES, "bugeyes", bugeyes,1, \
				 ArgType((float).1), 
				 ArgType((float)0));
    getController()->addCommand( GC_SHOWTEETH, "showteeth", showTeeth,1, \
				 ArgType((float).1), 
				 ArgType((float)0));
    getController()->addCommand( GC_BARK, "bark", bark,1, \
				 ArgType((float).2), 
				 ArgType((float)1.0), 
				 ArgType((char *)NULL));
    getController()->addCompoundCommand( GC_LFTFWD, "leftforward", GC_LFT, GC_FWD);
    getController()->addCompoundCommand( GC_RGTFWD, "rightforward", GC_RGT, GC_FWD);
    getController()->addCompoundCommand( GC_GOSTRAIGHT, "gostraight", GC_STRAIGHT, GC_FWD);
    getController()->addCompoundCommand( GC_GOSTRAIGHTBOUND, "straightbound", GC_STRAIGHT, GC_BOUND);
    getController()->addCompoundCommand( GC_HALT ,"halt", GC_STOP, GC_STRAIGHT);
    getController()->addCompoundCommand( GC_STANDALL ,"stand", GC_STAND, GC_STANDFROMCROUCH,GC_STANDFROMBEG );

}

void Dog::getDOFInterpolators()
{
    // NOTE !!!
    // Order must agree with values in DogInterp.h 
    // NOTE !!!
    
    /* this gets the leg dofs */
    ArticulatedBody *ab =  (ArticulatedBody *)getByName("FLLEG");
    if(ab){
	BSInterpV3f *foo = (BSInterpV3f *)getConnectedEngine(&ab->goal);
	dofs.append(new Dof(foo, ab));
    }
    
    ab =  (ArticulatedBody *)getByName("FRLEG");
    if(ab){
	BSInterpV3f *foo = (BSInterpV3f *)getConnectedEngine(&ab->goal);
	dofs.append(new Dof(foo, ab));
    }
    
    ab =  (ArticulatedBody *)getByName("RLLEG");
    if(ab){
	BSInterpV3f *foo = (BSInterpV3f *)getConnectedEngine(&ab->goal);
	dofs.append(new Dof(foo, ab));
    }
    
    ab =  (ArticulatedBody *)getByName("RRLEG");
    if(ab){
	BSInterpV3f *foo = (BSInterpV3f *)getConnectedEngine(&ab->goal);
	dofs.append(new Dof(foo, ab));
    }
    
    
    /* now get the shoulder interpolator */
    installXYZInterpolator("SHOULDER_JOINT");
     
    /* now get the front-back interpolator */
    installXYZInterpolator("FRONT_BACK_JOINTY");
    installXYZInterpolator("FRONT_BACK_JOINTZ");
    
    /* now get the back-back interpolator */
    installXYZInterpolator("BACK_BACK_JOINTY");
    installXYZInterpolator("BACK_BACK_JOINTZ");
    
    /* get tail */
    installXYZInterpolator("TAIL_JOINTX");
    installXYZInterpolator("TAIL_JOINTZ");
    
    /* neck and head */
    SoInterpolateVec3f *svf;
    if(newStyleNeck.getValue()){
	SoTransform *shn =  (SoTransform *)getByName("HEADJOINT");
	svf = (SoInterpolateVec3f *)getConnectedEngine(&shn->translation);
	dofs.append(new Dof(svf, shn));
    }
    else
	installXYZInterpolator("NECK_JOINTZ");
    
    installXYZInterpolator("HEAD_JOINTY");
    
    installXYZInterpolator("HEAD_JOINT");
    
    /* eyeballs and eye-brows */
    SoPendulum *p ; //= (SoPendulum *)getByName("EYE_JOINT");
    SoTransform *eye =  (SoTransform *)getByName("EYE_JOINT");
    SoInterpolateVec3f *sv3f = (SoInterpolateVec3f *)getConnectedEngine(&eye->translation);
    dofs.append(new Dof(sv3f, eye));
    
//    dofs.append(new Dof(p));
    
    SoShuttle *shut = (SoShuttle *)getByName("EYEBROW_SHUTTLE");
    dofs.append(new Dof(shut));
    
    installXYZInterpolator("EYEBROWR");
    installXYZInterpolator("EYEBROWL");
    
    /* ears */
    installXYZInterpolator("EARL");
    installXYZInterpolator("EARR");
    
    /* mouth */
    SoTransform *sh =  (SoTransform *)getByName("MOUTH_JOINT");
    SoInterpolateRotation *sif = (SoInterpolateRotation *)getConnectedEngine(&sh->rotation);
    dofs.append(new Dof(sif, sh));
    
    /* nose */
    p = (SoPendulum *)getByName("SNIFF_JOINT");
    dofs.append(new Dof(p));

    /* shoulder height transform */
    SoTransform *sxf =(SoTransform *)getByName("SHOULDERHEIGHT");
    svf = (SoInterpolateVec3f *)getConnectedEngine(&sxf->translation);
    dofs.append(new Dof(svf, sxf));
    
    /* Head Joint Z */
    installXYZInterpolator("HEAD_JOINTZ");
    
    /* Rear left leg lifter */
    installXYZInterpolator("RLLEGLIFT");
    
    /* Rear right leg lifter */
    installXYZInterpolator("RRLEGLIFT");
    
    /* install neck and head articulated body */
    ab =  (ArticulatedBody *)getByName("NECK_JOINT_AB");
    if(ab){
	//BSInterpV3f *foo = (BSInterpV3f *)getConnectedEngine(&ab->goal);
	dofs.append(new Dof(NULL, ab));
    }
    installXYZInterpolator("EARL1");
    
    SoTransform *teeth =  (SoTransform *)getByName("TEETHR");
    sv3f = (SoInterpolateVec3f *)getConnectedEngine(&teeth->translation);
    dofs.append(new Dof(sv3f, teeth));
    teeth =  (SoTransform *)getByName("TEETHL");
    sv3f = (SoInterpolateVec3f *)getConnectedEngine(&teeth->translation);
    dofs.append(new Dof(sv3f, teeth));
}



int Dog::stepTo(UpdateStateData &usd)
{
    int res =  MotorCreature::stepTo(usd);
#ifdef 0
    float t = usd.time;
    float ssa, angv, sa, ca;
    
    angv = ssa = angle;
    sa = sin(angv);
    ca = cos(angv);
    if(fabs(sa)>1.0e-4)
	angv = speed/(5.5/sa);
    else
	angv = 0;
    if(angv)
	setHeading(getHeading()+angv); 
    
    if(speed){
	  
	Spx baseFrame;
	baseFrame.e.setRotate(getBodyTransform()->rotation.getValue());
	Spv xx(SbVec3f(0, 0, 0), SbVec3f(ca*speed, 0, -sa*speed));
	baseFrame.multInvSpv(&xx);
	setPosition(xx.v0 + getBodyTransform()->translation.getValue());
	
	// creature's velocity in world space 
	vel.v = SbVec3f(0, angv, 0);
	vel.v0 = xx.v0;
    }

    SbBox3f wbox;
    calcBoundingBox(wbox);
    SbVec3f m1, m2;
    m1 = wbox.getMin();
    m2 = wbox.getMax();
    printf("%f %f %f %f %f %f \n", m1[0], m1[1], m1[2], m2[0], m2[1], m2[2]); 
#endif
    return res;

}


void Dog::getEyeView(SoCamera *xf)
{
    SbRotation r = getBodyTransform()->rotation.getValue();
    SbRotation rr = r*SbRotation(SbVec3f(0, 1, 0), M_PI);
    SbVec3f x = getBodyTransform()->translation.getValue();
    x = x*2;
    x[1]+=4.0;
    xf->orientation.setValue(rr);
    xf->position.setValue(x); 
}


int Dog::checkIfNeedToChangePosture(MotorCommandBlock &mcb, MotorSkill *thisOwner)
{
    Dof *aDof = (Dof *)dofs[FRONTBACKZ_INDEX];
    if((aDof->use == Dof::GRABBED)&&
	(thisOwner != aDof->owner)){
	if (aDof->owner == getMotorSkillByName("sitting"))
	    mcb.cmd = GC_STAND;
	else if(aDof->owner == getMotorSkillByName("begging"))
	    mcb.cmd = GC_STANDFROMBEG;
	else if(aDof->owner == getMotorSkillByName("crouching"))
	    mcb.cmd = GC_STANDFROMCROUCH;
	else if(aDof->owner == getMotorSkillByName("lying down"))
	    mcb.cmd = GC_STANDFROMLIEDOWN;
	return 1;
    }

    else {
	Dof *aDof = (Dof *)dofs[RRLSTEPPER_INDEX];
	Dof *bDof = (Dof *)dofs[RLLEGLIFTER_INDEX];
	Dof *cDof = (Dof *)dofs[RRLEGLIFTER_INDEX];
	if(((aDof->use == Dof::GRABBED) && (aDof->owner == getMotorSkillByName("liftleg")))
	    ||(bDof->use == Dof::GRABBED)
	    ||(cDof->use == Dof::GRABBED)) {
	    mcb.cmd = GC_LOWERLEG;
	    return 1;
	}
	else
	    return 0;
    }

}

int Dog::checkIfNeedToChangePosture(MotorCommandBlock &mcb, int cmd)
{
    return checkIfNeedToChangePosture(mcb,controller->getRequiredMotorSkill(cmd));
}

int Dog::checkIfReadyToMove(MotorCommandBlock &mcb)
{
   Dof *aDof = (Dof *)dofs[FRONTBACKZ_INDEX];
   static int ft = 1;
   static void *ms[5];
   if(ft){
       ms[0] = getMotorSkillByName("sitting");
       ms[1] = getMotorSkillByName("begging");
       ms[2] = getMotorSkillByName("crouching");
       ms[3] = getMotorSkillByName("lying down");
       ms[4] = getMotorSkillByName("liftleg");
       ft = 0;
   }
    if(aDof->use == Dof::GRABBED){
	if (aDof->owner == ms[0])
	    mcb.cmd = GC_STAND;
	else if(aDof->owner == ms[1])
	    mcb.cmd = GC_STANDFROMBEG;
	else if(aDof->owner == ms[2])
	    mcb.cmd = GC_STANDFROMCROUCH;
	else if(aDof->owner == ms[3])
	    mcb.cmd = GC_STANDFROMLIEDOWN;
	return 0;
    }
 
    else {
	Dof *aDof = (Dof *)dofs[RRLSTEPPER_INDEX];
	Dof *bDof = (Dof *)dofs[RLLEGLIFTER_INDEX];
	Dof *cDof = (Dof *)dofs[RRLEGLIFTER_INDEX];
	if(((aDof->use == Dof::GRABBED) && (aDof->owner == ms[4]))
	    ||(bDof->use == Dof::GRABBED)
	    ||(cDof->use == Dof::GRABBED)) {
	    mcb.cmd = GC_LOWERLEG;
	    return 0;
	}
	else
	    return 1;
    }

}






