#include "DogController.h"
#include "DogMotorSkills.h"
#include "DogDefines.h"

DogController::DogController(Dog *myDog, int shouldTrace)
:GenericController(myDog)
{
    ticks = 0;
    trace = shouldTrace;
    buildCommandTable();
}

/*
    addCommand( GC_GOSTRAIGHT,       "straight",           TURNINGSKILL, 1, DEFAULT_STRAIGHT, DEFAULT_SPEED);
    addCommand( GC_LFTFWD ,          "leftforward",        TURNINGSKILL, 1,1- DEFAULT_TURN, DEFAULT_SPEED);
    addCommand( GC_RGTFWD ,    	     "rightforward",   	   TURNINGSKILL, 1, DEFAULT_TURN, DEFAULT_SPEED);
*/

int DogController::executeMotorCommand(MotorCommandBlock &mcb)
{
    if( trace ){
	printf("dog controller trace at % d tick \n", ticks);
	mcb.print();
    }
    int result = GenericController::executeMotorCommand(mcb);
    ticks++;
    
    return result;
}



/*************************************************************************
 * 
 *	OLD VERSION OF executeMotorCommand() BELOW FOR REFERENCE
 * 
 * 
 * 
 *************************************************************************/





/*
int DogController::executeMotorCommand(MotorCommandBlock &mcb)
{
    if( trace ){
	printf("dog controller trace at % d tick \n", ticks);
	mcb.print();
    }
    switch(mcb.cmd){
	case GC_FWD:{
	    ((MotorSkill *)master->motorSkills[WALKINGSKILL])->turnOn(&mcb.args);
	    break;
	}
	case GC_RGT:
	case GC_RGTFWD:{
	    if(mcb.getNumArgs() == 0){
		mcb.setNextArg(DEFAULT_TURN);
		mcb.setNextArg(DEFAULT_SPEED);
	    }
	    ((MotorSkill *)master->motorSkills[TURNINGSKILL])->turnOn(&mcb.args);
	    if(mcb.cmd ==  GC_RGTFWD)
		((MotorSkill *)master->motorSkills[WALKINGSKILL])->turnOn(&mcb.args);
	    break;
	}
	case GC_LFT:
	case GC_LFTFWD:{
	    if(mcb.getNumArgs() == 0){
		mcb.setNextArg(1-DEFAULT_TURN);
		mcb.setNextArg(DEFAULT_SPEED);
	    }
	    ((MotorSkill *)master->motorSkills[TURNINGSKILL])->turnOn(&mcb.args);
	    if(mcb.cmd ==  GC_LFTFWD)
		((MotorSkill *)master->motorSkills[WALKINGSKILL])->turnOn(&mcb.args);
	    break;
	}
	case GC_GOSTRAIGHT:{
	    if(mcb.getNumArgs()<3){
		mcb.setNextArg(DEFAULT_STRAIGHT);
		mcb.setNextArg(DEFAULT_SPEED);
	    }
	    ((MotorSkill *)master->motorSkills[TURNINGSKILL])->turnOn(&mcb.args);
	    ((MotorSkill *)master->motorSkills[WALKINGSKILL])->turnOn(&mcb.args); 
	    break;
	}
	case GC_HALT:{
	    mcb.setNextArg(DEFAULT_STRAIGHT);
	    mcb.setNextArg(DEFAULT_SPEED);
	    ((MotorSkill *)master->motorSkills[TURNINGSKILL])->turnOn(&mcb.args);
	   ((MotorSkill *)master->motorSkills[WALKINGSKILL])->turnOff(); 
	    break;
	}
	case GC_WAGTAIL:
	case GC_MOVETAIL:{
	    if(mcb.getNumArgs() == 0)
		mcb.setNextArg(DEFAULT_SPEED);
	    if(mcb.getNumArgs() == 1){
		mcb.setNextArg(0.0);
		mcb.setNextArg(0.0);
	    }
	    else if(mcb.getNumArgs() == 2)
		mcb.setNextArg(DEFAULT_SPEED);
	    ((MotorSkill *)master->motorSkills[WAGGINGSKILL])->turnOn(&mcb.args); 
	    break;
	}
	case GC_STAND:
	case GC_SIT: {
	    if(mcb.getNumArgs() == 0){
		mcb.setNextArg((mcb.cmd == GC_STAND)?DEFAULT_STAND:DEFAULT_SIT);
	    }
	    if(mcb.getNumArgs() == 1)
		mcb.setNextArg(DEFAULT_SPEED*2);
	    ((MotorSkill *)master->motorSkills[SITTINGSKILL])->turnOn(&mcb.args); 
	    break;   
	}
	case GC_STANDFROMBEG:
	case GC_BEG: {
	    if(mcb.getNumArgs() == 0){
		mcb.setNextArg((mcb.cmd == GC_STANDFROMBEG)?DEFAULT_STAND:DEFAULT_SIT);
	    }
	    if(mcb.getNumArgs() == 1)
		mcb.setNextArg(DEFAULT_SPEED);
	    ((MotorSkill *)master->motorSkills[BEGGINGSKILL])->turnOn(&mcb.args); 
	    break;   
	}
	case GC_SHAKEHEAD:
	case GC_TURNHEAD:{
	    if(mcb.cmd ==  GC_SHAKEHEAD && mcb.getNumArgs()==0){
		mcb.setNextArg(.1);
		mcb.setNextArg(0);
		mcb.setNextArg(1);
	    }
	    ((MotorSkill *)master->motorSkills[TURNINGHEADSKILL])->turnOn(&mcb.args); 
	    break;
	  }
	case GC_MOVEHEAD: {
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(DEFAULT_SPEED);
	  if(mcb.getNumArgs() == 1)
	    mcb.setNextArg(1.0);  // default is lowering your head.... 
	  
	  ((MotorSkill *)master->motorSkills[LOWERINGHEADSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_MOVEEARS: {
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(0.0);
	  if(mcb.getNumArgs() == 1)
	    mcb.setNextArg(DEFAULT_SPEED);
	  
	  ((MotorSkill *)master->motorSkills[ROTATINGEARSSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_SNIFF:{
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(1.0);
	  
	  ((MotorSkill *)master->motorSkills[SNIFFINGSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_MOVEBROWS:{
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(DEFAULT_SPEED);
	  if(mcb.getNumArgs() == 1)
	    mcb.setNextArg(0);
	  
	  ((MotorSkill *)master->motorSkills[MOVINGEYEBROWSSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_SHIFTEYES: {
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(1.0);
	  
	  ((MotorSkill *)master->motorSkills[SHIFTINGEYESSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_SHAKEPAW: {
	  ((MotorSkill *)master->motorSkills[SHAKEPAWSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_TILTHEAD:{
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(DEFAULT_SPEED);
	  if(mcb.getNumArgs() == 1)
	    mcb.setNextArg(0);
	  ((MotorSkill *)master->motorSkills[TILTHEADSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_CHEW:
	case GC_MOVEMOUTH: {
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(DEFAULT_SPEED);
	  if(mcb.getNumArgs() == 1)
	    mcb.setNextArg((mcb.cmd == GC_CHEW)?1:0);
	  
	  ((MotorSkill *)master->motorSkills[CHEWINGSKILL])->turnOn(&mcb.args);
	  break;
	}
	case GC_NOD:
	case GC_MOVEHEADZ:{
	  if(mcb.getNumArgs() == 0)
	    mcb.setNextArg(DEFAULT_SPEED);
	  if(mcb.getNumArgs() == 1){
	    mcb.setNextArg((mcb.cmd == GC_NOD)?0:1);
	    }
	    ((MotorSkill *)master->motorSkills[NODDINGHEADSKILL])->turnOn(&mcb.args); 
	    break;
	  }	
	default: {
	  printf("NOT IMPLEMENTED \n");
	  return GC_NOTIMPLEMENTED;
	}
	}
    ticks++;
    return GC_OK;
  }

*/





