#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoPath.h>
#include "BasicDefines.h"
#include "DogMotorSkills.h"
#include "LimbModel.h"
#include "JointModel.h"
#include "ArticulatedBody.h"
#include "PendulumEiEo.h"
#include "BSInterpV3f.h"
#include "CreatureLib.h"
#define DEFTOL 1.0e-4
#define MAX(x, y) ((x)>(y))?(x):(y)
extern int useSound;
/*********************************************************************
 * 
 *		    WALKING SKILL
 * 
 * 
 *********************************************************************/
     
static float gaitPatterns[2][4] = {
    {.75, -.5, -.5, 0} /* walk */, 
    {.25, 0, .5, .5} /* bound */
    };

Walking::Walking(Dog *p,char *pname, float sl, float cycletime,int useAvg)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[FLLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLSTEPPER_INDEX]);
    useAverage = useAvg;
    stepLen = sl;
    cycleTime = cycletime;
    fwd = 1;
    gait = DOG_GAIT_WALK;
    currentPattern = &gaitPatterns[gait][0];
    firstTime = 1;
    
    /* initialize constraints */
    SoTransform *bodyTransform = p->getBodyTransform();
    /* first front feet relative to shoulder */
    SoNode *shoulder = p->getByName("SHOULDER");
    SoNode *hip = p->getByName("HIP");
    SoNode *shoulderh = p->getByName("SHOULDERHEIGHT");
    SoNode *hipj = p->getByName("HIP_JOINT");
    int i;
    for(i=0;i<4;i++)
	abcWalk[i] = new ABConstraint((ArticulatedBody *)dofs[i]->geometry, 
	    ((i<2)?shoulder : hip),
	    ((i<2)?shoulderh : hipj), FALSE);
}

int Walking::prepareToTurnOff(void *)
{
    int i;
    ArticulatedBody *ab;
    
    int okToTurnOff = ((fabs(counter-.70)<=.2) ||
	(fabs(counter-.20)<=.2)) ?1:0;

    if(okToTurnOff){
	for(i=0;i<4;i++){
	    ab = (ArticulatedBody *)dofs[i]->geometry;
	    ab->goal.setValue(initFeetPos[i]);
	    ab->moveToGoalImmediate();
	}
    }
    return okToTurnOff;
}

int Walking::modifyRun(void *arg)
{
    if(arg){ 
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	if(argVals->getArgCount() >0){
	    fwd = argVals->getFloatAt(0);
	    if(argVals->getArgCount()>1){
		gait = argVals->getFloatAt(1);
		currentPattern = &gaitPatterns[gait][0];
	    }
	    if(argVals->getArgCount()>2)
		cycleTime = argVals->getFloatAt(2);
	}
    }
    return 1;  
}



int Walking::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    
    if(okToTurnOn){
	modifyRun(arg);
	staticSetup();
    }
	
    return okToTurnOn;
}

void Walking::updateStateAt(UpdateStateData &usd)
{
    float delta = fwd*1.0/cycleTime;
    float currentTick;
    SbVec3f vfl, vfr, vrl, vrr;
    SbVec3f vf[4];
    int i;
    
    BSInterpV3f *stepper;
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	for(i=0;i<4;i++){
	   ((ArticulatedBody *)dofs[i]->geometry)->enableNotify(FALSE); 
	}
	    
	for(i=0;i<4;i++){   
	    stepper = (BSInterpV3f *)dofs[i]->interpolator;
	    if(i==0){
		counter = cycleWrap(stepper->alpha.getValue()+ delta);
		currentTick = counter;
	    }
	    else 
		currentTick = cycleWrap(counter + currentPattern[i]);
//	    stepper->deltaOnCurve(stepper->alpha.getValue(), currentTick, vf[i]);

	    abcWalk[i]->beginConstraint();
		dofs[i]->setAlpha(currentTick, TRUE);
		((ArticulatedBody *)dofs[i]->geometry)->moveToGoalImmediate();
	    abcWalk[i]->endConstraint();
	}
	float fx = MAX(abcWalk[0]->delta[0], abcWalk[1]->delta[0]);
	float bx = MAX(abcWalk[2]->delta[0], abcWalk[3]->delta[0]);
//	printf("l -> %f r -> %f \n", abcWalk[2]->delta[0], abcWalk[3]->delta[0]);
	if(gait == DOG_GAIT_WALK)
	    ((Dog *)parent)->speed = fwd*fx;
	else
	    ((Dog *)parent)->speed = fwd*MAX(fx, bx); 
	for(i=0;i<4;i++){
	   ((ArticulatedBody *)dofs[i]->geometry)->enableNotify(TRUE);
	   ((ArticulatedBody *)dofs[i]->geometry)->touch();
	}
    }
    else
	parent->speed = 0.0;
    MotorSkill::updateStateAt(usd);
    return;
}

void
Walking::staticSetup()
{
  // call parent function first
  MotorSkill::staticSetup();

  int i;
  BSInterpV3f *stepper;

  /* swap interpolators */
  ArticulatedBody *ab;
  for(i=0;i<4;i++){
    ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 0 ){
      ab->setPickInterpolator(0);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
//      ((Dof *)parent->dofs[FRONTBACKZ_INDEX])->setAlpha(0);
//      ((Dof *)parent->dofs[BACKBACKZ_INDEX])->setAlpha(0);
    }
    if(firstTime){
      ab->getEndPoint(initFeetPos[i]);
      ab->addToCache(initFeetPos[i]);
    }
  }
  firstTime = 0;
  currentPattern = &gaitPatterns[gait][0];
  for(i=0;i<4;i++){
    stepper = (BSInterpV3f *)dofs[i]->interpolator;
    float alphav = ((i==0)? currentPattern[0] :
	    cycleWrap(currentPattern[0] + currentPattern[i]));
    if((fabs(alphav)-stepper->alpha.getValue())>1.0e-5)
	    stepper->alpha.setValue(alphav);
  }
}  

bool_t Walking::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &stepLen))
    return FALSE;
  if (!xdr_float(xdrs, &cycleTime))
    return FALSE;
  if (!xdr_int(xdrs, &fwd))
    return FALSE;
  if (!xdr_int(xdrs, &gait))
    return FALSE;
  if (xdrs->x_op == XDR_DECODE)
    currentPattern = &gaitPatterns[gait][0];
  return TRUE;
}
  
/*********************************************************************
 * 
 *		    TURNING SKILL
 * 
 * 
 *********************************************************************/
Turning::Turning(Dog *p,char *pname, float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[SHOULDER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRONTBACKY_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[BACKBACKY_INDEX]);
//    dofs[numDofs++] = ((Dof *)parent->dofs[HEADJOINTY_INDEX]);
    speed = sp;
}

int Turning::prepareToTurnOff(void *)
{
    int okToTurnOff=1;

    return okToTurnOff;
}

int Turning::modifyRun(void *arg)
{
    if(arg){ 
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	target = argVals->getFloatAt(0);
	speed = argVals->getFloatAt(1);
    }
    return 1;  
}

int Turning::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    
    if(arg){ 
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	target = argVals->getFloatAt(0);
	speed = argVals->getFloatAt(1);
    }
    else {
	speed = .1;
	target = .5;
    }
    return okToTurnOn;
}

void Turning::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    //speed = .1;
    if(state == MotorSkill::SKILL_ON){
	if(target<0.0)
	    target = 0.0;
	else if(target>1.0)
	    target = 1.0;
	    
	SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[0]->interpolator;
	if(fabs(sif->alpha.getValue() -target)<1.0e-2)
	    turnOff();
	else {
	    counter = dofs[0]->setAlphaUsingSpline(speed, target);
	    dofs[1]->setAlphaUsingSpline(speed, 1.0-target);
	    dofs[2]->setAlphaUsingSpline(speed, 1.0-target);
	   
//	    dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
//	    dofs[1]->setAlphaFromSpeedAndTarget(speed, 1.0-target);
//	    dofs[2]->setAlphaFromSpeedAndTarget(speed, 1.0-target);
	}
    }
    /* when done */
    ((SoInterpolateFloat *)dofs[0]->interpolator)->output.getForwardConnections(sfl);
    ((Dog *)parent)->angle = ((SoSFFloat *)sfl[0])->getValue();
    MotorSkill::updateStateAt(usd);
    return;
}

bool_t Turning::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}
  
/*********************************************************************
 * 
 *		    WAG AND RAISE/LOWER TAIL  SKILL
 *  args (optional):
 *	0 - speed of wag
 *	1 - up/down target for raising and lowering tail 
 *	2 - speed for (1) 
 *********************************************************************/
 
 WaggingAndMovingTail::WaggingAndMovingTail(Dog *p,char *pname, float spx, float spz)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[TAILX_INDEX]);
  dofs[numDofs++] = ((Dof *)parent->dofs[TAILZ_INDEX]);
  speed = spx;
  speedz = spz;
}

int WaggingAndMovingTail::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    /* stop wag */
    SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[0]->interpolator;
    sif->alpha.setValue(.5);
    sif = (SoInterpolateFloat *)dofs[1]->interpolator;
    sif->alpha.setValue(0.5);
  }

  return okToTurnOff;

}

int WaggingAndMovingTail::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    target = 1;
    modifyRun(arg);
  }
  else {
    speed = .1;
    speedz = 0;
    target = 1.0;
    targetz = 0;
  }
  return okToTurnOn;
}

int WaggingAndMovingTail::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
	targetz = (argVals->getArgCount()>1)?argVals->getFloatAt(1):targetz;
	speedz = (argVals->getArgCount()>2)?argVals->getFloatAt(2):0;
    }
    return 1;  
}
      
void WaggingAndMovingTail::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	else {
	  if(speed){
	    counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	    
	    if(fabs(counter- 1.0)<1e-4)
	      target = 0.0;
	    else if(counter<1e-4)
	      target = 1.0;
	  }
	  else counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, .5);
	  if(speedz) {
	    counter = dofs[1]->setAlphaFromSpeedAndTarget(speedz, targetz);
	    
	    /* turn this off if not also wagging */
	   // if((fabs(counter-target)<1.e-4)&& (speed == 0.0)
	   //    && parent->runBehavior.getValue())
	   //   turnOff();
	  }
	}
      MotorSkill::updateStateAt(usd);
      }
    return;
  }

bool_t WaggingAndMovingTail::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_float(xdrs, &targetz))
    return FALSE;
  if (!xdr_float(xdrs, &speedz))
    return FALSE;
  return TRUE;
}


/*********************************************************************
 * 
 *		    CROUCH  SKILL
 *  args (optional):
 *	0 - whichpaw	(0 left,  1 right, 2 both)
 *	1 - speed
 *
 * 
 *********************************************************************/
 
Crouch::Crouch(Dog *p,char *pname,float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[FLLSTEPPER_INDEX]); //0
    dofs[numDofs++] = ((Dof *)parent->dofs[FRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[SHOULDERHEIGHT_INDEX]); //2
    dofs[numDofs++] = ((Dof *)parent->dofs[FRONTBACKZ_INDEX]); //3
    dofs[numDofs++] = ((Dof *)parent->dofs[BACKBACKZ_INDEX]); //4
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLSTEPPER_INDEX]);
    
    speed = sp;
    
    /* initialize constraints */
    /* first front feet relative to shoulder */
    SoNode *shoulder = p->getByName("SHOULDER");
    abcCrouch[0] = new ABConstraint((ArticulatedBody *)dofs[0]->geometry, 
	shoulder, (SoNode *)dofs[2]->geometry);
    abcCrouch[0]->weight = SbVec3f(0, 1, 0);
    
    /* now rear feet */
    abcCrouch[1] = new ABConstraint((ArticulatedBody *)dofs[5]->geometry, 
	shoulder, (SoNode *)dofs[5]->geometry);
    abcCrouch[1]->weight = SbVec3f(1, 0, 0);
    
    /* now rear feet */
    abcCrouch[2] = new ABConstraint((ArticulatedBody *)dofs[6]->geometry, 
	shoulder, (SoNode *)dofs[6]->geometry);
    abcCrouch[2]->weight = SbVec3f(1, 0, 0);
    
}

int Crouch::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    okToChangePaws = 1;
    if(arg){
	modifyRun(arg);
    }
    target = 1;
    speed = .1;
    staticSetup();
    static int ft = 1;
    if(ft){
	int i;
	SbVec3f ep;
	for(i=0;i<2;i++){
	    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
	    ab->getEndPoint(ep);
	    ab->goal.setValue(ep);
	    ab->moveToGoalImmediate();
	}
	for(i=5;i<7;i++){
	    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
	    ab->getEndPoint(ep);
	    ab->goal.setValue(ep);
	    ab->moveToGoalImmediate();
	}
	ft = 0;
    }
    return okToTurnOn;
}

int Crouch::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	(argVals->getArgCount()>0)?(int)argVals->getFloatAt(0):2;
	speed = (argVals->getArgCount()>1)?argVals->getFloatAt(1):speed;
    }
    return 1;  
}

int Crouch::prepareToTurnOff(void *)
{
    target = 0;
    speed = .2;
    
    if(fabs(counter)<=.01){
	((SoRotationXYZ *)dofs[3]->geometry)->angle.setValue(0);
	((SoRotationXYZ *)dofs[4]->geometry)->angle.setValue(0);
	int i;
	for(i=5;i<7;i++){
	    ((BSInterpV3f *)dofs[i]->interpolator)->alpha.setValue(0.0);
	    ((ArticulatedBody *)dofs[i]->geometry)->moveToGoalImmediate();
	}
	return 1;
    }
    else return 0;
}

void Crouch::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	if(speed){
/*
	    ((ArticulatedBody *)dofs[0]->geometry)->enableNotify(FALSE);
	    ((ArticulatedBody *)dofs[1]->geometry)->enableNotify(FALSE);
	    ((ArticulatedBody *)dofs[5]->geometry)->enableNotify(FALSE); 
	    ((ArticulatedBody *)dofs[6]->geometry)->enableNotify(FALSE);
*/
	    abcCrouch[0]->beginConstraint();
	    
	    counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    counter = dofs[1]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    
	    if((target != 0)&&(fabs(counter-target)<1.0e-4)){
		target = (fabs(target-1)<1.0e-4)?.7:1;
	    }
	    /* ok force immediate movement */
	    int i;
	    for(i=0;i<2;i++)
		((ArticulatedBody *)dofs[i]->geometry)->moveToGoalImmediate();
	    
	    abcCrouch[0]->endConstraint(); 
//	    printf("beginning crouch \n");
	    abcCrouch[2]->beginConstraint();
	    abcCrouch[1]->beginConstraint();
		
		// now get delta of shoulder 
		SbVec3f delta =((SoTransform *)dofs[2]->geometry)->translation.getValue();
		float ang = asin(fabs(delta[1])/11.0398);
		//printf("%f %f\n", delta[1], ang);
	    
		// now adjust back to compensate 
		((SoRotationXYZ *)dofs[3]->geometry)->angle.setValue(-ang);
		((SoRotationXYZ *)dofs[4]->geometry)->angle.setValue(ang);
	    
	    abcCrouch[1]->endConstraint();
	    abcCrouch[2]->endConstraint();
//	    printf("ending crouch \n");
	    /* turn off if at 0 again */
	    if((target == 0.0) && fabs(counter)<1.0e-4){
		turnOff();
	    }
/*
	    ((ArticulatedBody *)dofs[0]->geometry)->enableNotify(TRUE);
	    ((ArticulatedBody *)dofs[1]->geometry)->enableNotify(TRUE);
	    ((ArticulatedBody *)dofs[5]->geometry)->enableNotify(TRUE); 
	    ((ArticulatedBody *)dofs[6]->geometry)->enableNotify(TRUE);
	    ((ArticulatedBody *)dofs[0]->geometry)->touch();
	    ((ArticulatedBody *)dofs[1]->geometry)->touch();
	    ((ArticulatedBody *)dofs[5]->geometry)->touch(); 
	    ((ArticulatedBody *)dofs[6]->geometry)->touch();
*/	    
	}
	if((!requestedThisTick) && (parent->runBehavior.getValue()))
	    turnOff();
	MotorSkill::updateStateAt(usd);
    }
    return;
}

void
Crouch::staticSetup()
{
  // call parent function first
  MotorSkill::staticSetup();

  int i;
  for(i=0;i<2;i++){
    // set shaking interpolators for front legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 6){
      ab->setPickInterpolator(6);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
    }
  }
  BSInterpV3f *stepper;
  
  for(i=5;i<7;i++){
    // set sitting interpolators for rear legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 9){
      ab->setPickInterpolator(9);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
      
      stepper = (BSInterpV3f *)dofs[i]->interpolator;
      stepper->alpha.setValue(0.0);
      ab->moveToGoalImmediate();
    }
  }
}  

bool_t Crouch::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichPaw))
    return FALSE;
  if (!xdr_int(xdrs, &okToChangePaws))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *		    LYINGDOWN  SKILL
 *  args (optional):
 *	0 - whichpaw	(0 left,  1 right, 2 both)
 *	1 - speed
 *
 * 
 *********************************************************************/
LyingDown::LyingDown(Dog *p,char *pname,float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[FLLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[SHOULDERHEIGHT_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRONTBACKZ_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[BACKBACKZ_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLSTEPPER_INDEX]);
    speed = sp;
    
    /* first front feet relative to shoulder */
    SoNode *shoulder = p->getByName("SHOULDER");
    SoNode *hip = p->getByName("HIP_JOINT");
    backbone = (SoRotationXYZ *)p->getByName("BACK_JOINT");
    /* add constraints */
    abcLD[0] = new ABConstraint((ArticulatedBody *)dofs[0]->geometry, 
	shoulder, (SoNode *)dofs[2]->geometry);
    abcLD[0]->weight = SbVec3f(0, 1, 0);
    
    /* now rear feet */
    abcLD[1] = new ABConstraint((ArticulatedBody *)dofs[5]->geometry, 
	p, hip);
    abcLD[1]->weight = SbVec3f(0, 1, 0);
}

int LyingDown::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    okToChangePaws = 1;
    if(arg){
	modifyRun(arg);
    }
    else {
	whichPaw = 2;
    }
    okToChangePaws = 0;
    target = 1;
    speed = .1;
    staticSetup();
    static int ft = 1;
    if(ft){
	int i;
	SbVec3f ep;
	for(i=0;i<2;i++){
	    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
	    ab->getEndPoint(ep);
	    ab->goal.setValue(ep);
	    ab->moveToGoalImmediate();
	}
	for(i=5;i<7;i++){
	    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
	    ab->getEndPoint(ep);
	    ab->goal.setValue(ep);
	    ab->moveToGoalImmediate();
	}
	ft = 0;
    }
    return okToTurnOn;
}

int LyingDown::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	if(okToChangePaws) 
	    whichPaw = (argVals->getArgCount()>0)?(int)argVals->getFloatAt(0):2;
	speed = (argVals->getArgCount()>1)?argVals->getFloatAt(1):speed;
    }
    return 1;  
}

int LyingDown::prepareToTurnOff(void *)
{
    target = 0;
    speed = .1;
    
    if(fabs(counter)<=.01){
	((SoInterpolateFloat *)dofs[4]->interpolator)->input0.setValue(0);
	((SoInterpolateFloat *)dofs[4]->interpolator)->input1.setValue(-1.3);
	
	((SoRotationXYZ *)dofs[3]->geometry)->angle.setValue(0);
	((SoRotationXYZ *)dofs[4]->geometry)->angle.setValue(0);
	int i;
	for(i=5;i<7;i++){
	    ((BSInterpV3f *)dofs[i]->interpolator)->alpha.setValue(0.0);
	    ((ArticulatedBody *)dofs[i]->geometry)->moveToGoalImmediate();
	}
	return 1;
    }
    else return 0;
}

void LyingDown::updateStateAt(UpdateStateData &usd)
{
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	if(speed){
		speed = .1;
	    
	    abcLD[0]->beginConstraint();
	    abcLD[1]->beginConstraint();
	    
	    counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    counter = dofs[1]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    counter = dofs[5]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    counter = dofs[6]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    
	    /* ok force immediate movement */
	    int i;
	    for(i=0;i<2;i++){
		((ArticulatedBody *)dofs[i]->geometry)->moveToGoalImmediate();
		((ArticulatedBody *)dofs[i+5]->geometry)->moveToGoalImmediate();
	    }
	    /* now adjust shoulder */
	    abcLD[0]->endConstraint();
	    
	    float tmpcounter =  dofs[4]->setAlphaFromSpeedAndTarget(speed, target);
	    
	    abcLD[1]->endConstraint();
	    
	    /* now fix up back */
	    float ang = asin((abcLD[1]->delta[1])/-11.0398);
	    backbone->angle.setValue(backbone->angle.getValue()+ang);

	    /* turn off if at 0 again */
	    if((target == 0.0) && fabs(counter)<1.0e-4){
		turnOff();
	    }
	}
	if((!requestedThisTick) && (parent->runBehavior.getValue()))
	    turnOff();
	MotorSkill::updateStateAt(usd);
    }
    return;
}

void
LyingDown::staticSetup()
{
  // call parent function first
  MotorSkill::staticSetup();

  int i;
  for(i=0;i<2;i++){
    // set shaking interpolators for front legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 6){
      ab->setPickInterpolator(6);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
    }
  }
  for(i=5;i<7;i++){
    // set shaking interpolators for front legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 6){
      ab->setPickInterpolator(6);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
    }
  }
  ((SoInterpolateFloat *)dofs[4]->interpolator)->input0.setValue(0);
  ((SoInterpolateFloat *)dofs[4]->interpolator)->input1.setValue(.6);
}  

  
bool_t LyingDown::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichPaw))
    return FALSE;
  if (!xdr_int(xdrs, &okToChangePaws))
    return FALSE;
  return TRUE;
}


/*********************************************************************
 * 
 *		    LIFT LEG  SKILL
 *  args (optional):
 *	0 - whichpaw	(0 left,  1 right, 2 both)
 *	1 - speed
 *
 * 
 *********************************************************************/
 
LiftLeg::LiftLeg(Dog *p,char *pname,float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLEGLIFTER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLEGLIFTER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLSTEPPER_INDEX]);
    speed = sp;
}
int LiftLeg::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    okToChangePaws = 1;
    if(arg){
	modifyRun(arg);
    }
    else {
	whichPaw = 0;
    }
    okToChangePaws = 0;
    target = 1;
    speed = .05;
    
// do once at start (aries)
    SoundSkill::setLoopingSound("pee", 1);
    SoundSkill::triggerSound("pee");
    
    return okToTurnOn;
}

int LiftLeg::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	if(okToChangePaws) 
	    whichPaw = (argVals->getArgCount()>0)?(int)argVals->getFloatAt(0):0;
	speed = (argVals->getArgCount()>1)?argVals->getFloatAt(1):speed;
    }
    return 1;  
}

int LiftLeg::prepareToTurnOff(void *)
{
    target = 0;
    if(fabs(counter)<=.001){
	dofs[0]->setAlpha(0);
	dofs[1]->setAlpha(0);
    }
    
    // aries
    SoundSkill::turnoffSound("pee");
    
    return (fabs(counter)<=.001) ? 1:0;
}

void LiftLeg::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    
    speed = .1;
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	if(whichPaw == 0){
	   // counter = dofs[0]->setAlphaUsingSpline(speed,target);
	   counter = dofs[0]->setAlphaFromSpeedAndTarget(speed,target);
	}
	else if(whichPaw ==1){
	  //  counter = dofs[1]->setAlphaUsingSpline(speed,target);
	  counter = dofs[1]->setAlphaFromSpeedAndTarget(speed,target);
	}
	/* turn off if at 0 again */
	if((target == 0.0) && (fabs(counter)<1.0e-2))
	  {
	    turnOff();
	  }
	if((!requestedThisTick)&&(parent->runBehavior.getValue()))
	   {
	     turnOff();
	   }
	MotorSkill::updateStateAt(usd);
    }
    return;
}

bool_t LiftLeg::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichPaw))
    return FALSE;
  if (!xdr_int(xdrs, &okToChangePaws))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *		    SHAKE PAW  SKILL
 *  args (optional):
 *	0 - whichpaw	(0 left,  1 right, 2 both)
 *	1 - speed
 *
 * 
 *********************************************************************/
 
ShakePaw::ShakePaw(Dog *p,char *pname,float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[FLLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRLSTEPPER_INDEX]);
    speed = sp;
}
int ShakePaw::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    okToChangePaws = 1;
    if(arg){
	modifyRun(arg);
    }
    else {
	whichPaw = 0;
    }
    okToChangePaws = 0;
    target = 1;
    speed = .1;
    staticSetup();
    return okToTurnOn;
}

int ShakePaw::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	if(okToChangePaws) 
	    whichPaw = (argVals->getArgCount()>0)?(int)argVals->getFloatAt(0):0;
	speed = (argVals->getArgCount()>1)?argVals->getFloatAt(1):speed;
    }
    return 1;  
}

int ShakePaw::prepareToTurnOff(void *)
{
    target = 0;
    speed = .2;
    return (fabs(counter)<=DEFTOL) ? 1:0;
}

void ShakePaw::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	if(speed){
	    if((whichPaw ==0)||(whichPaw ==2)){
		counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    }
	    if((whichPaw ==1)||(whichPaw ==2)) {
		counter = dofs[1]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    }
	    if((target != 0)&&(fabs(counter-target)<1.0e-4)){
		target = (fabs(target-1)<1.0e-4)?.7:1;
	    }
	    /* turn off if at 0 again */
	    if((target == 0.0) && fabs(counter)<1.0e-4)
		turnOff();
	}
	if((!requestedThisTick)&&(parent->runBehavior.getValue()))
	   turnOff();
	MotorSkill::updateStateAt(usd);
    }
    return;
}

void
ShakePaw::staticSetup()
{
  // call parent function first
  MotorSkill::staticSetup();
  
  int i;
  for(i=0;i<2;i++){
    // set shaking interpolators for front legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 3){
      ab->setPickInterpolator(3);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
    }
  }
}

bool_t ShakePaw::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichPaw))
    return FALSE;
  if (!xdr_int(xdrs, &okToChangePaws))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *		    BEGGING SKILL
 *  args (optional):
 *	0 - target	(0 is standing,  1 is sitting)
 *	1 - speed
 * 
 *********************************************************************/
Begging::Begging(Dog *p,char *pname,float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRONTBACKZ_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[BACKBACKZ_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[SHOULDERHEIGHT_INDEX]);
    speed = sp;
    SoNode *shoulder = p->getByName("SHOULDER");
    abcBeg = new ABConstraint((ArticulatedBody *)dofs[0]->geometry, 
	shoulder, (SoNode *)dofs[4]->geometry);
    abcBeg->weight = SbVec3f(1, 1, 0);
}

int Begging::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    
    if(arg){
	modifyRun(arg);
    }
    else {
	speed = .1;
	target = 1;
    }
    staticSetup();
    return okToTurnOn;
}

int Begging::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	target = argVals->getFloatAt(0);
	speed = (argVals->getArgCount()>1)?argVals->getFloatAt(1):speed;
    }
    return 1;  
}

int Begging::prepareToTurnOff(void *)
{
    target = 0;
    speed = .1;
    return (fabs(counter)<=DEFTOL) ? 1:0;
}

void Begging::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
	if(speed){
	
	    /* before you start get the world position of the foot */
	    ArticulatedBody *ab = (ArticulatedBody *)dofs[0]->geometry;
	abcBeg->beginConstraint();   
	    /* first do front-back Z rotation */
	    SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[2]->interpolator;
	 
	    if(sif->input1[0] != 1.3)
		sif->input1.set1Value(0, 1.3);
	    
	    counter =  dofs[2]->setAlphaFromSpeedAndTarget(speed, target);
	    
	    /* first do back-back Z rotation */
	    counter =  dofs[3]->setAlphaFromSpeedAndTarget(speed, target);
	    ab->moveToGoalImmediate();
	abcBeg->endConstraint();
	    /* turn off if at 0 again */
	    if((target == 0.0) && fabs(counter)<1.0e-4){
		/* set this back to exactly 0 */
		((SoTransform *)dofs[4]->geometry)->translation.setValue(SbVec3f(0, 0, 0));
		sif = (SoInterpolateFloat *)dofs[2]->interpolator;
		if(sif->input1[0] != .7)
		    sif->input1.set1Value(0, .7);
		turnOff();
	    }
	}
	if((!requestedThisTick) && (parent->runBehavior.getValue()))
	    turnOff();
	MotorSkill::updateStateAt(usd);
    }
    return;
}

void
Begging::staticSetup()
{
  // call parent function first
  MotorSkill::staticSetup();

  int i;
  for(i=0;i<2;i++){
    // set sitting interpolators for rear legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 3){
      ab->setPickInterpolator(3);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
    }
  }
}  

bool_t Begging::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *		    SITTING/STANDING SKILL
 *  args (optional):
 *	0 - target	(0 is standing,  1 is sitting)
 *	1 - speed
 * 
 *********************************************************************/
 
Sitting::Sitting(Dog *p,char *pname,float sp)
    :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[RLLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[RRLSTEPPER_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[FRONTBACKZ_INDEX]);
    speed = sp;
}
int Sitting::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
    static int ptr=0;
    static FILE *fd[10] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

    if (ptr >= 10)
      {
	for (ptr=0; ptr<10; ptr++)
	  pclose(fd[ptr]);
	ptr = 0;
      }
 
    if(useSound)
	fd[ptr++] = popen("rsh blackfriars '/usr/sbin/sfplay /mas/vision/projects/ive/Alive/Alive_7.2/two_woofs.aiff'", "w");
   	     
    if(arg){
	modifyRun(arg);
    }
    else {
	speed = .1;
	target = 1;
    }
  
    staticSetup();
    return okToTurnOn;
}

int Sitting::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	target = argVals->getFloatAt(0);
	speed = (argVals->getArgCount()>1)?argVals->getFloatAt(1):speed;
    }
    return 1;  
}

int Sitting::prepareToTurnOff(void *)
{
    target = 0;
    speed = .1;
    return (fabs(counter)<=.05) ? 1:0;
}

void Sitting::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    int i;
        
    if(state == MotorSkill::SKILL_ON){
	if(speed){
	    /* first do front-back Z rotation */
	    counter =  dofs[2]->setAlphaFromSpeedAndTarget(speed, target);
	    ((ArticulatedBody *)dofs[0]->geometry)->enableNotify(FALSE);
	    ((ArticulatedBody *)dofs[1]->geometry)->enableNotify(FALSE);
	    for(i=0;i<2;i++){
		/* now do legs */
		counter =  dofs[i]->setAlphaFromSpeedAndTarget(speed, target, TRUE);
	    }
	    ((ArticulatedBody *)dofs[0]->geometry)->enableNotify(TRUE);
	    ((ArticulatedBody *)dofs[1]->geometry)->enableNotify(TRUE);
	    ((ArticulatedBody *)dofs[0]->geometry)->touch();
	    ((ArticulatedBody *)dofs[1]->geometry)->touch();
	    
	    /* turn off if at 0 again */
	    if((target == 0.0) && fabs(counter)<1.0e-4)
		turnOff();
	}
	if((!requestedThisTick) && (parent->runBehavior.getValue()))
	    turnOff();
	MotorSkill::updateStateAt(usd);
    }
    return;
}

void
Sitting::staticSetup()
{
  // call parent function first
  MotorSkill::staticSetup();
  
  int i;
  for(i=0;i<2;i++){
    // set sitting interpolators for rear legs 
    ArticulatedBody *ab = (ArticulatedBody *)dofs[i]->geometry;
    if(ab->pickInterpolator.getValue() != 3){
      ab->setPickInterpolator(3);
      dofs[i]->interpolator = getConnectedEngine(&ab->goal);
      dofs[i]->auxInterpolator0 =  getConnectedEngine(&ab->angleHints);
      dofs[i]->auxInterpolator1 =  getConnectedEngine(&ab->angleHintStrengths);
    }
  }
}

bool_t Sitting::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	LOWERING/RAISING HEAD SKILL
 *    args (optional):
 *	0 - speed of movement
 *      1 - target; desired alpha value*	
 *********************************************************************/
 
LoweringAndRaisingHead::LoweringAndRaisingHead(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[NECKJOINTZ_INDEX]);
  speed = sp;
}

int LoweringAndRaisingHead::prepareToTurnOff(void *)
{
  int okToTurnOff = (fabs(counter-0)<=.01)?1:0;
  if(!okToTurnOff)
    target = 0.0;
  return okToTurnOff;

}

int LoweringAndRaisingHead::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = .1;
    target = 1.0;
  }
  return okToTurnOn;
}

int LoweringAndRaisingHead::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
	target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
	target = (target>1.0)?1.0 : target;
    }
    return 1;  
}
      
void LoweringAndRaisingHead::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
      if(!requestedThisTick && parent->runBehavior.getValue())
	turnOff();
      if(speed){
	counter =  dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	Dog *dog = (Dog *)parent;
	if(dog->newStyleNeck.getValue()){
	    SoInterpolateRotation *svf = (SoInterpolateRotation *)getConnectedEngine(
		&((SoTransform *)dofs[0]->geometry)->rotation);
	    if(fabs(counter - svf->alpha.getValue())> 1.0e-5)
		svf->alpha.setValue(counter);
	}
	else {
	     SoInterpolateFloat *svf = (SoInterpolateFloat *)getConnectedEngine(
		&((SoRotationXYZ *)dofs[0]->geometry)->angle);
	     svf->alpha.setValue(counter);
	}
//	if(fabs(counter-target)<1.e-4 && parent->runBehavior.getValue())
//	  turnOff();
      }
    MotorSkill::updateStateAt(usd);
    }
    return;
  }

bool_t LoweringAndRaisingHead::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	SHAKING AND TURNING HEAD SKILL
 *    args (optional):
 *	0 - speed of movement
 *      1 - which mode (0 is shaking , 1 is turning)
 *	2 - target; desired alpha value 
 *********************************************************************/

 
ShakingAndTurningHead::ShakingAndTurningHead(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[HEADJOINTY_INDEX]);
  speed = sp;
}

int ShakingAndTurningHead::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff ) {
    SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[0]->interpolator;
    sif->alpha.setValue(.5);
  }
  return okToTurnOff;
}

int ShakingAndTurningHead::prepareToRun(void *arg)
{
  int okToTurnOn = 1;

  if(arg){
    modifyRun(arg);
  }
  else {
    target = 1.0;
    speed = .1;
    whichMode = 0;
  }
  return okToTurnOn;
}

int ShakingAndTurningHead::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
	target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
	whichMode = (argVals->getArgCount()>2)?argVals->getFloatAt(2):0;
      }
    return 1;  
}
      
void ShakingAndTurningHead::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
     	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	else {
	    counter =  dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	    if(!whichMode){
	      if(fabs(counter- 1.0)<1e-4)
		target = 0.0;
	      else if(counter<1e-4)
		target = 1.0;
	    }
	  }
      MotorSkill::updateStateAt(usd);
      }
    return;
  }

bool_t ShakingAndTurningHead::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichMode))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	SNIFFING SKILL
 *    args (optional):
 *	0 - speed of movement
 *
 *********************************************************************/

Sniffing::Sniffing(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[NOSEPEND_INDEX]);
  speed = sp;
}

int Sniffing::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  SoPendulum *sif = (SoPendulum *)dofs[0]->interpolator;
  sif->on.setValue(FALSE);
  return okToTurnOff;
}

int Sniffing::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = 1;
  }
  return okToTurnOn;
}

int Sniffing::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
      }
    return 1;  
}
      
void Sniffing::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
     	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	else {
	  SoPendulum *sif = (SoPendulum *) dofs[0]->interpolator;
	  sif->on.setValue(TRUE);
	}
      MotorSkill::updateStateAt(usd);
      }
    return;
  }
    

/*********************************************************************
 * 
 *	MOVING EARS SKILL
 *    args (optional):
 *	0 - target; desired alpha value 
 *	1 - speed of movement
 *      2 - tilt of ear
 *
 ********************************************************************/

RotatingEars::RotatingEars(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)	
{
    dofs[numDofs++] = ((Dof *)parent->dofs[EARLZ_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[EARRZ_INDEX]);
    dofs[numDofs++] = ((Dof *)parent->dofs[EARTIP_INDEX]);
    speed = sp;
    earLeft = 1.0;
    earRight = 1.0;
}


int RotatingEars::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    dofs[0]->setAlpha(.5);
    dofs[1]->setAlpha(.5);
    dofs[2]->setAlpha(1);
  }
  return okToTurnOff;
}

int RotatingEars::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = 0.1;
    earLeft = 1.0;
    earRight = 1.0;
    earTilt = 1.0;
    target = 0.0;
  }
  return okToTurnOn;
}

int RotatingEars::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 

      speed = argVals->getFloatAt(0);
      target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
 //     earLeft = (argVals->getArgCount()>2)?argVals->getFloatAt(2):1.0;
 //     earRight = (argVals->getArgCount()>3)?argVals->getFloatAt(3):1.0;
      earTilt = (argVals->getArgCount()>2)?argVals->getFloatAt(2):1.0;
    }
    return 1;  
}

void RotatingEars::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
      if(!requestedThisTick && parent->runBehavior.getValue())
	turnOff();
      if(speed){
	if(earLeft) {
	    counter =  dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	}
	if(earRight){
	    counter =  dofs[1]->setAlphaFromSpeedAndTarget(speed, target);
	}
	counter = dofs[2]->setAlphaFromSpeedAndTarget(speed, earTilt);
      }
    MotorSkill::updateStateAt(usd);
    }
    return;
}
bool_t RotatingEars::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_float(xdrs, &earLeft))
    return FALSE;
  if (!xdr_float(xdrs, &earRight))
    return FALSE;
  if (!xdr_float(xdrs, &earTilt))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	SHOWTEETH MOVEMENTS
 *    args (optional):
 *      0 - speed
 *      1 - targetl
 *
 *
 ********************************************************************/

ShowTeeth::ShowTeeth(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[TEETHR_INDEX]);
  dofs[numDofs++] = ((Dof *)parent->dofs[TEETHL_INDEX]);
  speed = sp;
}

int ShowTeeth::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;
  
    if(okToTurnOff) {
	dofs[0]->setAlpha(0);
	dofs[1]->setAlpha(0);
    }
    return okToTurnOff;
}

int ShowTeeth::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = .1;
    target = 0.0;
  }
  return okToTurnOn;
}

int ShowTeeth::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
      speed = argVals->getFloatAt(0);
      target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):1.0;
    }
    return 1;  
}

void ShowTeeth::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	counter = dofs[1]->setAlphaFromSpeedAndTarget(speed, target);
    }
    MotorSkill::updateStateAt(usd);
    return;
}
bool_t ShowTeeth::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}
/*********************************************************************
 * 
 *	BUGEYES MOVEMENTS
 *    args (optional):
 *      0 - speed
 *      1 - targetl
 *
 *
 ********************************************************************/

BugEyes::BugEyes(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[EYEBALLPEND_INDEX]);
  speed = sp;
}

int BugEyes::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;
  
    if(okToTurnOff) {
	dofs[0]->setAlpha(.5);
    }
    return okToTurnOff;
}

int BugEyes::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = .1;
    target = 0.0;
  }
  return okToTurnOn;
}

int BugEyes::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
      speed = argVals->getFloatAt(0);
      target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):1.0;
    }
    return 1;  
}

void BugEyes::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
    }
    MotorSkill::updateStateAt(usd);
    return;
}
bool_t BugEyes::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}
/*********************************************************************
 * 
 *	EYEBROW MOVEMENTS
 *    args (optional):
 *      0 - speed
 *      1 - targetl
 *      2 - whichMode (0 shuttle 1 off)
 *
 ********************************************************************/

MovingEyeBrows::MovingEyeBrows(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[EYEBROWSHUTTLE_INDEX]);
  dofs[numDofs++] = ((Dof *)parent->dofs[EYEBROWLX_INDEX]);
  dofs[numDofs++] = ((Dof *)parent->dofs[EYEBROWRX_INDEX]);
  speed = sp;
}

int MovingEyeBrows::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    if(!whichMode) {
      SoShuttle *shut = (SoShuttle *)dofs[0]->interpolator;
      shut->on.setValue(FALSE);
    }
    else {
	dofs[1]->setAlpha(.5);
	dofs[2]->setAlpha(.5);
    }
  }
  return okToTurnOff;
}

int MovingEyeBrows::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = 1.0;
    whichMode = 0;    
  }
  return okToTurnOn;
}

int MovingEyeBrows::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
      speed = argVals->getFloatAt(0);
      targetl = (argVals->getArgCount()>1)?argVals->getFloatAt(1):1.0;
      targetr = targetl;
      whichMode = (argVals->getArgCount()>2)?argVals->getFloatAt(2):0;
    }
    return 1;  
}

void MovingEyeBrows::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	if(!whichMode){
	    SoShuttle *shutt = (SoShuttle *)dofs[0]->interpolator;
	    shutt->on.setValue(TRUE);
	}
	else {
	    counter = dofs[1]->setAlphaFromSpeedAndTarget(speed, targetl);
	    counter = dofs[2]->setAlphaFromSpeedAndTarget(speed, targetr);
	}
    }
    MotorSkill::updateStateAt(usd);
    return;
}
bool_t MovingEyeBrows::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &targetl))
    return FALSE;
  if (!xdr_float(xdrs, &targetr))
    return FALSE;
  if (!xdr_int(xdrs, &whichMode))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	LOOK AT SKILL
 *    args x, y, z => world point to look at
 *
 *********************************************************************/
SoRotation *headjoint;
SoRotation *eyejoint;
SoGetMatrixAction *ma2;
SoPath *pathToHead1;

LookAt::LookAt(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[EYEBALLPEND_INDEX]);
    headjoint = (SoRotation *)p->getByName("HEADROT");
    eyejoint = (SoRotation *)p->getByName("EJ");
    SbViewportRegion vr;
    ma = new SoGetMatrixAction(vr);
    ma2 = new SoGetMatrixAction(vr);
    
    SoSearchAction sa;
    sa.setNode((SoNode *)eyejoint);
    sa.apply(p);
    pathToEye = sa.getPath();
    pathToEye->ref();
    
    sa.setNode(headjoint);
    sa.apply(p);
    pathToHead1 = sa.getPath();
    pathToHead1->ref();
    
    speed = sp;
}

int LookAt::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;
    SoPendulum *sif = (SoPendulum *)dofs[0]->interpolator;
    sif->on.setValue(FALSE);
    return okToTurnOff;
}

int LookAt::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    lookAtPoint = SbVec3f(1, 0, 0);
    speed = 1;
  }
  return okToTurnOn;
}

int LookAt::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	lookAtPoint = SbVec3f(argVals->getFloatAt(0), argVals->getFloatAt(1), argVals->getFloatAt(2));
	speed = argVals->getFloatAt(3);
      }
    return 1;  
}
      
void LookAt::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	SbVec3f localPt;
	SbVec3f tpt;
	SbVec3f axis;
	float ang;
	
	SbRotation startingRot = headjoint->rotation.getValue();
	
	// get the look-at point in x and z
	headjoint->rotation.setValue(SbVec3f(0, 1, 0), 0);
	tpt = lookAtPoint;
	ma->apply(pathToHead1);
	SbMatrix lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPoint, localPt);
	localPt[1] = 0;
	
	// compute the rotation around y to point x axis to the
	// look-at point.
	SbRotation deltaRot = SbRotation(SbVec3f(1, 0, 0), localPt);

	SbRotation lookAtRot = deltaRot*headjoint->rotation.getValue();
	lookAtRot.getValue(axis, ang);
	if(fabs(ang)>2.8){
	    ang = (ang<0)?-2.8 : 2.8;
	    lookAtRot.setValue(axis, ang);
	}
	headjoint->rotation.setValue(lookAtRot);
	
	// now do rotation around new axis to lower head
	ma->apply(pathToHead1);
	lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPoint, localPt);
	deltaRot = SbRotation(SbVec3f(1, 0, 0), localPt);

	lookAtRot = deltaRot*headjoint->rotation.getValue();
	lookAtRot.getValue(axis, ang);
	if(fabs(ang)>1.45){
	    ang = (ang<0)?-1.45 : 1.45;
	    lookAtRot.setValue(axis, ang);
	}
	headjoint->rotation.setValue(SbRotation::slerp(startingRot, 
	    lookAtRot, speed));
	
/* now do eye movement
	ma2->apply(pathToEye);
	lXw = ma2->getInverse();
	lXw.multVecMatrix(lookAtPoint, localPt);
	
	deltaRot = SbRotation(SbVec3f(1, 0, 0), localPt);
	lookAtRot = deltaRot*eyejoint->rotation.getValue();
	eyejoint->rotation.setValue(lookAtRot);
*/	
	// New Eye stuff
	// get the look-at point in x and z
	eyejoint->rotation.setValue(SbVec3f(0, 1, 0), 0);
	tpt = lookAtPoint;
	ma->apply(pathToEye);
	lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPoint, localPt);
	localPt[1] = 0;
	
	// compute the rotation around y to point x axis to the
	// look-at point.
	deltaRot = SbRotation(SbVec3f(1, 0, 0), localPt);

	lookAtRot = deltaRot*eyejoint->rotation.getValue();
	lookAtRot.getValue(axis, ang);
	if(fabs(ang)>1){
	    ang = (ang<0)?-1 : 1;
	    lookAtRot.setValue(axis, ang);
	}
	eyejoint->rotation.setValue(lookAtRot);
	
	//turnOff();
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	
      }
    MotorSkill::updateStateAt(usd);
    return;
}
    
bool_t LookAt::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lookAtPoint))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	LOOK AT SKILL USING ARTICULATED BODY
 *    args x, y, z => world point to look at
 *
 *********************************************************************/
void quadIntersection(SbMatrix q, SbVec3f &pt)
{
    float k0, k1, k2;
    SbVec3f a = pt;
    a.normalize();
    k2 = a[0]*(a[0]*q[0][0] + a[1]*(q[0][1]+q[1][0]))
	+ a[1]*(a[1]*q[1][1] + a[2]*(q[1][2]+q[2][1]))
	+ a[2]*(a[2]*q[2][2] + a[0]*(q[0][2]+q[2][0]));
    k1 = a[0]*(q[0][3] + q[3][0]) 
	+ a[1]*(q[1][3] + q[3][1])
	+ a[2]*(q[2][3] + q[3][2]);
    k0 = q[3][3];
    
    // now solve quadratic equation k2*t*t + k1*t + k0 = 0
    // (-b +- sqrt(b^2 - 4*a*c))/(2*a)
    // k2 = a; k1 = b; k0 = c;
    
    float f1 = sqrt(k1*k1 - 4*k2*k0);
    float r1 = (-k1 +  f1)/(2*k2);
    float r2 = (-k1 -  f1)/(2*k2); 
    SbVec3f p1 = r1*a;
    SbVec3f p2 = r2*a;
    if((p1-pt).length() <(p2-pt).length()){
//	printf("p1 is intersection point %f %f %f \n", 
//	    p1[0], p1[1], p1[2]);
	pt = p1;
    }
    else {
//	printf("p2 is intersection point %f %f %f \n", 
//	    p2[0], p2[1], p2[2]);
	pt = p2;
    }
}

static void ptOnSphere(SbVec3f &pt, SbVec3f &lp_old,  float r, float interp)
{
    SbVec3f npt = pt;
//    printf("before %f %f %f \n", pt[0],pt[1], pt[2]);
    float length = npt.normalize();
    pt = r * npt;
    float limit = 0;
    if(pt[0]<limit){
	pt[0] = limit;
	int sgn = (pt[2]<0)?-1:1;
	pt[2] = sgn*sqrt(r*r - pt[1]*pt[1] - limit*limit);
    }
//    printf("after %f %f %f \n", pt[0],pt[1], pt[2]); 
    SbRotation r0(SbVec3f(1, 0, 0), lp_old);
    SbRotation r1(SbVec3f(1, 0, 0), pt);
    SbRotation r2 = SbRotation::slerp(r0, r1, interp);
    r2.multVec(SbVec3f(1, 0, 0),lp_old) ;
    pt = lp_old = r *lp_old;
}

LookAtAB::LookAtAB(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[NECKAB_INDEX]);
    pathToHead = parent->getPathToName(parent, SbName("HEADJOINT"));
    pathToHead->ref();
    pathToNeck = parent->getPathToName(parent, SbName("NECK_JOINT_AB"));
    pathToNeck->ref();
    pathToEye = parent->getPathToName(parent, SbName("EJ"));
    pathToEye->ref();
    lp_old = SbVec3f(1, 0, 0);
    ma = new SoGetMatrixAction(SbViewportRegion());
    eyejoint = (SoRotation *)p->getByName("EJ");
    speed = sp;
}

int LookAtAB::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;
    return okToTurnOff;
}

int LookAtAB::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    lookAtPoint = SbVec3f(1, 0, 0);
    speed = 1;
  }
  return okToTurnOn;
}

int LookAtAB::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	lookAtPoint = SbVec3f(argVals->getFloatAt(0), argVals->getFloatAt(1), argVals->getFloatAt(2));
	speed = argVals->getFloatAt(3);
      }
    return 1;  
}
      
void LookAtAB::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	SbVec3f lookAtPointLocal,lookAtPointWorld ;
	((ArticulatedBody *)dofs[0]->geometry)->enableNotify(FALSE);
	
	// get point in head's base frame and calculate appropriate point
	// on sphere, and convert it back to world
	ma->apply(pathToHead);
	SbMatrix wXl = ma->getMatrix();
	SbMatrix lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPoint, lookAtPointLocal);
	SbVec3f lpl = lookAtPointLocal;

	lookAtPointLocal = lpl;
	ptOnSphere(lookAtPointLocal, lp_old, 1, .2);
	wXl.multVecMatrix(lookAtPointLocal, lookAtPointWorld);
    
	// now convert it to neck's base frame
	ma->apply(pathToNeck);
	lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPointWorld, lookAtPointLocal);
    
	// now set goal and move immediate
	ArticulatedBody *ab = (ArticulatedBody *)dofs[0]->geometry;
	if((ab->goal.getValue() - lookAtPointLocal).length()>
	    ab->tol.getValue()){
	    ab->goal.setValue(lookAtPointLocal);
	    ab->moveToGoalImmediate();
	}

	// New Eye stuff
	// get the look-at point in x and z
	eyejoint->rotation.setValue(SbVec3f(0, 1, 0), 0);
	ma->apply(pathToEye);
	lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPoint, lookAtPointLocal);
	lookAtPointLocal[1] = 0;
	
	// compute the rotation around y to point x axis to the
	// look-at point.
	SbRotation deltaRot = SbRotation(SbVec3f(1, 0, 0), lookAtPointLocal);
	SbRotation lookAtRot = deltaRot*eyejoint->rotation.getValue();
	SbVec3f axis;
	float ang;
	lookAtRot.getValue(axis, ang);
	if(fabs(ang)>1){
	    ang = (ang<0)?-1 : 1;
	    lookAtRot.setValue(axis, ang);
	}
	eyejoint->rotation.setValue(lookAtRot);
	((ArticulatedBody *)dofs[0]->geometry)->enableNotify(TRUE);
	((ArticulatedBody *)dofs[0]->geometry)->touch();
	//turnOff();
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	
      }
    MotorSkill::updateStateAt(usd);
    return;
}

bool_t LookAtAB::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lookAtPoint))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lp_old))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	SHAKEHEAD USING ARTICULATED BODY
 *    args x, y, z => world point to look at
 *
 *********************************************************************/

ShakeHeadAB::ShakeHeadAB(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[NECKAB_INDEX]);
    pathToHead = parent->getPathToName(parent, SbName("HEADJOINT"));
    pathToHead->ref();
    pathToNeck = parent->getPathToName(parent, SbName("NECK_JOINT_AB"));
    pathToNeck->ref();
    lp_old = SbVec3f(1, 0, 0);
    ma = new SoGetMatrixAction(SbViewportRegion());
    leftLimit = SbVec3f(.5, 0, .35);
    leftLimit.normalize();
    rightLimit = SbVec3f(.5, 0, -.35);
    rightLimit.normalize();
    speed = sp;
}

int ShakeHeadAB::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;
    return okToTurnOff;
}

int ShakeHeadAB::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = .4;
  }
  lookAtPoint = leftLimit;
  return okToTurnOn;
}

int ShakeHeadAB::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	speed = argVals->getFloatAt(0);
    }
    return 1;  
}
      
void ShakeHeadAB::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	SbVec3f lookAtPointLocal,lookAtPointWorld ;
	
	// get point in head's base frame and calculate appropriate point
	// on sphere, and convert it back to world
	ma->apply(pathToHead);
	SbMatrix wXl = ma->getMatrix();
	SbMatrix lXw = ma->getInverse();
	lookAtPointLocal = lookAtPoint;
	ptOnSphere(lookAtPointLocal, lp_old, 1, speed);
	wXl.multVecMatrix(lookAtPointLocal, lookAtPointWorld);
	if((lookAtPointLocal-leftLimit).length()<.12)
	    lookAtPoint = rightLimit;
	else if((lookAtPointLocal-rightLimit).length()<.12)
	    lookAtPoint = leftLimit;
	// now convert it to neck's base frame
	ma->apply(pathToNeck);
	lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPointWorld, lookAtPointLocal);
    
	// now set goal and move immediate
	ArticulatedBody *ab = (ArticulatedBody *)dofs[0]->geometry;
	ab->goal.setValue(lookAtPointLocal);
	ab->moveToGoalImmediate();

	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	
      }
    MotorSkill::updateStateAt(usd);
    return;
}

bool_t ShakeHeadAB::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lookAtPoint))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &leftLimit))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &rightLimit))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lp_old))
    return FALSE;
  return TRUE;
}
/*********************************************************************
 * 
 *	NODHEAD USING ARTICULATED BODY
 *    args x, y, z => world point to look at
 *
 *********************************************************************/

NodHeadAB::NodHeadAB(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
    dofs[numDofs++] = ((Dof *)parent->dofs[NECKAB_INDEX]);
    pathToHead = parent->getPathToName(parent, SbName("HEADJOINT"));
    pathToHead->ref();
    pathToNeck = parent->getPathToName(parent, SbName("NECK_JOINT_AB"));
    pathToNeck->ref();
    lp_old = SbVec3f(1, 0, 0);
    ma = new SoGetMatrixAction(SbViewportRegion());
    topLimit = SbVec3f(1, .45, 0);
    topLimit.normalize();
    bottomLimit = SbVec3f(.5, -.05, 0);
    bottomLimit.normalize();
    speed = sp;
}

int NodHeadAB::prepareToTurnOff(void *)
{
    int okToTurnOff = 1;
    return okToTurnOff;
}

int NodHeadAB::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = .1;
  }
  lookAtPoint = topLimit;
  return okToTurnOn;
}

int NodHeadAB::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){
	speed = argVals->getFloatAt(0);
    }
    return 1;  
}
      
void NodHeadAB::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON){
	SbVec3f lookAtPointLocal,lookAtPointWorld ;
	
	// get point in head's base frame and calculate appropriate point
	// on sphere, and convert it back to world
	ma->apply(pathToHead);
	SbMatrix wXl = ma->getMatrix();
	SbMatrix lXw = ma->getInverse();
	lookAtPointLocal = lookAtPoint;
	ptOnSphere(lookAtPointLocal, lp_old, 1, speed);
	wXl.multVecMatrix(lookAtPointLocal, lookAtPointWorld);
	if((lookAtPointLocal-topLimit).length()<.12)
	    lookAtPoint = bottomLimit;
	else if((lookAtPointLocal-bottomLimit).length()<.12)
	    lookAtPoint = topLimit;
	// now convert it to neck's base frame
	ma->apply(pathToNeck);
	lXw = ma->getInverse();
	lXw.multVecMatrix(lookAtPointWorld, lookAtPointLocal);
    
	// now set goal and move immediate
	ArticulatedBody *ab = (ArticulatedBody *)dofs[0]->geometry;
	ab->goal.setValue(lookAtPointLocal);
	ab->moveToGoalImmediate();

	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	
      }
    MotorSkill::updateStateAt(usd);
    return;
}

bool_t NodHeadAB::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lookAtPoint))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &topLimit))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &bottomLimit))
    return FALSE;
  if (!xdr_SbVec3f(xdrs, &lp_old))
    return FALSE;
  return TRUE;
}
/*********************************************************************
 * 
 *	SHIFTING EYES SKILL
 *    args (optional):
 *	0 - speed of movement
 *
 *********************************************************************/

ShiftingEyes::ShiftingEyes(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[EYEBALLPEND_INDEX]);
  speed = sp;
}

int ShiftingEyes::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  SoPendulum *sif = (SoPendulum *)dofs[0]->interpolator;
  sif->on.setValue(FALSE);
  return okToTurnOff;
}

int ShiftingEyes::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    speed = 1;
  }
  return okToTurnOn;
}

int ShiftingEyes::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
      }
    return 1;  
}
      
void ShiftingEyes::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
     	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	else {
	  SoPendulum *sif = (SoPendulum *) dofs[0]->interpolator;
	  sif->on.setValue(TRUE);
	}
  MotorSkill::updateStateAt(usd);
      }
    return;
  }
    

/*********************************************************************
 * 
 *	TILTING SKILL
 *    args (optional):
 *	0 - speed of movement
 *	1 - target; desired alpha value 
 *      2 - whichMode JUST FOR MY ENJOYMENT!!!!
 *
 *********************************************************************/

 
TiltingHead::TiltingHead(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[HEADPEND_INDEX]);
  speed = sp;
}

int TiltingHead::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    SoInterpolate *sif = (SoInterpolate *)dofs[0]->interpolator;
    sif->alpha.setValue(0.5);
  }
  return okToTurnOff;
}

int TiltingHead::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    target = 1.0;
    speed = .1;
    whichMode = 0;
  }
  return okToTurnOn;
}

int TiltingHead::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
	target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
	whichMode = (argVals->getArgCount()>2)?argVals->getFloatAt(2):0;
      }
    return 1;  
}
      
void TiltingHead::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
     	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
	else {
	    counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	    if(!whichMode){
		if(fabs(counter- 1.0)<1e-4)
		    target = 0.0;
		else if(counter<1e-4)
		    target = 1.0;
	    }
	}
    MotorSkill::updateStateAt(usd);
      }
    return;
  }
    
bool_t TiltingHead::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichMode))
    return FALSE;
  return TRUE;
}

/*********************************************************************
 * 
 *	OPEN/CLOSE MOUTH SKILL
 *    args (optional):
 *	0 - speed of movement	
 *	1 - whichMode (1 is chewing,  0 is open/close mouth)
 *	2 - target; desired alpha value 
 *
 *********************************************************************/

 
OpeningAndClosingMouth::OpeningAndClosingMouth(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[MOUTH_INDEX]);
  speed = sp;
}

int OpeningAndClosingMouth::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    whichMode = 0;
    target = .1;
    okToTurnOff = (fabs(counter-target)<1.0e-1)?1 :0;
  }
  return okToTurnOff;
}

int OpeningAndClosingMouth::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    whichMode = 1; 
    speed = .1;
  }
  return okToTurnOn;
}

int OpeningAndClosingMouth::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    speed = argVals->getFloatAt(0);
    target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
    whichMode = (argVals->getArgCount()>2)?argVals->getFloatAt(2):1;

    return 1;  
}
      
void OpeningAndClosingMouth::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED))
	{
     	counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	if(whichMode){
	    if(fabs(counter- 1.0)<1e-4)
		target = 0.0;
	    else if(counter<1e-4)
		target = 1.0;
	}
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
    }
    MotorSkill::updateStateAt(usd);
    return;
}

bool_t OpeningAndClosingMouth::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichMode))
    return FALSE;
  return TRUE;
}
/*********************************************************************
 * 
 *	BARK MOUTH SKILL
 *    args (optional):
 *	0 - speed of movement	
 *	1 - target (1 is open)
 *	2 - target; desired alpha value 
 *
 *********************************************************************/
Bark::Bark(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[MOUTH_INDEX]);
  speed = sp;
  startedBark = 0;
}

int Bark::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    okToTurnOff = ((fabs(target - .1)<1.0e-4)&&
	(fabs(counter-target)<1.0e-1))?1 :0;
  }
  if(okToTurnOff)
    startedBark = 0;
  return okToTurnOff;
}

int Bark::prepareToRun(void *arg)
{
    int okToTurnOn = 1;
  
    if(arg){
	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	speed = argVals->getFloatAt(0);
	target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
	const char *snd = (argVals->getArgCount()>2)?argVals->getCharPtrAt(2):NULL;
	if(!startedBark && (snd != NULL)){
	    parent->controller->executeMotorCommand(
	    MotorCommandBlock(GC_SOUND_ON, ArgType((char *)snd)));
	    startedBark = 1;
	}
    }
    else {
	speed = .1;
    }
    return okToTurnOn;
}

int Bark::modifyRun(void *arg)
{
    return 1;  
}
      
void Bark::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if((state == MotorSkill::SKILL_ON)||
	(state == MotorSkill::SKILL_OFF_BLOCKED)){
     	counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
//	printf("counter-> %f target-> %f \n", counter, target);
	if((target>0.2)&&(fabs(counter- target)<1e-4))
		target = 0.1;
	else if((target<0.2)&&(fabs(counter- target)<1e-4))
		turnOff();
	if(!requestedThisTick && parent->runBehavior.getValue())
	    turnOff();
    }
    MotorSkill::updateStateAt(usd);
    return;
}

bool_t Bark::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  return TRUE;
}
/*********************************************************************
 * 
 *	NODDING AND LOWERING/RAISING HEAD
 *    args (optional):
 *	0 - speed of movement
 *      1 - which mode (0 is nodding, 1 is lowering/raising)
 *	2 - target; desired alpha value 
 *********************************************************************/

 
NoddingAndLoweringHead::NoddingAndLoweringHead(Dog *p,char *pname, float sp)
     :MotorSkill(p,pname)
{
  dofs[numDofs++] = ((Dof *)parent->dofs[HEADJOINTZ_INDEX]);
  speed = sp;
}

int NoddingAndLoweringHead::prepareToTurnOff(void *)
{
  int okToTurnOff = 1;
  
  if(okToTurnOff) {
    SoInterpolateFloat *sif = (SoInterpolateFloat *)dofs[0]->interpolator;
    sif->alpha.setValue(.5);
  }
  return okToTurnOff;
}

int NoddingAndLoweringHead::prepareToRun(void *arg)
{
  int okToTurnOn = 1;
  
  if(arg){
    modifyRun(arg);
  }
  else {
    target = 1.0;
    speed = .1;
    whichMode = 0;
  }
  return okToTurnOn;
}

int NoddingAndLoweringHead::modifyRun(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    if(arg){ 
	speed = argVals->getFloatAt(0);
	target = (argVals->getArgCount()>1)?argVals->getFloatAt(1):target;
	whichMode = (argVals->getArgCount()>2)?argVals->getFloatAt(2):0;
      }
    return 1;  
}
      
void NoddingAndLoweringHead::updateStateAt(UpdateStateData &usd)
{
    SoFieldList sfl;
    
    if(state == MotorSkill::SKILL_ON){
     	if(!requestedThisTick && parent->runBehavior.getValue())  
	    turnOff();
	else {
	    counter = dofs[0]->setAlphaFromSpeedAndTarget(speed, target);
	    if(!whichMode){
		if(fabs(counter- 1.0)<1e-4)
		    target = 0.0;
		else if(counter<1e-4)
		    target = 1.0;
	    }
	}
    MotorSkill::updateStateAt(usd);
      }
    return;
  }

bool_t NoddingAndLoweringHead::toOrFromStream(XDR *xdrs)
{
  if (!MotorSkill::toOrFromStream(xdrs))
    return FALSE;
  if (!xdr_float(xdrs, &target))
    return FALSE;
  if (!xdr_int(xdrs, &whichMode))
    return FALSE;
  return TRUE;
}









