#ifndef DOGSKILLS
#define DOGSKILLS

#include "MotorSkill.h"
#include "Dog.h"
#include <Inventor/SoLists.h>

class SoPath;
class UpdateStateData;
class ABConstraint;

/*********************************************************************
 * 
 *		   WALKING
 *  args (optional):
 *	0 - forward (1) or back (-1)
 *	1 - gait: (0) walk (1) bound 
 *	2 - cycleTime 
 *********************************************************************/


/*
 * This MotorSkill implements walking in whatever direction the shoulders
 * are pointed. Currently there is one speed. Turning on the skill starts 
 * walk,  turning it off stops it. Because the skill wants to stop in
 * a neutral.
 */
class Walking: public MotorSkill {
    public:
    float stepLen;
    float cycleTime;
    int useAverage;
    ABConstraint *abcWalk[4];
    int fwd;
    int gait;
    float *currentPattern;
    SbVec3f initFeetPos[4];	// new variables !!
    int firstTime;
    
    Walking(Dog *p,char *pname, float sl, float cycletime, int useAvg);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    int prepareToRun(void *arg);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual void staticSetup();
    virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		    TURNING
 *  args (optional):
 *	0 - desired alpha value
 *	1 - speed for (1) 
 *
 *********************************************************************/

class Turning: public MotorSkill {
    public:
    float target;
    Turning(Dog *p,char *pname, float sp);
    int modifyRun(void *args);
    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg); 
    virtual void updateStateAt(UpdateStateData &usd);
    virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		    WAG AND RAISE/LOWER TAIL  SKILL
 *  args (optional):
 *	0 - speed of wag
 *	1 - up/down target for raising and lowering tail 
 *	2 - speed for (1) 
 *********************************************************************/

class WaggingAndMovingTail: public MotorSkill {
    public:
    float target;
    float targetz;
    float speedz;
    
    WaggingAndMovingTail(Dog *p,char *pname, float spx = .1, float spz = 0.0);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    int prepareToRun(void *arg); 
    virtual void updateStateAt(UpdateStateData &usd);
    virtual bool_t toOrFromStream(XDR *xdrs);
};
/*********************************************************************
 * 
 *		    SITTING/STANDING SKILL
 *  args (optional):
 *	0 - target	(0 is standing,  1 is sitting)
 *	1 - speed
 * 
 *********************************************************************/

class Sitting: public MotorSkill {
    public:
    float target;
    Sitting(Dog *p,char *pname,float sp);
    int prepareToRun(void *arg);
    int modifyRun(void *arg);
    int prepareToTurnOff(void *);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual void staticSetup();
    virtual bool_t toOrFromStream(XDR *xdrs);
};


/*********************************************************************
 * 
 *		    BEGGING/STANDING SKILL
 *  args (optional):
 *	0 - target	(0 is standing,  1 is sitting)
 *	1 - speed
 * 
 *********************************************************************/
class Begging: public MotorSkill {
    public:
    float target;
    ABConstraint *abcBeg;
    Begging(Dog *p,char *pname,float sp);
    int prepareToRun(void *arg);
    int modifyRun(void *arg);
    int prepareToTurnOff(void *);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual void staticSetup();
    virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		    LIFT LEG  SKILL
 *  args (optional):
 *	0 - whichpaw	(0 left,  1 right)
 *	1 - speed
 *
 * 
 *********************************************************************/

class LiftLeg: public MotorSkill {
    public:
    float target;
    int whichPaw;
    int okToChangePaws;
    LiftLeg(Dog *p,char *pname,float sp);
    int prepareToRun(void *arg);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		    SHAKE PAW  SKILL
 *  args (optional):
 *	0 - whichpaw	(0 left,  1 right)
 *	1 - speed
 *
 * 
 *********************************************************************/

class ShakePaw: public MotorSkill {
    public:
    float target;
    int whichPaw;
    int okToChangePaws;
    ShakePaw(Dog *p,char *pname,float sp);
    int prepareToRun(void *arg);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual void staticSetup();
    virtual bool_t toOrFromStream(XDR *xdrs);
};
/*********************************************************************
 * 
 *		    CROUCH SKILL
 *  args (ignore for now):
 *	0 - whichpaw	(0 left,  1 right)
 *	1 - speed
 *
 * 
 *********************************************************************/

class Crouch: public MotorSkill {
    public:
    float target;
    int whichPaw;
    int okToChangePaws;
    ABConstraint *abcCrouch[3];
    Crouch(Dog *p,char *pname,float sp);
    int prepareToRun(void *arg);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual void staticSetup();
    virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		    LYINGDOWN SKILL
 *  args (ignore for now):
 *	0 - whichpaw	(0 left,  1 right)
 *	1 - speed
 *
 * 
 *********************************************************************/

class LyingDown: public MotorSkill {
    public:
    float target;
    int whichPaw;
    int okToChangePaws;
    SoRotationXYZ *backbone;
    ABConstraint *abcLD[2];
    LyingDown(Dog *p,char *pname,float sp);
    int prepareToRun(void *arg);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    virtual void updateStateAt(UpdateStateData &usd);
    virtual void staticSetup();
    virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		  RAISING/LOWERING HEAD SKILL
 *  args (optional):
 *      0 - target; desired alpha value
 *	1 - speed 
 *	
 *********************************************************************/

class LoweringAndRaisingHead:public MotorSkill {
 public:
  float target;
  LoweringAndRaisingHead(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *		  SHAKING AND TURNING HEAD SKILL
 *  args (optional):
 *	0 - speed      
 *      1 - whichMode (0 is shaking, 1 is turning)
 *      2 - target; desired alpha value
 *
 *********************************************************************/

class ShakingAndTurningHead:public MotorSkill {
 public:
  float target;
  int whichMode;
  ShakingAndTurningHead(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	TILT HEAD SKILL:MOVING HEAD X-AXIS
 *  args (optional):
 *      0 - target; desired alpha value
 *	1 - speed      
 *      2 - whichMode (0 is back-forth, 1 is rotation)
 *
 *********************************************************************/

class TiltingHead:public MotorSkill {
 public:
  float target;
  int whichMode;
  TiltingHead(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};



/*********************************************************************
 * 
 *		 SNIFFING SKILL
 *  args (optional):
 *	1 - speed      
 *
 *********************************************************************/
class Sniffing:public MotorSkill {
 public:
  Sniffing(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
};


/*********************************************************************
 * 
 *       ROTATING EARS		
 *  args (optional):
 *      0 - speed
 *	1 - position (0 is forward,  1 is back)
 *      2 - tilt (0 is straight,  1 is tilted)
 *
 *********************************************************************/

class RotatingEars:public MotorSkill {
 public:
  float target;
  float earLeft;
  float earRight;
  float earTilt;
  RotatingEars(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *      EYEBROW MOVEMENT
 *  args (optional):
 *      0 - speed
 *      1 - tilt (1 is scared,  0 is angry)
 *
 *********************************************************************/

class MovingEyeBrows:public MotorSkill {
 public:
  float targetl;
  float targetr;
  int whichMode;
  MovingEyeBrows(Dog *p,char *pname, float sp = 1.0);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd); 
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *      TEETH MOVEMENT
 *  args (optional):
 *      0 - speed
 *      1 - target (1 is dilated,  0 constricted)
 *
 *********************************************************************/

class ShowTeeth:public MotorSkill {
 public:
    float target;
    ShowTeeth(Dog *p,char *pname, float sp = .1);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    int prepareToRun(void *arg); 
    virtual void updateStateAt(UpdateStateData &usd);
    virtual bool_t toOrFromStream(XDR *xdrs); 
};

/*********************************************************************
 * 
 *      BUGEYES MOVEMENT
 *  args (optional):
 *      0 - speed
 *      1 - target (1 is dilated,  0 constricted)
 *
 *********************************************************************/

class BugEyes:public MotorSkill {
 public:
    float target;
    BugEyes(Dog *p,char *pname, float sp = .1);
    int prepareToTurnOff(void *arg);
    int modifyRun(void *arg);
    int prepareToRun(void *arg); 
    virtual void updateStateAt(UpdateStateData &usd);
    virtual bool_t toOrFromStream(XDR *xdrs); 
};

/*********************************************************************
 * 
 *	SHIFTING EYE SKILL
 *  args (optional):
 *	1 - speed      
 *
 *********************************************************************/
class ShiftingEyes:public MotorSkill {
 public:
  ShiftingEyes(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
};

/*********************************************************************
 * 
 *	LOOKAT SKILL
 *  args (optional):
 *	1, 2, 3 - x y and z in world coordinates     
 *
 *********************************************************************/
class LookAt:public MotorSkill {
 public:
  LookAt(Dog *p,char *pname, float sp = .1);
  SoGetMatrixAction *ma;
  SoPath *pathToEye;
  SbVec3f lookAtPoint;
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	LOOKAT SKILL
 *  args (optional):
 *	1, 2, 3 - x y and z in world coordinates     
 *
 *********************************************************************/
class LookAtAB:public MotorSkill {
 public:
  LookAtAB(Dog *p,char *pname, float sp = .1);
  SoGetMatrixAction *ma;
  SoPath *pathToEye;
  SoPath *pathToHead;
  SoPath *pathToNeck;
  SbVec3f lookAtPoint;
  SbVec3f lp_old;
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	SHAKEHEADAB SKILL
 *  args (optional):
 *	1 speed     
 *
 *********************************************************************/
class ShakeHeadAB:public MotorSkill {
 public:
  ShakeHeadAB(Dog *p,char *pname, float sp = .1);
  SoGetMatrixAction *ma;
  SoPath *pathToHead;
  SoPath *pathToNeck;
  SbVec3f lookAtPoint;
  SbVec3f leftLimit;
  SbVec3f rightLimit;
  SbVec3f lp_old;
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	NODHEADAB SKILL
 *  args (optional):
 *	1 speed     
 *
 *********************************************************************/
class NodHeadAB:public MotorSkill {
 public:
  NodHeadAB(Dog *p,char *pname, float sp = .1);
  SoGetMatrixAction *ma;
  SoPath *pathToHead;
  SoPath *pathToNeck;
  SbVec3f lookAtPoint;
  SbVec3f topLimit;
  SbVec3f bottomLimit;
  SbVec3f lp_old;
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	Bark
 *  args (optional):
 *        0 -speed
 *        1 - target
 *	  2 - name of sound
 *
 ********************************************************************/

class Bark:public MotorSkill {
 public:
  float target;
  int startedBark;
  Bark(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	OPEN/CLOSE MOUTH
 *  args (optional):
 *        0 -speed
 *        1 - whichMode (1 is chewing, 0 is either open/close
 *        2 - target
 *
 ********************************************************************/

class OpeningAndClosingMouth:public MotorSkill {
 public:
  float target;
  int whichMode;
  OpeningAndClosingMouth(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *	NODDING AND RAISING/LOWERING HEAD
 *  args (optional):
 *      0 - target; desired alpha value
 *      1 - whichMode (0 is shaking, 1 is turning)
 *	2 - speed      
 *
 *********************************************************************/

class NoddingAndLoweringHead:public MotorSkill {
 public:
  float target;
  int whichMode;
  NoddingAndLoweringHead(Dog *p,char *pname, float sp = .1);
  int prepareToTurnOff(void *arg);
  int modifyRun(void *arg);
  int prepareToRun(void *arg); 
  virtual void updateStateAt(UpdateStateData &usd);
  virtual bool_t toOrFromStream(XDR *xdrs);
};


#endif











