#include <iostream.h>
#include <stdio.h>
#include <Inventor/SoPath.h>
#include <Inventor/fields/SoSFPath.h>
#include "BaseBehaviorNode.h"
#include "BehaviorCreatureLib.h"
#include "BehaviorPerson.h"
#include "mediaCreatureRMCallbacks.h"
#include "MediaCreature.h"
#include "MediaLeaf.h"
#include "util.h" // macros defined in file live here
#include "mediaCreatureActionCallbacks.h"
#include "CreatureRM.h"
#include "CreatureBehavior.h"
#include "SensorLib.h"
#include "util.h" // macros defined in file live here

const float PREY_RANGE = 60.0;
const float PREDATOR_RANGE = 65.0;
const float PREDATOR_KILL_RANGE = 10.0;
const float PERSON_RANGE = 10;

/*
 * Function types:
 *	typedef SbBool (*rmFilterCB)(ReleasingMechanism *rm, 
 *	    BehaviorUpdateAction *action);
 *	typedef float (*rmWeightCB)(ReleasingMechanism *rm, 
 *	    BehaviorUpdateAction *action);
 *	typedef void (*activityCB)(BaseActivity *act, 
 *	    BehaviorUpdateAction *action);
 *
 * Use these to add callbacks:
 *	addCallback(const char *name, rmFilterCB func);
 *	
 */
SbBool wallProximityOnly(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
    CBUA(dua, action);
    CRM(creatureRM, rm);
    PRONOME_RM(p, rm);
    PFSENSOR(s, dua);
    RayData rd;

    if(s->range<1.0e6){
	rd.pt = s->closestPt;
	rd.range = s->range;
	if(TESTRANGE((&rd), creatureRM)){

	  SniffSensor *ss = (SniffSensor *)creatureRM->getSensor(dua->creature);
if (ss == NULL){
  cerr << "Sniff Sensor == NULL \n " << endl;
}
  RayData *rd2 = NULL;
	  if(ss){
	    rd2 = ss->getDataForClosestWithTag(100);
	    if (rd2 == NULL) {
//cerr << ((MediaCreature *)dua->creature)->filename.getValue().getString() << " Ray reads NULL \n" << endl;
	      p->status = Pronome::FAILEDFIND;
	      return FALSE;
	      }
	      SbVec3f diff = rd2->who->position() - dua->creature->position();
	      if (diff.length() < (float)70){
		p->status = Pronome::FAILEDFIND;
		return FALSE;
	      }
	      
	      p->object = NULL;
	      p->location = rd.pt;
	      p->status = Pronome::PASSEDFIND;
	      p->type = Pronome::LOCATION;
	      p->range = rd.range;
	      p->bearing = rd.bearing; // not really valid.  hardts.
	 
	      p->vel = rd.vel;
	      p->validVel = rd.validVel;

	      creatureRM->value.setValue(creatureRM->maxValue.getValue());
	    }
	  }
	  
	}
	else p->status = Pronome::FAILEDFIND;
	return (p->status == Pronome::PASSEDFIND)?TRUE : FALSE;
}

SbBool
preyWithinRange(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  PRONOME_RM(p, rm);

  MediaLeaf *creature = (MediaLeaf *)p->object;
  MediaLeaf *self = (MediaLeaf *) dua->creature;
  	 
 if (self->dominance.getValue() > creature->dominance.getValue())
	{
	  // calculate distance to creature
	   SbVec3f creature_position = creature->position();
	   SbVec3f my_position = self->position();
	   SbVec3f diff = creature_position - my_position;
	   float len = diff.length();
	   if (len <= PREDATOR_RANGE)
	     {
	       return TRUE;
	     }
	 }
     return FALSE;
}



SbBool
predatorWithinRange(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  PRONOME_RM(p, rm);

  MediaLeaf *creature = (MediaLeaf *)p->object;
  MediaLeaf *self = (MediaLeaf *) dua->creature;
  	 
 if (self->dominance.getValue() < creature->dominance.getValue() )
	{
	  // calculate distance to creature
	   SbVec3f creature_position = creature->position();
	   SbVec3f my_position = self->position();
	   SbVec3f diff = creature_position - my_position;
	   float len = diff.length();
	   if (len <= PREDATOR_RANGE)
	     {
	       return TRUE;
	     }
	 }
     return FALSE;
}

SbBool
predatorWithinKillRange(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  PRONOME_RM(p, rm);

  MediaLeaf *creature = (MediaLeaf *)p->object;
  MediaLeaf *self = (MediaLeaf *) dua->creature;
  	 
 if (self->dominance.getValue() > creature->dominance.getValue())
	{
	  // calculate distance to creature
	   SbVec3f creature_position = creature->position();
	   SbVec3f my_position = self->position();
	   SbVec3f diff = creature_position - my_position;
	   float len = diff.length();
	   if (len <= PREDATOR_KILL_RANGE)
	     {
	       return TRUE;
	     }
	 }
     return FALSE;
}

SbBool
objectWithinRange(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  PRONOME_RM(p, rm);

  BaseCreature *creature = (BaseCreature *)p->object;
  BaseCreature *self = (BaseCreature *) dua->creature;

  // calculate distance to creature
  SbVec3f creature_position = creature->position();
  SbVec3f my_position = self->position();
  SbVec3f diff = creature_position - my_position;
  float len = diff.length();
  if (len <= PERSON_RANGE)
    {
    	      return TRUE;
    }
  return FALSE;
}

SbBool
personExtendedLeft(ReleasingMechanism *rm, BehaviorUpdateAction *action)
{
  CBUA(dua, action);
  PRONOME_RM(p, rm);

  BehaviorPerson *person = (BehaviorPerson *)(p->object);
  return (person->extendedLeft.getValue()
	  && !(person->extendedRight.getValue()));

}

void
initMediaCreatureRMCallbackList()
{
  BaseBehaviorNode::addCallback("objectWithinRange", objectWithinRange);
  BaseBehaviorNode::addCallback("personExtendedLeft",personExtendedLeft);
  BaseBehaviorNode::addCallback("predatorWithinRange",predatorWithinRange);
  BaseBehaviorNode::addCallback("preyWithinRange",preyWithinRange);
  BaseBehaviorNode::addCallback("wallProx",wallProximityOnly);
  BaseBehaviorNode::addCallback("predatorKill",predatorWithinKillRange);

}


