#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include "BasicMotorSkills.h"
#include "BasicDefines.h"
#include "MotorCreature.h"
#include "GenericMotorControl.h"
/*************************************************************
 * 
 * 
 *************************************************************/
GenericMove::GenericMove(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    
}

int GenericMove::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	start_speed = parent->speed;
	time_since  = 0.0; 

	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_speed = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);
    }

//fprintf(stderr, "Prepare To run : ds %f, secs %f\n", desired_speed, secs_to_reach);    

    return okToTurnOn;
}


int GenericMove::modifyRun(void *arg)
{
	start_speed = parent->speed;
	time_since  = 0.0; 


	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_speed = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);

        return 1;  
}



int GenericMove::prepareToTurnOff(void *)
{
    int okToTurnOff;

    if ((fabs(parent->speed) <= .05) && (desired_speed == 0.0)) 
	okToTurnOff = 1;
    else 
	okToTurnOff = 0;

    
    if (okToTurnOff) {
	parent->speed = 0.0;
        desired_speed = 0.0;
        start_speed = 0.0;
        secs_to_reach = 0.0;
    }
	
    return okToTurnOff;
}

void GenericMove::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {

    time_since += usd.time;

    if (parent->speed != desired_speed) {
	if (time_since < secs_to_reach) {
	    float tmp = time_since / secs_to_reach;
	    parent->speed = start_speed + (tmp * (desired_speed - start_speed));
	} else {
	    parent->speed = desired_speed;
	}
    } else 
	turnOff();

    }

    MotorSkill::updateStateAt(usd);
    return;
}


bool_t GenericMove::toOrFromStream(XDR *xdrs)
{
  if (!xdr_float(xdrs, &desired_speed))
    return FALSE;
  if (!xdr_float(xdrs, &start_speed))
    return FALSE;
  if (!xdr_float(xdrs, &time_since))
    return FALSE;
  if (!xdr_float(xdrs, &secs_to_reach))
    return FALSE;
  return MotorSkill::toOrFromStream(xdrs);
}

/*********************************************************************
 * 
 * GenericTurn Skill
 * 
 *********************************************************************/

GenericTurn::GenericTurn(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
}

int GenericTurn::prepareToRun(void *arg)
{
    int okToTurnOn = 1;

    if(okToTurnOn){
	start_angle = parent->angle;
	time_since  = 0.0; 

	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_angle = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);
    }

//fprintf(stderr, "Prepare To run : da %f, secs %f\n", desired_angle, secs_to_reach);    

    return okToTurnOn;
}


int GenericTurn::modifyRun(void *arg)
{
	start_angle = parent->angle;
	time_since  = 0.0; 


	MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
	desired_angle = argVals->getFloatAt(0);
	secs_to_reach = argVals->getFloatAt(1);

        return 1;  
}



int GenericTurn::prepareToTurnOff(void *)
{
    int okToTurnOff;

    if (parent->angle == desired_angle)
	okToTurnOff = 1;
    else 
	okToTurnOff = 0;

    
    if (okToTurnOff) {
	parent->angle = desired_angle;
        secs_to_reach = 0.0;
    }
	
    return okToTurnOff;
}

void GenericTurn::updateStateAt(UpdateStateData &usd)
{
    if(state == MotorSkill::SKILL_ON) {

    time_since += usd.time;

    if (parent->angle != desired_angle) {
	if (time_since < secs_to_reach) {
	    float tmp = time_since / secs_to_reach;
	    parent->angle = start_angle + (tmp * (desired_angle - start_angle));
	} else {
	    parent->angle = desired_angle;
	}
    } else 
	turnOff();

    }

    MotorSkill::updateStateAt(usd);
    return;
}

bool_t GenericTurn::toOrFromStream(XDR *xdrs)
{
  if (!xdr_float(xdrs, &desired_angle))
    return FALSE;
  if (!xdr_float(xdrs, &start_angle))
    return FALSE;
  if (!xdr_float(xdrs, &time_since))
    return FALSE;
  if (!xdr_float(xdrs, &secs_to_reach))
    return FALSE;
  return MotorSkill::toOrFromStream(xdrs);
}

/*********************************************************************
 * 
 * PositionSkill
 *
 *	arg 0 x value
 *	arg 1 y value
 *	arg 2 z value
 * 
 *********************************************************************/

PositionSkill::PositionSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void PositionSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    ((MotorCreature *)parent)->setPosition(SbVec3f(argVals->getFloatAt(0),
						   argVals->getFloatAt(1),
						   argVals->getFloatAt(2)));
}


/*********************************************************************
 * 
 * getPositionSkill
 * 
 *********************************************************************/

getPositionSkill::getPositionSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void getPositionSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;

    SbVec3f pos;
    parent->getPosition(pos);
    retVals->setNextReturnArg(pos);

    return;
}

/*********************************************************************
 * 
 * getHeadingSkill
 * 
 *********************************************************************/

getHeadingSkill::getHeadingSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void getHeadingSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;

    SbVec3f head;
    parent->getHeading(head);
    retVals->setNextReturnArg(head);

    return;
}


/*********************************************************************
 * 
 * HeadingSkill
 *
 *	arg 0 x value
 *	arg 1 y value
 *	arg 2 z value
 * 
 *********************************************************************/

HeadingSkill::HeadingSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void HeadingSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    SbVec3f pt;

    parent->getPosition(pt);
    pt += SbVec3f(argVals->getFloatAt(0),argVals->getFloatAt(1),argVals->getFloatAt(2));	

    MotorCommandBlock mcb;
    mcb.cmd = GC_HEADTOWARD;
    mcb.setNextArgGeneric(pt[0]);
    mcb.setNextArgGeneric(pt[1]);
    mcb.setNextArgGeneric(pt[2]);
    ((MotorCreature *)parent)->controller->executeMotorCommand(mcb);

    return;
}

/*********************************************************************
 * 
 * HeadTowardSkill
 *
 *	arg 0 x value
 *	arg 1 y value
 *	arg 2 z value
 * 
 *********************************************************************/

HeadTowardSkill::HeadTowardSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void HeadTowardSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    SbMatrix V;
    float bearing, range;
    SbVec3f p, vn;
    
    // This rotates creature around Y axis
    vn = SbVec3f(argVals->getFloatAt(0),argVals->getFloatAt(1),argVals->getFloatAt(2));
    parent->calcBearingAndRange(vn, bearing,range);
    SoTransform *xf =((MotorCreature *)parent)->getBodyTransform();
    SbVec3f center = xf->center.getValue();
    xf->recenter(SbVec3f(0, 0, 0));
    xf->rotation.setValue(SbRotation(SbVec3f(0, 1, 0), -bearing)*
	xf->rotation.getValue());
    
    // This rotates it around Z axis
    parent->vecToBody(vn);
    float elevation = atan2(vn[1], vn[0]);
    xf->rotation.setValue(SbRotation(SbVec3f(0, 0, 1), elevation)*
	xf->rotation.getValue());
	
    // This will remove any un-wanted rotation around X
    SbVec3f wup(0, 1, 0), lup;
    parent->dirToBody(wup);
    lup = wup - wup.dot(SbVec3f(0, 0, 1))*SbVec3f(0, 0, 1);
    SbRotation rot(lup, wup);
    xf->rotation.setValue(rot* xf->rotation.getValue());
    if(lup[1]<0.0){
	printf("correcting up vector \n");
	xf->rotation.setValue(SbRotation(SbVec3f(1,0,0),M_PI)*
		xf->rotation.getValue());
    }
    xf->recenter(center);
    return;
}
/*********************************************************************
 * 
 *			GrabbedSkill (arguments are no longer optional)
 *	arg 0 is a pointer to a BaseCreature.  If NULL, just keep the
 *            motor skill turned on.
 *	arg 1 is a char * for a named attach point
 *	arg 2 is whether to warp (int) 
 *	arg 3 if warping, then this is the transform between the grabber and
 *            grabee.  If set to NULL, will use identity Matrix. (void *)
 *********************************************************************/

grabbedConstraint::grabbedConstraint() : CreatureConstraint()
{
    ma = new SoGetMatrixAction(SbViewportRegion());
}

void grabbedConstraint::updateStateAt(UpdateStateData &)
{
    ma->apply(grabberPath);
    SbMatrix grabberXworld = ma->getMatrix();
    SbMatrix gdXgr(grabbedXgrabber);
    SbMatrix grabbedXworld  = gdXgr.multRight(grabberXworld);
    grabbedObject->getBodyTransform()->setMatrix(grabbedXworld);

    return;
}

//
//
//
GrabberSkill::GrabberSkill(MotorCreature *p, char *pname) : MotorSkill(p,
pname)
{
    ma = new SoGetMatrixAction(SbViewportRegion());
}

int GrabberSkill::prepareToRun(void *arg)
{
    return modifyRun(arg);
}

int GrabberSkill::modifyRun(void *arg)
{
  int okToTurnOn = 1;
  MotorCommandBlock *argVals = (MotorCommandBlock *)arg;

  BaseCreature *creature = argVals->getBaseCreaturePtrAt(0);

  // If creature is NULL, we are simply pinging the skill to keep it on.
  if ((!creature) && state == MotorSkill::SKILL_ON)
    return 1;

  if (creature && !creature->isGrabbed.getValue())
    {
      // fixed memory leak, hardts
      grabbedConstraint *gd = new grabbedConstraint;
      gd->grabbedObject = creature; 

      SoTransform *bt = gd->grabbedObject->getBodyTransform(); 
      
      /* get path to grabber node */
      SbName grabberName(argVals->getCharPtrAt(1));
      gd->grabberPath = parent->getPathToName(parent, grabberName);
      
      if(gd->grabberPath)
	{
	  if (argVals->getIntAt(2)) // Whether to warp
	    {
	      if (argVals->getArgCount()== 4){
//		  SoTransform *trans = (SoTransform *)argVals->getVoidPtrAt(3);
//		  ma->apply(trans);
//		  gd->grabbedXgrabber = ma->getMatrix();
		   gd->grabbedXgrabber = argVals->arglist.args[3].matval();  
		}
	      else
		gd->grabbedXgrabber = SbMatrix::identity();
	    }
	  else
	    {
	      SoTransform *bt = gd->grabbedObject->getBodyTransform(); 
	      ma->apply(bt);
	      SbMatrix grabbedXworld = ma->getMatrix();
	      ma->apply(gd->grabberPath);
	      SbMatrix worldXgrabber = ma->getInverse();
	      // This is the transform to use to maintain constraint
	      gd->grabbedXgrabber = grabbedXworld.multRight(worldXgrabber);
	    }
	  
	  // set something in grabbed object
	  gd->grabbedObject->setIsGrabbed(TRUE, parent);
//	  printf("added %x to list of length %d in parent %x\n", gd,grabbedList.getLength(), parent );
	  grabbedList.append(gd);
	  parent->myCache->addConstraint(gd);
	}
      else 
	{
	  delete gd;
	  okToTurnOn = 0;
	}
    }
  else 
    okToTurnOn = 0;
  
  return okToTurnOn;
}
#ifdef 0
int GrabberSkill::prepareToTurnOff(void *arg)
{

    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;

    grabbedConstraint *gc;
    int len = grabbedList.getLength();
    int i;

    if (argVals->getArgCount() == 1)  {
	BaseCreature *c = (BaseCreature *)argVals->getBaseCreaturePtrAt(0);
	for(i=0;i<len; i++) {
	     gc = ((grabbedConstraint *)grabbedList[i]);
	     if (gc->grabbedObject == c) {
   	     	gc->grabbedObject->setIsGrabbed(FALSE);
	     	gc->grabberPath->unref();
             	parent->myCache->removeConstraint(gc);
	     	delete gc;
	      	grabbedList.remove(i);
//		printf("I GOT HERE \n");	
		if (len == 1) 
		    return 1;
		else 
		    return 0;
	     }
	}
	fprintf(stderr, "Release not find object\n");
	return 0;
     } else {

	for(i=0;i<len; i++) {
	     gc = ((grabbedConstraint *)grabbedList[i]);
   	     gc->grabbedObject->setIsGrabbed(FALSE);
	     gc->grabberPath->unref();
//	     printf("removing %x from list of length %d of parent %x\n", gc, len, parent);
             parent->myCache->removeConstraint(gc);
	     delete gc;
	}
	grabbedList.truncate(0);
//	for(i=0;i<len; i++) {
//	     grabbedList.remove(i);
//	}
	return 1;
	}
}
#endif

int GrabberSkill::prepareToTurnOff(void *arg)
{

    int len = grabbedList.getLength();
    // want it to be handled by blocked
    return (len)? 0 :1;
}

void GrabberSkill::updateStateAt(UpdateStateData &usd)
{
    if (state == MotorSkill::SKILL_ON || 
	state == MotorSkill::SKILL_OFF_BLOCKED){
	if(state == MotorSkill::SKILL_OFF_BLOCKED){
	    int i, len = grabbedList.getLength();
	    grabbedConstraint *gc = NULL;
	    for(i=0;i<len; i++) {
		gc = ((grabbedConstraint *)grabbedList[i]);
   		gc->grabbedObject->setIsGrabbed(FALSE);
		gc->grabberPath->unref();
//		printf("removing %x from list of length %d of parent %x\n", gc, len, parent);
		parent->myCache->removeConstraint(gc);
		delete gc;
	    }
	    grabbedList.truncate(0);      
	}
	if((!requestedThisTick) && (parent->runBehavior.getValue()))
	    turnOff();
    }
    MotorSkill::updateStateAt(usd);
}


/*********************************************************************
 * 
 * SetTransformSkill
 *
 *	arg 0 is an SbMatrix
 * 
 *********************************************************************/

SetTransformSkill::SetTransformSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void SetTransformSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    SbMatrix sbm = argVals->arglist.args[0].matval();  
    ((MotorCreature *)parent)->getBodyTransform()->setMatrix(sbm);
}

/*********************************************************************
 * 
 * PositionRelSkill
 *
 *	arg 0 is a pointer to an object
 *	arg 1 x value
 *	arg 2 y value
 *	arg 3 z value
 * 
 *********************************************************************/

PositionRelSkill::PositionRelSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void PositionRelSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    BaseCreature *cr = argVals->getBaseCreaturePtrAt(0);

    SbVec3f pos(argVals->getFloatAt(1),
		argVals->getFloatAt(2),
		argVals->getFloatAt(3));
    cr->vecToWorld(pos);

    ((MotorCreature *)parent)->setPosition(pos);
}


/*********************************************************************
 * 
 * HeadingRelSkill
 *
 *	arg 0 is a pointer to an object
 *	arg 1 x value
 *	arg 2 y value
 *	arg 3 z value
 * 
 *********************************************************************/

HeadingRelSkill::HeadingRelSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void HeadingRelSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    BaseCreature *cr = argVals->getBaseCreaturePtrAt(0);

    SbVec3f head(argVals->getFloatAt(1),
		 argVals->getFloatAt(2),
		 argVals->getFloatAt(3));
    cr->dirToWorld(head);

    MotorCommandBlock mcb;
    float x, y, z;

    head.getValue(x,y,z);
    mcb.cmd = GC_HEADING;
    mcb.setNextArgGeneric(x);
    mcb.setNextArgGeneric(y);
    mcb.setNextArgGeneric(z);

    ((MotorCreature *)parent)->controller->executeMotorCommand(mcb);
}

/*********************************************************************
 * 
 * SoundSkill
 *
 *	arg 0 name of sound to trigger (char *)
 * 
 *********************************************************************/
ASound::ASound(const char *n, MotorCreature *p)
{
	name = new SbName(n);
	tiggerflag = 0;
	parent = p;
	aftersound = (ASound *)NULL;
	volume = 1.0;
}

SbPList SoundSkill::sounds;
FILE   *SoundSkill::pipe = (FILE *)NULL;
float   SoundSkill::rangeMin = 0.0;
float   SoundSkill::rangeMax = 10000.0;


void
SoundSkill::connectToServer(const char* server)
{
  pipe = popen(server, "w");
  if (!pipe)
    {
      fprintf(stderr, "Couldn't open soundserver: %s\n", server);
    }
}

void 
SoundSkill::disconnectFromServer()
{
  pclose(pipe);
  pipe = NULL;
}


SoundSkill::SoundSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void
SoundSkill::setSoundVolume(const char* n, float vol)
{
  ASound* as = getSoundByName(n);
  if (!as) return;
  as->volume = vol;
}

ASound *
SoundSkill::getSoundByName(const char *n)
{
  int i, num = sounds.getLength();  
  ASound *as;
  for(i = 0; i < num; i++) {
	as = (ASound *)sounds[i];
	if (n == *as->name)
		return as;
  }
  fprintf(stderr, "SoundSkill::getSoundByName: No such sound %s loaded\n", n);
  return (ASound *)NULL;
}

void
SoundSkill::addSound(const char *n, const char *filename, MotorCreature *p)
{
  sounds.append(new ASound(n, p));

  if (pipe) 
	fprintf(pipe, "c %s %s\n", n, filename);
}

int
SoundSkill::isServerConnected()
{
  return (pipe != NULL);
}


/* 
   update sounds using the current view matrix.  This is useful for
   doing head-mounted stuff.  Will adjust balance and volume based on
   relative position of the sound creator relative to the view matrix.
   */
void
SoundSkill::updateSounds(SbMatrix &mat)
{
  if (!pipe)
	return;

  int i, num = SoundSkill::sounds.getLength();  
  ASound *as;

  for(i = 0; i < num; i++) {
	as = (ASound *)SoundSkill::sounds[i];

	if (as->tiggerflag) {
	 	fprintf(pipe, "t %s\n", as->name->getString());
		as->tiggerflag = 0;
	}

        // 
        // amplitude created from the range
	//
        SbVec3f vec;
	as->parent->getPosition(vec);
        mat.multVecMatrix(vec, vec);
	float distance, amplitude, balance;
	float relx, rely, relz;
	vec.getValue(relx, rely, relz);

        distance = vec.length() + 0.00000001;  // add small # to stop div by 0
        amplitude = 1.0 - ((distance - rangeMin) / rangeMax);
	if (amplitude > 1.0) amplitude = 1.0;  
	if (amplitude < 0.0) amplitude = 0.0;  
        balance = acosf(relz/distance)/M_PI;

	fprintf(pipe, "b %s %f %f\n", as->name->getString(),
				      balance, amplitude);
  }

  fflush(pipe);
}


/*
  update the sounds assuming a fixed system with speakers on either
  side of the screen and assuming a coordinate system where facing the
  screen, right is +x, toward screen +z and up +y, origin centered on
  screen but in front by a bit. 
*/
void
SoundSkill::updateSoundsFixedView()
{
  int person_tag = 300;
  
  if (!pipe)
	return;

  int i, num = SoundSkill::sounds.getLength();  
  ASound *as;
  
  for(i = 0; i < num; i++) {
	as = (ASound *)SoundSkill::sounds[i];
	BaseCreature* person = as->parent->myCache->findBaseCreatureWithTag(person_tag);
	if (!person)
	  {
	    fprintf(stderr, "SoundSkill::updateSoundsFixedView: can't find person in cache!\n");
	    return;
	  }
	
	
        // 
        // amplitude created from the range
	//
                
	SbVec3f person_pos;
	person->getPosition(person_pos);

	SbVec3f left_speaker = SbVec3f(-speaker_x, 0.0, speaker_z);
	SbVec3f right_speaker = SbVec3f(speaker_x, 0.0, speaker_z);

	float d2 = (left_speaker - person_pos).length();
	float d1 = (right_speaker - person_pos).length();
	

	float balance; /* 0.0 full left, 1.0 full right */
	float adj_balance; 
	float amplitude; /* [0,1] */
	float distance; /* distance from person */
	float radial; /* distance from center (origin) to creature */
	float x, y, z;

	SbVec3f vec;
	as->parent->getPosition(vec);
#if 0	
	vec.getValue(x, y, z);
	radial = vec.length(); /* distance to origin of coords */
	if (radial < 1.0) /* inches */
	  {
	    /* angle gets unstable, but we know it is in center */
	    balance = 0.5;
	  }
	else 
	  {
	    balance = acosf(x/radial)/M_PI; 
	  }
#endif

	// reflect dog into mirror space
	vec += SbVec3f(0.0, 0.0, speaker_z - vec[2]);
	radial = (vec - person_pos).length(); 
	x = vec[0] - person_pos[0];
	balance = acosf(x/radial)/M_PI; // angle between person and dog
	
	amplitude = as->volume;
	//adj_balance = 1.0 - (d1 * (1.0 - balance)) / 
		  (d2 * balance  + d1 * (1.0 - balance));
	// adj_balance = balance;
	
	float dist_adj = 1.0 - d2/d1;
	adj_balance = balance - 0.2 * dist_adj;

	// printf("d1: %f  d2: %f  adapted balance: %f\n", d1, d2, adj_balance);

	fprintf(pipe, "b %s %f %f\n", as->name->getString(),
				      adj_balance, amplitude);

	if (as->tiggerflag) {
	 	fprintf(pipe, "t %s\n", as->name->getString());
		as->tiggerflag = 0;
	}

  }

  fflush(pipe);
}

float SoundSkill::speaker_x = 80.0;
float SoundSkill::speaker_z = 80.0;


void SoundSkill::turnoffSound(const char *n)
{
  if (!pipe) return;
   fprintf(pipe, "o %s\n", n);
}

void SoundSkill::setLoopingSound(const char *n, int i)
{
    if (!pipe) return;
   fprintf(pipe, "l %s %d\n", n, i);
}

void SoundSkill::triggerSound(const char *n)
{
  if (!pipe) return;
  ASound* as = getSoundByName(n);
  if (!as) 
    return;
  else 
    as->tiggerflag = 1;
}


void SoundSkill::triggerSoundAfter(const char *n, const char *an)
{
  if (!pipe) return;
   fprintf(pipe, "a %s %s\n", n, an);
}

void SoundSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;

    if (argVals->cmd == GC_SOUND_ON) { 	
	    triggerSound(argVals->getCharPtrAt(0));
    } else if (argVals->cmd == GC_SOUND_ON_AFTER) { 	
	    triggerSoundAfter(argVals->getCharPtrAt(0), argVals->getCharPtrAt(1));
    } else if (argVals->cmd == GC_SOUND_OFF) {	
	    turnoffSound(argVals->getCharPtrAt(0));
    } else if (argVals->cmd == GC_SOUND_LOOP) { 	
	    setLoopingSound(argVals->getCharPtrAt(0), 
			    argVals->getIntAt(1));
    }	 
}


/*********************************************************************
 * 
 * getFloorPositionSkill
 * 
 *********************************************************************/

getFloorPositionSkill::getFloorPositionSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void getFloorPositionSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;

    SbVec3f pos;
    parent->getPosition(pos);
    pos[1] = 0.0;	
    retVals->setNextReturnArg(pos);

    return;
}



/*********************************************************************
 * 
 * getFloorHeadingSkill
 * 
 *********************************************************************/

getFloorHeadingSkill::getFloorHeadingSkill(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void getFloorHeadingSkill::updateImmediate(void *arg)
{
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;

    SbVec3f head;
    parent->getHeading(head);
    head[1] = 0.0;	
    retVals->setNextReturnArg(head);

    return;
}

/*********************************************************************
 * 
 * GetMotorSkillStateVector
 * 
 *********************************************************************/

GetMotorSkillStateVector::GetMotorSkillStateVector(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void GetMotorSkillStateVector::updateImmediate(void *arg)
{
    MotorCommandBlock *retVals = (MotorCommandBlock *)arg;

    MotorSkill *ms;
    int i;
    int msnum = parent->motorSkills.getLength();
//    int *sv = new int[msnum];
//   retVals->setNextReturnArg(msnum);
//    retVals->setNextReturnArg((void *)sv);
    for(i=0;i<msnum;i++){
	ms = (MotorSkill *)parent->motorSkills[i];
	if((ms->state == MotorSkill::SKILL_ON) ||
	    (ms->state == MotorSkill::SKILL_ALREADY_ON) ||
	    (ms->state == MotorSkill::SKILL_OFF_BLOCKED)){
	    
	    retVals->setNextReturnArg((char *)ms->name->getString());
	    retVals->setNextReturnArg((int)ms->state);
	}
    }

    return;
}

/*********************************************************************
 * 
 * ListCommands
 * 
 *********************************************************************/

ListCommands::ListCommands(MotorCreature *p, char *pname) : MotorSkill(p, pname)
{
    isImmediate = TRUE;
}

void ListCommands::updateImmediate(void *)
{
    parent->getController()->printCommandTable();
    return;
}








