#ifndef _BASIC_SKILLS_
#define _BASIC_SKILLS_

#include <stdio.h>
#include "MotorSkill.h"
#include <Inventor/SbLinear.h>
#include <Inventor/SbPList.h>
#include <Inventor/SbString.h>

class SoPath;
class SoGetMatrixAction;
/**********************************************************************
 *
 * GenericMove
 *
 * This MotorSkill implements moving in what ever direction the wheels are
 * pointing.
 *
 *  args (optional):
 *    0 - new desired speed
 *    1 - second to reach desired speed
 *
 **********************************************************************/

class GenericMove: public MotorSkill {
  public:
    float desired_speed;
    float start_speed;
    float time_since;
    float secs_to_reach;

    GenericMove(MotorCreature *p, char *pname);

    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg);
    void updateStateAt(UpdateStateData &usd);
    int modifyRun(void *args);
    bool_t toOrFromStream(XDR *xdrs);
};

/*********************************************************************
 * 
 *  GenericTURN
 *
 *  args (optional):
 *	0 - new desired angle
 *	1 - seconds to reach that angle
 *
 *********************************************************************/

class GenericTurn: public MotorSkill {
    public:

    float desired_angle;
    float start_angle;
    float time_since;
    float secs_to_reach;

    GenericTurn(MotorCreature *p, char *pname);

    int modifyRun(void *args);
    int prepareToTurnOff(void *arg);
    int prepareToRun(void *arg); 
    void updateStateAt(UpdateStateData &usd);
    bool_t toOrFromStream(XDR *xdrs);
};
/**********************************************************************
 *
 * getMotorSkillStateVector
 *
 **********************************************************************/

class GetMotorSkillStateVector: public MotorSkill {
  public:
    GetMotorSkillStateVector(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};

/**********************************************************************
 *
 * getMotorSkillStateVector
 *
 **********************************************************************/

class ListCommands: public MotorSkill {
  public:
    ListCommands(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};

/**********************************************************************
 *
 * SetTransformSkill
 * takes an SbMatrix and sets the body transform to it.
 *
 **********************************************************************/

class SetTransformSkill: public MotorSkill {
  public:
    SetTransformSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};


/**********************************************************************
 *
 * Position
 *
 **********************************************************************/

class PositionSkill: public MotorSkill {
  public:
    PositionSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};


/**********************************************************************
 *
 * getPosition
 *
 **********************************************************************/

class getPositionSkill: public MotorSkill {
  public:
    getPositionSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};


/**********************************************************************
 *
 * getFloorPosition
 *
 **********************************************************************/

class getFloorPositionSkill: public MotorSkill {
  public:
    getFloorPositionSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};


/**********************************************************************
 *
 * getHeading
 *
 **********************************************************************/

class getHeadingSkill: public MotorSkill {
  public:
    getHeadingSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};

/**********************************************************************
 *
 * getFloorHeading
 *
 **********************************************************************/

class getFloorHeadingSkill: public MotorSkill {
  public:
    getFloorHeadingSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};

/**********************************************************************
 *
 * Heading
 *
 **********************************************************************/

class HeadingSkill: public MotorSkill {
  public:

    HeadingSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};


/**********************************************************************
 *
 * HeadToward
 *
 **********************************************************************/

class HeadTowardSkill: public MotorSkill {
  public:

    HeadTowardSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
    
};

/********************************************************************
 *
 *				Grabber Skill
 *
 *******************************************************************/

class GrabberSkill;

class grabbedConstraint: public CreatureConstraint {
    friend class GrabberSkill;
    SoGetMatrixAction *ma;

    BaseCreature *grabbedObject;
    SoPath *grabberPath;
    SbMatrix grabbedXgrabber;

    grabbedConstraint();

public:
  virtual void updateStateAt(UpdateStateData &usd);
  BaseCreature* getGrabbedCreature() 
  {
    return grabbedObject;
  }

};

class GrabberSkill: public MotorSkill
{
public:
  SoGetMatrixAction *ma;
  SbPList grabbedList;

  GrabberSkill(MotorCreature *p, char *pname);
  int prepareToRun(void *arg);
  int prepareToTurnOff(void *arg);
  void updateStateAt(UpdateStateData &usd);
  int modifyRun(void *args);
};

/**********************************************************************
 *
 * PositionRel
 *
 * args: <BaseCreature> <x y z>
 *
 **********************************************************************/

class PositionRelSkill: public MotorSkill {
  public:
    PositionRelSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};

/**********************************************************************
 *
 * HeadingRel
 *
 * args: <BaseCreature> <x y z>
 *
 **********************************************************************/

class HeadingRelSkill: public MotorSkill {
  public:

    HeadingRelSkill(MotorCreature *p, char *pname);
    void updateImmediate(void *arg);
};

/**********************************************************************
 *
 * SoundSkill
 *
 **********************************************************************/
class SoundSkill;

class ASound {
  friend class SoundSkill;

  SbName        *name;
  int            tiggerflag;
  ASound        *aftersound;
  MotorCreature *parent;
  float volume;

  public:
    ASound(const char *n, MotorCreature *p);
};

class SoundSkill: public MotorSkill {
  
  static SbPList sounds;
  static ASound  *getSoundByName(const char *n);
  static float   rangeMin, rangeMax;
  static FILE    *pipe;
  static float speaker_x, speaker_z; // speaker pos for right speaker
  // left speaker is at -x, z
  
public:
  
  static void addSound(const char *n, const char *filename, MotorCreature *p);
  
  static void connectToServer(const char* server);
  static void disconnectFromServer();
  static void triggerSound(const char *n);
  static void triggerSoundAfter(const char *n, const char *an);
  static void setLoopingSound(const char *n, int i);
  static void turnoffSound(const char *n);
  static int isServerConnected();
  static void setSoundVolume(const char* n, float vol);
  
  static void updateSounds(SbMatrix &mat);
  static void updateSoundsFixedView();
  
  // sets the right speaker.  Left is -x,z.
  static void setSpeakerPositions(float x, float z)
  {
    speaker_x = x; 
    speaker_z = z;
  }
  
  static void setRange(float min, float max)   
  { 
    rangeMin = min;
    rangeMax = max; 
  };
  
  static void getRange(float *min, float *max) 
  { 
    *min = rangeMin;
    *max = rangeMax; 
  };
  
  SoundSkill(MotorCreature *p, char *pname);
  void updateImmediate(void *arg);
};

#endif








