#include <iostream.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoSearchAction.h>
#include "DistributedSwitch.h"
#include "GenericMotorControl.h"
#include "MotorCreature.h"

SO_NODE_SOURCE(DistributedSwitch);

DistributedSwitch::DistributedSwitch() : SoGroup()
{
  SO_NODE_CONSTRUCTOR(DistributedSwitch);
  SO_NODE_ADD_FIELD(showFor, (DistributedSwitch::MASTER));
  SO_NODE_DEFINE_ENUM_VALUE(MotorCreatureType, MASTER);
  SO_NODE_DEFINE_ENUM_VALUE(MotorCreatureType, DROID);
  SO_NODE_SET_SF_ENUM_TYPE(showFor, MotorCreatureType);
  showChildren = -1;
}

void
DistributedSwitch::initClass()
{
  SO_NODE_INIT_CLASS(DistributedSwitch, SoGroup, "SoGroup");
}

DistributedSwitch::~DistributedSwitch()
{
}

SbBool
#ifdef SO_VERSION_REVISION
DistributedSwitch::readInstance(SoInput *in, unsigned short flags)
#else
DistributedSwitch::readInstance(SoInput *in)
#endif
{
#ifdef SO_VERSION_REVISION
  if (SoGroup::readInstance(in, flags))
#else
  if (SoGroup::readInstance(in))
#endif
    {
      ref();
      return (1);
    }
  return (0);
}

void
DistributedSwitch::doAction(SoAction *action)
{
  if (showChildren == -1)
    {
      MotorCreature *mc = NULL;
      // get a pointer to the MotorCreature containing this switch

      SoSearchAction *sa = new SoSearchAction;
      SoPath *creaturePath;
      SoNode *tailNode;
      sa->setNode(this);
      sa->apply(globals.creatureCache->sceneGraph);
      creaturePath = sa->getPath();
      if (creaturePath)
	{
	  // walk up the path to get a creature
	  do
	    {
	      tailNode = creaturePath->getTail();
	      if (tailNode)
		{
		  if (tailNode->isOfType(MotorCreature::getClassTypeId()))
		    {
		      mc = (MotorCreature *) tailNode;
		      // find out whether this creature is a master or droid
		      if (((mc->motorCreatureType.getValue() == MotorCreature::MASTER) &&
			   (showFor.getValue() == DistributedSwitch::MASTER)) ||
			  ((mc->motorCreatureType.getValue() == MotorCreature::DROID) &&
			   (showFor.getValue() == DistributedSwitch::DROID)))
			{
			  cerr << "DistributedSwitch::doAction: setting showChildren to 1" << endl; 
			  cerr << "Creature ID was " << mc->registryNumber.getValue() << endl;
			  showChildren = 1;
			}
		      else
			{
			  cerr << "DistributedSwitch::doAction: setting showChildren to 0" << endl;
			  cerr << "Creature ID was " << mc->registryNumber.getValue() << endl;
 			  showChildren = 0;
			}
		    }
		}
	      else
		{
		  cerr << "DistributedSwitch::doAction: error getting tail node" << endl;
		}
	      creaturePath->pop();
	    }
	  while ((!mc) && tailNode);
	}
      else
	{
	  cerr << "DistributedSwitch::doAction: error getting creature path" << endl;
	}
    }

  if (showChildren > -1)
    {
      // means we set our parent (the switch) appropriately.
      // do the proper traversal

      if (showChildren == 1)
	{
	  // show our children. do proper traversal

	  int numIndices;
	  const int *indices;

	  if (action->getPathCode(numIndices, indices) == SoAction::IN_PATH)
	    children->traverse(action, 0, indices[numIndices - 1]);
	  else
	    children->traverse(action);
	}
//      else
//	{
//	  cerr << "DistributedSwitch:: skipping action" << endl;
//	}
    }
  else
    {
      cerr << "DistributedSwitch::doAction: did not find my parent creature" << endl;
    }
}

void
DistributedSwitch::GLRender(SoGLRenderAction *action)
{
  DistributedSwitch::doAction(action);
}

/*
void
DistributedSwitch::handleEvent(SoHandleEventAction *action)
{
  DistributedSwitch::doAction(action);
}
*/
