#include "GenericMotorControl.h"
#include "CreatureLib.h"
#include "MotorCreature.h"
#include "MotorSkill.h"
#include "AliveDispatch.h"
#include <stdio.h>
#include <assert.h>

/****************************************************************
                            ArgType
 ****************************************************************/

ArgType::ArgType(const ArgType &other)
{
  // Make a copy of another ArgType.
  internal.tag = other.internal.tag;
  switch(internal.tag) 
    {
    case ArgType::INT: 
      internal.ArgTypeInternal_u.i = other.internal.ArgTypeInternal_u.i; 
      break;
    case ArgType::FLOAT: 
      internal.ArgTypeInternal_u.f = other.internal.ArgTypeInternal_u.f;
      break;
    case ArgType::SBVEC3F: 
      internal.ArgTypeInternal_u.vec[0] = other.internal.ArgTypeInternal_u.vec[0];
      internal.ArgTypeInternal_u.vec[1] = other.internal.ArgTypeInternal_u.vec[1];
      internal.ArgTypeInternal_u.vec[2] = other.internal.ArgTypeInternal_u.vec[2];
      break;
    case ArgType::CHARPTR: 
      internal.ArgTypeInternal_u.cptr = other.internal.ArgTypeInternal_u.cptr;
      break;
    case ArgType::VOIDPTR:
      internal.ArgTypeInternal_u.vptr = other.internal.ArgTypeInternal_u.vptr;
      break;
    case ArgType::BASECREATUREPTR:
      internal.ArgTypeInternal_u.creature_registry_number = other.internal.ArgTypeInternal_u.creature_registry_number;
      break;
    case ArgType::MATRIX: {
	int i,j;
	for (i = 0; i < 16; i++){
	    internal.ArgTypeInternal_u.mat[i] = other.internal.ArgTypeInternal_u.mat[i];
	}
	break;
    }
    default:
      assert(0);
    }
}

void ArgType::print()
{ 
  switch(internal.tag) {
  case ArgType::INT : printf("%d\n", internal.ArgTypeInternal_u.i); break;
  case ArgType::FLOAT : printf("%f\n", internal.ArgTypeInternal_u.f); break;
  case ArgType::SBVEC3F : printf("%f %f %f\n", internal.ArgTypeInternal_u.vec[0],internal.ArgTypeInternal_u.vec[1], internal.ArgTypeInternal_u.vec[2] ); break;
  case ArgType::CHARPTR : printf("%s\n", internal.ArgTypeInternal_u.cptr); break;
  case ArgType::VOIDPTR : printf("%x\n", internal.ArgTypeInternal_u.vptr); break; 
  case ArgType::BASECREATUREPTR : printf("%d\n", internal.ArgTypeInternal_u.creature_registry_number); break; 
  }
}


ArgTypeArray::ArgTypeArray()
{
    count = 0;
    capacity = MAXARGS;
}

ArgTypeArray::ArgTypeArray(const ArgTypeArray& ata)
{
    int i;
    for(i=0;i<ata.count;i++)
	args[i] = ata.args[i];
    count = ata.count;
}
    
void ArgTypeArray::append(const ArgType& a)
{
    if(count<capacity){
	args[count] = a;
	count++;
    }
    else printf("capacity exceeded \n");
}

void ArgTypeArray::truncate()
{
    count = 0;
}

int ArgTypeArray::getLength()
{
    return count;
}
    
/***************************************************************
 * 
 *	    MotorCommandBlock
 * 
 **************************************************************/
MotorCommandBlock::MotorCommandBlock()
{
    initGeneric();
}

MotorCommandBlock::MotorCommandBlock(int theCmd)
{
    initGeneric();
    cmd = theCmd;
}
MotorCommandBlock::MotorCommandBlock(int theCmd, const ArgType &arg1)
{
    initGeneric();
    cmd = theCmd;
    setNextArgGeneric(arg1);
}
MotorCommandBlock::MotorCommandBlock(int theCmd, const ArgType &arg1, 
				     const ArgType &arg2)
{
    initGeneric();
    cmd = theCmd;
    setNextArgGeneric(arg1);
    setNextArgGeneric(arg2);
}

MotorCommandBlock::MotorCommandBlock(int theCmd,  const ArgType & arg1, 
				     const ArgType & arg2,const ArgType & arg3)
{
    initGeneric();
    cmd = theCmd;
    setNextArgGeneric(arg1);
    setNextArgGeneric(arg2);
    setNextArgGeneric(arg3);
}

MotorCommandBlock::MotorCommandBlock(const MotorCommandBlock& mcb)
{
  cmd = mcb.cmd;
  returnResult = mcb.returnResult;
  arglist = mcb.arglist;
  retlist = mcb.retlist;
}

MotorCommandBlock::~MotorCommandBlock()
{
  initGeneric();		// frees memory
}

MotorCommandBlock&
MotorCommandBlock::operator=(const MotorCommandBlock& mcb)
{
  initGeneric();		// clears memory
  cmd = mcb.cmd;
  returnResult = mcb.returnResult;
  arglist = mcb.arglist;
  retlist = mcb.retlist;
  return *this;
}

// this deletes the storage made for the argtypes.
// which is safe since WE created it.
void
MotorCommandBlock::initGeneric()
{
  arglist.truncate();
  retlist.truncate();
  
  cmd = GC_NONESPECIFIED;
  returnResult = 0;		// correct?  aries 3/22/95
}

// protected aux func
SbPList
MotorCommandBlock::copyPointerList(const SbPList& from)
{
  int i;
  SbPList new_list;
  for(i=0; i<from.getLength(); i++)
    {
      ArgType* old_entry = (ArgType*) from[i];
      ArgType* new_entry = new ArgType(*old_entry);
      new_list.append(new_entry);
    }
  return new_list;
}

void MotorCommandBlock::print(GenericController *gc)
 {
     switch(cmd){
	 case GC_NONESPECIFIED:{
	    printf("MotorCommand -> %s \n", "None Specified");
	    break;
	 }
	 case GC_NOTIMPLEMENTED:{
	    printf("MotorCommand -> %s \n", "Not Implemented");
	    break;
	 }
	 default: {
	     if(gc){
		SbName *pname = gc->getCommandName(cmd);
		printf("MotorCommand -> %s \n", pname->getString());
	     }
	     else
		printf("MotorCommand -> %d \n", cmd);
	}
    }
/*     if(getNumArgs())
       args.print("args"); */
 }

void MotorCommandBlock::printArgs(GenericController *gc, SbBool retArgsToo)
{
     switch(cmd){
	 case GC_NONESPECIFIED:{
	    printf("MotorCommand -> %s \n", "None Specified");
	    break;
	 }
	 case GC_NOTIMPLEMENTED:{
	    printf("MotorCommand -> %s \n", "Not Implemented");
	    break;
	 }
	 default: {
	     if(gc){
		SbName *pname = gc->getCommandName(cmd);
		printf("MotorCommand -> %s \n", pname->getString());
	     }
	     else
		printf("MotorCommand -> %d \n", cmd);
	}
    }
    int i;
    for(i=0;i<arglist.getLength();i++)
	arglist.args[i].print();
    if(retArgsToo == TRUE)
	for(i=0;i<retlist.getLength();i++)
	    retlist.args[i].print();
}

void  MotorCommandBlock::fillArgsFromVector(Vector &v)
{
    int i;
    for(i=0;i<v.m;i++){
	setNextArgGeneric(v.mm[i]);
    }
} 
 /***************************************************************
 * 
 *	    Generic Controller
 * 
 ****************************************************************/
GenericController::GenericController(MotorCreature *mc)
{
    motorCreature = mc;
    buildCommandTable();
}

int GenericController::executeMotorCommand(MotorCommandBlock &mcb)
{
  if (motorCreature->myCache->isDistributed == TRUE)
    if (motorCreature->motorCreatureType.getValue() == MotorCreature::MASTER)
      {
	AliveDispatch::serializeMCB(&mcb);
      }

  CommandData *cmd = NULL;
  cmd = getCommand(mcb.cmd);

  if(!cmd)
    return GC_NOTIMPLEMENTED;
  else
    return cmd->executeMotorCommand(motorCreature, mcb);
}

void GenericController::addCommand(CommandData *cd)
{
    CommandData *cmdToOverride = getCommand(cd->commandId);
    if(cmdToOverride){
	commandTable.remove(commandTable.find(cmdToOverride));
	delete cmdToOverride;
    }
    commandTable.append(cd);
}


void GenericController::addCommand(int num, const char *name, 
				   MotorSkill *skill,int turnOn, 
				   const SbPList &defs)
{
  addCommand(new CommandData(name, num, skill, turnOn, defs));
}

void GenericController::addCommand(int num, const char *name, 
				   MotorSkill *skill, int turnOn)
{
  SbPList x;
  addCommand(new CommandData(name, num, skill, turnOn, x));
}

void GenericController::addCommand(int num, const char *name,MotorSkill *skill,
				   int turnOn, const ArgType &def1)
{
  SbPList x;
  x.append(new ArgType(def1));
  addCommand(new CommandData(name, num, skill, turnOn, x));
}

void GenericController::addCommand(int num, const char *name,MotorSkill *skill,
				   int turnOn, const ArgType &def1, 
				   const ArgType &def2)
{
  SbPList x;
  x.append(new ArgType(def1));
  x.append(new ArgType(def2));
  addCommand(new CommandData(name, num, skill, turnOn, x));
}

void GenericController::addCommand(int num, const char *name,MotorSkill *skill,
				   int turnOn, const ArgType &def1, 
				   const ArgType &def2, const ArgType &def3)
{
  SbPList x;
  x.append(new ArgType(def1));
  x.append(new ArgType(def2));
  x.append(new ArgType(def3));
  addCommand(new CommandData(name, num, skill, turnOn, x));
}

void GenericController::addCommand(int num, const char *name,MotorSkill *skill,
				   int turnOn, const ArgType &def1, 
				   const ArgType &def2, const ArgType &def3,
				   const ArgType &def4)
{
  SbPList x;
  x.append(new ArgType(def1));
  x.append(new ArgType(def2));
  x.append(new ArgType(def3));
  x.append(new ArgType(def4));
  addCommand(new CommandData(name, num, skill, turnOn, x));
}

void GenericController::addCompoundCommand(int num, const char *name, 
					   int cmd1, int cmd2)
{
  int cmds[2];
  cmds[0] = cmd1;
  cmds[1] = cmd2;
  addCompoundCommand(num, name, cmds, 2);
}

void GenericController::addCompoundCommand(int num, const char *name, 
					   int cmd1, int cmd2, int cmd3)
{
  int cmds[3];
  cmds[0] = cmd1;
  cmds[1] = cmd2;
  cmds[2] = cmd3;
  addCompoundCommand(num, name, cmds, 3);
}

void GenericController::addCompoundCommand(int num, const char *name, int *ids,
					   int c)
{
  CommandData *cds[10];
  int i;
  for(i=0;i<c;i++)
    cds[i] = getCommand(ids[i]);
  CompoundCommandData *ccd = 
    new CompoundCommandData(name, num, (CommandData **)cds, c);
  addCommand(ccd);
}  

void GenericController::buildCommandTable()
{
}

void GenericController::printCommandTable()
{
  int numCommands = commandTable.getLength();
  int i;
  CommandData *ce;
  for(i=0;i<numCommands;i++){
    ce = (CommandData *)commandTable[i];
    printf("%s %d \n", ce->commandName->getString(), ce->commandId);
  }
}

int GenericController::getCommandId(const char *cmdname)
{
  int numCommands = commandTable.getLength();
  int i;
  CommandData *ce;
  for(i=0;i<numCommands;i++){
    ce = (CommandData *)commandTable[i];
    if(cmdname == *ce->commandName)
      return ce->commandId;
  }
  return -1;
}

CommandData *GenericController::getCommandData(const char *cmdname)
{
  int numCommands = commandTable.getLength();
  for(int i=0;i<numCommands;i++)
    {
      CommandData *ce = (CommandData *)commandTable[i];
      if(cmdname == *ce->commandName)
	return ce;
    }
  return NULL;
}

CommandData *GenericController::getCommand(int id)
{
  int numCommands = commandTable.getLength();
  int i;
  CommandData *ce;
  for(i=0;i<numCommands;i++){
    ce = (CommandData *)commandTable[i];
    if(id == ce->commandId)
      return ce;
  }
  return NULL;
}

SbName *GenericController::getCommandName(int index)
{
  int numCommands = commandTable.getLength();
  int i;
  CommandData *ce;
  for(i=0;i<numCommands;i++){
    ce = (CommandData *)commandTable[i];
    if(index == ce->commandId)
      return ce->commandName;
  }
  return NULL;
}

/***************************************************************
 * 
 *	    Suggestion Stuff
 * 
 ****************************************************************/
SuggestedMCB::SuggestedMCB(MotorCommandBlock &cmd, int typeOfSuggestion, 
    void *who, float howMuch,  SbBool persistent)
    :MotorCommandBlock(cmd)	

{
    type = typeOfSuggestion;
    owner = who;
    value = howMuch;
    isPersistent = persistent;
}
    
void GenericController::clearSuggestions()
{
    int i;
    SbPList sugg;
    for(i=0;i<currentSuggestions.getLength();i++){
	if(((SuggestedMCB *)currentSuggestions[i])->isPersistent)
	    sugg.append(currentSuggestions[i]);
	else
	    delete ((SuggestedMCB *)currentSuggestions[i]);
    }
    currentSuggestions.truncate(0);
    currentSuggestions = sugg;
}

void GenericController::removeSuggestions(int cmd)
{

  // Should we deal with isPersistent?

  Boolean doAgain = False;
  do {
    doAgain = False;
    for (int i = 0; i < currentSuggestions.getLength() ; i++)
      if (((SuggestedMCB *)currentSuggestions[i])->cmd == cmd)
	{
	  doAgain = True;
	  delete currentSuggestions[i];
	  currentSuggestions.remove(i);
	  break;
	}
  }while (doAgain);
}  

SuggestedMCB *GenericController::getSuggestionFor(int cmd)
{
    int i;
    for(i=0;i<currentSuggestions.getLength();i++){
	if(((SuggestedMCB *)currentSuggestions[i])->cmd == cmd)
	    return (SuggestedMCB *)currentSuggestions[i];
    }
    return NULL; 
}

SbPList GenericController::getListOfSuggestionsFor(int cmd)
{
    SbPList sugg;
    int i;
    for(i=0;i<currentSuggestions.getLength();i++){
	if(((SuggestedMCB *)currentSuggestions[i])->cmd == cmd)
	    sugg.append(currentSuggestions[i]);
    }
    return sugg; 
}

int GenericController::suggestMotorCommand(MotorCommandBlock &mcb, int kind, 
    void *who, float hm,  SbBool isPersistent)
{
    SuggestedMCB *smcb = getSuggestionFor(mcb.cmd);
    if(!smcb){
	smcb = new SuggestedMCB(mcb, kind, who, hm, isPersistent);
	currentSuggestions.append(smcb);
    }
    else {
	if(smcb->value < hm){
//	    printf("suggestion %s with value &f by %x replaced by %x with value %f\n", 
//		getCommand(smcb->cmd)->commandName->getString(), 
//		smcb->owner, smcb->value, 
//		who, hm);

	  // warning!! Bug?  No assignment op.  Copies ptrs.  Delete fails.
	    *smcb = SuggestedMCB(mcb, kind, who, hm, isPersistent);
	}
	else {
//	    printf("suggestion %s by %x with value %f not replaced by %x with value %f\n", 
//		getCommand(smcb->cmd)->commandName->getString(), 
//		smcb->owner, smcb->value, 
//		who, hm);  
	}
    }
    return 1;
}

int GenericController::executeSuggestions()
{
    int i, j;
    for(i=0;i<currentSuggestions.getLength();i++){
	if(((SuggestedMCB *)currentSuggestions[i])->type == SuggestedMCB::DO)
	    j = executeMotorCommand(*((SuggestedMCB *)currentSuggestions[i]));
    }
    return j;
}
    
/***************************************************************
 * 
 *	    Command Data
 * 
 ****************************************************************/
 
CommandData::CommandData(const char *name, int num, MotorSkill *sk, int turnOn,  const SbPList &defs) 
:defaults(defs)
{
  commandName = new SbName(name);
  commandId = num;
  skill = sk;
  turnOnByDefault = turnOn;
}

int CommandData::executeMotorCommand(MotorCreature *,MotorCommandBlock &mcb)
{
  if(turnOnByDefault)
    {
      /* fill in defaults */
      for (int i = mcb.arglist.getLength(); i < defaults.getLength(); i++)
	mcb.setNextArgGeneric(*((ArgType *)defaults[i]));
      
      /* now execute command */
      return skill->turnOn(&mcb);
    }
  else
    return skill->turnOff(&mcb);
}

CompoundCommandData::CompoundCommandData(const char *name, int num, 
					 CommandData**helpers, int c)
    :CommandData(name, num,  (MotorSkill *)NULL, 1, SbPList())
{
    int i;
    for(i=0;i<c;i++)
	subCommands.append(helpers[i]);
}

int CompoundCommandData::executeMotorCommand(MotorCreature *master, 
    MotorCommandBlock &mcb)
{
    int count = subCommands.getLength();
    int i, result;
    for(i=0;i<count;i++){
//	mcb.init();
	mcb.initGeneric();
	mcb.cmd = ((CommandData *)subCommands[i])->commandId;
	result = ((CommandData *)subCommands[i])->executeMotorCommand(master, mcb);
    }
//    mcb.init();
    mcb.initGeneric();   // Added.  hardts
    return result;
}
