
#ifndef GENERICMOTORCONTROL
#define GENERICMOTORCONTROL
#include "ArgTypeInternal.h"
#include "CreatureLib.h"
#include "/usr/local/include/tcl.h"
#include "../AliveApp/globals.h"   // see comment below about globals -- kbrussel 6/9/95
#include <Inventor/SbString.h>
#include <Inventor/SbPList.h>
#include <Inventor/SbLinear.h>
/***************************************************************
 *
 *	    GENERIC MOTOR ERRORS ETC...
 * 
 * 
 **************************************************************/

#define GC_NONESPECIFIED  -1
#define GC_OK -2
#define GC_NOTIMPLEMENTED -3
#define GC_END -4

/* This will go away as soon as the motor and behavior systems
   use the registryNumber instead of a BaseCreature * .
   This will remove need for includes of tcl.h and globals.h */
extern Globals globals;

/****************************************************************
 * 
 * 
 * 
 ****************************************************************/
class MotorCreature;
class GenericController;
class MotorSkill;
class MotorCommandBlock;
class AliveDispatch;

class ArgType {
public:
  enum Type {
    FLOAT, 
      INT, 
      SBVEC3F, 
      CHARPTR, 
      VOIDPTR,
      BASECREATUREPTR,
      MATRIX,
      };
private:
  ArgTypeInternal internal;
  // changed "Type t" to "int t" for added compatibility with
  // external definition of ArgTypeInternal -- kbrussel 6/9/95
  void check(int t){if(internal.tag != t) printf("wrong type error\n");}
  
public:
  // changed return value from Type to int for added compatibility
  // with external definition of ArgTypeInternal -- kbrussel 6/9/95
  int getType() {return internal.tag;}
  ArgType(){internal.ArgTypeInternal_u.i = -1; internal.tag = ArgType::INT;}
  ArgType(const ArgType &other);
  ArgType(int ii){internal.ArgTypeInternal_u.i = ii; internal.tag = ArgType::INT;}
  ArgType(float ff){internal.ArgTypeInternal_u.f = ff; internal.tag = ArgType::FLOAT;}
  ArgType(char *p){internal.ArgTypeInternal_u.cptr = p; internal.tag = ArgType::CHARPTR;}
  ArgType(void *p){internal.ArgTypeInternal_u.vptr = p; internal.tag = ArgType::VOIDPTR;}
  ArgType(SbVec3f sb)
    {
      sb.getValue(internal.ArgTypeInternal_u.vec[0], internal.ArgTypeInternal_u.vec[1], internal.ArgTypeInternal_u.vec[2]);
      internal.tag = ArgType::SBVEC3F;
    }
  ArgType(BaseCreature *p) {
    if (p)
      internal.ArgTypeInternal_u.creature_registry_number = p->registryNumber.getValue(); 
    else
      internal.ArgTypeInternal_u.creature_registry_number = -1;
    internal.tag = ArgType::BASECREATUREPTR;
  }
  ArgType(const SbMatrix &mat) {
    internal.tag = ArgType::MATRIX;
    int i,j;
    for (i = 0; i < 4; i++)
      {
	for (j = 0; j < 4; j++)
	  {
	    internal.ArgTypeInternal_u.mat[(4*i)+j] = mat[i][j];
	  }
      }
  }

  int &ival(){check(ArgType::INT);return internal.ArgTypeInternal_u.i;}
  float &fval(){check(ArgType::FLOAT);return internal.ArgTypeInternal_u.f;}
  SbVec3f vec3fval(){check(ArgType::SBVEC3F);return SbVec3f(internal.ArgTypeInternal_u.vec);}
  SbMatrix matval(){check(ArgType::MATRIX);return SbMatrix(internal.ArgTypeInternal_u.mat[0],
							   internal.ArgTypeInternal_u.mat[1],
							   internal.ArgTypeInternal_u.mat[2],
							   internal.ArgTypeInternal_u.mat[3],
							   internal.ArgTypeInternal_u.mat[4],
							   internal.ArgTypeInternal_u.mat[5],
							   internal.ArgTypeInternal_u.mat[6],
							   internal.ArgTypeInternal_u.mat[7],
							   internal.ArgTypeInternal_u.mat[8],
							   internal.ArgTypeInternal_u.mat[9],
							   internal.ArgTypeInternal_u.mat[10],
							   internal.ArgTypeInternal_u.mat[11],
							   internal.ArgTypeInternal_u.mat[12],
							   internal.ArgTypeInternal_u.mat[13],
							   internal.ArgTypeInternal_u.mat[14],
							   internal.ArgTypeInternal_u.mat[15]);}
  const char *&cptrval(){check(ArgType::CHARPTR);return internal.ArgTypeInternal_u.cptr;}
  const void *&vptrval(){check(ArgType::VOIDPTR);return internal.ArgTypeInternal_u.vptr;}
  BaseCreature *&bcptrval()
    {check(ArgType::BASECREATUREPTR);return globals.creatureCache->getCreatureFromRegistryNumber(internal.ArgTypeInternal_u.creature_registry_number);}
  
  void print();

  friend bool_t xdr_ArgType(XDR *, ArgType *);
  friend class AliveDispatch;
};

#define MAXARGS 32

class ArgTypeArray {
    public:
    ArgType args[MAXARGS];
    int count;
    int capacity;
    ArgTypeArray();
    ArgTypeArray(const ArgTypeArray& ata);
    void append(const ArgType& a);
    void truncate();
    int getLength();
};

/*
  aries 3/22/95
  The following objects, MotorCommandBlock and SuggestionMCB, use
  SbPList to store lists of ArgType pointers, which are malloc'd in
  this class.  ArgType's really have no reason to be pointers, and
  this was done just to use a premade dynamic sized array.  I cleaned
  up many of the memory leaks caused by this approach, but there may
  be more.  My suggestion for a REAL fix of this code, which will also
  speed it up by avoiding memory alloc syscalls, is to use a dynamic
  array of ArgType's, with no freestore allocation.  This is much
  safer and faster.  My DynArray<T> template is the right answer, but
  this will cause Makefile's to change and I want to make the
  transition smooth.
  */
class MotorCommandBlock
{
public:
  int cmd;
  int returnResult;
  ArgTypeArray arglist;
  ArgTypeArray retlist;
  
  MotorCommandBlock();
  MotorCommandBlock(int theCmd);
  MotorCommandBlock(int theCmd,  const ArgType & arg1);
  MotorCommandBlock(int theCmd,  const ArgType & arg1, const ArgType & arg2);
  MotorCommandBlock(int theCmd,  const ArgType & arg1, const ArgType &arg2, const ArgType & arg3);
  MotorCommandBlock(const MotorCommandBlock& mcb); 
  
  ~MotorCommandBlock(); 
  MotorCommandBlock& operator=(const MotorCommandBlock& mcb);
  
  /* new style commands */
  void setNextArgGeneric(const ArgType &value) {arglist.append(ArgType(value));}
  void setNextArgGeneric(float value) {arglist.append(ArgType(value));}
  void setNextArgGeneric(int value){arglist.append(ArgType(value));}
  void setNextArgGeneric(char *value){arglist.append(ArgType(value));}
  void setNextArgGeneric(void *value){arglist.append(ArgType(value));}
  void setNextArgGeneric(SbVec3f value){arglist.append(ArgType(value));}
  void setNextArgGeneric(BaseCreature *value){arglist.append(ArgType(value));}
  
  void setNextReturnArg(const ArgType &value) {retlist.append(ArgType(value));}
  void setNextReturnArg(float value) {retlist.append(ArgType(value));}
  void setNextReturnArg(int value){retlist.append(ArgType(value));}
  void setNextReturnArg(char *value){retlist.append(ArgType(value));}
  void setNextReturnArg(void *value){retlist.append(ArgType(value));}
  void setNextReturnArg(SbVec3f value){retlist.append(ArgType(value));}
  void setNextReturnArg(BaseCreature *value){retlist.append(ArgType(value));}
  
  void initGeneric(); // fixed memory leaks.  aries 3/22/95
  
  int getArgCount(){return arglist.getLength();}
  int getRetArgCount(){return retlist.getLength();}
  // changed from ArgType::Type to int for compatibility
  // with external definition of ArgTypeInternal. -- kbrussel 6/9/95
  int getRetTypeAt(int i){return retlist.args[i].getType();}
  
  // took the following 2 unused observers out of the interface
  // to avoid future better solution (non-ptr) to this code. aries 3/22/95

  // These are not unused.  See tclCreature.c++ hardts 3/23/95
  // I changed them to return (const ArgType &) instead of a pointer.
  const ArgType &getArg(int i) {return arglist.args[i];}
  const ArgType &getReturnArg(int i) {return retlist.args[i];}
    
  int	  getIntAt(int i){return (arglist.args[i].ival());}
  float getFloatAt(int i){return (arglist.args[i].fval());}
  SbVec3f getSbVec3fAt(int i){return (arglist.args[i].vec3fval());}
  const char * getCharPtrAt(int i){return (arglist.args[i].cptrval());}
  const void * getVoidPtrAt(int i){return (arglist.args[i].vptrval());}
  BaseCreature * getBaseCreaturePtrAt(int i)
  {return (arglist.args[i].bcptrval());}
  
  int	  getRetIntAt(int i){return (retlist.args[i].ival());}
  float getRetFloatAt(int i){return (retlist.args[i].fval());}
  SbVec3f getRetSbVec3fAt(int i){return (retlist.args[i].vec3fval());}
  const char * getRetCharPtrAt(int i){return (retlist.args[i].cptrval());}
  const void * getRetVoidPtrAt(int i){return (retlist.args[i].vptrval());}
  
  /* utility function */
  void  fillArgsFromVector(Vector &v);
  
  /* old style print */
  void print(GenericController *gc = NULL);
  
  /* new style print */
  void printArgs(GenericController *gc = NULL, SbBool retArgsToo = FALSE);

// auxilary functions to follow
protected:
  // copy an SbPList, but allocate new pointer for each ArgType in it.
  SbPList copyPointerList(const SbPList& from);
};

class SuggestedMCB : public MotorCommandBlock {
    public:
    enum SuggType {
	DO, 
	MODIFY, 
	POT_FIELD_MODIFY
    };
    int type;
    SbBool isPersistent;
    void *owner;
    float value;
    SuggestedMCB(MotorCommandBlock &cmd, int typeOfSuggestion, 
	void *who, float howMuch,  SbBool persistent=FALSE);
    
};

class CommandData
{
public:
  SbName *commandName;
  int commandId;
  MotorSkill *skill;
  int turnOnByDefault;
  SbPList defaults;
  
  CommandData(const char *name, int num, MotorSkill *sk, int onOff, 
	      const SbPList &defs);
  virtual int executeMotorCommand(MotorCreature *owner,MotorCommandBlock &mcb);
  SbPList getDefaults() {return defaults;}
};

class CompoundCommandData:public CommandData {
public:
  SbPList subCommands;
  CompoundCommandData(const char *name, int num, CommandData **cds, int c);
  virtual int executeMotorCommand(MotorCreature *owner,MotorCommandBlock &mcb);
};

class GenericController {
public:
  SbPList commandTable;
  SbPList currentSuggestions;
  MotorCreature *motorCreature;
  
  void printCommandTable();
  void addCommand(CommandData *cd);
  void addCommand(int num, const char *name);
  void addCommand(int num, const char *name, MotorSkill *skill, int onOff, 
		  const SbPList &defs);
  void addCommand(int num, const char *name, MotorSkill *skill, int onOff);
  void addCommand(int num, const char *name, MotorSkill *skill,int onOff, 
		  const ArgType &def1);
  void addCommand(int num, const char *name, MotorSkill *skill,int onOff, 
		  const ArgType &def1, const ArgType &def2);
  void addCommand(int num, const char *name, MotorSkill *skill,int onOff, 
		  const ArgType &def1,const ArgType &def2,
		  const ArgType &def3);
  void addCommand(int num, const char *name, MotorSkill *skill,int onOff, 
		  const ArgType &def1,const ArgType &def2,
		  const ArgType &def3,const ArgType &def4);
  void addCompoundCommand(int num, const char *name, int cmd1, int cmd2);
  void addCompoundCommand(int num,const char *name,int cmd1,int cmd2,int cmd3);
  void addCompoundCommand(int num, const char *name, int *ids, int c);   
  int getCommandId(const char *cmdname);
  CommandData *getCommandData(const char *cmdname);
  SbName *getCommandName(int index);

  CommandData *getCommand(int id);
  // WARNING: Be careful if more than one CommandData for each id.

  MotorSkill *getRequiredMotorSkill(int id) 
    {CommandData *cd = getCommand(id); return (cd)?cd->skill : NULL;}
  virtual void buildCommandTable();
  GenericController(MotorCreature *owner);
  void clearSuggestions();
  
  void removeSuggestions(int cmd);
  // Remove all suggestions which match cmd;
  
  virtual int executeSuggestions();
  virtual SuggestedMCB *getSuggestionFor(int cmd);
  SbPList getListOfSuggestionsFor(int cmd);
    virtual int suggestMotorCommand(MotorCommandBlock &mcb, int kind, 
				    void *who, float hm,  SbBool isPersistent = FALSE);
  virtual int executeMotorCommand(MotorCommandBlock &mcb);
};
#endif
