#include <Inventor/engines/SoCompose.h>
#include <Inventor/engines/SoElapsedTime.h>
#include <Inventor/engines/SoTimeCounter.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include <Inventor/SoPath.h>
#include <Inventor/SoLists.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include "MoreMotorSkills.h"
#include "BasicDefines.h"
#include "MotorCreature.h"
#include "RepSwitch.h"
#include "GenericMotorControl.h"

ChangeRep::ChangeRep(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void ChangeRep::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int i;
    int newRep = argVals->getIntAt(0);
    parent->changeRep(newRep); 
}


ResetAction::ResetAction(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void ResetAction::updateImmediate(void *arg)
{
    if(parent->resetCB != NULL)
	parent->resetCB(parent);
}

SetVecField::SetVecField(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetVecField::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    /* name of field */
    SbName fName(argVals->getCharPtrAt(0));
    SoSFVec3f *vf = (SoSFVec3f *)parent->getField(fName);
    SbVec3f nvf = argVals->getSbVec3fAt(1);
    if(vf){
	vf->setValue(nvf);
    }
}

SetBoolFieldGeneric::SetBoolFieldGeneric(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetBoolFieldGeneric::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    /* name of field */
    SbName fName(argVals->getCharPtrAt(0));
    SoSFBool *vf = (SoSFBool *)parent->getField(fName);
    SbBool nvf = argVals->getIntAt(1);
    if(vf){
	vf->setValue(nvf);
    }
}


SetFloatFieldGeneric::SetFloatFieldGeneric(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetFloatFieldGeneric::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    /* name of field */
    SbName fName(argVals->getCharPtrAt(0));
    SoSFFloat *vf = (SoSFFloat *)parent->getField(fName);
    float nvf = argVals->getFloatAt(1);
    if(vf){
	vf->setValue(nvf);
    }
}

SetVecFieldInNamedNode::SetVecFieldInNamedNode(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetVecFieldInNamedNode::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int bv = argVals->getIntAt(3);
    SoNode *root = (bv)?parent->behaviorRoot:parent;
    if(bv && !parent->behaviorRoot){
	printf("no behavior system installed \n");
	return;
    }
    SbName nName(argVals->getCharPtrAt(0));
    SoNode *node = root->getByName(nName);
    if(node){
	/* name of field */
	SbName fName(argVals->getCharPtrAt(1));
	SoSFVec3f *vf = (SoSFVec3f *)node->getField(fName);
	SbVec3f nvf = argVals->getSbVec3fAt(2);
	if(vf){
	    vf->setValue(nvf);
	}
    }
    else printf("not found \n");
}

SetBoolFieldInNamedNode::SetBoolFieldInNamedNode(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetBoolFieldInNamedNode::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int bv = argVals->getIntAt(3);
    SoNode *root = (bv)?parent->behaviorRoot:parent;
    if(bv && !parent->behaviorRoot){
	printf("no behavior system installed \n");
	return;
    }
    SbName nName(argVals->getCharPtrAt(0));
    SoNode *node = root->getByName(nName);
    if(node){
	/* name of field */
	SbName fName(argVals->getCharPtrAt(1));
	SoSFBool *vf = (SoSFBool *)node->getField(fName);
	SbBool nvf = argVals->getIntAt(2);
	if(vf){
	    vf->setValue(nvf);
	}
    }
    else printf("not found \n");
}

SetFloatFieldInNamedNode::SetFloatFieldInNamedNode(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void SetFloatFieldInNamedNode::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int bv = argVals->getIntAt(3);
    SoNode *root = (bv)?parent->behaviorRoot:parent;
    if(bv && !parent->behaviorRoot){
	printf("no behavior system installed \n");
	return;
    }
    SbName nName(argVals->getCharPtrAt(0));
    SoNode *node = root->getByName(nName);
    if(node){
	SbName fName(argVals->getCharPtrAt(1));
	SoSFFloat *vf = (SoSFFloat *)node->getField(fName);
	float nvf = argVals->getFloatAt(2);
	if(vf){
	    vf->setValue(nvf);
	}
    }
    else printf("not found \n");
}



GetVecField::GetVecField(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void GetVecField::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    /* name of field */
    SbName fName(argVals->getCharPtrAt(0));
    SoSFVec3f *vf = (SoSFVec3f *)parent->getField(fName);
    if(vf)
      argVals->setNextReturnArg(vf->getValue());
    else
      printf("Could not find node: %s\n",argVals->getCharPtrAt(0));
  }

GetBoolFieldGeneric::GetBoolFieldGeneric(MotorCreature *p, char *pname) 
:MotorSkill(p,pname)
{
  isImmediate = TRUE;
}

void GetBoolFieldGeneric::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    /* name of field */
    SbName fName(argVals->getCharPtrAt(0));
    SoSFBool *vf = (SoSFBool *)parent->getField(fName);
    if (vf)
      argVals->setNextReturnArg(vf->getValue());
    else
      printf("Could not find node: %s\n",argVals->getCharPtrAt(0));
}

GetFloatFieldGeneric::GetFloatFieldGeneric(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void GetFloatFieldGeneric::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    /* name of field */
    SbName fName(argVals->getCharPtrAt(0));
    SoSFFloat *vf = (SoSFFloat *)parent->getField(fName);
    if(vf)
      argVals->setNextReturnArg(vf->getValue());
    else
      printf("Could not find node: %s\n",argVals->getCharPtrAt(0));
}

GetVecFieldInNamedNode::GetVecFieldInNamedNode(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void GetVecFieldInNamedNode::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int bv = argVals->getIntAt(2);
    SoNode *root = (bv)?parent->behaviorRoot:parent;
    if(bv && !parent->behaviorRoot){
	printf("no behavior system installed \n");
	return;
    }
    SbName nName(argVals->getCharPtrAt(0));
    SoNode *node = root->getByName(nName);
    if(node){
	/* name of field */
	SbName fName(argVals->getCharPtrAt(1));
	SoSFVec3f *vf = (SoSFVec3f *)node->getField(fName);
	if(vf)
	  argVals->setNextReturnArg(vf->getValue());
	else
	  printf("Could not find field: %s\n",argVals->getCharPtrAt(1));
      }
    else
      printf("Could not find node: %s\n",argVals->getCharPtrAt(0));
}

GetBoolFieldInNamedNode::GetBoolFieldInNamedNode(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void GetBoolFieldInNamedNode::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int bv = argVals->getIntAt(2);
    SoNode *root = (bv)?parent->behaviorRoot:parent;
    if(bv && !parent->behaviorRoot){
	printf("no behavior system installed \n");
	return;
    }
    SbName nName(argVals->getCharPtrAt(0));
    SoNode *node = root->getByName(nName);
    if(node){
	/* name of field */
	SbName fName(argVals->getCharPtrAt(1));
	SoSFBool *vf = (SoSFBool *)node->getField(fName);
	if(vf)
	  argVals->setNextReturnArg(vf->getValue());
	else
	  printf("Could not find field: %s\n",argVals->getCharPtrAt(1));
      }
    else
      printf("Could not find node: %s\n",argVals->getCharPtrAt(0));
}

GetFloatFieldInNamedNode::GetFloatFieldInNamedNode(MotorCreature *p, char *pname) :
MotorSkill((MotorCreature *)p, pname)
{
    isImmediate = TRUE;
}

void GetFloatFieldInNamedNode::updateImmediate(void *arg)
{
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    int bv = argVals->getIntAt(2);
    SoNode *root = (bv)?parent->behaviorRoot:parent;
    if(bv && !parent->behaviorRoot){
	printf("no behavior system installed \n");
	return;
    }
    SbName nName(argVals->getCharPtrAt(0));
    SoNode *node = root->getByName(nName);
    if(node){
	/* name of field */
	SbName fName(argVals->getCharPtrAt(1));
	SoSFFloat *vf = (SoSFFloat *)node->getField(fName);
	if(vf)
	  argVals->setNextReturnArg(vf->getValue());
	else
	  printf("Could not find field: %s\n",argVals->getCharPtrAt(1));
      }
    else
      printf("Could not find node: %s\n",argVals->getCharPtrAt(0));
}

GetBoolFields::GetBoolFields(MotorCreature *p, char *pname) 
:MotorSkill(p,pname)
{
  isImmediate = TRUE;
}

void GetBoolFields::updateImmediate(void *arg)
{
    SoFieldList sfl;
    SoField *sf;
    SbBool v;
    parent->getFields(sfl);
    int i;
    SbName fn;
    MotorCommandBlock *argVals = (MotorCommandBlock *)arg;
    
    for(i=0;i<sfl.getLength();i++){
	sf = sfl[i];
	if(sf->isOfType(SoSFBool::getClassTypeId())){
	    parent->getFieldName(sf, fn);
	    v = ((SoSFBool *)sf)->getValue();
	    printf("%s %d \n", 
		fn.getString(), v);
	    argVals->setNextReturnArg((char *)fn.getString());
	}
    }
}
