#include <stdio.h>
#include <iostream.h>
#include <math.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoPath.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>
#include <Inventor/engines/SoInterpolate.h>
#include <Inventor/engines/SoCalculator.h>
#include <Inventor/engines/SoCompose.h>
#include <Inventor/nodes/SoRotationXYZ.h>
#include <Inventor/nodes/SoPendulum.h>
#include <Inventor/nodes/SoShuttle.h>
#include "AliveDispatch.h"
#include "ArticulatedBody.h"
#include "JointModel.h"
#include "MotorCreature.h"
#include "MotorLib.h"
#include "MoreMotorSkills.h"
#include "BasicDefines.h"
#include "BasicMotorSkills.h"
#include "SocketMan.h"
#include "ArgTypeArray_xdr.h"

extern int continuousMode;

/*********************************************************************
 * 
 *	    MotorCreature  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(MotorCreature);

void
MotorCreature::initClass()
{
    SO_NODE_INIT_CLASS(MotorCreature, BaseCreature, "BaseCreature");
}

MotorCreature::MotorCreature() : BaseCreature()
{
    SO_NODE_CONSTRUCTOR(MotorCreature);
    SO_NODE_ADD_FIELD(shouldTrace, (FALSE));
    SO_NODE_ADD_FIELD(runBehavior, (FALSE));
    SO_NODE_ADD_FIELD(traceBehavior, (FALSE));
    SO_NODE_ADD_FIELD(behaviorFileName, ("none"));
    SO_NODE_ADD_FIELD(useRealTime, (TRUE));
    SO_NODE_ADD_FIELD(defaultFwdSpeed, (16.0));
    SO_NODE_ADD_FIELD(defaultAngSpeed, (.01));
    SO_NODE_ADD_FIELD(turnBase, (1.0));
    SO_NODE_ADD_FIELD(sounds, (""));
    SO_NODE_ADD_FIELD(motorCreatureType, (MotorCreature::MASTER));
    SO_NODE_DEFINE_ENUM_VALUE(MotorCreatureType, MASTER);
    SO_NODE_DEFINE_ENUM_VALUE(MotorCreatureType, DROID);
    SO_NODE_SET_SF_ENUM_TYPE(motorCreatureType, MotorCreatureType);
    SO_NODE_ADD_FIELD(lastRemoteUpdateTick, (0.0));
    SO_NODE_ADD_FIELD(updateReps, (FALSE));
    controller = NULL;
    speed = 0.0;
    angle = 0.0;
}

MotorCreature::~MotorCreature()
{
}

void
MotorCreature::buildSounds()
{
  int i, num = sounds.getNum();
  if (num < 2)
	return;
  for(i = 0; i < num; i +=2) 
	SoundSkill::addSound(sounds[i].getString(),sounds[i+1].getString(),this);
}

#ifdef SO_VERSION_REVISION
SbBool MotorCreature::readInstance(SoInput *in, unsigned short flags) 
#else
SbBool MotorCreature::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (BaseCreature::readInstance(in, flags)) {
#else
    if (BaseCreature::readInstance(in)) {
#endif
	buildMotorSystem();
	buildBehaviorSystem();
        buildSounds();
	return 1;
    }
    else
	return 0;
}

/*
 * 
 *	Methods for motor system and behavior system 
 * 
 */

void MotorCreature::installXYZInterpolator(const char *name)
{
    SoRotationXYZ *sh =  (SoRotationXYZ *)getByName(name);
    SoInterpolateFloat *sif = (SoInterpolateFloat *)getConnectedEngine(&sh->angle);
    Dof *dof = new Dof(sif, sh);
    dofs.append(dof);
}

void MotorCreature::buildMotorSystem()
{
    getDOFInterpolators();

    //
    // NOTE: addMotor Skills may add things to the controller
    //       therefore the controller should be added first
    // 
    controller = addController();    
    addMotorSkills();
}

/* use this method to add class-specific controller */
GenericController *MotorCreature::addController()
{
    return new GenericController(this);
}
    
/* use this class to add class-specific motor skills */
void MotorCreature::addMotorSkills()
{
  PositionSkill *pos = new PositionSkill(this, "position");
  getController()->addCommand( GC_POSITION, "position", pos, TRUE, 
			       ArgType((float)DEFAULT_X), 
			       ArgType((float)DEFAULT_Y), 
			       ArgType((float)DEFAULT_Z));
  
  getPositionSkill *getpos = new getPositionSkill(this, "getPosition");
  getController()->addCommand( GC_GET_POSITION, "getPosition", getpos, TRUE);
  
  getFloorPositionSkill *getflpos = 
    new getFloorPositionSkill(this, "getFloorPosition");
  getController()->addCommand( GC_GET_FLOOR_POSITION, 
			       "getFloorPosition", getflpos, TRUE);
  
  HeadingSkill *head = new HeadingSkill(this, "heading");
  getController()->addCommand( GC_HEADING, "heading", head, TRUE, 
			       ArgType((float)DEFAULT_HEADING_X), 
			       ArgType((float)DEFAULT_HEADING_Y), 
			       ArgType((float)DEFAULT_HEADING_Z));
  
  getHeadingSkill *gethead = new getHeadingSkill(this, "getHeading");
  getController()->addCommand( GC_GET_HEADING, "getHeading", gethead, TRUE);
  
  getFloorHeadingSkill *getflhead = 
    new getFloorHeadingSkill(this, "getFloorHeading");
  getController()->addCommand( GC_GET_FLOOR_HEADING, "getFloorHeading", 
			       getflhead, TRUE);
  
  HeadTowardSkill *headtoward = new HeadTowardSkill(this, "headtoward");
  getController()->addCommand( GC_HEADTOWARD, "headtoward", headtoward, TRUE,
			       ArgType((float)DEFAULT_X), ArgType((float)DEFAULT_Y), 
			       ArgType((float)DEFAULT_Z));
  
  GrabberSkill *grab = new GrabberSkill(this, "grabskill");
  getController()->addCommand( GC_GRAB, "grab", grab, TRUE, 
			       ArgType((BaseCreature *)NULL),ArgType("none"),
			       ArgType((int)0));
  getController()->addCommand( GC_UNGRAB, "release", grab, FALSE,
			       ArgType((BaseCreature *)NULL),ArgType("none"),
			       ArgType((int)0));
  
  PositionRelSkill *posRel = new PositionRelSkill(this, "positionRel");
  getController()->addCommand( GC_POSITION_REL, "positionRel", posRel, TRUE,
			       ArgType((BaseCreature *)NULL),ArgType((float)0),
			       ArgType((float)0),ArgType((float)0));
  
  HeadingRelSkill *headRel = new HeadingRelSkill(this, "headingRel");
  getController()->addCommand( GC_HEADING_REL, "headingRel", headRel, TRUE,
			       ArgType((BaseCreature *)NULL),ArgType((float)0),
			       ArgType((float)0),ArgType((float)0));
  
  SoundSkill *snd = new SoundSkill(this, "sound");
  getController()->addCommand( GC_SOUND_ON, "soundOn", snd, TRUE, "none");
  getController()->addCommand( GC_SOUND_OFF, "soundOff", snd, TRUE, "none");
  getController()->addCommand( GC_SOUND_LOOP, "soundLoop", snd, TRUE, "none");
  getController()->addCommand( GC_SOUND_ON_AFTER, "soundOnAfter", snd, TRUE,
			       "none");
  
  GenericMove *move = new GenericMove(this, "move");
  getController()->addCommand( GC_MOVE_GEN, "move", move, TRUE, 
			       ArgType((float)DEFAULT_SPEED_GEN), 
			       ArgType((float)DEFAULT_SECS_TO_SPEED_GEN));
  
  GenericTurn *turn = new GenericTurn(this, "turn");
  getController()->addCommand( GC_TURN_GEN, "turn", turn, TRUE, 
			       ArgType((float)DEFAULT_TURN_GEN), 
			       ArgType((float)DEFAULT_SECS_TO_TURN_GEN));
  
  SetVecField *svf = new SetVecField(this, "setvecfield");
  getController()->addCommand( GC_SETVECFIELD_GEN, "setVec", svf, TRUE,
			       ArgType("none"),ArgType(SbVec3f()));
  
  SetBoolFieldGeneric *sbf = 
    new SetBoolFieldGeneric(this, "setboolfieldgen");
  getController()->addCommand( GC_SETBOOLFIELD_GEN, "setBool", sbf, TRUE,
			       ArgType("none"),ArgType((int)0));
  
  SetFloatFieldGeneric *sff = 
    new SetFloatFieldGeneric(this, "setfloatfieldgen");
  getController()->addCommand( GC_SETFLOATFIELD_GEN, "setFloat", sff, TRUE,
			       ArgType("none"),ArgType((float)0));
  
  SetVecFieldInNamedNode *svfn = 
    new SetVecFieldInNamedNode(this, "setvecfieldinnode");
  getController()->addCommand( GC_SETVECFIELD_NAMED, "setVecInNode", 
			       svfn, TRUE,ArgType("none"),
			       ArgType("none"),ArgType(SbVec3f()), 
			       ArgType((int)0));
  
  SetBoolFieldInNamedNode *sbfn = 
    new SetBoolFieldInNamedNode(this, "setboolfieldinnode");
  getController()->addCommand( GC_SETBOOLFIELD_NAMED, "setBoolInNode", 
			       sbfn, TRUE,ArgType("none"),
			       ArgType("none"),ArgType((int)0),
			       ArgType((int)0));
  
  SetFloatFieldInNamedNode *sffn = 
    new SetFloatFieldInNamedNode(this, "setfloatfieldinnode");
  getController()->addCommand( GC_SETFLOATFIELD_NAMED, "setFloatInNode", 
			       sffn, TRUE,ArgType("none"),
			       ArgType("none"),ArgType((float)0), 
			       ArgType((int)0));
  
  getController()->addCommand( GC_SETVECFIELD_BEVNAMED, 
			       "setVecInBehaviorNode", svfn, TRUE, 
			       ArgType("none"),
			       ArgType("none"),ArgType(SbVec3f()),
			       ArgType((int)1));
  getController()->addCommand( GC_SETBOOLFIELD_BEVNAMED, 
			       "setBoolInBehaviorNode", sbfn, TRUE, 
			       ArgType("none"),
			       ArgType("none"),ArgType((int)0),
			       ArgType((int)1));
  getController()->addCommand( GC_SETFLOATFIELD_BEVNAMED, 
			       "setFloatInBehaviorNode", sffn, TRUE,
			       ArgType("none"),
			       ArgType("none"),ArgType((float)0),
			       ArgType((int)1),);
  
  GetVecField *gvf = new GetVecField(this, "getvecfield");
  getController()->addCommand( GC_GETVECFIELD_GEN, "getVec", gvf, TRUE,
			       ArgType("none"));
  
  GetBoolFieldGeneric *gbf = 
    new GetBoolFieldGeneric(this, "getboolfieldgen");
  getController()->addCommand( GC_GETBOOLFIELD_GEN, "getBool", gbf, TRUE,
			       ArgType("none"));
  
  GetFloatFieldGeneric *gff = 
    new GetFloatFieldGeneric(this, "getfloatfieldgen");
  getController()->addCommand( GC_GETFLOATFIELD_GEN, "getFloat", gff, TRUE,
			       ArgType("none"));
  
  GetVecFieldInNamedNode *gvfn = 
    new GetVecFieldInNamedNode(this, "getvecfieldinnode");
  getController()->addCommand( GC_GETVECFIELD_NAMED, "getVecInNode", 
			       gvfn, TRUE,ArgType("none"),
			       ArgType("none"), ArgType((int)0));
  
  GetBoolFieldInNamedNode *gbfn = 
    new GetBoolFieldInNamedNode(this, "getboolfieldinnode");
  getController()->addCommand( GC_GETBOOLFIELD_NAMED, "getBoolInNode", 
			       gbfn, TRUE,ArgType("none"),
			       ArgType("none"), ArgType((int)0));
  
  GetFloatFieldInNamedNode *gffn = 
    new GetFloatFieldInNamedNode(this, "getfloatfieldinnode");
  getController()->addCommand( GC_GETFLOATFIELD_NAMED, "getFloatInNode", 
			       gffn, TRUE,ArgType("none"),
			       ArgType("none"), ArgType((int)0));

  getController()->addCommand( GC_GETVECFIELD_BEVNAMED, 
			       "getVecInBehaviorNode", gvfn, TRUE, 
			       ArgType("none"),
			       ArgType("none"),ArgType((int)1));
  getController()->addCommand( GC_GETBOOLFIELD_BEVNAMED, 
			       "getBoolInBehaviorNode", gbfn, TRUE, 
			       ArgType("none"),
			       ArgType("none"),ArgType((int)1));
  getController()->addCommand( GC_GETFLOATFIELD_BEVNAMED, 
			       "getFloatInBehaviorNode", gffn, TRUE, 
			       ArgType("none"),
			       ArgType("none"),ArgType((int)1));
  
   GetMotorSkillStateVector *gmssv = 
     new GetMotorSkillStateVector(this, "get-state-vector");
   getController()->addCommand( GC_GETMSVECTOR, 
			       "getActiveMotorSkills", gmssv, TRUE);
   
   ListCommands *lcmds = new  ListCommands(this, "list-commands");
   getController()->addCommand( GC_LISTCOMMANDS, "?", lcmds, TRUE);
   
   ResetAction *ra = new  ResetAction(this, "reset-creature");
   getController()->addCommand( GC_RESETCOMMAND, "reset", ra, TRUE);
   
   GetBoolFields *gbfs = new  GetBoolFields(this, "get-bool-fields");
   getController()->addCommand( GC_GETBOOLFIELDS, "getboolfields", gbfs, TRUE);
   
   ChangeRep *cr = new  ChangeRep(this, "change-Rep");
   getController()->addCommand( GC_CHANGEREP, "changeRep", cr, TRUE, ArgType((int)-1000));
   
   SetTransformSkill *sts = new  SetTransformSkill(this, "setTransform");
   SbMatrix sbm;
   sbm.makeIdentity();
   getController()->addCommand( GC_SETTRANSFORM, "setTransform", sts, TRUE, ArgType((SbMatrix)sbm));

}

/* use this class to add class-specific behavior system */
void MotorCreature::buildBehaviorSystem()
{
    SoInput mySceneInput;
    behaviorRoot = NULL;
    if(behaviorFileName.getValue().getString() != SbString("none")){
	if (!mySceneInput.openFile(behaviorFileName.getValue().getString())) {
	   printf("Cannot open behavior graph file: %s\n", 
	    behaviorFileName.getValue().getString());
	   return;
	}
	behaviorRoot = SoDB::readAll(&mySceneInput);
	behaviorRoot->ref();
	if(!behaviorRoot){
	    printf("problem reading in behavior graph \n");
	    return;
	}
    }
    return;
}

int MotorCreature::checkIfReadyToMove(MotorCommandBlock &)
{
    return 1;
}

int MotorCreature::checkIfNeedToChangePosture(MotorCommandBlock &, int)
{
    return 0;
}

int MotorCreature::checkIfNeedToChangePosture(MotorCommandBlock &, MotorSkill *)
{
    return 0;
}

void MotorCreature::traceBehaviorSystemAt(UpdateStateData &)
{

}

SoAction *MotorCreature::getBehaviorUpdateAction()
{
    return NULL;
}

/* use this method to install class-specific interpolators */
void MotorCreature::getDOFInterpolators()
{
    
}

/* use this method to update class-specific behavior system */
void MotorCreature::updateBehaviorSystemAt(UpdateStateData &)
{
    
}

void MotorCreature::updateMotorSkillsAt(UpdateStateData &usd)
{
    int i;
    /* now let the motor skills execute */
    for(i=0;i<motorSkills.getLength();i++)
      ((MotorSkill *)motorSkills[i])->updateStateAt(usd);
    /*
     *  NEW CODE HERE
     */
    if(updateReps.getValue() == TRUE)
	updateRepsAt(usd);
}

MotorSkill *MotorCreature::getMotorSkillByName(const char *name)
{
    int i;
    /* now let the motor skills execute */
    for(i=0;i<motorSkills.getLength();i++)
      if (name == *((MotorSkill *)motorSkills[i])->name) 
	return ((MotorSkill *)motorSkills[i]);
}

SbBool MotorCreature::beginBehaviorSystemUpdate(UpdateStateData &usd)
{
  if (myCache->isDistributed == FALSE)
    {
      if (runBehavior.getValue() == TRUE)
	return TRUE;
      return FALSE;
    }
  else
    {
      if ((runBehavior.getValue() == TRUE) &&
	  (motorCreatureType.getValue() == MotorCreature::MASTER))
	{
	  AliveDispatch::beginMCBSerialization(this);
	  return TRUE;
	}
      return FALSE;
    }
}

void MotorCreature::endBehaviorSystemUpdate(UpdateStateData &usd)
{
  if (myCache->isDistributed == TRUE)
    if ((runBehavior.getValue() == TRUE)&&
	(motorCreatureType.getValue() == MotorCreature::MASTER))
      {
	AliveDispatch::endMCBSerialization();
      }
  return;
}

void MotorCreature::updateStateAt(UpdateStateData &usd)
{
    /* do normal update (i.e. the Sensor ) */
    BaseCreature::updateStateAt(usd);
    
    if(beginBehaviorSystemUpdate(usd)){
	updateBehaviorSystemAt(usd);
	if(traceBehavior.getValue())
	    traceBehaviorSystemAt(usd);
    }
    else
      {
	if ((myCache->isDistributed == TRUE) &&
	    (motorCreatureType.getValue() == MotorCreature::DROID))
	  {
	    int32_t creature_id;
	    int num_mcbs;
	    int counter;
	    char mcb_type;
	    float tick_number;

	    // get up to one block of MCBs off 
	    // the list and execute them.

	    if (remoteMCBList.getLength() > 0)
	      {
		SocketBuf *sb = (SocketBuf *) remoteMCBList[0];
		remoteMCBList.remove(0);
		
		// set up an XDR stream on this chunk of memory

		XDR xdrs;
		xdrmem_create(&xdrs,
			      (caddr_t) sb->buf, 
			      SB_MAX_UDP_PACKET_SIZE,
			      XDR_DECODE);

		if (!xdr_char(&xdrs, &mcb_type))  // ignored here
		  {
		    cerr << "MotorCreature::updateStateAt: error reading type" << endl;
		    xdr_destroy(&xdrs);
		    return;
		  }
		long creature_id_long;
		if (!xdr_long(&xdrs, &creature_id_long))  // ignored here
		  {
		    cerr << "MotorCreature::updateStateAt: error getting creature id" << endl;
		    xdr_destroy(&xdrs);
		    return;
		  }
		else creature_id = creature_id_long;
		// check to see whether we've dropped a packet
		
		if (!xdr_float(&xdrs, &tick_number))
		  {
		    cerr << "MotorCreature::updateStateAt: error reading tick number" << endl;
		    return;
		  }
		else
		  {
		    if (lastRemoteUpdateTick.getValue() != 0)
		      {
			if (continuousMode == 1)
			  {
			    // we know from examining idleCB in main.c++ that
			    // the updates should be in increments of 0.1.
			    
			    if (fabs(tick_number - (lastRemoteUpdateTick.getValue() + 0.1)) > 0.001)
			      {
				cerr << "MotorCreature::updateStateAt: lost a packet for creature " << creature_id <<
				  "; last update was at " << lastRemoteUpdateTick.getValue() << ", incoming update " <<
				    "is for " << tick_number << endl;
			      }
			  }
			else
			  {
			    if (tick_number != (lastRemoteUpdateTick.getValue() + 1))
			      {
				cerr << "MotorCreature::updateStateAt: lost a packet for creature " << creature_id <<
				  "; last update was at " << lastRemoteUpdateTick.getValue() << ", incoming update " <<
				    "is for " << tick_number << endl;
			      }
			  }
		      }
		    lastRemoteUpdateTick.setValue(tick_number);
		  }

//  cerr << "Updating creature " << creature_id << " on tick " << myCache->wallClock << endl;

		if (!xdr_int(&xdrs, &num_mcbs))
		  {
		    cerr << "MotorCreature::updateStateAt: error getting number of MCBs" << endl;
		    xdr_destroy(&xdrs);
		    return;
		  }

		for (counter = 0; counter < num_mcbs; counter++)
		  {
		    MotorCommandBlock mcb;
		    AliveDispatch::flushMCB(&mcb);
		    if (!xdr_MotorCommandBlock(&xdrs, &mcb))
		      {
			cerr << "MotorCreature::updateStateAt: error getting next motor command block" << endl;
			xdr_destroy(&xdrs);
			return;
		      }
		    getController()->executeMotorCommand(mcb);
		  }

		// destroy the XDR stream
		xdr_destroy(&xdrs);
		delete sb;

		updatedRemotelyThisTick = 1;
	      }
	    else
	      {
		updatedRemotelyThisTick = 0;
	      }
	  }
      }

    endBehaviorSystemUpdate(usd);
    if ((myCache->isDistributed == FALSE) ||
	((myCache->isDistributed == TRUE) && motorCreatureType.getValue() == MASTER) ||
	((myCache->isDistributed == TRUE) &&
	 (motorCreatureType.getValue() == MotorCreature::DROID) &&
	 (updatedRemotelyThisTick == 1)))
      updateMotorSkillsAt(usd);
    return;
}

int MotorCreature::stepTo(UpdateStateData &usd)
{
  if ((myCache->isDistributed == FALSE) ||
      ((myCache->isDistributed == TRUE) && motorCreatureType.getValue() == MASTER) ||
      ((myCache->isDistributed == TRUE) &&
       (motorCreatureType.getValue() == MotorCreature::DROID) &&
       (updatedRemotelyThisTick == 1)))
    {
      float t = usd.time;
      float distance = (useRealTime.getValue())? speed * t :speed;

      SbVec3f dir(1, 0, 0);
      if(speed != 0.0 || angle != 0.0){
	if(angle == 0.0){
	  dirToWorld(dir);
	  dir.normalize();
	  setPosition(position() + dir*distance);
	}
	else {
	  float delta;
	  if(turnBase.getValue()>0.0)
	    delta = (distance/turnBase.getValue())*ftan(angle);
	  else
	    delta = distance*sin(angle);
	  //printf("delta -> %f \n", delta);
	  setHeading(SbRotation(SbVec3f(0, 1, 0), delta)
		     *getBodyTransform()->rotation.getValue());
	  dirToWorld(dir);
	  dir.normalize();
	  setPosition(position() + dir*distance);
	}
      }
      isMoving.setValue((speed != 0.0)?TRUE:FALSE);
    }
  return 1;  
}

void MotorCreature::restart()
{
}

/*
int MotorCreature::isMoving()
{
  for(int i=0;i<3;i++){
    if(fabs(vel.v[i])>.1 || fabs(vel.v0[i])>.1)
      return 1; 
  }
  return 0;
}
*/
void MotorCreature::getVelocity(Spv &v)
{
    v = vel;
    return;
} 

SoNode *MotorCreature::getSensorOfType(SoType stype)
{
    return (SoNode *)getFirstTDCOfType(stype);
}

SoNode *MotorCreature::getSensorWithName(SbName &aName)
{
    return (SoNode *)getFirstTDCWithName(aName);
}

bool_t
MotorCreature::syncToOrFromStream(XDR *xdrs)
{
  int counter;
  
  if (!BaseCreature::syncToOrFromStream(xdrs))
    return FALSE;

  // read/write the state of the motor skills
  // note that this automatically takes care of
  // reading and writing the state of the degrees of freedom.

  for (counter = 0; counter < motorSkills.getLength(); counter++)
    {
      if (!((MotorSkill *) motorSkills[counter])->toOrFromStream(xdrs))
	return FALSE;
    }

  return TRUE;
}
