#ifndef _BASESENSOR_
#define _BASESENSOR_

#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/fields/SoMFVec3f.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoTransform.h>
#include "CreatureLib.h"
#include <Inventor/SbColor.h>
class SoPath;
class SoSwitch;
class Timestat;

class BaseSensor : public TimeDependentComponent {
    SO_NODE_HEADER(BaseSensor);
    SoSFBool shouldTime;
    
public:
    static void initClass();

    SoTransform *sensorFrame;
    SoPath *pathToCreature;
    SoPath *pathToWorld;
    
    /*
     * provided for use of subclasses. Up to them to set values during _updateStateAt
     */
    SbBool detected;
    float range;
    float bearing;
    SbVec3f closestPt;
    
    
    void initializeComponent(SoGroup *root);
    void vecFromSensorFrameUsing(SbVec3f &ep, SoPath *p);
    void vecToSensorFrameUsing(SbVec3f &ep, SoPath *p);
    SbVec3f getVecInSensorFrame(const SbVec3f pt);
    SbVec3f getVecInWorldFrame(const SbVec3f pt);
    SbVec3f calcBearingAndRange(SbVec3f pt, float &bearing, float &range);
    void prepareForUpdateAt(UpdateStateData &usd);
    void endUpdateAt(UpdateStateData &usd);
    void updateStateAt(UpdateStateData &usd);
    virtual void _updateStateAt(UpdateStateData &usd); // over-ride this method
    

protected:
    SbMatrix lXw;	    // world to local transform (updated by cacheMatrices)
    SbMatrix wXl;	    // local to world transform (updated by cacheMatrices)
    SoSwitch *swNode;	    // first switch node in creature (turned off during ray cast)
    void cacheMatrices();   // called as part of updateStateAt()
    int cachedSwitchValue;
    Timestat *timer;
    
    BaseSensor();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool	readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool	readInstance(SoInput *in);
#endif
    virtual ~BaseSensor();

};

#endif

