#include <stdio.h>

#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/elements/SoCacheElement.h>
#include <Inventor/projectors/SbPlaneProjector.h>
#include "CreatureLib.h"
#include "DepthMapSensor.h"
#include <GL/gl.h>
/*********************************************************************
 * 
 *	    DepthMapSensor  Class
 * 
 *********************************************************************/


SO_NODE_SOURCE(DepthMapSensor);

void
DepthMapSensor::initClass()
{
    SO_NODE_INIT_CLASS(DepthMapSensor, CamSensor, "CamSensor");
}

DepthMapSensor::DepthMapSensor() : CamSensor()
{
    SO_NODE_CONSTRUCTOR(DepthMapSensor);
}

DepthMapSensor::~DepthMapSensor()
{
}


SbBool	
#ifdef SO_VERSION_REVISION
DepthMapSensor::readInstance(SoInput *in, unsigned short flags) 
#else
DepthMapSensor::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (CamSensor::readInstance(in, flags)) {
#else
    if (CamSensor::readInstance(in)) {
#endif
	int i;
	SbVec2f s = size.getValue();
	depthData = new DepthData[(int)s[0]];
	for(i=0;i<(int)s[0];i++){
	    depthData[i].xyn = SbVec2f((float)i/s[0], 0.0);
	}
	return (1);
    }
    else {
	return(0);
   }
}

void DepthMapSensor::_updateStateAt(UpdateStateData &usd)
{
    int i;
    static int ft = 1;
    static SoOffscreenRenderer *sor = NULL;
    
    CamSensor::_updateStateAt(usd);
    
    if(!sor){
	SbVec2f vpsize = size.getValue();
	SbViewportRegion vp;
	vp.setWindowSize(SbVec2s(vpsize[0], vpsize[1]));
	sor = new SoOffscreenRenderer(vp);
	sor->setComponents(SoOffscreenRenderer::LUMINANCE);
    }
    SbBool ok = FALSE;
    ok = sor->render(ra->getSceneGraph());
    if(!ok){
	 printf("OFFSCREEN RENDER FAILED\n");
    }
    else {
	SbVec2f vpsize = size.getValue();
	genDepthDataAt(vpsize[1],vpsize[0],sor->getBuffer());
    }
    return;
}

void DepthMapSensor::genDepthDataAt(int h, int w, unsigned char *data)
{
    int hi, wj, offset;
    SbPlane plane(SbVec3f(0, 1, 0), 0);
    SbPlaneProjector pp(plane);
    SbViewVolume vv; 
    pp.setViewVolume(camera->getViewVolume());
    
    //printf("NEW FRAME \n");
    for(wj=0;wj<w;wj++){
	depthData[wj].fnd = FALSE;
	for(hi=0;(hi<h && (!depthData[wj].fnd));hi++){
	    offset = hi*w + wj;
	    if(((unsigned short)data[offset])>(unsigned short)0){
		depthData[wj].xyn[1] = ((float)hi/(float)h);
		depthData[wj].xyz = pp.project(depthData[wj].xyn);
		lXw.multVecMatrix(depthData[wj].xyz, depthData[wj].xyz);
		depthData[wj].fnd = TRUE;
		/*
		printf("nonfloor found at %f %f %f for column %d \n", 
		    depthData[wj].xyz[0], 
		    depthData[wj].xyz[1],
		    depthData[wj].xyz[2], wj);
		*/
	    }
	    else {
		depthData[wj].xyn[1] = 1;
	    }
	}
    }
}
