#include <stdio.h>

#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoTranslation.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoQuadMesh.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>

#include "PotentialField.h"
#include "../CreatureLib/CreatureLib.h"

#define DEFAULTDECAY 1.0

/*********************************************************************
 * 
 *	    PotentialField  Class
 * 
 *********************************************************************/
PotentialField::PotentialField(int dimx, int dimz)
{
    commonInit(dimx, dimz, DEFAULTDECAY);
}

PotentialField::PotentialField(int dimx, int dimz, float decayValue)
{
     commonInit(dimx, dimz, decayValue);
}

void PotentialField::commonInit(int dimx, int dimz, float decayValue)
{
    pfield = new Matrix(dimx, dimz);
    gradMag = new Matrix(dimx/2, dimz/2);
    grad = new Matrix(dimx/2, dimz/2);
    depth = dimx;
    width = dimz;
    decay = decayValue;
    buildVisualizer();
}

#define SX (-16.0)
#define S1 4.0
    
void PotentialField::buildVisualizer()
{
    int sx = (int)((-S1)*(grad->m/2));
    visualizer = new SoSeparator;
    coords = new SoCoordinate3;
    lines = new SoLineSet;
    SoBaseColor *bc = new SoBaseColor();
    bc->rgb.setValue(1, 0, 0);
    
    visualizer->addChild(bc);
    visualizer->addChild(coords);
    visualizer->addChild(lines);
    
    int i, j, k= 0;
    SbVec3f tmp, tmp1;
    for(i=0;i<(depth/2)*(depth/2);i++)
	lines->numVertices.set1Value(i, 2);
    /* now build coords */
    for(i=0;i<depth/2;i++){
	for(j = 0;j<width/2;j++){
	    tmp = SbVec3f((float)(sx + j*S1), (float)(-sx - i*S1), 0.0);
	    tmp1 = tmp + SbVec3f(0, 1, 0);
	    coords->point.set1Value(k++, tmp);
	    coords->point.set1Value(k++, tmp1);
	}
    }
    
    
    
    k=0;
    coords1 = new SoCoordinate3;
   // SoLineSet *lines1 = new SoLineSet;
    SoQuadMesh *qm = new SoQuadMesh;
    SoMaterial *mat = new SoMaterial;
    mat->diffuseColor.setValue(0, .7, 0);
    mat->specularColor.setValue(1, 1, 1);
    mat->shininess.setValue(.8);
    qm->verticesPerColumn.setValue(depth/2);
    qm->verticesPerRow.setValue(width/2);
    qm->startIndex.setValue(0);
    visualizer->addChild(mat);
    visualizer->addChild(coords1);
    visualizer->addChild(qm);
    
//    for(i=0;i<(depth/2)*(depth/2);i++)
//	lines1->numVertices.set1Value(i, 2);
    /* now build coords */
    for(i=0;i<depth/2;i++){
	for(j = 0;j<width/2;j++){
	    tmp = SbVec3f((float)(sx + j*S1), (float)(-sx - i*S1), 0.0);
	    tmp1 = tmp + SbVec3f(0, 1, 0);
	    coords1->point.set1Value(k++, tmp);
//	    coords1->point.set1Value(k++, tmp1);
	}
    }
    
    SoCube *cube = new SoCube;
    SoSeparator *sep = new SoSeparator;
    cube->width = cube->height  = 32.0;
    cube->depth = .1;
    SoTranslation *xf = new SoTranslation;
    xf->translation.setValue(0, 0, -.2);
    mat = new SoMaterial;
    mat->diffuseColor.setValue(.4, .4,.4);
    mat->transparency.setValue(.5);
    
    sep->addChild( xf);
    sep->addChild( mat);
    sep->addChild( cube);
    visualizer->addChild(sep); 
}

void PotentialField::clear()
{
    int i, j;
    float w2 = floor(width/2.0);
    float cw;
    for(i=0;i<depth;i++)
	for(j=0;j<width;j++){
	    cw = 1.0 - fabs((j-w2)/w2);
	    pfield->mm[i][j] = cw*(float)((i-depth)/4.0);
	}
    //pfield->print("after clear");
}


void PotentialField::addPotentialField(Matrix &pf1)
{
    *pfield = *pfield + pf1;
}

void PotentialField::scalePotentialField(float scaleFactor)
{
    float sf = scaleFactor;
    *pfield = (*pfield) * sf;
}
void PotentialField::addPotentialFieldAtOffset(Matrix &pf1, int d, int w)
{
    
    // This function adds another potential field to this one at an offset
    // w,d from the center of this field.
    // It is assumed that it is either the same size or bigger.
    // The start,stop conditions in the for loops below are to limit the
    // operation to the valid range of this field.
    
    // d and w are offsets from center of potential field. d is negative
    // to reflect sense
    d = depth/2 - d;
    w = width/2 + w;
    
    int pf1dim = pf1.m;
    int pfCenter = pf1dim/2;
    int i0, i1, j0, j1;
    int startW0, startW1;
    int startD0, startD1;
    int endW = ((w + pfCenter)>pfield->n)?pfield->n :w + pfCenter ;
    int endD = ((d + pfCenter)>pfield->m)?pfield->m :d + pfCenter ;
    
    if(w-pfCenter<0){
	startW0 = 0;
	startW1 = abs(w-pfCenter);
    }
    else {
	startW0 = w-pfCenter;
	startW1 = 0;
    }
    if(d-pfCenter<0){
	startD0 = 0;
	startD1 = abs(d-pfCenter);
    }
    else {
	startD0 = d-pfCenter;
	startD1 = 0;
    }
    for(i0 = startD0, i1 = startD1; i0<endD ;i0++, i1++)
	for(j0 = startW0, j1 = startW1; j0<endW ;j0++, j1++)
	    pfield->mm[i0][j0] += pf1.mm[i1][j1];
   // pfield->print("pfield after add");
}

void PotentialField::addGradient(Matrix &pg1)
{
    *grad = *grad + pg1;
}

void PotentialField::addGradientAtOffset(Matrix &grad1, int d, int w)
{
    
    // This function adds another gradient to this one at an offset
    // w,d from the center of this field.
    // It is assumed that it is either the same size or bigger.
    // The start,stop conditions in the for loops below are to limit the
    // operation to the valid range of this field.
    
    int grad1dim = grad1.m;
    int grad1Center = grad1dim/2;
    int i, j;
    int startW = w - grad1Center;
    int startD = d - grad1Center;
    int endW = w + grad1Center;
    int endD = d + grad1Center;
	
    for(i=(startD<0)?abs(startD):0; i<((endD<=grad->m)?grad1dim :
	(grad1dim-(endD- grad->m)));i++)
	for(j=(startW<0)?abs(startW):0; j<((endW<=pfield->n)?grad1dim :
	    (grad1dim-(endW- grad->n)));j++)
	     grad->mm[i][j] += grad1.mm[startD + i][startW + j];
}


SbBool PotentialField::spreadPotential(SbVec3f lpt, int maxPotential)
{
    // reverse sense of lpt[0] to acccount for structure of
    // pfield (bottom rows correspond to nearer depths
    //    lpt /= 1.25;
    int spread = abs((int)(maxPotential/decay));
    lpt[0] = depth - lpt[0] - depth/2;
    
    // shift over by width/2 so that center is in center of pfield
    lpt[2] += width/2;
    
    // test if it will have an effect. If so, then spread potential
    if(!((lpt[2]< -spread) ||
	lpt[2]> width + spread ||
	lpt[0]> depth + spread)){
	_spreadPotential((maxPotential<0?TRUE : FALSE), (int)lpt[2], (int)lpt[0],spread, spread );
	return TRUE;    
    }
    else
	return FALSE;
}

/*
void PotentialField::spreadPotential(int seedx, int seedy, int pvalue)
{
    _spreadPotential(FALSE, seedx,seedy,pvalue, pvalue);
}
*/

void PotentialField::_spreadPotential(SbBool ifAttractor, int seedx, int seedy, int pvalue, int iter)
{
    int i;
    int j;
    if(iter>1)
	_spreadPotential(ifAttractor, seedx,seedy,pvalue, iter-1);
    for(i= seedx-pvalue+iter;i<=seedx+pvalue-iter;i++)
	for(j=seedy-pvalue+iter;j<=seedy+pvalue-iter;j++)
	    if(((i>=0) && (i<width)) 
		&& ((j>=0) && (j<depth))
		/* && (pfield->mm[j][i] < pvalue) */)
		    pfield->mm[j][i] += (ifAttractor)?-1.0/decay:1.0/decay;
}

float PotentialField::convolveMaskAt(Matrix &mask, int d, int w)
{
    return convolveMaskAt(mask, d, w, 0, 0);
}
float PotentialField::convolveMaskAt(Matrix &mask, int d, int w, int skipr, int skipc)
{
    int maskDim = mask.m;
    float result = 0.0;
    if(maskDim % 2){
	int maskCenter = maskDim/2;
	int i, j;
	int startW = w - maskCenter;
	int startD = d - maskCenter;;
	
	for(i=0;i<maskDim;i+= 1 + skipr)
	    for(j=0;j<maskDim;j+= 1 + skipc)
		result += mask.mm[i][j]*pfield->mm[startD + i][startW + j];
    }
    else {
	printf("convolveMask assumes odd-sized mask \n");
    }
    return result;
}

float PotentialField::doWeightedBearing(int start, int end)
{
    const SbVec3f *vals = coords->point.getValues(0);
    int i, j, numNonZero = 0;
    
    /* now get weighted vector out of it */
    float cumweight = 0.0;
    float wb = 0.0, lm = 0.0;
    int widthDiv2 = (width/2);
    SbVec3f weightedVector(0, 0, 0);
    SbVec3f tv;
    int mi = -1, mj = -1;
    
    for(j=start;j<end;j++){
	lm = 0.0;
	for(i=(depth/2)-2;i>=0;i--){
	    if(fabs(gradMag->mm[i][j])>5.0e-3){
		if(fabs(gradMag->mm[i][j])<lm)
		    break;
		else {
		    lm = fabs(gradMag->mm[i][j]);
		    tv = vals[i*widthDiv2*2 + j*2 + 1] - vals[i*widthDiv2*2 + j*2];
		    weightedVector += tv;
		    numNonZero++;
		    if((mi<0) || gradMag->mm[i][j]>gradMag->mm[mi][mj]){
			mi = i;
			mj = j;
		    }
		}
	    }
	}
    }

    if(numNonZero)
	weightedVector /= ((float)numNonZero);
    wb = -atan2( -weightedVector[0], weightedVector[1]);
    maxGrad = (mi>-1)? gradMag->mm[mi][mj] :2.0;
//     printf("wb -> %f %f\n", wb, maxGrad);
    return wb;
}

void PotentialField::doGradient()
{
    Matrix maskWidth(3, 3);
    Matrix maskDepth(3, 3);
    
    maskWidth.mm[0][0] = -1;
    maskWidth.mm[1][0] = -2;
    maskWidth.mm[2][0] = -1;
    maskWidth.mm[0][2] =  1;
    maskWidth.mm[1][2] =  2;
    maskWidth.mm[2][2] =  1;
    
    maskDepth.mm[2][0] = -1;
    maskDepth.mm[2][1] = -2;
    maskDepth.mm[2][2] = -1;
    maskDepth.mm[0][0] =  1;
    maskDepth.mm[0][1] =  2;
    maskDepth.mm[0][2] =  1;
      
    int i, j, k = 1, ii = 0, jj = 0, kk=0;
    float w, d, r;
    SbVec3f *vals = coords->point.startEditing();
    SbVec3f *vals1 = coords1->point.startEditing();
    SbVec3f newPt;
    for(i=1;i<depth;i+=2, ii++){
	jj = 0;
	for(j=1;j<width;j+=2, jj++){
	    w = .25 * convolveMaskAt(maskWidth, i, j, 0, 1);
	    d = .25 * convolveMaskAt(maskDepth, i, j, 1, 0);
//	    if((fabs(w)>1.0e-4)|| (fabs(d)>1.0e-4)){
		grad->mm[ii][jj] = r = atan2( -w, -d);
		gradMag->mm[ii][jj] = sqrt(w*w + d*d)/2.0;
//	    }
//	    else {
//		grad->mm[ii][jj] = r = gradMag->mm[ii][jj] = 0.0;
//	    }
	    // set the z of the surface to be equal to the value of
	    // the potential field
	    newPt = vals1[kk];
	    newPt[2] =  (fabs(pfield->mm[i][j])>1.0e-1)?pfield->mm[i][j]:.1;
	    //coords1->point.set1Value(kk++, newPt);
	    vals1[kk++] = newPt;
	    
	    // set the vector so its base rests on the surface of the
	    // potential field, and it points in x and y.
	    //coords->point.set1Value(k-1, newPt);
	    vals[k-1] = newPt;
	    newPt = vals[k-1] + SbVec3f(sin(r)*gradMag->mm[ii][jj]
		,cos(r)*gradMag->mm[ii][jj], 0.0); 
	    //coords->point.set1Value(k, newPt);
	     vals[k] = newPt;
	    k += 2;
	}
    }
    coords->point.finishEditing();
    coords1->point.finishEditing();
    static float wb = 0;
    static float wmg = 0;
    float wwb;
    wwb = weightedBearing = doWeightedBearing(0, (int)(width/2));
//    printf("wwb -> %f \n", wwb);
    float Y = (wmg/(maxGrad + wmg))*cos(wb) + (maxGrad/(maxGrad + wmg))*cos(wwb);
    float X = (wmg/(maxGrad + wmg))*sin(wb) + (maxGrad/(maxGrad + wmg))*sin(wwb);
    float ATYX = atan2(X, Y);
//    printf("avg wwb -> %f \n", ATYX);
    weightedBearing = ATYX;
    
    if(weightedBearing> M_PI)
	weightedBearing -= (2.0*M_PI);
    wmg = maxGrad;
    wb = wwb;
}
