#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>
#include <Inventor/Sb.h>
#include <Inventor/SoInput.h>
#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/nodes/SoCamera.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoRotation.h>
#include <Inventor/nodes/SoBaseColor.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoPickStyle.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/SoPath.h>
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
#include <Inventor/SoOffscreenRenderer.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/actions/SoRayPickAction.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoFaceSet.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoCube.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoLightModel.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoMFInt32.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/elements/SoCacheElement.h>
#include "PotentialFieldSensor.h"
#include "PotentialField.h"
#include "CreatureCache.h"
#include "BaseCreature.h"
#include "MotorCreature.h"
#include "CreatureBehavior.h"
#include "RepSwitch.h"
#include <GL/gl.h>
#include "TimeStat.h"
float testbearing;
/*********************************************************************
 * 
 *	    PotentialFieldSensor  Class
 * 
 *********************************************************************/
PotentialField *pf1 = NULL;

SO_NODE_SOURCE(PotentialFieldSensor);

void
PotentialFieldSensor::initClass()
{
    SO_NODE_INIT_CLASS(PotentialFieldSensor, BaseSensor, "BaseSensor");
}

PotentialFieldSensor::PotentialFieldSensor() : BaseSensor()
{
    SO_NODE_CONSTRUCTOR(PotentialFieldSensor);
    SO_NODE_ADD_FIELD(cameraSlant, (0.01));
    SO_NODE_ADD_FIELD(potentialSpread, (12));
    SO_NODE_ADD_FIELD(sensorDir, (1, 0, 0));
    SO_NODE_ADD_FIELD(fieldWidth, (17));
    SO_NODE_ADD_FIELD(size,  (64, 64));
    SO_NODE_ADD_FIELD(ifShowViewer, (TRUE));
    SO_NODE_ADD_FIELD(camera, (NULL));
    SO_NODE_ADD_FIELD(lumaThreshold, (.05));
    SO_NODE_ADD_FIELD(useGndMat, (TRUE));
    SO_NODE_ADD_FIELD(toggleListTags, (0));
    SO_NODE_ADD_FIELD(sampleInterval, (1));
    pf = NULL;
    pfViewer = NULL;
    ra = NULL;
    retina = new GLubyte[128*128];
    obstaclesInRange = 1;
    ticksSinceLastSample = 0;
}

PotentialFieldSensor::~PotentialFieldSensor()
{
}



SbBool	
#ifdef SO_VERSION_REVISION
PotentialFieldSensor::readInstance(SoInput *in, unsigned short flags) 
#else
PotentialFieldSensor::readInstance(SoInput *in) 
#endif
{
#ifdef SO_VERSION_REVISION
    if (BaseSensor::readInstance(in, flags)) {
#else
    if (BaseSensor::readInstance(in)) {
#endif
	if(camera.getValue() != NULL)
	    camera.getValue()->ref();
	else {
	    SoPerspectiveCamera *cam = new SoPerspectiveCamera();
	    cam->viewportMapping = SoCamera::CROP_VIEWPORT_FILL_FRAME;
	    cam->aspectRatio.setValue(SO_ASPECT_SQUARE);
	    cam->nearDistance = .1;
	    cam->farDistance = 100;
	    cam->heightAngle.setValue(M_PI/2);
	    cam->ref();
	    camera.setValue(cam);
	}
	pf = new PotentialField((int)fieldWidth.getValue(), (int)fieldWidth.getValue(), (float)1.0);
	SbVec3f dir = sensorDir.getValue();
	SbRotation(SbVec3f(0, 0, 1), cameraSlant.getValue()).multVec(dir, dir);
	sensorDir.setValue(dir);
	return (1);
    }
    else {
	return(0);
   }
}

void PotentialFieldSensor::showGradientViewer()
{
    if(!pfViewer){
	SbVec2f vpsize(256, 256);
	SbViewportRegion vp;
	vp.setWindowSize(SbVec2s((short)vpsize[0], (short)vpsize[1]));
	pfViewer = new SoXtExaminerViewer();
	SoSeparator *sep = new SoSeparator;
	
	sep->addChild(pf->visualizer);
	
	pfViewer->setSceneGraph(sep);
	//pfViewer->setDecoration(FALSE);
	pfViewer->viewAll();
	pfViewer->setSize(SbVec2s((short)vpsize[0], (short)vpsize[1]));
	pfViewer->show();
    }
   
}
#define RETINA_COMPONENT_SIZE 3
#define ABGR_OFFSET ((RETINA_COMPONENT_SIZE) - 3)
#define RETINA_OFFSET(r,c)(dim[0]*(RETINA_COMPONENT_SIZE)*r + c*(RETINA_COMPONENT_SIZE))
// the pixel format is agbr and unsigned byte
#define RETINA_LUM(r,c) (((int)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET)]+ (int)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET +1)]+(int)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET+2)]))
#define RETINA_LUM1(r,c) (retina[RETINA_OFFSET(r,c)])
#define RETINA_COLOR1(r,c)(SbColor(((float)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET)]/255.0), \
    ((float)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET +1)]/255.0), \
    ((float)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET +2)]/255.0)))

#define RETINA_COLOR_COMP(r, c, rgb)((int)retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET)+rgb])
#define RETINA_CC(r, c, rgb) retina[RETINA_OFFSET(r,c)+(ABGR_OFFSET)+rgb]
SbVec2s PotentialFieldSensor::grabComponentFromImage(SoXtRenderArea *vwr, GLubyte *buf)
{
    dim = SbVec2s(0, 0);
    Display *d = SoXt::getDisplay();
    Window w = vwr->getNormalWindow();
    GLXContext c = vwr->getNormalContext();
    
    if (!glXMakeCurrent(d, w, c)) {
	printf("Unable to read z buffers\n");
	return dim;
    }
    
    // Read in the pixels into the retina array, whose size is set to
    // the dimensions ("dim") of the ViewportRegion.
    
    dim = vwr->getViewportRegion().getViewportSizePixels();
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadPixels(0, 0, dim[0], dim[1], 
	((RETINA_COMPONENT_SIZE ==4)?(GL_RGBA /* GL_ABGR_EXT */) : (GL_RGB) ),GL_UNSIGNED_BYTE, buf);
/*
    int i, j;
    for(i=0, j=0;i<(dim[0]*dim[1]-2)*3 ;i+= 3, j++){
	buf[j] = buf[i]+buf[i+1]+buf[i+2];
    }
*/
//    glReadPixels(0, 0, dim[0], dim[1], whichComp,GL_FLOAT, buf);
    return dim;  
}

int PotentialFieldSensor::findPixelWithColor(const SbColor &searchColor, int &r, int &c, int xstart, 
    int xend, SbVec2s &dim, int ignoreTopY)
{
    int j, k;
    int srgb[3];
    int diff;
    int cornerFnd = 0;
    int mink;
    int extentx =xend-1;
    int extenty = dim[1]-ignoreTopY;
    int matches = 0;
    int jj = 0, kk = 0;
#ifdef TIMEMARKER
    static Timestat *timer = NULL;
    if(!timer)
	timer = new Timestat("marker timing", 20); 
    timer->startTime();
#endif
    srgb[0] =  searchColor[0]*255;
    srgb[1] =  searchColor[1]*255;
    srgb[2] =  searchColor[2]*255;
    markerMinX = -100;
    markerMaxX = 100;
    for(j=xstart;j<extentx; j++){
	for(k=0;k<extenty; k++){
	    diff = abs(RETINA_COLOR_COMP(k, j, 0)-srgb[0]) +
		   abs(RETINA_COLOR_COMP(k, j, 1)-srgb[1]) +
		   abs(RETINA_COLOR_COMP(k, j, 2)-srgb[2]) ;
	    if(diff <10){
		if(!cornerFnd){
		    extentx = ((j+8)< extentx)? (j+8) : extentx;
		    extenty = ((k+4)< extenty)? (k+8) : extenty;
		    cornerFnd = 1;
		    mink = k;
		    markerMinX = j; 
		}
		markerMaxX = j;
		matches++;
		jj += j;
		kk += k;
	    }
	}
    }
    if(cornerFnd){
	jj /= matches;
	kk /= matches;
//	printf("color match centroid found at %d %d area %d \n", jj, kk, matches);
	r = mink;
	c = jj;
    }
#ifdef TIMEMARKER
    timer->endTime();
#endif
    return (cornerFnd)?1:0;
}

void PotentialFieldSensor::updateField()
{
    int j, k;
    SbVec3f lpt, wpt;
    
    // Grab the luminance data from the image and stuff it into
    // the retina. dim is the size of the image
    dim = grabComponentFromImage(ra, retina);
    
    // Create a floor plane on which to project points
    SbPlaneProjector pp(SbPlane(SbVec3f(0, 1, 0), 0), 0);
    pp.setViewVolume(((SoPerspectiveCamera *)camera.getValue())->getViewVolume());
    
    SbVec2f pt2d;
    // vpsize is the requested size of the image. Since the viewport may
    // be larger due to the size of the window, you want to extract relevant
    // portion of viewport. This is what is done below
    SbVec2f vpsize = size.getValue();
    int st = (int)(dim[0]/2 - vpsize[1]/2);
    int end = (int)(dim[0]/2 + vpsize[1]/2);
    
    scanStart = st;
    scanEnd = end;
    

/* 
//    for debugging purposes  
 
    printf("new update \n");
    for(k=0;k<dim[1];k++){
	printf("\n");
	for(j=st;j<end;j++){
	    //printf("%d", (retina[dim[0]*(dim[1]-k-1)+j]>.1)?1:0);
	    printf("%d", (RETINA_LUM((dim[1]-k-1), j)>.1)?1:0);
	}
    }
*/				   
    // clear the potential field
    pf->clear();
    
    // iterate through the retina. The retina is laid out with the first
    // row corresponding to the bottom row of the image.
    float rvalue;
    SbVec3f center((end-st)/2, 0,(end-st)/2); 
    float td;
    int spreadAny = 0;
    float lasttd = 0;
    int karray[32];
    int kindex = 0;
    obstaclesInRange = 0;
    for(j=st;j<end-1;j+=2){
	for(k=0;k<dim[1]-1;k+=2){
	    rvalue = .25*(1.0/(3*255.0))*(RETINA_LUM((k), (j)) + RETINA_LUM((k+1), (j))
		    + RETINA_LUM((k), (j+1)) + RETINA_LUM((k+1), (j+1)));
	    if(rvalue>lumaThreshold.getValue()){
		
		// if above threshold, then project point onto floor
		// and get world coordinates of point
		pt2d.setValue(((float)j-st)/(float)(end-st),((float)(k)/dim[1]));
		wpt = pp.project(pt2d);
		
		// map to local frame
		lXw.multVecMatrix(wpt, lpt);
		td = lpt.length();
			
		if(td<range){
		    closestPt = wpt;
		    range = td;
		}
		// now send it to potential field for spreading
		pf->spreadPotential(lpt, potentialSpread.getValue());
		spreadAny = (k<8)?1:0;
		if(spreadAny)
		    obstaclesInRange = 1;
		// go on to next column
		break;
	    }
	}
	karray[kindex++] = k;
    }
    markerFnd = findPixelWithColor(SbColor(0, 1, 0), k, j, scanStart, scanEnd, dim, 10);
    if(markerFnd){
	pt2d.setValue(((float)j-st)/(float)(end-st),((float)(k)/dim[1]));
	markerPos = pp.project(pt2d);
	SbVec3f ptmp;
	float r;
	int center = (end + st)/2;
	closestEdge =  (fabs(markerMinX - center)< fabs(markerMaxX - center))?
			markerMinX : markerMaxX;
//	pt2d.setValue(((float)closestEdge-st)/(float)(end-st),((float)(k)/dim[1]));
//	ptmp = pp.project(pt2d);
//	calcBearingAndRange(ptmp, markerEdgeBearing, r);
	
//	printf("%d %f %d %d\n", closestEdge,  markerEdgeBearing,  markerMinX, markerMaxX); 
    }

/*
    printf("column heights : ");
    for(k = 0;k<kindex;k++){
	printf(" %d", karray[k]);
    }
    printf("\n");
*/
}

void PotentialFieldSensor::updateGradient()
{
   
    
    // calculate the gradient
    pf->doGradient();

//    pf->grad->print("gradient");
//    pf->gradMag->print("magnitude of gradient");
//    printf("weighted bearing -> %f \n", pf->weightedBearing);
}

SoSeparator *PotentialFieldSensor::buildGndMat()
{
    float corners[4][3] = {{-.5, 0, .5}, {.5, 0, .5}, {.5, 0, -.5}, {-.5, 0, -.5}};
    SoSeparator *sep = new SoSeparator;
    SoTransform *xf = new SoTransform;
    SoCoordinate3 *matcoords = new SoCoordinate3;
    SoFaceSet *fs = new SoFaceSet;
    SoBaseColor *bc = new SoBaseColor;
    
    bc->rgb.setValue(0, 0, 0);
    xf->scaleFactor.setValue(5.0*fieldWidth.getValue(), 1, 5.0*fieldWidth.getValue());
    matcoords->point.setValues(0, 4, corners);
    fs->numVertices = 4;
    fs->startIndex = 0;
   
    sep->addChild(xf);
    sep->addChild(bc);
    sep->addChild(matcoords);
    sep->addChild(fs);
    return sep;
}

void PotentialFieldSensor::toggleVisForSensing(UpdateStateData &usd)
{
    const int32_t *vals = toggleListTags.getValues(0);
    int len = toggleListTags.getNum();
    int i, j;
    CreatureCache *cc = usd.currCreatureCache;
    BaseCreature *bc;
    if((len > 1)|| ((len>0) && vals[0] != 0)){
	for(i=0;i<cc->creatureCount;i++){
	    bc = (BaseCreature *)cc->creatures[i];
	    for(j=0;j<len;j++)
		if((bc->tag.getValue() == vals[j]) && bc->getSwitch()){
		    bc->getSwitch()->whichChild =
			(bc->getSwitch()->whichChild.getValue() == SO_SWITCH_NONE)
			? SO_SWITCH_ALL
			: SO_SWITCH_NONE;
		}
	}
    }
}


void PotentialFieldSensor::changeRepsForSensing(UpdateStateData &usd, int whichRep)
{
    CreatureCache *cc = usd.currCreatureCache;
    BaseCreature *bc;
    CreatureBehaviorUpdateAction *cba = (CreatureBehaviorUpdateAction *)
	((MotorCreature *)base)->getBehaviorUpdateAction();
    int ooi = -1;
    if(cba)
	ooi  = cba->objectOfInterestTag;
    int i;
    for(i=0;i<cc->creatureCount;i++){
	bc = (BaseCreature *)cc->creatures[i]; 
	if(bc != (BaseCreature *)base){
	    if((bc->tag.getValue() == ooi)&&
		(whichRep == RepSwitch::REP_SENSING))
		    bc->changeRep(RepSwitch::REP_SENSING_OO);
	    else
		bc->changeRep(whichRep);
	}
    }       
}
void PotentialFieldSensor::_updateStateAt(UpdateStateData &usd)
{
    if(!ra){
	static int numCreated = 0;
	SbVec2f vpsize = size.getValue();
	SbViewportRegion vp((short)vpsize[0],(short)vpsize[1]) ;
	//ra = new SoXtExaminerViewer(NULL, NULL, TRUE, SoXtFullViewer::BUILD_NONE, SoXtViewer::BROWSER);
	ra = new SoXtRenderArea();
	ra->setDoubleBuffer(FALSE);
	//((SoXtExaminerViewer *)ra)->setDrawStyle(SoXtViewer::STILL,
	//    SoXtViewer::VIEW_BBOX);
	SoSeparator *sep = new SoSeparator;

	SoDirectionalLight *dl = new SoDirectionalLight;
	dl->direction.setValue(0, 0, -1);
//	sep->addChild(dl);
/*
	SoEnvironment *env = new SoEnvironment;
	env->ambientIntensity.setValue(1000.0);
	sep->addChild(env);
*/	
	SoLightModel *lm = new SoLightModel();
	lm->model = SoLightModel::BASE_COLOR;
	sep->addChild(lm);
	sep->addChild(camera.getValue());
	
	gndMat = buildGndMat();
	sep->addChild(gndMat);
	
	sep->addChild((SoSeparator *)usd.currSceneGraph);
	ra->setAutoRedraw(FALSE);
	ra->setSceneGraph(sep);
	ra->setSize(SbVec2s((short)vpsize[0], (short)vpsize[1]));
	ra->setBorder(FALSE);
	ra->setTitle(" ");
	ra->show();
	XtMoveWidget(ra->getShellWidget(), 800 , 20 + numCreated*128);
	numCreated++;
    }
    if((ticksSinceLastSample >= sampleInterval.getValue())
	|| obstaclesInRange){
	    obstaclesInRange = 0;
	    ticksSinceLastSample = 0;
	if(camera.getValue() != NULL){
	    SoPerspectiveCamera *cam = (SoPerspectiveCamera *)camera.getValue();
	    SbVec3f campt(-(fieldWidth.getValue()), 0, 0); 
	    SbLine camdata(campt, campt + sensorDir.getValue());
	    wXl.multLineMatrix(camdata, camdata);
	    cam->position.setValue(camdata.getPosition());
	    cam->pointAt(camdata.getPosition()+ camdata.getDirection());
	    
	    if(ifShowViewer.getValue() == TRUE && (!pfViewer))
		showGradientViewer();
	    
	    // Switch off sensing of this and any camera creatures that may
	    // be racketing around.
	    if(swNode)
		swNode->whichChild = SO_SWITCH_NONE;
	    if (usd.currCreatureCache->getCameraCreature()){
		usd.currCreatureCache->getCameraCreature()->
		    getSwitch()->whichChild.setValue( SO_SWITCH_NONE);
	    }
	    
	    //toggleVisForSensing(usd);
	    changeRepsForSensing(usd, RepSwitch::REP_SENSING);
	    
	    // gag gag gag, set gndMat
	    SbVec3f gndMatPos(0, 0, 0);
	    wXl.multVecMatrix(gndMatPos, gndMatPos);
	    gndMatPos[1] = 1;
	    ((SoTransform *)gndMat->getChild(0))->translation.setValue(gndMatPos);
	    
	    ra->render();
	    
	    // restore switches that may have been dinked with
	    if(swNode)
		swNode->whichChild = SO_SWITCH_ALL;
	    if (usd.currCreatureCache->getCameraCreature()){
		usd.currCreatureCache->getCameraCreature()->
		    getSwitch()->whichChild.setValue( SO_SWITCH_ALL);
	    }
	    
	    //toggleVisForSensing(usd);
	    changeRepsForSensing(usd, RepSwitch::REP_REVERT);
	    
	    closestPt = SbVec3f(1.0E6,1.0E6, 1.0E6);
	    range = 1.0e6; 
	    updateField();
	    
	    calcBearingAndRange(closestPt, bearing, range);	
	}
    }
//    else
//	printf("skipping sensing \n");
    ticksSinceLastSample++;
    return;
}

