#ifndef _PFSENSOR_
#define _PFSENSOR_
#include <GL/gl.h>
#include <Inventor/SbBasic.h>
#include <Inventor/SbLinear.h>
#include <Inventor/fields/SoSFVec3f.h>
#include <Inventor/fields/SoSFFloat.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/fields/SoMFVec3f.h>
#include <Inventor/fields/SoSFVec2f.h>
#include <Inventor/fields/SoSFInt32.h>
#include <Inventor/fields/SoMFInt32.h>
#include <Inventor/fields/SoMFNode.h>
#include <Inventor/fields/SoSFNode.h>
#include <Inventor/fields/SoSFBool.h>
#include <Inventor/nodes/SoTransform.h>
#include "BaseSensor.h"
#include <Inventor/SbColor.h>
class SoPath;
class SoSwitch;
class SoCamera;
class PotentialField;
class SoXtExaminerViewer;

class switchStateData {
    BaseCreature *creature;
    int switchValue;
};

class PotentialFieldSensor: public BaseSensor {
    SO_NODE_HEADER(PotentialFieldSensor);

public:
    SoSFVec3f sensorDir;
    SoSFFloat cameraSlant;
    SoSFFloat potentialSpread;
    SoSFVec2f size;
    SoSFNode camera;
    SoSFFloat lumaThreshold;
    SoSFFloat fieldWidth;
    SoSFBool ifShowViewer;
    SoSFBool useGndMat;
    SoMFInt32 toggleListTags;
    SoSFInt32 sampleInterval;
    SoXtRenderArea *ra;
    PotentialField *pf;
    SoXtExaminerViewer *pfViewer;
    int markerFnd;
    SbVec3f markerPos;
    int markerMinX;
    int markerMaxX;
    int closestEdge;
    float markerEdgeBearing;
    
    SbVec2s grabComponentFromImage(SoXtRenderArea *vwr, GLubyte *buf);
    int findPixelWithColor(const SbColor &searchColor, int &r, int &c, 
	int xstart, int xend, SbVec2s &dim, int ignoreTopY);

    void updateField();
    void updateGradient();
    
    void showGradientViewer();
    
    static void initClass();
    virtual void _updateStateAt(UpdateStateData &usd); // over-ride this method

protected:
    GLubyte *retina;
    SoSeparator *PotentialFieldSensor::buildGndMat();
    SoSeparator *gndMat;
    SbPList toggleList;
    void toggleVisForSensing(UpdateStateData &usd);
    void changeRepsForSensing(UpdateStateData &usd, int whichRep);

    int ticksSinceLastSample;
    int obstaclesInRange;
    SbVec2s dim;
    int scanStart;
    int scanEnd;
    
    PotentialFieldSensor();       // Note only should be call during a read
#ifdef SO_VERSION_REVISION
    virtual SbBool	readInstance(SoInput *in, unsigned short flags);
#else
    virtual SbBool	readInstance(SoInput *in);
#endif
    virtual ~PotentialFieldSensor();

};

#endif

