// GenericIterativeServer.c++
// kbrussel@media.mit.edu

#include <iostream.h>
#include <signal.h>
#include <stdlib.h>
#ifdef __GNUG__
#include "Dlist.c++"
#endif
#include "SocketMan.h"

const int port = 14159;

SocketMan *sm;
SocketBuf *sbptr;

int main(int argc, char **argv)
{
  int use_udp;
  int use_sys_v_signals;
  
  if (argc != 2)
    {
      cerr << "usage: smtest <use UDP: [0,1]>" << endl;
      exit(1);
    }
  else
    {
      use_udp = atoi(argv[1]);
    }

  // We'll assume System V signals here.
  // On the SGI and Alpha using -lbsd, this could just
  // as well be 0.

  use_sus_v_signals = 1;

  // Note that the read timeout values are both 0.
  // This means the server won't wait AT ALL for data to
  // come in via any socket.
  // In the case of large reads, there will inevitably be
  // pauses in data transfer. Thus, recoverFromReadError
  // is used below.

  sm = new SocketMan(port, 0, 0, 0, 500000, 0, 0, 0, use_sys_v_signals, use_udp);

  int counter = 0;
  
  while (1)
    {
      // Simulates the server's main loop.
      // The code below performs the socket work

      // Accept a new connection

      sm->acceptConnectionWithPoll();
      
      // Iterate over all connected sockets

      sm->rewindSockets();
      
      while (sm->nextSocket())
	{
	  if ((sm->getLastMessage() == SocketMan::ERR_TIMEOUT_READING_PACKET_LENGTH) ||
	      (sm->getLastMessage() == SocketMan::ERR_TIMEOUT_READING_PACKET_DATA_TYPE) ||
	      (sm->getLastMessage() == SocketMan::ERR_TIMEOUT_READING_PACKET))
	    {
	      sbptr = sm->recoverFromReadError();
	    }
	  else
	    {
	      sbptr = sm->readWithPollAndTimeout();
	    }
	  
	  // If we got data from a socket, it means there was a request
	  // for data from the client. The code here was left in as a
	  // simple example of how to interpret the SocketBuf.

	  if (sbptr)
	    {
	      if (sbptr->num_bytes < 1000)
		{
		  cerr << "READ SUCCEEDED" << endl;
		  
		  switch (sbptr->type)
		    {
		    case SOCKET_CHAR:
		      {
			cerr << "The secret message of the day is: " << (char *)sbptr->buf << endl;
			break;
		      }
		      
		    case SOCKET_INT:
		      {
			int counter;
			
			cerr << "Integer vector received. Numbers: ";
			
			for(counter = 0; counter < sbptr->num_bytes; counter += sizeof(int))
			  {
			    cerr << *((int *) (sbptr->buf + counter )) << " ";
			  }
			
			cerr << endl;
			break;
		      }
		      
		    case SOCKET_FLOAT:
		      {
			int counter;
			
			cerr << "Float vector received. Numbers: ";
			
			for(counter = 0; counter < sbptr->num_bytes; counter += sizeof(float))
			  {
			    cerr << *((float *) (sbptr->buf + counter )) << " ";
			  }
			
			cerr << endl;
			break;
		      }
		    }
		}
	      else
		{
		  // Note that this will not happen under UDP.

		  if (sbptr->num_bytes_successful == 1048576)
		    cerr << "BIG READ COMPLETED" << endl;
		  else
		    cerr << "**BIG READ FAILED**: num_bytes_successful == " << sbptr->num_bytes_successful << endl;
		}

	      // NOTE! The next line is absolutely necessary! Since 
	      // the read functions allocate memory internally (in the
	      // form of a SocketBuf) you must delete this object
	      // when you are done with it!

	      delete sbptr;
	    }
	}
    }
}


