// GenericSharedMemoryServerStub.c++
// kbrussel@media.mit.edu

#define _BSD_SIGNALS
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#ifdef __GNUG__
#include "Dlist.c++"
#endif
#include "SocketMan.h"
#include "SemaphoreClient.h"

// Take a fast look at this one line file

#include "GenericSharedMemoryServer.h"

SocketMan *sm;
SemaphoreClient *sc;

// This will ordinarily be in a header file #included
// by both the SocketMan and SocketClient programs.
const int SOCKET_PORT = 10666;   // random number

// This is the data type that you will exchange with the
// other thread via shared memory.
typedef void your_data_type;

void
cleanup(...)
{
  delete sm;
  delete sc;

  cerr << "shared memory stub cleaned up" << endl;
  exit(1);
}

int
main(int argc, char **argv)
{
  SocketBuf *sbuf;
  static SocketBuf sendbuf;
  your_data_type *ydtemp;
  int select_result;

  // Create a new sub-process. See "man fork"
  if (!fork())   // child gets a zero return value from fork()
    {
      // Execute the process that will generate
      // the data we will serve to clients over sockets.

      execv("your_exec", argv);  // pass in command line options to your_exec
      exit(1);                   // NOTE: This should never be reached
    }

  // This is important when using shared memory segments.
  signal(SIGINT, cleanup);
  
  // Start the SemaphoreClient
  sc = new SemaphoreClient(SHARED_MEMORY_ID, sizeof(your_data_type));
  
  // Start the Socket Manager. Using UDP sockets for speed.
  // If you were sending megabytes of data, you'd make the
  // last '1' a '0'.
  sm = new SocketMan(SOCKET_PORT, 1, 0, 1, 0, 0, 0, 0, 0, 1);
  
  // Go into main loop
  while(!sc->getDieFlag())
    {
      // This halts execution, waiting for one of two conditions:
      // 1. A new client attempts connection
      // 2. A connected client makes a data request
      select_result = sm->smartSelect();
      
      switch (select_result)
	{
	case SocketMan::ERR_DURING_SELECT:
	  {
	    cerr << "generic_stub: error during select" << endl;
	    break;
	  }
	  
	case SocketMan::MSG_SS_ACCEPT_CONNECTION:
	  {
	    // This is what allows handling of multiple clients
	    sm->acceptConnectionWithPoll();
	    break;
	  }

	case SocketMan::MSG_SS_READ_FROM_CURRENT_SOCKET:
	  {
	    // smartSemect took care of setting the
	    // current socket for us. All we have
	    // to do is read.

	    sbuf = sm->readWithPollAndTimeout();
	    if (sbuf)  // request for data on the other end
	      {
		// get a pointer to the last complete buffer
		
		ydtemp = (your_data_type *) sc->getCurrentBuffer();
		
		// set up the data structure to send
		
		sendbuf.num_bytes = sizeof(ContourPacketStruct);
		sendbuf.type = CHAR;
		sendbuf.buf = cp->bytes;
		sm->writeWithTimeout(&sendbuf);
		
		// clean up after ourselves
		
		sc->unlockBuffer();
		delete sbuf;
	      }
	    break;
	  }

	case SocketMan::ERR_SS_UNEXPECTED_FAILURE:
	  {
	    cerr << "backsub: unexpected failure during select" << endl;
	    break;
	  }
	}
    }
  
  cerr << "Semaphore client caught dieFlag" << endl;
  fflush(stderr);
  cleanup();
}

