#ifdef __GNUG__
extern "C" {
#endif
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef __GNUG__
}
#endif
#include <iostream.h>
#ifdef __sgi
#include <bstring.h>
#endif
#ifdef __GNUG__
#include "Dlist.c++"
#include "SocketMan.h"
#endif
#include "SocketBuf.h"
#include "Tunnel.h"

Tunnel::Tunnel(char *multicast_group_addr,
	       char *remote_host_name,
	       int port)
{
  msc = new SocketClient(port);
  msc->setUsingMulticast(1);
  if (msc->connectToServer(multicast_group_addr))
    cerr << "connect to multicast server failed" << endl;

  usc = new SocketClient(port+1);
  if (usc->connectAsTunnel(remote_host_name))
    cerr << "connect to remote host failed" << endl;
}

int
Tunnel::selectOnSocketClients(SocketClient *cli, ...)
{
  fd_set readfds;
  int width;
  int result;
  
  // find the maximum number of open file descriptors
#ifndef __hpux
  width = getdtablesize();
#else
  width = FD_SETSIZE;
#endif

  FD_ZERO(&readfds);

  FD_SET(cli->current_socket->fd, &readfds);

  va_list args;
  va_start(args, cli);

  SocketClient *cli2;
  while (cli2 = va_arg(args, SocketClient*))
    FD_SET(cli2->current_socket->fd, &readfds);

  va_end(args);
    
  // go into a blocking select on all these sockets

  cerr << "Entering select" << endl;
#ifndef __hpux
  result = select(width, &readfds, NULL, NULL, 0);
#else
  result = select(width, (int *) &readfds, NULL, NULL, 0);
#endif  
  cerr << "out of select" << endl;
  if (result < 0)
    {
      return -1;
    }

  return 0;
}

void
Tunnel::mainLoop()
{
  while(1)
    {
      SocketBuf *sb;

      if (selectOnSocketClients(msc, usc, 0) < 0)
	{
	  cerr << "Tunnel::mainLoop: select failed, exiting" << endl;
	  cleanup();
	}
      sb = msc->readWithPollAndTimeout();
      if (msc->getGlobalMessage() == SocketClient::ERR_CURRENT_SOCKET_NOT_OPEN)
	{
	  cerr << "Tunnel::mainLoop: multicast socket closed, exiting" << endl;
	  cleanup();
	}
      if (sb)
	{
	  cerr << "Tunneling packet" << endl;
	  usc->writeWithTimeout(sb);
	  delete sb;
	}
      else
	{
	  sb = usc->readWithPollAndTimeout();
	  if (usc->getGlobalMessage() == SocketClient::ERR_CURRENT_SOCKET_NOT_OPEN)
	    {
	      cerr << "Tunnel::mainLoop: udp socket closed, exiting" << endl;
	      cleanup();
	    }
	  if (sb)
	    {
	      cerr << "Rebroadcasting packet" << endl;
	      msc->writeWithTimeout(sb);
	      delete sb;
	    }
	}
    }
}

void
Tunnel::cleanup()
{
  delete msc;
  delete usc;
  exit(1);
}
