// smtest3.c++
// Reads data from a socket and then
// writes a megabyte back down it.
// Can be used to test SIGPIPE facility on server end.
// Uses and tests smartSelect.

#include <iostream.h>
#include <unistd.h>
#include <signal.h>
#ifdef __GNUG__
#include "Dlist.c++"
#endif
#include "SocketMan.h"
#include "sockettests.h"

SocketMan *sm;

void
cleanup(...)
{
  delete sm;

  cerr << "smtest3 cleaned up" << endl;
  exit(1);
}

int
main(int argc, char **argv)
{
  SocketBuf *sbuf;
  int select_result;
  SocketBuf *megbuf;

  signal(SIGINT, cleanup);
  
  // Start the Socket Manager
  sm = new SocketMan(port, 0, 0, 1, 0);
  
  // allocate the memory for the megabyte buffer
  megbuf = new SocketBuf;
  megbuf->num_bytes = 1048576;
  megbuf->type = SOCKET_CHAR;
  megbuf->buf = new sm_byte[megbuf->num_bytes];

  // Go into main loop
  while(1)
    {
      cerr << "Starting select" << endl;

      select_result = sm->smartSelect();
      
      cerr << "Out of select" << endl;

      switch (select_result)
	{
	case SocketMan::ERR_DURING_SELECT:
	  {
	    cerr << "smtest2: error during select" << endl;
	    break;
	  }
	  
	case SocketMan::MSG_SS_ACCEPT_CONNECTION:
	  {
	    sm->acceptConnectionWithPoll();
	    break;
	  }

	case SocketMan::MSG_SS_READ_FROM_CURRENT_SOCKET:
	  {
	    if ((sm->getLastMessage() == SocketMan::ERR_TIMEOUT_READING_PACKET_LENGTH) ||
		(sm->getLastMessage() == SocketMan::ERR_TIMEOUT_READING_PACKET_DATA_TYPE) ||
		(sm->getLastMessage() == SocketMan::ERR_TIMEOUT_READING_PACKET))
	      {
		sbuf = sm->recoverFromReadError();
	      }
	    else
	      {
		sbuf = sm->readWithPollAndTimeout();
	      }
	    
	    if (sbuf)
	      {
		if (sbuf->num_bytes < 1000)
		  {
		    cerr << "READ SUCCEEDED" << endl;
		    
		    switch (sbuf->type)
		      {
		      case SOCKET_CHAR:
			{
			  cerr << "The secret message of the day is: " << (char *)sbuf->buf << endl;
			  break;
			}
			
		      case SOCKET_INT:
			{
			  int counter;
			  
			  cerr << "Integer vector received. Numbers: ";
			  
			  for(counter = 0; counter < sbuf->num_bytes; counter += sizeof(int))
			    {
			      cerr << *((int *) (sbuf->buf + counter )) << " ";
			    }
			  
			  cerr << endl;
			  break;
			}
			
		      case SOCKET_FLOAT:
			{
			  int counter;
			  
			  cerr << "Float vector received. Numbers: ";
			  
			  for(counter = 0; counter < sbuf->num_bytes; counter += sizeof(float))
			    {
			      cerr << *((float *) (sbuf->buf + counter )) << " ";
			    }
			  
			  cerr << endl;
			  break;
			}
		      }
		  }
		else
		  {
		    if (sbuf->num_bytes_successful == 1048576)
		      cerr << "BIG READ COMPLETED" << endl;
		    else
		      cerr << "**BIG READ FAILED**: num_bytes_successful == " << sbuf->num_bytes_successful << endl;
		  }
		delete sbuf;

		// Now write a megabyte back down that socket

		sm->writeWithTimeout(megbuf);
	      }
	    break;
	  }

	case SocketMan::ERR_SS_UNEXPECTED_FAILURE:
	  {
	    cerr << "smtest2: unexpected failure during select" << endl;
	    break;
	  }
	}
    }
}

