#ifndef __USERMODEL__
#define __USERMODEL__

#define BUFFER_LENGTH 64 /* 3 secs at 20hz. */
#define  FRAMES_FOR_AVERAGE  20
#define FRAMES_BEFORE_AVERAGE 10

#define DEFAULT_HYSTERESIS_PERCENT 0.05
#define DEFAULT_POINTING_WIDTH 1.5
#define DEFAULT_SITTING_HEIGHT 0.5


#define NORMAL_WIDTH 16.0 /* inches */
#define WIDTH_FRACTION_OF_HEIGHT .3333

typedef enum { 
  CENTROID_X=0,
  CENTROID_Y,
  LEFT,
  RIGHT,
  TOP,
  BOTTOM,
  DLEFT,
  DRIGHT,
  DTOP,
  DBOTTOM,
  HAND_LEFT_X,
  HAND_LEFT_Y,
  HAND_RIGHT_X,
  HAND_RIGHT_Y,
  DHAND_LEFT_X,
  DHAND_LEFT_Y,
  DHAND_RIGHT_X,
  DHAND_RIGHT_Y,  
  FOOT_LEFT_X,
  FOOT_LEFT_Y,
  FOOT_RIGHT_X,
  FOOT_RIGHT_Y,
  DFOOT_LEFT_X,
  DFOOT_LEFT_Y,
  DFOOT_RIGHT_X,
  DFOOT_RIGHT_Y,  
  HAND_LEFT_VALID,
  HAND_RIGHT_VALID,
  TIME}
rbIndex;

#define NUMBER_OF_BUFFERS 29

typedef enum { LEFT_HAND, RIGHT_HAND, BOTH_HANDS } Hand;

typedef enum { LEFT_FOOT, RIGHT_FOOT } Foot;

typedef struct
{
  float hysteresis;
  float pointing_width; /* must be > */
  float sitting_height; /* must be < */
  float standing_height; /* must be > */
  float hand_over_head_height;
  float spread_eagle_width; /* normalized width, must be > */
  float pointing_still_time; /* secs */
  float max_dist_for_hand_not_moving; /* world */
  float throwing_thresh;
  float throwing_time;
  int oscillations_for_wave;
  float waving_thresh;
  float kicking_thresh;
  float min_hand_dist_for_wave;
  float min_hand_move_dist_for_wave;
  float min_hand_dist_for_extend;
  float min_foot_height_for_kick;
  float bending_over_height;
} GestureParams;

typedef struct
{
  unsigned char standing;
  unsigned   char sitting;
  unsigned char hand_over_head;
  unsigned char bending_over;

  unsigned   char point_left;
  unsigned   char point_right;

  unsigned   char kick_left;
  unsigned   char kick_right;

  unsigned   char wave_left;
  unsigned   char wave_right;

  unsigned   char throw_left;
  unsigned   char throw_right;

  unsigned char extend_left;
  unsigned char extend_right;
  
  unsigned char hand_moving_left;
  unsigned char hand_moving_right;
  
  unsigned   char user_exists;

} GestureFlags;


typedef struct
{
  boolean implemented;
} CameraToWorld;

typedef struct timeval Timer;

typedef struct 
{
  GestureParams *gp;
  CameraToWorld *camera; /* the camera model that changes 
			    points from camera to world coords. */
  
  RingBuffer **buffer; /* array of ring buffers */
  
  float normal_width; /* the user's basic shape, from first frames */
  float normal_height;
  
  Timer start_time; /* some thing used to tell what time it is */
  long ticks;
  boolean user_exists; /* if there is a blob */
  boolean acquired_model; /* got basic shape already, ie can process */
  
} UserModel;


#endif









