
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "curvature.h"

#include "curvature.proto"


/* return valid number of points in curvature output.
 Curvature must be at least MAX_POINTS * COORD_PER_POINT in size */
int
  calculateCurvature(float *contour, int len, float *curvature)
{
  int i;
  int numpoints = len/COORDS_PER_POINT;
  Vector v1, v2; /* v1 is older than v2 */
  
  v1 = makeVectorFromPoints(getPoint(contour, 0),
			    getPoint(contour, 1));
  
  for (i=0; i < numpoints-2; i++)
    {
      v2 = makeVectorFromPoints(getPoint(contour, i+1), 
				getPoint(contour, i+2)); 
      curvature[i] = angleBetweenVectors(v1, v2);
      v1 = v2;
    }
  return numpoints - 2;
}

void
printFloatArray(char *filename, float *array, int len)
{
  FILE *fp;
  int i;
  
  if (filename == NULL)
    fp = stdout;
  else
    {
      fp = fopen(filename, "w");
      if (fp == NULL)
	{
	  perror("printFloatArray");
	  return;
	}
    }
  
  for(i=0; i<len; i++)
    fprintf(fp, "%f ", array[i]);
  fprintf(fp, "\n");
  if (filename != NULL)
    fclose(fp);
}



Vector
makeVector(float x, float y)
{
  Vector v;
  
  v.x = x;
  v.y = y;
  return v;
}


Vector
makeVectorFromPoints(Vector p1, Vector p2)
{
  Vector v;

  v.x = p2.x - p1.x;
  v.y = p2.y - p1.y;
  return v;

}


Vector
getPoint(float *contour, int i)
{
  int index = i*COORDS_PER_POINT;
  Vector p;

  p.x = contour[index];
  p.y = contour[index+1];
  return p;
}


float
angleBetweenVectors(Vector v1, Vector v2)
{
  float dot; 

  dot = dotProduct(v1, v2);
  dot /= magnitudeVector(v1) * magnitudeVector(v2);
  if (dot >= 1.0) dot = 1.0;
  if (dot <= -1.0) dot = -1.0;

  return (float) acos(dot);
}


float 
dotProduct(Vector v1, Vector v2)
{
  
  return (v1.x * v2.x + v1.y * v2.y);

}


float 
magnitudeVector(Vector v)
{
  return (float) sqrt(v.x * v.x + v.y * v.y);
}

