#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include "common.h"
#include "/d/vision11/aries/socketlib/slib.h"
#include <math.h>
#include <stropts.h>

#include <gl.h>
#include <device.h>
#include <vl/vl.h>
#include <vl/dev_sirius.h>
#include <math.h>

#include "contour.h"
#include "packet.h"

#include "find_features.proto"
#include "track_features.proto"

#define BS_DEBUG 0
#define BS_DEBUG_THAD 0

#define LEFT_FOOT_EXTREMITY_ID 1
#define LEFT_HAND_EXTREMITY_ID 2
#define HEAD_EXTREMITY_ID 3
#define RIGHT_HAND_EXTREMITY_ID 4
#define RIGHT_FOOT_EXTREMITY_ID 5

int
InBoundingBox(float tlx, float tly,
	      float brx, float bry,
	      float xcor, float ycor)
{
  if ((xcor >= tlx) && (xcor <= brx) &&
      (ycor <= tly) && (ycor >= bry))
    return TRUE;
  else
    {
#if BS_DEBUG
      fprintf(stderr, "InBoundingBox: returned false\n");
#endif
      return FALSE;
    }
}

int
InTopOfBoundingBox(float tlx, float tly, 
		   float brx, float bry,
		   float yValue)
{
  if ((yValue <= tly) && 
      (yValue >= (bry + (1. / 3.) * (tly - bry))))
    return TRUE;
  else
    return FALSE;
}

int
  InBottomOfBoundingBox(float tlx, float tly, 
			float brx, float bry,
			float yValue)
{
  if ((yValue >= bry) && 
      (yValue <= (bry + (.4 * (tly - bry)))))
    return TRUE;
  else
    return FALSE;
}

int
  Quadrant(float centroidx,
	   float centroidy,
	   float testPointx,
	   float testPointy)
{
  if ((testPointx <= centroidx) && (testPointy < centroidy))
    return 1;
  else
    if ((testPointx < centroidx) && (testPointy >= centroidy))
      return 2;
    else
      if ((testPointx >= centroidx) && (testPointy > centroidy))
	return 3;
      else
	return 4;
}

float
  Distance(float ax,
	   float ay,
	   float bx,
	   float by)
{
  return (float)(sqrt ((bx - ax) * (bx - ax) + (by - ay) * (by - ay)));
}

int
  ThresholdScan(float *oldPos,
		float extremities[NUMBER_OF_EXTREMA * 3],
		int startCounter,
		int endCounter,
		float *newPos,
		int *newIndex,
		float Threshold)
{
  float tempPos[2];
  int flag = 0;
  float calcedDistance = 0.;

  tempPos[0] = newPos[0];
  tempPos[1] = newPos[1];

  while (startCounter < endCounter)
    {
      calcedDistance = Distance(oldPos[0], 
				oldPos[1], 
				extremities[startCounter], 
				extremities[startCounter + 1]);

#if BS_DEBUG
      fprintf(stderr, "TScan: calcedDistance = %f\n", calcedDistance);
#endif     

      if ((!flag) || 
	  (calcedDistance < Threshold && calcedDistance < Distance(oldPos[0],
								   oldPos[1],
								   newPos[0],
								   newPos[1])))
	{
	  newPos[0] = extremities[startCounter];
	  newPos[1] = extremities[startCounter + 1];
	  *newIndex = startCounter;
	  flag = 1;
	}
      startCounter += 3;
    }

/* KBR 7/1/94: inserted this to fix MAJOR bug that allowed a tracked point
   to leap major distances when going from tracking -> nontracking. */

  if (Distance(oldPos[0],
	       oldPos[1],
	       newPos[0],
	       newPos[1]) < Threshold)
    flag = 1;
  else
    {
      flag = 0;
      newPos[0] = tempPos[0];
      newPos[1] = tempPos[1];
    }
  return(flag);
}

float
  FloatAbs(float number)
{
  if (number < 0.)
    return (-1. * number);
  else
    return(number);
}

int
  ValidPick(int whichExtremity,
	    float tlx, float tly,
	    float brx, float bry,
	    float centroidx,
	    float centroidy,
	    float *latestPos)
{
  int flag = 0;

  switch(whichExtremity)
    {
    case LEFT_FOOT_EXTREMITY_ID:
      {
	if ((Quadrant(centroidx,
		      centroidy,
		      latestPos[0],
		      latestPos[1]) == 1) &&
	    (!(InTopOfBoundingBox(tlx, tly,
				  brx, bry,
				  latestPos[1]))))
	  flag = 1;
	else
	  flag = 0;
      }
      break;
    case LEFT_HAND_EXTREMITY_ID:
      {
	if ((InTopOfBoundingBox(tlx, tly,
				brx, bry,
				latestPos[1])) &&
	    ((centroidx - latestPos[0]) > 10.))
	  flag = 1;
	else
	  flag = 0;
      }
      break;
    case HEAD_EXTREMITY_ID:
      {
	if ((FloatAbs(centroidx - latestPos[0]) < 10.) &&
	    (centroidy < latestPos[1]))
	  flag = 1;
	else
	  flag = 0;
      }
      break;
    case RIGHT_HAND_EXTREMITY_ID:
      {
	if ((InTopOfBoundingBox(tlx, tly,
				brx, bry,
				latestPos[1])) &&
	    ((latestPos[0] - centroidx > 10.)))
	  flag = 1;
	else
	  flag = 0;
      }
      break;
    case RIGHT_FOOT_EXTREMITY_ID:
      {
	if ((Quadrant(centroidx,
		      centroidy,
		      latestPos[0],
		      latestPos[1]) == 4) &&
	    (!(InTopOfBoundingBox(tlx, tly,
				  brx, bry,
				  latestPos[1]))))
	  flag = 1;
	else
	  flag = 0;
      }
      break;
    }
  return(flag);
}

void
  ScanContourByExtremityID(int whichExtremity,
			   float centroidx,
			   float centroidy,
			   float *contour,
			   int numOfPoints,
			   float tlx, float tly,
			   float brx, float bry,
			   float *newPos)
{
  int counter = 0;
  int pointQuad;
  int flag = 0;

  newPos[0] = centroidx;
  newPos[1] = centroidy;
  
  switch (whichExtremity)
    {

    case LEFT_FOOT_EXTREMITY_ID:
      {
	while (Quadrant(centroidx,
			centroidy,
			contour[counter],
			contour[counter + 1]) == 1)
	  {
	    if ((contour[counter] <= newPos[0]) &&
		(contour[counter + 1] <= newPos[1]))
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	  counter += 3;
	  }
	
	if (flag == 0)
	  {
	    newPos[0] = tlx + (1. / 3.) * (brx - tlx);
	    newPos[1] = bry;
	  }
      }
      break;

    case LEFT_HAND_EXTREMITY_ID:
      {

	while (counter < ((numOfPoints - 1) * COORDS_PER_POINT))
	  {
	    if ((FloatAbs((bry + 0.5 * (tly - bry)) - contour[counter + 1]) < .5) &&
		(contour[counter] < newPos[0]))
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	    counter += 3;
	  }

#if BS_DEBUG
      fprintf(stderr, "*******In left hand of ScanContourByExtremityID %d %d\n",
	      newPos[0], newPos[1]);
#endif     

	if (flag == 0)
	  {
	    newPos[0] = tlx;
	    newPos[1] = bry + 0.6 * (tly - bry);
	  }
      }
      break;

    case HEAD_EXTREMITY_ID:
      {
	pointQuad = Quadrant(centroidx,
			     centroidy,
			     contour[counter],
			     contour[counter + 1]);
	
	while (counter < 3 * (numOfPoints - 1))
	  {
	    if (contour[counter] >= newPos[1]) /*&&
		(FloatAbs(contour[counter] - (.5 * (brx + tlx))) < 10.))*/
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
		flag = 1;
	      }
	    counter += 3;	    
	  }

	if (flag == 0)
	  {
	    newPos[0] = 0.5 * (tlx + brx);
	    newPos[1] = tly;
	  }
      }
      break;

    case RIGHT_HAND_EXTREMITY_ID:
      {
#if BS_DEBUG
      fprintf(stderr, "In right hand of ScanContourByExtremityID %d %d\n",
	      newPos[0], newPos[1]);
#endif     

      counter = 3 * (numOfPoints - 1);


      while (counter > 0)
	{
	  if ((FloatAbs((bry + 0.5 * (tly - bry)) - contour[counter + 1]) < .5) &&
	      (contour[counter] > newPos[0]))
	    {
	      newPos[0] = contour[counter];
	      newPos[1] = contour[counter + 1];
	      flag = 1;
	    }
	  counter -= 3;
	}

      if (flag == 0)
	{
	  newPos[0] = brx;
	  newPos[1] = bry + 0.6 * (tly - bry);
	}
      
#if BS_DEBUG
      fprintf(stderr, "In right hand of ScanContourByExtremityID %f %f counter: %d\n",
	      newPos[0], newPos[1], counter);
#endif     

      }
      break;
    case RIGHT_FOOT_EXTREMITY_ID:
      {
	counter = 3 * (numOfPoints - 1);
       
	while (Quadrant(centroidx,
			centroidy,
			contour[counter],
			contour[counter + 1]) == 4)
	  {
	    if ((contour[counter] >= newPos[0]) &&
		(contour[counter + 1] <= newPos[1]))
	      {
		newPos[0] = contour[counter];
		newPos[1] = contour[counter + 1];
	      }
	    counter -= 3;
	  }

	if (newPos[0] == centroidx && newPos[1] == centroidy)
	  {
	    newPos[0] = tlx + (COORD_TYPE) ((2. / 3.) * (float) (brx - tlx));
	    newPos[1] = bry;
	  }
      }
      break;
    }
}

int
  PickPointByExtremityID(int whichExtremity,
			 float *extremities,
			 int startCounter,
			 int endCounter,
			 float *newPos,
			 int *newIndex)
{
  int flag = 0;

  switch (whichExtremity)
    {
    case LEFT_FOOT_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((!flag) ||
	      (extremities[startCounter] < newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case LEFT_HAND_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((!flag) ||
	      (extremities[startCounter] < newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case HEAD_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((!flag) ||
	      (extremities[startCounter + 1] > newPos[1]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case RIGHT_HAND_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((!flag) ||
	      (extremities[startCounter] > newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    case RIGHT_FOOT_EXTREMITY_ID:
      while (startCounter < endCounter)
	{
	  if ((!flag) ||
	      (extremities[startCounter] > newPos[0]))
	    {
	      newPos[0] = extremities[startCounter];
	      newPos[1] = extremities[startCounter + 1];
	      *newIndex = startCounter;
	      flag = 1;
	    }
	  startCounter += 3;
	}
      break;
    default:
      fprintf(stderr, "Fatal: PickPointByExtremityID\n");
    }
  return(flag);
}

void
  LookForNewPoint(int whichExtremity,
		  float centroidx,
		  float centroidy,
		  float tlx, float tly,
		  float brx, float bry,
		  float *contour,
		  int contourLength,
		  unsigned long int frameNumber,
		  float *extremityCentroidList,
		  int startCounter,
		  int endCounter,
		  int *centIndex,
		  float *oldPos,
		  unsigned char oldValidBit,
		  float *lastValidPos,
		  unsigned long int lastValidFrame,
		  unsigned char *validBit,
		  float *newPos)
{
  float latestPos[2] = {0., 0.};
  int latestIndex = 0;
  static float Threshold = 10.;
  static int validCounter[5] = {0,0,0,0,0};

/* thad-changes first 2 %f's to %d */
#if BS_DEBUG
  fprintf(stderr, "LFNP: %d, %d (ID: %d)\n", newPos[0], newPos[1], whichExtremity);
#endif

#if BS_DEBUG
      fprintf(stderr, "Looking at extremity %d\n", whichExtremity);
#endif

  if (oldValidBit)  /* last position value was valid */
    {

#if BS_DEBUG
      fprintf(stderr, "oldValidBit set\n");
#endif

      if (ThresholdScan(oldPos,
			extremityCentroidList,
			startCounter,
			endCounter,
			latestPos,
			&latestIndex,
			Threshold))  /* new point within threshold of old */
	{
#if BS_DEBUG
      fprintf(stderr, "threshold scan passed\n");
#endif
	  *centIndex = (3 + latestIndex);  
	  newPos[0] = latestPos[0];
	  newPos[1] = latestPos[1];
	  *validBit = 1;  /* flag this value as valid too */
	}  /* end successful threshold scan */
      else  /* gotta be invalid */
	{
#if BS_DEBUG
	  fprintf(stderr, "threshold scan did not pass\n");
#endif
	  /*  thad-why would this ever become 0,0 */
	  
	  if (PickPointByExtremityID(whichExtremity,
				     extremityCentroidList,
				     startCounter,
				     endCounter,
				     latestPos,
				     &latestIndex))
	    {
	      if (!(ValidPick(whichExtremity,
			      tlx, tly,
			      brx, bry,
			      centroidx,
			      centroidy,
			      latestPos)))
		if (!(lastValidPos[0] == 0. && lastValidPos[1] == 0.))
		  {
#if BS_DEBUG
		    fprintf(stderr, "lastvalidpos's == 0\n");
#endif
		    newPos[0] = lastValidPos[0];
		    newPos[1] = lastValidPos[1];
		  }
		else
		  {
#if BS_DEBUG
		    fprintf(stderr, "lastvalidpos's == 0\n");
#endif
		    newPos[0] = oldPos[0];
		    newPos[1] = oldPos[1];
		  }
	    }
	  else
	    {
	      if (!(ValidPick(whichExtremity,
			      tlx, tly,
			      brx, bry,
			      centroidx,
			      centroidy,
			      latestPos)))
		if (!(lastValidPos[0] == 0. && lastValidPos[1] == 0.))
		  {
#if BS_DEBUG
		    fprintf(stderr, "lastvalidpos's == 0\n");
#endif
		    newPos[0] = lastValidPos[0];
		    newPos[1] = lastValidPos[1];
		  }
		else
		  {
#if BS_DEBUG
		    fprintf(stderr, "lastvalidpos's == 0\n");
#endif
		    newPos[0] = oldPos[0];
		    newPos[1] = oldPos[1];
		  }
	    }	      
	  *validBit = 0;
	}
    }  /* end old position valid */
  else	/* last point was invalid */

    if (lastValidFrame > 0 && (frameNumber - lastValidFrame) < 5)  /* last valid value is still okay to
								      use as a reference */
      {
#if BS_DEBUG
      fprintf(stderr, "lastvalid value ok to reference\n");
#endif	  
	if (ThresholdScan(lastValidPos,
			  extremityCentroidList,
			  startCounter,
			  endCounter,
			  latestPos,
			  &latestIndex,
			  Threshold))  /* new point within threshold of old */
	  {
#if BS_DEBUG
	    fprintf(stderr, "thresh scan ok\n");
#endif	  
	    *centIndex = (3 + latestIndex);
	    newPos[0] = latestPos[0];
	    newPos[1] = latestPos[1];
	    *validBit = 1;  /* flag this value as valid too */
	  }  /* end successful threshold scan */
	else  /* gotta be invalid */
	  {
#if BS_DEBUG
	    fprintf(stderr, "thresh scan not ok\n");
#endif	  
	    
	    if (PickPointByExtremityID(whichExtremity,
					 extremityCentroidList,
					 startCounter,
					 endCounter,
					 latestPos,
					 &latestIndex))
	      {
		if (!(ValidPick(whichExtremity,
			      tlx, tly,
			      brx, bry,
			      centroidx,
			      centroidy,
			      latestPos)))
		  if (!(lastValidPos[0] == 0. && lastValidPos[1] == 0.))
		    {
#if BS_DEBUG
		      fprintf(stderr, "lastvalidpos's == 0\n");
#endif	  
		      newPos[0] = lastValidPos[0];
		      newPos[1] = lastValidPos[1];
		    }
		  else
		    {
		      
#if BS_DEBUG
		      fprintf(stderr, "lastvalidpos's != 0\n");
#endif	  
		      newPos[0] = oldPos[0];
		      newPos[1] = oldPos[1];
		    }
	      }
	    else
	      if (!(lastValidPos[0] == 0. && lastValidPos[1] == 0.))
		{
#if BS_DEBUG
		  fprintf(stderr, "lastvalidpos's == 0\n");
#endif	  
		  newPos[0] = lastValidPos[0];
		  newPos[1] = lastValidPos[1];
		}
	      else
		{
		  
#if BS_DEBUG
		  fprintf(stderr, "lastvalidpos's != 0\n");
#endif	  
		  newPos[0] = oldPos[0];
		  newPos[1] = oldPos[1];
		}
	    *validBit = 0;
	  }	  
    }	 /* end last valid value ok */	  
    else
      if (PickPointByExtremityID(whichExtremity,
				 extremityCentroidList,
				 startCounter,
				 endCounter,
				 latestPos,
				 &latestIndex))
	{
#if BS_DEBUG
	  fprintf(stderr, "pickpointbyextermityid passes\n");
#endif	  
	  
	  if (ValidPick(whichExtremity,
			tlx, tly,
			brx, bry,
			centroidx,
			centroidy,
			latestPos))
	    {
#if BS_DEBUG
	      fprintf(stderr, "s3\n");
	      fprintf(stderr, "latestPos: %f, %f   oldPos: %f, %f\n",latestPos[0], latestPos[1],
		      oldPos[0], oldPos[1]);
#endif	  
	      
	      newPos[0] = latestPos[0];
	      newPos[1] = latestPos[1];
	      if (Distance(latestPos[0],
			   latestPos[1],
			   oldPos[0],
			   oldPos[1]) < Threshold)
		{
#if BS_DEBUG
		  fprintf(stderr, "dist < thresh\n");
#endif	  
		  validCounter[whichExtremity - 1]++;
		  if (validCounter[whichExtremity - 1] > 5)
		    {
#if BS_DEBUG
		      fprintf(stderr, "validcounter > 5\n");
#endif	  
		      *validBit = 1;  /* IMPORTANT!
					 This is how we get back on track
					 after a series of invalid tracks. */
		      validCounter[whichExtremity - 1] = 0;
		      *centIndex += (3 + latestIndex);
		    }
		  else
		    *validBit = 0;
		}
	      else
		{
		  *validBit = 0;
		  validCounter[whichExtremity - 1] = 0;
		}
	    }
	  else
	    {
#if BS_DEBUG
	      fprintf(stderr, "to scan contourbyextremity\n");
#endif	  
	      ScanContourByExtremityID(whichExtremity,
				       centroidx,
				       centroidy,
				       contour,
				       contourLength,
				       tlx, tly,
				       brx, bry,
				       newPos);
	      *validBit = 0;
	    }
	}  /* end got leftmost point */
      else
	{
#if BS_DEBUG
	  fprintf(stderr, "s4\n");
#endif
#if BS_DEBUG
	  fprintf(stderr, "to scan contourbyextremity2\n");
#endif	  
	  ScanContourByExtremityID(whichExtremity,
				   centroidx,
				   centroidy,
				   contour,
				   contourLength,
				   tlx, tly,
				   brx, bry,
				   newPos);
	  *validBit = 0;
	}  /* end didn't get leftmost point */
}  /* end of procedure */
  
void
  LookForLeftFoot(float centroidx, float centroidy,
		  float tlx, float tly,
		  float brx, float bry,
		  float *contour,
		  int contourLength,
		  unsigned long int frameNumber,
		  float *extremityCentroidList,
		  int extListLength,
		  int *centIndex,
		  float *oldPos,
		  bitMaskType oldBitMask,
		  float *lastValidPos,
		  unsigned long int lastValidFrame,
		  bitMaskType *newBitMask,
		  float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = *centIndex;

  while(endCounter < (extListLength - 3) && 
	(Quadrant(centroidx, 
		  centroidy, 
		  extremityCentroidList[endCounter], 
		  extremityCentroidList[endCounter + 1]) == 1) &&
	(InBottomOfBoundingBox(tlx, tly, brx, bry, extremityCentroidList[endCounter + 1])))
    endCounter += 3;
  
  LookForNewPoint(LEFT_FOOT_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.left_foot_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->left_foot_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForLeftHand(float centroidx, float centroidy,
		  float tlx, float tly,
		  float brx, float bry,
		  float *contour,
		  int contourLength,
		  unsigned long int frameNumber,
		  float *extremityCentroidList,
		  int extListLength,
		  int *centIndex,
		  float *oldPos,
		  bitMaskType oldBitMask,
		  float *lastValidPos,
		  unsigned long int lastValidFrame,
		  bitMaskType *newBitMask,
		  float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = *centIndex;
  int pointQuad;

  if (!oldBitMask.left_hand_status)
    {

      /* thad-do we need to flush out bad extremities here ?*/
      pointQuad = Quadrant(centroidx, 
			   centroidy, 
			   extremityCentroidList[endCounter], 
			   extremityCentroidList[endCounter + 1]);
      
      while(endCounter < (extListLength - 3) && 
	    (pointQuad == 1 || pointQuad == 2))
	{
	  endCounter += 3;
	  if (endCounter < extListLength)
	    pointQuad = Quadrant(centroidx,
				 centroidy, 
				 extremityCentroidList[endCounter], 
				 extremityCentroidList[endCounter + 1]);
	}
    }
  else
    {
      startCounter = 0;
      endCounter = extListLength + 1;
    }

  LookForNewPoint(LEFT_HAND_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.left_hand_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->left_hand_status),
		  newPos);
  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForHead(float centroidx, float centroidy,
	      float tlx, float tly,
	      float brx, float bry,
	      float *contour,
	      int contourLength,
	      unsigned long int frameNumber,
	      float *extremityCentroidList,
	      int extListLength,
	      int *centIndex,
	      float *oldPos,
	      bitMaskType oldBitMask,
	      float *lastValidPos,
	      unsigned long int lastValidFrame,
	      bitMaskType *newBitMask,
	      float *newPos)
{
  int startCounter = *centIndex;
  int endCounter;
  int pointQuad = 0;
  static float HeadThreshold = 10.;
/*
  if (!oldBitMask.head_status)
    {
*/      
      /* Flush out bad extremities in case left hand and foot were both invalid */
      
/*      while((startCounter < extListLength - 1) &&
	    (Quadrant(centroidx,
		      centroidy, 
		      extremityCentroidList[startCounter],
		      extremityCentroidList[startCounter + 1]) == 1))
	startCounter += 3;
      
#if BS_DEBUG
      fprintf(stderr, "Head1: start = %d, end = %d\n", startCounter, endCounter);
#endif
      
      if (startCounter < extListLength - 1)
	pointQuad = Quadrant(centroidx,
			     centroidy,
			     extremityCentroidList[startCounter],
			     extremityCentroidList[startCounter + 1]);
      
#if BS_DEBUG
      fprintf(stderr, "Head2: start = %d, end = %d\n", startCounter, endCounter);
#endif
      
      while((startCounter < extListLength - 1) && 
	    (pointQuad == 2 || pointQuad == 3) &&
	    ((FloatAbs(centroidx - extremityCentroidList[startCounter]) >= HeadThreshold) ||
	     (extremityCentroidList[startCounter + 1] >= centroidy)))
	{
	  startCounter += 3;
	  if (startCounter < extListLength - 1)
	    pointQuad = Quadrant(centroidx,
				 centroidy,
				 extremityCentroidList[startCounter],
				 extremityCentroidList[startCounter + 1]);
	}
      
#if BS_DEBUG
      fprintf(stderr, "Head3: start = %d, end = %d\n", startCounter, endCounter);
#endif
      
      endCounter = extListLength - 1;
*/      
      /* Walk in from the other end of the list to "wall in" all possible extremities
	 that *could be* the head. */
      
#if BS_DEBUG
      fprintf(stderr, "Head4: start = %d, end = %d\n", startCounter, endCounter);
#endif
      
/*      while((endCounter > startCounter) &&
	    (Quadrant(centroidx,
		      centroidy,
		      extremityCentroidList[endCounter],
		      extremityCentroidList[endCounter + 1]) == 4))
	endCounter -= 3;
      
#if BS_DEBUG
      fprintf(stderr, "Head5: start = %d, end = %d\n", startCounter, endCounter);
#endif
      
      if(endCounter > startCounter) 
	pointQuad = Quadrant(centroidx,
			     centroidy,
			     extremityCentroidList[endCounter],
			     extremityCentroidList[endCounter + 1]);
      
#if BS_DEBUG
      fprintf(stderr, "Head6: start = %d, end = %d\n", startCounter, endCounter);
#endif
      
      while((endCounter > startCounter) && 
	    (pointQuad == 2 || pointQuad == 3) &&
	    ((FloatAbs(centroidx - extremityCentroidList[endCounter]) >= HeadThreshold) ||
	     (extremityCentroidList[endCounter + 1] >= centroidy)))
	{
	  endCounter -= 3;
	  if (endCounter > startCounter)
	    pointQuad = Quadrant(centroidx,
				 centroidy,
				 extremityCentroidList[endCounter],
				 extremityCentroidList[endCounter + 1]);
	}
      
#if BS_DEBUG
      fprintf(stderr, "Head7: start = %d, end = %d\n", startCounter, endCounter);
#endif
*/      
      /* Leave a "pad" at the end so the other procedures will work */
      
      if ((endCounter == startCounter) && (endCounter != extListLength - 1))
	endCounter += 3;
/*      
#if BS_DEBUG
      fprintf(stderr, "Head8: start = %d, end = %d\n", startCounter, endCounter);
#endif
    }
  else
    {
      startCounter = 0;
      endCounter = extListLength + 1;
    }
*/

  startCounter = *centIndex;
  endCounter = extListLength + 1;

  LookForNewPoint(HEAD_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.head_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->head_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForRightHand(float centroidx, float centroidy,
		   float tlx, float tly,
		   float brx, float bry,
		   float *contour,
		   int contourLength,
		   unsigned long int frameNumber,
		   float *extremityCentroidList,
		   int extListLength,
		   int *centIndex,
		   float *oldPos,
		   bitMaskType oldBitMask,
		   float *lastValidPos,
		   unsigned long int lastValidFrame,
		   bitMaskType *newBitMask,
		   float *newPos)
{
  int startCounter = *centIndex;
  int endCounter;
  int pointQuad;

  
  if (!oldBitMask.right_hand_status)
    {
      
      /* Flush out bad extremities in case left hand, foot and head were all invalid */
      
      while((startCounter < extListLength - 1) &&
	    (Quadrant(centroidx,
		      centroidy,
		      extremityCentroidList[startCounter],
		      extremityCentroidList[startCounter + 1]) < 3))
	startCounter += 3;
      
      
      pointQuad = Quadrant(centroidx,
			   centroidy,
			   extremityCentroidList[startCounter],
			   extremityCentroidList[startCounter + 1]);
      
      endCounter = startCounter;
      
      while((endCounter <= extListLength) && 
	    (pointQuad == 3 || pointQuad == 4))
	{
	  endCounter += 3;
	  if (endCounter <= extListLength)
	    pointQuad = Quadrant(centroidx,
				 centroidy,
				 extremityCentroidList[endCounter],
				 extremityCentroidList[endCounter + 1]);
	}
    }
  else
    {
      startCounter = 0;
      endCounter = extListLength + 1;
    }

  LookForNewPoint(RIGHT_HAND_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.right_hand_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->right_hand_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  LookForRightFoot(float centroidx, float centroidy,
		   float tlx, float tly,
		   float brx, float bry,
		   float *contour,
		   int contourLength,
		   unsigned long int frameNumber,
		   float *extremityCentroidList,
		   int extListLength,
		   int *centIndex,
		   float *oldPos,
		   bitMaskType oldBitMask,
		   float *lastValidPos,
		   unsigned long int lastValidFrame,
		   bitMaskType *newBitMask,
		   float *newPos)
{
  int startCounter = *centIndex;
  int endCounter = *centIndex;

  /* Flush out bad extremities in case previous ones weren't valid and
     didn't take care of this for us */

  while ((startCounter < extListLength - 1) && 
	 (Quadrant(centroidx,
		   centroidy,
		   extremityCentroidList[startCounter],
		   extremityCentroidList[startCounter + 1]) < 4))
    startCounter += 3;
  
  while ((startCounter < extListLength - 1) &&
	 (Quadrant(centroidx,
		   centroidy,
		   extremityCentroidList[startCounter],
		   extremityCentroidList[startCounter + 1]) == 4) &&
	 (InTopOfBoundingBox(tlx, tly, brx, bry, extremityCentroidList[startCounter + 1])))
    startCounter += 3;
  
  endCounter = startCounter;
  
  while(endCounter <= extListLength && 
	(Quadrant(centroidx,
		  centroidy,
		  extremityCentroidList[startCounter],
		  extremityCentroidList[startCounter + 1]) == 4) &&
	(!(InTopOfBoundingBox(tlx, tly, brx, bry, extremityCentroidList[endCounter + 1]))))
    endCounter += 3;

  LookForNewPoint(RIGHT_FOOT_EXTREMITY_ID,
		  centroidx,
		  centroidy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  contourLength,
		  frameNumber,
		  extremityCentroidList,
		  startCounter,
		  endCounter,
		  centIndex,
		  oldPos,
		  oldBitMask.right_foot_status,
		  lastValidPos,
		  lastValidFrame,
		  &(newBitMask->right_foot_status),
		  newPos);

  if (*centIndex > (extListLength - 3))
    *centIndex = extListLength - 3;
}

void
  DoFinalSanityCheck(float *newLeftFoot,
		     float *newLeftHand,
		     float *newHead,
		     float *newRightHand,
		     float *newRightFoot,
		     bitMaskType *newBitMask,
		     float Threshold)
{
  if (Distance(newLeftFoot[0], 
	       newLeftFoot[1], 
	       newLeftHand[0], 
	       newLeftHand[1]) < Threshold)
    {
      newBitMask->left_foot_status = 0;
      newBitMask->left_hand_status = 0;
    }
/*thad-for good measure */
  if (Distance(newRightHand[0], 
	       newRightHand[1], 
	       newLeftHand[0], 
	       newLeftHand[1]) < Threshold)
    {
      newBitMask->right_hand_status = 0;
      newBitMask->left_hand_status = 0;
      exit;
    }
  if (Distance(newHead[0], 
	       newHead[1], 
	       newLeftHand[0], 
	       newLeftHand[1]) < Threshold)
    {
      newBitMask->right_hand_status = 0;
      newBitMask->left_hand_status = 0;
      exit;
    }
  if (Distance(newRightFoot[0], 
	       newRightFoot[1], 
	       newLeftHand[0], 
	       newLeftHand[1]) < Threshold)
    {
      newBitMask->right_hand_status = 0;
      newBitMask->left_hand_status = 0;
      exit;
    }
  if (Distance(newLeftFoot[0], 
	       newLeftFoot[1], 
	       newLeftHand[0], 
	       newLeftHand[1]) < Threshold)
    {
      newBitMask->right_hand_status = 0;
      newBitMask->left_hand_status = 0;
      exit;
    }
  if (Distance(newLeftHand[0], 
	       newLeftHand[1], 
	       newHead[0], 
	       newHead[1]) < Threshold)
    {
      newBitMask->left_hand_status = 0;
      newBitMask->head_status = 0;
    }
  if (Distance(newHead[0], 
	       newHead[1], 
	       newRightHand[0], 
	       newRightHand[1]) < Threshold)
    {
      newBitMask->head_status = 0;
      newBitMask->right_hand_status = 0;
    }
  if (Distance(newRightHand[0], 
	       newRightHand[1], 
	       newRightFoot[0], 
	       newRightFoot[1]) < Threshold)
    {
      newBitMask->right_hand_status = 0;
      newBitMask->right_foot_status = 0;
    }
#if BS_DEBUG_THAD
  fprintf(stderr, "head %d %d\n", newHead[0], newHead[1]);
  fprintf(stderr, "left hand %d %d\n", newLeftHand[0], newLeftHand[1]);
  fprintf(stderr, "right hand %d %d\n", newRightHand[0], newRightHand[1]);
  fprintf(stderr, "left foot %d %d\n", newLeftFoot[0], newLeftFoot[1]);
  fprintf(stderr, "right foot %d %d\n", newRightFoot[0],
	  newRightFoot[1]);
#endif
#if SLEEP
  sleep(1);
#endif
}      

void
  trackBodyFeatures(const float *contour, /* 3 coords */
		    int numPoints,
		    const float *extrema, /* 3 coords */
		    int numExtrema,
		    float cx, float cy,   /* centroid x, centroid y */
		    float tlx, float tly, /* bounding box: top left x, top left y */
		    float brx, float bry, /* bounding box: bottom right x, bottom right y */
		    float *head,          /* index 0 = x coord, index 1 = y coord */
		    float *left_hand,
		    float *right_hand,
		    float *left_foot,
		    float *right_foot,
		    boolean *head_valid,
		    boolean *left_hand_valid,
		    boolean *right_hand_valid,
		    boolean *left_foot_valid,
		    boolean *right_foot_valid)
{
  static unsigned long int frameNumber = 0;
  static unsigned long int lastValidLeftFootFrame = 0;
  static unsigned long int lastValidRightFootFrame = 0;
  static unsigned long int lastValidLeftHandFrame = 0;
  static unsigned long int lastValidRightHandFrame = 0;
  static unsigned long int lastValidHeadFrame = 0;
  static float sanityThreshold = 8.;    /* old value was 6.375 */

  static bitMaskType oldBitMask = {0,0,0,0,0}, bitMask = {0,0,0,0,0};
  int centIndex = 0;
  static float oldHead[2] = {0., 0.};
  static float oldLeftHand[2] = {0., 0.};
  static float oldRightHand[2] = {0., 0.};
  static float oldLeftFoot[2] = {0., 0.};
  static float oldRightFoot[2] = {0., 0.};
  static float lastValidHead[2] = {0., 0.};
  static float lastValidLeftHand[2] = {0., 0.};
  static float lastValidRightHand[2] = {0., 0.};
  static float lastValidLeftFoot[2] = {0., 0.};
  static float lastValidRightFoot[2] = {0., 0.};

  sanityCheck(extrema,
	      numExtrema,
	      tlx,
	      tly,
	      brx,
	      bry);

  LookForLeftFoot(cx, cy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  (numPoints * 3),
		  frameNumber,
		  extrema,
		  (numExtrema * 3),
		  &centIndex,
		  oldLeftFoot,
		  oldBitMask,
		  lastValidLeftFoot,
		  lastValidLeftFootFrame,
		  &bitMask,
		  left_foot);
  LookForLeftHand(cx, cy,
		  tlx, tly,
		  brx, bry,
		  contour,
		  (numPoints * 3),
		  frameNumber,
		  extrema,
		  (numExtrema * 3),
		  &centIndex,
		  oldLeftHand,
		  oldBitMask,
		  lastValidLeftHand,
		  lastValidLeftHandFrame,
		  &bitMask,
		  left_hand);
  LookForHead(cx, cy,
	      tlx, tly,
	      brx, bry,
	      contour,
	      (numPoints * 3),
	      frameNumber,
	      extrema,
	      (numExtrema * 3),
	      &centIndex,
	      oldHead,
	      oldBitMask,
	      lastValidHead,
	      lastValidHeadFrame,
	      &bitMask,
	      head);
  LookForRightHand(cx, cy,
		   tlx, tly,
		   brx, bry,
		   contour,
		   (numPoints * 3),
		   frameNumber,
		   extrema,
		   (numExtrema * 3),
		   &centIndex,
		   oldRightHand,
		   oldBitMask,
		   lastValidRightHand,
		   lastValidRightHandFrame,
		   &bitMask,
		   right_hand);
  LookForRightFoot(cx, cy,
		   tlx, tly,
		   brx, bry,
		   contour,
		   (numPoints * 3),
		   frameNumber,
		   extrema,
		   (numExtrema * 3),			   
		   &centIndex,
		   oldRightFoot,
		   oldBitMask,
		   lastValidRightFoot,
		   lastValidRightFootFrame,
		   &bitMask,
		   right_foot);
  DoFinalSanityCheck(left_foot,
		     left_hand,
		     head,
		     right_hand,
		     right_foot,
		     &bitMask,
		     sanityThreshold);
  

#if BS_DEBUG_THAD
  fprintf(stderr, "bb upper left %f %f lower right %f %f\n", 
	  tlx,
	  tly,
	  brx,
	  bry);
#endif
	
  if (bitMask.head_status)
    {
      lastValidHead[0] = head[0];
      lastValidHead[1] = head[1];
      lastValidHeadFrame = frameNumber;
      head[0] = (.1 * oldHead[0]) + (.9 * head[0]);
      head[1] = (.1 * oldHead[1]) + (.9 * head[1]);
    }
  else
    {
      head[0] = (.4 * oldHead[0]) + (.6 * head[0]);
      head[1] = (.4 * oldHead[1]) + (.6 * head[1]);
    }

  if (bitMask.left_foot_status)
    {
      lastValidLeftFoot[0] = left_foot[0];
      lastValidLeftFoot[1] = left_foot[1];
      lastValidLeftFootFrame = frameNumber;
      left_foot[0] = (.1 * oldLeftFoot[0]) + (.9 * left_foot[0]);
      left_foot[1] = (.1 * oldLeftFoot[1]) + (.9 * left_foot[1]);
    }
  else
    {
      left_foot[0] = (.4 * oldLeftFoot[0]) + (.6 * left_foot[0]);
      left_foot[1] = (.4 * oldLeftFoot[1]) + (.6 * left_foot[1]);
    }

  if (bitMask.left_hand_status)
    {
      lastValidLeftHand[0] = left_hand[0];
      lastValidLeftHand[1] = left_hand[1];
      lastValidLeftHandFrame = frameNumber;
      left_hand[0] = (.1 * oldLeftHand[0]) + (.9 * left_hand[0]);
      left_hand[1] = (.1 * oldLeftHand[1]) + (.9 * left_hand[1]);
    }
  else
    {
      left_hand[0] = (.4 * oldLeftHand[0]) + (.6 * left_hand[0]);
      left_hand[1] = (.4 * oldLeftHand[1]) + (.6 * left_hand[1]);
    }      

  if (bitMask.right_hand_status)
    {
      lastValidRightHand[0] = right_hand[0];
      lastValidRightHand[1] = right_hand[1];
      lastValidRightHandFrame = frameNumber;
      right_hand[0] = (.1 * oldRightHand[0]) + (.9 * right_hand[0]);
      right_hand[1] = (.1 * oldRightHand[1]) + (.9 * right_hand[1]);
    }
  else
    {
      right_hand[0] = (.4 * oldRightHand[0]) + (.6 * right_hand[0]);
      right_hand[1] = (.4 * oldRightHand[1]) + (.6 * right_hand[1]);
    }

  if (bitMask.right_foot_status)
    {
      lastValidRightFoot[0] = right_foot[0];
      lastValidRightFoot[1] = right_foot[1];
      lastValidRightFootFrame = frameNumber;
      right_foot[0] = (.1 * oldRightFoot[0]) + (.9 * right_foot[0]);
      right_foot[1] = (.1 * oldRightFoot[1]) + (.9 * right_foot[1]);
    }
  else
    {
      right_foot[0] = (.4 * oldRightFoot[0]) + (.6 * right_foot[0]);
      right_foot[1] = (.4 * oldRightFoot[1]) + (.6 * right_foot[1]);
    }
  
  frameNumber += 1;

  oldBitMask = bitMask;

  *head_valid = bitMask.head_status;
  *left_hand_valid = bitMask.left_hand_status;
  *left_foot_valid = bitMask.left_foot_status;
  *right_hand_valid = bitMask.right_hand_status;
  *right_foot_valid = bitMask.right_foot_status;
  

}


