#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "common.h"
#include "ring_buffer.h"
#include "filter.h"

#include "common.proto"
#include "ring_buffer.proto"
#include "filter.proto"


/* make a sliding window filter with window size _n_. */
SWFilter *
makeSWFilter(int n)
{
  SWFilter *swf;

  swf = (SWFilter *) getMemory(sizeof(SWFilter));
  swf->n = n;
  swf->sum = 0.0;
  swf->rb = makeRingBuffer(n);
  return swf;
}


void
freeSWFilter(SWFilter *swf)
{
  freeRingBuffer(swf->rb);
  free(swf);
}


/* Given the new value _value_, updates the filter and returns the new
   average. */
float
updateSWFilter(SWFilter *swf, float value)
{
  int i; 
  boolean error;
  
  /* if first element, fill the array with it */
  if (rbValid(swf->rb) < swf->n)
    {
      swf->sum = ((float) swf->n) * value;
      for(i=0; i<swf->n; i++)
	rbAddElement(swf->rb, value);
    }

  else
    {
      /* remove oldest */
      swf->sum -= rbGetElement(swf->rb, swf->n-1, &error);
      if(error) printf("updateSWFilter: error in get\n");
      /* and add newewst */
      swf->sum += value;
      rbAddElement(swf->rb, value);
    }

  return swf->sum / (float)swf->n;
}


float 
getAverageSWFilter(SWFilter *swf)
{
  return swf->sum / (float) swf->n;
}
