#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "ring_buffer.h"


#include "ring_buffer.proto"
#include "common.proto"


RingBuffer *
makeRingBuffer(int size)
{
  RingBuffer *rb = (RingBuffer *) getMemory(sizeof(RingBuffer));
  
  rb->ptr = 0;
  rb->size = size;
  rb->buf = (RBtype *) getMemory(sizeof(RBtype) * size);
  rb->added = 0;
  return rb;
}

void
freeRingBuffer(RingBuffer *rb)
{
  free(rb->buf);
  free(rb);       
}


int
rbSize(RingBuffer *rb)
{
  return rb->size;
}

void
rbAddElement(RingBuffer *rb, RBtype elt)
{
  rb->ptr = (rb->ptr + 1) % (rb->size);
  rb->buf[rb->ptr] = elt;
  rb->added++;
}


/* 
  rbGetElement
  return the element entered _back_ adds ago.  
  index 0 is the head, ie the last added object.  
  index 1 is the one entered before that.
  If the index is > the size of the buffers, signals 
  an error.  If the index reaches into invalid data as specified
  by the number of adds since the last reset, also signals error.
  Return on error will be garbage.  Error signaled by filling
  in _errror_ arg.
  */
RBtype
rbGetElement(RingBuffer *rb, int back, boolean *error)
{
  int index;
  RBtype crap=0.0;
  *error = FALSE;

  if (back >= rb->size)
    {
      *error = TRUE;
      return crap;
    }
  if (back >= rb->added) 
    {
      *error = TRUE;
      return crap;
    }
  if (back < 0)
    {
      *error = TRUE;
      return crap;
    }
  index = (rb->ptr - back + rb->size) % rb->size;
  return rb->buf[index];
}


void
rbReset(RingBuffer *rb)
{
  rb->added = 0;
}

/* return the number of valid values in _rb_ */
int
rbValid(RingBuffer *rb)
{
  return MIN(rb->added, rb->size);
}


