// VLImage.c++

#include <stdlib.h>
#include <string.h>
#include "VLImage.h++"

VLServer VLImage::server = 0;
VLNode VLImage::drain = 0;
VLNode VLImage::source = 0;
VLPath VLImage::path = 0;
VLBuffer VLImage::buffer = 0;
VLInfoPtr VLImage::info = 0;
VLControlValue VLImage::val;
int VLImage::capture_width = 0;
int VLImage::capture_height = 0;
int VLImage::capture_size = 0;
int VLImage::capture_type = VLImage::RGB;
int VLImage::nbuffers = 40;
int VLImage::input = VLImage::INDY_SVIDEO;
int VLImage::verbose = 0;
int VLImage::zoom = 1;
int VLImage::transfer_begun = 0;

void
VLImage::Error( char *message )
{
  char tag[128];
  sprintf( tag, "VLImage: %s", message );
  vlPerror( tag );
  exit( 1 );
}

void
VLImage::Init()
{
  // this code works for Indy and Indigo (should work for all VL)

  if( transfer_begun )
     return;

  // (0) Connect to the server

  if( !( server = vlOpenVideo("") ) )
     Error( "open video" );

  // (1) Find the "Analog Video Input" node number
  //     It is different for Indy and Indigo, and probably
  //     different still for the Sirius

  VLDevList listOfDevices;
  if( vlGetDeviceList( server, &listOfDevices ) < 0 )
     Error( "getting device list" );

  for( int nodeIndex = 0;
       nodeIndex < listOfDevices.devices[0].numNodes &&
       strcmp( listOfDevices.devices[0].nodes[nodeIndex].name,
	       (input == -1 ? "Digital Video Input" :
		"Analog Video Input" ));
       nodeIndex++ );
  if( nodeIndex >= listOfDevices.devices[0].numNodes )
     Error( "couldn't find Analog Video Input node" );
  int analVidInputNodeNum = listOfDevices.devices[0].nodes[nodeIndex].number;

  // (2) Get the AVI node we just located, and any memory drain
  //     The path has to be setup before trying to set controls

  if( -1 == (drain = vlGetNode( server, VL_DRN, VL_MEM, VL_ANY )) )
    Error( "get drain node" );
  if( -1 == (source = vlGetNode( server, VL_SRC, VL_VIDEO, analVidInputNodeNum )))
    Error( "get source node" );
  if( -1 == (path = vlCreatePath( server, VL_ANY, source, drain )) )
    Error( "create path" );
  if( vlSetupPaths( server, (VLPathList)&path, 1, VL_SHARE, VL_SHARE ) < 0 )
     Error( "setup paths" );

  // (3) Set controls
  //     For some reason (??) you have to set Input (VL_MUXSWITCH)
  //     before the other controls or the Indy will barf (bug, IMHO)

  if (input >= 0) Input( input );
  Encoding( capture_type );
  Zoom( zoom );

  // (4) Make and register buffers, start the transfer

  buffer = vlCreateBuffer( server, path, drain, nbuffers );
  if( !buffer )
     Error( "create buffers" );
  vlRegisterBuffer( server, path, drain, buffer );
  if( vlBeginTransfer( server, path, 0, NULL ) )
     Error( "begin transfer" );
}

void
VLImage::Close()
{
  if( !transfer_begun )
     return;
  vlEndTransfer( server, path );
  vlDeregisterBuffer( server, path, drain, buffer );
  vlDestroyBuffer( server, buffer );
  vlDestroyPath( server, path );
  vlCloseVideo( server );
  transfer_begun = 0;
}

int
VLImage::NBuffers( int n )
{
  if( QUERY != n && !transfer_begun )
     nbuffers = n;
  return nbuffers;
}

int
VLImage::Encoding( int mode )
{
  if( QUERY != mode )
     {
       if( !server )
	  return( capture_type = mode );
       switch( mode )
	  {
	   case YUV: val.intVal = VL_PACKING_YVYU_422_8; break;
	   case RGB: default: val.intVal = VL_PACKING_RGB_8; break;
	  }
       if( vlSetControl( server, path, drain, VL_PACKING, &val ) < 0 )
	  Error( "set packing control" );
     }
  vlGetControl( server, path, drain, VL_PACKING, &val );
  switch( val.intVal )
     {
      case VL_PACKING_YVYU_422_8: capture_type = YUV; break;
      case VL_PACKING_RGB_8: capture_type = RGB; break;
      default: capture_type = -1; break;
     }
  return capture_type;
}

int
VLImage::Zoom( int decimation )
{
  if( QUERY != decimation )
     {
       if( !server )
	  return( zoom = decimation );
       val.fractVal.numerator = 1;
       val.fractVal.denominator = decimation;
       if( vlSetControl( server, path, drain, VL_ZOOM, &val ) < 0 )
	  Error( "set zoom control" );
       vlGetControl( server, path, drain, VL_SIZE, &val );
       capture_height = val.xyVal.y;
       capture_width = val.xyVal.x;
       capture_size = capture_height * capture_width;
     }
  vlGetControl( server, path, drain, VL_ZOOM, &val );
  return val.intVal;
}

int
VLImage::Input( int in )
{
  if( !server )
     return( input = in );
  if( QUERY != in )
     {
       val.intVal = in;
       if ( vlSetControl( server, path, source, VL_MUXSWITCH, &val ) < 0 )
	 ;
       // Error( "set source control" );
     }
  vlGetControl( server, path, source, VL_MUXSWITCH, &val );
  return val.intVal;
}

int
VLImage::Verbose( int in )
{
  // Only set if 0 or 1; else just echo value
  if( 0 == in )
     verbose = 0;
  if( 1 == in )
     verbose = 1;
  return verbose;
}

VLImage::VLImage()
  : size(0), width(0), dataPtr(0)
{}

VLImage&
VLImage::Grab()
{
#if 0
  if( Type() != capture_type )
     Error( "type mismatch" );
#endif
#define NOLAG
#ifdef NOLAG
  vlGetLatestValid( server, buffer );
  vlPutFree(server,buffer);
#endif
  while( !( info = vlGetLatestValid( server, buffer ) ) )
    sginap( 0 );

  dataPtr = (unsigned char *) vlGetActiveRegion( server, buffer, info );
  size = capture_size;
  width = capture_width;
  return *this;
}

// -------- VLRgbImage --------

int VLRgbImage::type = VLImage::RGB;

ostream&
VLRgbImage::Write( ostream& os )
{
  unsigned char *ucp = dataPtr;
  for( int count = size; count--; )
     {
       os.put( ucp[3] );
       os.put( ucp[2] );
       os.put( ucp[1] );
       ucp += 4;
     }
  return os;
}  

VLPixel
VLRgbImage::Pixel( int i )
{
  static VLRgbPixel pixel;
  unsigned char *ucp = dataPtr + i*4 + 1;
  pixel.Blue() = *ucp++;
  pixel.Green() = *ucp++;
  pixel.Red() = *ucp++;
  return pixel;
}

// -------- VLYuvImage --------

int VLYuvImage::type = VLImage::YUV;

ostream&
VLYuvImage::Write( ostream& os )
{
  unsigned char *ucp = dataPtr;
  for( int count = size; count--; )
     {
       os.put( ucp[1] );
       os.put( ucp[0] );
       os.put( ucp[3] );
       os.put( ucp[2] );
       ucp += 4;
     }
  return os;
}

VLPixel
VLYuvImage::Pixel( int i )
{
  static VLYuvPixel pixel;
  pixel.Y() = Y(i);
  pixel.U() = U(i);
  pixel.V() = V(i);
  return pixel;
}

