// VLImage.h++

#ifndef _VLImage_H_
#define _VLImage_H_

#include <vl/vl.h>
#include <unistd.h>
#include <stream.h>

// -------- VLPixel --------

class VLPixel
{
public:
  unsigned char ch1;
  unsigned char ch2;
  unsigned char ch3;
};

// -------- VLRgbPixel --------

class VLRgbPixel : public VLPixel
{
public:
  inline unsigned char& Red() { return ch1; }
  inline unsigned char& Green() { return ch2; }
  inline unsigned char& Blue() { return ch3; }
  inline unsigned char& R() { return ch1; }
  inline unsigned char& G() { return ch2; }
  inline unsigned char& B() { return ch3; }
};

// -------- VLYuvPixel --------

class VLYuvPixel : public VLPixel
{
public:
  inline unsigned char& Y() { return ch1; }
  inline unsigned char& U() { return ch2; }
  inline unsigned char& V() { return ch3; }
};

// -------- VLImage --------

class VLImage
{
private:

  static VLServer server;	// Set by Init
  static VLNode drain;		// Set by Init
  static VLNode source;		// Set by Init
  static VLPath path;		// Set by Init
  static VLBuffer buffer;	// Set by Init
  static VLInfoPtr info;
  static VLControlValue val;
  static int transfer_begun;	// Set by Init and Close
  static int capture_width;	// Set by Zoom
  static int capture_height;	// Set by Zoom
  static int capture_size;	// Set by Zoom (height * width)
  static int capture_type;	// Set by Encoding (RGB or YUV)
  static int nbuffers;		// Set by NBuffers
  static int input;
  static int zoom;
  static int verbose;

  static void Error( char * = "VLImage" );

public:

  enum { QUERY = -1, RGB = 0, YUV = 1 };
  enum { GALILEO_SVIDEO_1 = 0,
	   GALILEO_SVIDEO_2 = 1,
	   GALILEO_SVIDEO_3 = 2,
	   GALILEO_COMPOSITE_1 = 3,
	   GALILEO_COMPOSITE_2 = 4,
	   GALILEO_COMPOSITE_3 = 5,
	   GALILEO_YRB_1 = 6, 
	   GALILEO_YRB_2 = 7 };
  enum { INDY_INDYCAM = -1, INDY_COMPOSITE = 0, INDY_SVIDEO = 1 };

  static void Init();
  static void Close();
  static int NBuffers( int = QUERY );
  static int Encoding( int = QUERY );
  static int Zoom( int = QUERY );
  static int Input( int = QUERY );
  static int Verbose( int = QUERY );
  static inline int GetSize( int*hp=0, int*wp=0 );
  static inline int GetHeight();
  static inline int GetWidth();

public:

  int size;
  int width;
  unsigned char *dataPtr;

public:

  VLImage();

  inline int Height() { return( size/((width)?width:1) ); }
  inline int Size() { return size; }
  inline int Width() { return width; }
  inline int Index( int row, int col ) { return( row*width + col ); }

  VLImage& Grab();
  inline void Free();
  operator unsigned long *() { return (unsigned long *)dataPtr; }
  operator unsigned char *() { return (unsigned char *)dataPtr; }
  virtual ostream& Write( ostream& ) = 0;
  virtual VLPixel Pixel( int i ) = 0;
  virtual int Type() = 0;

};

inline
int
VLImage::GetSize( int *hp, int *wp )
{
  if( hp ) *hp = capture_height;
  if( wp ) *wp = capture_width;
  return capture_size;
}

inline
int
VLImage::GetHeight()
{
  return capture_height;
}

inline
int
VLImage::GetWidth()
{
  return capture_width;
}

inline
void
VLImage::Free()
{ 
  vlPutFree( server, buffer );
  size = 0;
  width = 0;
  dataPtr = 0;
}

inline ostream& operator<<( ostream& os, VLImage& vli ) { return vli.Write( os ); }

// -------- VLRgbImage --------

class VLRgbImage : public VLImage
{
  static int type;
public:
  unsigned char Red( int i ) { return dataPtr[i*4 + 3]; }
  unsigned char Green( int i ) { return dataPtr[i*4 + 2]; }
  unsigned char Blue( int i ) { return dataPtr[i*4 + 1]; }
  unsigned char R( int i ) { return dataPtr[i*4 + 3]; }
  unsigned char G( int i ) { return dataPtr[i*4 + 2]; }
  unsigned char B( int i ) { return dataPtr[i*4 + 1]; }
  ostream& Write( ostream& );
  VLPixel Pixel( int i );
  int Type() { return type; }
};

// -------- VLYuvImage --------

class VLYuvImage : public VLImage
{
  static int type;
public:
  unsigned char Y( int i ) { return dataPtr[i*2 + 1]; }
  unsigned char U( int i ) { return dataPtr[((i-i%2)*2)]; }
  unsigned char V( int i ) { return dataPtr[((i-i%2)*2)+2]; }
  ostream& Write( ostream& );
  VLPixel Pixel( int i );
  int Type() { return type; }
};

#if 0
inline float max( float a, float b ) { return( (a > b) ? a : b ); }
inline float min( float a, float b ) { return( (a < b) ? a : b ); }
inline float limit( float a, float low, float high ) { return min( high, max(low, a ) ); }
#endif

#endif /* _VLImage_H_ */
