// KalmanFilter.h++

#ifndef _KalmanFilter_H_
#define _KalmanFilter_H_

// This is a simple linear Kalman filter.
// Unlike my Ekf class, it is not resizeable, so it is meant
// for simple linear problems.

#include <stream.h>
#include "QuickMath/QuickMath.h"

class KalmanFilter
{
  static const float veryBig;
protected:
  int n;		// dimension of state
  int p;		// dimension of mmt
  int t;		// prediction time (1 + estimate time)
  AllocVector xP;
  AllocVector xE;
  AllocVector dX;	
  AllocVector yP;
  AllocVector dY;
  AllocMatrix Pp;
  AllocMatrix Pe;
  AllocMatrix K;
  AllocMatrix I;
  AllocMatrix tempNxP;
  AllocMatrix tempNxN;
  AllocMatrix tempPxP;
public:
  // Leave these public so the user can initialize or change them
  AllocMatrix H;
  AllocMatrix Phi;
  AllocMatrix Q;
  AllocMatrix R;
public:
  KalmanFilter( int nstates, int nmmts,
		QuickVector& xp, QuickMatrix& pp,
		QuickMatrix& h, QuickMatrix& r,	
		QuickMatrix& ph, QuickMatrix& q );
  KalmanFilter( int ns, float rv = 1., float qv = 1., float dt = 1. );
  KalmanFilter& Update( QuickVector& y );
  KalmanFilter& Update( float* y ) { return Update(QuickVector(p,y)); }
  QuickVector& StatePrediction() { return xP; }
  QuickVector& StateEstimate() { return xE; }
  QuickVector& Innovation() { return dY; }
  int PredictionTime() { return t; }
  int EstimateTime() { return( t-1 ); }
};

#endif /* _KalmanFilter_H_ */
