// SubVector.h

#ifndef _SubVector_H_
#define _SubVector_H_

#include <stddef.h>
#include <stream.h>
#include <assert.h>
#include "QuickMath/real.h"
#include "QuickMath/SubMatrix.h"

class SubVector
{
  friend class SubMatrix;
private:
protected:
  real *array;
  int elementCount;
  int spacing; // if 0, then packed
public:
  SubVector( )
    : array(0), elementCount(0) {}
  SubVector( int n, int s, real* a )
    : array(a), elementCount(n), spacing(s) {}
  // SubVector( const SubVector& );

  SubVector& Copy( const SubVector& sv );
  SubVector& operator=( const SubVector& sv ) { return Copy(sv); }

  int Length() const { return elementCount; }
  int ElementCount() const { return elementCount; }
  int Spacing() const { return spacing; }
  real* Array( int i = 0 ) { return( array + i ); }

  SubVector& Insert( real *p );
#if 0
  real* Extract( real *p );
  SubVector& operator<<( real *p ) { Insert(p); return *this; }
  real* operator>>( real *p ) { return Extract(p); }
#endif

#if 0
  SubVector SubVector();
  SubVector SubVector( int ulr, int ulc, int dh, int dw );
#endif

  SubVector& Add( const SubVector&, const SubVector& );
  SubVector& Subtract(const SubVector&, const SubVector& );
  // SubVector& Multiply(const SubVector&, const SubVector& );
  SubVector& Multiply( const SubMatrix&, const SubVector& );
  // SubVector& Divide( const SubVector&, const SubVector& );
  // SubVector& Negate( const SubVector& );
  // SubVector& Invert( const SubVector& );
  // SubVector& Transpose( const SubVector& );
  // SubVector& Shift( const SubVector&, real val );
  // SubVector& Scale( const SubVector&, real val  );
  SubVector& AddEq( const SubVector& op );
  SubVector& SubtractEq( const SubVector& op );
  // SubVector& MultiplyEq( const SubVector& op );
  // SubVector& DivideEq( const SubVector& op );
  // SubVector& NegateEq();
  // SubVector& InvertEq();
  SubVector& ShiftEq( real val );
  SubVector& ScaleEq( real val );
  SubVector& ConstEq( real val );
  SubVector& Zero() { return ConstEq( 0. ); }
  SubVector& operator+=( const SubVector& op ) { return AddEq(op); }
  // SubVector& operator-=( const SubVector& op ) { return SubtractEq(op); }
  // SubVector& operator*=( const SubVector& op ) { return MultiplyEq(op); }
  // SubVector& operator/=( const SubVector& op ) { return DivideEq(op); }
  // SubVector& operator+=( const real val ) { return ShiftEq(val); }
  SubVector& operator-=( const real val ) { return ShiftEq(-val); }
  // SubVector& operator*=( const real val ) { return ScaleEq(val); }
  SubVector& operator/=( const real val ) { return ScaleEq(real(1)/val); }
  // SubVector& operator=( const real val ) { return ConstEq(val); }
  real InnerProduct( const SubVector& ) const;
  int operator==( const SubVector& ) const;
  SubVector& Abs();

  ostream& Print( ostream& );
  istream& Scan( istream& );
  ostream& Write( ostream& );
  istream& Read( istream& );
  
  friend ostream& operator<<( ostream&, SubVector& );

  real& operator()( int i ) { return array[i]; }

};

#endif /* _SubVector_H_ */
