 /* bsClassify.c */

/* RCS: $Header: /mas/vision/projects/pfinder/RCS/bsClassify.c,v 6.10 1995/09/24 00:48:37 cwren Exp $ */

#include "bsClassify.h"
#include <string.h>
#include "bsColor.h"

typedef struct {
  color_class **background_classes;
  color_class **foreground_classes;
  int num_foreground_classes;
  int max_foreground_classes;
  int width,height;
} Backsub;

Backsub *bsClassifyState;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsClassifyGlobals
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int verbose = 0;


float CXYWEIGHT = 0.50;
int NORMALIZE      = ALL_NORMALIZE;
int FOREGROUND      = 1;
float bsClassifyTHRESH  = 110.0;
float bsClassifyOFFSET  = 1.0;
extern float bsSegmentTHRESH;
float UPDATEBG = 0.00;

float bsClassifyGetThreshMax()
{
  return (300.0);
}

float bsClassifyGetThresh()
{
  return (bsClassifyTHRESH);
}

void bsClassifySetThresh(float newvalue)
{
  bsClassifyTHRESH = newvalue;
}

float bsClassifyGetUpdateBGMax()
{
  return (1.0);
}

float bsClassifyGetUpdateBG()
{
  return (UPDATEBG);
}

void bsClassifySetUpdateBG(float newvalue)
{
  UPDATEBG = newvalue;
}

float bsClassifyGetOffsetMax()
{
  return (10.0);
}

float bsClassifyGetOffset()
{
  return (bsClassifyOFFSET);
}

void bsClassifySetOffset(float newvalue)
{
  bsClassifyOFFSET = newvalue;
}

void bsClassifySetNormalize(int newvalue) 
{
  if (newvalue)
    NORMALIZE = ALL_NORMALIZE;
  else
    NORMALIZE = NO_NORMALIZE;    
}

int bsClassifyGetNormalize() 
{
  return NORMALIZE;
}
void bsClassifySetForeground(int newvalue) 
{
  static int num = -1;
  
  if (num < 0)
    num = bsClassifyState->max_foreground_classes;

  if (!newvalue)
    bsClassifyState->num_foreground_classes = 1 ;
  else 
    {
      bsClassifyState->num_foreground_classes = num ;
      bsClassifyInitForegroundPriors(num);
    }

}

int bsClassifyGetForeground() 
{
  if (bsClassifyState->num_foreground_classes > 1)
    return 1;
  else
    return 0;
}

void
bsClassifySetVerbose( int val )
{
  verbose = val;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsClassifyInitialize
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


static color_class *bsClassifyNewBackgroundClass()
{
  color_class *new;

  if (!(new = (color_class *) malloc(sizeof(color_class))))
    {
      perror("failed to allocate class");
      return NULL;
    }

  new->fgp = 0;      

  if (!(new->inv_covar = MatrixCreate(3,3)))
    perror("failed to allocate covar");
  if (!(new->norm_inv_covar = MatrixCreate(2,2)))
    perror("failed to allocate norm covar");

  return new;
}
  
static void bsClassifyFreeBackgroundClass(color_class *ptr)
{
  MatrixFree(ptr->inv_covar);
  MatrixFree(ptr->norm_inv_covar);

  free(ptr);
}
  

void
bsClassifyInitialize(int width, int height, int maxfore)
{
  int i;
  int w1 = width, h1 = height;

  bsClassifyState = (Backsub *) malloc(sizeof(Backsub));
  bsClassifyState->background_classes = (color_class **)
    malloc(width * height * sizeof(color_class *));
  for (i=0; i<width*height; i++) 
    {
      bsClassifyState->background_classes[i] = 
	bsClassifyNewBackgroundClass();
    }
  bsClassifyState->foreground_classes = (color_class **)
    malloc(NCLASSES * sizeof(color_class *));
  bsClassifyState->max_foreground_classes = NCLASSES;
  bsClassifyState->num_foreground_classes = ((maxfore < NCLASSES) ? 
					     maxfore : 
					     NCLASSES);
  bsClassifyState->width = width;
  bsClassifyState->height = height;

  for (i = 0; i < NCLASSES; i++)
    {
      bsClassifyState->foreground_classes[i] = (color_class *) malloc(sizeof(color_class));
      bsClassifyState->foreground_classes[i]->fg = (fg_class *) malloc(sizeof(fg_class));

      bsClassifyState->foreground_classes[i]->fgp = 1;      

      bsClassifyState->foreground_classes[i]->inv_covar = MatrixCreate(3,3);
      bsClassifyState->foreground_classes[i]->fg->prior_inv_covar = 
	MatrixCreate(3,3);
      bsClassifyState->foreground_classes[i]->fg->covar_acc = 
	MatrixCreate(3,3);

      bsClassifyState->foreground_classes[i]->norm_inv_covar = 
	    MatrixCreate(2,2);
      bsClassifyState->foreground_classes[i]->fg->prior_norm_inv_covar = 
	MatrixCreate(2,2);
      bsClassifyState->foreground_classes[i]->fg->norm_covar_acc = 
	MatrixCreate(2,2);

      bsClassifyState->foreground_classes[i]->fg->xy_inv_covar = MatrixCreate(2,2);
      bsClassifyState->foreground_classes[i]->fg->prior_xy_inv_covar = 
	MatrixCreate(2,2);
      bsClassifyState->foreground_classes[i]->fg->xy_covar_acc = 
	MatrixCreate(2,2);

      bsClassifyState->foreground_classes[i]->singular = 1;
      bsClassifyState->foreground_classes[i]->fg->prior_set = 0;
      bsClassifyState->foreground_classes[i]->fg->prior_singular = 1;
      bsClassifyState->foreground_classes[i]->fg->last_num_points = 0;
      bsClassifyState->foreground_classes[i]->fg->num_points = 0;
    }
  bsClassifyInitForegroundPriors(NCLASSES);
  bsColorInitClasses(bsClassifyState->foreground_classes,
		     maxfore);
}


int bsClassifyXdim()
{
  return bsClassifyState->width;
}

int bsClassifyYdim()
{
  return bsClassifyState->height;
}

float
bsClassifyXYDiscrim(int x, int y, int i)
{
  bsXYDiscrim(x, y, bsClassifyState->foreground_classes[i]);
}


static int writematrix(FILE *dump, Matrix matrix)
{
  int i,j;
  for (i=0;i<matrix->width;i++)
    for (j=0;j<matrix->height;j++)
      {
	fwrite((void *) &(matrix->data[i][j]), sizeof(Real), 1, dump);
	if (ferror(dump)) 
	  {perror("error dumping inv_covar");return(0);}
      }
  return(1);
}

static int readmatrix(FILE *dump, Matrix matrix)
{
  int i,j;
  for (i=0;i<matrix->width;i++)
    for (j=0;j<matrix->height;j++)
      {
	fread((void *) &(matrix->data[i][j]), sizeof(Real), 1, dump);
	if (ferror(dump)) 
	  {perror("error dumping inv_covar");return(0);}
      }
  return(1);
}

int
  bsClassifyWriteBackgroundClasses( )
{
  FILE *dump; 
  int i;

  puts("Writing background file to /usr/tmp/pfinder.dump.");
  if ((dump = fopen ("/usr/tmp/pfinder.dump","w")) == NULL)
    {
      perror("failed to dump classes");
      return(0);
    }
  
  if (!fwrite((void *) &(bsClassifyState->num_foreground_classes),
	      sizeof(int),   1, dump))
    {perror("error dumping num_foreground_classes");return(0);}

  if (!fwrite((void *) &(bsClassifyState->height),
	      sizeof(int),   1, dump))
    {perror("error dumping height");return(0);}

  if (!fwrite((void *) &(bsClassifyState->width),
	      sizeof(int),   1, dump))
    {perror("error dumping width");return(0);}
  

  for (i=0; i<bsClassifyState->width*bsClassifyState->height; i++) 
    {
      register color_class *back = bsClassifyState->background_classes[i];


      if (!writematrix(dump,back->inv_covar)) return(0);
      if (!writematrix(dump,back->norm_inv_covar)) return(0);

      fwrite((void *) &(back->singular),		       sizeof(int),   1, dump);
      if (ferror(dump)) {perror("error dumping singular ");return(0);}
      fwrite((void *)  (back->mean),                    sizeof(ftype), 3, dump);
      if (ferror(dump)) 	{perror("error dumping mean");return(0);}
      fwrite((void *) &(back->ln_det_covar),            sizeof(ftype), 1, dump);
      if (ferror(dump)) 	{perror("error dumping ln_det_covar");return(0);}
      fwrite((void *)  (back->norm_mean),               sizeof(ftype), 2, dump);
      if (ferror(dump)) 	{perror("error dumping norm_mean");return(0);}
      fwrite((void *) &(back->norm_ln_det_covar),       sizeof(ftype), 1, dump);
      if (ferror(dump)) 	{perror("error dumping norm_ln_det_covar");return(0);}


    }
  if (fclose(dump) == EOF)
    {perror("error closing dump");return(0);}
  
  return(1);
}

int
  bsClassifyReadBackgroundClasses( )
{
  FILE *dump; 
  int i;
  int width,height,numfg;
  
  if ((dump = fopen ("/usr/tmp/pfinder.dump","r")) == NULL)
    {
      perror("failed to dump classes");
      return(0);
    }
  
  if (!fread((void *) &(numfg),
	      sizeof(int),   1, dump))
    {perror("error loading num_foreground_classes");return(0);}

  if (!fread((void *) &(height),
	      sizeof(int),   1, dump))
    {perror("error loading height");return(0);}

  if (!fread((void *) &(width),
	      sizeof(int),   1, dump))
    {perror("error loading width");return(0);}

  bsClassifyInitialize( width,  height, numfg);

  bsClassifyState->num_foreground_classes = bsClassifyState->max_foreground_classes;

  for (i=0; i<(bsClassifyState->width*bsClassifyState->height); i++)
    {
      register color_class *back = bsClassifyState->background_classes[i];



      if (!readmatrix(dump,back->inv_covar)) return(0);
      if (!readmatrix(dump,back->norm_inv_covar)) return(0);

      fread((void *) &(back->singular),		       sizeof(int),   1, dump);
      if (ferror(dump)) {perror("error loading singular ");return(0);}
      fread((void *)  (back->mean),                    sizeof(ftype), 3, dump);
      if (ferror(dump)) 	{perror("error loading mean");return(0);}
      fread((void *) &(back->ln_det_covar),            sizeof(ftype), 1, dump);
      if (ferror(dump)) 	{perror("error loading ln_det_covar");return(0);}
      fread((void *)  (back->norm_mean),               sizeof(ftype), 2, dump);
      if (ferror(dump)) 	{perror("error loading norm_mean");return(0);}
      fread((void *) &(back->norm_ln_det_covar),       sizeof(ftype), 1, dump);
      if (ferror(dump)) 	{perror("error loading norm_ln_det_covar");return(0);}

    }
  if (fclose(dump) == EOF)
    {perror("error closing load");return(0);}

  return(1);
}

void bsClassifyDisplayClasses (int n)
{
  int i;
  
  for(i=0;i<bsClassifyState->num_foreground_classes;i++)
    {
      color_class *fore = bsClassifyState->foreground_classes[i];
      if(bsClassifyGetConfidence(i)>0.5)
	bsXDrawCross((int)fore->fg->xy_mean[0],
		     (int)fore->fg->xy_mean[1],
		     n);
    }
}

void
  bsClassifyMakeBackgroundClasses( iveImage **frames, int n)
{

  int i;
  
  bsColorComputeBackgroundClassesBatch( frames, n,
				       bsClassifyState->background_classes );
  
}


void
  bsClassifyChromakey(char *chroma)
{

  int i;
  char buf[9];
  int r,g,b;
  
  if (strlen(chroma) < 6) {
    fprintf(stderr,"Invalid chroma, must be rrggbb, using 000000\n");
    r = g = b = 0;
  } else {
    buf[0] = chroma[0];
    buf[1] = chroma[1];
    buf[2] = ' ';
    buf[3] = chroma[2];
    buf[4] = chroma[3];
    buf[5] = ' ';
    buf[6] = chroma[4];
    buf[7] = chroma[5];
    buf[8] = '\000';
    
    sscanf(buf, "%x %x %x", &r, &g, &b);
  }
  
  if (verbose) 
    {
      printf("enabling chromakey mode to color triplet (%d,%d,%d)\n",
	     r,g,b);
    }
    
  bsColorInstallChromakey(r, g, b,
			  bsClassifyState->background_classes,
			  bsClassifyState->width, 
			  bsClassifyState->height);
}

int
  bsClassifyNewForegroundClass( )
{
	if ( bsClassifyState->num_foreground_classes >=
	    bsClassifyState->max_foreground_classes )
	  return -1;
	return ((bsClassifyState->num_foreground_classes)++);

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * printClasses
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


void bsClassifyPrintClasses(int width, int height, FILE *file)
{
  color_class **class= bsClassifyState->foreground_classes;
  int num_classes = bsClassifyState->num_foreground_classes;
  int i;
  
  for (i=0; i<num_classes ; i++)
    {
      if (class[i]->singular)  
	{
	  fprintf(file,"%e %e %e %e %e %e", 
		 0.0,0.0,0.0,0.0,0.0,0.0);
	}
      else
	{
	  float l[2],v[4];
	  float size;

	  size = (float) ((width < height) ? height : width);

	  l[0] = class[i]->fg->xyL[0]/size;
	  l[1] = class[i]->fg->xyL[1]/size;

	  v[0] = class[i]->fg->xyS[0][0] * l[0];
	  v[1] = class[i]->fg->xyS[1][0] * l[0];
	  v[2] = class[i]->fg->xyS[0][1] * l[1];
	  v[3] = class[i]->fg->xyS[1][1] * l[1];

	  fprintf(file,"%e %e %e %e %e %e",
		 class[i]->fg->xy_mean[0] / size,		 
		 class[i]->fg->xy_mean[1] / size,
		 v[0],v[1],v[2],v[3]);
	  
	}
      
      if (i<num_classes-1)
	{
	  fprintf(file,"\t");
	}
      else
	{
	  fprintf(file,"\n");
	}
    }
}

int
  bsClassifySetForegroundStats( iveImage *frame, iveBitmap *inMap )
{
  bsColorComputeForegroundClasses(inMap, frame, 
				  bsClassifyState->foreground_classes,
				  bsClassifyState->num_foreground_classes);
  bsColorInitClasses(bsClassifyState->foreground_classes,
		     bsClassifyState->num_foreground_classes);
}

int
  bsClassifySetForegroundStatsI( iveImage *frame, iveBitmap *inMap )
{
  int x,y;
  bsColorInitClasses(bsClassifyState->foreground_classes,
		     bsClassifyState->num_foreground_classes);

  for(x=0; x<inMap->width; x++)
    for(y=0; y<inMap->height; y++)
      {
	ivePixel ivep;
	
	iveImageGetPixel(frame,y,x,&ivep);
	bsColorAddExample(iveBitmapGetBit(inMap,y,x),
			  x, y,
			  &ivep,
			  bsClassifyState->foreground_classes,
			  bsClassifyState->num_foreground_classes);
      }
  
  bsColorComputeForegroundClasses(inMap, frame, 
				  bsClassifyState->foreground_classes,
				  bsClassifyState->num_foreground_classes);
}

int bsClassifySetClassPosition(int num, float x, float y)
{
  if (num < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[num];
      bsColorSetClassPosition(fore, x, y);
    }
}

int bsClassifyGetNumClasses()
{
  return (bsClassifyState->num_foreground_classes);
}

int bsClassifyGetScreenSize(int *w, int *h)
{
  *w = (float) bsClassifyState->width;
  *h = (float) bsClassifyState->height;
  return(1);
}

int bsClassifyGetClassPosition(int num, float *x, float *y)
{
  if (num < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[num];

      *x = fore->fg->xy_mean[0];
      *y = fore->fg->xy_mean[1];
      return(1);
    }
  return(0);
}

int bsClassifyGetClassColor(int num, float *r, float *g, float *b)
{
  if (num < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[num];

      *r = fore->mean[0];
      *g = fore->mean[1];
      *b = fore->mean[2];
      return(1);
    }
  return(0);
}

int bsClassifyGetClassBB(int num, float *x, float *y)
{
  if (num < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[num];

      *x = fore->fg->bb[0];
      *y = fore->fg->bb[1];
      return(1);
    }
  return(0);
}

int bsClassifyGetClassEigen(int num, float L[2], float S[2][2])
{
  if (num < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[num];

      L[0] = fore->fg->xyL[0];
      L[1] = fore->fg->xyL[1];
      S[0][0] = fore->fg->xyS[0][0];
      S[1][0] = fore->fg->xyS[1][0];
      S[0][1] = fore->fg->xyS[0][1];
      S[1][1] = fore->fg->xyS[1][1];

      return(1);
    }
  return(0);
}

float bsClassifyGetConfidence(int num)
{
  if (num < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[num];
      register int np = fore->fg->last_num_points;
      
      if (np < 5) {
	return(0.0);
      } else if (np <10) {
	return(((float)np-5.0) * 0.2);
      } else {
	return(1.0);
      }
    }
  return(0.0);
}

 
int
bsClassifyInitForegroundPrior( int index )
{
  if (index < bsClassifyState->num_foreground_classes)
    {
      register color_class *fore = bsClassifyState->foreground_classes[index];
      int width = bsClassifyState->width;
      int height = bsClassifyState->height;

      fore->singular = 1;
      fore->fg->prior_set = 1;
      fore->fg->prior_singular = 0;
      
      switch(index){
      case pfGLOBAL:
	bsColorSetClassSize(fore, XYWIDE);
	bsColorSetClassColor(fore,pfBROWN);
	bsColorSetClassPosition(fore, (float) width/2., (float) height/2.);
	fore->fg->xympriorweight = 0.0;
	fore->fg->xyvpriorweight = 0.0;
	fore->fg->cmpriorweight  = 0.0;
	fore->fg->cvpriorweight  = 0.0;
	fore->fg->cxyweight  = CXYWEIGHT;
	bsColorBlendInPriors(fore);
	fore->fg->prior_set = 0;
	break;
	
	break;
	
      case pfHEAD:
	bsColorSetClassColor(fore,pfFLESH);
	bsColorSetClassSize(fore, XYNARROW);
	bsColorSetClassPosition(fore, (float) width/2., (float) height/2.);
	fore->fg->xympriorweight = 0.0;
	fore->fg->xyvpriorweight = 0.5;
	fore->fg->cmpriorweight  = 0.9;
	fore->fg->cvpriorweight  = 0.9;
	fore->fg->cxyweight  = CXYWEIGHT;
	bsColorBlendInPriors(fore);
	break;
	
      case pfRIGHTHAND:
      case pfLEFTHAND:
	bsColorSetClassColor(fore,pfFLESH);
	bsColorSetClassSize(fore, XYNARROW);
	bsColorSetClassPosition(fore, (float) width/2., (float) height/2.);
	fore->fg->xympriorweight = 0.0;
	fore->fg->xyvpriorweight = 0.5;
	fore->fg->cmpriorweight  = 0.9;
	fore->fg->cvpriorweight  = 0.9;
	fore->fg->cxyweight  = CXYWEIGHT;
	bsColorBlendInPriors(fore);
	break;
	
      case pfCHEST:
      case pfWAIST:
      case pfTHIGHS:
	bsColorSetClassColor(fore,pfGREY);
	bsColorSetClassSize(fore, XYMEDIUM);
	bsColorSetClassPosition(fore, (float) width/2., (float) height/2.);
	fore->fg->xympriorweight = 0.0;
	fore->fg->xyvpriorweight = 1.0;
	fore->fg->cmpriorweight  = 0.0;
	fore->fg->cvpriorweight  = 1.0;
	fore->fg->cxyweight  = 0.5;
	bsColorBlendInPriors(fore);
	fore->fg->prior_set = 1;
	break;
	
      case pfRIGHTFOOT:
      case pfLEFTFOOT:
	bsColorSetClassColor(fore,pfBROWN);
	bsColorSetClassSize(fore, XYTINY);
	bsColorSetClassPosition(fore, (float) width/2., (float) height/2.);
	fore->fg->xympriorweight = 0.0;
	fore->fg->xyvpriorweight = 1.0;
	fore->fg->cmpriorweight  = 0.0;
	fore->fg->cvpriorweight  = 0.0;
	fore->fg->cxyweight  = CXYWEIGHT;
	bsColorBlendInPriors(fore);
	fore->fg->prior_set = 1;
	break;

      default:
	bsColorSetClassColor(fore,pfZERO); 
	bsColorSetClassSize(fore, XYMEDIUM);
	bsColorSetClassPosition(fore, (float) width/2., (float) height/2.);
	fore->fg->xympriorweight = 0.0;
	fore->fg->xyvpriorweight = 0.5;
	fore->fg->cmpriorweight  = 0.5;
	fore->fg->cvpriorweight  = 0.5;
	fore->fg->cxyweight  = 1.0;
	bsColorBlendInPriors(fore);
	fore->fg->prior_set = 0;
	break;
      }
    }
}

int
bsClassifyInitForegroundPriors( int maxfore )
{
  int i;
  for(i=0;i<maxfore;i++)
    {
      bsClassifyInitForegroundPrior( i );
    }
}

int
bsClassifyUnsetForegroundPriors( )
{
  int i;
  for(i=1;i<bsClassifyState->num_foreground_classes;i++)
    {
      bsClassifyState->foreground_classes[i]->fg->prior_set = 0;
    }
}  

int
bsClassifySetForegroundPriors( )
{
  int i;
  for(i=1;i<bsClassifyState->num_foreground_classes;i++)
    {
      color_class *this = bsClassifyState->foreground_classes[i];
      this->fg->prior_set = 1;
      this->fg->prior_singular = this->singular;
      memcpy(this->fg->prior_mean,     this->mean,     3*sizeof(ftype));
      memcpy(this->fg->prior_norm_mean,this->norm_mean,2*sizeof(ftype));
      memcpy(this->fg->prior_xy_mean,  this->fg->xy_mean,  2*sizeof(ftype));
      
      memcpy(this->fg->prior_inv_covar->data[0], 
	     this->inv_covar->data[0], 
	     (this->inv_covar->width*
	      this->inv_covar->height*
	      sizeof(Real)));
      memcpy(this->fg->prior_norm_inv_covar->data[0], 
	     this->norm_inv_covar->data[0], 
	     (this->norm_inv_covar->width*
	      this->norm_inv_covar->height*
	      sizeof(Real)));
      memcpy(this->fg->prior_xy_inv_covar->data[0], 
	     this->fg->xy_inv_covar->data[0], 
	     (this->fg->xy_inv_covar->width*
	      this->fg->xy_inv_covar->height*
	      sizeof(Real)));

      this->fg->prior_ln_det_covar = this->ln_det_covar;
      this->fg->prior_norm_ln_det_covar= this->ln_det_covar;
      this->fg->prior_xy_ln_det_covar = this->ln_det_covar;
    }
}

void bsClassifyUpdateBG( iveImage *im, iveBitmap *Map )
{
  static int base = 0;
  register color_class **back_class = bsClassifyState->background_classes;
  register int width = im->width;
  register int total = width * im->height; 
  register int chunk = total * UPDATEBG;
  register int i;

  for (i=0; i<chunk; i++)
    {
      register int index = (base + i) % total;
      
      if(!iveBitmapGetBit(Map, index/width, index%width))
	{
	  ivePixel ivep;
	  iveImageGetPixelI(im, index, &ivep);
	  
	  bsColorUpdateBackgroundPixel(&ivep, back_class[index]); 
	}
    }
  
  base = (base + chunk) % total;
}
  
  

PixelClassStruct *bsClassifyPixel( iveImage *im, int x, int y ) 
{
  static PixelClassStruct pix;
  ftype color[3];
  int c;
  color_class **fore_classes = bsClassifyState->foreground_classes;
  color_class *back_class;
  ftype fore_color_dist, fore_xy_dist, fore_dist, max_fore_dist = 1.0;
  ftype back_color_dist, back_dist;
  ivePixel ivep;
  int index = idx(x,y,im->width);
  
  iveImageGetPixel(im,y,x,&ivep);
  
  color[0] = ivePixelChannel( &ivep, 0 );
  color[1] = ivePixelChannel( &ivep, 1 );
  color[2] = ivePixelChannel( &ivep, 2 );
  
  back_class = bsClassifyState->background_classes[index];
  
  back_color_dist = bsColorDiscrim( color, back_class, BG_NORMALIZE);
  
  pix.dist = back_color_dist;
  pix.c = (pix.dist < -(bsSegmentTHRESH));

  if (!FOREGROUND) 
    {
      return (&pix);
    }
  
#define XYFACTOR -4.6 /* = Log(0.01) */

  back_dist = ((0.5 * back_color_dist) + 
	       (0.5 *  XYFACTOR));
  pix.dist = back_dist;

  if (! (fore_classes[pfGLOBAL])->singular) /* is it foreground? */
    {
      pix.c = 0;

      fore_color_dist = bsColorDiscrim(color,fore_classes[pfGLOBAL],
					 BG_NORMALIZE);
      fore_xy_dist = bsXYDiscrim(x,y,fore_classes[pfGLOBAL]);
      

      fore_dist = ((0.5 * fore_color_dist) + 
		   (0.5 * fore_xy_dist));
      
      if (back_color_dist < fore_color_dist)
	{
	  pix.c = 1;
	  max_fore_dist = -100000.0;
	  
	  for (c = pfRIGHTHAND;
	       c < bsClassifyState->num_foreground_classes; 
	       c++)
	    {
	      if ((!fore_classes[c]->singular) && 
		  bsXYBoundingBox(x,y,fore_classes[c]))
		{
		  float cxyweight = fore_classes[c]->fg->cxyweight;
		  fore_color_dist = bsColorDiscrim(color,fore_classes[c],
						   NORMALIZE);
		  fore_xy_dist = bsXYDiscrim(x,y,fore_classes[c]);
		  
		  fore_dist = ((cxyweight * fore_color_dist) + 
			       ((1.0 - cxyweight) * fore_xy_dist));
		  
		  
		  if ((fore_dist > -(bsClassifyTHRESH) &&
		       fore_dist > max_fore_dist))
		    {	
		      pix.c = c+1;
		      max_fore_dist = fore_dist;
		    }
		}
	    }
	}      
    }

  bsColorAddExample(pix.c,
		    x, y,
		    &ivep,
		    fore_classes,
		    bsClassifyState->num_foreground_classes);

  return (&pix);
}


