
/* bsColor.c */

/* RCS: $Header: /mas/vision/projects/pfinder/RCS/bsColor.c,v 6.8 1995/09/24 00:49:55 cwren Exp $ */

#include "bsColor.h"

static int verbose = 0;

static int NORMALIZE=1;

static float NORMALFLOOR=0.6;
static float BGUPDATERATE=0.1;

#include "eigen.h"

float bsColorGetUpdateBG() {
  return BGUPDATERATE;
}

void bsColorSetUpdateBG(float newvalue) {
  BGUPDATERATE = newvalue;
}

float bsColorGetNormalFloor() {
  return NORMALFLOOR;
}

void bsColorSetNormalFloor(float newvalue) {
  NORMALFLOOR = newvalue;
}

bsColorGetNormalize() {
	return NORMALIZE;
}
void bsColorSetNormalize(int newvalue) {
	NORMALIZE = newvalue;
}


void
bsColorSetVerbose( int val )
{
  verbose = val;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * printMatrix
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void
printMatrix(Matrix A)
{
  int c1, c2;
  
  fprintf(stderr,"\nNew Matrix:\n");
  for (c1 = 0; c1 < A->height; c1++)
    {
      fprintf(stderr,"\n");
      for (c2 = 0; c2 < A->width; c2++)
	fprintf(stderr,"%8.5f ",A->data[c1][c2]);
    }
  
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorDiscrim
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

ftype
bsColorDiscrim(ftype *rawcolor, color_class *cclass, int norm_type)
{
  extern int NORMALIZE;  
  ftype dist[3], tmp_t[3];
  register ftype a, result, ch0;
  register ftype *meanp = cclass->mean;
  register ftype **matp;
  ftype norm;
  ftype color[3];

#ifdef DIGITIZEYUV
	  a = rawcolor[0]/(meanp[0]);
#else
	  a = (rawcolor[0]+rawcolor[1]+rawcolor[2]) /
	    (meanp[0]+meanp[1]+meanp[2]);
#endif
  
  switch(norm_type)
    {
    case BG_NORMALIZE:
      if (!NORMALIZE || a<NORMALFLOOR || a>=1.0) 
	break;
      
    case ALL_NORMALIZE:
      
      norm = rawcolor[0];
      color[0] = rawcolor[1] / norm;
      color[1] = rawcolor[2] / norm;

      meanp = cclass->norm_mean;

      diff2(color,meanp,dist);
      matp = cclass->norm_inv_covar->data;
      diag_transpose_mul_mat2(dist,matp,tmp_t);
      result = - dot2(tmp_t,dist) - cclass->norm_ln_det_covar;
      return result;

    case NO_NORMALIZE:
      break;
      
    }

  diff3(rawcolor,meanp,dist);
  matp = cclass->inv_covar->data;
  diag_transpose_mul_mat3(dist,matp,tmp_t);
  result = - dot3(tmp_t,dist) - cclass->ln_det_covar;
  
  return result;
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXYDiscrim
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

ftype
bsXYDiscrim(int x, int y, color_class *cclass)
{
  ftype result;

  if (!cclass->fgp) {
    fprintf(stderr,"bsXYDiscrim called on a background pixel!!!\n");
    result = 0.0;
  } else {
    ftype data[2], dist[2], tmp_t[2];
    register ftype *meanp = cclass->fg->xy_mean;
    register ftype **matp = cclass->fg->xy_inv_covar->data;
    
    data[0] = x; data[1] = y;
    
    diff2(data,meanp,dist);
    transpose_mul_mat2(dist,matp,tmp_t);
    result = - dot2(tmp_t,dist) - cclass->fg->xy_ln_det_covar;
  }
  return result;
}

int bsXYBoundingBox(int x, int y, color_class *cclass)
{
  int result;

  if (!cclass->fgp) {
    fprintf(stderr,"bsXYBoundingBox called on a background pixel!!!\n");
    result = 0;
  } else {
    ftype data[2];
    ftype dist[2];
    int result;
    register ftype *mean   = cclass->fg->xy_mean;
    register ftype *bb   = cclass->fg->bb;
    register ftype **covar = cclass->fg->xy_inv_covar->data;
    
    data[0] = (ftype) x;
    data[1] = (ftype) y;
    
    dist[0] = bsABS(data[0] - mean[0]); 
    dist[1] = bsABS(data[1] - mean[1]);
  
    result = ((dist[0] < (5.0 * bb[0])) &&
	      (dist[1] < (5.0 * bb[1])));
  }

  return (result);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorComputeForegroundClass
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void updatepixel(ivePixel *ivepixel,
			color_class **fore,
			int num_classes,
			int class, 
			int x, int y)
{
  register int c,k,n;
  register ftype p, p0;
  ftype xy[2];
  
  xy[0] = x;
  xy[1] = y;
  
  fore[class]->fg->num_points += 1;

  for (c = 0; c < 3; c++) {
    p = ivePixelChannel( ivepixel, c );
    fore[class]->fg->mean_acc[c] += p;
  }

  p0 = ivePixelChannel( ivepixel, 0 );
  for (c = 0; c < 2; c++) {
    p = ivePixelChannel(ivepixel,c+1)/p0;
    fore[class]->fg->norm_mean_acc[c] += p;
  }

  for (c = 0; c < 3; c++) {
    p = ivePixelChannel( ivepixel, c );
    fore[class]->fg->covar_acc->data[c][c] += p*p;
  }

  p0 = ivePixelChannel( ivepixel, 0 );
  for (c = 0; c < 2; c++) {
    p = ivePixelChannel(ivepixel,c+1)/p0;
    fore[class]->fg->norm_covar_acc->data[c][c] += p*p;
  }

  for (c = 0; c < 2; c++) {
    fore[class]->fg->xy_mean_acc[c] += xy[c];
  }

  for (c = 0; c < 2; c++) {
    p = xy[c];
    for (k = 0; k < 2; k++) {
      fore[class]->fg->xy_covar_acc->data[c][k] += p * xy[k];
    }
  }
}

void bsColorInitClasses(color_class **fore,
			int num_classes)
{
  int class;
  int c,k;

  /* clear mean and covar for each class */
  
  for (class=0; class<num_classes; class++)
    {
      fore[class]->fg->last_num_points=fore[class]->fg->num_points;
      fore[class]->fg->num_points=0;

      for (c = 0; c < 3; c++) {
	fore[class]->fg->mean_acc[c] = 0.0;
      }
      
      for (c = 0; c < 2; c++) {
	fore[class]->fg->norm_mean_acc[c] = 0.0;
      }
      
      for (c = 0; c < 3; c++) {
	fore[class]->fg->covar_acc->data[c][c] = 0.0;
      }
      for (c = 0; c < 2; c++) {
	fore[class]->fg->norm_covar_acc->data[c][c] = 0.0;
      }

      for (c = 0; c < 2; c++) {
	fore[class]->fg->xy_mean_acc[c] = 0.0;
      }
  
      for (c = 0; c < 2; c++) {
	for (k = 0; k < 2; k++) {
	  fore[class]->fg->xy_covar_acc->data[c][k] = 0.0;
	}
      }
    }
}


int bsColorAddExample(int classnum,
		      int xpos,
		      int ypos,
		      ivePixel *ivep,
		      color_class **fore,
		      int num_classes)
{
  /* for each pixel... add it to the appropriate class */
  /* if it's foreground.  always add to class 1 */
	
  switch (classnum)
    {
    case 0:
      return;
      break;

    case 1:
      updatepixel(ivep,fore,num_classes,0,xpos,ypos);
      break;
      
    default:
      updatepixel(ivep,fore,num_classes,classnum-1,xpos,ypos);
      updatepixel(ivep,fore,num_classes,0,xpos,ypos);
      break;
    }
}
  

void bsColorBlendInPriors(color_class *fore)
{
  int c,k;
  register int   copy = (fore->singular); 
  register ftype cmpriorweight = (copy ? 1.0 : fore->fg->cmpriorweight);
  register ftype cvpriorweight = (copy ? 1.0 : fore->fg->cvpriorweight);
  register ftype xympriorweight= (copy ? 1.0 : fore->fg->xympriorweight);
  register ftype xyvpriorweight= (copy ? 1.0 : fore->fg->xyvpriorweight);
  register ftype cmdataweight  = 1.0 - cmpriorweight;
  register ftype cvdataweight  = 1.0 - cvpriorweight;
  register ftype xymdataweight = 1.0 - xympriorweight;
  register ftype xyvdataweight = 1.0 - xyvpriorweight;

  fore->singular = 0;

  for (c = 0; c < 3; c++) {
    fore->mean[c] =
      (fore->mean[c] * cmdataweight +
       fore->fg->prior_mean[c] * cmpriorweight);
  }

  for (c = 0; c < 2; c++) {
    fore->norm_mean[c] =
      (fore->norm_mean[c] * cmdataweight +
       fore->fg->prior_norm_mean[c] * cmpriorweight);
  }
  for (c = 0; c < 2; c++) {
    fore->fg->xy_mean[c] =
      (fore->fg->xy_mean[c] * xymdataweight +
       fore->fg->prior_xy_mean[c] * xympriorweight);
  }
  for (c = 0; c < 3; c++) {
    fore->inv_covar->data[c][c] =
      ((fore->inv_covar->data[c][c] * 
	cvdataweight) +
       (fore->fg->prior_inv_covar->data[c][c] * 
	cvpriorweight));
  }
  for (c = 0; c < 2; c++) {
    fore->norm_inv_covar->data[c][c] =
      ((fore->norm_inv_covar->data[c][c] * 
	cvdataweight) +
       (fore->fg->prior_norm_inv_covar->data[c][c] * 
	cvpriorweight));
  }
  for (c = 0; c < 2; c++) {
    for (k = 0; k < 2; k++) {
      fore->fg->xy_inv_covar->data[c][k] =
	((fore->fg->xy_inv_covar->data[c][k] * 
	  xyvdataweight) +
	 (fore->fg->prior_xy_inv_covar->data[c][k] * 
	  xyvpriorweight));
    }
  }
  
  
  fore->ln_det_covar = ((fore->ln_det_covar * 
				cvdataweight) +
			       (fore->fg->prior_ln_det_covar * 
				cvpriorweight));
  
  fore->norm_ln_det_covar = ((fore->norm_ln_det_covar * 
				     cvdataweight) +
				    (fore->fg->prior_norm_ln_det_covar * 
				     cvpriorweight));
  
  fore->fg->xy_ln_det_covar = ((fore->fg->xy_ln_det_covar * 
				  xyvdataweight) +
				 (fore->fg->prior_xy_ln_det_covar * 
				  xyvpriorweight));
  
  fore->fg->bb[0] = ((fore->fg->bb[0]  * 
			xyvdataweight) +
		       (fore->fg->prior_bb[0]  * 
			xyvpriorweight));
  fore->fg->bb[1] = ((fore->fg->bb[1] * 
			xyvdataweight) +
		       (fore->fg->prior_bb[1] * 
			xyvpriorweight));
  
}

void diag_MatrixInvert(Matrix A)
{
  register int i;
  
  for (i=0;i<A->height;i++)
    A->data[i][i] = 1.0 / A->data[i][i];
}

static ftype diag_MatrixDeterminant(Matrix A)
{
  register ftype result =1.0;
  register int i;
  
  for (i=0;i<A->height;i++)
    result *= A->data[i][i];

  return result;
}


  
int
bsColorComputeForegroundClasses( iveBitmap *inMap,
				iveImage *frame,
				color_class **fore,
				int num_classes)
{
  static ivePixel ivepixel;
  static ftype *num = NULL;
  int i,j,c,k,pixel,n;
  register ftype p, p0, xy[2];
  int class;
  
  int m = frame -> pixelCount;
  int w = frame -> width;
  
  /* compute Sum(x), Sum(x^T x) */
  
  /* calculate stats for each class */
  
  for (class=0; class<num_classes; class++)
    {
      /* compute E[x], E[x^T x] */
      
      if (fore[class]->fg->num_points > 3 )
	{
	  for (c = 0; c < 3; c++) {
	    fore[class]->mean[c] = 
	      fore[class]->fg->mean_acc[c] / 
		fore[class]->fg->num_points;

	  }
	  for (c = 0; c < 2; c++) {
	    fore[class]->norm_mean[c] = 
	      fore[class]->fg->norm_mean_acc[c] / 
		fore[class]->fg->num_points;
	  }

	  for (c = 0; c < 3; c++) {
	    fore[class]->inv_covar->data[c][c] =
	      fore[class]->fg->covar_acc->data[c][c] /
		fore[class]->fg->num_points;
	  }   
	  for (c = 0; c < 2; c++) {
	    fore[class]->norm_inv_covar->data[c][c] = 
	      fore[class]->fg->norm_covar_acc->data[c][c] / 
		fore[class]->fg->num_points;
	  }

	  for (c = 0; c < 2; c++) {
	    fore[class]->fg->xy_mean[c] = 
	      fore[class]->fg->xy_mean_acc[c] / 
		fore[class]->fg->num_points;
	  }

	  for (c = 0; c < 2; c++) {
	    for (k = 0; k < 2; k++) {
	      fore[class]->fg->xy_inv_covar->data[c][k] = 
		fore[class]->fg->xy_covar_acc->data[c][k] /
		  fore[class]->fg->num_points;
	    }
	  }
	  
	  
	  /* compute E[(x-m)^T (x-m)] = E[x^T x] - m m^T] */
	  
	  for (c = 0; c < 3; c++) {
	    p = fore[class]->mean[c];
	    fore[class]->inv_covar->data[c][c] -= p*p;
	  }
	  for (c = 0; c < 2; c++) {
	    p = fore[class]->norm_mean[c];
	    fore[class]->norm_inv_covar->data[c][c] -= p*p;
	  }

	  for (c = 0; c < 2; c++) {
	    for (k = 0; k < 2; k++) {
	      fore[class]->fg->xy_inv_covar->data[c][k] 
		-= fore[class]->fg->xy_mean[c] * 
		  fore[class]->fg->xy_mean[k];
	    }
	  }
	  
	  for (c = 0; c < 2; c++) 
	    {
	      fore[class]->fg->bb[c] = fore[class]->fg->xy_inv_covar->data[c][c];
	      fore[class]->fg->bb[c] = bsSQRT(fore[class]->fg->bb[c]); 
	    }

	  /* compute inverse and log determinant of E[(x-m)^T (x-m)] */
	  /* compute ln of det later because you want that value now */
	  
	  fore[class]->ln_det_covar = 
	    diag_MatrixDeterminant(fore[class]->inv_covar);
	  fore[class]->norm_ln_det_covar = 
	    diag_MatrixDeterminant(fore[class]->norm_inv_covar);
	  
	  fore[class]->fg->xy_ln_det_covar = 
	    MatrixDeterminant(fore[class]->fg->xy_inv_covar);
	  
	  
	  /* compute principle axes of blob from covariance matrix */

	  eigen(fore[class]->fg->xy_inv_covar->data,
		fore[class]->fg->xyL,
		fore[class]->fg->xyS);
	  fore[class]->fg->xyL[0] = sqrt(fore[class]->fg->xyL[0]);
	  fore[class]->fg->xyL[1] = sqrt(fore[class]->fg->xyL[1]);

	  /* Deal with Non-invertible matrices */
	  
	  fore[class]->singular = 1;
	  
	  if (fore[class]->ln_det_covar != 0.0) 
	    {
	      diag_MatrixInvert(fore[class]->inv_covar);
	      if (fore[class]->norm_ln_det_covar != 0.0) 
		{
		  diag_MatrixInvert(fore[class]->norm_inv_covar);
		  fore[class]->ln_det_covar = 
		    bsLOG(bsABS(fore[class]->ln_det_covar));
		  fore[class]->norm_ln_det_covar = 
		    bsLOG(bsABS(fore[class]->norm_ln_det_covar));
		  
		  fore[class]->singular = 0;
		}
	    }
	  
	  
	  if (fore[class]->fg->xy_ln_det_covar != 0.0) 
	    {
	      MatrixInvert(fore[class]->fg->xy_inv_covar);
	      fore[class]->fg->xy_ln_det_covar = 
		bsLOG(bsABS(fore[class]->fg->xy_ln_det_covar));
	      
	      fore[class]->singular |= 0;
	    }
	}
      else
	{
	  fore[class]->singular = 1;
	}	  
      
      if (fore[class]->fg->prior_set)  
	{
	  bsColorBlendInPriors(fore[class]);
	}
    }
  return 1;
}


static int bsColorPrintClass(color_class **fore, int class)
{
      printf("class %d is marked %s\n",class,
	     (fore[class]->singular ? "singular" : "valid"));
      printf("\t color mean %f %f %f\n",
	     (float) fore[class]->mean[0],
	     (float) fore[class]->mean[1],
	     (float) fore[class]->mean[2]);
      printf("\t color covar %g %g %g\n",
	     (float) fore[class]->inv_covar->data[0][0],
	     (float) fore[class]->inv_covar->data[0][1],
	     (float) fore[class]->inv_covar->data[0][2]);
      printf("\t             %g %g %g\n",
	     (float) fore[class]->inv_covar->data[1][0],
	     (float) fore[class]->inv_covar->data[1][1],
	     (float) fore[class]->inv_covar->data[1][2]);
      printf("\t             %g %g %g\n",
	     (float) fore[class]->inv_covar->data[2][0],
	     (float) fore[class]->inv_covar->data[2][1],
	     (float) fore[class]->inv_covar->data[2][2]);
      printf("\t ln_det_covar %f\n",
	     (float) fore[class]->ln_det_covar);
      printf("\t norm2 mean %f %f\n",
	     (float) fore[class]->norm_mean[0],
	     (float) fore[class]->norm_mean[1]);
      printf("\t norm2 covar %g %g\n",
	     (float) fore[class]->norm_inv_covar->data[0][0],
	     (float) fore[class]->norm_inv_covar->data[0][1]);
      printf("\t             %g %g\n",
	     (float) fore[class]->norm_inv_covar->data[1][0],
	     (float) fore[class]->norm_inv_covar->data[1][1]);
      printf("\t norm_ln_det %f\n",
	     (float) fore[class]->norm_ln_det_covar);
      printf("\t xy    mean %f %f\n",
	     (float) fore[class]->fg->xy_mean[0],
	     (float) fore[class]->fg->xy_mean[1]);
      printf("\t xy    covar %g %g\n",
	     (float) fore[class]->fg->xy_inv_covar->data[0][0],
	     (float) fore[class]->fg->xy_inv_covar->data[0][1]);
      printf("\t             %g %g\n\n",
	     (float) fore[class]->fg->xy_inv_covar->data[1][0],
	     (float) fore[class]->fg->xy_inv_covar->data[1][1]);
      printf("\t xy_ln_det %f\n",
	     (float) fore[class]->fg->xy_ln_det_covar);
}	     

int bsColorSetClassColor(color_class *fore, pfColors pfc)
{
  switch(pfc){

  case pfFLESH:
    fore->fg->prior_mean[0] = 167.3392;
    fore->fg->prior_mean[1] = 125.0676;
    fore->fg->prior_mean[2] = 116.2794;
    fore->fg->prior_inv_covar->data[0][0] =  0.0046;
    fore->fg->prior_inv_covar->data[0][1] = -0.0035;
    fore->fg->prior_inv_covar->data[0][2] = -0.0022;
    fore->fg->prior_inv_covar->data[1][0] = -0.0035;
    fore->fg->prior_inv_covar->data[1][1] =  0.1271;
    fore->fg->prior_inv_covar->data[1][2] = -0.1269;
    fore->fg->prior_inv_covar->data[2][0] = -0.0022;
    fore->fg->prior_inv_covar->data[2][1] = -0.1269;
    fore->fg->prior_inv_covar->data[2][2] =  0.1350;
    fore->fg->prior_ln_det_covar = 14.2607;
    
    fore->fg->prior_norm_mean[0] = 0.7482;
    fore->fg->prior_norm_mean[1] = 0.6944;
    fore->fg->prior_norm_inv_covar->data[0][0] =  3.3721e+03;
    fore->fg->prior_norm_inv_covar->data[0][1] = -3.3068e+03;
    fore->fg->prior_norm_inv_covar->data[1][0] = -3.3068e+03;
    fore->fg->prior_norm_inv_covar->data[1][1] =  3.5530e+03;
    fore->fg->prior_norm_ln_det_covar = -13.8604;
    break;
    
  case pfBROWN:
    fore->fg->prior_mean[0] = 74.0000;
    fore->fg->prior_mean[1] = 66.7500;
    fore->fg->prior_mean[2] = 47.1250;

    fore->fg->prior_inv_covar->data[0][0] =  1.0798e-02;
    fore->fg->prior_inv_covar->data[0][1] = -1.2753e-03;
    fore->fg->prior_inv_covar->data[0][2] = -4.1791e-04;
    fore->fg->prior_inv_covar->data[1][0] = -1.2753e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.8555e-02;
    fore->fg->prior_inv_covar->data[1][2] = -1.7816e-05;
    fore->fg->prior_inv_covar->data[2][0] = -4.1791e-04;
    fore->fg->prior_inv_covar->data[2][1] = -1.7816e-05;
    fore->fg->prior_inv_covar->data[2][2] =  1.1935e-02;
    fore->fg->prior_ln_det_covar =  1.2953e+01;
    
    fore->fg->prior_norm_mean[0] = 8.9189e-01;
    fore->fg->prior_norm_mean[1] = 6.3514e-01;
    fore->fg->prior_norm_inv_covar->data[0][0] =  2.0000e+01;
    fore->fg->prior_norm_inv_covar->data[0][1] =  0.0000e+00;
    fore->fg->prior_norm_inv_covar->data[1][0] =  0.0000e+00;
    fore->fg->prior_norm_inv_covar->data[1][1] =  2.0000e+01;
    fore->fg->prior_norm_ln_det_covar = -2.9957e+00;
    break;
    
    
  case pfGREY:
    fore->fg->prior_mean[0] = 100.0000;
    fore->fg->prior_mean[1] = 100.0000;
    fore->fg->prior_mean[2] = 100.0000;

    fore->fg->prior_inv_covar->data[0][0] =  1.0798e-02;
    fore->fg->prior_inv_covar->data[0][1] = -1.2753e-03;
    fore->fg->prior_inv_covar->data[0][2] = -4.1791e-04;
    fore->fg->prior_inv_covar->data[1][0] = -1.2753e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.8555e-02;
    fore->fg->prior_inv_covar->data[1][2] = -1.7816e-05;
    fore->fg->prior_inv_covar->data[2][0] = -4.1791e-04;
    fore->fg->prior_inv_covar->data[2][1] = -1.7816e-05;
    fore->fg->prior_inv_covar->data[2][2] =  1.1935e-02;
    fore->fg->prior_ln_det_covar =  1.2953e+01;
    
    fore->fg->prior_norm_mean[0] = 1.0000e-00;
    fore->fg->prior_norm_mean[1] = 1.0000e-00;
    fore->fg->prior_norm_inv_covar->data[0][0] = 1.0000e+03;
    fore->fg->prior_norm_inv_covar->data[0][1] = 0.0000e+00;
    fore->fg->prior_norm_inv_covar->data[1][0] = 0.0000e+00;
    fore->fg->prior_norm_inv_covar->data[1][1] = 1.0000e+03;
    fore->fg->prior_norm_ln_det_covar = -1.7269e+00;
    break;
    
    
  case pfZERO:
    fore->fg->prior_mean[0] = 0.0;
    fore->fg->prior_mean[1] = 0.0;
    fore->fg->prior_mean[2] = 0.0;
    fore->fg->prior_inv_covar->data[0][0] = 0.0;
    fore->fg->prior_inv_covar->data[0][1] = 0.0;
    fore->fg->prior_inv_covar->data[0][2] = 0.0;
    fore->fg->prior_inv_covar->data[1][0] = 0.0;
    fore->fg->prior_inv_covar->data[1][1] = 0.0;
    fore->fg->prior_inv_covar->data[1][2] = 0.0;
    fore->fg->prior_inv_covar->data[2][0] = 0.0;
    fore->fg->prior_inv_covar->data[2][1] = 0.0;
    fore->fg->prior_inv_covar->data[2][2] = 0.0;
    fore->fg->prior_ln_det_covar =  1.0e3;

    fore->fg->prior_norm_inv_covar->data[0][0] = 0.0;
    fore->fg->prior_norm_inv_covar->data[0][1] = 0.0;
    fore->fg->prior_norm_inv_covar->data[1][0] = 0.0;
    fore->fg->prior_norm_inv_covar->data[1][1] = 0.0;
    fore->fg->prior_norm_ln_det_covar = 1.0e3;
    break;
    
  case pfRED:
    /*(r > 1.5 * b && r > 1.5 * g && r > 20.0) */
    fore->fg->prior_mean[0] = 2.4774e+02;
    fore->fg->prior_mean[1] = 6.4478e+01;
    fore->fg->prior_mean[2] = 9.9087e+01;
    fore->fg->prior_inv_covar->data[0][0] =  1.0798e-02;
    fore->fg->prior_inv_covar->data[0][1] = -1.2753e-03;
    fore->fg->prior_inv_covar->data[0][2] = -4.1791e-04;
    fore->fg->prior_inv_covar->data[1][0] = -1.2753e-03;
    fore->fg->prior_inv_covar->data[1][1] =  1.8555e-02;
    fore->fg->prior_inv_covar->data[1][2] = -1.7816e-05;
    fore->fg->prior_inv_covar->data[2][0] = -4.1791e-04;
    fore->fg->prior_inv_covar->data[2][1] = -1.7816e-05;
    fore->fg->prior_inv_covar->data[2][2] =  1.1935e-02;
    fore->fg->prior_ln_det_covar =  1.2953e+01;
    
    fore->fg->prior_norm_mean[0] = 2.6058e-01;
    fore->fg->prior_norm_mean[1] = 4.0060e-01;
    fore->fg->prior_norm_inv_covar->data[0][0] =  1.0541e+03;
    fore->fg->prior_norm_inv_covar->data[0][1] = -7.6038e+01;
    fore->fg->prior_norm_inv_covar->data[1][0] = -7.6038e+01;
    fore->fg->prior_norm_inv_covar->data[1][1] =  6.0674e+02;
    fore->fg->prior_norm_ln_det_covar = -1.3359e+01;
    break;

  case pfGREEN:
    /* (g - r > 30 && g - b > 30) */
    fore->fg->prior_mean[0] = 1.3516e+02;
    fore->fg->prior_mean[1] = 2.5492e+02;
    fore->fg->prior_mean[2] = 2.0860e+02;
    fore->fg->prior_inv_covar->data[0][0] =  9.4118e-03;
    fore->fg->prior_inv_covar->data[0][1] = -4.7922e-02;
    fore->fg->prior_inv_covar->data[0][2] = -7.9882e-03;
    fore->fg->prior_inv_covar->data[1][0] = -4.7922e-02;
    fore->fg->prior_inv_covar->data[1][1] =  1.4290e+01;
    fore->fg->prior_inv_covar->data[1][2] =  7.4731e-02;
    fore->fg->prior_inv_covar->data[2][0] = -7.9882e-03;
    fore->fg->prior_inv_covar->data[2][1] =  7.4731e-02;
    fore->fg->prior_inv_covar->data[2][2] =  9.3067e-03;
    fore->fg->prior_ln_det_covar = 8.0375e+00;
    
    fore->fg->prior_norm_mean[0] = 1.9533e+00;
    fore->fg->prior_norm_mean[1] = 1.5649e+00;
    fore->fg->prior_norm_inv_covar->data[0][0] =  1.7391e+01;
    fore->fg->prior_norm_inv_covar->data[0][1] = -4.3550e+01;
    fore->fg->prior_norm_inv_covar->data[1][0] = -4.3550e+01;
    fore->fg->prior_norm_inv_covar->data[1][1] =  1.4486e+02;
    fore->fg->prior_norm_ln_det_covar = -6.4343e+00;
    break;
  }    
}	     

int bsColorSetClassSize(color_class *fore, int flag)
{
  switch (flag) {
  case XYWIDE:
    fore->fg->prior_xy_inv_covar->data[0][0] =  3.310e-03;
    fore->fg->prior_xy_inv_covar->data[0][1] =  1.096e-03;
    fore->fg->prior_xy_inv_covar->data[1][0] =  1.096e-03;
    fore->fg->prior_xy_inv_covar->data[1][1] =  5.473e-03;
    fore->fg->prior_xy_ln_det_covar = 16.8388;
    fore->fg->bb[0] = 17.0;
    fore->fg->bb[1] = 17.0;
    break;

  case XYMEDIUM:
    fore->fg->prior_xy_inv_covar->data[0][0] =  2.50000e-01;
    fore->fg->prior_xy_inv_covar->data[0][1] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][0] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][1] =  2.50000e-01;
    fore->fg->prior_xy_ln_det_covar = 4.00000e+00;
    fore->fg->prior_bb[0] = 2.0;
    fore->fg->prior_bb[1] = 2.0;
    break;

  case XYNARROW:
    fore->fg->prior_xy_inv_covar->data[0][0] =  1.00000e+00;
    fore->fg->prior_xy_inv_covar->data[0][1] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][0] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][1] =  1.00000e+00;
    fore->fg->prior_xy_ln_det_covar = 0.00000e+00;
    fore->fg->prior_bb[0] = 1.0;
    fore->fg->prior_bb[1] = 1.0;
    break;

  case XYTINY:
    fore->fg->prior_xy_inv_covar->data[0][0] =  4.0000e+00;
    fore->fg->prior_xy_inv_covar->data[0][1] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][0] =  0.0;
    fore->fg->prior_xy_inv_covar->data[1][1] =  4.0000e+00;
    fore->fg->prior_xy_ln_det_covar = -1.3862944e+00;
    fore->fg->prior_bb[0] = 0.5;
    fore->fg->prior_bb[1] = 0.5;
    break;
  }
}	     

int bsColorSetClassPosition(color_class *fore, ftype x, ftype y)
{
  fore->fg->xy_mean[0] = x;
  fore->fg->xy_mean[1] = y;
  fore->fg->prior_xy_mean[0] = x;
  fore->fg->prior_xy_mean[1] = y;
}	     


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorComputeBackgroundClassesBatch
 *
 \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
bsColorComputeBackgroundClassesBatch( iveImage **frames,
				     int frameCount,  
				     color_class **background )

{
  static ivePixel ivepixel;
  int pixelCount = frames[0]->pixelCount;
  int w = frames[0]->width;
  int i,j,c,k,pixel,rankzero=0;
  register ftype p, p0;

  /* fprintf(stderr,"background pixelCount: %d\n",pixelCount); */

  /* clear mean and covar */

  if( verbose )
    {
      fprintf(stderr,"Computing background class now\n");
    }
  for (i=0; i<pixelCount; i++) {
    for (c=0; c<3; c++) {
      background[i]->mean[c] = 0.0;
      background[i]->inv_covar->data[c][c] = 0.0;
    }
    for (c=0; c<2; c++) {
      background[i]->norm_mean[c] = 0.0;
      background[i]->norm_inv_covar->data[c][c] = 0.0;
    }
    
  }
  
  /* compute Sum(x), Sum(x^T x) */
  
  for ( j = 0; j < frameCount; j++ ) {
    for (i=0; i<pixelCount; i++) {
      iveImageGetPixelI( frames[j], i, &ivepixel );
      for ( c=0; c<3; c++ ) {
	p = ivePixelChannel( &ivepixel, c );
	background[i]->mean[c] += p;
	background[i]->inv_covar->data[c][c] += p*p;
      }
      p0 = ivePixelChannel( &ivepixel, 0 );
      for ( c=0; c<2; c++ ) {
	p = ivePixelChannel(&ivepixel,c+1)/p0;
	background[i]->norm_mean[c] += p;
	background[i]->norm_inv_covar->data[c][c] += p*p;
      }
    }
  }
  
  /* compute E[x], E[x^T x] */
  
  for (i=0; i<pixelCount; i++) {
    for (c=0; c<3; c++) {
      background[i]->mean[c] /= (ftype) frameCount;
      background[i]->inv_covar->data[c][c] /= (ftype) frameCount;
    }
    for (c=0; c<2; c++) {
      background[i]->norm_mean[c] /= (ftype) frameCount;
      background[i]->norm_inv_covar->data[c][c] /= (ftype) frameCount;
    }
  }
  
  /* compute E[(x-m)^T (x-m)] = E[x^T x] - m m^T  */
  
  for (i=0; i<pixelCount; i++) {
    for (c=0; c<3; c++) {
      p = background[i]->mean[c];
      background[i]->inv_covar->data[c][c] -= p*p;
    }
    for (c=0; c<2; c++) {
      p = background[i]->norm_mean[c];
	background[i]->norm_inv_covar->data[c][c] -= p*p;
    }
  }
  
  /* compute inverse and log determinant of E[(x-m)^T (x-m)] */
  
  for (i=0; i<pixelCount; i++) {

    background[i]->ln_det_covar = 
      diag_MatrixDeterminant(background[i]->inv_covar);
    background[i]->norm_ln_det_covar = 
      diag_MatrixDeterminant(background[i]->norm_inv_covar);

    if (background[i]->ln_det_covar == 0.0) {
      background[i]->inv_covar->data[0][0] = 4.0;
      background[i]->inv_covar->data[1][0] = 0.0;
      background[i]->inv_covar->data[2][0] = 0.0;
      background[i]->inv_covar->data[0][1] = 0.0;
      background[i]->inv_covar->data[1][1] = 4.0;
      background[i]->inv_covar->data[2][1] = 0.0;
      background[i]->inv_covar->data[0][2] = 0.0;
      background[i]->inv_covar->data[1][2] = 0.0;
      background[i]->inv_covar->data[2][2] = 4.0;
      background[i]->ln_det_covar = -4.1589;
      rankzero++; 
    } else {
      diag_MatrixInvert(background[i]->inv_covar);
      background[i]->ln_det_covar = 
	bsLOG(bsABS(background[i]->ln_det_covar));
    }


    if (background[i]->norm_ln_det_covar == 0.0) {
      background[i]->norm_inv_covar->data[0][0] =  3.2127;
      background[i]->norm_inv_covar->data[1][0] = -3.1999;
      background[i]->norm_inv_covar->data[0][1] = -3.1999;
      background[i]->norm_inv_covar->data[1][1] =  3.1875;
      background[i]->norm_ln_det_covar = 7.1390;
    } else {
      diag_MatrixInvert(background[i]->norm_inv_covar);
      background[i]->norm_ln_det_covar = 
	bsLOG(bsABS(background[i]->norm_ln_det_covar));
    }
  }
  
  if (rankzero>pixelCount/10) {
    fprintf(stderr, "%d of %d background classes are null!\n", 
	    rankzero, pixelCount);
    fprintf(stderr, "Default covariences set for these classes.\n");
  }
  if( verbose ) 
    fprintf(stderr,"Computed Background Classes\n");

}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsColorInstallChromakey
 *
 \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
bsColorInstallChromakey(int r, int g, int b,
			color_class **background,
			int width, int height)
			
{
  register ivePixel ivepixel;
  register int pixelCount = width*height;
  int i,j,c,k,pixel,rankzero=0;
  register ftype p, p0;

  ivePixelChannel( &ivepixel, 0 ) = r;
  ivePixelChannel( &ivepixel, 1 ) = g;
  ivePixelChannel( &ivepixel, 2 ) = b;

  /* fprintf(stderr,"background pixelCount: %d\n",pixelCount); */

  /* clear mean and covar */

  if( verbose )
    {
      fprintf(stderr,"Setting Chromakey background class now\n");
    }

  for (i=0; i<pixelCount; i++) {

    background[i]->mean[0] = (float) r;
    background[i]->mean[1] = (float) g;
    background[i]->mean[2] = (float) b;

    if (r == 0)
      {
	background[i]->norm_mean[0] = (float) g;
	background[i]->norm_mean[1] = (float) b;
      }
    else 
      {
	background[i]->norm_mean[0] = (float) g / (float) r;
	background[i]->norm_mean[1] = (float) b / (float) r;
      }

    background[i]->inv_covar->data[0][0] = 4.0;
    background[i]->inv_covar->data[1][0] = 0.0;
    background[i]->inv_covar->data[2][0] = 0.0;
    background[i]->inv_covar->data[0][1] = 0.0;
    background[i]->inv_covar->data[1][1] = 4.0;
    background[i]->inv_covar->data[2][1] = 0.0;
    background[i]->inv_covar->data[0][2] = 0.0;
    background[i]->inv_covar->data[1][2] = 0.0;
    background[i]->inv_covar->data[2][2] = 4.0;
    background[i]->ln_det_covar = -4.1589;
    
    background[i]->norm_inv_covar->data[0][0] =  3.2127;
    background[i]->norm_inv_covar->data[1][0] = -3.1999;
    background[i]->norm_inv_covar->data[0][1] = -3.1999;
    background[i]->norm_inv_covar->data[1][1] =  3.1875;
    background[i]->norm_ln_det_covar = 7.1390;
  }
  
  if( verbose ) 
    fprintf(stderr,"Computed Background Classes\n");

}

void
bsColorUpdateBackgroundPixel(ivePixel *ivep, color_class *bg)
{
  register int c;
  register ftype p,p0;
  
  for ( c=0; c<3; c++ ) 
    {
      p = ivePixelChannel( ivep, c );
      bg->mean[c] = ((   BGUPDATERATE    * p) + 
		     ((1.0-BGUPDATERATE) * bg->mean[c]));
    }

  p0 = ivePixelChannel( ivep, 0 );
  for ( c=0; c<2; c++ ) 
    {
      p = ivePixelChannel(ivep,c+1)/p0;
      bg->norm_mean[c] = ((  BGUPDATERATE     * p) + 
			  ((1.0-BGUPDATERATE) * bg->norm_mean[c]));
    }
}
