/* bsSegment.c */

/* RCS: $Header: /mas/vision/projects/pfinder/RCS/bsRtrackGrow.c,v 1.1 1995/09/16 22:35:12 trevor Exp $ */

#include "bsSegment.h"
#include "bsSegmentP.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * GLOBAL VARIABLES
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static ZoomGrowFunction zoomfunction = Init; 

static int SMOOTH 	  = 1;
static int ZOOM_GROW      = 1;
static int BICHSEL        = 0;
static int CHAIN_ITERATE  = 0;
static int DOUBLE_ITERATE = 0;
float bsSegmentTHRESH  = 150.0;


float region_centroid[2] = { -1, -1};
float region_major_axis[2] = { -1, -1};
float region_minor_axis[2] = { -1, -1};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * GLOBAL FUNCTIONS
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

float bsSegmentGetThresh()
{
  return (bsSegmentTHRESH);
}

float bsSegmentGetThreshMax()
{
  return (300.0);
}

void bsSegmentSetThresh(float newvalue)
{
  bsSegmentTHRESH = newvalue;
}


int bsSegmentGetZoomGrow()
{
  return(ZOOM_GROW);
}

void bsSegmentSetZoomGrow(int newvalue)
{
  ZOOM_GROW = newvalue;
  SetZoomGrowFunction();
}


int bsSegmentGetSmooth() 
{
  return SMOOTH;
}
void bsSegmentSetSmooth(int newvalue) {
  SMOOTH = newvalue;
}

int bsSegmentGetBichsel()
{
  return(BICHSEL);
}

void bsSegmentSetBichsel(int newvalue)
{
  BICHSEL = newvalue;
  SetZoomGrowFunction();
}

int bsSegmentGetDoubleIterate()
{
  return(DOUBLE_ITERATE);
}

void bsSegmentSetDoubleIterate(int newvalue)
{
  DOUBLE_ITERATE = newvalue;
  SetZoomGrowFunction();
}

int bsSegmentGetChainIterate()
{
  return(CHAIN_ITERATE);
}

void bsSegmentSetChainIterate(int newvalue)
{
  CHAIN_ITERATE = newvalue;
  SetZoomGrowFunction();
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Init
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static int *visitedbulk = NULL;
static int **visited = NULL;
static PixelClassStruct *memo = NULL;

static int Init(iveImage *im,
		 iveBitmap *outMap,
		 iveBitmap *inMap,
		 int zoom,
		 Point *centroid )
{
  static int init = 0;

  if (im == (iveImage *) NULL)
    {
      return init;
    }
  else
    {
      register int width = im->width;
      register int height = im->height;
      int i;

      init = 1;

      visitedbulk = (int *) malloc(sizeof(int)*width*height);
      bzero(visitedbulk, sizeof(int)*width*height);

      visited = (int **) malloc(sizeof(int *)*width);
      for (i=0;i<width;i++)
	{
	  visited[i] = &visitedbulk[i*height];
	}
      memo = (PixelClassStruct *) 
	malloc(sizeof(PixelClassStruct)*width*height);
      bzero(memo, sizeof(PixelClassStruct)*width*height);

      SetZoomGrowFunction();
      return ((*zoomfunction)( im, outMap, inMap, zoom, centroid));
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsSegmentFindSilhouette
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int 
bsSegmentFindSilhouette(iveImage *im,
			iveBitmap *outMap,
			iveBitmap *inMap,
			int zoom,
			Point *centroid )
{

  register int width = im->width;
  register int height = im->height;

  iveBitmapZero( outMap );

  (*zoomfunction)( im, outMap, inMap, zoom, centroid);

  bzero(memo, sizeof(PixelClassStruct)*width*height);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * LOCAL FUNCTIONS
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * SetZoomGrowFunction
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


static void SetZoomGrowFunction()
{
  if (!Init((iveImage *) NULL, (iveBitmap *) NULL,
	   (iveBitmap *) NULL, 0, (Point *) NULL))
    {
      zoomfunction = Init;
    }
  else
    {
      if (ZOOM_GROW)
	{
	    zoomfunction = Grow;
	}
      else
	{
	  zoomfunction = Raster;
	}
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * PixelClass
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* return boolean decision about pixel being n background   */
/* use neighborhood information */

static PixelClassStruct *CheckCoarserGrid(int x, int y, 
					  int zoomin, int w, int h)
{
  static PixelClassStruct pix;
  register int   zoom  = zoomin * 2;
  register int   gridx = x - (x % zoom);
  register int   gridy = y - (y % zoom);
  register int   index = w*(gridy)+gridx;
  register int  gindex = index;
  register int   zeros = 0;
  register int    ones = 0;
  
  
  if (!memo[index].valid || !memo[index].c ) 
    {
      zeros++;
    } 
  else 
    {
      gindex = index;
      ones++;
    }
  
  if (gridx+zoom < w) 
    {
      index = w*(gridy)+(gridx+zoom);
      if(!memo[index].valid || !memo[index].c) 
	{
	  zeros++;
	} 
      else
	{
	  gindex = index;
	  ones++;
	}

      if (gridy+zoom < h) 
	{
	  index = w*(gridy+zoom)+(gridx);
	  if(!memo[index].valid || !memo[index].c) 
	    {
	      zeros++;
	    } 
	  else 
	    {
	      gindex = index;
	      ones++;
	    }
	  
	  index = w*(gridy+zoom)+(gridx+zoom);
	  if(!memo[index].valid || !memo[index].c) 
	    {
	      zeros++;
	    } 
	  else
	    {
	      gindex = index;
	      ones++;
	    }
	}
    }
  
  if (zeros > ones) 
    return (NULL);

  pix.c    = memo[gindex].c;
  pix.dist = memo[gindex].dist;
  
  return (&pix);
}
  

float getonepix(iveImage *im, int zoom, int w, int h, int x, int y) 
{
  register index = w*y+x;
  
  if (x<0 || y<0 || x>=w || y>=h)
    {
      return (0);
    }

  if (!memo[index].valid)
    {
      PixelClassStruct *pix;
      
      pix = bsClassifyPixel(im, x, y);

      memcpy((void *)(&memo[index]),
	     (void *)pix,
	     sizeof(PixelClassStruct));
      memo[index].valid = 1;
    }

  return ((memo[index].dist));
}


static int PixelClass( iveImage *im,
		      int zoom,
		      int x,
		      int y )
{
  float thresh = -bsSegmentTHRESH; 
  int w = im->width;
  int h = im->height;
  int index = y*w + x;
  int ret;
  float Pneighbor;
  float PnMax, PnSum;
  float result;
  extern SMOOTH;
  
  result = getonepix(im, zoom, w, h, x, y); 
  if (SMOOTH) 
    {
      
      Pneighbor =  getonepix(im, zoom, w, h, x+zoom, y); 
      PnSum = PnMax = Pneighbor; 
      
      Pneighbor =  getonepix(im, zoom, w, h, x-zoom, y);  
      PnSum += Pneighbor;
      PnMax = (Pneighbor > PnMax ? Pneighbor : PnMax);
      
      Pneighbor =  getonepix(im, zoom, w, h, x, y+zoom); 
      PnSum += Pneighbor;
      PnMax = (Pneighbor > PnMax ? Pneighbor : PnMax);
      
      Pneighbor =  getonepix(im, zoom, w, h, x, y-zoom);  
      PnSum += Pneighbor;
      PnMax = (Pneighbor > PnMax ? Pneighbor : PnMax);
      
      if (BICHSEL) {
	result += PnMax; 
      } else {
	result = result * 0.5 + PnSum * (0.125 * 4) ;  
      }
    }
  
  if (result <= thresh)
    {
      register int c = memo[w*y+x].c;
      register int result = (c ? c : 1);
      bsXPaintBit(result,x,y);
      return (result);
    }
  else
    {
      return (0);
    }
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * PixelSimple()
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* return a boolean decision about pixel being in background */

static int
PixelSimple( iveImage *im, int x, int y)
{
  PixelClassStruct *pix = bsClassifyPixel( im, x, y) ;
  return (pix->c);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * GetEdgePoint
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int 
GetEdgePoint(iveImage *im, int *x, int *y )
{
  int i,j;
  float sum = 0.0;
  float gamma=EDGE_DECAY;
  float theta=EDGE_THRESH;
  
  int width = im->width;
  int height = im->height;
  int got_one = BSSEG_FALSE;
  int lastx;
  int zoom = im->width / 10;
  
  for(j=zoom; j < height; j+=zoom)
    {
      got_one = BSSEG_FALSE;
      sum = 0.0;
      
      for(i=0; i < width && !got_one; i++)
	{
	  sum = gamma*sum + 
	    ( PixelSimple( im, i, j )
	     ?
	     1.0
	     :
	     0.0 );
	  if (sum >= theta)
	    {
	      *x = i;
	      *y = j;
	      return BSSEG_TRUE;
	    }
	}
    }
  return BSSEG_FALSE;
}





/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Raster
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


static int Raster(iveImage *im,
		 iveBitmap *outMap,
		 iveBitmap *inMap,
		 int zoom,
		 Point *centroid )
{
  int i,j;
  static iveBitmap *doneMap = NULL;
  register int width = im->width;
  register int height = im->height;
  if ( doneMap == NULL ) 
    doneMap = iveBitmapMake( height, width );
  
  for(i=1; i<(width-1); i++) 
    for(j=1;j<(height-1);j++) 
      {
	iveBitmapPutBit( outMap, j, i, 
			PixelClass(im, zoom, i, j )); 
	
	iveBitmapSetBit( doneMap, j, i );
      }

  return (1);
}

findaxes(float C[2][2], float S[2][2]) {
      float L[2];
      static double **shadowc = NULL;
      if (shadowc == NULL) {
	    shadowc = (double **)malloc(2*sizeof(double *));
	    shadowc[0] = (double *)malloc(2*sizeof(double));
	    shadowc[1] = (double *)malloc(2*sizeof(double));
      }
      shadowc[0][0] = C[0][0]; shadowc[0][1] = C[0][1];
      shadowc[1][0] = C[1][0]; shadowc[1][1] = C[1][1];
      eigen(shadowc,L,S);

      S[0][0] = S[0][0] / sqrt(S[0][0] * S[0][0] + S[1][0]*S[1][0]);
      S[1][0] = S[1][0] / sqrt(S[0][0] * S[0][0] + S[1][0]*S[1][0]);

      S[0][1] = S[0][1] / sqrt(S[0][1] * S[0][1] + S[1][1]*S[1][1]);
      S[1][1] = S[1][1] / sqrt(S[0][1] * S[0][1] + S[1][1]*S[1][1]);

      L[0] = sqrt(L[0]);
      L[1] = sqrt(L[1]);

      S[0][0] *= L[0]; S[1][0] *= L[0];
      S[0][1] *= L[1]; S[1][1] *= L[1];
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * Grow
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void
pfMainGetBoundingBox( Point* left, Point* right )
{
  extern Point lowerLeft;
  extern Point lowerRight;
  left->x = lowerLeft.x;
  left->y = lowerLeft.y;
  right->x = lowerRight.x;
  right->y = lowerRight.y;
}


#define NHOOD 1
/* use neighborhood information and do a grow */
static int Grow(iveImage *im,
		     iveBitmap *outMap,
		     iveBitmap *inMap,
		     int zoom,
		     Point *centroid )
{
  static Point stack1[MAX_POINTS_IN_STACK];
  static Point stack2[MAX_POINTS_IN_STACK];
  static Point neighbors[8];
  register Point *burning = stack1;
  register Point *catching = stack2;
  register Point *temp;
  register int burnptr = 0;
  register int catchptr = 0;
  int burn_timer = 0;
  int done = BSSEG_FALSE;
  register int width = im->width;
  register int height = im->height;
  int oldcx, oldcy, tx, ty;
  int startx, starty;
  int found_blob = BSSEG_FALSE;
  int points = 0;
  int i;
  float growcovar[2][2];

  bzero(visitedbulk, sizeof(int)*width*height);


  /* Light the first point on fire */
  if (!(centroid->x == 0 && centroid->y == 0) && 
      (PixelClass(im, 
		  zoom,
		  centroid->x,
		  centroid->y)))
    {
      /* snap to zoom grid */
      burning[burnptr].x = centroid->x - (centroid->x % zoom);
      burning[burnptr].y = centroid->y - (centroid->y % zoom);
      burnptr++;
    }
  else
    {
      
      found_blob = GetEdgePoint(im, &startx, &starty);
      if (!found_blob)
	{
	  return (0);
	}
      
      burning[burnptr].x = startx - (startx % zoom) + zoom;
      burning[burnptr].y = starty - (starty % zoom) + zoom;
      burnptr++;
    }

  oldcx = centroid->x;
  oldcy = centroid->y;

  centroid->x = 0;
  centroid->y = 0;
  growcovar[0][0] = growcovar[0][1] = growcovar[1][0] = growcovar[1][1] = 0.0;
	
  while (!done)
    {
      
      while (burnptr--) /* burning stack loop */

	{
	  int dx, dy;
	  int x, y;
	  int count;
	  int class;
	  
	  /* Get next burning point */
	  x = burning[burnptr].x;
	  y = burning[burnptr].y;
	  count = 0;
	  /* Look at all neighbors, using subsample factor */
	  for(dx = -NHOOD; dx <= NHOOD; dx++) /* neighbors */
	    for(dy = -NHOOD; dy <= NHOOD; dy++)
	      {
		int nx = x + (dx*zoom);
		int ny = y + (dy*zoom);
		
		if (nx < 1 || ny < 1 ||
		    nx >= (width-1) || ny >= (height-1))
		  continue; /* neighbors */	
		
		if (visited[nx][ny] )
		  continue; /* neighbors */
		
		/* implicit dilation */
		iveBitmapSetBit( outMap, ny, nx );
		points++;
		tx = nx-oldcx; ty = ny-oldcy;
		centroid->x += tx;
		centroid->y += ty;
		growcovar[0][0] += tx*tx;
		growcovar[0][1] += tx*ty;
		growcovar[1][0] += ty*tx;
		growcovar[1][1] += ty*ty;

		/* If foreground pixel, add to catching */
		if (class = PixelClass(im, 
				       zoom,
				       nx,
				       ny))
		  {
		    count++;
		    catching[catchptr].x = nx;
		    catching[catchptr].y = ny;
		    catchptr++;
		    iveBitmapPutBit( outMap, ny, nx, class);
		  }
	     
	  
		visited[nx][ny] = 1;
	      } /* End neighbors loop */
	  
	} /* burning stack loop */
      
      /* swap burning and  non burning stacks */
      burnptr = catchptr;
      catchptr = 0;
      temp = catching;
      catching = burning;
      burning = temp;
      
      /* If no more burning, we are done */
      if (burnptr == 0) 
	done = BSSEG_TRUE;
      
      burn_timer++;
    } /* While (!done) */

  if(points != 0)
    {
      float S[2][2];
      float norm;

      centroid->x /= points;
      centroid->y /= points;
      
      growcovar[0][0] /= points;
      growcovar[0][1] /= points;
      growcovar[1][0] /= points;
      growcovar[1][1] /= points;

      growcovar[0][0] -= (centroid->x * centroid->x);
      growcovar[0][1] -= (centroid->x * centroid->y);
      growcovar[1][0] -= (centroid->y * centroid->x);
      growcovar[1][1] -= (centroid->y * centroid->y);

      centroid->x += oldcx;
      centroid->y += oldcy;

      findaxes(growcovar, S);

      region_centroid[0] = ((float)centroid->x)/(float)width;
      region_centroid[1] = ((float)centroid->y)/(float)height;
      region_major_axis[0] = S[0][0];
      region_major_axis[1] = S[1][0];
      region_minor_axis[0] = S[0][1];
      region_minor_axis[1] = S[1][1];
      norm = sqrt(region_major_axis[0]*region_major_axis[0] + 
		  region_major_axis[1]*region_major_axis[1]);
      region_major_axis[0] /= norm; region_major_axis[1] /= norm;
      region_minor_axis[0] /= norm; region_minor_axis[1] /= norm;

      bsXNewCanvas(0);
      bsXDrawColorLine(centroid->x, centroid->y,
		       (int) (S[0][0]), (int)(S[1][0]),
		       2, 0);
      bsXDrawColorLine(centroid->x, centroid->y,
		       (int) -(S[0][0]), (int)-(S[1][0]),
		       2, 0);
      bsXDrawColorLine(centroid->x, centroid->y,
		       (int) (S[0][1]), (int)(S[1][1]),
		       3, 0);
      bsXDrawColorLine(centroid->x, centroid->y,
		       (int) -(S[0][1]), (int)-(S[1][1]),
		       3, 0);

    }
  else
    {
      centroid->x = 0;
      centroid->y = 0;
      growcovar[0][0] = 0;
      growcovar[0][1] = 0;
      growcovar[1][0] = 0;
      growcovar[1][1] = 0;
    }

  return (1);
}


