/* backsub-Xclient.c */

#include "bsXclient.h"
#define MAX_NUM_CANVAS (12)
int num_canvas = 0;

static Widget shell;			/* shell widget */
static Widget widgetshell;		/* shell widget */
static Widget widgetmanager;		/* top level for widgets*/
static Widget editshell;		/* shell widget */
static Widget displaymanager;		/* top level for editor*/
static Widget editnext;
static XtAppContext context;
static GC gc = NULL;
static int width;
static int height;

#define maxcontour 5000	/* see pfContour.h++ */
			/* if this were C++ code instead of C code, you */
			/* could just use pfContour::maxContourLength */

typedef struct {
  Widget widget;
  Window window;
  Pixmap display;
} CanvasStruct;
  
static CanvasStruct canvas[MAX_NUM_CANVAS];	/* display windows*/
static CanvasStruct editor;				/* edit window */

static Display *display;
static int screen;
static int depth;

static double IMAGETHRESH  = 0.40;

int bsXxpos = 200;
int bsXypos = 30;
int bsXexpos = 600;
int bsXeypos = 30;


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * GLOBAL FUNCTIONS
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


double bsXGetThresh()
{
  return (IMAGETHRESH);
}

void bsXSetThresh(float newvalue)
{
 IMAGETHRESH = (double) newvalue;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * local procedure decls
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void edit_action (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);

static void edit_next   (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);

static void edit_clear  (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);

static void edit_quit   (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * callback structs
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

typedef struct {
  XtPointer addr;
  float mult;
  float off;
  char *title;
  int funcp;
} sliderCallbackStruct;

typedef struct {
  XtPointer addr;
  char *title;
  int funcp;
} buttonCallbackStruct;
 
typedef struct {
  PushBCallbackFunc addr;
  char *title;
} pushbCallbackStruct;
 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * slider_notify
 * button_notify
 * draw_notify
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void slider_notify (Widget gw, 
			   XtPointer client_data, 
			   XtPointer call_data)
{
  sliderCallbackStruct *client = (sliderCallbackStruct *) client_data;
  XmScaleCallbackStruct *call = (XmScaleCallbackStruct *) call_data;
  int value;
  float position;

  XmScaleGetValue(gw,&value);
  
  position= ((float) value)/100.0;

  if (client->funcp) 
    {
      SliderCallbackFunc callback = (SliderCallbackFunc) client->addr;
      (callback)(position);
    }
  else 
    {
      float *variable = (float *) client->addr;
      *variable = position;
    }

  if (call->reason == XmCR_VALUE_CHANGED) /* XmCR_DRAG */
    {
      fprintf (stderr, "***%s: %lf\n", client->title, position);
    }
} 

static void button_notify (Widget gw, XtPointer client_data, XtPointer call_data)
{
  buttonCallbackStruct *data = (buttonCallbackStruct *) client_data;
  int value;
  if (XmToggleButtonGetState(gw))
    {
      value = 1;
    }
  else
    {
      value = 0;
    }
  
  if (data->funcp) 
    {
      ButtonCallbackFunc callback = (ButtonCallbackFunc) data->addr;
      (callback)(value);
    }
  else 
    {
      int *variable = (int *) data->addr;
      *variable = value;
    }
  
  fprintf (stderr, "***%s: %s\n", data->title, 
	   (value? "ON" : "OFF"));
  
}
 
static void pushb_notify (Widget gw, XtPointer client_data, XtPointer call_data)
{
  pushbCallbackStruct *data = (pushbCallbackStruct *) client_data;
  
  (*data->addr)();
  fprintf (stderr, "***called: %s\n", data->title);
}
 
static void draw_notify (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data)
{
  Pixmap *p = (Pixmap*) client_data;
  XmDrawingAreaCallbackStruct *call = (XmDrawingAreaCallbackStruct *) call_data;
  XCopyArea(display, *p, call->window, gc, 0, 0, width, height, 0, 0);
  XFlush(display);
} 


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXInitialize
 * bsXHandleEvents
 *
 * bsXInitializeGC
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void
InitializeXlib()
{
  static int init = 0;

  if (!init)
    {
      init = 1;
      {				/* set basic globals */
	Arg args[10];
	int nargs;
	
	display = XtDisplay(canvas[0].widget);
	screen = DefaultScreen(display);
	
	nargs = 0;
	XtSetArg(args[nargs],XmNdepth, &depth); nargs++;
	XtGetValues(canvas[0].widget,args,nargs);
      }
      
      {				/* build GC's */
	XGCValues values;
	XtGCMask value_mask;
	
	values.background = WhitePixel(display,screen);
	values.foreground = BlackPixel(display,screen);
	values.fill_style = FillSolid;
	
	value_mask = GCBackground | GCForeground | GCFillStyle;
	
	gc = XtGetGC(canvas[0].widget,value_mask,&values);
      }
      
      {				/* build Pixmaps */
	int index;
	
	for(index=0; index<num_canvas; index++) 
	  {
	    /* get the wid of the canvas */
	    canvas[index].window = XtWindow(canvas[index].widget);
	    
	    /* make a pixmap for it */
	    canvas[index].display = XCreatePixmap(display, 
						  canvas[index].window, 
						  width, 
						  height, 
						  depth);
	    
	    /* clear it */
	    XSetForeground(display,gc,WhitePixel(display,screen));
	    XFillRectangle(display, canvas[index].display, 
			   gc, 0, 0, width, height);

	    XSetForeground(display,gc,BlackPixel(display,screen));
	    XDrawLine(display, canvas[index].display, 
		      gc, 0, 0, width, height);
	    XDrawLine(display, canvas[index].display, 
		      gc, 0, height, width, 0);

	    /* add callback */
	    XtAddCallback (canvas[index].widget, 
			   XmNexposeCallback, 
			   draw_notify, 
			   &(canvas[index].display));
	  }
      }

      XFlush(display);
      bsXHandleEvents();
    }
}

InitializeEditor()
{				/* set up editing window */
  static int init = 0;
  
  if(!init) {

    init = 1;
    /* XtPopup (editshell,XtGrabNone); */
    /* get the wid of the canvas */
    editor.window = XtWindow(editor.widget);
    
    /* make a pixmap for it */
    editor.display = XCreatePixmap(display, 
				   editor.window, 
				   width, 
				   height, 
				   depth);
    
    XtAddCallback (editor.widget, XmNexposeCallback,
		   draw_notify, &(editor.display));
    XtAddCallback (editor.widget, XmNinputCallback,
		   edit_action, (XtPointer) NULL);

  }
}

#define NUMCOLORS (11)

typedef struct {
  Pixel greys[NUMCOLORS];
  Pixel colors[NUMCOLORS];
} ApplicationResources;
static ApplicationResources appres;

void
bsXInitialize( int w, int h, int n, int argc, char **argv )
{

  Arg args[10];
  int nargs=0;
  Widget frame;
  Widget manager;
  Widget button;
  XmString xmstitle;
  static XtResource resources[] = {
    {"color0","Color0",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[0]),XtRString,"white"},
    {"color1","Color1",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[1]),XtRString,"black"},
    {"color2","Color2",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[2]),XtRString,"red"},
    {"color3","Color3",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[3]),XtRString,"green"},
    {"color4","Color4",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[4]),XtRString,"light blue"},
    {"color5","Color5",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[5]),XtRString,"orange"},
    {"color6","Color6",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[6]),XtRString,"yellow"},
    {"color7","Color7",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[7]),XtRString,"violet"},
    {"color8","Color8",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[8]),XtRString,"lightseagreen"},
    {"color9","Color9",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[9]),XtRString,"blue violet"},
    {"color10","Color10",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,colors[10]),XtRString,"brown"},
    {"grey0","Grey0",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[0]),XtRString,"black"},
    {"grey1","Grey1",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[1]),XtRString,"grey10"},
    {"grey2","Grey2",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[2]),XtRString,"grey20"},
    {"grey3","Grey3",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[3]),XtRString,"grey30"},
    {"grey4","Grey4",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[4]),XtRString,"grey40"},
    {"grey5","Grey5",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[5]),XtRString,"grey50"},
    {"grey6","Grey6",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[6]),XtRString,"grey60"},
    {"grey7","Grey7",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[7]),XtRString,"grey70"},
    {"grey8","Grey8",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[8]),XtRString,"grey80"},
    {"grey9","Grey9",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[9]),XtRString,"grey90"},
    {"grey10","Grey10",XtRPixel,sizeof(Pixel),
       XtOffsetOf(ApplicationResources,greys[10]),XtRString,"white"}
  };    
  static char *fallback_resources[] = {
    "*foreground:            grey80",
    "*background:            black",
    "*font:                  -*-fixed-*-*-*-*-*-60-*-*-*-*-*-*",
    (char *) NULL
  };

  width = w;
  height = h;

  nargs=0;
  shell = XtAppInitialize( &context,
			  "pfinder",
			  NULL, 0,
			  &nargs, 0,
			  fallback_resources,
			  0, nargs );

  nargs=0;
  XtGetApplicationResources(shell,
			    &appres,
			    resources,
			    XtNumber(resources),
			    args, nargs);
			    
  nargs=0;
  widgetshell = XtCreatePopupShell ("controls", 
				    shellWidgetClass,
				    shell,
				    args,nargs);
  
  
  nargs=0;
  XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
  XtSetArg(args[nargs],XmNadjustLast,False); nargs++;
  XtSetArg(args[nargs],XmNnumColumns,2); nargs++;
  XtSetArg(args[nargs],XmNpacking,XmPACK_COLUMN); nargs++;
  widgetmanager = XtCreateManagedWidget ("widgetpane", 
					 xmRowColumnWidgetClass, 
					 widgetshell, 
					 args,nargs);
  nargs=0;
  editshell = XtCreatePopupShell ("classeditor", 
				    shellWidgetClass,
				    shell,
				    args,nargs);
  
  nargs=0;
  XtSetArg(args[nargs],XmNsashWidth, 1); nargs++;
  XtSetArg(args[nargs],XmNsashHeight, 1); nargs++;
  XtSetArg(args[nargs],XmNseparatorOn, False); nargs++;
  manager = XtCreateManagedWidget ("widgetpane", 
				   xmPanedWindowWidgetClass, 
				   editshell, 
				   args,nargs);
  
  
  frame = XtCreateManagedWidget ("frame",
				 xmFrameWidgetClass,
				 manager,
				 args,nargs);

  nargs=0;
  XtSetArg(args[nargs],XmNwidth, width); nargs++;
  XtSetArg(args[nargs],XmNheight, height); nargs++;
  XtSetArg(args[nargs],XmNresizePolicy, XmRESIZE_NONE); nargs++;
  editor.widget = XtCreateManagedWidget ("edit",
					 xmDrawingAreaWidgetClass,
					 frame,
					 args,nargs);
  
  xmstitle = XmStringCreateSimple("Next Class");
  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNlabelString, xmstitle); nargs++;
  editnext = XtCreateManagedWidget ("next", 
				  xmPushButtonWidgetClass,
				  manager, 
				  args,nargs);
  
  XmStringFree(xmstitle);
  XtAddCallback (editnext, XmNactivateCallback, edit_next, NULL);

  xmstitle = XmStringCreateSimple("Clear Class");
  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNlabelString, xmstitle); nargs++;
  button = XtCreateManagedWidget ("clear", 
				  xmPushButtonWidgetClass,
				  manager, 
				  args,nargs);
  
  XmStringFree(xmstitle);
  XtAddCallback (button, XmNactivateCallback, edit_clear, NULL);
  
  xmstitle = XmStringCreateSimple("Continue");
  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNlabelString, xmstitle); nargs++;
  button = XtCreateManagedWidget ("quit", 
				  xmPushButtonWidgetClass,
				  manager, 
				  args,nargs);
  
  XmStringFree(xmstitle);
  XtAddCallback (button, XmNactivateCallback, edit_quit, NULL);
  

  nargs=0;
  XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
  XtSetArg(args[nargs],XmNadjustLast,False); nargs++;
  XtSetArg(args[nargs],XmNnumColumns,1); nargs++;
  XtSetArg(args[nargs],XmNpacking,XmPACK_COLUMN); nargs++;
  displaymanager = XtCreateManagedWidget ("displaypane", 
				xmRowColumnWidgetClass, 
				shell, 
				args,nargs);
  {
    int i;
    num_canvas = 2;

    if ( num_canvas > MAX_NUM_CANVAS ) {
      num_canvas = MAX_NUM_CANVAS;
      fprintf(stderr,"too many classes, will only display %d.\n",
	      num_canvas-1);
    }

    if ( num_canvas < 1 ) num_canvas = 1;

    for(i=0;i<num_canvas;i++)
      {
	
	nargs=0;
	XtSetArg(args[nargs],XmNshadowThickness, 4); nargs++;
	XtSetArg(args[nargs],XmNshadowType, XmSHADOW_OUT); nargs++;
	XtSetArg(args[nargs],XmNskipAdjust, True); nargs++;
	frame = XtCreateManagedWidget ("frame",
				       xmFrameWidgetClass,
				       displaymanager,
				       args,nargs);
	
	nargs=0;
	XtSetArg(args[nargs],XmNwidth, width); nargs++;
	XtSetArg(args[nargs],XmNheight, height); nargs++;
	XtSetArg(args[nargs],XmNresizePolicy, XmRESIZE_NONE); nargs++;
	canvas[i].widget = XtCreateManagedWidget ("canvas",
						   xmDrawingAreaWidgetClass,
						   frame,
						   args,nargs);
      }
  }
}

void
bsXHandleEvents()
{
  static init = 0;

  if (!init)
    {
      init = 1;
      XtRealizeWidget (shell);
      XtMoveWidget( shell, bsXxpos, bsXypos );
      XtPopup (widgetshell,XtGrabNone);
      XtMoveWidget( widgetshell, bsXexpos, bsXeypos );
      /* XtPopup (editshell,XtGrabNone); */
    }
  while(XtAppPending(context))
    {
      XEvent event;
      XtAppNextEvent(context, &event);
      XtDispatchEvent(&event); 
    }
}



bsXClose()
{
  int index;
  for(index=0; index<num_canvas; index++) 
    {
      /* free pixmaps */
      XFreePixmap(display,canvas[index].display);
    }

  if(editor.display)
    XFreePixmap(display,editor.display);
  
  if(gc)
    XFreeGC(display,gc);
} 

  
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXMakeSlider
 * bsXMakeButton
 * bsXMakeSliderProc
 * bsXMakeButtonProc
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


void
bsXMakeSliderDo(XPointer addr, char *title,
		float mult, float off, float def,
		int funcp) 
{
  Widget slider;
  Arg args[10];
  int nargs=0;
  sliderCallbackStruct *scb = (sliderCallbackStruct *) 
    malloc (sizeof(sliderCallbackStruct));
  XmString xmstitle;
  scb->addr = addr;
  scb->title = strdup(title);
  scb->mult = mult;
  scb->off = off;
  scb->funcp = funcp;

  xmstitle = XmStringCreateSimple(title);

  nargs=0;
  XtSetArg(args[nargs],XmNshowValue, True); nargs++;
  XtSetArg(args[nargs],XmNtitleString, xmstitle); nargs++;
  XtSetArg(args[nargs],XmNdecimalPoints, 2); nargs++;
  XtSetArg(args[nargs],XmNorientation, XmHORIZONTAL); nargs++;
  XtSetArg(args[nargs],XmNminimum, (int) scb->off*100.0); nargs++;
  XtSetArg(args[nargs],XmNmaximum, (int) (scb->mult+scb->off)*100.0); nargs++;
  XtSetArg(args[nargs],XmNvalue,   (int) (def*100.0)); nargs++;
  XtSetArg(args[nargs],XmNscaleHeight, 20); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  slider = XtCreateManagedWidget (scb->title, 
				  xmScaleWidgetClass, 
				  widgetmanager, 
				  args,nargs);
  
  XtAddCallback (slider, XmNdragCallback, slider_notify, scb);
  XtAddCallback (slider, XmNvalueChangedCallback, slider_notify, scb);
  XmScaleSetValue(slider, (int) (def*100.0));

  XmStringFree(xmstitle);
}

void
bsXMakeSlider( float *addr, char *title,
	       float mult, float off, float def) 
{
  bsXMakeSliderDo((XPointer)addr,title,mult,off,def,0);
}
void
bsXMakeSliderProc(SliderCallbackFunc addr, char *title,
	       float mult, float off, float def) 
{
  bsXMakeSliderDo((XPointer)addr,title,mult,off,def,1);
}


static void
bsXMakeButtonDo(XPointer addr, char *title,
		int def,
		int funcp) 
{
  Widget button;
  Arg args[10];
  int nargs=0;
  buttonCallbackStruct *bcb = (buttonCallbackStruct *)
    malloc (sizeof(buttonCallbackStruct));

  bcb->addr = addr;
  bcb->title = strdup(title);
  bcb->funcp = funcp;

  nargs=0;
  XtSetArg(args[nargs],XmNindicatorType, XmN_OF_MANY); nargs++;
  XtSetArg(args[nargs],XmNset, (def ? True : False)); nargs++;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  button = XtCreateManagedWidget (bcb->title, 
				  xmToggleButtonWidgetClass, 
				  widgetmanager, 
				  args,nargs);

  XtAddCallback (button, XmNvalueChangedCallback, button_notify, bcb);
  XmScaleSetValue(button, (int) (def*100.0));
}

void
bsXMakeButton( int *addr, char *title,
	       int  def) 
{
  bsXMakeButtonDo((XPointer)addr,title,def,0);
}
void
bsXMakeButtonProc( ButtonCallbackFunc addr, char *title,
	       int def) 
{
  bsXMakeButtonDo((XPointer)addr,title,def,1);
}


 void
bsXMakePushButton(PushBCallbackFunc func, char *title)
{
  Widget button;
  Arg args[10];
  int nargs=0;
  pushbCallbackStruct *pcb = (pushbCallbackStruct *)
    malloc (sizeof(pushbCallbackStruct));
  
  pcb->addr = func;
  pcb->title = strdup(title);

  nargs=0;
  XtSetArg(args[nargs],XmNheight, 30); nargs++;
  XtSetArg(args[nargs],XmNpaneMinimum, 30); nargs++;
  button = XtCreateManagedWidget (pcb->title, 
				  xmPushButtonWidgetClass,
				  widgetmanager, 
				  args,nargs);

  XtAddCallback (button, XmNactivateCallback, pushb_notify, pcb);
}

void
bsXMakeSeparator()
{
  Widget sep;
  Arg args[10];
  int nargs=0;

  nargs=0;
  sep = XtCreateManagedWidget ("sep", 
			       xmSeparatorWidgetClass,
			       widgetmanager, 
			       args,nargs);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXDisplayBitmap
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define MAXPOINTS 19360

static XPoint points[NUMCOLORS][MAXPOINTS];
static int numpoints[NUMCOLORS];

int
bsXPaintBit(int color, int x, int y)
{
  color %= NUMCOLORS;
  points[color][numpoints[color]].x = (short) x;
  points[color][numpoints[color]].y = (short) y;
  numpoints[color] = (numpoints[color]+1) % MAXPOINTS;
}

int
bsXSync()
{
  XFlush(display);
}

int
bsXCommitPaint(int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDisplayBitmap: arg out of range: %d\n",n);
      return (-1);
    }
  
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  {
    register int i;

    for (i=0;i<NUMCOLORS;i++) {
      XSetForeground(display,gc,appres.colors[i]);
      XDrawPoints(display, target->display, gc, 
		  points[i], numpoints[i], 
		  CoordModeOrigin);
    }
    XCopyArea(display, target->display, target->window, 
	      gc, 0, 0, width, height, 0, 0);

    for (i=0;i<NUMCOLORS;i++)
      numpoints[i] = 0;
  }
}

int
bsXNewCanvas(int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDisplayBitmap: arg out of range: %d\n",n);
      return (-1);
    }
  
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  {
    register int i;

    XSetForeground(display,gc,appres.colors[0]);
    XFillRectangle(display, target->display, gc,
		   0,0,width,height);
  } 
}

int
bsXDisplayBitmap( iveBitmap *bm , int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDisplayBitmap: arg out of range: %d\n",n);
      return (-1);
    }
  
  
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  {
    int i;
    int x,y;
    XPoint points[NUMCOLORS][307200];
    int numpoints[NUMCOLORS];
    {
      register int i;
      for (i=0;i<NUMCOLORS;i++)
	numpoints[i] = 0;
    }

    for (x=0; x<width; x++)
      for (y=0; y<height; y++)
	{
	  int tot=0;
	  
	  tot = iveBitmapGetBit(bm,y,x);
	  tot = ((tot<NUMCOLORS) ? tot : NUMCOLORS-1);
	  points[tot][numpoints[tot]].x = (short) x;
	  points[tot][numpoints[tot]].y = (short) y;
	  numpoints[tot]++;
	}
    
    for (i=1;i<NUMCOLORS;i++) {
      XSetForeground(display,gc,appres.colors[i]);
      XDrawPoints(display, target->display, gc, 
		  points[i], numpoints[i], 
		  CoordModeOrigin);
    }
    XCopyArea(display, target->display, target->window, 
	      gc, 0, 0, width, height, 0, 0);

  }
}

int
bsXDrawCross( int x, int y, int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDrawCross: arg out of range: %d\n",n);
      return (-1);
    }
  
  
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  XSetForeground(display,gc,appres.greys[5]);
  XDrawLine(display, target->display, gc, 
	    x-20,y,
	    x+20,y);
  XDrawLine(display, target->display, gc, 
	    x,y-20,
	    x,y+20);
  XCopyArea(display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

  /* fprintf(stderr,"bsXDrawCross: drew it dude...\n"); */
}

int
bsXDrawColorCross( int x, int y, int size, int color, int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDrawCross: arg out of range: %d\n",n);
      return (-1);
    }
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  XSetForeground( display, gc, appres.colors[color] );
  XDrawLine( display, target->display, gc, x-size, y, x+size, y );
  XDrawLine( display, target->display, gc, x, y-size, x, y+size );
  XCopyArea( display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

  /* fprintf(stderr,"bsXDrawCross: drew it dude...\n"); */
}


int
bsXDrawColorLine( int x, int y, int dx, int dy, int color, int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDrawLine: arg out of range: %d\n",n);
      return (-1);
    }
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  XSetForeground( display, gc, appres.colors[color] );
  XDrawLine( display, target->display, gc, x, y, x+dx, y+dy );
  XCopyArea( display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

}

void bsXSetColor(int i, int r, int g, int b)
{
  XColor ccell;
  
  ccell.pixel = appres.colors[i];
  ccell.red = r;
  ccell.green = g;
  ccell.blue = b;
  ccell.flags = DoRed | DoGreen | DoBlue;
  
  /* XStoreColor(display, 
     DefaultColormap(display,DefaultScreen(display)),
     &ccell);
     */
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXDisplayContour
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int
bsXDisplayContour( int* pointArray, int nPoints, int n )
/*
  pointArray is { x(0), y(0), x(1), y(1), ..., x(nPoints-1), y(nPoints-1) }
*/
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDisplayContour: arg out of range: %d\n",n);
      return (-1);
    }
  
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  {
    int i;
    XPoint points[ maxcontour ];
    
    for (i=0; i<nPoints; i++)
      {
	points[i].x = (short) pointArray[ 2*i ];
	points[i].y = (short) pointArray[ 2*i + 1 ];
      }

    XSetForeground(display,gc,BlackPixel(display,screen));
    XFillRectangle(display, target->display, gc, 0, 0, width, height);

    XSetForeground(display,gc,WhitePixel(display,screen));
    XDrawLines(display, target->display, gc, points, nPoints, CoordModeOrigin);
  }
  
  XCopyArea(display, target->display, target->window, 
	    gc, 0, 0, width, height, 0, 0);

  XFlush(display);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXDisplayImage
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

int
bsXDisplayImage ( iveImage *image , int n)
{
  static int init = 0;
  CanvasStruct *target;
  
  if (!init) 
    {
      InitializeXlib();
      init = 1;
    }
  
  if (n>num_canvas)
    {
      fprintf(stderr,"bsXDisplayImage: arg out of range: %d\n",n);
      return (-1);
    }
  
  
  if (n<0)
    {
     target = &editor;
    }
  else
    {
      target = &(canvas[n]);
    }
  
  {
    int i;
    int x,y;
    XPoint points[NUMCOLORS][307200];
    int numpoints[NUMCOLORS];
    register double thresh;

    {
      register int i;
      for (i=0;i<NUMCOLORS;i++)
	numpoints[i] = 0;
    }

    thresh = IMAGETHRESH;

    for (x=0; x<width; x++)
      for (y=0; y<height; y++)
	{
	  ivePixel pix;
	  int tot=0;
	  
	  iveImageGetPixel(image,y,x, &pix );
#ifdef DIGITIZEYUV
	  tot = ivePixelChannel(&pix,0)  * 0.18750 * thresh;
#else
	  tot = ((ivePixelGetRed(&pix)   * 0.06250 * thresh) +
		 (ivePixelGetGreen(&pix) * 0.09375 * thresh) +
		 (ivePixelGetBlue(&pix)  * 0.03125 * thresh));
#endif
	  tot = ((tot<NUMCOLORS) ? tot : NUMCOLORS-1);
	  points[tot][numpoints[tot]].x = (short) x;
	  points[tot][numpoints[tot]].y = (short) y;
	  numpoints[tot]++;
	}
    
    for (i=0;i<NUMCOLORS;i++) {
      XSetForeground(display,gc,appres.greys[i]);
      XDrawPoints(display, target->display, gc, 
		  points[i], numpoints[i], 
		  CoordModeOrigin);
    }
    XCopyArea(display, target->display, target->window, 
	      gc, 0, 0, width, height, 0, 0);

  }
  
  XFlush(display);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * bsXEditClasses
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static editing = -1;		/* -1 not editing, else class number */
static iveImage *current_img = NULL; /* current input image */
static iveBitmap *oldMap = NULL; /* current input image */
static int max_classes = 0;	

				/* handle editing events and  */
				/* call the classification routines */

static void edit_quit(Widget gw, 
		      XtPointer client_data, 
		      XtPointer call_data)
{
  editing = -1;
}

static void edit_clear(Widget gw, 
		      XtPointer client_data, 
		      XtPointer call_data)
{
  fprintf(stderr,"clear class %d definition\n",editing+1);
  iveBitmapReplaceValue( oldMap, editing+1, 1);
}

static void edit_next(Widget gw, 
		      XtPointer client_data, 
		      XtPointer call_data)
{
  Arg args[10];
  int nargs=0;
  
  editing = (editing+1) % max_classes;
  fprintf(stderr,"now editing class %d\n",editing+1);

  nargs=0;
  XtSetArg(args[nargs],XmNbackground, 
	   appres.colors[(((editing+1)<NUMCOLORS) ? editing+1 : NUMCOLORS-1)]);
  nargs++;
  XtSetValues(editnext,args,nargs);

}

static void edit_action (Widget gw, 
			 XtPointer client_data, 
			 XtPointer call_data)

{
  static Point centroid;
  static iveBitmap *newMap = (iveBitmap *)NULL;
  int x,y;
  int decimate = 1;

  XmDrawingAreaCallbackStruct *call;
  call = (XmDrawingAreaCallbackStruct *) call_data;

  if( newMap == (iveBitmap *)NULL )
    {
      centroid.x = 0;
      centroid.y = 0;
      newMap = iveBitmapMake( height, width );
    }
  
  if (call->event->type == ButtonRelease) 
    {
      x = (int) call->event->xbutton.x;
      y = (int) call->event->xbutton.y;

      iveBitmapPutBit(oldMap, 
		      y+0,
		      x+0, 
		      editing+1);
      iveBitmapPutBit(oldMap, 
		      y+0,
		      (x+decimate>width ? width : x+decimate), 
		      editing+1);
      iveBitmapPutBit(oldMap,  
		      (y+decimate>height ? height : y+decimate),
		      x+0,
		      editing+1);
      iveBitmapPutBit(oldMap,  
		      y+0,
		      (x-decimate < 0 ? 0 : x-decimate),
		      editing+1);
      iveBitmapPutBit(oldMap,  
		      (y-decimate < 0 ? 0 : y-decimate),
		      x+0,
		      editing+1);

      bsXDisplayBitmap( oldMap , -1);
      centroid.x = x;
      centroid.y = y;
      fprintf(stderr,"added cluster at %d,%d to class %d\n",
	      x,y,editing+1);
      bsClassifySetForegroundStatsI(current_img, 
				    oldMap);
      iveBitmapZero( newMap );
      bsSegmentFindSilhouette(current_img, newMap, newMap, 
			      decimate, &centroid);;
      bsXCommitPaint(1);
    }
} 

				/* start an editing session */
				/* pops up the editor and  */
				/* kills it off when done */
				/* actual editing happes in */
				/*        edit_action */

void bsXEditClasses(iveImage *image, 
		    iveBitmap *bm,
		    int numclasses)
{
  bsClassifyUnsetForegroundPriors();

  XtPopup (editshell,XtGrabNone);

  InitializeEditor();
  
  bsXDisplayBitmap( bm , -1);
  
  editing = 0;
  max_classes = numclasses;
  edit_next(NULL,NULL,NULL);

  current_img = image;
  oldMap = bm;
  
  {
    while(editing >= 0)
      {
	XtInputMask m;
	bsXDisplayImage(current_img,-1);
	while( m = XtAppPending(context) )
	   XtAppProcessEvent( context, m );
      }
  }
  
  bsClassifySetForegroundPriors();

  XtPopdown (editshell);
}


