#include "eigen.h"

#ifdef DEBUGEIGEN
#include <stdio.h>
#endif

void eigen(double **A, float L[2], float S[2][2])
{
  double len[2];
  double l[2];

  double a = A[0][0];
  double b = A[0][1];
  double c = A[1][0];
  double d = A[1][1];

  double B = -(a+d);
  double C = a*d - b*c;
  
  double rad = sqrt((B*B) - (4.*C));
  double l0 = (-B - rad) / 2.;
  double l1 = (-B + rad) / 2.;
  
  double row[4][2];
  double norm[4];
  int best = 0;
  int i;

  row[0][0] = a - l0;
  row[0][1] = b;
  row[1][0] = c;
  row[1][1] = d - l0;
  row[2][0] = a - l1;
  row[2][1] = b;
  row[3][0] = c;
  row[3][1] = d - l1;
  
  norm[0] = row[0][0]*row[0][0] + row[0][1]*row[0][1];
  norm[1] = row[1][0]*row[1][0] + row[1][1]*row[1][1];
  norm[2] = row[2][0]*row[2][0] + row[2][1]*row[2][1];
  norm[3] = row[3][0]*row[3][0] + row[3][1]*row[3][1];

  best = 0;
  for(i=1; i<4; i++)
    {
      if (norm[i] > norm[best]) 
	best = i;
    }

  if ( norm[best] < 1e-10 )
    {
      L[0] = L[1] = 1.0;
      S[0][0] = S[1][1] = 1.0;
      S[0][1] = S[0][1] = 0.0;
      return;
    }
  
  
  {
    int one,two;
    if(best < 2) 
      one = 1, two = 0;
    else
      one = 0, two = 1;
	
    norm[best] = sqrt(norm[best]);

    S[0][one] =  row[best][0] / norm[best];
    S[1][one] =  row[best][1] / norm[best];
    S[0][two] =  S[1][one];
    S[1][two] = -S[0][one];
  }
  
  L[0] = l0;
  L[1] = l1;
  
  if (fabs(l0) < fabs(l1))
    {
      double tmp; 
      
      tmp = L[0];
      L[0] = L[1];
      L[1] = tmp;
      
      tmp = S[0][0];
      S[0][0] = S[0][1];
      S[0][1] = tmp;
      
      tmp = S[1][0];
      S[1][0] = S[1][1];
      S[1][1] = tmp;
    }
  
}

#ifdef DEBUGEIGEN
#include <stdlib.h>

void main (int argc,char **argv)
{
  if (argc != 5) 
    {
      fprintf(stderr,"usage: %s a b c d\n",argv[0]);
      exit (1);
    }
  
  {
    double **A = malloc (2*sizeof(double *));
    A[0] = malloc (2*sizeof(double));
    A[1] = malloc (2*sizeof(double));
    
    A[0][0] = atof(argv[1]);
    A[0][1] = atof(argv[2]);
    A[1][0] = atof(argv[3]);
    A[1][1] = atof(argv[4]);

    printf("matrix: %4.2f %4.2f\n",A[0][0],A[0][1]);
    printf("        %4.2f %4.2f\n",A[1][0],A[1][1]);
    printf("\n");
    
    {
      float L[2];
      float S[2][2]; 

      eigen(A,L,S);


      printf("eigenvalues: %g %g\n", L[0],L[1]);
	     
      printf("eigenvectors: %g %g\n",S[0][0],S[0][1]);
      printf("              %g %g\n",S[1][0],S[1][1]);
      printf("\n");
    }
  }    
}
#endif
