#include <stream.h>
#include <rpc/rpc.h>
#include <string.h>
#include <bstring.h>
#include <unistd.h>

#include "pfinder_clnt.h"

#include <Inventor/SoDB.h>
#include <Inventor/Xt/SoXt.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoCone.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransformSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedLineSet.h>
#include <X11/IntrinsicP.h>

static int headSize = 5.;
static int handSize = 1.5;

static Widget window = 0;
static SoSeparator* root = 0;

static SoSeparator* head = 0;
static SoMaterial* headMaterial = 0;
static SoTransform* headLocation = 0;
static SoSphere* headShape = 0;

static SoSphere* handShape = 0;

static SoSeparator* leftHand = 0;
static SoMaterial* leftHandMaterial = 0;
static SoTransform* leftHandLocation = 0;

static SoSeparator* rightHand = 0;
static SoMaterial* rightHandMaterial = 0;
static SoTransform* rightHandLocation = 0;

// static SoPerspectiveCamera *camera = 0;

static SoXtExaminerViewer* viewer = 0;

static
void
_handle()
{
  XtInputMask m;

  while( m = XtAppPending( SoXt::getAppContext() ) )
     XtAppProcessEvent( SoXt::getAppContext(), m );
}

static
int
_initialize( char *title = "Sick Viewer" )
{
  window = SoXt::init( "inv call pf" );
  if ( window == NULL) 
      return 1;

  root = new SoSeparator;
  root->ref();

  // NODES

  head = new SoSeparator;
  headMaterial = new SoMaterial;
  headMaterial->diffuseColor.setValue( .1, .1, .8 );
  headLocation = new SoTransform;
  headShape = new SoSphere;
  headShape->radius.setValue( headSize );

  handShape = new SoSphere;
  handShape->radius.setValue( handSize );

  leftHand = new SoSeparator;
  leftHandMaterial = new SoMaterial;
  leftHandMaterial->diffuseColor.setValue( .8, .1, .2 );
  leftHandLocation = new SoTransform;

  rightHand = new SoSeparator;
  rightHandMaterial = new SoMaterial;
  rightHandMaterial->diffuseColor.setValue( .5, .8, .1 );
  rightHandLocation = new SoTransform;

  // camera = new SoPerspectiveCamera;
  // camera->position.setValue( 0., 0., 5. );
  // camera->pointAt( SbVec3f( 0., 0., 0.) );
  // camera->orientation.setValue( SbVec3f( -0.067824, -0.706937, 0.704017 ), 3.16827 );

  // camera->position.setValue( -0.291696, -3.7881, 1.30955 );
  // camera->orientation.setValue( SbVec3f( -0.143925, -0.576789, 0.804114 ), 3.29064 );
  // SCENE GRAPH

  // root->addChild( camera );

  root->addChild( head );
  head->addChild( headMaterial );
  head->addChild( headLocation );
  head->addChild( headShape );

  root->addChild( leftHand );
  leftHand->addChild( leftHandMaterial );
  leftHand->addChild( leftHandLocation );
  leftHand->addChild( handShape );

  root->addChild( rightHand );
  rightHand->addChild( rightHandMaterial );
  rightHand->addChild( rightHandLocation );
  rightHand->addChild( handShape );

  // Set up viewer:
  viewer = new SoXtExaminerViewer( window );
  viewer->setSceneGraph(root);
  viewer->setTitle( title );
  viewer->show();

  SoXt::show( window );
  _handle();
  // XtResizeWidget( window, 650, 500, 0 );
  // XtMoveWidget( window, 6, 0 );
  // XtConfigureWidget( window, 6, 0, 640, 480, 0 );
  // SoXt::mainLoop();
  return 0;
}

static
int
_update( pfv3f& head, pfv3f& leftHand, pfv3f& rightHand )
{
  // fprintf( stderr, "head: %g %g %g\n", head.x, head.y, head.z );
  // fprintf( stderr, "rightHand: %g %g %g\n", rightHand.x, rightHand.y, rightHand.z );
  // fprintf( stderr, "leftHand: %g %g %g\n\n", leftHand.x, leftHand.y, leftHand.z );
  headLocation->translation.setValue( head.x, head.y, head.z );
  leftHandLocation->translation.setValue( leftHand.x, leftHand.y, leftHand.z );
  rightHandLocation->translation.setValue( rightHand.x, rightHand.y, rightHand.z );
  return 0;
}

void
main ( int argc, char *argv[] )
{
  CLIENT *cl;
  pfperson *p;
  pfopts o;
  bzero(&o,sizeof(pfopts));

  if (argc<2) exit(0); // first arg is hostname of pfinder

  if (!( cl = clnt_create( argv[1], PFINDERPROG, PFINDERVERS, "tcp")))
     clnt_pcreateerror( argv[1] ), exit(1);

  if (argc>2) {
    o.kalman = atoi(argv[2]);
    printf(" kalman is %d\n", o.sync);
  }
  if (argc>3) {
    int t = atoi(argv[3]);
    printf(" trgger is %d\n", t);
    pftrigger_1(&t, cl);
  }

  if( o.kalman )
     _initialize( "Kalman filtered" );
  else
     _initialize( "Raw" );
  o.f3 = o.f2 = YES;
  for( ;; )
     {
       p = pf_1(&o, cl);
       if (!p)
	  clnt_perror(cl, argv[1]), exit(-1);
       if (p->f3.opt)
	  {
	    pffeat3d_s *feats = p->f3.pffeat3d_u.feats;
	    _update( feats->head, feats->left_hand, feats->right_hand );
	  }
       else
	  fprintf( stderr, "no f3\n" );

       _handle();
     }
}

