/* iveImage.c */
/* RCS: $Header: /mas/vision/projects/pfinder/RCS/iveImage.c,v 6.1 1995/10/06 21:42:15 cwren Exp $ */

#include "iveImage.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class ivePixel
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DIGITIZEYUV

/* these functions are return by value, whereas the macro is writable */

unsigned char
ivePixelGetColor( ivePixel *pix, ivePixelCh ch )
{
  return ( ivePixelColor( pix, ch ) );
}

unsigned char
ivePixelGetRed( ivePixel *pix )
{
  return ( pix->red );
}

unsigned char
ivePixelGetGreen( ivePixel *pix )
{
  return ( pix->green );
}

unsigned char
ivePixelGetBlue( ivePixel *pix )
{
  return ( pix->blue );
}

void
ivePixelPutColor( ivePixel *pix, ivePixelCh ch, unsigned char val )
{
  ivePixelColor( pix, ch ) = val;
}

void
ivePixelPutRed( ivePixel *pix, unsigned char val )
{
  pix->red = val;
}

void
ivePixelPutGreen( ivePixel *pix, unsigned char val )
{
  pix->green = val;
}

void
ivePixelPutBlue( ivePixel *pix, unsigned char val )
{
  pix->blue = val;
}

#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class iveImage
 *
 * iveImageMake
 * iveImageMakeCopy
 * iveImageFree
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

iveImage *
iveImageMake( int height, int width )
{
  iveImage *im = ( iveImage * ) malloc( sizeof( iveImage ) );

  if ( im ) 
    {
      im -> height = height;
      im -> width = width;
      im -> pixelCount = width * height;
      im -> byteCount = im->pixelCount * iveBPP;
      im -> data = (unsigned char *) malloc( im->byteCount );
      if ( ! im->data )
	{
	  free( im );
	  return 0;
	}
    }
  return im;
}

iveImage *
iveImageCopy( iveImage *src, iveImage *dst )
{
  if ( dst && src ) 
    {
      dst -> height = src->height;
      dst -> width = src->width;
      dst -> pixelCount = src->pixelCount;
      dst -> byteCount = src->byteCount;
      bcopy( src->data, dst->data, dst->byteCount );
    }
  return dst;
}

iveImage *
iveImageMakeCopy( iveImage *src )
{
  iveImage *im = ( iveImage * ) malloc( sizeof( iveImage ) );

  if ( im && src ) 
    {
      im -> height = src->height;
      im -> width = src->width;
      im -> pixelCount = src->pixelCount;
      im -> byteCount = src->byteCount;
      im -> data = (unsigned char *) malloc( im->byteCount );
      if ( ! im->data )
	{
	  free( im );
	  return 0;
	}
      bcopy( src->data, im->data, im->byteCount );
    }
  return im;
}

iveImage *
iveImageInit( iveImage *im, int height, int width, unsigned char *buffer )
{
  im -> height = height;
  im -> width = width;
  im -> pixelCount = width * height;
  im -> byteCount = im->pixelCount * iveBPP;
  im -> data = buffer;
  return im;
}


/* this should only be called on Make or MakeCopy results */
void iveImageFree(iveImage *im)	
{
  free(im->data);
  free(im);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * iveImageReadRaw
 * iveImageWriteRaw
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DIGITIZEYUV

int
iveImageReadRaw( iveImage *im, FILE *stream )
{
  unsigned char rgb[3];
  int num;
  int i;
  
  for( i = 0; i < im->pixelCount; i++ )
    {
      num = fread( rgb, sizeof(unsigned char), 3, stream );
      if( EOF == num )
	return 1;
      if ( num != 3 )
	{
	  fprintf( stderr,
		  "iveImage: EOF or read raw error (%d, %d) \n", i, num );
	  return 1;
	}
      iveImageColorI( im, i, iveImageRed ) = rgb[0];
      iveImageColorI( im, i, iveImageGreen ) = rgb[1];
      iveImageColorI( im, i, iveImageBlue ) = rgb[2];
    }
  return 0;
}

int
iveImageWriteRaw( iveImage *im, FILE *stream )
{
  int i;
  
  for( i = 0; i < im->pixelCount; i++ )
    {
      putc( iveImageColorI( im, i, iveImageRed ), stream );
      putc( iveImageColorI( im, i, iveImageGreen ), stream );
      putc( iveImageColorI( im, i, iveImageBlue ), stream );
    }
  return 0;
}

#else

/* YUV */
#define MIN( a, b ) ( ( (a) < (b) ) ? (a) : (b) )
#define MAX( a, b ) ( ( (a) > (b) ) ? (a) : (b) )
#define LIMIT( x, low, high ) MAX( (low), MIN( (x), (high) ) )


int
iveImageReadRaw( iveImage *im, FILE *stream )
{
  unsigned char rgb[3];
  unsigned char y0, y1, u, v;
  unsigned char *ucp = im->data;
  int num;
  int i;
  
  for( i = 0; i < im->pixelCount; i++ )
    {
      num = fread( rgb, sizeof(unsigned char), 3, stream );
      if( feof( stream ) )
	return 1;
      if ( num != 3 )
	{
	  fprintf( stderr, "iveImage: read raw error (%d, %d) \n", i, num );
	  return 2;
	}
      y0 = LIMIT( (.299 * rgb[0] + .587 * rgb[1] + .114 * rgb[2]), 0., 255. );
      u = LIMIT( ((rgb[2] + 227. - y0 )/2.), 0., 255. );
      v = LIMIT( ((rgb[0] + 179. - y0 )/2.), 0., 255. );
      i++;
      num = fread( rgb, sizeof(unsigned char), 3, stream );
      y1 = LIMIT( (.299 * rgb[0] + .587 * rgb[1] + .114 * rgb[2]), 0., 255. );
      *ucp++ = u;
      *ucp++ = y0;
      *ucp++ = v;
      *ucp++ = y1;
    }
  return 0;
}

int
iveImageWriteRaw( iveImage *im, FILE *stream )
{
  int i;
  float y, u, v;
  float red, green, blue;
  unsigned char r, g, b;
  
  for( i = 0; i < im->pixelCount ; i ++ )
    {
      ivePixel ivepixel;
      iveImageGetPixelI( im, i, &ivepixel );

      y = ivePixelChannel( &ivepixel, 0 );
      u = ivePixelChannel( &ivepixel, 1 );
      v = ivePixelChannel( &ivepixel, 2 );

      red = (1.164*(y-16.0) + 0.000*(u-128.0) + 1.596*(v-128.0));
      blue = (1.164*(y-16.0) + 2.017*(u-128.0) + 0.000*(v-128.0));
      green = (1.164*(y-16.0) - 0.392*(u-128.0) - 0.813*(v-128.0));

      r = LIMIT( red, 0., 255. );
      b = LIMIT( blue, 0., 255. );
      g = LIMIT( green, 0., 255. );

#ifdef OUTPUT_YUV
      putc( (unsigned char)y, stream );
      putc( (unsigned char)u, stream );
      putc( (unsigned char)v, stream );
#else
#ifdef OUTPUT_Y
      putc( y, stream );
#else
      putc( r, stream );
      putc( g, stream );
      putc( b, stream );
#endif
#endif
    }
  return 0;
}

#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * iveImageGetPixel
 * iveImagePutPixel
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DIGITIZEYUV

ivePixel *
iveImageGetPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  unsigned char *ucp = im->data + iveImageIndex( im, y, x, 0 );
  pix->alpha = *ucp++;
  pix->blue = *ucp++;
  pix->green = *ucp++;
  pix->red = *ucp++;
}

ivePixel *
iveImageGetPixelI( iveImage *im, int index, ivePixel *pix )
{
  unsigned char *ucp = im->data + index * iveBPW;
  pix->alpha = *ucp++;
  pix->blue = *ucp++;
  pix->green = *ucp++;
  pix->red = *ucp++;
}

void
iveImagePutPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  unsigned char *ucp = im->data + iveImageIndex( im, y, x, 0 );
  *ucp++ = pix->alpha;
  *ucp++ = pix->blue;
  *ucp++ = pix->green;
  *ucp++ = pix->red;
}

#else

ivePixel *
iveImageGetPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  pix -> y = iveImageColor( im, y, x, iveImageY );
  pix -> u = iveImageColor( im, y, x, iveImageU );
  pix -> v = iveImageColor( im, y, x, iveImageV );
}

ivePixel *
iveImageGetPixelI( iveImage *im, int index, ivePixel *pix )
{
  pix -> y = iveImageColorI( im, index, iveImageY );
  pix -> u = iveImageColorI( im, index, iveImageU );
  pix -> v = iveImageColorI( im, index, iveImageV );
}

void
iveImagePutPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  iveImageColor( im, y, x, iveImageY ) = pix->y;
  iveImageColor( im, y, x, iveImageU ) = pix->u;
  iveImageColor( im, y, x, iveImageV ) = pix->v;
}

#endif
