// lite.h++

#ifndef _lite_H_
#define _lite_H_

#include <stdio.h>
#include "pfCamera.h++"
#include "pfFeatures.h++"
#include "pfPersonSize.h++"
#include "pfGesture.h++"
#include "pfSemaphore.h++"
#include "pfContour.h++"
#include "KalmanFilter/vtiKalmanFilter.h++"

class lite
{
private:
    static pfCamera* camera;
    static pfFeatures features;
    static pfPersonSize personSize;
    static pfGesture gestures;
    
public:
    static void initialize();
    static void interpretPfContour( pfContour& );
    static void dumpContour( FILE* );
    static void handleEvents();
    
    static inline void Project( float in[3], float out[3] ); // out is y-down
    static inline void BackProject( float in[2], float out[3], float z ); // in is y-down
    static inline void BackProject( float in[2], float out[3] ); // in is y-down
    
    static inline float* PersonDimensions() { return personSize.Size(); }
    static inline int PersonDimensionsValid() { return 1; }
    static inline float aspectRatio() { return camera->AspectRatio(); }
    
    // Kalman filters
private:
    static vtiKalmanFilter *khead_kal;
    static vtiKalmanFilter *krightHand_kal;
    static vtiKalmanFilter *kleftHand_kal;
    static vtiKalmanFilter *krightFoot_kal;
    static vtiKalmanFilter *kleftFoot_kal;
    static vtiKalmanFilter *kbodyCentroid_kal;
    static _feature_data kalmanFeatures;
    static _feature_data& kalmanUpdate( _feature_data& in );
public:
    static pfSemaphore kalmanSemaphore;
    static float mmtNoise;
    static float accelNoise;
    static int kfGrain;
    static inline _feature_data& FeatureDataKalman() { return kalmanFeatures; }
    
private:
    static int personExists; // if the contour is big enough
public:
    static inline pfGesture& getGestures() { return gestures; }
    static inline pfFeatures& getFeatures() { return features; }
    static inline int exists() { return personExists; }
    static int reportFreq;
    static int reportMove;
    static int contourGroove;
    static int printGestures;
};

inline
void
lite::Project( float in[3], float out[3] )
{
    camera -> projectToScreen( in, out );
}

inline
void
lite::BackProject( float in[2], float out[3], float z )
{
    camera -> projectScreenYtoWorldY( in, out, z );
}

inline
void
lite::BackProject( float in[2], float out[3] )
{
    BackProject( in, out, features.getCurrent().userplaneCentroid[2] );
}

#endif /* _lite_H_ */
