// pfContour.c++

#include <stream.h>
#include "pfContour.h++"

pfContour::pfContour() { clear(); }

pfContour&
pfContour::generate( iveBitmap *bm, ivePoint *pcenter, int )
{
  static int horizLUT[8] = {-1, 0, 1, 1, 1, 0, -1, -1};
  static int vertLUT[8] = {-1, -1, -1, 0, 1, 1, 1, 0};

  height = bm->height;
  width = bm->width;

  clear();
  centroid[0] = bBoxMin[0] = bBoxMax[0] = pcenter -> x;
  centroid[1] = bBoxMin[1] = bBoxMax[1] = pcenter -> y;

#if 0
  // Find a low point on the contour
  int y;
  if( !getBit( bm, pcenter->y, pcenter->x ) )
    {
      y = pcenter->y+1;
      while( y < bm->height && !getBit( bm, y, pcenter->x ) )
	y++;
      if( y == bm->height )
	return *this;
    }
  else
    {
      for( y = pcenter->y;
	  y < bm->height-1 && getBit( bm, y+1, pcenter->x );
	  y++);
    }
  vec2iSet( points, pcenter->x, y );
#endif
  // Find a good point on the side of the contour
  int fmDir = 7;
  int x = 0;
  while( x < pcenter->x && !getBit( bm, pcenter->y, x ) )
    x++;
  if( x == pcenter ->x )
    {
      x = bm->width-1;
      while( x > pcenter->x && !getBit( bm, pcenter->y, x ) )
	x--;
      if( x == pcenter->x )
	return *this;
      fmDir = 3;
    }
  vec2iSet( points, x, pcenter->y );

  int *current = points;
  length = 1;
  
  // Directions:
  //   0 1 2
  //   7 + 3
  //   6 5 4

  // int dirToBGpoint_start = 5; // DOWN (cuz we just searched up)
  int dirToBGpoint_start = fmDir;  // either left or right
  int direction;

  while( ! vec2iEquiv( current, points ) || 1==length )
    {
      int *nextPoint = current + 2;

      vec2iUpdateBox( bBoxMin, bBoxMax, current );

      // Find direction, and coords of next CW contour point 
      direction = dirToBGpoint_start;
      do {
	direction = (direction + 1) % 8;
	vec2iSet( nextPoint,
		 current[0] + horizLUT[direction],
		 current[1] + vertLUT[direction] );
	if( direction == dirToBGpoint_start )
	  cerr << "pfContour::generate: ERROR" << endl;
      }
      while( !getBit( bm, nextPoint ) );

      // Reset the dirToBGpoint for the next time around:
      // +4 points back to current point
      // +1 rotates one more in CW dir, which is guaranteed to
      //    be in BG.
      dirToBGpoint_start = ( direction + 5 ) % 8;
      
      current = nextPoint;
      length++;
    }
  return *this;
}

