// pfContour.h++

#ifndef _pfContour_H_
#define _pfContour_H_

extern "C"
{
#include "iveBitmap.h"
#include "ivePoint.h"
typedef Point ivePoint;
}

class pfContour
{
  inline int getBit( iveBitmap *, int y, int x );
  inline int getBit( iveBitmap *, int *v );
public:
  const int maxContourLength = 5000;
  // these are in pixels
  int centroid[2];
  int bBoxMin[2];
  int bBoxMax[2];
  int length;
  int points[ 2 * maxContourLength ];
  inline void clear();
  int height;
  int width;
  pfContour();
  pfContour& generate( iveBitmap *, ivePoint *, int );
};

inline void vec2iZero( int *v ) { v[0] = v[1] = 0; }
inline void vec2iCopy( int *src, int *dst ) 
{
  dst[0] = src[0];
  dst[1] = src[1];
}
inline void vec2iSet( int *v, int x, int y )
{
  v[0] = x;
  v[1] = y;
}
inline void vec2iUpdateBox( int *bmin, int *bmax, int *v )
{
  if( v[0] < bmin[0] ) bmin[0] = v[0];
  if( v[1] < bmin[1] ) bmin[1] = v[1];
  if( v[0] > bmax[0] ) bmax[0] = v[0];
  if( v[1] > bmax[1] ) bmax[1] = v[1];
}
inline int vec2iEquiv( int *v1, int *v2 )
{
  return( v1[0]==v2[0] && v1[1]==v2[1] );
}

inline
int
pfContour::getBit( iveBitmap *bm, int y, int x )
{
  return( bm->data[ y * bm->width + x ] );
}

inline
int
pfContour::getBit( iveBitmap *bm, int *v )
{
  if( v[0] < 0 || v[0] >= bm->width || v[1] < 0 || v[1] >= bm->height )
    return 0;
  return( bm->data[ v[1] * bm->width + v[0] ] );
}

inline
void
pfContour::clear()
{
  length = 0;
  vec2iZero( centroid );
  vec2iZero( bBoxMin );
  vec2iZero( bBoxMax );
}


#endif // _pfContour_H_
