// pfFeatures.c++

#define SIGGRAPH

#include "pfFeatures.h++"

extern "C" 
{
#include "bsXclient.h"
}

float pfFeatures::hmin_pu = .3;
float pfFeatures::hmax_pu = .6;
float pfFeatures::feetTol = 1.;
float pfFeatures::headTol = 3.;
float pfFeatures::extendedTol = 10.;
float pfFeatures::xLeft = -72.; // inches, left of ALIVE space
			       // initial condition for centroid

pfFeatures::pfFeatures()
  : features( featureBuffers ),
    featuresLatest( featureBuffers + 1 ),
    numExtremePoints( 0 ),
    scale( 1. )
{}

int handMin = 40;
int handMax = 80;

static
inline
int
inLowBox( _feature_data* fts, float *fp )
{
  return( fp[1] < (fts->userplaneBoxBR[1] + pfFeatures::feetTol) );
}

static
inline
int
inHiMidBox( _feature_data* fts, float *fp )
{
  float dx = fp[0] - fts->userplaneCentroid[0];
  float absdx = ( dx < 0 ) ? -dx : dx;
  return( ( fp[1] > fts->userplaneCentroid[1] ) &&
	 ( absdx < pfFeatures::headTol ) );
}

static
inline
int
inHiBox( _feature_data* fts, float *fp )
{
  float dx = fp[0] - fts->userplaneCentroid[0];
  float absdx = ( dx < 0 ) ? -dx : dx;
  return( ( fp[1] > fts->userplaneCentroid[1] ) &&
	 ( absdx < 2.*pfFeatures::headTol ) );
}

static
inline
int
ringMod( int x, int m )
{
  if( x>=0 )
    return x%m;
  while( x < 0 )
    x+=m;
  return x;
}

void
pfFeatures::PFind( _feature_data* fts, pfCamera* )
{
  ContourPointSet lowSet;
  ContourPointSet hiMidSet;
  ContourPointSet hiSet;
  ContourPointSet rightSet;
  ContourPointSet leftSet;

  // Go over the contour, finding lowSet and himidSet

  float *cpt = fts -> userplaneContour;
  float* wayHighPt = cpt;
  for( int i = 0; i < fts->numContourPoints; i++ )
    {
      if( inLowBox( fts, cpt ) )
	lowSet.add( cpt );
      else if( inHiMidBox( fts, cpt ) )
	hiMidSet.add( cpt );
      if( inHiBox( fts, cpt ) )
	hiSet.add( cpt );
      if( cpt[1] > wayHighPt[1] )
	wayHighPt = cpt;
      cpt += 3;
    }

  // Analyze the pfeet

  float xmin = lowSet.points[0][0];
  float xmax = lowSet.points[0][0];
  float* leftFootPt = lowSet.points[0];
  float* rightFootPt = lowSet.points[0];
  for( i = 1; i < lowSet.n; i++ )
    {
      if( lowSet.points[i][0] < xmin )
	{
	  xmin = lowSet.points[i][0];
	  leftFootPt = lowSet.points[i];
	}
      if( lowSet.points[i][0] > xmax )
	{
	  xmax = lowSet.points[i][0];
	  rightFootPt = lowSet.points[i];
	}
    }
  vector3Copy( leftFootPt, fts -> userplaneLeftFoot );
  vector3Copy( rightFootPt, fts -> userplaneRightFoot );
  fts -> leftFootValid = pfTRUE;
  fts -> rightFootValid = pfTRUE;

  // Analyze the head
  // make the head be the point with max y in the hiMidSet

  float ymax;
  float *headPt;
  if( hiMidSet.n > 0 )
    {
      ymax = hiMidSet.points[0][1];
      headPt = hiMidSet.points[0];
      for( i = 1; i < hiMidSet.n; i++ )
	if( hiMidSet.points[i][1] > ymax )
	  {
	    ymax = hiMidSet.points[i][1];
	    headPt = hiMidSet.points[i];
	  }
    }
  else if( hiSet.n > 0 )
    {
      ymax = hiSet.points[0][1];
      headPt = hiSet.points[0];
      for( i = 1; i < hiSet.n; i++ )
	if( hiSet.points[i][1] > ymax )
	  {
	    ymax = hiSet.points[i][1];
	    headPt = hiSet.points[i];
	  }
    }
  else 
    {
      cerr << "well oh darn, there's not much we can do then" << endl;
      headPt = wayHighPt;
    }
  vector3Copy( headPt, fts -> userplaneHead );
  fts -> headValid = pfTRUE;

  // pfind hands

  int headIndex = ( headPt - fts->userplaneContour ) / 3;
  int leftFootIndex = ( leftFootPt - fts->userplaneContour ) / 3;
  int rightFootIndex = ( rightFootPt - fts->userplaneContour ) / 3;
  int leftSideLength = ringMod( headIndex - leftFootIndex,
			       fts->numContourPoints );
  int rightSideLength = ringMod( rightFootIndex - headIndex,
				fts->numContourPoints );

  // (LEFT HAND)
  int index = ringMod( int(headIndex - hmin_pu * leftSideLength),
		      fts->numContourPoints );
  int endIndex = ringMod( int(headIndex - hmax_pu * leftSideLength),
			 fts->numContourPoints );
  float* pt;
  for( i = index; i != endIndex; i = ringMod(i-1, fts->numContourPoints) )
    {
      pt = fts->userplaneContour + 3*i;
      leftSet.add( pt );
    }
  xmax = leftSet.points[0][0];
  ymax = leftSet.points[0][1];
  float *widePt = leftSet.points[0];
  float *highPt = leftSet.points[0];
  for( i = 1; i < leftSet.n; i++ )
    {
      if( leftSet.points[i][0] < xmax )
	{
	  xmax = leftSet.points[i][0];
	  widePt = leftSet.points[i];
	}
      if( leftSet.points[i][1] > ymax )
	{
	  ymax = leftSet.points[i][1];
	  highPt = leftSet.points[i];
	}
    }
  float *leftPt = widePt;
  if( ymax > headPt[1] )
    leftPt = highPt;
  else if( (headPt[0]-xmax ) < extendedTol )
    { 
      index = ringMod( int(headIndex - hmax_pu*leftSideLength),
		      fts->numContourPoints );
      leftPt = fts->userplaneContour + 3*index;
    }
  vector3Copy( leftPt, fts->userplaneLeftHand );
  fts -> leftHandValid = pfTRUE;
	  
  // (RIGHT HAND)
  index = ringMod( int(headIndex + hmin_pu * rightSideLength),
		  fts->numContourPoints );
  endIndex = ringMod( int(headIndex + hmax_pu * rightSideLength),
		     fts->numContourPoints );
  for( i = index; i != endIndex; i = ringMod((i+1), fts->numContourPoints) )
    {
      pt = fts->userplaneContour+ 3*i;
      rightSet.add( pt );
    }
  xmax = rightSet.points[0][0];
  ymax = rightSet.points[0][1];
  widePt = rightSet.points[0];
  highPt = rightSet.points[0];
  for( i = 1; i < rightSet.n; i++ )
    {
      if( rightSet.points[i][0] > xmax )
	{
	  xmax = rightSet.points[i][0];
	  widePt = rightSet.points[i];
	}
      if( rightSet.points[i][1] > ymax )
	{
	  ymax = rightSet.points[i][1];
	  highPt = rightSet.points[i];
	}
    }
  float* rightPt = widePt;
  if( ymax > headPt[1] )
    rightPt = highPt;
  else if( (xmax-headPt[0]) < extendedTol )
    {
      index = ringMod( int(headIndex + hmax_pu * rightSideLength),
		      fts->numContourPoints );
      rightPt = fts->userplaneContour + 3*index;
    }
  vector3Copy( rightPt, fts->userplaneRightHand );
  fts -> rightHandValid = pfTRUE;

  
}

static void
_copyColorClassData(_feature_data* fts)
{
  register int len = fts->numColorClasses = bsClassifyGetNumClasses();
  for(int i=0; i<len; i++)
    {
      bsClassifyGetClassColor(i, 
			      &fts->colorclasses2d[i].cmean[0],
			      &fts->colorclasses2d[i].cmean[1],
			      &fts->colorclasses2d[i].cmean[2]);
      bsClassifyGetClassPosition(i, 
				 &fts->colorclasses2d[i].gmean[0],
				 &fts->colorclasses2d[i].gmean[1]);
      bsClassifyGetClassEigen(i, 
			      fts->colorclasses2d[i].gL, 
			      fts->colorclasses2d[i].gS);
      
    }
}

static void
_projColorClassTo3D(_feature_data* fts, pfCamera* camera)
{
  register int len = fts->numColorClasses;
  for(int i=0; i<len; i++)
    {
      int iwidth, iheight;
      bsClassifyGetScreenSize(&iwidth, &iheight);
      float width  = (float)iwidth;
      float height = (float)iheight;
      
      float bottom[2]; // average bbox bottom in screen coords (y-down)
      bottom[0] = 0.5;
      bottom[1] = fts->imageBoxMax[1];
      float bottom3d[3];
      camera -> projectYtoZFloor( bottom, bottom3d );
      float& user_z = bottom3d[2];
      
      // fill in 3D structure from 2d structure
      fts->colorclasses3d[i].cmean[0] = fts->colorclasses2d[i].cmean[0];
      fts->colorclasses3d[i].cmean[1] = fts->colorclasses2d[i].cmean[1];
      fts->colorclasses3d[i].cmean[2] = fts->colorclasses2d[i].cmean[2];
      
      // project 2d centroid into 3D
      float tmp2[2];
      tmp2[0] = fts->colorclasses2d[i].gmean[0] / width;
      tmp2[1] = fts->colorclasses2d[i].gmean[1] / height;
      camera -> projectScreenYtoWorldY(tmp2,
				       fts->colorclasses3d[i].gmean,
				       user_z );
      
      for(int j=0; j<2; j++)
	{
	  float tmp3[3];		
	  // build the vector in 2d
	  tmp2[0] = ((fts->colorclasses2d[i].gL[j] * 
		      fts->colorclasses2d[i].gS[0][j]) + 
		     fts->colorclasses2d[i].gmean[0]) /
		       width;
	  tmp2[1] = ((fts->colorclasses2d[i].gL[j] * 
		      fts->colorclasses2d[i].gS[1][j]) +
		     fts->colorclasses2d[i].gmean[1]) /
		       height;
	  // project to 3D
	  camera -> projectScreenYtoWorldY(tmp2,
					   tmp3,
					   user_z );
	  // take off the projected centroid
	  tmp3[0] -= fts->colorclasses3d[i].gmean[0];
	  tmp3[1] -= fts->colorclasses3d[i].gmean[1];
	  tmp3[2] -= fts->colorclasses3d[i].gmean[2];
	  // find the length
	  fts->colorclasses3d[i].gL[j] = sqrt(tmp3[0]*tmp3[0] + 
					      tmp3[1]*tmp3[1] +
					      tmp3[2]*tmp3[2]);
	  // normalize the components
	  fts->colorclasses3d[i].gS[0][j]  = tmp3[0];
	  fts->colorclasses3d[i].gS[0][j] /= fts->colorclasses3d[i].gL[j];
	  fts->colorclasses3d[i].gS[1][j]  = tmp3[1];
	  fts->colorclasses3d[i].gS[1][j] /= fts->colorclasses3d[i].gL[j];
	  fts->colorclasses3d[i].gS[2][j]  = tmp3[2];
	  fts->colorclasses3d[i].gS[2][j] /= fts->colorclasses3d[i].gL[j];
	}
      
      // 2d blob in 3d implies singular S
      // make last vector perp to other two
      // with e.value=0 -> sigma=0 -> determinisitc
      fts->colorclasses3d[i].gL[2] = 0.0;
      fts->colorclasses3d[i].gS[0][2]  = ((fts->colorclasses3d[i].gS[1][0]*
					   fts->colorclasses3d[i].gS[2][1])-
					  (fts->colorclasses3d[i].gS[2][0]*
					   fts->colorclasses3d[i].gS[1][1]));
      fts->colorclasses3d[i].gS[1][2]  = ((fts->colorclasses3d[i].gS[2][0]*
					   fts->colorclasses3d[i].gS[0][1])-
					  (fts->colorclasses3d[i].gS[0][0]*
					   fts->colorclasses3d[i].gS[2][1]));
      fts->colorclasses3d[i].gS[2][2]  = ((fts->colorclasses3d[i].gS[0][0]*
					   fts->colorclasses3d[i].gS[1][1])-
					  (fts->colorclasses3d[i].gS[1][0]*
					   fts->colorclasses3d[i].gS[0][1]));
    }
}

void
pfFeatures::Find( pfCamera* camera )
{

  PFind( features, camera );
	
  for( int i = 0; i < features->numContourPoints; i++ )
    vector3SubtractEq( features->userplaneContour + 3*i,
		      features->userplaneCentroid );
  
  camera -> projectToScreen2(features->userplaneHead, 
			     features->imageHead );
  camera -> projectToScreen2(features->userplaneLeftHand, 
			     features->imageLeftHand );
  camera -> projectToScreen2(features->userplaneRightHand, 
			     features->imageRightHand );
  camera -> projectToScreen2(features->userplaneLeftFoot, 
			     features->imageLeftFoot );
  camera -> projectToScreen2(features->userplaneRightFoot, 
			     features->imageRightFoot );
  
  _copyColorClassData(features);
  _projColorClassTo3D(features, camera);
}

#if 0
int
pfFeatures::ProjectContourToUserPlane( ContourPacket* cp, pfCamera* camera )
{
    // (1) UNPACK CONTOUR PACKET
    //     ---------------------
    packetPoint *pt;
    float *npt;
    pt = &(cp->packet.centroid);
    npt = features->imageCentroid;
    normalizePacketPoint(pt->x, pt->y, npt, npt+1);
    pt = &(cp->packet.bBoxMin);
    npt = features->imageBoxMin;
    normalizePacketPoint(pt->x, pt->y, npt, npt+1);
    pt = &(cp->packet.bBoxMax);
    npt = features->imageBoxMax;
    normalizePacketPoint(pt->x, pt->y, npt, npt+1);
    COORD_TYPE *contour = cp->packet.contour;
    for( int i = 0; i < (int) cp->packet.numOfPoints; i++ )
       normalizePacketPoint( contour[i*COORDS_PER_POINT],
			     contour[i*COORDS_PER_POINT + 1],
			     features->imageContour + i*2,
			     features->imageContour + i*2 + 1 );

    COORD_TYPE *extremities = cp->packet.extremityList;
    for( i = 0; i < (int) cp->packet.numExtremities; i++ )
       normalizePacketPoint( extremities[i*COORDS_PER_POINT], 
			     extremities[i*COORDS_PER_POINT + 1], 
			     imageExtreme + i*2, 
			     imageExtreme + i*2 + 1 );
    numExtremePoints = cp->packet.numExtremities;
    features->numContourPoints = cp->packet.numOfPoints;
    

    // (2) CHECK FOR DEGENERACY
    //     --------------------
    if( features->numContourPoints < 10 ) 
       {
	   vectorSetValue( features->userplaneCentroid, xLeft, 0., 0. );
	   return 0;  // this return value is used to set !userExists
       }


    // (3) ATTEMPT SOME LAME-ASS HEURISTICS FOR ROBUST FILTERING THE
    //     BOTTOM POINT
    //     ---------------------------------------------------------
    float bottom[2]; // average bbox bottom in screen coords
    bottom[0] = 0.;
#if 0
    boolean error;
    bottom[1] = updateSWFilter( swf, screen_features.boxMax[1] );
    bottom[1] = MAX( MAX( rbGetElement( swf->rb, 0, &error ),
			  rbGetElement( swf->rb, 1, &error ) ),
		     rbGetElement( swf->rb, 2, &error ) );
    if ( bottom[1] < 0.45 )
       bottom[1] = 0.45;	// ugly fucking hacks TJD
#endif
    bottom[1] = features->imageBoxMax[1];


    // (4) PROJECT CONTOUR PTS, EXTREME PTS, CENTROID, and BOUNDING BOX
    //     ------------------------------------------------------------
    float bottom3d[3];
    camera -> projectYtoZFloor( bottom, bottom3d );
    float& user_z = bottom3d[2];
    camera -> projectScreenYtoWorldY( features->numContourPoints,
				      features->imageContour,
				      features->userplaneContour,
				      user_z );
    camera -> projectScreenYtoWorldY( numExtremePoints,
				      imageExtreme,
				      userplaneExtreme,
				      user_z );
    camera -> projectScreenYtoWorldY( features->imageCentroid,
				      features->userplaneCentroid,
				      user_z );
    // recalc bounding box 
    vector3Copy( features->userplaneContour, features->userplaneBoxUL );
    vector3Copy( features->userplaneContour, features->userplaneBoxBR );
    for ( i = 1; i < features->numContourPoints; i++ )
       {
	   features->userplaneBoxUL[0] 
	      = MIN( features->userplaneBoxUL[0],
		     features->userplaneContour[3*i] );
	   features->userplaneBoxUL[1]
	      = MAX( features->userplaneBoxUL[1],
		     features->userplaneContour[3*i+1] );
	   features->userplaneBoxBR[0]
	      = MAX( features->userplaneBoxBR[0],
		     features->userplaneContour[3*i+0] );
	   features->userplaneBoxBR[1]
	      = MIN( features->userplaneBoxBR[1],
		     features->userplaneContour[3*i+1] );
       }

    return features->numContourPoints;
}
#endif

int
pfFeatures::ProjectPfContourToUserPlane( pfContour& cont, pfCamera* camera )
{
  if( cont.length < minContourLength ) 
    {
      vectorSetValue( features->userplaneCentroid, xLeft, 0., 0. );
      return 0;
    }

  float width = cont.width;
  float height = cont.height;

  features -> imageCentroid[0] = float(cont.centroid[0]) / width;
  features -> imageCentroid[1] = float(cont.centroid[1]) / height;
  features -> imageBoxMin[0] = float(cont.bBoxMin[0]) / width;
  features -> imageBoxMin[1] = float(cont.bBoxMin[1]) / height;
  features -> imageBoxMax[0] = float(cont.bBoxMax[0]) / width;
  features -> imageBoxMax[1] = float(cont.bBoxMax[1]) / height;

  features -> numContourPoints = cont.length;
  for( int i = 0; i < cont.length; i++ )
    {
      features->imageContour[i*2] = float(cont.points[i*2]) / width;
      features->imageContour[i*2+1] = float(cont.points[i*2+1]) / height;
    }

  float bottom[2]; // average bbox bottom in screen coords (y-down)
  bottom[0] = 0.5;
  bottom[1] = features->imageBoxMax[1];
  float bottom3d[3];
  camera -> projectYtoZFloor( bottom, bottom3d );
  float& user_z = bottom3d[2];

  camera -> projectScreenYtoWorldY( features->numContourPoints,
				   features->imageContour,
				   features->userplaneContour,
				   user_z );
#if 0
  camera -> projectScreenYtoWorldY( numExtremePoints,
				   imageExtreme,
				   userplaneExtreme,
				   user_z );
#endif
  camera -> projectScreenYtoWorldY( features->imageCentroid,
				   features->userplaneCentroid,
				   user_z );
    // recalc bounding box 
    vector3Copy( features->userplaneContour, features->userplaneBoxUL );
    vector3Copy( features->userplaneContour, features->userplaneBoxBR );
    for ( i = 1; i < features->numContourPoints; i++ )
       {
	   features->userplaneBoxUL[0] 
	      = MIN( features->userplaneBoxUL[0],
		     features->userplaneContour[3*i] );
	   features->userplaneBoxUL[1]
	      = MAX( features->userplaneBoxUL[1],
		     features->userplaneContour[3*i+1] );
	   features->userplaneBoxBR[0]
	      = MAX( features->userplaneBoxBR[0],
		     features->userplaneContour[3*i+0] );
	   features->userplaneBoxBR[1]
	      = MIN( features->userplaneBoxBR[1],
		     features->userplaneContour[3*i+1] );
       }
  return features->numContourPoints;
}

ostream&
pfFeatures::Print( ostream& os )
{
    _print_vector( os << "CENTROID: ", featuresLatest->userplaneCentroid ) << endl;

    if( featuresLatest->headValid )
       _print_vector( os << "    HEAD: ", featuresLatest->userplaneHead ) << endl;
    else
       os << "    HEAD: not valid" << endl;

    if( featuresLatest->leftHandValid )
       _print_vector( os << "  L HAND: ", featuresLatest->userplaneLeftHand ) << endl;
    else
       os << "  L HAND: not valid" << endl;

    if( featuresLatest->rightHandValid )
       _print_vector( os << "  R HAND: ", featuresLatest->userplaneRightHand ) << endl;
    else
       os << "  R HAND: not valid" << endl;

    if( featuresLatest->leftFootValid )
       _print_vector( os << "  L FOOT: ", featuresLatest->userplaneLeftFoot ) << endl;
    else
       os << "  L FOOT: not valid" << endl;

    if( featuresLatest->rightFootValid )
       _print_vector( os << "  R FOOT: ", featuresLatest->userplaneRightFoot ) << endl;
    else
       os << "  R FOOT: not valid" << endl;

    os << endl;
    return os;
}

ostream&
pfFeatures::Print2d( ostream& os )
{
    _print_vector2( os << "CENTROID: ", featuresLatest->imageCentroid ) << endl;
    _print_vector2( os << "    HEAD: ", featuresLatest->imageHead ) << endl;
    _print_vector2( os << "  L HAND: ", featuresLatest->imageLeftHand ) << endl;
    _print_vector2( os << "  R HAND: ", featuresLatest->imageRightHand ) << endl;
    _print_vector2( os << "  L FOOT: ", featuresLatest->imageLeftFoot ) << endl;
    _print_vector2( os << "  R FOOT: ", featuresLatest->imageRightFoot ) << endl;
    os << endl;
    return os;
}

