// pfGesture.c++

#include "pfGesture.h++"
#include "QuickMath/QuickMath.h++"

float plMmtNoise = 1.;
float plAccelNoise = 1.;
int plGrain = 3; // steps per prediction

float pfGesture::bodyMoveXThreshold = 2.;
float pfGesture::postureCrouchThreshold = .9;
float pfGesture::postureSitThreshold = .8;
float pfGesture::handXSmall = .3;
float pfGesture::handXLarge = .38;
float pfGesture::handYSmall = -.35;
float pfGesture::handYLarge = .00;

static char* handGstrings[5] = {
    "Neutral", "Point Down", "Point Side", "Over Head", "Throwing" };

static char* posGstrings[3] = { "Standing", "Bent Over", "Sitting" };

static char *bodyGstrings[2] = { "Stationary", "Moving" };

pfGesture::pfGesture()
{
    Clear();
}

void
pfGesture::Clear()
{
    leftHandAction = handNeutral;
    rightHandAction = handNeutral;
    posture = posStanding;
    bodyAction = still;
    
    vector9Zero( headSave );
    vector9Zero( lhandSave );
    vector9Zero( rhandSave );
    vector9Zero( centSave );
}

void
pfGesture::Update( float* head,
		   float* lhand,
		   float* rhand,
		   float* cent,
		   float height )
// each input is a 9-vector, pos(3), vel(3), acc(3)
{
    semaphore.localLOCK();

    // (1) BODY ACTION
    //     -----------
    bodyAction = ( fabsf( cent[0]-centSave[0] ) > bodyMoveXThreshold ) ? moving : still;

    // (2) POSTURE
    //     -------
    float heightFraction = head[1] / height;
    posture = ( heightFraction < postureSitThreshold ) ? posSitting
       : ( heightFraction < postureCrouchThreshold ) ? posCrouched
       : posStanding;

    // (3) HANDS
    //     -----
    QuickVector headPos( 3, head );
    QuickVector leftPos( 3, lhand );
    QuickVector rightPos( 3, rhand );
    AllocVector xtend( 3 );

    xtend.Subtract( leftPos, headPos );
    xtend(0) = fabsf( xtend(0) );
    xtend /= height;
    leftHandAction = ( xtend(0) > handXSmall && xtend(1) < handYSmall ) ? handPointDown
       : ( xtend(0) > handXLarge && xtend(1) < handYLarge ) ? handPointSide
       : ( xtend(1) > handYLarge ) ? handOverHead
       : handNeutral;

    xtend.Subtract( rightPos, headPos );
    xtend(0) = fabsf( xtend(0) );
    xtend /= height;
    rightHandAction = ( xtend(0) > handXSmall && xtend(1) < handYSmall ) ? handPointDown
       : ( xtend(0) > handXLarge && xtend(1) < handYLarge ) ? handPointSide
       : ( xtend(1) > handYLarge ) ? handOverHead
       : handNeutral;

    semaphore.localUNLOCK();

    vector9Copy( head, headSave );
    vector9Copy( lhand, lhandSave );
    vector9Copy( rhand, rhandSave );
    vector9Copy( cent, centSave );
}

ostream&
pfGesture::Print( ostream& os )
{
    os << "    BODY: " << bodyGstrings[ bodyAction ] << endl;
    os << " POSTURE: " << posGstrings[ posture ] << endl;
    os << "  L HAND: " << handGstrings[ leftHandAction ] << endl;
    os << "  R HAND: " << handGstrings[ rightHandAction ] << endl;
    os << endl;
    return os;
}
