/* bsContour.c */

#include "ivePoint.h"
#include "iveBitmap.h"
#include "bsXclient.h"
#include "ivePoint.h"
typedef Point ivePoint;



static int pfMorph8Connected(iveBitmap *bm, int x, int y, int decimate)
{
  static int horizLookup[8] = {-1,  0,  1, 1, 1, 0, -1, -1};
  static int vertLookup[8]  = {-1, -1, -1, 0, 1, 1,  1,  0};

  register int i;
  register int width  =bm->width;
  register int height =bm->height;
  register int tempx;
  register int tempy;

  register int k = 0;

  for(i=0;i<8;i++)
    {
      tempx = x + decimate * horizLookup[i];
      tempy = y + decimate * vertLookup[i];
      if (!(tempx < 0 || tempx >= width || 
	    tempy < 0 || tempy >= height) &&
	  iveBitmapGetBit(bm,tempy,tempx))
	{
	  k++;
	}
    }
  return k;
}

static int pfMorph4Connected(iveBitmap *bm, int x, int y, int decimate)
{
  static int horizLookup[4] = { 0,  1,  0,  -1};
  static int vertLookup[4] =  {-1,  0,  1,   0};
  register int i;
  register int width  =bm->width;
  register int height =bm->height;

  register int k = 0;

  for(i=0;i<4;i++)
    {
      register int tempx;
      register int tempy;

      tempx = x + decimate * horizLookup[i];
      tempy = y + decimate * vertLookup[i];
      if (!(tempx < 0 || tempx >= width || 
	    tempy < 0 || tempy >= height) &&
	  iveBitmapGetBit(bm,tempy,tempx))
	k++;
    }

  return k;
}

static void pfMorphDilateSlow(iveBitmap *inMap,
	 int decimate,
	 int thresh)
{
  static iveBitmap *buffer = NULL;
  register int width  =inMap->width;
  register int height =inMap->height;
  register int i;
  register int j;

  /* still the same size? */
  if ((buffer == NULL) ||
      (width  != buffer->width) || 
      (height != buffer->height))
    {				/* nope: make a new buffer */
      if (buffer != NULL) iveBitmapFree(buffer);	
      buffer = iveBitmapMakeCopy(inMap);
    }
  else
    {				/* yup: just copy the data over */
      if (!iveBitmapCopy(inMap,buffer))
	fprintf(stderr,"fatal flaw in pfMorph: can't copy bitmaps!\n");
    }
  
  for(i=0;i<width;i+=decimate)
    for(j=0;j<height;j+=decimate)
      {
	if(!iveBitmapGetBit(buffer,j,i) &&
	   pfMorph4Connected(buffer,i,j,decimate) > thresh)
	  {
	    iveBitmapPutBit(inMap,j,i,10);
	  }
      }
}

static void pfMorphErodeSlow(iveBitmap *inMap, iveBitmap *origMap,
	 int decimate, 
         int thresh)
{
  static iveBitmap *buffer = NULL;
  register int width  =inMap->width;
  register int height =inMap->height;
  register int i;
  register int j;

  /* still the same size? */
  if ((buffer == NULL) ||
      (width  != buffer->width) || 
      (height != buffer->height))
    {				/* nope: make a new buffer */
      if (buffer != NULL) iveBitmapFree(buffer);	
      buffer = iveBitmapMakeCopy(inMap);
    }
  else
    {				/* yup: just copy the data over */
      if (!iveBitmapCopy(inMap,buffer))
	fprintf(stderr,"fatal flaw in pfMorph: can't copy bitmaps!\n");
    }
  
  for(i=0;i<width;i+=decimate)
    for(j=0;j<height;j+=decimate)
      if(!iveBitmapGetBit(origMap, j, i) &&
         iveBitmapGetBit(buffer,j,i) &&
	 pfMorph4Connected(buffer,i,j,decimate) < thresh)
	{
	  iveBitmapUnSetBit(inMap,j,i);
	}
}


/* this is the contour fcn from pfContour....
 eventually it should be merged into a single function
 which covers the same functionality via a callback or whatever.
 but for now this will do.
 also note that I think chris has a faster algorithm in mind
 that infers which pixels need to be set/cleared rather than
 explicitly calling morph4connected on them.
 --tjd
*/

void safeset(iveBitmap *b, int j, int i, int w, int h) {

     if (i>=0 && j >=0 && i < w && h < w) iveBitmapSetBit( b, j, i) ;
}

void safeput(iveBitmap *b, int j, int i, int w, int h) {

     if (i>=0 && j >=0 && i < w && h < w &&
	 !iveBitmapGetBit(b, j, i))
       iveBitmapPutBit( b, j, i, 10) ;
}

void safeunset(iveBitmap *b, int j, int i, int w, int h) {

     if (i>=0 && j >=0 && i < w && h < w) iveBitmapUnSetBit( b, j, i) ;
}

static
int
dilateABlob( iveBitmap *in,
	   iveBitmap *out,
	   iveBitmap *visited,
	   ivePoint *pcenter,
	   int decimate,
	   int thresh )
{
  static int horizLUT[8] = {-1, 0, 1, 1, 1, 0, -1, -1};
  static int vertLUT[8] = {-1, -1, -1, 0, 1, 1, 1, 0};
  ivePoint pstart;
  ivePoint pts[3];
  ivePoint* current = &pstart;
  ivePoint* next = pts;
  int next_i = 0;
  int length = 1;
  int dirToBGpoint_start = 5;
  int direction;
  int hit;
  int debug = 0;
  int h = in->height;
  int w = in->width;

  /* Find a low point on the contour */
  pstart.x = pcenter->x;
  for( pstart.y = pcenter->y;
      pstart.y < in->height-1 && iveBitmapGetBit( in, pstart.y+1, pstart.x );
      pstart.y++);


  while( ! (current->x==pstart.x && current->y==pstart.y) || 1==length )
    {
      if( iveBitmapGetBit( visited, current->y, current->x ) )
	return 0;
      iveBitmapSetBit( visited, current->y, current->x );

      if( 0 == thresh )
	{
	  safeput( out, current->y, current->x - 1, w, h);
	  safeput( out, current->y, current->x + 1 , w, h);
	  safeput( out, current->y - 1, current->x, w, h);
	  safeput( out, current->y + 1, current->x, w, h);
	}
      else
	fprintf( stderr, "unimplemented!\n" );

      direction = dirToBGpoint_start;
      do {
	if( debug )
	  {
	    fprintf( stderr, "direction= %d\n", direction );
	    fprintf( stderr, "dirToBGpoint_start= %d\n", dirToBGpoint_start );
	    fprintf( stderr, "current->x= %d\n", current->x );
	    fprintf( stderr, "current->y= %d\n", current->y );
	    fprintf( stderr, "next->x= %d\n", next->x );
	    fprintf( stderr, "next->y= %d\n", next->y );
	    fprintf( stderr, "in->width= %d\n", in->width );
	    fprintf( stderr, "in->height= %d\n", in->height );
	  }
	direction = (direction + 1) % 8;
	next->x = current->x + horizLUT[direction];
	next->y = current->y + vertLUT[direction];
	hit = (next->x>=0 && next->x<in->width && 
	       next->y>=0 && next->y<in->height );
	if( debug )
	  fprintf( stderr, "hit (1)= %d\n", hit );

	hit = hit && iveBitmapGetBit( in, next->y, next->x );
	if( debug )
	  fprintf( stderr, "hit (2)= %d\n", hit );
	if( direction == dirToBGpoint_start )
	  {
	    fprintf( stderr, "pfMorph: dilate ERROR x= %d, y= %d\n",
		    next->x, next->y );
	    debug = 1;
	    return 1;
	  }
      }
      while( !hit );
      dirToBGpoint_start = ( direction + 5 ) % 8;
      current = next;
      next_i = ( next_i + 1 ) % 3;
      next = pts + next_i;
      length++;
    }
  return 0;
}

static pfMorphDilateFast2( bm, N, pcenter, decimate,  thresh)
iveBitmap *bm;
int N;
ivePoint *pcenter;
int decimate;
int thresh;
{
  static iveBitmap *buffer = NULL;
  static iveBitmap *visited = NULL;
  register int width  =bm->width;
  register int height =bm->height;
  register int i;
  register int j;

  /* still the same size? */
  if ((buffer == NULL) ||
      (width  != buffer->width) || 
      (height != buffer->height))
    {				/* nope: make a new buffer */
      if (buffer != NULL) iveBitmapFree(buffer);	
      buffer = iveBitmapMakeCopy(bm);
    }
  else
    {				/* yup: just copy the data over */
      if (!iveBitmapCopy(bm,buffer))
	fprintf(stderr,"fatal flaw in pfMorph: can't copy bitmaps!\n");
    }

  /* still the same size? */
  if ((visited == NULL) ||
      (width  != visited->width) || 
      (height != visited->height))
    {				/* nope: make a new buffer */
      if (visited != NULL) iveBitmapFree(visited);	
      visited = iveBitmapMakeCopy(bm);
    }
  iveBitmapZero( visited );

  for( i = 0; i < N; i++ ) {
    int numc = pfMorph4Connected(buffer, pcenter[i].x, pcenter[i].y, decimate);
    if (numc) dilateABlob( buffer, bm, visited, pcenter+i, decimate, thresh );
  }
}

static
int
erodeABlob( iveBitmap *in,
	   iveBitmap *obm,
	   iveBitmap *out,
	   iveBitmap *visited,
	   ivePoint *pcenter,
	   int decimate,
	   int thresh )
{
  static int horizLUT[8] = {-1, 0, 1, 1, 1, 0, -1, -1};
  static int vertLUT[8] = {-1, -1, -1, 0, 1, 1, 1, 0};
  ivePoint pstart;
  ivePoint pts[3];
  ivePoint* current = &pstart;
  ivePoint* next = pts;
  int next_i = 0;
  int length = 1;
  int dirToBGpoint_start = 5;
  int direction;
  int hit;
  int h = in->height;
  int w = in->width;

  /* Find a low point on the contour */
  pstart.x = pcenter->x;
  for( pstart.y = pcenter->y;
      pstart.y < in->height-1 && iveBitmapGetBit( in, pstart.y+1, pstart.x );
      pstart.y++);

  while( ! (current->x==pstart.x && current->y==pstart.y) || 1==length )
    {
      if( iveBitmapGetBit( visited, current->y, current->x ) )
	return 0;
      safeset(visited, current->y, current->x, w, h);

      if(!iveBitmapGetBit(obm, current->y, current->x ) &&
	 pfMorph4Connected(in, current->x, current->y,decimate) < thresh)
	{
	  safeunset(out, current->y, current->x, w, h);
	}

      direction = dirToBGpoint_start;
      do {
	direction = (direction + 1) % 8;
	next->x = current->x + horizLUT[direction];
	next->y = current->y + vertLUT[direction];
	hit = (next->x>=0 && next->x<in->width && 
	       next->y>=0 && next->y<in->height );
	hit = hit && iveBitmapGetBit( in, next->y, next->x );
	if( direction == dirToBGpoint_start )
	  {
	    fprintf( stderr, "pfMorph: erode ERROR\n" );
	    return 1;
	  }
      }
      while( !hit );
      dirToBGpoint_start = ( direction + 5 ) % 8;
      current = next;
      next_i = ( next_i + 1 ) % 3;
      next = pts + next_i;
      length++;
    }
  return 0;
}

static
void
pfMorphErodeFast(iveBitmap *bm, iveBitmap *obm, int N, ivePoint* pcenter, int decimate,
		 int thresh )
{
  static iveBitmap *buffer = NULL;
  static iveBitmap *visited = NULL;
  register int width  =bm->width;
  register int height =bm->height;
  register int i;
  register int j;

  /* still the same size? */
  if ((buffer == NULL) ||
      (width  != buffer->width) || 
      (height != buffer->height))
    {				/* nope: make a new buffer */
      if (buffer != NULL) iveBitmapFree(buffer);	
      buffer = iveBitmapMakeCopy(bm);
    }
  else
    {				/* yup: just copy the data over */
      if (!iveBitmapCopy(bm,buffer))
	fprintf(stderr,"fatal flaw in pfMorph: can't copy bitmaps!\n");
    }

  /* still the same size? */
  if ((visited == NULL) ||
      (width  != visited->width) || 
      (height != visited->height))
    {				/* nope: make a new buffer */
      if (visited != NULL) iveBitmapFree(visited);	
      visited = iveBitmapMakeCopy(bm);
    }
  iveBitmapZero( visited );

  for( i = 0; i < N; i++ ) {
    int numc = pfMorph4Connected(buffer, pcenter[i].x, pcenter[i].y, decimate);
    if (numc) erodeABlob( buffer, obm, bm, visited, pcenter+i, decimate, thresh );
  }
  return;
}


extern erodemultiple;

void pfMorphBridgeGaps(iveBitmap *inMap,
		       Point *seeds,
		       int numseeds,
		       int decimate,
		       int dilation_factor)
{
  static iveBitmap *buffer = NULL;
  register int width  =inMap->width;
  register int height =inMap->height;
  int i,k;


  if ((buffer == NULL) ||
      (width  != buffer->width) || 
      (height != buffer->height))
    {				/* nope: make a new buffer */
      if (buffer != NULL) iveBitmapFree(buffer);	
      buffer = iveBitmapMakeCopy(inMap);
    }
  else
    {				/* yup: just copy the data over */
      if (!iveBitmapCopy(inMap,buffer))
	fprintf(stderr,"fatal flaw in pfMorph: can't copy bitmaps!\n");
    }
  
    for (k=0; k<dilation_factor; k++) 
      pfMorphDilateFast2(inMap, numseeds, seeds, decimate, 0);

    for (k=0; k<dilation_factor*erodemultiple; k++) 
      pfMorphErodeFast( inMap, buffer, numseeds, seeds, decimate, 3);
} 
  

#ifdef TEST_MORPH

main()
{
  Point seed;
  iveBitmap *bm = iveBitmapMake(120,160);
  iveBitmap *orig = NULL;

  iveBitmapReadImage( bm, stdin ); /* get image from cin */
  
  orig = iveBitmapMakeCopy(bm);

  seed.x = 0;			/* bogus seed */
  seed.y = 0;

  pfMorphBridgeGaps(bm,&seed,1,1);  /* try it out */

  {
    int i,j;
    
    for(i=0;i<bm->width;i+=1)
      for(j=0;j<bm->height;j+=1)
	{
	  if(iveBitmapGetBit(orig,j,i) &&
	     !iveBitmapGetBit(bm,j,i))
	    {
	      iveBitmapPutBit(bm,j,i,11);
	    }
	}

  }
  
  iveBitmapWriteImage(bm,stdout); /* write it to sdtout */
}

#endif
