/* pfinder-main.c++ */

/* RCS: $Header: /mas/vision/projects/pfinder/RCS/pfinder-main.c++,v 6.13 1995/10/03 02:31:03 ajazar Exp $ */

#include <stdio.h>
#include <stdlib.h>
#define _BSD_SIGNALS
#include <signal.h>
#include <unistd.h>
#include <iostream.h>
#include <bstring.h>
#include <math.h>

#include "Parse/Parse.h"
#include "VLImage.h++"
#include "pfGesture.h++"
#include "pfContour.h++"

#define Time MYTIME
#include "lite.h++"
#undef Time

extern "C"
{
#include "ivePoint.h"
#include "iveBitmap.h"
#include "iveImage.h"
#include "bsXclient.h"
#include "bsColor.h"
#include "bsClassify.h"
#include "bsSegment.h"
#include "pfMorphology.h"
}

static int saveMode = 0;	// save output to savestream
static int readMode = 0;	// read input from stdin
static int pauseMode = 0;	// loop on current image
static int dumpMode = 0;	// dump input to dumpstream
static int statMode = 0;	// dump class stats to statstream
static int contourMode = 0;	// dump contour to contourstream
static int expert = 0;		// show all the gory sliders
static int dumpCount = 0;
static int dumpDelay = 0;
static char *chroma = NULL;
static char *dumpFile = NULL;
static char *statFile = NULL;
static char *saveFile = NULL;
static char *contourFile = NULL;
static FILE *dumpstream = stdout;
static FILE *savestream = stdout;
static FILE *statstream = stdout;
static FILE *contourstream = stdout;
static int decimate = 1;
static int useALIVEInterface = 1;
static int useXInterface = 1;
static int verbose = 0;
static int maxForegroundClasses = 8;
static int numberOfBGFrames = 40;
static iveImage **BackgroundFrames = NULL;
static iveImage **imageBuffers = NULL;
Point lowerLeft;
Point lowerRight;

extern "C" {
Point seed_points_for_morphology[1024];
int erodemultiple = 10;
}

int THAD_TRACKER = 0; /* track green and red blobs for thad */
int INITCLASSFLAG = 0;

static int MORPHOLOGY = 1;
static int CONTOUR=1;
extern int FOREGROUND;  // should be in bsClassify.h (or something)
static int HOWSLOW = 1;
static int FRAMEZERO = 0;
static int SHOWSEEDS=0;

int SEGMENT=1; /* continous */
int SEGMENT_TRIGGER=0; /* one shot */
static int EVILBITMAPS=1;

void cleanup( void );
void cleanup_sig( ... );

extern void startRpcServer(void);
extern void stopRpcServer(void);
void call_cached_pf_reqests(void);



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * howslow
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static const int NFRAMES = 25;

static void howslow() 
{
  static int total = -1;
  static int count = 0;
  static struct timeval start;
  static struct timeval end;
  static double rate;
  static double time_sec;

  total++;
  // count = number of calls since init, *including* this one
  // Thusly, when count == NFRAMES, the time elapsed is for NFRAMES
  //  calls to processFrame()

  if (count==0)  // only happens once
     {
	gettimeofday(&start);
	count = 1;
	return;
     } 
  else if ( count++ == NFRAMES ) 
     {
	gettimeofday( &end );
	time_sec = (end.tv_usec - start.tv_usec)/1000000.0 + 
	   (end.tv_sec - start.tv_sec);
	rate = (double) NFRAMES / time_sec; // frames per second
	// fprintf ( stderr, "[%2.2f Hz]\n", rate );
	fprintf ( stderr, "[%2.2f Hz, (%d frames, %d total, %.6f sec) %d]\n", 
		  rate, NFRAMES, (int)(30.*time_sec), time_sec, total );
	start = end;
	count = 1;
	return;
     } 
  // putc('.',stderr);fflush(stderr);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * image flow control
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int STEP = 1;

void bsMainSetPause(int newvalue) {
  pauseMode = newvalue;
}

int bsMainGetPause() {
  return pauseMode;
}

void bsMainStep()
{
  STEP =1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * clickframe
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int CLICKME=0;

void
bsMainDumpNextFrame()
{
  CLICKME =1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * processFrame
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static int frameNumber = -1;

int bsMainGetFrameNumber()
{
  return frameNumber;
}

void
processFrame( iveImage *image)
{
  /* does not need frameNumber as an arg; can keep track internally */

  int j, nums;
  static Point centroid;
  static iveBitmap *newMap = (iveBitmap *)NULL;
  static iveBitmap *oldMap = (iveBitmap *)NULL;
  iveBitmap *tempMap = NULL;
#ifdef DONOTHING
  howslow(); return;
  /* NOTREACHED */
#endif
  frameNumber++;

  if (FRAMEZERO) {
    frameNumber = 0;
    FRAMEZERO = 0;
  }  

  if( newMap == (iveBitmap *)NULL )
    {
      int height = image->height;
      int width = image->width;
      
      newMap = iveBitmapMake( height, width );
      oldMap = iveBitmapMake( height, width );
      iveBitmapZero( oldMap );
      centroid.x = 0;
      centroid.y = 0;	
    }
  
  if( dumpMode )
    {
      if( dumpCount>0 && frameNumber >= dumpCount )
	{
	  if( imageBuffers )
	    {
	      for( int i = 0; i < dumpCount; i++ )
		iveImageWriteRaw( imageBuffers[i], dumpstream );
	    }
	  cleanup();
	}
      if( verbose )
	cerr << "pfinder: dumping frame " << frameNumber << endl;
      if( useXInterface )
	{
	  bsXDisplayImage( image, 1);
	}
      if( imageBuffers )
	iveImageCopy( image, imageBuffers[frameNumber] );
      else
	iveImageWriteRaw( image, dumpstream );

      if( statMode )
	{	
	  bsClassifyPrintClasses(image->width,image->height,statstream);
	}		

      if( contourMode )
	{	
	  lite::dumpContour(contourstream);
	}		
  
      if ( saveMode ) 
	{
	  iveBitmapWriteImage( newMap, savestream );
	}
      return;
    }
  

  if (frameNumber < numberOfBGFrames) {
    
    BackgroundFrames[frameNumber] = iveImageMakeCopy(image);
    
    if( useXInterface ) {
      bsXDisplayImage( image , 1);
    }
    
    if( statMode )
      {	
	bsClassifyPrintClasses(image->width,image->height,statstream);
      }		
    
    if( contourMode )
      {	
        lite::dumpContour(contourstream);
      }		

    if ( saveMode ) 
      {
	iveBitmapWriteImage( newMap, savestream );
      }
    return;
  }

  if (frameNumber == numberOfBGFrames) 
    {
      if (!chroma) 
	{
	  bsClassifyMakeBackgroundClasses(BackgroundFrames, numberOfBGFrames);
	} else {
	  bsClassifyChromakey(chroma);
	}
    }
  
  /* classify */

  if( HOWSLOW ) howslow();
  int dosegment = SEGMENT || SEGMENT_TRIGGER;
  SEGMENT_TRIGGER = 0;
  if (dosegment)     
    {
      int RASTERPAINT = 1;
      
      if( useXInterface && EVILBITMAPS && SHOWSEEDS ) 
	{
	  bsXNewCanvas( 1 );  
	  bsClassifyDisplayClasses(1);
	}

      iveBitmapZero( newMap );
      bsSegmentFindSilhouette( image, newMap, oldMap, 
			      decimate, &(centroid));

      if (FOREGROUND) bsClassifySetForegroundStats(image, 
						   newMap);
    
      
      bsClassifyUpdateBG( image, newMap );
	
      if( useXInterface && EVILBITMAPS ) 
	{
	  if (!SHOWSEEDS) bsXNewCanvas( 1 );  
	  
	  if (!RASTERPAINT)
	    bsXCommitPaint( 1 );

	}
      
  
      if (MORPHOLOGY) {
	      for (j = 0; seed_points_for_morphology[j].x && seed_points_for_morphology[j].y ; j++) ;
	      nums = j;
	      pfMorphBridgeGaps(newMap, seed_points_for_morphology,nums , decimate, MORPHOLOGY);
      }
      if( useXInterface && EVILBITMAPS ) 
	{
	  if (RASTERPAINT)
	    bsXDisplayBitmap( newMap , 1);
	  else
	    bsXCommitPaint( 1 ); 
	  
	}
    } else {
      if( useXInterface && EVILBITMAPS)
	bsXDisplayImage(image, 1);
    }
  
  if (dosegment && CONTOUR) {
      static pfContour contour;
      contour.generate( newMap, &centroid, decimate );

      lowerLeft.x = contour.bBoxMin[0];
      lowerLeft.y = contour.bBoxMax[1];
      lowerRight.x = contour.bBoxMax[0];
      lowerRight.y = contour.bBoxMax[1];
      lite::interpretPfContour( contour );

      if( useXInterface ) {
	  bsXDisplayContour( contour.points, contour.length, 0 );
	  bsXSync();	
      }
      if( useALIVEInterface ) {
	  /* sync requests are only replied to when
	     segment and contour are on */
	  call_cached_pf_reqests();
      }
  }

  if( statMode )
    {	
      bsClassifyPrintClasses(image->width,image->height,statstream);
    }		
  
  if( contourMode )
    {	
      lite::dumpContour(contourstream);
    }		

  if ( saveMode ) 
    {
      iveBitmapWriteImage( newMap, savestream );
    }

  if ( CLICKME )
    {
      if (verbose)
	cerr << "dumping this frame's data" << endl; 

      CLICKME = 0;
      if (statFile != (char *) NULL)
	bsClassifyPrintClasses(image->width,image->height, statstream);
      if (contourFile != (char *) NULL)
	lite::dumpContour(contourstream);
      if (dumpFile != (char *) NULL)
	iveImageWriteRaw( image, dumpstream );
      if (saveFile != (char *) NULL)
	iveBitmapWriteImage( newMap, savestream );
    }

  tempMap = oldMap;
  oldMap = newMap;
  newMap = tempMap;
}

 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * cleanup
 * cleanup_sig
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

void cleanup_do( void ) {
      
      if (useALIVEInterface) {
	    cerr << getpid() << " stopping rpc server" << endl;
	    stopRpcServer();
      }

      if ( useXInterface ) {
	    cerr << getpid() << " cleaning up X" << endl;
	    bsXClose();
      }

      if (!readMode) {
	    cerr << getpid() << " cleaning up VL" << endl;
	    VLImage::Close();
      }
}

void cleanup( void )
{
  cleanup_do();

  exit(1);
}

void cleanup_sig( ... )
{
  cerr << "Caught Signal" << getpid()  << endl;       
  cleanup();
}

static void 	FrameZero() { FRAMEZERO = 1; }
static int  bsMainGetContour() { return CONTOUR; }
static void bsMainSetContour(int newvalue) { CONTOUR = newvalue; }
static float  bsMainGetMorphology() { return (float) MORPHOLOGY; }
static void bsMainSetMorphology(float newvalue) { MORPHOLOGY = (int) newvalue; printf("morph %d\n", MORPHOLOGY);}
static int  bsMainGetEvilbitmaps() { return EVILBITMAPS; }
static void bsMainSetEvilbitmaps(int newvalue) { EVILBITMAPS = newvalue; }
static int  bsMainGetSegment() { return !SEGMENT; }
static void bsMainSetSegment(int newvalue) { SEGMENT = !newvalue; }

void verboseCallback( void * )
{
  bsColorSetVerbose( 1 );
  bsClassifySetVerbose( 1 );
  VLImage::Verbose( 1 );
}

void thadModeCallback( void * )
{
  THAD_TRACKER = !THAD_TRACKER;
  if (THAD_TRACKER) maxForegroundClasses =3;
}

void chromaModeCallback( void * )
{
  bsSegmentSetThresh(10.0);
}

void dumpModeCallback( void * )
{
  dumpMode = 1;
  statMode = 0;
  readMode = 0;
  saveMode = 0;
  maxForegroundClasses =1;
}

void readModeCallback( void * )
{
  readMode = 1;
}

void saveModeCallback( void * )
{
  statMode = 0;
  dumpMode = 0;
  saveMode = 1;
}

void statModeCallback( void * )
{
  statMode = 1;
  dumpMode = 0;
  saveMode = 0;
}

void aliveCallback (void *)
{
  bsMainSetEvilbitmaps(0);
  maxForegroundClasses = 1;
  bsClassifySetUpdateBG(0.01);
  expert = 0;
}  

void expertCallback (void *)
{
  bsMainSetEvilbitmaps(1);
  maxForegroundClasses = 8;
  expert = 1;
}  


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * main
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


void
help( void *progname )
{
  Parse::VerboseUsage( 1 );
  Parse::Usage( (char *) progname );
  cleanup();
}

int 
main(int argc, char *argv[])
{
  int zoom = 4;

  int width = 640;
  int height = 480;
  int nbuffers = 3;
  int input_connector = VLImage::INDY_SVIDEO;
  static char *inputHelp = "\n"\
     "\t\t\t0: INDY_COMPOSITE\n"\
     "\t\t\t1: INDY_SVIDEO\n"\
     "\t\t\t0: GALILEO_SVIDEO_1\n"\
     "\t\t\t1: GALILEO_SVIDEO_2\n"\
     "\t\t\t2: GALILEO_SVIDEO_3\n"\
     "\t\t\t3: GALILEO_COMPOSITE_1\n"\
     "\t\t\t4: GALILEO_COMPOSITE_2\n"\
     "\t\t\t5: GALILEO_COMPOSITE_3\n"\
     "\t\t\t6: GALILEO_YRB_1\n"\
     "\t\t\t7: GALILEO_YRB_2\n"\
     "\t\t\t\t";

  ParseRuleTag( "alive", aliveCallback, 0, "Alive Demo" );
  ParseRuleTag( "expert", expertCallback, 0, "Expert Mode (anti-ALIVE)" );
  ParseRuleFlag( 'n', &useALIVEInterface, !useALIVEInterface,
		 "Toggle whether to use ALIVE interface" );
  ParseRuleFlag( 'v', &verbose, 1, "Main program verbose mode on" );
  ParseRuleFlag( 'v', verboseCallback, 0, "bsColor and VL modules verbose" );
  ParseRuleFlag( 'X', &useXInterface, !useXInterface, "Use X Interface" );
  ParseRuleTaggedArgInt( "b", &numberOfBGFrames, "Number of BG Frames" );
  ParseRuleTaggedArgInt( "d", &decimate, "Decimation factor" );
  ParseRuleTag( "D", dumpModeCallback, 0, 
		"Dump mode on, images from VL to cout" );
  ParseRuleTaggedArgInt( "D", &dumpCount, "Number of frames to dump" );
  ParseRuleTag( "T", thadModeCallback, 0, "Thad mode: track green and red blobs" );
  ParseRuleTag( "t", statModeCallback, 0, "Stat mode: vlob centers to cout");
  ParseRuleTag( "R", readModeCallback, 0, "Read mode on, images fm cin" );
  ParseRuleTag( "single", &pauseMode, !pauseMode, 
	       "loop on first captured image");
  ParseRuleTag( "seeds", &SHOWSEEDS, !SHOWSEEDS, 
	       "mark zoomgrow seeds on image");
  ParseRuleTag( "S", saveModeCallback, 0, "Save mode on, bitmaps to cout" );
  ParseRuleTaggedArgInt( "E", &erodemultiple,"erode multiple" );
  ParseRuleTaggedArgInt( "w", &width, "Width of images" );
  ParseRuleTaggedArgInt( "h", &height, "Height of images" );
  ParseRuleTaggedArgInt( "size", &height, "Height of images", "height" );
  ParseRuleTaggedArgInt( "size", &width, "Width of images", "width" );
  ParseRuleTaggedArgInt( "z", &zoom, "Zoom factor" );
  ParseRuleTaggedArgInt( "f", &maxForegroundClasses, 
			"Max number of foreground classes" );
  ParseRuleTaggedArgInt( "in", &input_connector, inputHelp, "conn" );
  ParseRuleTaggedArgInt( "nb", &nbuffers, "Number of VL buffers to use", "number" );

  // This group of args is for gesture testing and should be removed
  // eventually
  ParseRuleTaggedArgFloat( "mn", &lite::mmtNoise, "KF: Std dev of mmt noise", "stdev" );
  ParseRuleTaggedArgFloat( "an", &lite::accelNoise, "KF: Std dev of accel noise", "stdev" );

  ParseRuleTaggedArgFloat( "gs", &pfGesture::handXSmall, "x small" );
  ParseRuleTaggedArgFloat( "gs", &pfGesture::handYSmall, "y small" );
  ParseRuleTaggedArgFloat( "gl", &pfGesture::handXLarge, "x large" );
  ParseRuleTaggedArgFloat( "gl", &pfGesture::handYLarge, "y large" );
  ParseRuleTaggedArgFloat( "gp", &pfGesture::postureSitThreshold, "sit" );
  ParseRuleTaggedArgFloat( "gp", &pfGesture::postureCrouchThreshold, "bent" );
  ParseRuleTaggedArgFloat( "gm", &pfGesture::bodyMoveXThreshold, "move" );
  ParseRuleTag( "pg", &lite::printGestures, 1, "print gestures" );

  ParseRuleTaggedArgInt( "plrf", &lite::reportFreq, "freq of report" );
  ParseRuleTag( "plrm", &lite::reportMove, 1, "report moves" );

  ParseRuleTaggedArgInt( "hand", &handMin, "min hand" );
  ParseRuleTaggedArgInt( "hand", &handMax, "max hand" );

  ParseRuleTag( "noslow", &HOWSLOW, 0, "Turn off rate reporting" );
  ParseRuleTag( "help", help, argv[0], "Prints usage" );
  
  // rules for getting file names for dumping output to files other than stdout

  ParseRuleTaggedArg("dumpfile", &dumpFile, "File to dump input to");
  ParseRuleTaggedArg("statfile", &statFile, "File to save statistics to");
  ParseRuleTaggedArg("contourfile", &contourFile, "File ot save contours to");
  ParseRuleTaggedArg("savefile", &saveFile, "File to save segments to");

  ParseRuleTag( "chroma", chromaModeCallback, 0, 
		"use chromakey instead of aquiring background" );
  ParseRuleTaggedArg("chroma", &chroma, "chroma to key on %2x%2x%2x");

  if( Parse::ParseCmdLine( argc, argv ) )
    {
      Parse::VerboseUsage( 0 );
      Parse::Usage( argv[0] );
      Parse::VerboseParse();
      Parse::ParseCmdLine( argc, argv );
      cleanup();
    }
  
  signal(SIGINT, cleanup_sig);
  signal(SIGPIPE, cleanup_sig); 
  signal(SIGHUP, cleanup_sig);
  signal(SIGILL, cleanup_sig);

  if (chroma)
    {
      numberOfBGFrames = 0;
    }

  if (dumpFile)
    {
      if (verbose)
	cerr << "dumpFile was " << dumpFile << endl;
      if ((dumpstream = fopen(dumpFile, "w")) == (FILE *) NULL)
	{
	  cerr << "Error opening dump file" << endl;
	  cleanup();
	}
    }

  if (statFile)
    {
      if (verbose)
	cerr << "statFile was " << statFile << endl;
      if ((statstream = fopen(statFile, "w")) == (FILE *) NULL)
	{
	  cerr << "Error opening stat file" << endl;
	  cleanup();
	}
    }

  if (contourFile)
    {
      if (verbose)
	cerr << "contourFile was " << contourFile << endl;
      if ((contourstream = fopen(contourFile, "w")) == (FILE *) NULL)
	{
	  cerr << "Error opening contour file" << endl;
	  cleanup();
	}
    }


  if (saveFile)
    {
      if (verbose)
	cerr << "saveFile was " << saveFile << endl;
      if ((savestream = fopen(saveFile, "w")) == (FILE *) NULL)
	{
	  cerr << "Error opening save file" << endl;
	  cleanup();
	}
    }


  if( !readMode )
    {
      VLImage::NBuffers( nbuffers );
#ifdef DIGITIZEYUV
      VLImage::Encoding( VLImage::YUV );
#else
      VLImage::Encoding( VLImage::RGB );
#endif
      VLImage::Zoom( zoom );
      VLImage::Input( input_connector );
      VLImage::Init();
      height = VLImage::GetHeight();
      width = VLImage::GetWidth();
    }
  if( dumpMode )
    {
      if( verbose )
	cerr << "pfinder: VL frame size: height= " << height
	     << " width= " << width << endl;
      if( dumpMode && dumpCount <= 100 && dumpCount>0)
	{
	  imageBuffers = (iveImage **) malloc( dumpCount*sizeof(iveImage*) );
	  if( imageBuffers )
	    {
	      for( int i = 0; i < dumpCount; i++ )
		{
		  imageBuffers[i] = iveImageMake( height, width );
		  if( ! imageBuffers[i] )
		    {
		      fprintf( stderr, "pfinder: failed to malloc image %d\n", i );
		      exit( 1 );
		    }
		}
	    }
	}
    }
  
  lite::initialize();

  if( !dumpMode )
     {
       bsClassifyInitialize(width, height, maxForegroundClasses);
       
       BackgroundFrames = (iveImage **) malloc( sizeof(iveImage *) * 
					       numberOfBGFrames );
     }
  
  if( useXInterface )
    {
      bsXInitialize(width, height, 
		    maxForegroundClasses, 
		    0, NULL );
      
      if ( !dumpMode ) 
	{
	  bsXMakeSliderProc( bsSegmentSetThresh, "FG/BG Threshhold",
			    bsSegmentGetThreshMax(), 0.0, 
			    bsSegmentGetThresh() );
	  bsXMakeSliderProc( bsColorSetNormalFloor, "Normalize Floor",
			    1.0, 0.0, 
			    bsColorGetNormalFloor() );
	  if (expert)
	    {
	      bsXMakeSliderProc( bsClassifySetThresh, "Sub-Class Threshhold",
				bsClassifyGetThreshMax(), 0.0, 
				bsClassifyGetThresh() );

	      bsXMakeSliderProc( bsMainSetMorphology, "Dilation",
				10.0, 0.0, bsMainGetMorphology() );
	    }
	}
      
      if (expert)
	{
	  bsXMakeSliderProc( bsXSetThresh, "Display Contrast",
			    1.0, 0.0, bsXGetThresh() );
	}
      
      if ( !dumpMode ) 
	{
	  bsXMakeButtonProc(bsMainSetSegment, "Pass-Through",
			    bsMainGetSegment() );
	  if (expert)
	    {
	      bsXMakeButtonProc(bsMainSetContour, "Compute Contour",
				bsMainGetContour() );
	    }
	  bsXMakeButtonProc(bsMainSetEvilbitmaps, "Display Segmentation",
			    bsMainGetEvilbitmaps() );

	  if (expert)
	    {
	      bsXMakeSliderProc( pfFeatures::setHandMin, "Hand min",
				1., 0., pfFeatures::getHandMin() );
	      bsXMakeSliderProc( pfFeatures::setHandMax, "Hand max",
				1., 0., pfFeatures::getHandMax() );
	    }

	  if (expert)
	    {
	      bsXMakeSeparator();
	      bsXMakeButtonProc(bsSegmentSetZoomGrow, "Grow",
				bsSegmentGetZoomGrow() );
	      bsXMakeButtonProc(bsSegmentSetRecursive, "Zoom",
				bsSegmentGetRecursive() );
	      bsXMakeButtonProc(bsSegmentSetSmooth, "Smooth",
				bsSegmentGetSmooth() );
	    }
	  
	  if (expert)
	    {
	      bsXMakeSeparator();
	      bsXMakeSliderProc(bsClassifySetUpdateBG, "Update Percentage",
				1.0, 0.0,
				bsClassifyGetUpdateBG() );
	      bsXMakeSliderProc(bsColorSetUpdateBG, "Update Rate",
				1.0, 0.0,
				bsColorGetUpdateBG() );
	    }
	  
	  if (expert) 
	    {
	      bsXMakeSeparator();
	      bsXMakeButtonProc(bsColorSetNormalize, "Normalize Enable",
				bsColorGetNormalize() );
	      bsXMakeButtonProc(bsClassifySetNormalize, "Normalize everywhere",
				bsClassifyGetNormalize());
	    }
	  bsXMakeButtonProc(bsClassifySetForeground, "Use Foreground",
			    bsClassifyGetForeground() );
	  
	  bsXMakeSeparator();
	  if (expert)
	    {
	      bsXMakeButtonProc( bsMainSetPause, "Pause",
				bsMainGetPause());
	      bsXMakePushButton( bsMainStep, "Step" );
	      bsXMakeSeparator();
	      bsXMakePushButton( bsMainDumpNextFrame, "Dump next frame" );
	    }
	  bsXMakePushButton( FrameZero, "Reaquire BG" );
	  
	}
      
      bsXMakePushButton( cleanup, "panic: core dump" );
    }

  if (useALIVEInterface) startRpcServer();

  if( readMode )		// read from a file
    {
      iveImage *image = iveImageMake( height, width ); 
      
      for(;;)
	 {
	   if( useXInterface ) 
	     bsXHandleEvents();
	   
	   if (pauseMode && !STEP)
	     {
	       processFrame( image );
	     }
	   else
	     {
	       if (iveImageReadRaw( image, stdin ))
		 {
		   fprintf(stderr,"end of input\n");
		   cleanup();
		 }
	       processFrame( image );
	     }
	   STEP = 0;
	 }
    }
  else				// live or dumpMode
    {
      sleep( dumpDelay );

#ifdef DIGITIZEYUV
      VLYuvImage picture;
#else
      VLRgbImage picture;
#endif
      iveImage wrapper;
      for(;;)
	 {
	   if( useXInterface )
	      bsXHandleEvents();
      
	   if (pauseMode && !STEP)
	     {
	       processFrame( &wrapper );
	     }
	   else
	     {
	       picture.Grab();
	       
	       iveImageInit(&wrapper, 
			    picture.Height(), 
			    picture.Width(), 
			    picture );
	       processFrame( &wrapper );
	       
	       picture.Free();
	     }

	   STEP = 0;

	 }
    }      
}



