/*
 * pfinder.x
 *
 * T. Darrell 5/95 
 *
 */

const YES=1;
const NO=0;

struct pfopts {
	bool sync; /* synchronous? (reply after next full frame) */

			/* known bug: if a client crashes in the middle of a 
			   sync RPC call to pfinder, it may crash the server. 
			   use sync only if needed. */

	bool alive_coords; /* alive coords or world coords? */
	bool kalman; /* kalman filtered or raw values? */
	bool f2;
	bool f3;
	bool c2;
	bool c3;
	bool g;
	bool cc2d;		/* color classes: m and (L,Q) */
	bool cc3d;		/* projected into alive coords */
};

struct pfv2f {
	float x;
	float y;
};

struct pfv3f {
	float x;
	float y;
	float z;
};

struct pfcc2d_s {
  float cmean[3];		/* color (RGB) mean */
  float gmean[2];		/* geometric (screen) mean normalized */
  float gL[2];			/* l0 l1  (normalized) */
  pfv2f gS[2];			/* s0x s0y s1x s1y */
				/* A*li = li*[six siy] */
};

struct pfcc3d_s {
  float cmean[3];		/* same as pfcc2d projected into 3D */
  float gmean[3];		
  float gL[3];			/* room for 3d blobs... but they're */
  pfv3f gS[3];			/* jsut 2d projected onto a plane in 3d */
				/* for now */
};

const MAXCONTOURLEN = 5000;
const MAXCLASSES    = 50;

struct pfgest_s {
    bool extend_left;
    bool extend_right;
    bool point_left;
    bool point_right;
    bool hand_moving_left;
    bool hand_moving_right;

    bool kick_left;
    bool kick_right;

    bool throw_left;
    bool throw_right;

    bool standing;
    bool sitting;
    bool bending_over;
    bool user_exists;
};


struct pffeat2d_s
{
  pfv2f left_hand;
  pfv2f right_hand;
  pfv2f head;
  pfv2f centroid;

  bool left_hand_valid;
  bool right_hand_valid;
  bool head_valid;
    
  pfv2f gaze_pos;
};


struct pffeat3d_s
{
  pfv3f left_hand;
  pfv3f right_hand;
  pfv3f head;
  pfv3f centroid;

  bool left_hand_valid;
  bool right_hand_valid;
  bool head_valid;
    
  pfv3f gaze_pos;

  pfv3f normal_size;  /* 0 is third coordinate: width, height */
  bool normal_valid;
};

union pfcont2d switch (int opt) {
case YES:
	pfv2f points<MAXCONTOURLEN>;
case NO:
	void;
default:
	void;
};

union pfcont3d switch (int opt) {
case YES:
	pfv3f points<MAXCONTOURLEN>;
case NO:
	void;
default:
	void;
};

union pfcc2d switch (int opt) {
case YES:
	pfcc2d_s cc<MAXCLASSES>;
case NO:
	void;
default:
	void;
};

union pfcc3d switch (int opt) {
case YES:
	pfcc3d_s cc<MAXCLASSES>;
case NO:
	void;
default:
	void;
};

union pfgest switch (int opt) {
case YES:
	pfgest_s *bits;
case NO:
	void;
default:
	void;
};

union pffeat2d switch (int opt) {
case YES:
	pffeat2d_s *feats;
case NO:
	void;
default:
	void;
};

union pffeat3d switch (int opt) {
case YES:
	pffeat3d_s *feats;
case NO:
	void;
default:
	void;
};

struct pfperson {
  int n;
  
  pffeat2d f2;
  pfcont2d c2;
  
  pffeat3d f3;
  pfcont3d c3;
  
  pfgest g;
  
  pfcc2d cc2d;
  pfcc3d cc3d;
};

program PFINDERPROG {
	version PFINDERVERS {
		void pfinit( void ) = 11;
		void pftrigger( int ) = 10;
		pfperson pf( pfopts ) = 12;
		void pfexit( void ) = 13;

		void pfsetgaze( pfv3f ) = 14;

		pfv2f pfscrsize( void ) = 15;

		float pfuserheight( void ) = 16;
	} = 1;
} = 0x20000095;



